/** @file
 *
 * 	VRR exports.
 *
 * 	PDF, PS and EPS export routines and related defines.
 */

#ifndef __EXPORT_EXPORT_H__
#define __EXPORT_EXPORT_H__

#define MM2POINT (72/25.4)

struct obj_doc;
struct obj_tlo;

/**
 * Exports all graphical objects in one doc into the postscript.
 * If pwidth and pheight are 0, bounding box will be computed, else width and
 * pheight determine the bounding box.
 *
 * @a doc document to be exported
 * @a f FILE the doc will be exported into
 * @a pwidth width of the postscript page in internal units
 * @a pheight height of the postscript page in internal units
 * @a fit_to_page set to nonzero if objects should be scaled to fit the page
 * @a without_fonts set to nonzero if you want to rely on document manager
 * 		to download required fonts
 */
void export_ps(struct obj_doc *doc, FILE *f, real pwidth, real pheight, int fit_to_page, int without_fonts);

/**
 * Exports all graphical objects in one tlo into the encapsulated postscript.
 *
 * @a tlo tlo to be exported
 * @a f opened FILE the doc will be exported into
 * @a without_fonts set to nonzero if you want to rely on document manager
 * 		to download required fonts
 */
void export_eps(struct obj_tlo *tlo, FILE *f, int without_fonts);

/**
 * Exports all graphical objects in one doc to the pdf.
 *
 * @a doc document to be exported
 * @a f opened FILE the doc will be exported into
 * @a without_fonts set to nonzero if you don't wish to embed font files
 * @a compression compression level, 0 = no compress, 1 = fast compress, 2 = best compress
 */
void export_pdf(struct obj_doc* doc, FILE *f, int without_fonts, int compression);

/**
  * Exports all graphical objects in on tlo into the svg format
  * @a tlo tlo to be exported
  * @a f opened FILE the tlo will be exported into
  */
void export_svg(struct obj_tlo *tlo, const char *filename, int *result);

#endif
