#undef LOCAL_DEBUG

#include "lib/lib.h"
#include "font/internal.h"
#include "geomlib/geomlib.h"

/*
 * TODO:
 *
 * Domyslet system pouzivanych znakovych sad.
 *
 * Zjednodusit *_render a string_bbox.
 *
 * Vymyslet vhodnou cache na uchovavani vyrenderovanych bitmap.
 * Nejaka LRU s kompresenymi bitmapami...
 *
 * Dodelat kerning v string_render. Ale je to sporne, protoze treba
 * TT rasterizer ghostscriptu kerning neumi, takze....
 */

#undef EXPERIMENTAL_CACHE

static int load_charmap(FT_Face face, uns charmap)
{
	FT_Encoding ch;
	switch (charmap) {
	case FONT_CHARMAP_NONE:
		ch = 0;
		break;
	case FONT_CHARMAP_STD:
		/*
		 * This should be the standard encoding.
		 * Not all fonts include this charmap, but this is the most suitable choice.
		 */
		ch = FT_ENCODING_UNICODE;
		break;
	case FONT_CHARMAP_TEX:
		/*
		 * I'm not sure if this is really the right charmap for TeX fonts
		 * as nothing is documented, but everything seems to work fine.
		 */
		ch = FT_ENCODING_ADOBE_CUSTOM;
		break;
	case FONT_CHARMAP_UNICODE:
		ch = FT_ENCODING_UNICODE;
		break;
	case FONT_CHARMAP_LATIN1:
		ch = FT_ENCODING_ADOBE_LATIN_1;
		break;
	default:
		BUG();
	}

	return font_Select_Charmap(face, ch);
}

#ifdef EXPERIMENTAL_CACHE
static int load_glyph(struct font_ctl *ctl, uns char_code, FT_Glyph *glyph)
{
#if 0
	/*
	 * Freetype is still full of bugs and betacode.
	 * Maybe in some future version this piece will work...
	 */

	FTC_CMapDescRec cmdr;

	cmdr.face_id = (FTC_FaceID)ctl->fnt_id;
	if (ctl->charmap) {
		cmdr.type = FTC_CMAP_BY_ENCODING;
		switch (charmap) {
		case FONT_CHARMAP_STD:
			/*
			 * This should be the standard encoding.
			 * Not all fonts include this charmap, but this is the most suitable choice.
			 */
			cmdr.u.encoding = FT_ENCODING_UNICODE;
			break;
		case FONT_CHARMAP_TEX:
			/*
			 * I'm not sure if this is really the right charmap for TeX fonts
			 * as nothing is documented, but everything seems to work fine.
			 */
			cmdr.u.encoding = FT_ENCODING_ADOBE_CUSTOM;
			break;
		case FONT_CHARMAP_UNICODE:
			cmdr.u.encoding = FT_ENCODING_UNICODE;
			break;
		case FONT_CHARMAP_LATIN1:
			cmdr.u.encoding = FT_ENCODING_ADOBE_LATIN_1;
			break;
		default:
			BUG();
		}
		char_code = font_CMapCache_Lookup(cmap_cache, &cmdr, char_code);
	}
#endif
	FTC_ImageTypeRec fitr;

	/* FIXME */
	char_code = font_CMapCache_Lookup(font_cmap_cache, (FTC_FaceID)ctl->font_id, 0, char_code);

	bzero(&fitr, sizeof(fitr));
	fitr.face_id = (FTC_FaceID)ctl->font_id;
	fitr.height = fitr.width = ctl->font_size * 100;
	fitr.flags = FT_LOAD_DEFAULT;

	if (font_ImageCache_Lookup(font_img_cache, &fitr, char_code, glyph, NULL))
		return -1;
	return 0;
}
#endif

int font_get_glyph_bbox(struct font_ctl *ctl, uns char_code, struct geom_rectangle *bbox)
{
	FT_Face face;
	FT_Size size;
	FT_Glyph glyph;
	FT_BBox bbx;
	FTC_ScalerRec scaler;
	FT_Matrix matrix;
	FT_Vector delta;

	ASSERT(bbox);

	/* This size setting has no effect actually for bbox computation */
	scaler.face_id = (FTC_FaceID)ctl->font_id;
	scaler.width = scaler.height = 64 * ctl->font_size;
	scaler.pixel = 1;
	scaler.x_res = scaler.y_res = 1000;
	if (font_Manager_LookupSize(font_mgr, &scaler, &size))
		return -1;
	face = size->face;
	load_charmap(face, ctl->charmap);
	if (ctl->charmap ? font_Load_Char(face, char_code, FT_LOAD_DEFAULT) :
			font_Load_Glyph(face, char_code, FT_LOAD_DEFAULT))
		return -1;
	if (font_Get_Glyph(face->glyph, &glyph))
		return -1;
	matrix.xx = ctl->transform.coef[0][0] * 0x10000;
	matrix.xy = ctl->transform.coef[0][1] * 0x10000;
	matrix.yx = ctl->transform.coef[1][0] * 0x10000;
	matrix.yy = ctl->transform.coef[1][1] * 0x10000;
	delta.x = ctl->transform.coef[0][2] * 0x40;
	delta.y = ctl->transform.coef[1][2] * 0x40;
	if (font_Glyph_Transform(glyph, &matrix, &delta))
		return -1;

	font_Glyph_Get_CBox(glyph, FT_GLYPH_BBOX_PIXELS, &bbx);

	bbox->ll.x = bbx.xMin / (real)0x40;
	bbox->ll.y = bbx.yMin / (real)0x40;
	bbox->ur.x = bbx.xMax / (real)0x40;
	bbox->ur.y = bbx.yMax / (real)0x40;
	font_Done_Glyph(glyph);
	return 0;
}

/* Freetype bug workaround */
static void extract_move(struct geom_transform *t, struct geom_transform *mv)
{
	struct geom_point p;

	bzero(mv, sizeof(*mv));
	bzero(&p, sizeof(p));
	geom_point_transform(&p, t, &p);
	geom_transform_set_move(mv, -p.x, -p.y);
	geom_transform_merge(t, mv, t);
	geom_transform_set_move(mv, p.x, p.y);
}

int font_get_string_bbox(struct font_ctl *ctl, const char *str, struct geom_rectangle *bbox)
{
	FT_Face face;
	FT_Size size;
	FT_Glyph glyph;
	FT_BBox bbx;
	FTC_ScalerRec scaler;
	FT_Matrix matrix;
	FT_Vector delta, adv;
	real minx, miny, maxx, maxy;
	int do_trn = 0;

	ASSERT(bbox);
	scaler.face_id = (FTC_FaceID)ctl->font_id;
	scaler.width = scaler.height = 64 * ctl->font_size;
	scaler.pixel = 1;
	scaler.x_res = scaler.y_res = 1000;
	if (font_Manager_LookupSize(font_mgr, &scaler, &size))
		return -1;
	face = size->face;
	if (!(ctl->transform.flags & GEOM_TRANSFORM_IDENTITY)) {
		matrix.xx = ctl->transform.coef[0][0] * 0x10000;
		matrix.xy = ctl->transform.coef[0][1] * 0x10000;
		matrix.yx = ctl->transform.coef[1][0] * 0x10000;
		matrix.yy = ctl->transform.coef[1][1] * 0x10000;
		delta.x = ctl->transform.coef[0][2] * 0x40;
		delta.y = ctl->transform.coef[1][2] * 0x40;
		do_trn = 1;
	}
	load_charmap(face, ctl->charmap);
	minx = miny = 1/0.0;
	maxx = maxy = -1/0.0;
	bzero(&adv, sizeof(adv));
	while (*str) {
		DBGLN("character '%c'\n", *str);
		if (ctl->charmap ? font_Load_Char(face, *str, FT_LOAD_DEFAULT) : font_Load_Glyph(face, *str, FT_LOAD_DEFAULT))
			return -1;
		if (font_Get_Glyph(face->glyph, &glyph))
			return -1;
		if (do_trn)
			if (font_Glyph_Transform(glyph, &matrix, &delta)) {
				font_Done_Glyph(glyph);
				return -1;
			}
		font_Glyph_Get_CBox(glyph, FT_GLYPH_BBOX_PIXELS, &bbx);
		DBGLN("got bbox %f..%f, %f..%f\n", bbx.xMin/(real)0x40, bbx.xMax/(real)0x40,
			bbx.yMin/(real)0x40, bbx.yMax/(real)0x40);

		minx = MIN(minx, bbx.xMin /(real)0x40 + adv.x /(real)0x10000);
		maxx = MAX(maxx, bbx.xMax /(real)0x40 + adv.x /(real)0x10000);
		maxy = MAX(maxy, bbx.yMax /(real)0x40 + adv.y /(real)0x10000);
		miny = MIN(miny, bbx.yMin /(real)0x40 + adv.y /(real)0x10000);
		DBGLN("actual bbox %f..%f  %f..%f\n", minx, maxx, miny, maxy);

		adv.x += glyph->advance.x / 64;
		adv.y += glyph->advance.y / 64 ;
		DBGLN("advance vector: %f %f\n", adv.x / (real)0x10000, adv.y / (real)0x10000);

		font_Done_Glyph(glyph);
		str++;
	}

	bbox->ll.x = minx;
	bbox->ll.y = miny;
	bbox->ur.x = maxx;
	bbox->ur.y = maxy;
        if (bbox->ur.x < bbox->ll.x) {
                real tmp = bbox->ur.x;
                bbox->ur.x = bbox->ll.x;
                bbox->ll.x = tmp;
        }
        if (bbox->ur.y < bbox->ll.y) {
                real tmp = bbox->ur.y;
                bbox->ur.y = bbox->ll.y;
                bbox->ll.y = tmp;
        }
	return 0;
}

int font_render_glyph(struct font_ctl *ctl, uns char_code)
{
	FT_Face face;
	FT_Bitmap bmp;
	FT_Size size;
	FT_Glyph glyph;
	FT_BBox bbx;
	FT_Matrix matrix;
	FT_Vector delta;
	FT_BitmapGlyph gbmp;
	FTC_ScalerRec scaler;
	struct geom_transform move_trans, tr;

	ASSERT(ctl);
	scaler.face_id = (FTC_FaceID)ctl->font_id;
	scaler.width = scaler.height = ctl->font_size * 100;
	scaler.pixel = 1;
	scaler.x_res = scaler.y_res = 1000;
	if (font_Manager_LookupSize(font_mgr, &scaler, &size))
		return -1;
	face = size->face;
	load_charmap(face, ctl->charmap);

#ifdef EXPERIMENTAL_CACHE
	FT_Glyph cglyph;
	if (load_glyph(ctl, char_code, &cglyph))
		return -1;
	if (font_Glyph_Copy(cglyph, &glyph))
		return -1;
#else
	if (ctl->charmap ? font_Load_Char(face, char_code, FT_LOAD_DEFAULT) : font_Load_Glyph(face, char_code, FT_LOAD_DEFAULT))
		return -1;
	if (font_Get_Glyph(face->glyph, &glyph))
		return -1;
#endif

	struct geom_point p = { 0, 0};
	geom_point_transform(&p, &ctl->transform, &p);
	tr = ctl->transform;

	extract_move(&tr, &move_trans);

	/*
	 * Ugly libfreetype bug workaround.
	 * Instead of scaling glyphs by transformation, we multiply the font_size by 100.
	 */
	struct geom_transform tmp_tr;
	geom_transform_set_scale(&tmp_tr, 0.01, 0.01);
	geom_transform_merge(&tr, &tmp_tr, &tr);

	matrix.xx = tr.coef[0][0] * 0x10000;
	matrix.xy = tr.coef[0][1] * 0x10000;
	matrix.yx = tr.coef[1][0] * 0x10000;
	matrix.yy = tr.coef[1][1] * 0x10000;
	delta.x = tr.coef[0][2] * 0x40;
	delta.y = tr.coef[1][2] * 0x40;

	if (font_Glyph_Transform(glyph, &matrix, &delta))
		goto err;

	font_Glyph_Get_CBox(glyph, FT_GLYPH_BBOX_PIXELS, &bbx);

	uns x = bbx.xMax - bbx.xMin;
	uns y = bbx.yMax - bbx.yMin;
	if (ctl->bsize < x*y) {
		if (!ctl->bitmap || ctl->bsize == 0)
			ctl->bsize = 1;
		while (ctl->bsize < x*y)
			ctl->bsize *= 2;
		ctl->bitmap = xrealloc(ctl->bitmap, ctl->bsize);
	}
	if (font_Glyph_To_Bitmap(&glyph, FT_RENDER_MODE_NORMAL, NULL, 1))
		goto err;
	gbmp = (FT_BitmapGlyph) glyph;
	bmp = gbmp->bitmap;

	ctl->width = bmp.width;
	ctl->height = bmp.rows;

	DBG("gbmpleft=%d gbmptop=%d px=%f py=%f\n", gbmp->left, gbmp->top, p.x, p.y);
	ctl->xoff = p.x + gbmp->left;
	ctl->yoff = p.y + gbmp->top - ctl->height;

	int i, j;
	for (i=0; i < bmp.rows; i++)
		for (j=0; j < bmp.width; j++)
			ctl->bitmap[(bmp.rows-1-i)*bmp.width + j] = bmp.buffer[i*bmp.pitch + j];
	font_Done_Glyph(glyph);
	return 0;
err:
	font_Done_Glyph(glyph);
	return -1;
}

int font_render_string(struct font_ctl *ctl, const char *str)
{
	FT_Face face;
	FT_Bitmap bmp;
	FT_Size size;
	FT_Glyph glyph;
	FTC_ScalerRec scaler;
	FT_Matrix matrix;
	FT_Vector delta, adv;
	FT_BitmapGlyph gbmp;
	FT_BBox bbx;
	int penx, peny;
	int minx, maxx, miny, maxy;
	const char *s;
	struct geom_transform tr, move_trans;

	ASSERT(ctl);
	DBG("render_string font_id=%d '%s'\n", ctl->fnt_id, font_filename(ctl->fnt_id));

	scaler.face_id = (FTC_FaceID)ctl->font_id;
	scaler.width = scaler.height = ctl->font_size * 100;
	scaler.pixel = 1;
	if (font_Manager_LookupSize(font_mgr, &scaler, &size))
		return -1;
	face = size->face;
	load_charmap(face, ctl->charmap);
	DBG("names: '%s'  '%s'\n", face->family_name, face->style_name);

	struct geom_point p = { 0, 0};
	geom_point_transform(&p, &ctl->transform, &p);
	tr = ctl->transform;

	/*
	 * Ugly libfreetype bug workaround.
	 * Instead of scaling glyphs by transformation, we multiply the font_size by 100.
	 */
	struct geom_transform tmp_tr;
	geom_transform_set_scale(&tmp_tr, 0.01, 0.01);
	geom_transform_merge(&tr, &tmp_tr, &tr);

	extract_move(&tr, &move_trans);
	matrix.xx = tr.coef[0][0] * 0x10000;
	matrix.xy = tr.coef[0][1] * 0x10000;
	matrix.yx = tr.coef[1][0] * 0x10000;
	matrix.yy = tr.coef[1][1] * 0x10000;
	delta.x = tr.coef[0][2] * 0x40;
	delta.y = tr.coef[1][2] * 0x40;

	ctl->width = ctl->height = 0;
	ctl->xoff = ctl->yoff = 0;

	/*
	 * First phase: compute size, boundingbox & positions
	 */
	s = str;
	minx = miny = INT_MAX;
	maxx = maxy = INT_MIN;
	bzero(&adv, sizeof(adv));
	while (*s) {
		DBGLN("character '%c'\n", *s);
		if ((ctl->charmap ? font_Load_Char(face, *s, FT_LOAD_DEFAULT) : font_Load_Glyph(face, *s, FT_LOAD_DEFAULT))
				|| font_Get_Glyph(face->glyph, &glyph))
			return -1;
		if (font_Glyph_Transform(glyph, &matrix, &delta)) {
			font_Done_Glyph(glyph);
			return -1;
		}
		font_Glyph_Get_CBox(glyph, FT_GLYPH_BBOX_PIXELS, &bbx);
		DBGLN("got bbox %d..%d, %d..%d\n", (int)bbx.xMin, (int)bbx.xMax, (int)bbx.yMin, (int)bbx.yMax);

		if (s == str) {
			penx = -bbx.xMin;
			peny = bbx.yMax;
		}
		else {
			if (minx > bbx.xMin + (font_RoundFix(adv.x) >> 16))
				penx += minx - (bbx.xMin + (font_RoundFix(adv.x) >> 16));
			if (maxy < bbx.yMax + (font_RoundFix(adv.y) >> 16))
				peny += bbx.yMax + (font_RoundFix(adv.y) >> 16) - maxy;
		}
		minx = MIN(minx, bbx.xMin + (font_RoundFix(adv.x) >> 16));
		maxx = MAX(maxx, bbx.xMax + (font_RoundFix(adv.x) >> 16));
		maxy = MAX(maxy, bbx.yMax + (font_RoundFix(adv.y) >> 16));
		miny = MIN(miny, bbx.yMin + (font_RoundFix(adv.y) >> 16));
		DBGLN("actual bbox %d..%d  %d..%d\n", minx, maxx, miny, maxy);
		DBGLN("actual pen x=%d y=%d\n", penx, peny);

		adv.x += glyph->advance.x;
		adv.y += glyph->advance.y;
		DBGLN("advance vector: %d %d\n",(int) adv.x >> 16,(int) adv.y >> 16);

		font_Done_Glyph(glyph);
		s++;
	}
	ctl->width = maxx - minx;
	ctl->height = maxy - miny;

	ctl->xoff = p.x - penx;
	ctl->yoff = p.y + peny - ctl->height;

	if (ctl->bsize < (uns)ctl->width * ctl->height) {
		if (!ctl->bitmap || ctl->bsize == 0)
			ctl->bsize = 1;
		while (ctl->bsize < (uns)ctl->width * ctl->height)
			ctl->bsize *= 2;
		ctl->bitmap = xrealloc(ctl->bitmap, ctl->bsize);
	}
	DBGLN("expected bitmap size h=%d w=%d\n", ctl->height, ctl->width);

	/*
	 * Second phase: render all bitmaps
	 */
	bzero(ctl->bitmap, ctl->bsize);
	s = str;
	DBGLN("penx=%d peny=%d\n", penx, peny);
	penx <<= 16;
	peny <<= 16;
	while (*s) {
		DBGLN("rendering '%c'\n", *s);
		if ((ctl->charmap ? font_Load_Char(face, *s, FT_LOAD_DEFAULT) : font_Load_Glyph(face, *s, FT_LOAD_DEFAULT))
				|| font_Get_Glyph(face->glyph, &glyph))
			return -1;
		if (font_Glyph_Transform(glyph, &matrix, &delta))
			return -1;
		adv = glyph->advance;
		DBGLN("advance %d, %d\n", (int)adv.x >> 16, (int)adv.y >> 16);

		if (font_Glyph_To_Bitmap(&glyph, FT_RENDER_MODE_NORMAL, NULL, 1))
			return -1;
		gbmp = (FT_BitmapGlyph)glyph;
		bmp = gbmp->bitmap;
		DBGLN("left=%d top=%d\n", gbmp->left, gbmp->top);

		int i, j, x, y;
		y = (font_RoundFix(peny) >> 16) - gbmp->top;
		DBGLN("y=%d\n", y);
		for (i=0; i < bmp.rows; i++, y++) {
			x = (font_RoundFix(penx) >> 16) + gbmp->left;
			for (j=0; j < bmp.width; j++, x++) {
				if (y<0 || x<0 || x>=ctl->width || y>=ctl->height) {
					DBGLN("got outside of bitmap!!! y=%d, x=%d\n", y, x);
					BUG();
				}
				ctl->bitmap[(ctl->height-1 - y)*ctl->width + x] = MIN(ctl->bitmap[(ctl->height-1 - y)*ctl->width + x] + bmp.buffer[i*bmp.pitch + j], 255);
			}
		}
		penx += adv.x;
		peny -= adv.y;
		DBGLN("pen position %d %d\n", (int) font_RoundFix(penx) >> 16, (int) font_RoundFix(peny) >> 16);
		s++;
	}
	return 0;
}


#define F2R(f) ((f) / (real)0x40)

static struct geom_pathset *output_pathset;
static struct geom_path current_path;
static struct geom_point cursor_pos;

static int outl_moveto(FT_Vector *to, void *user UNUSED)
{
	DBG("moveto [%f, %f]\n", F2R(to->x), F2R(to->y));

	if (!geom_group_empty(&current_path)) {
		/*geom_path_close_epsilon_with_segment(&current_path); FIXME: docasne nepodporovana fce */
		geom_group_add_tail(output_pathset, &current_path);	// FIXME
		geom_group_free_all(&current_path);
	}
	cursor_pos.x = F2R(to->x);
	cursor_pos.y = F2R(to->y);

	return 0;
}

static int outl_lineto(FT_Vector *to, void *user UNUSED)
{
	struct geom_segment s;
	struct geom_point p;

	p.x = F2R(to->x);
	p.y = F2R(to->y);
	DBG("lineto [%f, %f]\n", p.x, p.y);

	GEOM_ITEM_INIT(&s, segment);
	geom_segment_set(&s, &cursor_pos, &p);
	geom_path_append_dup(&current_path, &s);
	geom_segment_destroy(&s);
	cursor_pos = p;

	return 0;
}

static int outl_conicto(FT_Vector *control, FT_Vector *to, void *user UNUSED)
{
	struct geom_bezier b;
	struct geom_point p1, p2;

	p1.x = F2R(control->x);
	p1.y = F2R(control->y);
	p2.x = F2R(to->x);
	p2.y = F2R(to->y);
	DBG("conicto [%f, %f]  [%f, %f]\n", p1.x, p1.y, p2.x, p2.y);

	GEOM_ITEM_INIT(&b, bezier);
	geom_bezier_set_quadratic_nonrational(&b, &cursor_pos, &p1, &p2);
	geom_path_append_dup(&current_path, &b);
	geom_bezier_destroy(&b);
	cursor_pos = p2;

	return 0;
}

static int outl_cubicto(FT_Vector *control1, FT_Vector *control2, FT_Vector *to, void *user UNUSED)
{
	struct geom_bezier b;
	struct geom_point p1, p2, p3;

	p1.x = F2R(control1->x);
	p1.y = F2R(control1->y);
	p2.x = F2R(control2->x);
	p2.y = F2R(control2->y);
	p3.x = F2R(to->x);
	p3.y = F2R(to->y);
	DBG("cubicto [%f, %f]  [%f, %f]  [%f, %f]\n", p1.x, p1.y, p2.x, p2.y, p3.x, p3.y);

	GEOM_ITEM_INIT(&b, bezier);
	geom_bezier_set_cubic_nonrational(&b, &cursor_pos, &p1, &p2, &p3);
	geom_path_append_dup(&current_path, &b);
	geom_bezier_destroy(&b);
	cursor_pos = p3;

	return 0;
}

int font_char_decompose(struct font_ctl *ctl, int char_code, struct geom_pathset *pathset)
{
	FT_Face face;
	FT_Size size;
	FTC_ScalerRec scaler;
	FT_Outline_Funcs outl_funcs;
	FT_OutlineGlyph outl_glyph;
	FT_Glyph glyph;
	FT_Matrix matrix;
	FT_Vector delta;

	scaler.face_id = (FTC_FaceID)ctl->font_id;
	scaler.width = scaler.height = 64 * ctl->font_size;
	scaler.pixel = 1;
	scaler.x_res = scaler.y_res = 1000;
	if (font_Manager_LookupSize(font_mgr, &scaler, &size))
		return -1;
	face = size->face;
	load_charmap(face, ctl->charmap);
	if (ctl->charmap ? font_Load_Char(face, char_code, FT_LOAD_DEFAULT) : font_Load_Glyph(face, char_code, FT_LOAD_DEFAULT))
		return -1;
	if (font_Get_Glyph(face->glyph, &glyph))
		return -1;

	matrix.xx = ctl->transform.coef[0][0] * 0x10000;
	matrix.xy = ctl->transform.coef[0][1] * 0x10000;
	matrix.yx = ctl->transform.coef[1][0] * 0x10000;
	matrix.yy = ctl->transform.coef[1][1] * 0x10000;
	delta.x = ctl->transform.coef[0][2] * 0x40;
	delta.y = ctl->transform.coef[1][2] * 0x40;
	if (font_Glyph_Transform(glyph, &matrix, &delta))
		return -1;

	if (glyph->format != FT_GLYPH_FORMAT_OUTLINE) {
		DBG("Glyph is not in outline format!\n");
		return -1;
	}
	outl_glyph = (FT_OutlineGlyph) glyph;

	bzero(&outl_funcs, sizeof(outl_funcs));
	outl_funcs.move_to = outl_moveto;
	outl_funcs.line_to = outl_lineto;
	outl_funcs.conic_to = outl_conicto;
	outl_funcs.cubic_to = outl_cubicto;

	output_pathset = pathset;
	GEOM_ITEM_CREATE(&current_path, path);
	if (font_Outline_Decompose(&outl_glyph->outline, &outl_funcs, NULL)) {
		DBG("Outline decomposition failed\n");
		return -1;
	}
	if (!geom_group_empty(&current_path)) {
		/* geom_path_close_epsilon_with_segment(&current_path); FIXME: docasne nepodporovano */
		geom_group_add_tail(output_pathset, &current_path);	// FIXME
	}
	font_Done_Glyph(glyph);
	geom_path_destroy(&current_path);

	return 0;
}
