/** @file
 *
 * 	GEOMLIB - Curves approximations.
 *
 */

#ifndef _VRR_GEOM_APPROX_H_
#define _VRR_GEOM_APPROX_H_

#include "geomlib/types.h"

#define GEOM_APPROX_TYPE                  0x00000003
#define GEOM_APPROX_SHAPE                 0x00000000
#define GEOM_APPROX_TIME                  0x00000001
#define GEOM_APPROX_LENGTH                0x00000002

#define GEOM_APPROX_LINEAR_NONRATIONAL    0x00000004
#define GEOM_APPROX_LINEAR_RATIONAL       0x00000008
#define GEOM_APPROX_QUADRATIC_NONRATIONAL 0x00000010
#define GEOM_APPROX_QUADRATIC_RATIONAL    0x00000020
#define GEOM_APPROX_CUBIC_NONRATIONAL     0x00000040
#define GEOM_APPROX_CUBIC_RATIONAL        0x00000080
#define GEOM_APPROX_POINT                 0x00000100
#define GEOM_APPROX_ARC                   0x00000200
#define GEOM_APPROX_ELLIPTIC_ARC          0x00000400

#define GEOM_APPROX_LINEAR     (GEOM_APPROX_LINEAR_NONRATIONAL | GEOM_APPROX_LINEAR_RATIONAL)
#define GEOM_APPROX_QUADRATIC  (GEOM_APPROX_QUADRATIC_NONRATIONAL | GEOM_APPROX_QUADRATIC_RATIONAL)
#define GEOM_APPROX_CUBIC      (GEOM_APPROX_CUBIC_NONRATIONAL | GEOM_APPROX_CUBIC_RATIONAL)

#define GEOM_APPROX_BEZIER     (GEOM_APPROX_LINEAR | GEOM_APPROX_QUADRATIC | GEOM_APPROX_CUBIC)



void geom_approx_init(struct geom_approx *approx);

void geom_approx_destroy(struct geom_approx *approx);

void geom_approx_clear(struct geom_approx *approx);

int geom_approx_add_bezier(struct geom_approx *approx, struct geom_bezier *bezier, uns flags, double len);

int geom_approx_bezier(struct geom_approx *approx, struct geom_bezier *bezier, uns flags);

int geom_approx_path(struct geom_approx *approx, const struct geom_path *path, uns flags);

#endif
