#define GEOM_SOURCE
#include "geomlib/bernstein.h"
#include "geomlib/vector.h"
#include "geomlib/bezier.h"


/* geom_bezier_control_box
 * Function finds bounding box of bezier control points. */

int geom_bezier_control_box(struct geom_bezier *b, struct geom_rectangle *bbox)
{
	ITEM_ASSERT(bezier, b);
	PASSERT(bbox);

	/* find control bounding box */
	bbox->ll.x = bbox->ur.x = b->pt[0].x;
	bbox->ll.y = bbox->ur.y = b->pt[0].y;
	for (int i = 1; i <= BEZIER_DEGREE(b); i++) {
		if (bbox->ll.x > b->pt[i].x)
			bbox->ll.x = b->pt[i].x;
		if (bbox->ur.x < b->pt[i].x)
			bbox->ur.x = b->pt[i].x;
		if (bbox->ll.y > b->pt[i].y)
			bbox->ll.y = b->pt[i].y;
		if (bbox->ur.y < b->pt[i].y)
			bbox->ur.y = b->pt[i].y;
	}

	return 0;
}


/* geom_bezier_bbox */

int geom_bezier_bbox(struct geom_bezier *bezier, struct geom_rectangle *bbox)
{
	FUNC;
	ITEM_ASSERT(bezier, bezier);
	PASSERT(bbox);

	if (!(INSTANCE_FLAGS(bezier) & GEOM_BEZIER_BBOX_VALID)) {
		bezier->bbox.ll.x = MIN(bezier->pt[0].x, bezier->pt[BEZIER_DEGREE(bezier)].x);
		bezier->bbox.ur.x = MAX(bezier->pt[0].x, bezier->pt[BEZIER_DEGREE(bezier)].x);
		bezier->bbox.ll.y = MIN(bezier->pt[0].y, bezier->pt[BEZIER_DEGREE(bezier)].y);
		bezier->bbox.ur.y = MAX(bezier->pt[0].y, bezier->pt[BEZIER_DEGREE(bezier)].y);
		if (BEZIER_DEGREE(bezier) > GEOM_BEZIER_LINEAR) {
			int i, n;
			real t[5];
			struct geom_point p;

			/* find points with vertical derivation */
			n = geom_bezier_direction_times(bezier, &geom_point_up, 0, t);

			/* update x bounding range */
			for (i = 0; i < n; i++) {
				geom_bezier_point_at_time(bezier, t[i], &p);
				if (bezier->bbox.ll.x > p.x)
					bezier->bbox.ll.x = p.x;
				else if (bezier->bbox.ur.x < p.x)
					bezier->bbox.ur.x = p.x;
			}

			/* find points with horizontal derivation */
			n = geom_bezier_direction_times(bezier, &geom_point_right, 0, t);

			/* update y bounding range */
			for (i = 0; i < n; i++) {
				geom_bezier_point_at_time(bezier, t[i], &p);
				if (bezier->bbox.ll.y > p.y)
					bezier->bbox.ll.y = p.y;
				else if (bezier->bbox.ur.y < p.y)
					bezier->bbox.ur.y = p.y;
			}
		}
		INSTANCE_FLAGS(bezier) |= GEOM_BEZIER_BBOX_VALID;
	}
	if (bbox)
		*bbox = bezier->bbox;

	ERET(0);
}


