#define GEOM_SOURCE
#include "geomlib/fpath.h"


#define GEOM_FPATH_INITIAL_SIZE  (2 * sizeof(struct geom_bezier))


/* geom_fpath_destroy */

void geom_fpath_destroy(struct geom_fpath *path)
{
	geom_fpath_clear(path);
	geom_item_expansion_release(path);
	geom_group_destroy(&path->path.group);
}


/* geom_fpath_copy */

void geom_fpath_copy(struct geom_fpath *path, struct geom_path *src)
{
	struct geom_curve *curve;

	INSTANCE_IS_ASSERT(fpath, path);
	ITEM_ASSERT(path, src);
	
	if (path->path.group.list.head.next)
		BASSERT(geom_group_empty(path));
	else
		geom_fpath_create(path);
	GEOM_GROUP_WALK(src, curve)
		geom_fpath_append_dup(path, curve);
	geom_fpath_finish(path);
}


/* geom_fpath_clear */

void geom_fpath_clear(struct geom_fpath *path)
{
	struct geom_fpath_block *block, *prev;
	
	INSTANCE_IS_ASSERT(fpath, path);
	
	geom_group_destroy_all(&path->path.group);
	block = path->block_last;
	while (block) {
		prev = block->prev;
		xfree(block);
		block = prev;
	}
	path->block_last = NULL;
	path->block_free = 0;
	path->total_size = 0;
}


/* geom_fpath_alloc */

void *geom_fpath_alloc(struct geom_fpath *path, uns size)
{
	void *result;
	struct geom_fpath_block *block;
	uns s;

	ITEM_ASSERT(fpath, path);
	BASSERT(size > 0);
	
	if (size > path->block_free) {
		if (path->block_last)
			s = path->total_size;
		else
			s = GEOM_FPATH_INITIAL_SIZE;
		s = MAX(s, size);
		block = xmalloc_zero(OFFSETOF(struct geom_fpath_block, start) + s);
		block->prev = path->block_last;
		path->block_last = block;
		path->block_ptr = &block->start;
		path->block_free = s;
		path->total_size += s;
	}
	result = path->block_ptr;
	path->block_ptr = path->block_ptr + size;
	path->block_free -= size;
	return result;
}


