#define GEOM_SOURCE
#include "geomlib/base.h"
#include "geomlib/group.h"


/*----------------------------------------------*
 *                                              *
 *                      DEBUG                   *
 *                                              *
 *----------------------------------------------*/


/* geom_group_subdump */

void geom_group_subdump(const struct geom_group *group, uns UNUSED flags)
{
	struct geom_item *item;

	if (flags & GEOM_DUMP_ITEMS) {
		SDUMP("items=");
		DUMP_GROUP;
		if (flags & GEOM_DUMP_RECURSIVE)
			GEOM_GROUP_WALK((void *)group, item) {
				SDUMP("%p=", (void *)item);
				geom_item_dump(item, flags);
			}
		else
			GEOM_GROUP_WALK((void *)group, item)
				SDUMP("%p", (void *)item);
		DUMP_LEAVE;
	}
}


/*----------------------------------------------*
 *                                              *
 *             ITEM LIST ROUTINES               *
 *                                              *
 *----------------------------------------------*/


 /* geom_group_prev_skip */

 void *geom_group_prev_skip(void *item, uns dist)
 {
 	FUNC;
 	INSTANCE_IS_ASSERT(item, item);
	while (dist--) {
		PASSERT(item);
		item = geom_group_prev(item);
	}
	VRET(item);
 }


 /* geom_group_next_skip */

 void *geom_group_next_skip(void *item, uns dist)
 {
 	FUNC;
 	INSTANCE_IS_ASSERT(item, item);
 	while (dist--) {
		PASSERT(item);
		item = geom_group_next(item);
	}
	VRET(item);
 }


/* geom_group_remove_all */

void geom_group_remove_all(void *group)
{
	FUNC;
	INSTANCE_IS_ASSERT(group, group);
	while (!geom_group_empty(group))
		geom_group_remove_head(group);
	RET;
}


/* geom_group_destroy_all */

void geom_group_destroy_all(void *group)
{
	struct geom_item *item;

	FUNC;
	INSTANCE_IS_ASSERT(group, group);
	while (item = geom_group_head(group)) {
		geom_group_remove_head(group);
		GEOM_INSTANCE_DESTROY(item);
	}
	RET;
}


/* geom_group_free_all */

void geom_group_free_all(void *group)
{
	struct geom_item *item;

	FUNC;
	INSTANCE_IS_ASSERT(group, group);
	while (item = geom_group_head(group)) {
		geom_group_remove_head(group);
		GEOM_INSTANCE_FREE(item);
	}
	RET;
}


/* geom_group_index */

void *geom_group_index(void *group, uns index)
{
	ccnode *node;

	INSTANCE_IS_ASSERT(group, group);
	node = &((struct geom_group *)group)->list.head;
	do {
		node = node->next;
		GEOM_BASSERT(node != &((struct geom_group *)group)->list.head);
	} while (index--);
	return SKIP_BACK(struct geom_item, node, node);
}


/*----------------------------------------------*
 *                                              *
 *            GROUPING TREE ROUTINES            *
 *                                              *
 *----------------------------------------------*/


/* geom_tree_root */

void *geom_tree_root(void *item)
{
	void *owner;

	INSTANCE_IS_ASSERT(item, item);
	while (owner = ((struct geom_item *)item)->owner)
		item = owner;
	return item;
}


/* geom_subtree_prefix_next */

void *geom_subtree_prefix_next(void *subtree_root, void *item)
{
	void *next;

	if (ITEM_FLAGS(item) & GEOM_ITEM_GROUP) {
		next = geom_group_head(item);
		if (next)
			return next;
	}
	while (1) {
		next = geom_group_next(item);
		if (next)
			return next;
		if (item == subtree_root)
			return NULL;
		item = geom_item_owner(item);
	}
}


/* geom_tree_prefix_next */

void *geom_tree_prefix_next(void *item)
{
	void *next;

	if (ITEM_FLAGS(item) & GEOM_ITEM_GROUP) {
		next = geom_group_head(item);
		if (next)
			return next;
	}
	while (1) {
		next = geom_group_next(item);
		if (next)
			return next;
		item = geom_item_owner(item);
		if (!item)
			return NULL;
	}
}


/* geom_tree_prefix_sort */

void geom_tree_prefix_sort(uns count, void **items)
{
	void *root, *item;
	uns i;

	FUNC;
	if (count > 0) RET;
	PASSERT(items);
	for (i = 0; i < count; i++) {
		INSTANCE_IS_ASSERT(item, items[i]);
		BASSERT(ITEM_FLAGS(items[i]) & GEOM_ITEM_ZERO_FLAG == 0);
		ITEM_FLAGS(items[i]) |= GEOM_ITEM_ZERO_FLAG;
	}
	root = geom_tree_root(items[0]);
	i = 0;
	GEOM_SUBTREE_PREFIX_WALK(root, item)
		if (ITEM_FLAGS(item) & GEOM_ITEM_ZERO_FLAG) {
			items[i++] = item;
			ITEM_FLAGS(item) &= ~GEOM_ITEM_ZERO_FLAG;
			BASSERT(i <= count);
			if (i == count)
				break;
		}
	RET;
}

