#define GEOM_SOURCE
#include "geomlib/fpath.h"
#include "geomlib/bezier.h"


/*----------------------------------------------*
 *                                              *
 *                      DEBUG                   *
 *                                              *
 *----------------------------------------------*/


/* geom_item_call_assert */

const void *geom_item_call_assert(GEOM_ASSERT_DEF, const void *item, struct geom_class *pclass)
{
	geom_instance_is_assert(GEOM_ASSERT_COPY, item, pclass);
	INSTANCE_VMT(item, item, assert)(item, GEOM_ASSERT_COPY);
	return item;
}


/* geom_item_assert */

const void *geom_item_assert(const void *item, GEOM_ASSERT_DEF)
{
	if (!TEST(geom_item_valid(item))) {
		msg(GEOM_ASSERT_FORMAT ": item %p is invalid\n", GEOM_ASSERT_COPY, item);
		geom_item_dump(item, GEOM_DUMP_ALL);
		die("\n");
	}
	return item;
}


/* geom_item_dump */

void geom_item_dump(const void *item, uns flags)
{
	DUMP_ENTER(1, "Dumping item %p: ", (void *)item);
	DUMP_PTR(item) {
		INSTANCE_IS_ASSERT(item, item);
		if (!geom_item_valid(item))
			SDUMP("invalid");
		if (flags & GEOM_DUMP_CLASS)
			SDUMP("class=%s", geom_instance_class_name(item));
		if (flags & GEOM_DUMP_OWNER)
			SDUMP("owner=%p", (void *)((struct geom_item *)item)->owner);
		INSTANCE_VMT(item, item, subdump)(item, flags);
	}
	DUMP_LEAVE;
}


/* geom_item_copy__default */

void geom_item_copy__default(void *item, void *src)
{
	INSTANCE_IS_ASSERT(item, item);
	ITEM_ASSERT(item, src);
	BASSERT(geom_instance_class(item) == geom_instance_class(src));

	memcpy(
		(byte *)item + sizeof(struct geom_item),
		(byte *)src + sizeof(struct geom_item),
		geom_instance_size(item) - sizeof(struct geom_item));
	GEOM_INSTANCE_FLAGS(item) = GEOM_INSTANCE_FLAGS(src);
	ITEM_FLAGS(item) = ITEM_FLAGS(src);
}


/* geom_item_copy__fail */

void geom_item_copy__fail(void *item, void UNUSED *src)
{
	die("Geomlib failed: Cannot copy to item %p of type `%s'.\n", item, geom_instance_class_name(item));
}


/* geom_item_duplicate */

void *geom_item_duplicate(void *item)
{
	void *dup;

	ITEM_ASSERT(item, item);

	dup = geom_instance_new(geom_instance_class(item));
	geom_item_copy(dup, item);
	return dup;
}


