#ifndef _VRR_GEOMLIB_NURBS_H_
#define _VRR_GEOMLIB_NURBS_H_


/**
 * @file geomlib/nurbs.h
 *
 * Geomlib - Nonuniform rational Bezier splines.
 */


#include "geomlib/vector.h"
#include "geomlib/curve.h"


#define geom_nurbs_VMT  geom_curve_VMT
#define geom_nurbs_INIT geom_curve_INIT

GEOM_CLASS_HEAD(nurbs, curve);


/*----------------------------------------------*
 *                                              *
 *              INITIALIZATION                  *
 *                                              *
 *----------------------------------------------*/


static inline void geom_nurbs_create(struct geom_nurbs *nurbs)
{
	geom_curve_create(&nurbs->curve);
	cclist_init(&nurbs->knots);
}


void geom_nurbs_destroy(struct geom_nurbs *nurbs);


static inline void geom_nurbs_after_change(struct geom_nurbs *nurbs)
{ geom_curve_after_change(&nurbs->curve); }

void geom_nurbs_copy(struct geom_nurbs *nurbs, struct geom_nurbs *src);

int geom_nurbs_from_bezier(struct geom_nurbs *nurbs, struct geom_bezier *bezier);


/*----------------------------------------------*
 *                                              *
 *               PARAMETRIZATIONS               *
 *                                              *
 *----------------------------------------------*/


real geom_nurbs_time_to_btime(struct geom_nurbs *nurbs, struct geom_expansion *expansion, real time);

real geom_nurbs_btime_to_time(struct geom_nurbs *nurbs, struct geom_expansion *expansion, real btime);


/*----------------------------------------------*
 *                                              *
 *                      EDIT                    *
 *                                              *
 *----------------------------------------------*/


int geom_nurbs_knot_insert(struct geom_nurbs *nurbs, real time);


/*----------------------------------------------*
 *                                              *
 *     CONVERSION TO GIVEN TYPES OF CURVES      *
 *                                              *
 *----------------------------------------------*/


int geom_nurbs_expansion_append(struct geom_nurbs *nurbs, struct geom_fpath *expansion);


#endif

