#ifndef _VRR_GEOMLIB_PATH_H_
#define _VRR_GEOMLIB_PATH_H_

/**
 * @file geomlib/path.h
 *
 * Geomlib - Continous paths from more curves.
 */

#include "geomlib/group.h"
#include "geomlib/curve.h"


/*----------------------------------------------*
 *                                              *
 *                  PATH CLASS                  *
 *                                              *
 *----------------------------------------------*/


#define geom_path_VMT  geom_group_VMT
#define geom_path_INIT geom_group_INIT \
	\
	.destroy = (void (*)(void *)) &geom_path_destroy, \
	.copy = (void (*)(void *, void *)) &geom_path_copy, \
	.after_change = (void (*)(void *)) &geom_path_after_change, \
	\
	.tlength = (real (*)(void *)) &geom_path_tlength, \
	.alength = (real (*)(void *)) &geom_path_alength, \
	\
	.time_to_atime = (real (*)(void *, real)) &geom_path_time_to_atime, \
	.times_to_atimes = (int (*)(void *, uns, real *, real *)) &geom_path_times_to_atimes, \
	.atime_to_time = (real (*)(void *, real)) &geom_path_atime_to_time, \
	.atimes_to_times = (int (*)(void *, uns, real *, real *)) &geom_path_atimes_to_times, \
	\
	.time_to_ratime = (real (*)(void *, real)) &geom_path_time_to_ratime, \
	.times_to_ratimes = (int (*)(void *, uns, real *, real *)) &geom_path_times_to_ratimes, \
	.ratime_to_time = (real (*)(void *, real)) &geom_path_ratime_to_time, \
	.ratimes_to_times = (int (*)(void *, uns, real *, real *)) &geom_path_ratimes_to_times, \
	\
	.expansion_append = (int (*)(void *, struct geom_fpath *)) &geom_path_expansion_append, \
	\
	.transform = (int (*)(void *, struct geom_transform *, void *)) &geom_path_transform, \
	\
	.point_start = (int (*)(void *, struct geom_point *)) &geom_path_point_start, \
	.point_end = (int (*)(void *, struct geom_point *)) &geom_path_point_end, \
	.point_at_time = (int (*)(void *, real, struct geom_point *)) &geom_path_point_at_time, \
	\
	.derivation_start = (int (*)(void *, struct geom_vector *)) &geom_path_derivation_start, \
	.derivation_end = (int (*)(void *, struct geom_vector *)) &geom_path_derivation_end, \
	.derivation_at_time = (int (*)(void *, real, struct geom_vector *)) &geom_path_derivation_at_time, \
	\
	.bbox = (int (*)(void *, struct geom_rectangle *)) &geom_path_bbox, \
	.nearest_to_point = (int (*)(void *, struct geom_point *, uns, struct geom_nearest *)) &geom_path_nearest_to_point, \

GEOM_CLASS_HEAD(path, group);


/*----------------------------------------------*
 *                                              *
 *                 INITIALIZATION               *
 *                                              *
 *----------------------------------------------*/



static inline void geom_path_create(struct geom_path *path)
{
	geom_group_create(&path->group);
	GEOM_ITEM_FLAGS(path) |= GEOM_ITEM_PATH;
}

static inline void geom_path_destroy(struct geom_path *path)
{
	geom_group_free_all(path);
	geom_item_expansion_release(path);
	geom_group_destroy(&path->group);
}

void geom_path_copy(struct geom_path *path, struct geom_path *src);

static inline void *geom_path_append_dup(struct geom_path *path, void *item)
{
	void *copy = geom_item_duplicate(item);
	geom_group_add_tail(path, copy);
	return copy;
}

static inline void geom_path_clear(struct geom_path *path)
{ geom_group_free_all(path); }


/*----------------------------------------------*
 *                                              *
 *          OVERRIDED VIRTUAL METHODS           *
 *                                              *
 *----------------------------------------------*/


/* Parametrizations */

static inline real geom_path_tlength(struct geom_path *path) { return geom_group_count(path); }
real geom_path_alength(struct geom_path *path);

real geom_path_time_to_atime(struct geom_path *path, real time);
int geom_path_times_to_atimes(struct geom_path *path, uns count, real *times, real *atimes);
real geom_path_atime_to_time(struct geom_path *path, real atime);
int geom_path_atimes_to_times(struct geom_path *path, uns count, real *atimes, real *times);


real geom_path_time_to_ratime(struct geom_path *path, real time);
int geom_path_times_to_ratimes(struct geom_path *path, uns count, real *times, real *ratimes);
real geom_path_ratime_to_time(struct geom_path *path, real ratime);
int geom_path_ratimes_to_times(struct geom_path *path, uns count, real *ratimes, real *times);


/* Expansions */

int geom_path_expansion_append(struct geom_path *path, struct geom_fpath *expansion);


/* Transformations */

int geom_path_transform(struct geom_path *path, struct geom_transform *transform, struct geom_path *result);


/* Geometric properties */

int geom_path_point_start(struct geom_path *path, struct geom_point *result);
int geom_path_point_end(struct geom_path *path, struct geom_point *result);
int geom_path_point_at_time(struct geom_path *path, real time, struct geom_point *result);

int geom_path_derivation_start(struct geom_path *path, struct geom_vector *result);
int geom_path_derivation_end(struct geom_path *path, struct geom_vector *result);
int geom_path_derivation_at_time(struct geom_path *path, real time, struct geom_vector *result);

int geom_path_bbox(struct geom_path *path, struct geom_rectangle *bbox);
int geom_path_nearest_to_point(struct geom_path *path, struct geom_point *point, uns flags, struct geom_nearest *nearest);
int geom_path_direction_times(struct geom_path *path, struct geom_vector *direction, struct garr *result);

int geom_paths_intersections(struct geom_path *path, struct geom_path *path2, uns flags, struct garr *result);



#endif
