#define GEOM_SOURCE
#include "geomlib/pathset.h"

#if 0 /* FIXME :-) */

// prvek v halde pri prochazeni roviny
struct geom_pathset_entry {
	enum geom_pathset_entry_type {
		GEOM_PATHSET_ENTRY_START,
		GEOM_PATHSET_ENTRY_END,
		GEOM_PATHSET_ENTRY_OPEN,
		GEOM_PATHSET_ENTRY_CLOSE,
		GEOM_PATHSET_ENTRY_INTERSECTION,
		GEOM_PATHSET_ENTRY_INTERNAL} type;
	uns index;
	real y;
	union {
		struct {
			struct geom_pathset_edge *edge;
			struct geom_pathset_entry *neighbour;
		} start;
		struct {
			struct geom_pathset_edge *edge;
		} end;
		struct {
			struct geom_pathset_edge *edge;
			struct geom_pathset_entry *neighbour;
		} internal;
		struct {
			struct geom_pathset_edge *edge1;
			struct geom_pathset_edge *edge2;
			struct geom_pathset_entry *neighbour1;
			struct geom_pathset_entry *neighbour2;
		} open;
		struct {
			struct geom_pathset_edge *edge1;
			struct geom_pathset_edge *edge2;
			uns ref_count;
		} close;
		struct {
			struct geom_pathset_sweep_trap *sweep_trap;
			struct geom_intersection geom;
		} intersection;
	} data;
};

// oblastina sweep line
struct geom_pathset_sweep_face {
	cnode sweep_node;
	clist poly_list;
	cclist sweep_trap_list;
};

// trapezoidy na sweep line
struct geom_pathset_sweep_trap {
	cnode sweep_node;
	ccnode sweep_face_node;
	struct geom_pathset_sweep_face *sweep_face;
	struct geom_pathset_half *left_half;
	struct geom_pathset_half *right_half;
	struct geom_pathset_entry *intersection;
	uns left_index;
	real left_min_time;
	real left_max_time;
};

// lokalni promenne
static struct {
	real y;
	struct geom_pathset *pathset;
	clist sweep_trap_list;
	clist sweep_face_list;
	uns heap_n;
	struct garr heap;
} geom_sweep;


#define HALF_SWEEP_TRAP(h) ((struct geom_pathset_sweep_trap *)((h)->trap))
#define SWEEP_TRAP_IS_FIRST(t) ((t)->left_half == NULL)
#define SWEEP_TRAP_IS_LAST(t) ((t)->right_half == NULL)
#define SWEEP_TRAP_PREV(t) (SKIP_BACK(struct geom_pathset_sweep_trap, sweep_node, (t)->sweep_node.prev))
#define SWEEP_TRAP_NEXT(t) (SKIP_BACK(struct geom_pathset_sweep_trap, sweep_node, (t)->sweep_node.next))


/* geom_pathset_half_to_edge */

static inline struct geom_pathset_edge *geom_pathset_half_to_edge(struct geom_pathset_half *half)
{
	if ((half->flags & GEOM_PATHSET_HALF_SIDE) == GEOM_PATHSET_HALF_SIDE_LEFT)
		return SKIP_BACK(struct geom_pathset_edge, left_half, half);
	else
		return SKIP_BACK(struct geom_pathset_edge, right_half, half);
}


/*-------------------------------------------------------*
 *                                                       *
 *                          DEBUG                        *
 *                                                       *
 *-------------------------------------------------------*/


/* geom_pathset_sweep_assert */

static void geom_pathset_sweep_assert(void)
{
	// assert list of trapezoids on sweep line
	{
		cnode *ptr;
		ccnode *ccptr;
		struct geom_pathset_sweep_trap *sweep_trap = NULL, *prev_sweep_trap;
		struct geom_pathset_sweep_face *sweep_face;
		uns sweep_face_count = 0;
		uns sweep_traps_count = 0;
		uns sweep_face_traps_count = 0;

		CLIST_WALK(ptr, geom_sweep.sweep_trap_list) {
			PASSERT(ptr);
			BASSERT(ptr->prev->next == ptr);
			sweep_traps_count++;
			prev_sweep_trap = sweep_trap;
			sweep_trap = SKIP_BACK(struct geom_pathset_sweep_trap, sweep_node, ptr);
			if (!clist_next(&geom_sweep.sweep_trap_list, ptr))
				BASSERT(sweep_trap->right_half == NULL);
			else {
				PASSERT(sweep_trap->right_half);
				BASSERT((sweep_trap->right_half->flags & GEOM_PATHSET_HALF_SIDE) == GEOM_PATHSET_HALF_SIDE_LEFT);
			}
			if (!clist_prev(&geom_sweep.sweep_trap_list, ptr))
				BASSERT(sweep_trap->left_half == NULL);
			else {
				PASSERT(sweep_trap->left_half);
				BASSERT((sweep_trap->left_half->flags & GEOM_PATHSET_HALF_SIDE) == GEOM_PATHSET_HALF_SIDE_RIGHT);
				BASSERT(
					geom_pathset_half_to_edge(sweep_trap->left_half) ==
					geom_pathset_half_to_edge(prev_sweep_trap->right_half));
			}
		}

		CLIST_WALK(ptr, geom_sweep.sweep_face_list) {
			PASSERT(ptr);
			BASSERT(ptr->prev->next == ptr);
			sweep_face_count++;
			sweep_face = SKIP_BACK(struct geom_pathset_sweep_face, sweep_node, ptr);

			CCLIST_WALK(ccptr, sweep_face->sweep_trap_list) {
				PASSERT(ccptr);
				BASSERT(ccptr->prev->next == ccptr);
				sweep_trap = SKIP_BACK(struct geom_pathset_sweep_trap, sweep_face_node, ccptr);
				BASSERT(sweep_trap->sweep_face == sweep_face);
				sweep_face_traps_count++;
			}
		}
		BASSERT(sweep_traps_count == sweep_face_traps_count);

	}

	// assert heap
	{
		uns i;
		struct geom_pathset_entry *entry;

		for (i = 1; i <= geom_sweep.heap_n; i++) {
			entry = *(struct geom_pathset_entry **)garr_get(&geom_sweep.heap, i - 1);
			switch (entry->type) {
			case GEOM_PATHSET_ENTRY_START:
				break;
			case GEOM_PATHSET_ENTRY_END:
				break;
			case GEOM_PATHSET_ENTRY_OPEN:
				break;
			case GEOM_PATHSET_ENTRY_CLOSE:
				break;
			case GEOM_PATHSET_ENTRY_INTERSECTION:
				break;
			case GEOM_PATHSET_ENTRY_INTERNAL:
				break;
			default:
				BASSERT(0);
			}
		}
	}
}


/*-------------------------------------------------------*
 *                                                       *
 *                   MEMORY MANAGEMENT                   *
 *                                                       *
 *-------------------------------------------------------*/

/* FIXME: az bude algoritmus funkcni, alokaci predelam */


/* geom_pathset_item_alloc & free */

static inline struct geom_pathset_item *geom_pathset_item_alloc(void)
{ return (struct geom_pathset_item *)xmalloc_zero(sizeof(struct geom_pathset_item)); }

static inline void geom_pathset_item_free(struct geom_pathset_item *item)
{ memset(item, 0, sizeof(struct geom_pathset_item)); xfree(item); }


/* geom_pathset_poly_alloc */

static inline struct geom_pathset_poly *geom_pathset_poly_alloc(void)
{ return (struct geom_pathset_poly *)xmalloc_zero(sizeof(struct geom_pathset_poly)); }

static inline void geom_pathset_poly_free(struct geom_pathset_poly *poly)
{ memset(poly, 0, sizeof(struct geom_pathset_poly)); xfree(poly); }


/* geom_pathset_vertex_alloc & free */

static inline struct geom_pathset_vertex *geom_pathset_vertex_alloc(void)
{ return (struct geom_pathset_vertex *)xmalloc_zero(sizeof(struct geom_pathset_vertex)); }

static inline void geom_pathset_vertex_free(struct geom_pathset_vertex *vertex)
{ memset(vertex, 0, sizeof(struct geom_pathset_vertex)); xfree(vertex); }


/* geom_pathset_edge_alloc & free */

static inline struct geom_pathset_edge *geom_pathset_edge_alloc(void)
{ return (struct geom_pathset_edge *)xmalloc_zero(sizeof(struct geom_pathset_edge)); }

static inline void geom_pathset_edge_free(struct geom_pathset_edge *edge)
{ memset(edge, 0, sizeof(struct geom_pathset_edge)); xfree(edge); }


/* geom_pathset_face_alloc & free */

static inline struct geom_pathset_face *geom_pathset_face_alloc(void)
{ return (struct geom_pathset_face *)xmalloc_zero(sizeof(struct geom_pathset_face)); }

static inline void geom_pathset_face_free(struct geom_pathset_face *face)
{ memset(face, 0, sizeof(struct geom_pathset_face)); xfree(face); }


/* geom_pathset_sweep_face_alloc & free */

static inline struct geom_pathset_sweep_face *geom_pathset_sweep_face_alloc(void)
{ return (struct geom_pathset_sweep_face *)xmalloc_zero(sizeof(struct geom_pathset_sweep_face)); }

static inline void geom_pathset_sweep_face_free(struct geom_pathset_sweep_face *sweep_face)
{ memset(sweep_face, 0, sizeof(struct geom_pathset_sweep_face)); xfree(sweep_face); }


/* geom_pathset_sweep_trap_alloc & free */

static inline struct geom_pathset_sweep_trap *geom_pathset_sweep_trap_alloc(void)
{ return (struct geom_pathset_sweep_trap *)xmalloc_zero(sizeof(struct geom_pathset_sweep_trap)); }

static inline void geom_pathset_sweep_trap_free(struct geom_pathset_sweep_trap *sweep_trap)
{ memset(sweep_trap, 0, sizeof(struct geom_pathset_sweep_trap)); xfree(sweep_trap); }


/* geom_pathset_entry_alloc & free */

static inline struct geom_pathset_entry *geom_pathset_entry_alloc(void)
{ return (struct geom_pathset_entry *)xmalloc_zero(sizeof(struct geom_pathset_entry)); }

static inline void geom_pathset_entry_free(struct geom_pathset_entry *entry)
{ memset(entry, 0, sizeof(struct geom_pathset_entry)); xfree(entry); }


/*-------------------------------------------------------*
 *                                                       *
 *                    HEAP OF ENTRIES                    *
 *                                                       *
 *-------------------------------------------------------*/


/* geom_pathset_heap_insert */

static void geom_pathset_heap_insert(struct geom_pathset_entry *entry)
{
	uns i, j, n, index;
	real y;
	struct geom_pathset_entry **heap;

	garr_get(&geom_sweep.heap, i = n = geom_sweep.heap_n++);
	heap = (struct geom_pathset_entry **)garr_get(&geom_sweep.heap, 0);
	index = entry->index;
	y = entry->y;
	GASSERT(r,y);
	while (i) {
		j = (i + 1) / 2 - 1;
		if (heap[j]->y <= y)
			break;
		(heap[i] = heap[j])->index = i;
		i = j;
	}
	(heap[i] = entry)->index = i;
}


/* geom_pathset_heap_remove */

static void geom_pathset_heap_remove(struct geom_pathset_entry *entry)
{
	uns i, j, n;
	real y;
	struct geom_pathset_entry **heap, *last;

	heap = (struct geom_pathset_entry **)garr_get(&geom_sweep.heap, 0);
	last = heap[n = --geom_sweep.heap_n];
	y = last->y;
	i = entry->index;
	while (1) {
		j = i * 2 + 1;
		if (j >= n)
			break;
		if (j + 1 < n && heap[j]->y > heap[j + 1]->y)
			j++;
		if (y <= heap[j]->y)
			break;
		(heap[i] = heap[j])->index = i;
		i = j;
	}
	(heap[i] = last)->index = i;
}


/* geom_pathset_heap_increase */

static inline void geom_pathset_heap_increase(struct geom_pathset_entry *entry)
{
	geom_pathset_heap_remove(entry);
	geom_pathset_heap_insert(entry);
}


/* geom_pathset_entry_new_start */

static inline struct geom_pathset_entry *geom_pathset_entry_new_start(struct geom_pathset_edge *edge)
{
	struct geom_pathset_entry *entry;

	entry = geom_pathset_entry_alloc();
	entry->type = GEOM_PATHSET_ENTRY_START;
	entry->y = edge->bottom_vertex->point.point.y;
	entry->data.start.edge = edge;
	geom_pathset_heap_insert(entry);
	return entry;
}


/* geom_pathset_entry_new_end */

static inline struct geom_pathset_entry *geom_pathset_entry_new_end(struct geom_pathset_edge *edge)
{
	struct geom_pathset_entry *entry;

	entry = geom_pathset_entry_alloc();
	entry->type = GEOM_PATHSET_ENTRY_END;
	entry->y = edge->top_vertex->point.point.y;
	entry->data.end.edge = edge;
	return entry;
}


/* geom_pathset_entry_new_open */

static inline struct geom_pathset_entry *geom_pathset_entry_new_open(struct geom_pathset_edge *edge1)
{
	struct geom_pathset_entry *entry;

	entry = geom_pathset_entry_alloc();
	entry->type = GEOM_PATHSET_ENTRY_OPEN;
	entry->y = edge1->bottom_vertex->point.point.y;
	entry->data.open.edge1 = edge1;
	geom_pathset_heap_insert(entry);
	return entry;
}


/* geom_pathset_entry_new_close */

static inline struct geom_pathset_entry *geom_pathset_entry_new_close(struct geom_pathset_edge *edge1)
{
	struct geom_pathset_entry *entry;

	entry = geom_pathset_entry_alloc();
	entry->type = GEOM_PATHSET_ENTRY_CLOSE;
	entry->y = edge1->top_vertex->point.point.y;
	entry->data.close.edge1 = edge1;
	entry->data.close.ref_count = 2;
	return entry;
}


/*-------------------------------------------------------*
 *                                                       *
 *                       VRCHOLY                         *
 *                                                       *
 *-------------------------------------------------------*/


/* geom_pathset_vertex_new
 * - vytvori novy vrchol a prida ho do seznamu v pathsetu */

static inline struct geom_pathset_vertex *geom_pathset_vertex_new(struct geom_point *point, real time)
{
	struct geom_pathset_vertex *vertex;

	GEOM_POINT_ASSERT(point);
	GASSERT(r,time);

	vertex = geom_pathset_vertex_alloc();
	vertex->point.point = *point;
	vertex->point.time = time;
	return vertex;
}


/*-------------------------------------------------------*
 *                                                       *
 *                 HRANY A "POLOHRANY"                   *
 *                                                       *
 *-------------------------------------------------------*/


/* geom_pathset_edge_split
 * - rozdeli hranu v danem bode */

static void geom_pathset_edge_split(struct geom_pathset_edge *top_edge, struct geom_pathset_vertex *vertex)
{
	struct geom_pathset_edge *bottom_edge;

	bottom_edge = geom_pathset_edge_alloc();
	*bottom_edge = *top_edge;
	if (top_edge->left_half.cw_half == &top_edge->right_half) {
		bottom_edge->left_half.cw_half = &bottom_edge->right_half;
		bottom_edge->right_half.ccw_half = &bottom_edge->left_half;
	}
	else {
		top_edge->left_half.cw_half->ccw_half = &bottom_edge->left_half;
		top_edge->right_half.ccw_half->cw_half = &bottom_edge->right_half;
		bottom_edge->left_half.cw_half = top_edge->left_half.cw_half;
		bottom_edge->right_half.ccw_half = top_edge->right_half.ccw_half;
	}
	bottom_edge->left_half.ccw_half = &top_edge->left_half;
	bottom_edge->right_half.cw_half = &top_edge->right_half;
	top_edge->left_half.cw_half = &bottom_edge->left_half;
	top_edge->right_half.ccw_half = &bottom_edge->right_half;
	top_edge->bottom_vertex = bottom_edge->top_vertex = vertex;
	if ((top_edge->left_half.flags & GEOM_PATHSET_EDGE_DIR) == GEOM_PATHSET_EDGE_DIR_INC) {
		bottom_edge->left_half.flags = top_edge->left_half.flags & ~GEOM_PATHSET_EDGE_END;
		bottom_edge->right_half.flags = top_edge->right_half.flags & ~GEOM_PATHSET_EDGE_END;
		top_edge->left_half.flags &= ~GEOM_PATHSET_EDGE_START;
		top_edge->right_half.flags &= ~GEOM_PATHSET_EDGE_START;
	}
	else {
		bottom_edge->left_half.flags = top_edge->left_half.flags & ~GEOM_PATHSET_EDGE_START;
		bottom_edge->right_half.flags = top_edge->right_half.flags & ~GEOM_PATHSET_EDGE_START;
		top_edge->left_half.flags &= ~GEOM_PATHSET_EDGE_END;
		top_edge->right_half.flags &= ~GEOM_PATHSET_EDGE_END;
	}
}


/* geom_pathset_edge_new
 * - vytvori novou hranu */

static struct geom_pathset_edge *geom_pathset_edge_new(struct geom_pathset_item *item, struct geom_pathset_vertex *start_vertex, struct geom_pathset_vertex *end_vertex, uns flags)
{
	struct geom_pathset_edge *edge;

	edge = geom_pathset_edge_alloc();
	edge->item = item;
	edge->left_half.flags = flags | GEOM_PATHSET_HALF_SIDE_LEFT;
	edge->right_half.flags = flags | GEOM_PATHSET_HALF_SIDE_RIGHT;
	if ((flags & GEOM_PATHSET_EDGE_DIR) == GEOM_PATHSET_EDGE_DIR_INC) {
		edge->bottom_vertex = start_vertex;
		edge->top_vertex = end_vertex;
	}
	else {
		edge->bottom_vertex = end_vertex;
		edge->top_vertex = start_vertex;
	}
	return edge;
}


/* geom_pathset_half_join
 * - propoji 2 sousedni polohrany */

static inline void geom_pathset_half_join(struct geom_pathset_half *ccw_half, struct geom_pathset_half *cw_half)
{
	cw_half->ccw_half = ccw_half;
	ccw_half->cw_half = cw_half;
}

/* geom_pathset_item_init_edges
 * - danou cestu rozlozi na monotonni hrany
 * - FIXME: podpora cyklickych cest */

static void geom_pathset_item_init_edges(struct geom_pathset_item *item)
{
	struct geom_path *path;
	struct geom_pathset_vertex *start_vertex, *vertex, *vertex2;
	struct geom_pathset_edge *edge;
	struct geom_pathset_entry *entry, *entry_prev;
	struct garr garr;
	uns index, flags, dir = 0, i, m;
	struct geom_path_item *path_item;
	struct geom_point point, prev_point;
	real time = 0, prev_time;

	FUNC;

	path = item->path;
	if (path->count) {
		flags = (uns)(-1);
		if (!ETEST(geom_path_point_start(path, &point)))
			point = geom_point_zero;
		vertex = start_vertex = geom_pathset_vertex_new(&point, 0);
		garr_init(&garr, 16, sizeof(real));

		entry_prev = NULL;
		GEOM_PATH_WALK_ITEMS_ALL(path, index, path_item) {
			m = geom_curve_horizontal_times(path_item->flags & GEOM_CURVE_TYPE, path_item->ptr, &garr);
			m = MAX((int)m, 0);
			*(real *)garr_get(&garr, m) = 1;
			for (i = 0; i <= m; i++) {
				prev_point = point;
				prev_time = time;
				time = *(real *)garr_get(&garr, i);
				if (!ETEST(geom_curve_point_at_time(path_item->flags & GEOM_CURVE_TYPE, path_item->ptr, time, &point)))
					point = prev_point;
				dir = (point.y >= prev_point.y) ? GEOM_PATHSET_EDGE_DIR_INC : GEOM_PATHSET_EDGE_DIR_DEC;
				if ((int)flags < 0)
					flags = GEOM_PATHSET_EDGE_START | dir;
				else if (dir != (flags & GEOM_PATHSET_EDGE_DIR)) {
					vertex2 = vertex;
					vertex = geom_pathset_vertex_new(&prev_point, inf2finr(prev_time + index));
					edge = geom_pathset_edge_new(item, vertex2, vertex, flags);
					if (dir == GEOM_PATHSET_EDGE_DIR_INC) {
						entry = geom_pathset_entry_new_open(edge);
						if (!entry_prev)
							entry_prev = geom_pathset_entry_new_end(edge);
						else
							entry_prev->data.close.edge2 = edge;
						entry->data.open.neighbour1 = entry_prev;
					}
					else {
						entry = geom_pathset_entry_new_close(edge);
						if (!entry_prev)
							geom_pathset_entry_new_start(edge)->data.start.neighbour = entry;
						else {
							entry_prev->data.open.neighbour2 = entry;
							entry_prev->data.open.edge2 = edge;
						}
					}
					entry_prev = entry;
					flags = dir;
				}
			}
		}
		vertex2 = geom_pathset_vertex_new(&point, path->count);
		edge = geom_pathset_edge_new(item, vertex, vertex2, flags | GEOM_PATHSET_EDGE_END);
		if (dir == GEOM_PATHSET_EDGE_DIR_DEC) {
			entry = geom_pathset_entry_new_start(edge);
			if (!entry_prev)
				entry_prev = geom_pathset_entry_new_end(edge);
			else
				entry_prev->data.close.edge2 = edge;
			entry->data.start.neighbour = entry_prev;
		}
		else {
			entry = geom_pathset_entry_new_end(edge);
			if (!entry_prev)
				geom_pathset_entry_new_start(edge)->data.start.neighbour = entry;
			else {
				entry_prev->data.open.neighbour2 = entry;
				entry_prev->data.open.edge2 = edge;
			}
		}
		garr_cleanup(&garr);
	}
	RET;
}


/* geom_pathset_poly_new
 * - vytvori strukturu pro cyklus hran a prida ji do prislusne steny */

static inline struct geom_pathset_poly *geom_pathset_poly_new(struct geom_pathset_half *half)
{
	struct geom_pathset_poly *poly;

	poly = geom_pathset_poly_alloc();
	poly->first_half = half;
	clist_add_tail(&HALF_SWEEP_TRAP(half)->sweep_face->poly_list, &poly->face_node);
	return poly;
}


/* geom_pathset_edge_search */

static void geom_pathset_edge_search(struct geom_pathset_edge *edge, struct geom_point *point, struct geom_curve_point *result)
{
	struct geom_path *path;
	real min_time, max_time, t;

	GEOM_POINT_ASSERT(point);
	path = edge->item->path;
	if ((edge->left_half.flags & GEOM_PATHSET_EDGE_DIR) == GEOM_PATHSET_EDGE_DIR_INC) {
		min_time = edge->bottom_vertex->point.time;
		max_time = edge->top_vertex->point.time;
	}
	else {
		min_time = edge->top_vertex->point.time;
		max_time = edge->bottom_vertex->point.time;
	} // FIXME: dotazy jsou setridene, jde zrychlit
	if (!GEOM_ETEST(geom_path_solve_y(path, min_time, max_time, point->y, GEOM_CURVE_POINT_ALL, result))) {
		TRACE("Using Y search hack!");
		t = (point->y - edge->bottom_vertex->point.point.y) * edge->top_vertex->point.point.y - edge->bottom_vertex->point.point.y;
		if (!isfinite(t)) t = 0;
		t = fminmaxr(t, 0, 1);
		result->time = fcoefrng(edge->bottom_vertex->point.time, edge->top_vertex->point.time, t);
		if (!GEOM_ETEST(geom_path_point_at_time(path, result->time, &result->point)))
			result->point = *point;
		else
			result->point.y = point->y;
	}
	GEOM_POINT_ASSERT(&result->point);
	GASSERT(r,result->time);
}


/* geom_pathset_half_intersection
 * - z left_half a right_half spocita souradnice pruseciku
 * - vrati, zda byl nalezen prusecik */

static int geom_pathset_half_intersection(struct geom_pathset_half *left_half, struct geom_pathset_half *right_half, struct geom_intersection *intersection, real epsilon UNUSED)
{
	// FIXME: HACK! zbytecne pomale a nestabilni
	//
	struct geom_pathset_edge *left_edge, *right_edge;
	struct geom_path *left_path, *right_path;
	struct garr garr;
	struct geom_intersection *inter;
	int i, n, found;
	real left_min_time, left_max_time, right_min_time, right_max_time, t;

	left_edge = geom_pathset_half_to_edge(left_half);
	right_edge = geom_pathset_half_to_edge(right_half);
	if ((left_edge->left_half.flags & GEOM_PATHSET_EDGE_DIR) == GEOM_PATHSET_EDGE_DIR_INC) {
		left_min_time = left_edge->bottom_vertex->point.time;
		left_max_time = left_edge->top_vertex->point.time;
	}
	else {
		left_min_time = left_edge->top_vertex->point.time;
		left_max_time = left_edge->bottom_vertex->point.time;
	}
	if ((right_edge->left_half.flags & GEOM_PATHSET_EDGE_DIR) == GEOM_PATHSET_EDGE_DIR_INC) {
		right_min_time = right_edge->bottom_vertex->point.time;
		right_max_time = right_edge->top_vertex->point.time;
	}
	else {
		right_min_time = right_edge->top_vertex->point.time;
		right_max_time = right_edge->bottom_vertex->point.time;
	}
	left_path = left_edge->item->path;
	right_path = right_edge->item->path;

	garr_init(&garr, 16, sizeof(struct geom_intersection));
	if (left_path < right_path)
		n = geom_paths_intersections(left_path, right_path, GEOM_INTERSECTION_ALL, &garr);
	else
		n = geom_paths_intersections(right_path, left_path, GEOM_INTERSECTION_ALL, &garr);

	found = 0;
	for (i = 0; i < n; i++) {
		inter = (struct geom_intersection *)garr_get(&garr, i);
		if (left_path > right_path) {
			t = inter->time1; inter->time1 = inter->time2; inter->time2 = t;
		}
		if (left_path == right_path && IS_EQUAL_L(inter->time1, inter->time2))
			continue;
		if (left_path == right_path && ((int)inter->time1) == ((int)inter->time2))
			continue;
		if (inter->time1 >= left_min_time && inter->time1 <= left_max_time &&
			inter->time2 >= right_min_time && inter->time2 <= right_max_time &&
			inter->point.y > geom_sweep.y) {
			if (!found || (inter->point.y < intersection->point.y)) {
				*intersection = *inter;
				found = 1;
			}
		}
	}
	garr_cleanup(&garr);
	if (found) {
		GEOM_POINT_ASSERT(&intersection->point);
		GASSERT(r,intersection->time1);
		GASSERT(r,intersection->time2);
	}
	return found;
}


/* geom_pathset_edge_order
 * - zjisti, ktera ze dvou hran zacinajicich ve stejnem bode smeruje vic doleva a ktera vic doprava
 * - vrati 0, pokud je prvni hrana vlevo a 1, pokud je druha hrana vlevo */

static int geom_pathset_edge_order(struct geom_pathset_edge *edge1, struct geom_pathset_edge *edge2, struct geom_pathset_sweep_trap *left_trap UNUSED, struct geom_pathset_sweep_trap *right_trap UNUSED)
{
	struct geom_vector deriv1, deriv2;

	FUNC;
	// FIXME: neni prilis stabilni
	if (
		!GEOM_ETEST(geom_path_derivation_at_time(edge1->item->path, edge1->bottom_vertex->point.time, &deriv1)) ||
		!GEOM_ETEST(geom_path_derivation_at_time(edge2->item->path, edge2->bottom_vertex->point.time, &deriv2))) {
		TRACE("Hack #3623732.");
		ERET(0);
	}
	if (edge1->left_half.flags & GEOM_PATHSET_EDGE_DIR) { deriv1.dx = -deriv1.dx; deriv1.dy = -deriv1.dy; }
	if (edge2->left_half.flags & GEOM_PATHSET_EDGE_DIR) { deriv2.dx = -deriv2.dx; deriv2.dy = -deriv2.dy; }
	if (IS_ZERO(deriv1.dy) && IS_ZERO(deriv2.dy))
		ERET((deriv1.dx <= deriv2.dx) ? 0 : 1);
	else if (IS_ZERO(deriv1.dy))
		ERET((deriv1.dx <= 0) ? 0 : 1);
	else if (IS_ZERO(deriv2.dy))
		ERET((deriv2.dx >= 0) ? 0 : 1);
	else
		ERET((deriv1.dx * deriv2.dy - deriv1.dy * deriv2.dx <= 0) ? 0 : 1);
}



/*-------------------------------------------------------*
 *                                                       *
 *                        STENY                          *
 *                                                       *
 *-------------------------------------------------------*/


/* geom_pathset_sweep_face_new
 * - vytvori novou stenu */

static inline struct geom_pathset_sweep_face *geom_pathset_sweep_face_new(void)
{
	struct geom_pathset_sweep_face *face;

	face = geom_pathset_sweep_face_alloc();
	clist_add_tail(&geom_sweep.sweep_face_list, &face->sweep_node);
	cclist_init(&face->sweep_trap_list);
	clist_init(&face->poly_list);
	return face;
}


/* geom_pathset_sweep_face_close */

static inline void geom_pathset_sweep_face_close(struct geom_pathset_sweep_face *sweep_face, uns flags)
{
	struct geom_pathset_face *face;

	FUNC;
	BASSERT(sweep_face->sweep_trap_list.count <= 1);
	face = geom_pathset_face_alloc();
	sweep_face->poly_list.head.next->prev = &face->poly_list.head;
	sweep_face->poly_list.head.prev->next = &face->poly_list.head;
	face->poly_list.head.next = sweep_face->poly_list.head.next;
	face->poly_list.head.prev = sweep_face->poly_list.head.prev;
	face->flags = flags;
	clist_add_tail(&geom_sweep.pathset->face_list, &face->pathset_node);
	clist_remove(&sweep_face->sweep_node);
	geom_pathset_sweep_face_free(sweep_face);
	RET;
}


/* geom_pathset_sweep_face_insert_trap
 * - vlozi trapezoid do dane steny */

static inline void geom_pathset_sweep_face_insert_trap(struct geom_pathset_sweep_face *sweep_face, struct geom_pathset_sweep_trap *sweep_trap)
{
	cclist_add_tail(&sweep_face->sweep_trap_list, &sweep_trap->sweep_face_node);
	sweep_trap->sweep_face = sweep_face;
}


/* geom_pathset_sweep_face_merge
 * - vola se pri zjisteni souvislosti mezi dvojici sten
 * - sjednoti je v jednu a druhou uvolni z pameti
 * - amortizonavne O(log(n)) */

static void geom_pathset_sweep_face_merge(struct geom_pathset_sweep_face *face1, struct geom_pathset_sweep_face *face2)
{
	void *ptr1, *ptr2;

	if (face1->sweep_trap_list.count < face2->sweep_trap_list.count) { ptr1 = face1; face1 = face2; face2 = ptr1; }

	CCLIST_WALK_DELSAFE(ptr1, face2->sweep_trap_list, ptr2)
		geom_pathset_sweep_face_insert_trap(face1, SKIP_BACK(struct geom_pathset_sweep_trap, sweep_face_node, ptr1));
	clist_remove(&face2->sweep_node);
	if (!clist_empty(&face2->poly_list)) {
		face2->poly_list.head.next->prev = &face1->poly_list.head;
		face2->poly_list.head.prev->next = face1->poly_list.head.next;
		face1->poly_list.head.next->prev = face2->poly_list.head.prev;
		face1->poly_list.head.next = face2->poly_list.head.next;

	}
	geom_pathset_sweep_face_free(face2);
}


/*-------------------------------------------------------*
 *                                                       *
 *               KRIVKOVITE TRAPEZOIDY                   *
 *                                                       *
 *-------------------------------------------------------*/


/* geom_pathset_sweep_trap_search
 * - najde, ve kterem trapezoidu se vyskytuje dany bod
 * - pokud se nejedna o krajni trapezoid, vrati vyplnene struktury popisujici bod na leve a prave hrane trapezoidu */

static struct geom_pathset_sweep_trap *geom_pathset_sweep_trap_search(struct geom_point *point, struct geom_curve_point *left_result, struct geom_curve_point *right_result)
{
	struct geom_pathset_sweep_trap *trap;
	void *ptr, *next;

	ptr = clist_head(&geom_sweep.sweep_trap_list);
	while (1) {
		next = clist_next(&geom_sweep.sweep_trap_list, ptr);
		trap = SKIP_BACK(struct geom_pathset_sweep_trap, sweep_node, ptr);
		if (!next)
			return trap;
		geom_pathset_edge_search(SKIP_BACK(struct geom_pathset_edge, left_half, trap->right_half), point, right_result);
		if (right_result->point.x >= point->x)
			return trap;
		*left_result = *right_result;
		ptr = next;
	}
}


/* geom_pathset_sweep_trap_delete
 * - vymaze trapezoid ze steny a sweep liny */

static inline void geom_pathset_sweep_trap_remove(struct geom_pathset_sweep_trap *sweep_trap)
{
	cclist_remove(&sweep_trap->sweep_face->sweep_trap_list, &sweep_trap->sweep_face_node);
	clist_remove(&sweep_trap->sweep_node);
}


/* geom_pathset_sweep_trap_remove_intersection
 * - zrusi prusecik prirazeny danemu trapezoidu (pokud existuje) */

static void geom_pathset_sweep_trap_remove_intersection(struct geom_pathset_sweep_trap *sweep_trap)
{
	if (sweep_trap->intersection) {
		geom_pathset_heap_remove(sweep_trap->intersection);
		geom_pathset_entry_free(sweep_trap->intersection);
		sweep_trap->intersection = NULL;
	}
}


/* geom_pathset_sweep_trap_update_intersection
 * - pro dany trapezoid spocita mozny prusecik */

static void geom_pathset_sweep_trap_update_intersection(struct geom_pathset_sweep_trap *sweep_trap)
{
	struct geom_intersection intersection;
	struct geom_pathset_entry *entry;

	BASSERT(sweep_trap->intersection == NULL);
	if (sweep_trap->left_half && sweep_trap->right_half) {
		if (geom_pathset_half_intersection(sweep_trap->left_half, sweep_trap->right_half, &intersection, 0)) {
			entry = geom_pathset_entry_alloc();
			entry->type = GEOM_PATHSET_ENTRY_INTERSECTION;
			entry->y = intersection.point.y;
			entry->data.intersection.sweep_trap = sweep_trap;
			entry->data.intersection.geom = intersection;
			geom_pathset_heap_insert(entry);
			sweep_trap->intersection = entry;
		}
	}
}

static void geom_pathset_sweep_trap_update_intersection_hack(struct geom_pathset_sweep_trap *sweep_trap)
{
	struct geom_intersection intersection;
	struct geom_pathset_entry *entry;

	BASSERT(sweep_trap->intersection == NULL);
	if (sweep_trap->left_half && sweep_trap->right_half) {
		if (geom_pathset_half_intersection(sweep_trap->left_half, sweep_trap->right_half, &intersection, LARGE_EPSILON)) {
			entry = geom_pathset_entry_alloc();
			entry->type = GEOM_PATHSET_ENTRY_INTERSECTION;
			entry->y = intersection.point.y;
			entry->data.intersection.sweep_trap = sweep_trap;
			entry->data.intersection.geom = intersection;
			geom_pathset_heap_insert(entry);
			sweep_trap->intersection = entry;
		}
	}
}


/* geom_pathset_sweep_trap_open
 * - za dany trapezoid vlozi novy trapezoid
 * - novy trapezoid ohranici dvojici sousednich hran
 * - trapezoidu priradi novou stenu */

static struct geom_pathset_sweep_trap *geom_pathset_sweep_trap_open(struct geom_pathset_sweep_trap *left_trap, struct geom_pathset_half *left_half, struct geom_pathset_half *right_half)
{
	struct geom_pathset_sweep_trap *mid_trap;

	geom_pathset_half_join(right_half, left_half);
	mid_trap = geom_pathset_sweep_trap_alloc();
	mid_trap->left_half = left_half;
	mid_trap->right_half = right_half;
	left_half->trap = (void *)mid_trap;
	right_half->trap = (void *)mid_trap;
	mid_trap->intersection = NULL;
	clist_insert_after(&mid_trap->sweep_node, &left_trap->sweep_node);
	geom_pathset_sweep_face_insert_trap(geom_pathset_sweep_face_new(), mid_trap);
	return mid_trap;
}


/* geom_pathset_sweep_trap_close
 * - uzavre trapezoid dany 2 krajnima hranama
 * - kdyz mezi hranama je vic nez 1 trapezoid, jedna se o neobecnou polohu nebo numerickou chybu, coz se nak vyresi :-) */

static void geom_pathset_sweep_trap_close(struct geom_pathset_half *left_half, struct geom_pathset_half *right_half)
{
	struct geom_pathset_sweep_trap *trap;
	struct geom_pathset_edge *l_edge;
	struct geom_intersection intersection;

	FUNC;

	trap = HALF_SWEEP_TRAP(left_half);
	l_edge = SKIP_BACK(struct geom_pathset_edge, right_half, left_half);

	intersection.point = l_edge->bottom_vertex->point.point;
	intersection.time1 = l_edge->bottom_vertex->point.time;
	while (trap != HALF_SWEEP_TRAP(right_half)) {

		// HACK!!!!

		TRACE("Using intersection hack!");

		struct geom_pathset_sweep_trap *sweep_trap = trap;
		struct geom_pathset_vertex *left_vertex, *right_vertex;
		struct geom_pathset_edge *left_edge, *right_edge;
		struct geom_pathset_sweep_trap *left_trap, *right_trap, *mid_trap;
		struct geom_curve_point curve_point;

		left_edge = SKIP_BACK(struct geom_pathset_edge, right_half, sweep_trap->left_half);
		right_edge = SKIP_BACK(struct geom_pathset_edge, left_half, sweep_trap->right_half);
		geom_pathset_edge_search(right_edge, &intersection.point, &curve_point);
		intersection.time2 = curve_point.time;
		left_vertex = geom_pathset_vertex_new(&intersection.point, intersection.time1);
		right_vertex = geom_pathset_vertex_new(&intersection.point, intersection.time2);
		geom_pathset_sweep_trap_remove_intersection(sweep_trap);
		geom_pathset_edge_split(left_edge, left_vertex);
		geom_pathset_edge_split(right_edge, right_vertex);
		left_trap = HALF_SWEEP_TRAP(&left_edge->left_half);
		right_trap = HALF_SWEEP_TRAP(&right_edge->right_half);
		geom_pathset_sweep_trap_remove_intersection(left_trap);
		geom_pathset_sweep_trap_remove_intersection(right_trap);
		geom_pathset_sweep_trap_close(left_edge->right_half.ccw_half, right_edge->left_half.cw_half);

		geom_pathset_half_join(&right_edge->left_half, left_edge->left_half.cw_half);
		geom_pathset_half_join(right_edge->right_half.ccw_half, &left_edge->right_half);
		left_trap->right_half = &right_edge->left_half;
		right_trap->left_half = &left_edge->right_half;
		right_edge->left_half.trap = (void *)left_trap;
		left_edge->right_half.trap = (void *)right_trap;
		mid_trap = geom_pathset_sweep_trap_open(left_trap, &right_edge->right_half, &left_edge->left_half);

		geom_pathset_sweep_trap_update_intersection(left_trap);
		geom_pathset_sweep_trap_update_intersection_hack(mid_trap);
		trap = right_trap;

	}
	geom_pathset_sweep_trap_remove_intersection(trap);
	geom_pathset_half_join(left_half, right_half);
	geom_pathset_sweep_trap_remove(trap);
	if (!trap->sweep_face->sweep_trap_list.count) {
		geom_pathset_poly_new(left_half);
		geom_pathset_sweep_face_close(trap->sweep_face, 0);
	}
	geom_pathset_sweep_trap_free(trap);

	RET;
}


/* geom_pathset_sweep_trap_merge
 * - spoji dva sousedni trapezoidy, uzavre krivku */

static void geom_pathset_sweep_trap_merge(struct geom_pathset_sweep_trap *left_trap)
{
	struct geom_pathset_sweep_trap *right_trap;

	right_trap = SWEEP_TRAP_NEXT(left_trap);
	geom_pathset_sweep_trap_remove_intersection(left_trap);
	geom_pathset_sweep_trap_remove_intersection(right_trap);
	geom_pathset_half_join(right_trap->left_half, left_trap->right_half);
	if (left_trap->sweep_face == right_trap->sweep_face)
		geom_pathset_poly_new(left_trap->right_half);
	else
		geom_pathset_sweep_face_merge(left_trap->sweep_face, right_trap->sweep_face);
	if (!SWEEP_TRAP_IS_LAST(right_trap))
		right_trap->right_half->trap = (void*)left_trap;
	left_trap->right_half = right_trap->right_half;
	geom_pathset_sweep_trap_remove(right_trap);
	geom_pathset_sweep_trap_free(right_trap);
}


/* geom_pathset_sweep_trap_split
 * - rozdeli trapezoid na dva sousedni
 * - vlozi mezi ne dvojisi dousednich hran
 * - oba trapezoidy patri stejne stene */

static struct geom_pathset_sweep_trap *geom_pathset_sweep_trap_split(struct geom_pathset_sweep_trap *left_trap, struct geom_pathset_half *left_half, struct geom_pathset_half *right_half)
{
	struct geom_pathset_sweep_trap *right_trap;

	geom_pathset_sweep_trap_remove_intersection(left_trap);
	geom_pathset_half_join(left_half, right_half);
	right_trap = geom_pathset_sweep_trap_alloc();
	right_trap->right_half = left_trap->right_half;
	if (!SWEEP_TRAP_IS_LAST(right_trap))
		right_trap->right_half->trap = (void*)right_trap;
	left_trap->right_half = left_half;
	right_trap->left_half = right_half;
	left_half->trap = (void*)left_trap;
	right_half->trap = (void*)right_trap;
	right_trap->intersection = NULL;
	clist_insert_after(&right_trap->sweep_node, &left_trap->sweep_node);
	geom_pathset_sweep_face_insert_trap(left_trap->sweep_face, right_trap);
	return right_trap;
}


/*-------------------------------------------------------*
 *                                                       *
 *                ZPRACOVANI Y-UDALOSTI                  *
 *                                                       *
 *-------------------------------------------------------*/


/* geom_pathset_do_start
 * - zpracovani bodu, kde zacina cesta
 * - dane souradnice bodu a dvojice sousednich hran u koncoveho bodu cesty */

static void geom_pathset_do_start(struct geom_pathset_edge *edge)
{
	struct geom_pathset_sweep_trap *left_trap, *right_trap;
	struct geom_curve_point left_point, right_point;

	FUNC;
		
	PASSERT(edge);

	left_trap = geom_pathset_sweep_trap_search(&edge->bottom_vertex->point.point, &left_point, &right_point);
	right_trap = geom_pathset_sweep_trap_split(left_trap, &edge->left_half, &edge->right_half);
	geom_pathset_sweep_trap_update_intersection(left_trap);
	geom_pathset_sweep_trap_update_intersection(right_trap);

	RET;
}


/* geom_pathset_do_end */

static void geom_pathset_do_end(struct geom_pathset_edge *edge)
{
	struct geom_pathset_sweep_trap *left_trap, *right_trap;

	FUNC;
		
	PASSERT(edge);

	left_trap = HALF_SWEEP_TRAP(&edge->left_half);
	right_trap = HALF_SWEEP_TRAP(&edge->right_half);
	geom_pathset_sweep_trap_merge(left_trap);
	geom_pathset_sweep_trap_update_intersection(left_trap);

	RET;
}


/* geom_pathset_do_open */

static void geom_pathset_do_open(struct geom_pathset_edge *edge1, struct geom_pathset_edge *edge2)
{
	struct geom_pathset_edge *left_edge, *right_edge;
	struct geom_pathset_sweep_trap *left_trap, *right_trap, *mid_trap;
	struct geom_curve_point left_point, right_point;

	FUNC;
		
	PASSERT(edge1);PASSERT(edge2);

	left_trap = geom_pathset_sweep_trap_search(&edge1->bottom_vertex->point.point, &left_point, &right_point);
	if (geom_pathset_edge_order(edge1, edge2, left_trap, left_trap) == 0) {
		left_edge = edge1;
		right_edge = edge2;
	}
	else {
		left_edge = edge2;
		right_edge = edge1;
	}
	right_trap = geom_pathset_sweep_trap_split(left_trap, &left_edge->left_half, &right_edge->right_half);
	mid_trap = geom_pathset_sweep_trap_open(left_trap, &left_edge->right_half, &right_edge->left_half);
	geom_pathset_sweep_trap_update_intersection(left_trap);
	geom_pathset_sweep_trap_update_intersection(right_trap);
	geom_pathset_sweep_trap_update_intersection_hack(mid_trap);

	RET;
}



/* geom_pathset_do_close */

static void geom_pathset_do_close(struct geom_pathset_edge *edge1, struct geom_pathset_edge *edge2)
{
	void *ptr;
	struct geom_pathset_edge *left_edge, *right_edge;
	struct geom_pathset_sweep_trap *trap;

	FUNC;
		
	PASSERT(edge1);PASSERT(edge2);

	if (HALF_SWEEP_TRAP(&edge1->right_half) == HALF_SWEEP_TRAP(&edge2->left_half)) {
EDGE1_FIRST:
		left_edge = edge1;
		right_edge = edge2;
	}
	else if (HALF_SWEEP_TRAP(&edge1->left_half) == HALF_SWEEP_TRAP(&edge2->right_half)) {
EDGE2_FIRST:
		left_edge = edge2;
		right_edge = edge1;
	}
	else {
		CLIST_WALK(ptr, geom_sweep.sweep_trap_list) {
			if (ptr == &HALF_SWEEP_TRAP(&edge1->left_half)->sweep_node)
				goto EDGE1_FIRST;
			else if (ptr == &HALF_SWEEP_TRAP(&edge2->left_half)->sweep_node)
				goto EDGE2_FIRST;
		}
		BUG();
	}
	geom_pathset_sweep_trap_close(&left_edge->right_half, &right_edge->left_half);
	trap = HALF_SWEEP_TRAP(&left_edge->left_half);
	geom_pathset_sweep_trap_merge(trap);
	geom_pathset_sweep_trap_update_intersection(trap);

	RET;
}


/* geom_pathset_do_intersection */

static void geom_pathset_do_intersection(struct geom_pathset_sweep_trap *sweep_trap)
{
	struct geom_intersection *intersection;
	struct geom_pathset_vertex *left_vertex, *right_vertex;
	struct geom_pathset_edge *left_edge, *right_edge;
	struct geom_pathset_sweep_trap *left_trap, *right_trap, *mid_trap;


	intersection = &sweep_trap->intersection->data.intersection.geom;

	FUNC;
		
	PASSERT(sweep_trap);

	left_edge = SKIP_BACK(struct geom_pathset_edge, right_half, sweep_trap->left_half);
	right_edge = SKIP_BACK(struct geom_pathset_edge, left_half, sweep_trap->right_half);
	left_vertex = geom_pathset_vertex_new(&intersection->point, intersection->time1);
	right_vertex = geom_pathset_vertex_new(&intersection->point, intersection->time2);
	geom_pathset_sweep_trap_remove_intersection(sweep_trap);
	geom_pathset_edge_split(left_edge, left_vertex);
	geom_pathset_edge_split(right_edge, right_vertex);
	left_trap = HALF_SWEEP_TRAP(&left_edge->left_half);
	right_trap = HALF_SWEEP_TRAP(&right_edge->right_half);
	geom_pathset_sweep_trap_remove_intersection(left_trap);
	geom_pathset_sweep_trap_remove_intersection(right_trap);
	geom_pathset_sweep_trap_close(left_edge->right_half.ccw_half, right_edge->left_half.cw_half);
	/*if (geom_pathset_edge_order(right_edge, left_edge, left_trap, right_trap) == 0)*/ {
		geom_pathset_half_join(&right_edge->left_half, left_edge->left_half.cw_half);
		geom_pathset_half_join(right_edge->right_half.ccw_half, &left_edge->right_half);
		left_trap->right_half = &right_edge->left_half;
		right_trap->left_half = &left_edge->right_half;
		right_edge->left_half.trap = (void *)left_trap;
		left_edge->right_half.trap = (void *)right_trap;
		mid_trap = geom_pathset_sweep_trap_open(left_trap, &right_edge->right_half, &left_edge->left_half);
	}
	/*else {  // dotykovy prusecik
		geom_pathset_half_join();
		BUG();
		geom_pathset_half_join(left_edge->left_half.ccw_half, &right_edge->left_half);
		geom_pathset_half_join(right_edge->right_half.cw_half, &left_edge->right_half);
		mid_trap = geom_pathset_sweep_trap_open(left_trap, &right_edge->right_half, &left_edge->left_half);
	}*/
	geom_pathset_sweep_trap_update_intersection(left_trap);
	geom_pathset_sweep_trap_update_intersection(right_trap);
	geom_pathset_sweep_trap_update_intersection_hack(mid_trap);

	RET;
}


/*-------------------------------------------------------*
 *                                                       *
 *                      ARRANGEMENT                      *
 *                                                       *
 *-------------------------------------------------------*/


/* geom_pathset_unref_neighbour */

static inline void geom_pathset_unref_neighbour(struct geom_pathset_entry *entry)
{
	if (entry->type == GEOM_PATHSET_ENTRY_END || !(--(entry->data.close.ref_count)))
		geom_pathset_heap_insert(entry);
}


/* geom_pathset_arrangement_compute */

static void geom_pathset_arrangement_compute(struct geom_pathset *pathset)
{
	void *ptr;
	struct geom_pathset_sweep_face *sweep_face;
	struct geom_pathset_sweep_trap *sweep_trap;
	struct geom_pathset_item *item;
	struct geom_pathset_entry *entry;

	FUNC;
	GEOM_PATHSET_ASSERT(pathset);

	// inicializace
	geom_sweep.pathset = pathset;
	clist_init(&pathset->face_list);
	clist_init(&geom_sweep.sweep_face_list);

	// inicializace haldy
	geom_sweep.heap_n = 0;
	garr_init(&geom_sweep.heap, 64, sizeof(struct geom_pathset_entry *));

	// inicializace trapezoidu
	sweep_face = geom_pathset_sweep_face_new();
	clist_init(&geom_sweep.sweep_trap_list);
	sweep_trap = geom_pathset_sweep_trap_alloc();
	sweep_trap->left_half = NULL;
	sweep_trap->right_half = NULL;
	sweep_trap->intersection = NULL;
	clist_add_tail(&geom_sweep.sweep_trap_list, &sweep_trap->sweep_node);
	geom_pathset_sweep_face_insert_trap(sweep_face, sweep_trap);

	// rozdeleni cest na monotonni useky
	CLIST_WALK(ptr, pathset->item_list) {
		geom_pathset_sweep_assert();
		item = SKIP_BACK(struct geom_pathset_item, pathset_node, ptr);
		geom_pathset_item_init_edges(item);
	}

	// pruchod rovinou
	while (geom_sweep.heap_n) {
		geom_pathset_sweep_assert();
		entry = *(struct geom_pathset_entry **)garr_get(&geom_sweep.heap, 0);
		geom_sweep.y = entry->y;

		switch (entry->type) {
		case GEOM_PATHSET_ENTRY_START:
			geom_pathset_heap_remove(entry);
			geom_pathset_do_start(entry->data.start.edge);
			geom_pathset_unref_neighbour(entry->data.start.neighbour);
			geom_pathset_entry_free(entry);
			break;

		case GEOM_PATHSET_ENTRY_END:
			geom_pathset_heap_remove(entry);
			geom_pathset_do_end(entry->data.end.edge);
			geom_pathset_entry_free(entry);
		break;

		case GEOM_PATHSET_ENTRY_OPEN:
			geom_pathset_heap_remove(entry);
			geom_pathset_do_open(entry->data.open.edge1, entry->data.open.edge2);
			geom_pathset_unref_neighbour(entry->data.open.neighbour1);
			geom_pathset_unref_neighbour(entry->data.open.neighbour2);
			geom_pathset_entry_free(entry);
			break;

		case GEOM_PATHSET_ENTRY_CLOSE:
			geom_pathset_heap_remove(entry);
			geom_pathset_do_close(entry->data.close.edge1, entry->data.close.edge2);
			geom_pathset_entry_free(entry);
			break;

		case GEOM_PATHSET_ENTRY_INTERSECTION:
			geom_pathset_do_intersection(entry->data.intersection.sweep_trap);
			break;

		case GEOM_PATHSET_ENTRY_INTERNAL:
			BUG();
		}

	}

	geom_pathset_sweep_assert();

	// uzavrit vnejsi stenu/trapezoid
	sweep_face = SKIP_BACK(struct geom_pathset_sweep_face, sweep_node, geom_sweep.sweep_face_list.head.next);
	sweep_trap = SKIP_BACK(struct geom_pathset_sweep_trap, sweep_node, geom_sweep.sweep_trap_list.head.next);
	geom_pathset_sweep_face_close(sweep_face, GEOM_PATHSET_FACE_OPEN);
	geom_pathset_sweep_trap_free(sweep_trap);

	// uvolneni pomocnych strukturu
	garr_cleanup(&geom_sweep.heap);

	RET;
}


/* geom_pathset_arrangement_clear */

static inline void geom_pathset_arrangement_clear(struct geom_pathset *pathset)
{
	if (pathset->flags & GEOM_PATHSET_ARRANGEMENT_VALID) {
		// FIXME: uvolnit expanzi
		pathset->flags &= ~GEOM_PATHSET_ARRANGEMENT_VALID;
	}
}


/* geom_pathset_arrangement_update */

int geom_pathset_arrangement_update(struct geom_pathset *pathset)
{
	FUNC;
	if (!(pathset->flags & GEOM_PATHSET_ARRANGEMENT_VALID)) {
		geom_pathset_arrangement_compute(pathset);
		pathset->flags |= GEOM_PATHSET_ARRANGEMENT_VALID;
	}
	ERET(0);
}


/*-------------------------------------------------------*
 *                                                       *
 *                                                       *
 *                                                       *
 *-------------------------------------------------------*/


/* geom_pathset_valid */

int geom_pathset_valid(const struct geom_pathset *pathset UNUSED)
{
	return 1;
}


/* geom_pathset_dump */

void geom_pathset_dump(const struct geom_pathset *pathset)
{
	struct geom_pathset_item *item;
	uns i;

	msg("Dumping geom_pathset structure at %p: ", (void*)pathset);
	geom_pathset_msg(pathset);
	msg("\n");
	i = 0;
	GEOM_PATHSET_WALK_ITEMS(pathset, item) {
		i++;
		msg("%d. item: flags=0x%x path=", i, item->flags);
		geom_path_msg(item->path);
		msg("\n");
	}
}


/* geom_pathset_msg */

void geom_pathset_msg(const struct geom_pathset *pathset)
{
	if (pathset) {
		if (!geom_pathset_valid(pathset)) msg("!");
		msg("(flags=0x%x)", pathset->flags);
	}
	else
		msg("NULL");
}


/* geom_pathset_clear */

int geom_pathset_clear(struct geom_pathset *pathset)
{
	void *ptr1, *ptr2;
	struct geom_pathset_item *item;

	FUNC;
	geom_pathset_arrangement_clear(pathset);
	CLIST_WALK_DELSAFE(ptr1, pathset->item_list, ptr2) {
		item = SKIP_BACK(struct geom_pathset_item, pathset_node, ptr1);
		if (!(item->flags & GEOM_PATHSET_ITEM_LINK))
			geom_path_destroy(item->path);
		geom_pathset_item_free(item);
	}
	clist_init(&pathset->item_list);
	ERET(0);
}


/* geom_pathset_init */

int geom_pathset_init(struct geom_pathset *pathset)
{
	FUNC;
	GEOM_PATHSET_ASSERT(pathset);

	pathset->flags = 0;
	clist_init(&pathset->item_list);

	ERET(0);
}


/* geom_pathset_destroy */

int geom_pathset_destroy(struct geom_pathset *pathset)
{
	FUNC;
	GEOM_PATHSET_ASSERT(pathset);
	geom_pathset_clear(pathset);
	ERET(0);
}


/* geom_pathset_insert */

int geom_pathset_insert(struct geom_pathset *pathset, uns flags, struct geom_path *path)
{
	struct geom_pathset_item *item;

	FUNC;
	GEOM_PATHSET_ASSERT(pathset);
	GEOM_PATH_ASSERT(path);

	geom_pathset_arrangement_clear(pathset);
	item = geom_pathset_item_alloc();
	item->flags = flags;
	item->path = path;
	clist_add_tail(&pathset->item_list, &item->pathset_node);

	ERET(0);
}


/* geom_pathset_delete */

int geom_pathset_delete(struct geom_pathset *pathset, struct geom_pathset_item *item)
{
	FUNC;
	GEOM_PATHSET_ASSERT(pathset);

	geom_pathset_arrangement_clear(pathset);
	clist_remove(&item->pathset_node);
	geom_pathset_item_free(item);

	ERET(0);
}

void geom_pathset_dump(struct geom_pathset *ps UNUSED)
{
	uns i, j;
	struct geom_pathset_item *item;
	struct geom_pathset_edge *edge;
	void *ptr, *ptr2;

	msg("Dumping geom_pathset structure at %p\n", ps);
	if (!ps)
		return;
	if (!geom_pathset_valid(ps))
		msg("(invalid)\n");
	msg("flags=0x%x\n", ps->flags);

	i = 0;
	CLIST_WALK(ptr, ps->item_list) {
		msg("Guts of embedded item %d:\n", i);
		item = SKIP_BACK(struct geom_pathset_item, item_node, ptr);
		geom_path_dump(item->path);

		msg("Edge list dump:\n");
		j = 0;
		CLIST_WALK(ptr2, item->edge_list) {
			edge = SKIP_BACK(struct geom_pathset_edge, edge_node, ptr2);
			msg("Edge %d: ", j);
			msg("top_time=%f bottom_time=%f top_point=[%f,%f] bottom_point=[%f,%f]\n",
					edge->top_time, edge->bottom_time, edge->top_point.x, edge->top_point.y,
					edge->bottom_point.x, edge->bottom_point.y);
			j++;
		}
		msg("End of edge list dump.\n");
		msg("End of item %d dump.\n", i);
		i++;
	}
}
#endif
