/** @file
 *
 * 	GEOMLIB - Pathsets.
 *
 */

#ifndef _VRR_GEOMLIB_PATHSET_H_
#define _VRR_GEOMLIB_PATHSET_H_

#include "geomlib/path.h"


#define geom_pathset_VMT    geom_group_VMT
#define geom_pathset_INIT   geom_group_INIT

GEOM_CLASS_HEAD(pathset, group);


/* Constructor and destructor */

static inline void geom_pathset_create(struct geom_pathset *pathset)
{ geom_group_create(&pathset->group); }

static inline void geom_pathset_destroy(struct geom_pathset *pathset)
{ geom_group_destroy(&pathset->group); }




#if 0

/* FIXME
	- definice presunout do types.h
 */


/** Path has references outside the pathset and is not destroyed with pathset. */
#define GEOM_PATHSET_ITEM_LINK 0x1

/** Structure to describe each path in pathset. */
struct geom_pathset_item {
	uns flags;                /** GEOM_PATHSET_ITEM_x flags. */
	struct geom_path *path;   /** Pointer to geom_path structure. */
	cnode pathset_node;       /** List of paths in pathset. */
};


#define GEOM_PATHSET_ARRANGEMENT_VALID 0x1

/** Set of paths. */
struct geom_pathset {
	uns flags;         /** GEOM_PATHSET_x flags. */
	clist item_list;   /** List of paths in pathset. */
	clist face_list;
};


/**
 * Macro makes assertions to check validity of geom_pathset structure.
 *
 * @ai  t  Pointer to geom_pathset structure.
 */
#ifdef DEBUG
#define GEOM_PATHSET_ASSERT(p) GEOM_PASSERT(p)
#else
#define GEOM_PATHSET_ASSERT(p) do{}while(0)
#endif


/**
 * Macro cycles all allocated items.
 *
 * @ai  pathset  Pointer to pathset.
 * @ao  item     Pointer to geom_pathset_item varriable name for current item.
 */
#define GEOM_PATHSET_WALK_ITEMS(pathset, item) \
	for (item = (void *)(pathset)->item_list.head.next; \
		item != (void *)&(pathset)->item_list.head && ((item = SKIP_BACK(struct geom_pathset_item, pathset_node, item)), 1); \
		item = (void *)item->pathset_node.next)


/**
 * Function check validity of geom_pathset structure.
 *
 * @ai  pathset  Pointer to geom_pathset structure.
 * @rv  nonzero  Structure is valid.
 * @rv  zero     Structure is invalid.
 */
int geom_pathset_valid(const struct geom_pathset *pathset);


/**
 * Dumps geom_pathset starture data.
 *
 * @ai  pathset  Pointer to geom_pathset structure (could contain invalid data).
 */
void geom_pathset_dump(const struct geom_pathset *pathset);


/**
 * Short dump of geom_pathset starture data.
 *
 * @ai  pathset  Pointer to geom_pathset structure (could contain invalid data).
 */
void geom_pathset_msg(const struct geom_pathset *pathset);


int geom_pathset_init(struct geom_pathset *pathset);

int geom_pathset_destroy(struct geom_pathset *pathset);

int geom_pathset_clear(struct geom_pathset *pathset);

int geom_pathset_insert(struct geom_pathset *pathset, uns flags, struct geom_path *path);

int geom_pathset_delete(struct geom_pathset *pathset, struct geom_pathset_item *item);




// vrchol v arrangementu
struct geom_pathset_vertex {
	struct geom_curve_point point;
};

// uzavrena krivka tvorici hranici oblasti
struct geom_pathset_poly {
	cnode face_node;
	struct geom_pathset_half *first_half;
};

// 0 rostouci Y, jinak klesajici
#define GEOM_PATHSET_EDGE_DIR             0x1
#define GEOM_PATHSET_EDGE_DIR_INC         0
#define GEOM_PATHSET_EDGE_DIR_DEC         GEOM_PATHSET_EDGE_DIR
// nenula, pokud zacina v pocatecnim bodu cesty
#define GEOM_PATHSET_EDGE_START           0x2
// nenula, pokus konci v koncovem bodu cesty
#define GEOM_PATHSET_EDGE_END             0x4
// 0 leva pulka, jinak prava hrany
#define GEOM_PATHSET_HALF_SIDE            0x8
#define GEOM_PATHSET_HALF_SIDE_LEFT       0
#define GEOM_PATHSET_HALF_SIDE_RIGHT      GEOM_PATHSET_HALF_SIDE

// "pulka" hrany v arrangementu
struct geom_pathset_half {
	uns flags;
	struct geom_pathset_half *cw_half;
	struct geom_pathset_half *ccw_half;
	struct geom_pathset_trap *trap;
};

// hrana v arrangementu
struct geom_pathset_edge {
	struct geom_pathset_item *item;
	struct geom_pathset_vertex *bottom_vertex;
	struct geom_pathset_vertex *top_vertex;
	struct geom_pathset_half left_half;
	struct geom_pathset_half right_half;
};

// pseudo-trapezoid v arrangementu
struct geom_pathset_trap {
	byte _unused;
};


#define GEOM_PATHSET_FACE_OPEN 0x1

// oblast v arrangementu
struct geom_pathset_face {
	uns flags;
	cnode pathset_node;
	clist poly_list;
};




int geom_pathset_arrangement_update(struct geom_pathset *pathset);

#endif

#endif /* _VRR_GEOMLIB_PATHSET_H */

