/** @file
 *
 * 	GEOMLIB - Polynomial solver and auxilliary routines.
 *
 */

#ifndef _VRR_GEOMLIB_POLYNOMIAL_H_
#define _VRR_GEOMLIB_POLYNOMIAL_H_


/* constants */

#define GEOM_SOLVE_MULTIPLICITY   0x00000001
#define GEOM_SOLVE_LEFT_ONLY      0x00000002
#define GEOM_SOLVE_UNIT_INTERVAL  0x00000004

/**
 * Test for validity of given polynomial.
 *
 * @a degree	polynomial degree
 * @a coef	array of size degree of polynomial coefficients
 * @r		boolean value
 */
int geom_polynomial_valid(int degree, double coef[]);

/**
    \brief Function evaluates a real polynomial at given parameter t.
    \param degree integer degree of polynomial
    \param coef array of polynomial coefficients (in order of increasing powers)
    \param t a real number
    \retval 0 success
    \retval -1 numeric overflow
*/
double geom_polynomial_evaluate(int degree, double coef[], double t);

/**
    \brief Function returns product of two real polynomials.
    \param degree1 integer degree of first polynomial
    \param coef1 array of first polynomial coefficients (in order of increasing powers)
    \param degree2 integer degree of second polynomial
    \param coef1 array of second polynomial coefficients
    \param result array of result polynomial coefficients (size degree1+degree2+1)
    \retval >=0 integer degree of result polynomial
    \retval -1 numeric overflow
*/
int geom_polynomial_mul(int degree1, double coef1[], int degree2, double coef2[], double result[]);

/**
    \brief Function finds all real roots of a real polynomial.
    \param degree integer degree of polynomial
    \param coef array of polynomial coefficients (in order of increasing powers)
    \param flags OR combinations of
        GEOM_SOLVE_MULTIPLICITY (returns multiple roots as multiple array entries),
        GEOM_SOLVE_LEFT_ONLY (returns one root only),
        GEOM_SOLVE_UNIT_INTERVAL (returns only roots in closed interval <0,1>)
    \param result array of found roots (degree entries needed, one entry with GEOM_SOLVE_LEFT_ONLY)
    \retval >=0 number of found roots
    \retval -1 numeric error
*/
int geom_polynomial_solve(int degree, double coef[], int flags, double result[]);


#endif
