/** @file
 *
 * 	GEOMLIB - Rectangles.
 *
 */

#ifndef _VRR_GEOMLIB_RECTANGLE_H_
#define _VRR_GEOMLIB_RECTANGLE_H_

#include "geomlib/vector.h"


int geom_rectangle_valid (const struct geom_rectangle *r);

const struct geom_rectangle *geom_rectangle_assert(GEOM_ASSERT_DEF, const struct geom_rectangle *r);

void geom_rectangle_dump(const struct geom_rectangle *r);





int geom_rectangle_is_valid (const struct geom_rectangle *r);

/**
   \brief Returns whether rectangle is empty (lower left point > upper right point)
 */

extern int
geom_rectangle_is_empty (const struct geom_rectangle *r);

/**
   \brief Sets rectangle to empty (lower left point > upper right point)
 */

extern int
geom_rectangle_set_empty (struct geom_rectangle *r);

/**
   \brief Function merges two rectangles to a rectangle enclosing both.
	  One rectangle can be used as many arguments.

   \retval 0 success
   \retval -1 numeric error
*/

extern int
geom_rectangle_merge (const struct geom_rectangle *r1,
		      const struct geom_rectangle *r2,
		      struct geom_rectangle *result);

/**
   \brief Function merges rectangle and point to a rectangle enclosing both.
	  One rectangle can be used as many arguments.

   \retval 0 success
   \retval -1 numeric error
*/

extern int
geom_rectangle_merge_point (const struct geom_rectangle *r,
		      const struct geom_point *p,
		      struct geom_rectangle *result);

/**
   \brief Function enlarges rectangle in every direction by value
  	  One rectangle can be used as both source and result.

   \retval 0 success
   \retval -1 numeric error
*/

extern int
geom_rectangle_enlarge (const struct geom_rectangle *source, real value,
		        struct geom_rectangle *result);


/**
   \brief Returns whether two rectangles have nonempty intersection
 */

extern int
geom_rectangle_have_intersection (const struct geom_rectangle *r1,
				  const struct geom_rectangle *r2);


/**
   \brief Returns whether first rectangle is completely in second
 */

extern int
geom_rectangle_is_in (const struct geom_rectangle *r1, const struct geom_rectangle *r2);

/**
   \brief Returns whether rectangle adjoins (their borders have common point)
 */

extern int
geom_rectangle_is_adjoin (const struct geom_rectangle *r1, const struct geom_rectangle *r2);

#endif
