#ifndef _VRR_GEOMLIB_SPECIAL_H_
#define _VRR_GEOMLIB_SPECIAL_H_


/**
 * @file geomlib/special.h
 *
 * Geomlib - Simple special objects.
 */


#include "geomlib/bezier.h"
#include "geomlib/fpath.h"


/*----------------------------------------------*
 *                                              *
 *                  POINT ITEM                  *
 *                                              *
 *----------------------------------------------*/


#define geom_point_item_VMT geom_item_VMT
#define geom_point_item_INIT geom_item_INIT \
	GEOM_CURVE_DEFAULTS \
	GEOM_CURVE_TIME_EQUALS_BTIME \
	.expansion_append = (int (*)(void *, struct geom_fpath *)) &geom_point_item_expansion_append,
GEOM_CLASS_HEAD(point_item, item);


/**
 * Create new point in the origin.
 * Deafult constructor.
 *
 * @ai  item  Instance pointer.
 */
static inline void geom_point_item_create(struct geom_point_item UNUSED *item)
{ GEOM_INSTANCE_IS_ASSERT(point_item, item); }

/**
 * Change point coordinates.
 * Can be used as constructor.
 *
 * @ai  item   Instacne pointer.
 * @ai  point  Wanted coordinates.
 */
static inline void geom_point_item_set(struct geom_point_item *item, struct geom_point *point)
{ GEOM_INSTANCE_IS_ASSERTX(point_item, item)->point = *point; }


/* Overrided methods */

static inline void geom_point_item_destroy(struct geom_point_item *item)
{ geom_item_expansion_release(item); geom_item_destroy(&GEOM_INSTANCE_IS_ASSERTX(point_item, item)->item); }

static inline int geom_point_item_expansion_append(struct geom_point_item *item, struct geom_fpath *path)
{ void *p; GEOM_FPATH_APPEND_NEW_CREATE2(p, path, bezier, geom_bezier_set_linear_nonrational, &item->point, &item->point); return 0; }


/*----------------------------------------------*
 *                                              *
 *             CALLBACK-EXPANSION ITEM          *
 *                                              *
 *----------------------------------------------*/


#define geom_callback_item_VMT geom_item_VMT
#define geom_callback_item_INIT geom_item_INIT \
	GEOM_CURVE_DEFAULTS \
	GEOM_CURVE_TIME_EQUALS_BTIME \
	.tlength = (real (*)(void *)) &geom_callback_item_tlength, \
	.expansion_append = (int (*)(void *, struct geom_fpath *)) &geom_callback_item_expansion_append,
GEOM_CLASS_HEAD(callback_item, item);


/**
 * Creates new item with undefined expansion.
 * Callback must be set manually before calls to geometric routines.
 * Default constructor.
 *
 * @ai  item  Instance pointer.
 */
static inline void geom_callback_item_create(struct geom_callback_item UNUSED *item)
{ GEOM_INSTANCE_IS_ASSERT(callback_item, item); }


/**
 * Creates new item with given expansion callback.
 * Constructor.
 *
 * @ai  item  Instance pointer.
 * @ai  func  Expansion callback pointer.
 */
static inline void geom_callback_item_create_set(struct geom_callback_item *item, geom_callback_item_func *func)
{ GEOM_INSTANCE_IS_ASSERTX(callback_item, item)->func = *func; }


/* Overrided methods */

static inline void geom_callback_item_destroy(struct geom_callback_item *item)
{ geom_item_expansion_release(item); geom_item_destroy(&GEOM_INSTANCE_IS_ASSERTX(callback_item, item)->item); }

static inline real geom_callback_item_tlength(struct geom_callback_item *item) { return geom_item_blength__expand(item); }

int geom_callback_item_expansion_append(struct geom_callback_item *item, struct geom_fpath *path);


#endif

