/** @file
 *
 * 	GEOMLIB - Transformations alternatives.
 *
 */

#ifndef _VRR_GEOMLIB_TRANSFORM2_H_
#define _VRR_GEOMLIB_TRANSFORM2_H_



int geom_transform2_is_valid(const struct geom_transform2 *t);
int geom_transform2_set_identity (struct geom_transform2 *t);
int geom_transform2_set_move (struct geom_transform2 *t, const double x, const double y);
int geom_transform2_set_rotation (struct geom_transform2 *t, const double angle);
int geom_transform2_set_scale (struct geom_transform2 *t, const double x, const double y);
/**
	\brief Initialize transformation matrix T so (T(0, 0) = (p1x, p1y), T(1, 0) = (p2x, p2y), T(0, 1) = (p3x, p3y)).

	\retval 0 success
	\retval -1 numeric overflow or singular matrix
*/
int geom_transform2_set_basis (struct geom_transform2 *t, real p1x, real p1y, real p2x, real p2y, real p3x, real p3y);
int geom_transform2_invert (const struct geom_transform2 *t, struct geom_transform2 *result);
int geom_transform2_merge (const struct geom_transform2 *t1, const struct geom_transform2 *t2, struct geom_transform2 *result);

// (T <= NT(x, y) * T)

int geom_transform2_translate (struct geom_transform2 *t, const double x, const double y);
int geom_transform2_rotate (struct geom_transform2 *t, const double angle);
int geom_transform2_scale (struct geom_transform2 *t, const double x, const double y);

int geom_transform2_set_rotate_center (struct geom_transform2 *t, const real x, const real y, const real angle);
int geom_transform2_set_scale_center (struct geom_transform2 *t, const real x, const real y, const real sx, const real sy);

int geom_transform2_set_skew (struct geom_transform2 *t,
			      real x1, real y1, real x2, real y2,
			      real xf, real yf, real xt, real yt);

/** Computes the skew affine transformation in which (zx,zy) and (ax,ay) are fixpoints
  and normalized (fx,fy) becomes normalized (tx,ty).

  \retval 0 success
  \retval -1 numeric overflow or singular matrix
*/
int geom_transform2_set_skew_normal (struct geom_transform2 *t, const real zx, const real zy,
	const real ax, const real ay, const real fx, const real fy, const real tx, const real ty);

#endif

