#include "lib/lib.h"
#include <gtk/gtk.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include "gui/main.h"
#include "gui/changeunits.h"
#include "gui/ruler.h"

static void destroy(GtkWidget *widget UNUSED, gpointer data)
{
	xfree(data);
}

void rulers_set_resolution(struct view *lw UNUSED, float resolution UNUSED)
{
  //	vert_ruler_set_resolution(RULER (lw->vruler), resolution);
  //	horiz_ruler_set_resolution(RULER (lw->hruler), resolution);
}

// returns 0 if unit_name hasn't been defined yet, or unit's multiplier
// in case it was defined
float get_multiplier(char *unit_name)
{
	GList *units = global_units_glist;
	GList *mult = global_multiplier_glist;
	int i = 1;

	while(units)
	{
		if(!strcmp((char *)units->data, unit_name))
			return  *((float *)g_list_nth(global_multiplier_glist, i - 1)->data);
		units = units->next;
		mult = mult->next;
		i++;
	}
	return 0;

}

static void set_new_combo_value (GtkWidget *combo UNUSED, struct change_units_window *cuw)
{
	char mult[MAX_UNIT_MULTIPLIER_LENGTH + 6];
	float multiplier;

	if( (multiplier = get_multiplier(xstrdup(gtk_entry_get_text(GTK_ENTRY((GTK_COMBO (cuw->combo))->entry)))) ))
		sprintf(mult, " Mult.: %f", multiplier);
	gtk_label_set_text(GTK_LABEL (cuw->multiplier), mult);
}

static void combo_value_changed(GtkWidget *combo UNUSED, struct change_units_window *cuw)
{
	char mult[MAX_UNIT_MULTIPLIER_LENGTH + 6];
	float multiplier;

	if(!(multiplier = get_multiplier(xstrdup(gtk_entry_get_text(GTK_ENTRY((GTK_COMBO (cuw->combo))->entry))))))
		sprintf(mult, " Mult.: ");
	else
		sprintf(mult, " Mult.: %f", multiplier);
	gtk_label_set_text(GTK_LABEL (cuw->multiplier), mult);
}

static gboolean control_text(char *unit_name)
{
	char c;
	int i;
	int len = strlen(unit_name);

	if((len > MAX_UNIT_NAME_LENGTH) || (len < 1))
		return 0;
	if(!(isalpha(unit_name[0])))
		return 0;
	for(i = 1; i < len; i++)
	{
		c = unit_name[i];
		if(!isdigit(c) && !isalpha(c) && (c != '_'))
			return 0;
	}

	return 1;
}

static float control_float_number(char *unit_name)
{
	double retval;

	if (!(unit_name[0]))
		return -1;
	if(!(retval = strtod(unit_name, 0)))
		return -1;
	if(retval < 0)
		return -1;
	return (float) retval;
}

static int unit_already_exists(const char *name)
{
	GList *ptr = global_units_glist;
	int i = 1;

	while(ptr)
	{
		if(!strcmp((char *)ptr->data, name))
			return i;
		ptr = ptr->next;
		i++;
	}
	return 0;
}

static void button_apply_pressed_global(GtkWidget *widget, struct change_units_window *cuw)
{
	char *name, *multiplier;
	float fmultiplier = 0;
	struct vrr_options_window *vow = cuw->vow;
	char tmp[65];

	if(!(cuw->act_id))	// newly defined units
	{
		name = str_trim(xstrdup(gtk_entry_get_text(GTK_ENTRY (cuw->entry_name))));
		multiplier = str_trim(xstrdup(gtk_entry_get_text(GTK_ENTRY (cuw->entry_multiplier))));
		if(!control_text(name))
		{
			create_err_window(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW), "Error",
				"Name of new unit isn't defined correctly. Use just letters, \ndigits or '_', start with letter. Max. length of number is %d characters.", " OK     ", MAX_UNIT_NAME_LENGTH);
			goto free_units_data;
		}
		if((fmultiplier = control_float_number(multiplier)) <= 0)
		{
			create_err_window(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW), "Error",
				"Multiplier of new unit isn't defined correctly. String must \ncontain at least 1 digit. Use just digits or decimal point ('.'), \nnumber must start with digit. Max. length of name is 16 characters.", " OK     ");
			goto free_units_data;
		}
		if(unit_already_exists(name))
		{
			create_err_window(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW), "Error", "Unit with this name (%s) already exists.", " OK     ", name);
			goto free_units_data;
		}
		global_units_glist = g_list_append(global_units_glist, name);
		global_multiplier_glist = g_list_append(global_multiplier_glist, new_float(fmultiplier));

		gtk_combo_set_popdown_strings(GTK_COMBO(cuw->combo), global_units_glist);

		strcpy(vow->unit.name, name);
		vow->unit.multiplier = fmultiplier;

		xfree(multiplier);
	}
	else	// choose units from list
	{
		int id;
		name = (char *)gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(cuw->combo)->entry));

		id = unit_already_exists(name);
		if(!id)
		{
			create_err_window(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW), "Error", "You can't choose a unit which doesn't exist (use list of defined unist).", " OK     ");
			goto free_units_data;
		}
		strcpy(vow->unit.name, name);
		vow->unit.multiplier = *((float *)g_list_nth(global_multiplier_glist, id - 1)->data);
	}

	gtk_label_set_text(GTK_LABEL(vow->label_units_name), name);
	sprintf(tmp, "Mult.: %5.3f", vow->unit.multiplier);
	gtk_label_set_text(GTK_LABEL(vow->label_multiplier), tmp);

	free_units_data:
	gtk_widget_destroy(cuw->window);
	xfree(cuw);
}

static void button_apply_pressed(GtkWidget *widget, struct change_units_window *cuw)
{
	char *name, *multiplier;
	float fmultiplier = 0;
	struct units *unit = &(cuw->lw->units_data);

	DBG("Changing units\n");
	if(!(cuw->act_id))	// newly defined units
	{
		name = str_trim(xstrdup(gtk_entry_get_text(GTK_ENTRY (cuw->entry_name))));
		multiplier = str_trim(xstrdup(gtk_entry_get_text(GTK_ENTRY (cuw->entry_multiplier))));
		if(!control_text(name))
		{
			create_err_window(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW), "Error",
				"Name of new unit isn't defined correctly. Use just letters, \ndigits or '_', start with letter. Max. length of number is %d characters.", " OK     ", MAX_UNIT_NAME_LENGTH);
			goto free_units_data;
		}
		if((fmultiplier = control_float_number(multiplier)) <= 0)
		{
			create_err_window(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW), "Error",
				"Multiplier of new unit isn't defined correctly. String must \ncontain at least 1 digit. Use just digits or decimal point ('.'), \nnumber must start with digit. Max. length of name is 16 characters.", " OK     ");
			goto free_units_data;
		}
		if(unit_already_exists(name))
		{
			create_err_window(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW), "Error", "Unit with this name (%s) already exists.", " OK     ", name);
			goto free_units_data;
		}
		global_units_glist = g_list_append(global_units_glist, name);
		global_multiplier_glist = g_list_append(global_multiplier_glist, new_float(fmultiplier));

		gtk_combo_set_popdown_strings(GTK_COMBO(cuw->combo), global_units_glist);

		strcpy(unit->name, name);
		unit->multiplier = fmultiplier;

		g_print("name : %s, multiplier: %f\n", name, unit->multiplier);
		xfree(multiplier);
	}
	else	// choose units from list
	{
		int id;
		name = (char *)gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(cuw->combo)->entry));

		id = unit_already_exists(name);
		if(!id)
		{
			create_err_window(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW), "Error", "You can't choose a unit which doesn't exist (use list of defined unist).", " OK     ");
			goto free_units_data;
		}
		strcpy(unit->name, name);
		unit->multiplier = *((float *)g_list_nth(global_multiplier_glist, id - 1)->data);
		g_print("name : %s, multiplier: %f\n", name, unit->multiplier);
	}

	free_units_data:
	gtk_widget_destroy(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW));
	rulers_set_resolution(cuw->lw, unit->multiplier);
	xfree(cuw);
}

static void button_cancel_pressed(GtkWidget *widget, gpointer data)
{
	if(data)
		xfree(data);
	gtk_widget_destroy(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW));
}

static void set_units(struct units *unit, char *name, float mult)
{
	unit->multiplier = mult;
	strcpy(unit->name, name);
}

static void vow_set_unit(struct vrr_options_window *vow UNUSED, char *name, float mult)
{
	char smult[MAX_UNIT_MULTIPLIER_LENGTH + 6];
	sprintf(smult, "Mult.: %f", mult);
	gtk_label_set_text(GTK_LABEL(vow->label_units_name), name);
	gtk_label_set_text(GTK_LABEL(vow->label_multiplier), smult);
}

static void button_delete_pressed(GtkWidget *widget UNUSED, struct change_units_window *cuw)
{
	// TODO dopsat kontroly, zda se jednotka nepouziva, pokud jo, warning
	char *name;
	GList *units = global_units_glist;
	GList *mult = global_multiplier_glist;
	struct vrr_options_window *vow = cuw->vow;

	name = (char *)gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(cuw->combo)->entry));

	if(!(strcmp(name, "mm")))
	{
		create_err_window(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW), "Error", "You can't delete units 'mm'.", " OK     ");
		return;
	}
	if(!strcmp(name, global_units.name))
	{
		set_units(&global_units, "mm", 1);
		set_units(&(vow->unit), "mm", 1);
		vow_set_unit(vow, "mm", 1);
	}
	while(units)
	{
		if(!strcmp((char *)units->data, name))
		{
			global_units_glist = g_list_remove_link(global_units_glist, units);
			global_multiplier_glist = g_list_remove_link(global_multiplier_glist, mult);
			gtk_combo_set_popdown_strings(GTK_COMBO(cuw->combo), global_units_glist);
			return;
		}
		units = units->next;
		mult = mult->next;
	}

	create_err_window(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW), "Error", "You can't delete a unit which doesn't exist (use list of defined unist).", " OK     ");
}


static void on_multiplier_set(GtkWidget *widget, struct change_units_window *cuw)
{
	char *multiplier;
	float fmultiplier;
	char name[MAX_UNIT_MULTIPLIER_LENGTH + 6];
	GList *units = global_units_glist;
	GList *mult = global_multiplier_glist;


	multiplier = str_trim(xstrdup(gtk_entry_get_text(GTK_ENTRY (cuw->edit_entry))));

	if((fmultiplier = control_float_number(multiplier)) <= 0)
	{
		create_err_window(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW), "Error", "Multiplier of new unit isn't defined correctly. String must \ncontain at least 1 digit. Use just digits or decimal point ('.'), \nnumber must start with digit. Max. length of name is 16 characters.", " OK     ");
		return;
	}
	sprintf(name, " Mult.: %f", fmultiplier);
	gtk_label_set_text(GTK_LABEL(cuw->multiplier), name);
	gtk_widget_destroy(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW));

	strcpy(name,(char *)gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(cuw->combo)->entry)));
	if(!strcmp(name, global_units.name))
		set_units(&global_units, name, fmultiplier);

	while(units)
	{
		if(!strcmp((char *)units->data, name))
		{
			*((float *)mult->data) = fmultiplier;
			return;
		}
		units = units->next;
		mult = mult->next;
	}
}

static void button_edit_pressed(GtkWidget *widget UNUSED, struct change_units_window *cuw)
{
	GtkWidget *win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	GtkWidget *table = gtk_table_new(2, 2, FALSE);
	GtkWidget *label = gtk_label_new("New multiplier: ");
	cuw->edit_entry = gtk_entry_new();
	GtkWidget *button_apply, *button_cancel;
	char name[MAX_UNIT_MULTIPLIER_LENGTH + 6];

	strcpy(name, (char *)gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(cuw->combo)->entry)));
	if(!(strcmp(name, "mm")))
	{
		create_err_window(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW), "Error", "You can't change multiplier of unit 'mm'.", " OK     ");
		return;
	}

	gtk_window_set_title(GTK_WINDOW(win), "Multiplier editation");
	gtk_container_set_border_width (GTK_CONTAINER (win), 5);
	gtk_window_set_position(GTK_WINDOW(win), GTK_WIN_POS_MOUSE);

	button_apply = gtk_button_new_from_stock("gtk-apply");
	button_cancel = gtk_button_new_from_stock("gtk-cancel");

	gtk_container_add(GTK_CONTAINER(win), table);
	gtk_table_set_row_spacings(GTK_TABLE(table), 3);
	gtk_table_set_col_spacings(GTK_TABLE(table), 3);

	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_table_attach(GTK_TABLE(table), cuw->edit_entry, 1, 2, 0, 1,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_table_attach(GTK_TABLE(table), button_apply, 0, 1, 1, 2,
			(GtkAttachOptions) (0),
			(GtkAttachOptions) (0), 0, 0);
	gtk_table_attach(GTK_TABLE(table), button_cancel, 1, 2, 1, 2,
			(GtkAttachOptions) (0),
			(GtkAttachOptions) (0), 0, 0);

	g_signal_connect(G_OBJECT(button_cancel), "clicked",
			G_CALLBACK(button_cancel_pressed), 0);
	g_signal_connect(G_OBJECT(button_apply), "clicked",
			G_CALLBACK(on_multiplier_set), cuw);

	gtk_widget_show_all(win);
}

static void toggle_radio(GtkContainer *container UNUSED,
		GtkWidget *widget UNUSED,
		GtkToggleButton *button)
{
	gtk_toggle_button_set_active(button, TRUE);

}

static void on_radio_combo_toggled(GtkToggleButton *togglebutton UNUSED, struct change_units_window *cuw)
{
	gtk_widget_set_sensitive(cuw->label_name, FALSE);
	gtk_widget_set_sensitive(cuw->label_multiplier, FALSE);
	gtk_widget_set_sensitive(cuw->entry_name, FALSE);
	gtk_widget_set_sensitive(cuw->entry_multiplier, FALSE);
	if(cuw->button_delete)
	{
		gtk_widget_set_sensitive(cuw->button_delete, TRUE);
		gtk_widget_set_sensitive(cuw->button_edit, TRUE);
	}
	cuw->act_id = 1;
}

static void on_radio_define_new_toggled(GtkToggleButton *togglebutton UNUSED, struct change_units_window *cuw)
{
	gtk_widget_set_sensitive(cuw->label_name, TRUE);
	gtk_widget_set_sensitive(cuw->label_multiplier, TRUE);
	gtk_widget_set_sensitive(cuw->entry_name, TRUE);
	gtk_widget_set_sensitive(cuw->entry_multiplier, TRUE);
	if(cuw->button_delete)
	{
		gtk_widget_set_sensitive(cuw->button_delete, FALSE);
		gtk_widget_set_sensitive(cuw->button_edit, FALSE);
	}

	cuw->act_id = 0;
}

struct change_units_window *create_units_window(struct view *lw)
{
	GtkWidget *table;
	GSList *radio_units_group = NULL;
	GtkWidget *radio_combo, *radio_define_new;
	GtkWidget *button_ok, *button_cancel;
	GtkWidget *hbox;
	char mult[MAX_UNIT_MULTIPLIER_LENGTH+6];
	struct change_units_window *cuw = xmalloc(sizeof(struct change_units_window));

	cuw->button_delete = cuw->button_edit = NULL;
	cuw->lw = lw;
	cuw->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_container_set_border_width (GTK_CONTAINER (cuw->window), 2);
	gtk_window_set_title (GTK_WINDOW (cuw->window), "Units");

	table = gtk_table_new(5, 2, FALSE);
	gtk_container_add(GTK_CONTAINER(cuw->window), table);
	gtk_container_set_border_width(GTK_CONTAINER(table), 7);
	gtk_table_set_row_spacings(GTK_TABLE(table), 5);
	gtk_table_set_col_spacings(GTK_TABLE(table), 4);

	cuw->combo = gtk_combo_new();
	gtk_combo_set_value_in_list(GTK_COMBO(cuw->combo), FALSE, FALSE);
	gtk_combo_set_use_arrows(GTK_COMBO(cuw->combo), TRUE);
	gtk_combo_disable_activate(GTK_COMBO(cuw->combo));
	gtk_combo_set_popdown_strings(GTK_COMBO(cuw->combo), global_units_glist);

	hbox = gtk_hbox_new(FALSE, 0);
	radio_combo = gtk_radio_button_new(0);
	gtk_box_pack_start(GTK_BOX(hbox), radio_combo, FALSE, FALSE, 0);

	gtk_box_pack_start(GTK_BOX(hbox), cuw->combo, 0, 0, 0);

	sprintf(mult, " Mult.: %f", lw->units_data.multiplier);
	cuw->multiplier = gtk_label_new_with_mnemonic(mult);
	gtk_box_pack_start(GTK_BOX(hbox), cuw->multiplier, 0, 0, 0);
	g_signal_connect(G_OBJECT (cuw->combo), "set-focus-child",
			G_CALLBACK(toggle_radio), radio_combo);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 2, 0, 1,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_radio_button_set_group (GTK_RADIO_BUTTON (radio_combo), radio_units_group);
	radio_units_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radio_combo));

	radio_define_new = gtk_radio_button_new_with_mnemonic(0, "define new");
	gtk_table_attach(GTK_TABLE(table), radio_define_new, 0, 1, 1, 2,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_radio_button_set_group(GTK_RADIO_BUTTON(radio_define_new), radio_units_group);
	radio_units_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radio_define_new));

	cuw->entry_name = gtk_entry_new();
	gtk_table_attach(GTK_TABLE(table), cuw->entry_name, 1, 2, 2, 3,
			(GtkAttachOptions) (GTK_EXPAND |GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(cuw->combo)->entry), lw->units_data.name);

	cuw->entry_multiplier = gtk_entry_new();
	gtk_table_attach(GTK_TABLE (table), cuw->entry_multiplier,
			1, 2, 3, 4, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);

	button_cancel = gtk_button_new_from_stock("gtk-cancel");
	gtk_table_attach(GTK_TABLE(table), button_cancel, 1, 2, 4, 5,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 34, 0);
	g_signal_connect(G_OBJECT (button_cancel), "clicked",
			G_CALLBACK(button_cancel_pressed), cuw);


	button_ok = gtk_button_new_from_stock("gtk-apply");
	gtk_table_attach(GTK_TABLE(table), button_ok, 0, 1, 4, 5,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 14, 9);
	g_signal_connect(G_OBJECT (button_ok), "clicked",
			G_CALLBACK(button_apply_pressed), cuw);

	cuw->label_multiplier = gtk_label_new ("multiplier (against mm):");
	gtk_table_attach (GTK_TABLE(table), cuw->label_multiplier, 0, 1, 3, 4,
			(GtkAttachOptions)(GTK_FILL),
			(GtkAttachOptions) (0), 7, 2);
	gtk_label_set_justify(GTK_LABEL(cuw->label_multiplier), GTK_JUSTIFY_RIGHT);

	cuw->label_name = gtk_label_new("name:");
	gtk_table_attach(GTK_TABLE (table), cuw->label_name, 0, 1, 2, 3,
		(GtkAttachOptions) (GTK_FILL),
		(GtkAttachOptions) (0), 7, 0);
	gtk_label_set_justify(GTK_LABEL(cuw->label_name), GTK_JUSTIFY_RIGHT);

	g_signal_connect(G_OBJECT (radio_combo), "toggled",
			G_CALLBACK(on_radio_combo_toggled), cuw);
	g_signal_connect(G_OBJECT (radio_define_new), "toggled",
			G_CALLBACK(on_radio_define_new_toggled), cuw);
	g_signal_connect(G_OBJECT (cuw->window), "destroy",
			G_CALLBACK(destroy), 0);
	g_signal_connect(G_OBJECT((GTK_COMBO (cuw->combo))->entry), "activate",
			G_CALLBACK(combo_value_changed), cuw);
	g_signal_connect(G_OBJECT((GTK_COMBO (cuw->combo))->entry), "changed",
			G_CALLBACK(set_new_combo_value), cuw);

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (radio_define_new), TRUE);
	on_radio_define_new_toggled(0, cuw);

	gtk_widget_show_all(cuw->window);

	return cuw;
}

struct change_units_window *global_units_change(struct vrr_options_window *vow)
{
	GtkWidget *table;
	GSList *radio_units_group = NULL;
	GtkWidget *radio_combo, *radio_define_new;
	GtkWidget *button_ok, *button_cancel;
	GtkWidget *hbox, *button_hbox;
	char mult[MAX_UNIT_MULTIPLIER_LENGTH+6];
	struct change_units_window *cuw = xmalloc(sizeof(struct change_units_window));

	cuw->vow = vow;
	cuw->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_container_set_border_width (GTK_CONTAINER (cuw->window), 2);
	gtk_window_set_title (GTK_WINDOW (cuw->window), "Units");

	table = gtk_table_new(6, 2, FALSE);
	gtk_container_add(GTK_CONTAINER(cuw->window), table);
	gtk_container_set_border_width(GTK_CONTAINER(table), 6);
	gtk_table_set_row_spacings(GTK_TABLE(table), 5);
	gtk_table_set_col_spacings(GTK_TABLE(table), 4);

	cuw->combo = gtk_combo_new();
	gtk_combo_set_value_in_list(GTK_COMBO(cuw->combo), FALSE, FALSE);
	gtk_combo_set_use_arrows(GTK_COMBO(cuw->combo), TRUE);
	gtk_combo_disable_activate(GTK_COMBO(cuw->combo));
	gtk_combo_set_popdown_strings(GTK_COMBO(cuw->combo), global_units_glist);

	hbox = gtk_hbox_new(FALSE, 0);
	radio_combo = gtk_radio_button_new(0);
	gtk_box_pack_start(GTK_BOX(hbox), radio_combo, FALSE, FALSE, 0);

	gtk_box_pack_start(GTK_BOX(hbox), cuw->combo, 0, 0, 0);

	sprintf(mult, " Mult.: %f", vow->unit.multiplier);
	cuw->multiplier = gtk_label_new_with_mnemonic(mult);
	gtk_box_pack_start(GTK_BOX(hbox), cuw->multiplier, 0, 0, 0);
	g_signal_connect(G_OBJECT (cuw->combo), "set-focus-child",
			G_CALLBACK(toggle_radio), radio_combo);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 2, 0, 1,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_radio_button_set_group (GTK_RADIO_BUTTON (radio_combo), radio_units_group);
	radio_units_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radio_combo));

	button_hbox = gtk_hbox_new(TRUE, 3);
	gtk_table_attach(GTK_TABLE(table), button_hbox, 0,1,1,2,
			(GtkAttachOptions) (0),
			(GtkAttachOptions) (0), 0, 0);

	cuw->button_delete = gtk_button_new_with_label("Delete");
	gtk_box_pack_start(GTK_BOX(button_hbox), cuw->button_delete, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(cuw->button_delete), "clicked",
			G_CALLBACK(button_delete_pressed), cuw);

	cuw->button_edit = gtk_button_new_with_label(" Edit ");
	gtk_box_pack_start(GTK_BOX(button_hbox), cuw->button_edit, 0, 0, 0);
	g_signal_connect(G_OBJECT(cuw->button_edit), "clicked",
			G_CALLBACK(button_edit_pressed), cuw);

	radio_define_new = gtk_radio_button_new_with_mnemonic(0, "define new");
	gtk_table_attach(GTK_TABLE(table), radio_define_new, 0, 1, 2, 3,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_radio_button_set_group(GTK_RADIO_BUTTON(radio_define_new), radio_units_group);
	radio_units_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radio_define_new));

	cuw->entry_name = gtk_entry_new();
	gtk_table_attach(GTK_TABLE(table), cuw->entry_name, 1, 2, 3, 4,
			(GtkAttachOptions) (GTK_EXPAND |GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(cuw->combo)->entry), vow->unit.name);

	cuw->entry_multiplier = gtk_entry_new();
	gtk_table_attach(GTK_TABLE (table), cuw->entry_multiplier,
			1, 2, 4, 5, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);

	button_cancel = gtk_button_new_from_stock("gtk-cancel");
	gtk_table_attach(GTK_TABLE(table), button_cancel, 1, 2, 5, 6,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 34, 0);
	g_signal_connect(G_OBJECT (button_cancel), "clicked",
			G_CALLBACK(button_cancel_pressed), cuw);


	button_ok = gtk_button_new_from_stock("gtk-apply");
	gtk_table_attach(GTK_TABLE(table), button_ok, 0, 1, 5, 6,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 14, 9);
	g_signal_connect(G_OBJECT (button_ok), "clicked",
			G_CALLBACK(button_apply_pressed_global), cuw);

	cuw->label_multiplier = gtk_label_new ("multiplier (against mm):");
	gtk_table_attach (GTK_TABLE(table), cuw->label_multiplier, 0, 1, 4, 5,
			(GtkAttachOptions)(GTK_FILL),
			(GtkAttachOptions) (0), 7, 2);
	gtk_label_set_justify(GTK_LABEL(cuw->label_multiplier), GTK_JUSTIFY_RIGHT);

	cuw->label_name = gtk_label_new("name:");
	gtk_table_attach(GTK_TABLE (table), cuw->label_name, 0, 1, 3, 4,
		(GtkAttachOptions) (GTK_FILL),
		(GtkAttachOptions) (0), 7, 0);
	gtk_label_set_justify(GTK_LABEL(cuw->label_name), GTK_JUSTIFY_RIGHT);

	g_signal_connect(G_OBJECT (radio_combo), "toggled",
			G_CALLBACK(on_radio_combo_toggled), cuw);
	g_signal_connect(G_OBJECT (radio_define_new), "toggled",
			G_CALLBACK(on_radio_define_new_toggled), cuw);
	g_signal_connect(G_OBJECT (cuw->window), "destroy",
			G_CALLBACK(destroy), 0);
	g_signal_connect(G_OBJECT((GTK_COMBO (cuw->combo))->entry), "activate",
			G_CALLBACK(combo_value_changed), cuw);
	g_signal_connect(G_OBJECT((GTK_COMBO (cuw->combo))->entry), "changed",
			G_CALLBACK(set_new_combo_value), cuw);

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (radio_define_new), TRUE);
	on_radio_define_new_toggled(0, cuw);

	gtk_widget_show_all(cuw->window);

	return cuw;
}

