#ifndef _GUI_ATTRIBUTES_H_
#define _GUI_ATTRIBUTES_H_

/**
 \file changeunits.h
 \brief Editor attributes (units, ...).
*/

#include "lib/lib.h"
#include <gtk/gtk.h>
#include "gui/main.h"


/**
 This structure "takes care" of units (unit_window).
*/
struct change_units_window{
	GtkWidget *combo;	/**< combo box */
	GtkWidget *multiplier;		/** label showing actual multiplier */
	GtkWidget *entry_name;		/**< entry (name of new unit) */
	GtkWidget *entry_multiplier;	/**< entry (name of multiplier) */
	GtkWidget *label_name;		/** label "name" */
	GtkWidget *label_multiplier;	/** label "multiplier" */
	GtkWidget *window;		/** pointer to window */
	struct view *lw;	/**< parent local window */
	struct vrr_options_window *vow;	/**< parent vrr options window */
	int act_id;		/**< id of selected unit. if act_id == 0, it's newly defined unit */
	GtkWidget *button_delete;	/**< button to delete unit */
	GtkWidget *button_edit;		/**< button to edit unit */
	GtkWidget *edit_entry;		/**< Edit entry */
};

/**
 returns multiplier of unit with name unit_name
*/
float get_multiplier(char *unit_name);

/**
 Creates window where localwindow units can be changed.
 \returns pointer to created window
*/
struct change_units_window *create_units_window(struct view *lw);

/**
 Window which changes global units (insert, edit, delete).
 */
struct change_units_window *global_units_change(struct vrr_options_window *vow);

/**
 Sets new resolution - according to chosen / newly defined unit - to both
 rulers (horizontal and vertical).
 \param lw pointer to structere view (parent view window)
 \param resolution multiplier of new unit (how many pixels creates 1 new unit)
*/
void rulers_set_resolution(struct view *lw, float resolution);


#endif
