#undef LOCAL_DEBUG

#include "lib/lib.h"
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "kernel/kernel.h"

#include "gui/cmdmgr.h"
#include "gui/main.h"
#include "gui/properties.h"

#include "build/path.h"

#define COMBINE_STATE(c,m) ((c) ? (cmd->state | (m)) : (cmd->state & ~(m)))

static char vrrvw[] = "vrrview";
static char vrrgroup[] = "vrrgroup";

static void on_command_activated( GtkWidget * widget, struct cmd * command );
static int stamp;
static void change_stamp(void){ stamp = (stamp + 1) % 256 + 1; }
static int sel_req_ok( uns mask, uns cmask, int cnt, int ccnt );
static int context_matches( struct cmd * cmd );
static void on_context_dump( struct cmd * cmd, void * subj );
static void load_icons_in_category( struct cmd_category * ctg );

//static int menu_big_bang( struct cmd * command );

static int sync_in_progress;
static void fill_menu_with_category( GtkWidget * menu, struct cmd_category * ctg, unsigned int location, struct cmddata cd );
static void sync_menu_item( struct cmd * cmd, GtkWidget * command, unsigned int location );
static void sync_menu_by_category( GtkWidget * menuitem, struct cmd_category * ctg, unsigned int location );
static void fill_toolbar_with_category( GtkBox * tb, struct cmd_category * ctg, unsigned int location, struct cmddata cd );
static void sync_toolbar_by_category( struct cmd_category * ctg, unsigned int location, GList ** list );
static void sync_tool_item( struct cmd * cmd, GtkWidget * command, unsigned int location, struct cmd_category * ctg );
static void sync_cmd_item( struct cmd * cmd, GtkWidget * command, unsigned int location, unsigned int hide );
static void calculate_selection(void);

/** Type bit masks. */
enum obj_vtm
{
	VTM_DOCUMENT		= 1 << OT_DOCUMENT,
	VTM_TLO			= 1 << OT_TLO,
	VTM_OBJ			= ( VTM_DOCUMENT | VTM_TLO ),

	VTM_GROUP		= 1 << (GOT_GROUP + OT_MAX),
	VTM_POINT		= 1 << (GOT_POINT + OT_MAX),
	VTM_SEGMENT		= 1 << (GOT_SEGMENT + OT_MAX),
	VTM_BEZIER		= 1 << (GOT_BEZIER + OT_MAX),
	VTM_ELARC		= 1 << (GOT_ELARC + OT_MAX),
	VTM_PARPOINT		= 1 << (GOT_PARAMETRIC_POINT + OT_MAX),
	VTM_INTERSECTION	= 1 << (GOT_INTERSECTION + OT_MAX),
	VTM_TEXT		= 1 << (GOT_TEXT + OT_MAX),
	VTM_TEX_TEXT		= 1 << (GOT_TEX_TEXT + OT_MAX),
	VTM_DECPOINT		= 1 << (GOT_DECORATOR_POINT + OT_MAX),
	VTM_GO			= ( VTM_GROUP | VTM_POINT | VTM_SEGMENT | VTM_BEZIER | VTM_ELARC | VTM_PARPOINT |
					VTM_INTERSECTION | VTM_TEXT | VTM_TEX_TEXT | VTM_DECPOINT ),

	VTM_ANYTHING		= ( VTM_OBJ | VTM_GO ),
	VTM_VIEW		= 1 << ( OT_MAX + GOT_MAX + WT_VIEW ),
	VTM_UNDO_HISTORY	= 1 << ( OT_MAX + GOT_MAX + WT_UNDO_HISTORY ),
	VTM_UNIVERSE_BROWSER	= 1 << ( OT_MAX + GOT_MAX + WT_UNIVERSE_BROWSER ),

	VTM_CT_DOC		= 1 << ( OT_MAX + GOT_MAX + WT_UNIVERSE_BROWSER + 1 ),
	VTM_CT_TLO		= 1 << ( OT_MAX + GOT_MAX + WT_UNIVERSE_BROWSER + 2 ),
	VTM_CT_GRP		= 1 << ( OT_MAX + GOT_MAX + WT_UNIVERSE_BROWSER + 3 ),
};

/* File category */
static struct cmd file_exit_cmd = { "Exit", "Terminate the program", GDK_Q, GDK_CONTROL_MASK, VTM_ANYTHING, 0,
				VCL_EVERYWHERE | VCC_META_SEL, VCS_VIS_EN, LIB_PATH"exit.png", NULL, NULL, &on_file_exit, NULL, NULL };
static struct cmd file_close_cmd = { "Close", "Close the file", 0, 0, VTM_DOCUMENT | VTM_CT_DOC, -1,
				VCL_LOCAL_MENUS | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, &on_file_close, NULL, &file_exit_cmd };
static struct cmd file_save_as_cmd = { "Save as...", "Save the file as..." , 0, 0, VTM_DOCUMENT | VTM_CT_DOC, 1,
				VCL_UB_MENU | VCL_VIEW_MENU | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, &on_file_save_as, NULL, &file_close_cmd };
static struct cmd file_save_cmd = { "Save", "Save the file" , GDK_S, GDK_CONTROL_MASK, VTM_DOCUMENT | VTM_CT_DOC, 1,
				VCL_UB_MENU | VCL_VIEW_MENU | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, &on_file_save, NULL, &file_save_as_cmd };
static struct cmd file_open_cmd = { "Open", "Open a file" , GDK_O, GDK_CONTROL_MASK, VTM_ANYTHING, 0,
				VCL_EVERYWHERE | VCC_META_SEL, VCS_VIS_EN, LIB_PATH"open.png", NULL, NULL, &on_file_open, NULL, &file_save_cmd };
static struct cmd file_new_cmd =  { "New", "Create a new file", GDK_N, GDK_CONTROL_MASK, VTM_ANYTHING, 0,
				VCL_EVERYWHERE | VCC_META_SEL, VCS_VIS_EN, LIB_PATH"new.png", NULL, NULL, &on_file_new, NULL, &file_open_cmd };

/* Import/Export category */
static struct cmd page_export_svg = {"Export page to SVG", "Export page to SVG", 0, 0, VTM_TLO | VTM_CT_TLO, 1,
				VCL_LOCAL_MENUS | VCC_META_SEL, VCS_VIS_EN, NULL, NULL , NULL, &on_export_svg, NULL, NULL };
static struct cmd page_export_eps = { "Export page to EPS", "Export page to EPS", 0, 0, VTM_TLO | VTM_CT_TLO, 1, VCL_LOCAL_MENUS | VCC_META_SEL,
				VCS_VIS_EN, NULL, NULL, NULL, &on_export_eps, NULL, &page_export_svg };
static struct cmd file_export_ps = { "Export document to PS", "Export document to PS", 0, 0, VTM_DOCUMENT | VTM_CT_DOC, 1,
				VCL_LOCAL_MENUS | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, &on_export_ps, NULL, &page_export_eps };
static struct cmd file_export_pdf = { "Export document to PDF", "Export document to PDF", 0, 0, VTM_DOCUMENT | VTM_CT_DOC, 1,
				VCL_LOCAL_MENUS | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, &on_export_pdf, NULL, &file_export_ps };

/* Page category */
static struct cmd import_ipe5_cmd = { "Import Ipe 5", "Import an Ipe 5 file", 0, 0, VTM_TLO | VTM_CT_TLO, 1,
				VCL_ALL_MENUS | VCC_META_SEL, VCS_VIS_EN, 0, NULL, NULL, &on_import_ipe5, NULL, NULL };
static struct cmd import_svg_cmd = { "Import SVG", "Import a SVG file", GDK_I, GDK_CONTROL_MASK, VTM_TLO | VTM_CT_TLO, 1,
				VCL_ALL_MENUS | VCC_META_SEL, VCS_VIS_EN, 0, NULL, NULL, &on_svg_file_import, NULL, &import_ipe5_cmd };
static struct cmd page_close_cmd = { "Delete", "Delete the page", 0, 0, VTM_TLO | VTM_CT_TLO, -1,
				VCL_LOCAL_MENUS | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, &on_file_close_page, NULL, &import_svg_cmd };
static struct cmd page_new_cmd =  { "New", "Create a new page", 0, 0, VTM_DOCUMENT | VTM_CT_DOC, -1,
				VCL_UB_MENU | VCL_VIEW_MENU | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, &on_file_new_page,
				NULL, &page_close_cmd };
static struct cmd new_view_cmd = { "Open page/group in new view", "Open new view", 0, 0, VTM_GROUP | VTM_CT_GRP, -1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, NULL, &on_group_new_view, NULL, &page_new_cmd };

/* Edit category */

static int paste_cmd_enabled( struct cmd * cmd UNUSED, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( cmd->state & VCS_ENABLED && context.tlo != tlo_clipboard &&
	group_get_first_selected_go(tlo_get_tlg(tlo_clipboard)), VCS_ENABLED ); }

static int cutcopy_enabled( struct cmd * cmd UNUSED, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( cmd->state & VCS_ENABLED && context.tlo != tlo_clipboard, VCS_ENABLED ); }

static int redo_enabled( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( tlo_universum->undo_present != trans_get_last_gui(tlo_universum), VCS_ENABLED ); }

static int undo_enabled( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( tlo_universum->undo_present != &tlo_universum->undo_start, VCS_ENABLED ); }

static int local_redo_enabled( struct cmd * cmd, GtkWidget * widget )
{
	void * v = g_object_get_data( G_OBJECT(widget), vrrgroup );
	if ( v )
	{
		struct obj_tlo * tlo = go_get_tlo(void2go(v));
		return COMBINE_STATE( tlo->undo_present != trans_get_last_gui(tlo), VCS_ENABLED );
	}
	else return COMBINE_STATE( context.tlo && context.tlo->undo_present !=
		trans_get_last_gui(context.tlo), VCS_ENABLED );
}
static int local_undo_enabled( struct cmd * cmd, GtkWidget * widget )
{
	void * v = g_object_get_data( G_OBJECT(widget), vrrgroup );
	if ( v )
	{
		struct obj_tlo * tlo = go_get_tlo(void2go(v));
		return COMBINE_STATE( tlo->undo_present != &tlo->undo_start, VCS_ENABLED );
	}
	else return COMBINE_STATE( context.tlo && context.tlo->undo_present != &context.tlo->undo_start, VCS_ENABLED );
}
static int history_jump_enabled( struct cmd * cmd, GtkWidget * widget )
{
	void * v = g_object_get_data( G_OBJECT(widget), vrrgroup );
	struct obj_tlo * tlo = NULL;
	if ( v ) tlo = go_get_tlo(void2go(v));
	return COMBINE_STATE( ( tlo || (tlo = tlo_universum) ) && trans_get_first_gui(tlo), VCS_ENABLED ); //I really mean this! :)
}
#define GET_CONTEXT_GO(_go) \
	struct go * _go = NULL; \
	if ( context.o && context.o->kind == T_GO && go_get_group(o2go(context.o)) == context.group ) _go = o2go(context.o); \
	else if (context.group) go = group_get_first_selected_go(context.group)

static int can_move_up( struct cmd * cmd, GtkWidget * widget UNUSED )
{
	GET_CONTEXT_GO(go);
	return COMBINE_STATE( go && group_get_prev_go(context.group, go), VCS_ENABLED );
}
static int can_move_down( struct cmd * cmd, GtkWidget * widget UNUSED )
{
	GET_CONTEXT_GO(go);
	return COMBINE_STATE( go && group_get_next_go(context.group, go), VCS_ENABLED );
}
static int can_move_front( struct cmd * cmd, GtkWidget * widget UNUSED )
{
	GET_CONTEXT_GO(go);
	return COMBINE_STATE( go && go != group_get_first_go(context.group), VCS_ENABLED );
}
static int can_move_back( struct cmd * cmd, GtkWidget * widget UNUSED )
{
	GET_CONTEXT_GO(go);
	return COMBINE_STATE( go && go != group_get_last_go(context.group), VCS_ENABLED );
}
static struct cmd edit_select_none =  { "Clear selection", "Clear selection", GDK_A, GDK_SHIFT_MASK | GDK_CONTROL_MASK, VTM_CT_GRP, 1,
				VCL_LOCAL_MENUS | VCC_META_SEL,	VCS_VIS_EN, NULL,
				NULL, NULL, &on_group_select_none, NULL, NULL };
static struct cmd edit_select_all =  { "Select all", "Select all GO", GDK_A, GDK_CONTROL_MASK, VTM_CT_GRP, 1,
				VCL_LOCAL_MENUS | VCC_META_SEL,	VCS_VIS_EN, NULL, NULL,
				NULL, &on_group_select_all, NULL, &edit_select_none };
static struct cmd edit_local_redo_cmd =  { "Redo", "Redo one local action", GDK_R, GDK_CONTROL_MASK, VTM_CT_TLO, 1,
				VCL_LOCAL_MENUS | VCL_LWUH_TBAR | VCC_META_SEL, VCS_VIS_EN, LIB_PATH"redo.png",
				NULL, &local_redo_enabled, &on_tlo_redo, NULL, &edit_select_all };
static struct cmd edit_local_hist_jump =  { "Jump", "Jump to selected action", GDK_J, GDK_CONTROL_MASK, VTM_UNDO_HISTORY, 1,
				VCL_LWUH_TBAR | VCC_META_SEL, VCS_VIS_EN, LIB_PATH"jump.png",
				NULL, &history_jump_enabled, &on_hist_jump, NULL, &edit_local_redo_cmd };
static struct cmd edit_local_undo_cmd =  { "Undo", "Undo one local action", GDK_Z, GDK_CONTROL_MASK, VTM_CT_TLO, 1,
				VCL_LOCAL_MENUS | VCL_LWUH_TBAR | VCC_META_SEL, VCS_VIS_EN, LIB_PATH"undo.png", NULL,
				&local_undo_enabled, &on_tlo_undo, NULL, &edit_local_hist_jump };
static struct cmd edit_redo_cmd =  { "Redo", "Redo one action", 0, 0, VTM_ANYTHING, 0,
				VCL_MAIN_MENU | VCL_UB_MENU | VCL_WUH_TBAR | VCC_META_SEL, VCS_VIS_EN,
				LIB_PATH"redo.png", NULL, &redo_enabled, &on_obj_redo, NULL, &edit_local_undo_cmd };
static struct cmd edit_hist_jump =  { "Jump", "Jump to selected action", 0, 0, VTM_UNDO_HISTORY, 1,
				VCL_WUH_TBAR | VCC_META_SEL, VCS_VIS_EN,
				LIB_PATH"jump.png", NULL, &history_jump_enabled, &on_hist_jump, NULL, &edit_redo_cmd };
static struct cmd edit_undo_cmd =  { "Undo", "Undo one action", 0, 0, VTM_ANYTHING, 0,
				VCL_MAIN_MENU | VCL_UB_MENU | VCL_WUH_TBAR | VCC_META_SEL, VCS_VIS_EN,
				LIB_PATH"undo.png", NULL, &undo_enabled, &on_obj_undo, NULL, &edit_hist_jump };
static struct cmd edit_delete_cmd = { "Delete", "Delete", GDK_Delete, 0, VTM_GO, -1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, NULL, &on_go_delete, NULL, &edit_undo_cmd };
static struct cmd edit_move_back = { "Move to back", "Move to back", GDK_D, GDK_SHIFT_MASK, VTM_GO, 1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, &can_move_back, &on_move_back, NULL, &edit_delete_cmd };
static struct cmd edit_move_front = { "Move to front", "Move to front", GDK_U, GDK_SHIFT_MASK, VTM_GO, 1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, &can_move_front, &on_move_front, NULL, &edit_move_back };
static struct cmd edit_move_down = { "Move down", "Move down", GDK_D, 0, VTM_GO, 1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, &can_move_down, &on_move_down, NULL, &edit_move_front };
static struct cmd edit_move_up = { "Move up", "Move up", GDK_U, 0, VTM_GO, 1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, &can_move_up, &on_move_up, NULL, &edit_move_down };
static struct cmd edit_enclose_sel = { "Select dependent", "Add dependent to selection", 0, 0, VTM_CT_GRP, 1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, NULL, &on_go_enclose_sel, NULL, &edit_move_up };
static struct cmd edit_ungroup = { "Ungroup", "Ungroup the group", 0, 0, VTM_GROUP, 1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, NULL, &on_group_ungroup, NULL, &edit_enclose_sel };
static struct cmd edit_group = { "Group selected", "Group the selected objects", 0, 0, VTM_GO, -1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, NULL, &on_group_selected, NULL, &edit_ungroup };
static struct cmd edit_paste_cmd =  { "Paste", "Paste clipboard contents", GDK_V, GDK_CONTROL_MASK, VTM_CT_TLO, 1,
				VCL_LOCAL_MENUS | VCC_STOCK | VCC_META_SEL,
				VCS_VIS_EN, "gtk-paste", NULL, &paste_cmd_enabled, &on_paste, NULL, &edit_group };
static struct cmd edit_cut_cmd =  { "Cut", "Cut selection", GDK_X, GDK_CONTROL_MASK, VTM_GO, -1,
				VCL_LOCAL_MENUS | VCC_STOCK, VCS_VIS_EN, "gtk-cut", NULL, &cutcopy_enabled, &on_cut, NULL, &edit_paste_cmd };
static struct cmd edit_copy_cmd =  { "Copy", "Copy selection", GDK_C, GDK_CONTROL_MASK, VTM_GO, -1,
				VCL_LOCAL_MENUS | VCC_STOCK, VCS_VIS_EN, "gtk-copy", NULL, &cutcopy_enabled, &on_copy, NULL, &edit_cut_cmd };
static struct cmd edit_text_cmd =  { "Text", "Edit text", 0, 0, VTM_TEXT | VTM_TEX_TEXT, 1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, NULL, &on_edit_text, NULL, &edit_copy_cmd };
static struct cmd edit_go_prop_cmd =  { "GO properties", "Change GO properties", 0, 0, VTM_GO, 1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, LIB_PATH"obj_properties.png", NULL, NULL, &on_change_go_prop, NULL, &edit_text_cmd };

static struct cmd edit_prop_cmd =  { "Object Properties", "Change object properties", GDK_P, GDK_CONTROL_MASK, VTM_OBJ, 1,
				VCL_UB_MENU | VCC_META_SEL, VCS_VIS_EN, LIB_PATH"obj_properties.png", NULL, NULL, &on_change_obj_prop,
				NULL, &edit_go_prop_cmd };
/* Basic transforms */
static struct cmd edit_rot270 = { "Rotate by 270 degrees", "Rotate by 270 degrees", 0, 0, VTM_GO, -1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, NULL, &on_rotate270, NULL, NULL };
static struct cmd edit_rot180 = { "Rotate by 180 degrees", "Rotate by 180 degrees", 0, 0, VTM_GO, -1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, NULL, &on_rotate180, NULL, &edit_rot270 };
static struct cmd edit_rot90 = { "Rotate by 90 degrees", "Rotate by 90 degrees", 0, 0, VTM_GO, -1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, NULL, &on_rotate90, NULL, &edit_rot180 };
static struct cmd edit_hflip = { "Flip horizontally", "Flip horizontally", 0, 0, VTM_GO, -1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, NULL, &on_flip_horizontal, NULL, &edit_rot90 };
static struct cmd edit_vflip = { "Flip vertically", "Flip vertically", 0, 0, VTM_GO, -1,
				VCL_LOCAL_MENUS, VCS_VIS_EN, NULL, NULL, NULL, &on_flip_vertical, NULL, &edit_hflip };
/* Snap category */
static int snap_depend_active( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( factory.snap_set & SNAP_DEPENDENCY, VCS_CHECKED ); }

static int snap_inters_active( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( factory.snap_set & SNAP_INTERS, VCS_CHECKED ); }

static int snap_lines_active( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( factory.snap_set & SNAP_LINES, VCS_CHECKED ); }

static int snap_grid_active( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( factory.snap_set & SNAP_GRID, VCS_CHECKED ); }

static int snap_hangers_active( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( factory.snap_set & SNAP_HANGERS, VCS_CHECKED ); }

static struct cmd snap_depend = { "Create dependencies", "Create dependencies", 0, 0, VTM_ANYTHING, 0,
				VCL_LOCAL_MENUS | VCL_VIEW_TBAR |  VCC_META_SEL | VCC_CHECKABLE | VCC_SAFE | VCC_STOCK, VCS_VIS_EN,
				"gtk-connect", NULL, &snap_depend_active, &on_snap_depend, NULL, NULL };
static struct cmd snap_inters_cmd = { "Snap to intersections", "Snap to intersections", 0, 0, VTM_ANYTHING, 0,
				VCL_LOCAL_MENUS | VCL_VIEW_TBAR |  VCC_META_SEL | VCC_CHECKABLE | VCC_SAFE, VCS_VIS_EN,
				LIB_PATH"snap_intersections.png", NULL, &snap_inters_active, &on_snap_inters, NULL, &snap_depend };
static struct cmd snap_lines_cmd = { "Snap to lines", "Snap to lines", 0, 0, VTM_ANYTHING, 0,
				VCL_LOCAL_MENUS | VCL_VIEW_TBAR |  VCC_META_SEL | VCC_CHECKABLE | VCC_SAFE, VCS_VIS_EN,
				LIB_PATH"snap_lines.png", NULL, &snap_lines_active, &on_snap_lines, NULL, &snap_inters_cmd };
static struct cmd snap_grid_cmd = { "Snap to grid", "Snap to grid", 0, 0, VTM_ANYTHING, 0,
				VCL_LOCAL_MENUS |  VCL_VIEW_TBAR | VCC_META_SEL | VCC_CHECKABLE | VCC_SAFE, VCS_VIS_EN,
				LIB_PATH"snap_grid.png", NULL, &snap_grid_active, &on_snap_grid, NULL, &snap_lines_cmd };
static struct cmd snap_hangers_cmd = { "Snap to hangers", "Snap to hangers", 0, 0, VTM_ANYTHING, 0,
				VCL_LOCAL_MENUS |  VCL_VIEW_TBAR | VCC_META_SEL | VCC_CHECKABLE | VCC_SAFE, VCS_VIS_EN,
				LIB_PATH"snap_hangers.png", NULL, &snap_hangers_active, &on_snap_hangers, NULL, &snap_grid_cmd };
/* New category */

static int state_active( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( cmd->of_state == factory.start_state, VCS_CHECKED ); }

static int can_step_back( struct cmd * cmd UNUSED, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( context.tlo == factory.tlo && factory.state != factory.start_state, VCS_ENABLED ); }

static struct cmd new_random_bezier_cubic = { "New random cubic bezier", "New random cubic bezier", 0, 0, VTM_GROUP | VTM_CT_GRP, 1,
				VCL_LOCAL_MENUS |  VCC_LOCAL_OP, VCS_VIS_EN,
				LIB_PATH"bezier_cubic.png", NULL, NULL, &on_random_bezier_cubic, NULL, NULL };
static struct cmd new_random_bezier_quad = { "New random quadratic bezier", "New random quadratic bezier", 0, 0, VTM_GROUP | VTM_CT_GRP, 1,
				VCL_LOCAL_MENUS |  VCC_LOCAL_OP, VCS_VIS_EN,
				LIB_PATH"bezier_quad.png", NULL, NULL, &on_random_bezier_quad, NULL, &new_random_bezier_cubic };
static struct cmd new_random_segment = { "New random segment", "New random segment", 0, 0, VTM_GROUP | VTM_CT_GRP, 1,
				VCL_LOCAL_MENUS |  VCC_LOCAL_OP, VCS_VIS_EN,
				LIB_PATH"segment.png", NULL, NULL, &on_random_segment, NULL, &new_random_bezier_quad };
static struct cmd new_random_point = { "New random point", "New random point", 0, 0, VTM_GROUP | VTM_CT_GRP, 1,
				VCL_LOCAL_MENUS |  VCC_LOCAL_OP, VCS_VIS_EN,
				LIB_PATH"point.png", NULL, NULL, &on_random_point, NULL, &new_random_segment };
static struct cmd op_break_cmd = { "Break creation", "Break creation", GDK_Escape, 0, VTM_CT_TLO, 1,
				VCL_VIEW_MENU | VCC_META_SEL | VCC_SAFE, VCS_VIS_EN,
#ifndef VRR_POTEMKIN
				NULL, NULL, &can_step_back, &on_break_op, NULL, &new_random_point };
#else
				NULL, NULL, &can_step_back, &on_break_op, NULL, NULL };
#endif
static struct cmd step_back_cmd = { "Step back", "Step back", GDK_BackSpace, 0, VTM_CT_TLO, 1,
				VCL_VIEW_MENU | VCC_META_SEL | VCC_SAFE, VCS_VIS_EN,
				NULL, NULL, &can_step_back, &on_step_back, NULL, &op_break_cmd };
static struct cmd new_text = { "New text", "Create a new text", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"text.png", NULL, &state_active, NULL, &ofs_text_1, NULL };
static struct cmd new_tex_text = { "New TeX text", "Create a new TeX text", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"tex.png", NULL, &state_active, NULL, &ofs_tex_text_1, &new_text };
static struct cmd new_ellarc_center_radii = { "New ellipse center radii", "Create a new elliptic arc by center and radii", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"ellarc_center_radii.png", NULL, &state_active, NULL, &ofs_ellarc_center_radii_1, NULL };
static struct cmd new_ellipse_center_radii = { "New ellipse center radii", "Create a new ellipse by center and radii", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"ellipse_center_radii.png", NULL, &state_active, NULL,
			&ofs_ellipse_center_radii_1, &new_ellarc_center_radii };
static struct cmd new_ellipse_xy1ecc = { "New ellipse center, rot, ecc", "Create a new ellipse by center, point, rotation and eccentricity",
			0, 0, VTM_CT_GRP, 1, VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"ellipse_center_pt.png", NULL, &state_active, NULL,
			&ofs_ellipse_xy1ecc_1, &new_ellipse_center_radii };
static struct cmd new_ellipse_foci_pt = { "New ellipse foci point", "Create a new ellipse by foci and point", 0, 0,
			VTM_CT_GRP, 1, VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"ellipse_foci_pt.png", NULL, &state_active, NULL,
			&ofs_ellipse_foci_pt_1, &new_ellipse_xy1ecc };
static struct cmd new_ellipse_3pecc = { "New ellipse 3 points ecc", "Create a new ellipse by 3 points, rotation and ecc", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"ellipse_3points.png", NULL, &state_active, NULL, &ofs_ellipse_3p_ecc_1,
			&new_ellipse_foci_pt };
static struct cmd new_ellipse_3p = { "New ellipse 3 points ", "Create a new smallest ellipse by 3 points", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"smallest_ellipse_3points.png", NULL, &state_active, NULL, &ofs_ellipse_3p_smallest_1,
			&new_ellipse_3pecc };
static struct cmd new_circ_center_rad = { "New circle center radius", "Create a new circle by center and radius", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"circle_center_radius.png", NULL, &state_active, NULL, &ofs_circ_center_radius_1, NULL };
static struct cmd new_circ_center_pt = { "New circle", "Create a new circle by center and point", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"circle_center_pt.png", NULL, &state_active, NULL, &ofs_circ_center_pt_1, &new_circ_center_rad };
static struct cmd new_circ_3p = { "New circle by 3 points", "Create a new circle by 3 points", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"circle_3points.png", NULL, &state_active, NULL, &ofs_circ_3p_1, &new_circ_center_pt };
static struct cmd new_arc_center_rad = { "New arc center radius", "Create a new arc by center and radius", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"arc_center_radius.png", NULL, &state_active, NULL, &ofs_arc_center_radius_1, &new_circ_3p };
static struct cmd new_arc_3p = { "New arc", "Create a new arc by 3 points", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"arc_3points.png", NULL, &state_active, NULL, &ofs_arc_3p_1, &new_arc_center_rad };
static struct cmd new_bez_cub = { "New cubic bezier", "Create a new cubic bezier", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"bezier_cubic.png", NULL, &state_active, NULL, &ofs_bez_cub_1, NULL };
static struct cmd new_bez_quad = { "New quadratic bezier", "Create a new quadratic bezier", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"bezier_quad.png", NULL, &state_active, NULL, &ofs_bez_quad_1, &new_bez_cub };
static struct cmd new_segment = { "New segment", "Create a new segment", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"segment.png", NULL, &state_active, NULL, &ofs_segment_1, NULL };

static struct cmd newpt_inters = { "New intersection point", "Create a new intersection point", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, 0, LIB_PATH"point.png", NULL, &state_active, NULL, &ofs_point_inters, NULL };
static struct cmd newdecpt_cmd = { "New decoration point", "Create a new decoration point", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"decoration.png", NULL, &state_active, NULL, &ofs_decorator_pt, &newpt_inters };
static struct cmd newpt_cmd = { "New point", "Create a new point", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW | VCC_META_SEL | VCC_START_OP | VCC_INSIDE_TRANS | VCC_CHECKABLE |
			VCC_LOCAL_OP, VCS_VIS_EN, LIB_PATH"point.png", NULL, &state_active, NULL, &ofs_point, &newdecpt_cmd };

/* Mode category */
static struct cmd mode_rehang = { "Rehang", "Rehang some anchors", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW_TBAR | VCC_META_SEL | VCC_START_OP | VCC_CHECKABLE | VCC_LOCAL_OP | VCC_INSIDE_TRANS,
			VCS_VIS_EN, LIB_PATH"rehang_mode.png", NULL, &state_active, NULL, &ofs_rehang_1, NULL };

static struct cmd mode_transform = { "Transform", "Transform selection", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW_TBAR | VCC_META_SEL | VCC_START_OP | VCC_CHECKABLE | VCC_LOCAL_OP,
			VCS_VIS_EN, LIB_PATH"transform_mode.png", NULL, &state_active, NULL, &ofs_transform_1, &mode_rehang };

static struct cmd mode_select = { "Select", "Transform selection", 0, 0, VTM_CT_GRP, 1,
			VCL_VIEW_TBAR | VCC_META_SEL | VCC_START_OP | VCC_CHECKABLE | VCC_LOCAL_OP,
			VCS_VIS_EN, LIB_PATH"select_mode.png", NULL, &state_active, NULL, &ofs_select, &mode_transform };

/* View category */
static int property_window_visible (struct cmd *cmd, GtkWidget *widget UNUSED)
{ return COMBINE_STATE( context_property_window, VCS_CHECKED); }

static int console_window_visible( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( console_window && console_window->visible, VCS_CHECKED ); }

static int univ_browser_visible( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( wnd_univ_browser.window && wnd_univ_browser.visible, VCS_CHECKED ); }

static int main_window_visible( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( main_window_open, VCS_CHECKED ); }

static int plugin_window_visible( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( wnd_plugins.visible, VCS_CHECKED ); }

static int settings_window_visible( struct cmd * cmd, GtkWidget * widget UNUSED )
{ return COMBINE_STATE( wnd_settings.window, VCS_CHECKED ); }

static struct cmd view_zoom_reset = { "Reset zoom and rotation", "Reset zoom and rotation", GDK_R, 0, VTM_VIEW, 1,
				VCL_VIEW_MENU | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, &on_view_zoom_reset, NULL, NULL };
static struct cmd view_out_kp = { "Zoom out", "Zoom out", GDK_KP_Subtract, 0, VTM_VIEW, 1, VCL_VIEW_MENU | VCC_META_SEL,
				VCS_ENABLED, NULL, NULL, NULL, &on_view_zoom_out, NULL, &view_zoom_reset };
static struct cmd view_in_kp = { "Zoom in", "Zoom in", GDK_KP_Add, 0, VTM_VIEW, 1, VCL_VIEW_MENU | VCC_META_SEL,
				VCS_ENABLED, NULL, NULL, NULL, &on_view_zoom_in, NULL, &view_out_kp };
static struct cmd view_out = { "Zoom out", "Zoom out", GDK_minus, 0, VTM_VIEW, 1, VCL_VIEW_MENU | VCC_META_SEL,
				VCS_VIS_EN, NULL, NULL, NULL, &on_view_zoom_out, NULL, &view_in_kp };
static struct cmd view_in = { "Zoom in", "Zoom in", GDK_equal, 0, VTM_VIEW, 1, VCL_VIEW_MENU | VCC_META_SEL,
				VCS_VIS_EN, NULL, NULL, NULL, &on_view_zoom_in, NULL, &view_out };
static struct cmd view_fit_all = { "Fit all in view", "Fit all in view", GDK_KP_Multiply, 0, VTM_VIEW, 1,
				VCL_VIEW_MENU | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, &on_view_fit_all, NULL, &view_in };
static struct cmd view_fit_selected = { "Fit selected in view", "Fit selected in view", GDK_F, 0, VTM_VIEW, 1,
				VCL_VIEW_MENU | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, &on_view_fit_selected, NULL, &view_fit_all };

/* Windows category */
static struct cmd view_settings = { "Global settings", "Global settings", 0, 0, VTM_ANYTHING, 0,
				VCC_CHECKABLE | VCL_ALL_MENUS | VCL_MAIN_TBAR | VCC_STOCK | VCC_META_SEL, VCS_VIS_EN, "gtk-preferences",
				NULL, &settings_window_visible, &on_show_settings_window, NULL, NULL };
static struct cmd view_plugins_cmd = { "Plugin manager", "Plugin manager", 0, 0, VTM_ANYTHING, 0,
				VCL_ALL_MENUS | VCL_MAIN_TBAR | VCC_CHECKABLE | VCC_META_SEL, VCS_VIS_EN, LIB_PATH"properties.png",
				NULL, &plugin_window_visible, &on_show_plugin_window, NULL, &view_settings };
static struct cmd view_mainwin_cmd = { "Main window", "Main window", GDK_M, GDK_CONTROL_MASK, VTM_ANYTHING, 0,
				VCL_LOCAL_MENUS | VCC_CHECKABLE | VCC_META_SEL, VCS_VIS_EN, 0, NULL, &main_window_visible,
				&on_show_main_window, NULL, &view_plugins_cmd } ;
static struct cmd view_clipboard = {"Clipboard", "View clipboard", 0, 0, VTM_ANYTHING, 0, VCL_ALL_MENUS | VCC_META_SEL,
				 VCS_VIS_EN, NULL, NULL, NULL, &on_view_clipboard, NULL, &view_mainwin_cmd };
static struct cmd view_local_uh = { "Local undo history", "Local undo history", 0, 0, VTM_TLO | VTM_CT_TLO, 1,
				VCL_LOCAL_MENUS | VCC_META_SEL, VCS_VIS_EN, LIB_PATH"history.png", NULL, NULL, &on_undo_history,
				NULL, &view_clipboard };
static struct cmd view_uh_cmd = { "Undo history", "Undo history", 0, 0, VTM_ANYTHING, 0, VCL_ALL_MENUS | VCL_MAIN_TBAR | VCC_META_SEL,
				VCS_VIS_EN, LIB_PATH"history.png", NULL, NULL, &on_show_undo_history, NULL, &view_local_uh };
static struct cmd view_ub_cmd = { "Universe browser", "Universe browser", GDK_U, GDK_CONTROL_MASK, VTM_ANYTHING, 0,
				VCL_ALL_MENUS | VCL_MAIN_TBAR | VCC_CHECKABLE | VCC_META_SEL, VCS_VIS_EN, LIB_PATH"browser.png",
				NULL, &univ_browser_visible, &on_show_universe_browser, NULL, &view_uh_cmd };
static struct cmd view_property_cmd = { "Property window", "Property window", 0, 0, VTM_ANYTHING, 0, VCL_ALL_MENUS |
				VCC_CHECKABLE | VCC_META_SEL, VCS_VIS_EN, 0, NULL, &property_window_visible,
				&on_show_property_window, NULL, &view_ub_cmd };
static struct cmd view_console_cmd = { "Console", "Scheme console", GDK_C, GDK_SHIFT_MASK, VTM_ANYTHING, 0,
				VCL_ALL_MENUS | VCL_MAIN_TBAR | VCC_CHECKABLE | VCC_META_SEL, VCS_VIS_EN, LIB_PATH"console.png", NULL,
				&console_window_visible, &on_show_console_window, NULL, &view_property_cmd };

/* Debug category */

static struct cmd debug_context = { "Dump context", "Dump context", GDK_F2, 0, VTM_ANYTHING, 0,
				VCL_MAIN_MENU | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, &on_context_dump,
				NULL, NULL };
static struct cmd debug_meta_op2 = { "Meta operation on two TLOs", "", 0, 0, VTM_TLO, 2,
				VCL_UB_MENU | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, NULL, NULL, &debug_context };
static struct cmd debug_meta_op = { "Meta operation on one TLO", "", 0, 0, VTM_TLO, 1,
				VCL_UB_MENU | VCC_META_SEL, VCS_VIS_EN, NULL, NULL, NULL, NULL, NULL, &debug_meta_op2 };
static struct cmd debug_kerntest = { "Kerntest", "Kernel test", 0, 0, VTM_ANYTHING, 0,
				VCL_ALL_MENUS | VCC_META_SEL, VCS_VIS_EN, 0, NULL, NULL, &on_kerntest, NULL, &debug_meta_op };
static struct cmd debug_dump = { "Kernel dump", "Kernel dump", 0, 0, VTM_ANYTHING, 0,
				VCL_EVERYWHERE | VCC_META_SEL, VCS_VIS_EN, LIB_PATH"debug.png", NULL, NULL, &on_dump, NULL, &debug_kerntest };

/* Help category */
static struct cmd vrr_online = {"VRR online", "VRR online", 0, 0, VTM_ANYTHING, 0, VCL_ALL_MENUS | VCC_META_SEL, VCS_VIS_EN,
			LIB_PATH"help.png", 0, NULL, &on_open_www, NULL, NULL };
static struct cmd print_help = {"Help", "Show help", GDK_F1, 0, VTM_ANYTHING, 0, VCL_EVERYWHERE | VCC_META_SEL, VCS_VIS_EN,
			LIB_PATH"help.png", 0, NULL, &on_help, NULL, &vrr_online };

/* Category initialization */
#ifdef VRR_POTEMKIN
static struct cmd_category debug_ctg = { "Debug", 0, NULL, NULL, &debug_dump, NULL };
#else
static struct cmd_category debug_ctg = { "Debug", VCL_MAIN_MENU | VCL_MAIN_TBAR | VCL_UB_MENU, NULL, NULL, &debug_dump, NULL };
#endif
static struct cmd_category help_ctg = { "Help", VCL_EVERYWHERE, NULL, &debug_ctg, &print_help, NULL };
static struct cmd_category export_ctg = { "Export", VCL_UB_MENU | VCL_VIEW_MENU, NULL, &help_ctg, &file_export_pdf, NULL };
static struct cmd_category windows_ctg = { "Windows", VCL_ALL_MENUS | VCL_MAIN_TBAR, NULL, &export_ctg, &view_console_cmd, NULL };
static struct cmd_category view_ctg = { "View", VCL_VIEW_MENU, NULL, &windows_ctg, &view_fit_selected, NULL };
static struct cmd_category snap_ctg = { "Snap", VCL_VIEW_MENU | VCL_VIEW_TBAR | VCC_CAN_CLPS, NULL, &view_ctg, &snap_hangers_cmd, NULL };
static struct cmd_category new_text_ctg = { "New text", VCL_VIEW_MENU | VCL_VIEW_TBAR2 | VCC_CAN_CLPS, NULL, NULL, &new_tex_text, NULL };
static struct cmd_category new_ellipse_ctg = { "New elliptic arc", VCL_VIEW_MENU | VCL_VIEW_TBAR2 | VCC_CAN_CLPS, NULL,
						&new_text_ctg, &new_ellipse_3p, &new_ellipse_3p };
static struct cmd_category new_circ_ctg = { "New circular arc", VCL_VIEW_MENU | VCL_VIEW_TBAR2 | VCC_CAN_CLPS, NULL,
						&new_ellipse_ctg, &new_arc_3p, &new_arc_3p };
static struct cmd_category new_bez_ctg = { "New Bezier", VCL_VIEW_MENU | VCL_VIEW_TBAR | VCC_CAN_CLPS, NULL,
					&new_circ_ctg, &new_bez_quad, NULL };
static struct cmd_category new_seg_ctg = { "New segment", VCL_VIEW_MENU | VCL_VIEW_TBAR | VCC_CAN_CLPS, NULL,
					&new_bez_ctg, &new_segment, NULL };
static struct cmd_category new_point_ctg = { "New point", VCL_VIEW_MENU | VCL_VIEW_TBAR | VCC_CAN_CLPS, NULL, &new_seg_ctg, &newpt_cmd, NULL };
static struct cmd_category new_ctg = { "Create", VCL_VIEW | VCL_UB_MENU, &new_point_ctg, &snap_ctg, &step_back_cmd, NULL };
static struct cmd_category mode_ctg = { "Mode", VCL_VIEW_TBAR | VCC_CAN_CLPS, NULL, &new_ctg, &mode_select, &mode_select };
#ifdef VRR_POTEMKIN
static struct cmd_category transform_ctg = { "Transform", 0, NULL, NULL, &edit_vflip, NULL };
#else
static struct cmd_category transform_ctg = { "Transform", VCL_VIEW_MENU, NULL, NULL, &edit_vflip, NULL };
#endif
static struct cmd_category edit_ctg = { "Edit", VCL_EVERYWHERE | VCL_WUH_TBAR | VCL_LWUH_TBAR, &transform_ctg, &mode_ctg,
					&edit_prop_cmd, NULL };
static struct cmd_category page_ctg = { "Page", VCL_UB_MENU | VCL_VIEW_MENU, NULL, &edit_ctg, &new_view_cmd, NULL };
static struct cmd_category file_ctg = { "File", VCL_ALL_MENUS | VCL_MAIN_TBAR, NULL, &page_ctg, &file_new_cmd, NULL };
static struct cmd_category main_ctg = { "All", VCL_EVERYWHERE, &file_ctg, NULL, NULL, NULL };

static void collapse_category( GtkWidget * w UNUSED, struct cmd_category * ctg )
{
	if ( sync_in_progress || !(ctg->flags & VCC_CAN_CLPS )) return;
	ctg->flags ^= VCC_COLLAPSED;
	if ( !(ctg->selected_cmd) ) ctg->selected_cmd = ctg->first_cmd;
	refresh_gui();
}
static void recurse_collapsion( struct cmd_category * ctg, struct cmd * clicked )
{
	struct cmd * co;
	struct cmd_category * cc;
	for ( co = ctg->first_cmd; co && co != clicked; co = co->next_cmd );
	if ( co )
	{ ctg->selected_cmd = co; collapse_category( NULL, ctg ); }
	else for ( cc = ctg->son; cc; cc = cc->brother ) recurse_collapsion(cc, clicked);
}
static gboolean command_icon_press( GtkWidget * widget UNUSED, GdkEventButton *event, struct cmd * command )
{
	if (event->button == 1) return FALSE;
	recurse_collapsion( &main_ctg, command );
	return TRUE;
}
static void on_command_activated( GtkWidget * widget, struct cmd * command )
{
	if ( sync_in_progress ) return; //ignore callbacks caused by checking items manually

	struct window * w;
	ASSERT( widget );
	ASSERT( command );
	void * v = g_object_get_data( G_OBJECT(widget), vrrgroup );
	struct go_group * group = NULL;
	if ( v )
	{
		group = (struct go_group*)void2go(v);
		ASSERT_GROUP(group);
	}
	w = (struct window *)g_object_get_data( G_OBJECT(widget), vrrvw );

	if ( w && w != context.w || group && group != context.group )
		change_context( w, group, (context.o && context.o->kind == T_GO &&
			go_get_group(o2go(context.o)) == group ) ? context.o : NULL );
	execute_command_item( command );
}

static GtkAccelGroup * accel_group = NULL;
static GtkTooltips * tooltips;
GtkAccelGroup * get_command_accel_group(void){ return accel_group; }

void command_manager_init(void)
{
	accel_group = gtk_accel_group_new();
	tooltips = GTK_TOOLTIPS(gtk_tooltips_new());
	context.sel_valid = 0;
	context.msel_valid = 0;
	context.o = NULL;
	context.group = NULL;
	context.tlo = NULL;
	context.w = NULL;
	load_icons_in_category( &main_ctg );

	change_stamp();
}

static void load_icons_in_category( struct cmd_category * ctg )
{
	struct cmd * co;
	struct cmd_category * cc;
	for ( co = ctg->first_cmd; co; co = co->next_cmd )
		if ((co->icon_path) && !(co->flags & VCC_STOCK))
			if (!(co->icon_pixbuf = icon_load(co->icon_path)))
			{
				co->flags |= VCC_STOCK; //set the icon of missing image instead
				co->icon_path = "gtk-missing-image";
			}
	for ( cc = ctg->son; cc; cc = cc->brother ) load_icons_in_category(cc);
}

void command_manager_finish(void) { }

struct cmd * register_new_command( struct cmd_category * category, struct cmd * cmd )
{
	struct cmd * brother;
	ASSERT( category );
	ASSERT( cmd );
	change_stamp();
	cmd->next_cmd = NULL;
	if ( category->first_cmd ){
		for ( brother = category->first_cmd; brother->next_cmd; brother = brother->next_cmd );
		brother->next_cmd = cmd;
	} else category->first_cmd = cmd;
	return cmd;
}
void remove_command( struct cmd * command UNUSED )
{
	change_stamp();
}

struct cmd_category * add_new_category( struct cmd_category * parent, char * title, unsigned int flags )
{
	struct cmd_category * c, * n;
	ASSERT( parent );
	DBG("Create a new category\n");
	n = xmalloc( sizeof(struct cmd_category) );
	n->title = title;
	n->flags = flags;
	n->brother = NULL;
	n->son = NULL;
	n->first_cmd = NULL;

	if ( !parent->son ){
		parent->son = n;
	} else {
		c = parent->son;
		while ( c->brother ) c = c->brother;
		c->brother = n;
	}
	change_stamp();
	return n;
}

/*static int menu_big_bang( struct cmd * command UNUSED )
{
	struct cmd_category * fa_ctg = add_new_category( &help_ctg, "First aid", VCL_ALL_MENUS );
	struct cmd_category * dang_ctg = add_new_category( &selection_ctg, "Dangerous actions", VCL_ALL_MENUS );
	add_new_category( &selection_ctg, "More dangerous actions", VCL_ALL_MENUS );
	register_new_command_full( fa_ctg, "CPR",
		NULL, "Cardiopulmonary resuscitation", 0, VTM_ANYTHING, VCL_EVERYWHERE, VCS_VIS_EN,
		NULL, NULL, NULL );
	register_new_command_full( dang_ctg, "DO NOT CLICK",
		NULL, "Danger, do not click!", 0, VTM_ANYTHING, VCL_EVERYWHERE, VCS_VISIBLE,
		NULL, NULL, NULL );
	register_new_command_full( &selection_ctg, "XXX",
		NULL, "Do something extraordinary", 0, VTM_ANYTHING, VCL_EVERYWHERE, VCS_VIS_EN,
		NULL, NULL, NULL );

	return VCS_VISIBLE;
}*/

void change_context( struct window * window, struct go_group * group, struct o * o )
{
	int refresh = (context.w != window) || (context.group != group) || (context.o != o);
	context.w = window;
	if ( context.group != group )
	{
		context.group = group;
		context.tlo = group ? go_get_tlo(&group->go) : NULL;
		context.sel_valid = 0;
	}
	if(context.o != o && context_property_window) prop_window_context_changed(o);
	
	context.o = o;

	if (refresh) refresh_gui();
}

static void calculate_selection(void)
{
	if ( context.sel_valid && context.msel_valid ) return;

	if ( !context.msel_valid )
	{
		context.msel_mask = ( obj_sel_count[OT_DOCUMENT] > 0 ) * VTM_DOCUMENT;
		context.msel_cnt_all = obj_sel_count[OT_DOCUMENT];
		context.msel_mask |= ( obj_sel_count[OT_TLO] > 0 ) * VTM_TLO;
		context.msel_cnt_all += obj_sel_count[OT_TLO];
		context.msel_valid = 1;
	}

	if ( !context.sel_valid )
	{
		struct go * go;
		context.sel_mask = 0;
		context.sel_cnt_all = 0;
		for (uns u = 0; u < GOT_MAX; u++) context.sel_cnt[u] = 0;

		if ( context.group ) WALK_GROUP_GO(context.group, go)
			if ( go_is_selected(go) )
			{
				context.sel_cnt[go->type]++;
				context.sel_mask |= GOT_TO_VTM(go->type);
				context.sel_cnt_all++;
			}
		context.sel_valid = 1;
	}
}

/** cmask (=command mask) must be a bitwise supset of the context mask */
static int sel_req_ok( uns mask, uns cmask, int cnt, int ccnt )
{
	int ok = ( mask & ~cmask ) == 0 && ( ccnt == 0 || (ccnt == -1 && cnt > 0) || ccnt == cnt );
	//DBG("Hit=%d, Context: M=%x, C=%d, Cmd: M=%x, C=%d\n", ok, mask, cnt, cmask, ccnt );
	return ok;
}

static int context_matches( struct cmd * cmd )
{
	/* An opened view determines a view, TLO and document */
	if ( cmd->count_request <= 1 &&
		((context.w && WT_TO_VTM(context.w->type) == cmd->type_mask) || //the window is what it wants
		((cmd->type_mask & VTM_CT_GRP || cmd->type_mask & VTM_CT_TLO || cmd->type_mask & VTM_CT_DOC)
		 	&& context.group) )) //a group determines it
		return 1;

	/* If there is a selection, use selection; if not, use current o (under cursor) */
	if ( cmd->flags & VCC_META_SEL ){
		if ( context.msel_cnt_all > 0 || cmd->count_request == 0 )
			return sel_req_ok( context.msel_mask, cmd->type_mask, context.msel_cnt_all, cmd->count_request );
	} else {
		if ( context.sel_cnt_all > 0 || cmd->count_request == 0 )
			return sel_req_ok( context.sel_mask, cmd->type_mask, context.sel_cnt_all, cmd->count_request );
	}
	/* This is a bit weird ... */
	if ( context.o )
		return sel_req_ok( ( context.o->kind == T_GO ) ? ( 1 << (context.o->type + OT_MAX) ) : ( 1 << context.o->type ),
			cmd->type_mask, 1, cmd->count_request );
	return 0;
}

static void sync_cmd_item( struct cmd * cmd, GtkWidget * command, unsigned int location, unsigned int hide )
{
	//evaluate the context conditions first
	if ( ((location & VCL_VIEW_TBAR || location & VCL_VIEW_TBAR2) && (cmd->flags & VCC_LOCAL_OP) || context_matches( cmd )) )
		cmd->state = cmd->state | VCS_ENABLED;
	else cmd->state = cmd->state & ~VCS_ENABLED;
	//add custom conditions
	if ( cmd->get_state_f ) cmd->state = cmd->get_state_f( cmd, command );

	/* set visibility */
	if ( !hide && (cmd->state & VCS_VISIBLE) ) gtk_widget_show_all( GTK_WIDGET(command) );
	else gtk_widget_hide_all( GTK_WIDGET(command) );

	/* set enabled */
	gtk_widget_set_sensitive( command, cmd->state & VCS_ENABLED );
}

static void sync_menu_item( struct cmd * cmd, GtkWidget * command, unsigned int location )
{
	sync_cmd_item( cmd, command, location, 0 );
	/* set checked */
	if(!GTK_IS_TEAROFF_MENU_ITEM(command))
	{
		if ( cmd->flags & VCC_CHECKABLE )
		{
			sync_in_progress = 1;
			gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(command), cmd->state & VCS_CHECKED );
			sync_in_progress = 0;
		}
	}
}
static void sync_tool_item( struct cmd * cmd, GtkWidget * command, unsigned int location, struct cmd_category * ctg )
{
	sync_cmd_item( cmd, command, location, (ctg->flags & VCC_COLLAPSED) && (ctg->selected_cmd != cmd ) );
	/* set checked */
	if (!GTK_IS_TEAROFF_MENU_ITEM(command))
	{
		if ( cmd->flags & VCC_CHECKABLE )
		{
			sync_in_progress = 1;
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(command), cmd->state & VCS_CHECKED );
			sync_in_progress = 0;
		}
	}
	gtk_tooltips_set_tip( tooltips, command, cmd->description, NULL );
}

GtkWidget * create_menu( unsigned int location, struct cmddata cd ){

	GtkWidget * msh;
	calculate_selection();

	switch( location ){
	case VCL_MAIN_MENU: msh = gtk_menu_bar_new(); break;
	case VCL_VIEW_MENU:
	case VCL_UB_MENU: msh = gtk_menu_new(); break;
	default: BUG();
	}
	//if ( cd.tlo == tlo_universum ) cd.tlo = NULL; FIXME ?
	fill_menu_with_category( GTK_WIDGET(msh), &main_ctg, location, cd );
	g_object_set_data( G_OBJECT(msh), "vstamp", (gpointer)stamp );
	return GTK_WIDGET(msh);
}


GtkWidget * create_toolbar( unsigned int location, int vertical, struct cmddata cd )
{
	GtkBox * tb = GTK_BOX( vertical ? gtk_vbox_new( FALSE, 1 ) : gtk_hbox_new( FALSE, 1 ));
	//if ( cd.tlo == tlo_universum ) cd.tlo = NULL; FIXME ?
	fill_toolbar_with_category( tb, &main_ctg, location, cd );
	g_object_set_data( G_OBJECT(tb), "vstamp", (gpointer)stamp );
	gtk_widget_show( GTK_WIDGET(tb) );
	return GTK_WIDGET(tb);
}
static void fill_menu_with_category( GtkWidget * menu, struct cmd_category * ctg, unsigned int location, struct cmddata cd )
{
	GtkWidget * cmi, * submenu, * mi, *tear;
	struct cmd_category * cc;
	struct cmd * co;
	ASSERT( ctg && menu );

	if ( ctg != &main_ctg ){
		tear = gtk_tearoff_menu_item_new();
		gtk_menu_shell_append( GTK_MENU_SHELL( menu ), tear);
		gtk_widget_show(tear);
	}

	/* process categories */
	for ( cc = ctg->son; cc; cc = cc->brother ){
		if ( cc->flags & location ){
			cmi = gtk_menu_item_new_with_label( cc->title );
			gtk_menu_item_set_submenu( GTK_MENU_ITEM(cmi), submenu = gtk_menu_new() );
			fill_menu_with_category( submenu, cc, location, cd );
			gtk_menu_shell_append( GTK_MENU_SHELL( menu ), cmi );
			gtk_widget_show(cmi);
			gtk_widget_show(submenu);
		}
	}

	/* process commands */
	for ( co = ctg->first_cmd; co; co = co->next_cmd ){
		if ( co->flags & location ){
			if ( co->flags & VCC_CHECKABLE ) mi = gtk_check_menu_item_new_with_label( co->title );
			else mi = gtk_menu_item_new_with_label( co->title );

			g_object_set_data( G_OBJECT(mi), vrrvw, (gpointer)cd.window );
			g_object_set_data( G_OBJECT(mi), vrrgroup, (gpointer)cd.group );
			gtk_menu_shell_append( GTK_MENU_SHELL( menu ), mi );
			if ( co->accel)
				gtk_widget_add_accelerator( mi, "activate", accel_group, co->accel, co->modifier, GTK_ACCEL_VISIBLE );
			sync_menu_item( co, mi, location );
			g_signal_connect( G_OBJECT(mi), ( co->flags & VCC_CHECKABLE ) ? "toggled" : "activate",
				G_CALLBACK(on_command_activated), (gpointer)(co) );
		}
	}
}
static void fill_toolbar_with_category( GtkBox * tb, struct cmd_category * ctg, unsigned int location, struct cmddata cd )
{
	ASSERT( ctg );
	ASSERT( tb );
	struct cmd * co;
	struct cmd_category * cc;
	GtkWidget * widget;

	for ( co = ctg->first_cmd; co; co = co->next_cmd ){
		if ( co->flags & location ){
			widget = ( co->flags & VCC_CHECKABLE ) ? gtk_toggle_button_new() : gtk_button_new();
			if ( co->icon_path){
				if ( co->flags & VCC_STOCK ) gtk_container_add( GTK_CONTAINER(widget),
					gtk_image_new_from_stock( co->icon_path, (location == VCL_MAIN_TBAR)
						? GTK_ICON_SIZE_LARGE_TOOLBAR : GTK_ICON_SIZE_SMALL_TOOLBAR) );
				else gtk_container_add( GTK_CONTAINER(widget), gtk_image_new_from_pixbuf( co->icon_pixbuf ) );
			}
			gtk_box_pack_start( tb, widget, FALSE, FALSE, 0 );
			gtk_tooltips_set_tip( tooltips, widget, co->description, NULL );
			g_object_set_data( G_OBJECT(widget), vrrvw, (gpointer)cd.window );
			g_object_set_data( G_OBJECT(widget), vrrgroup, (gpointer)cd.group );
			g_signal_connect( G_OBJECT(widget), "button-press-event", G_CALLBACK(command_icon_press), (gpointer)co );
			g_signal_connect( G_OBJECT(widget), "clicked", G_CALLBACK(on_command_activated), (gpointer)co );
			sync_tool_item( co, widget, location, ctg );
		}
	}
	if ( ctg->flags & VCC_CAN_CLPS )
	{
		widget = gtk_toggle_button_new();
		if ( GTK_IS_VBOX(tb) ) gtk_container_add( GTK_CONTAINER(widget), gtk_hseparator_new() );
		else gtk_container_add( GTK_CONTAINER(widget), gtk_vseparator_new() );
		g_object_set( G_OBJECT(widget), "can-focus", FALSE, "relief", GTK_RELIEF_NONE, NULL );
		gtk_box_pack_start( tb, widget, FALSE, FALSE, 0 );
		gtk_tooltips_set_tip( tooltips, widget, "(un)collapse icons", NULL );
		g_signal_connect( G_OBJECT(widget), "clicked", G_CALLBACK(collapse_category), (gpointer)ctg );
	}
	cc = ctg->son;
	for ( cc = ctg->son; cc; cc = cc->brother ){
		if ( cc->flags & location ){
			fill_toolbar_with_category( tb, cc, location, cd );
		}
	}
}

int sync_menu( GtkWidget * menu, unsigned int location ){
	calculate_selection();
	if ( (int)g_object_get_data( G_OBJECT(menu), "vstamp") != stamp ) return 0;
	sync_menu_by_category( menu, &main_ctg, location );
	return 1;
}

int sync_toolbar( GtkWidget * toolbar, unsigned int location )
{
	GList * list, *origlist;
	calculate_selection();
	if ( (int)g_object_get_data( G_OBJECT(toolbar), "vstamp") != stamp ) return 0;
	origlist = list = gtk_container_get_children( GTK_CONTAINER( toolbar ) );
	sync_toolbar_by_category( &main_ctg, location, &list );
	g_list_free(origlist);
	return 1;
}
static void sync_menu_by_category( GtkWidget * menu, struct cmd_category * ctg, unsigned int location )
{
	struct cmd_category * cc;
	struct cmd * co;
	GList * list, * cur;
	GtkMenuItem * mi = NULL;

	list = gtk_container_get_children( GTK_CONTAINER(menu) );

	for( cur = list, cc = ctg->son; cur && cc; ){
		if ( cc->flags & location ){
			//gtk_menu_item_set_label( , cc->title ); TODO!
			mi = GTK_MENU_ITEM(cur->data);
			if(!GTK_IS_TEAROFF_MENU_ITEM(mi)){
				sync_menu_by_category( gtk_menu_item_get_submenu(mi), cc, location );
				cc = cc->brother;
			}
			cur = cur->next; //proceed in the existing items list
		} else cc = cc->brother;
	}
	for ( co = ctg->first_cmd; co && cur; )
		if ( co->flags & location ){
			if(!GTK_IS_TEAROFF_MENU_ITEM(cur->data)){
				sync_menu_item( co, GTK_WIDGET(cur->data), location );
				co = co->next_cmd;
			}
			cur = cur->next;
		} else co = co->next_cmd;
	g_list_free(list);
}


static void sync_toolbar_by_category( struct cmd_category * ctg, unsigned int location, GList ** list )
{
	struct cmd_category * cc;
	struct cmd * co;

	for ( co = ctg->first_cmd; co && *list; co = co->next_cmd ){
		if ( co->flags & location ){
			sync_tool_item( co, GTK_WIDGET( (*list)->data), location, ctg );
			*list = (*list)->next; //proceed in the existing items list
		}
	}
	if ( ctg->flags & VCC_CAN_CLPS )
	{
		sync_in_progress = 1;
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON((*list)->data), !(ctg->flags & VCC_COLLAPSED) );
		sync_in_progress = 0;
		*list = (*list)->next;
	}
	for( cc = ctg->son; cc; cc = cc->brother ){
		if ( cc->flags & location ){
			sync_toolbar_by_category( cc, location, list );
		}
	}
}

void execute_command_item( struct cmd * cmd )
{
	if ( sync_in_progress ) return; //ignore callbacks caused by checking items manually
	ASSERT( cmd );
	//Check if context matches
	if (!context_matches(cmd)) { DBG("Context does not match!\n"); return; }

	if ( !( cmd->flags & VCC_SAFE ) ) factory_op_break();
	if ( cmd->flags & VCC_START_OP ) //start factory operation
	{
		factory_op_start( context.group, cmd->of_state, cmd->description, cmd->flags & VCC_INSIDE_TRANS );
	}
	else if ( cmd->activate_f ) {
		void * param = NULL;
		if ( context.w && WT_TO_VTM(context.w->type) == cmd->type_mask ) param = context.w; //window context
		else
		{
			if ( context.group && cmd->count_request <= 1 ) //local context
			{
				if ( !param && (cmd->type_mask & VTM_CT_GRP) ) param = context.group;
				if ( !param && (cmd->type_mask & VTM_CT_TLO) ) param = context.tlo;
				if ( !param && (cmd->type_mask & VTM_CT_DOC) ) param = context.tlo->obj.father;
			}
		}
		if ( !param && context.o && (( cmd->flags & VCC_META_SEL ) ? context.msel_cnt_all : context.sel_cnt_all ) == 0 &&
			cmd->count_request != 0 )
			param = context.o; //The context matched, so the context.o must be all it needs
		cmd->activate_f( cmd, param );
	}
	refresh_gui();
}

static void on_context_dump( struct cmd * cmd UNUSED, void * subj UNUSED )
{
	msg("\nCurrent context is:\nVIEW = %p, GROUP = %p, O = %p\n",
		context.w, context.group, context.o );
	msg("msel_mask = %x, msel_cnt_all = %u; sel_mask = %x, sel_cnt_all = %u\n",
		context.msel_mask, context.msel_cnt_all, context.sel_mask, context.sel_cnt_all );
	for ( uns u = 0; u < GOT_MAX; u++ )
		if ( context.sel_cnt[u] ) msg("%u %s\n", context.sel_cnt[u], string_go_type[u]->text);
	msg("\n");
}
