#undef LOCAL_DEBUG
#include "lib/lib.h"
#include "kernel/kernel.h"
#include "font/font.h"
#include "gui/main.h"
#include "gui/visualisation.h"
#include "gui/properties.h"

struct factory factory;
static void factory_set_state( struct of_state * state );
static void factory_op_finish( void );
static void factory_inner_moving_start(struct geom_point po);
static void factory_inner_moving_end(void);

// non-public factory states (the public ones are in main.h).
static struct of_state ofs_start, ofs_end;
static struct of_state ofs_point_inters_2;
static struct of_state ofs_segment_2;
static struct of_state ofs_bez_cub_2, ofs_bez_cub_3, ofs_bez_cub_4;
static struct of_state ofs_bez_quad_2, ofs_bez_quad_3;
static struct of_state ofs_arc_3p_2, ofs_arc_3p_3;
static struct of_state ofs_circ_3p_2, ofs_circ_3p_3;
static struct of_state ofs_circ_center_pt_2;
static struct of_state ofs_ellipse_3p_smallest_2, ofs_ellipse_3p_smallest_3;
static struct of_state ofs_ellipse_3p_ecc_2, ofs_ellipse_3p_ecc_3;
static struct of_state ofs_ellipse_foci_pt_2, ofs_ellipse_foci_pt_3;
static struct of_state ofs_ellipse_xy1ecc_2;
static struct of_state ofs_transform_2;

//creating functions
static void create_point( struct of_state * state );
static void create_decorator_pt( struct of_state * state );
static void create_intersection( struct of_state * state );
static void create_segment_1( struct of_state * state );
static void create_bez_cub_1( struct of_state * state );
static void create_bez_quad_1( struct of_state * state );
static void create_arc_3p_1( struct of_state * state );
static void create_circ_3p_1( struct of_state * state );
static void create_circ_center_pt_1( struct of_state * state );
static void create_arc_center_radius_1( struct of_state * state );
static void create_ellipse_3p_smallest_2( struct of_state * state );
static void create_ellipse_3p_ecc_1( struct of_state * state );
static void create_ellipse_segment( struct of_state * state );
static void create_ellipse_foci_pt_2( struct of_state * state );
static void fix_ellipse_anchor_2( struct of_state * state );
static void create_ellarc_center_radii_1( struct of_state * state );
static void create_ellipse_xy1ecc_1( struct of_state * state );
static void create_text_1( struct of_state * state );
static void create_tex_text_1( struct of_state * state );
static void do_transform_1( struct of_state * s );
static void do_transform_2( struct of_state * s );
static void do_rehang_1( struct of_state * s );
static void do_rehang_2( struct of_state * s );

// cleanup functions
/** Decrease reference count of the hanger stored as the input value of the given state;
    or, if the hanger is part of a GO, unref the GO.  */
static void clean_hanger( struct of_state * state );
static void clean_hanger_move( struct of_state * state );
static void clean_move( struct of_state * state );
static void clean_transform( struct of_state * state );
static void clean_go( struct of_state * state );
static void clean_anchor_move( struct of_state * state );
// rehanging functions
static void fix_anchor_2( struct of_state * state );
static void fix_anchor_3( struct of_state * state );
static void fix_anchor_4( struct of_state * state );

#define VAL_INIT {DIST_PASS_POINT, {.point = {0.0,{0.0, 0.0}}}}

//! a special state which means end of operation
static struct of_state ofs_end = { OFIT_POINT, VAL_INIT, "",
				NULL, NULL, NULL, &ofs_end, &ofs_end, NULL, { NULL, NULL } };
//! a special state which means before start of an operation
static struct of_state ofs_start = { OFIT_POINT, VAL_INIT, "",
				NULL, NULL, NULL, &ofs_start, &ofs_start, NULL, { NULL, NULL } };

struct of_state ofs_transform_x = { OFIT_TRANSFORM, VAL_INIT, "Drag the point and transform",
				do_transform_2, NULL, NULL, &ofs_select, &ofs_end, NULL, { NULL, NULL } };

struct of_state ofs_select = { OFIT_SEL, VAL_INIT, "Select mode",
				do_transform_1, clean_transform, NULL, &ofs_start, &ofs_transform_x, NULL, { NULL, NULL } };

static struct of_state ofs_transform_2 = { OFIT_TF, VAL_INIT, "Transform",
				do_transform_2, NULL, NULL, &ofs_transform_1, &ofs_end, NULL, { NULL, NULL } };
struct of_state ofs_transform_1 = { OFIT_TF, VAL_INIT, "Transform mode",
				do_transform_1, clean_transform, NULL, &ofs_start, &ofs_transform_2, NULL, { NULL, NULL } };

struct of_state ofs_rehang_2 = { OFIT_HANGER, VAL_INIT, "Choose the new position of the anchor",
				do_rehang_2, clean_hanger, NULL, &ofs_rehang_1, &ofs_end, NULL, { NULL, NULL } };
struct of_state ofs_rehang_1 = { OFIT_ANCHOR, VAL_INIT, "Choose an anchor",
				do_rehang_1, clean_anchor_move, NULL, &ofs_start, &ofs_rehang_2, NULL, { NULL, NULL } };

static struct of_state ofs_point_inters_2 = { OFIT_GO, VAL_INIT, "Choose the second object for the new intersection",
				create_intersection, clean_go, NULL, &ofs_point_inters, &ofs_end, NULL, { NULL, NULL } };
struct of_state ofs_point_inters = { OFIT_GO, VAL_INIT, "Choose the first object for the new intersection",
				NULL, clean_go, NULL, &ofs_start, &ofs_point_inters_2, NULL, { NULL, NULL } };

struct of_state ofs_point = { OFIT_HANGER, VAL_INIT, "Choose a position for the new point",
				create_point, clean_hanger, NULL, &ofs_start, &ofs_end, NULL, { NULL, NULL } };

struct of_state ofs_decorator_pt = { OFIT_HANGER, VAL_INIT, "Choose a position for the new decoration point",
				create_decorator_pt, clean_hanger, NULL, &ofs_start, &ofs_end, NULL, { NULL, NULL } };

static struct of_state ofs_segment_2 = { OFIT_HANGER, VAL_INIT, "Choose the endpoint of the new segment",
				fix_anchor_2, clean_hanger, NULL, &ofs_segment_1, &ofs_end, NULL, { NULL, NULL } };
struct of_state ofs_segment_1 = { OFIT_HANGER, VAL_INIT, "Choose the startpoint of the new segment",
				create_segment_1, clean_hanger_move, NULL, &ofs_start, &ofs_segment_2, NULL, { NULL, NULL } };

//FIXME: wieghts!
static struct of_state ofs_bez_cub_4 = { OFIT_HANGER, VAL_INIT, "Choose the endpoint of the new cubic Bezier curve",
				fix_anchor_4, clean_hanger, NULL, &ofs_bez_cub_3, &ofs_end, NULL, { NULL, NULL } };
static struct of_state ofs_bez_cub_3 = { OFIT_HANGER, VAL_INIT, "Choose the second control point of the new cubic Bezier curve",
				fix_anchor_3, clean_hanger, NULL, &ofs_bez_cub_2, &ofs_bez_cub_4, NULL, { NULL, NULL } };
static struct of_state ofs_bez_cub_2 = { OFIT_HANGER, VAL_INIT, "Choose the first control point of the new cubic Bezier curve",
				fix_anchor_2, clean_hanger, NULL, &ofs_bez_cub_1, &ofs_bez_cub_3, NULL, { NULL, NULL } };
struct of_state ofs_bez_cub_1 = { OFIT_HANGER, VAL_INIT, "Choose the startpoint of the new cubic Bezier curve",
				create_bez_cub_1, clean_hanger_move, NULL, &ofs_start, &ofs_bez_cub_2, NULL, { NULL, NULL } };

static struct of_state ofs_bez_quad_3 = { OFIT_HANGER, VAL_INIT, "Choose the endpoint of the new quadratic Bezier curve",
				fix_anchor_3, clean_hanger, NULL, &ofs_bez_quad_2, &ofs_end, NULL, { NULL, NULL } };
static struct of_state ofs_bez_quad_2 = { OFIT_HANGER, VAL_INIT, "Choose the control point of the new quadratic Bezier curve",
				fix_anchor_2, clean_hanger, NULL, &ofs_bez_quad_1, &ofs_bez_quad_3, NULL, { NULL, NULL } };
struct of_state ofs_bez_quad_1 = { OFIT_HANGER, VAL_INIT, "Choose the startpoint of the new quadratic Bezier curve",
				create_bez_quad_1, clean_hanger_move, NULL, &ofs_start, &ofs_bez_quad_2, NULL, { NULL, NULL } };

static struct of_state ofs_arc_3p_3 = { OFIT_HANGER, VAL_INIT, "Choose the endpoint of the new circular arc",
				fix_anchor_3, clean_hanger, NULL, &ofs_arc_3p_2, &ofs_end, NULL, { NULL, NULL } };
static struct of_state ofs_arc_3p_2 = { OFIT_HANGER, VAL_INIT, "Choose the midpoint of the new circular arc",
				fix_anchor_2, clean_hanger, NULL,  &ofs_arc_3p_1, &ofs_arc_3p_3, NULL, { NULL, NULL } };
struct of_state ofs_arc_3p_1 = { OFIT_HANGER, VAL_INIT, "Choose the startpoint of the new circular arc",
				create_arc_3p_1, clean_hanger_move, NULL,  &ofs_start, &ofs_arc_3p_2, NULL, { NULL, NULL } };

static struct of_state ofs_circ_3p_3 = { OFIT_HANGER, VAL_INIT, "Choose the third point on the new circle",
				fix_anchor_3, clean_hanger, NULL, &ofs_circ_3p_2, &ofs_end, NULL, { NULL, NULL } };
static struct of_state ofs_circ_3p_2 = { OFIT_HANGER, VAL_INIT, "Choose the second point on the new circle",
				fix_anchor_2, clean_hanger, NULL, &ofs_circ_3p_1, &ofs_circ_3p_3, NULL, { NULL, NULL } };
struct of_state ofs_circ_3p_1 = { OFIT_HANGER, VAL_INIT, "Choose the first point on the new circle",
				create_circ_3p_1, clean_hanger_move, NULL, &ofs_start, &ofs_circ_3p_2, NULL, { NULL, NULL } };

static struct of_state ofs_circ_center_pt_2 = { OFIT_HANGER, VAL_INIT, "Choose the point on the new circle",
				fix_anchor_2, clean_hanger, NULL, &ofs_circ_center_pt_1, &ofs_end, NULL, { NULL, NULL } };
struct of_state ofs_circ_center_pt_1 = { OFIT_HANGER, VAL_INIT, "Choose the center point of the new circle",
				create_circ_center_pt_1, clean_hanger_move, NULL, &ofs_start, &ofs_circ_center_pt_2, NULL, { NULL, NULL } };

struct of_state ofs_arc_center_radius_1 = { OFIT_HANGER, VAL_INIT, "Choose the center of the new arc",
				create_arc_center_radius_1, clean_hanger, NULL, &ofs_start, &ofs_end, NULL, { NULL, NULL } };

struct of_state ofs_circ_center_radius_1 = { OFIT_HANGER, VAL_INIT, "Choose the center of the new circle",
				create_arc_center_radius_1, clean_hanger, NULL, &ofs_start, &ofs_end, NULL, { NULL, NULL } };


static struct of_state ofs_ellipse_3p_smallest_3 = { OFIT_HANGER, VAL_INIT, "Choose the third point on the new ellipse",
				fix_ellipse_anchor_2, clean_hanger, NULL, &ofs_ellipse_3p_smallest_2, &ofs_end,
				NULL, { NULL, NULL } };
static struct of_state ofs_ellipse_3p_smallest_2 = { OFIT_HANGER, VAL_INIT, "Choose the second point on the new ellipse",
				create_ellipse_3p_smallest_2, clean_hanger, NULL, &ofs_ellipse_3p_smallest_1, &ofs_ellipse_3p_smallest_3,
				NULL, { NULL, NULL } };
struct of_state ofs_ellipse_3p_smallest_1 = { OFIT_HANGER, VAL_INIT, "Choose the first point on the new ellipse",
				create_ellipse_segment, clean_hanger_move, NULL, &ofs_start, &ofs_ellipse_3p_smallest_2,
				NULL, { NULL, NULL } };

static struct of_state ofs_ellipse_3p_ecc_3 = { OFIT_HANGER, VAL_INIT, "Choose the third point on the new ellipse",
				fix_anchor_3, clean_hanger, NULL, &ofs_ellipse_3p_ecc_2, &ofs_end,
				NULL, { NULL, NULL } };
static struct of_state ofs_ellipse_3p_ecc_2 = { OFIT_HANGER, VAL_INIT, "Choose the second point on the new ellipse",
				fix_anchor_2, clean_hanger, NULL, &ofs_ellipse_3p_ecc_1, &ofs_ellipse_3p_ecc_3,
				NULL, { NULL, NULL } };
struct of_state ofs_ellipse_3p_ecc_1 = { OFIT_HANGER, VAL_INIT, "Choose the first point on the new ellipse",
				create_ellipse_3p_ecc_1, clean_hanger_move, NULL, &ofs_start, &ofs_ellipse_3p_ecc_2,
				NULL, { NULL, NULL } };

struct of_state ofs_ellarc_center_radii_1 = { OFIT_HANGER, VAL_INIT, "Choose the center of the new elliptic arc",
				create_ellarc_center_radii_1, clean_hanger, NULL, &ofs_start, &ofs_end, NULL, { NULL, NULL } };

struct of_state ofs_ellipse_center_radii_1 = { OFIT_HANGER, VAL_INIT, "Choose the center of the new ellipse",
				create_ellarc_center_radii_1, clean_hanger, NULL, &ofs_start, &ofs_end, NULL, { NULL, NULL } };

static struct of_state ofs_ellipse_foci_pt_3 = { OFIT_HANGER, VAL_INIT, "Choose a point on the new ellipse",
				fix_ellipse_anchor_2, clean_hanger, NULL, &ofs_ellipse_foci_pt_2, &ofs_end,
				NULL, { NULL, NULL } };
static struct of_state ofs_ellipse_foci_pt_2 = { OFIT_HANGER, VAL_INIT, "Choose the second focus of the new ellipse",
				create_ellipse_foci_pt_2, clean_hanger, NULL, &ofs_ellipse_foci_pt_1, &ofs_ellipse_foci_pt_3,
				NULL, { NULL, NULL } };
struct of_state ofs_ellipse_foci_pt_1 = { OFIT_HANGER, VAL_INIT, "Choose the first focus of the new ellipse",
				create_ellipse_segment, clean_hanger_move, NULL, &ofs_start, &ofs_ellipse_foci_pt_2,
				NULL, { NULL, NULL } };

static struct of_state ofs_ellipse_xy1ecc_2 = { OFIT_HANGER, VAL_INIT, "Choose the point on the new ellipse",
				fix_anchor_2, clean_hanger, NULL, &ofs_ellipse_xy1ecc_1, &ofs_end, NULL, { NULL, NULL } };
struct of_state ofs_ellipse_xy1ecc_1 = { OFIT_HANGER, VAL_INIT, "Choose the center point of the new ellipse",
				create_ellipse_xy1ecc_1, clean_hanger_move, NULL, &ofs_start, &ofs_ellipse_xy1ecc_2, NULL, { NULL, NULL } };

struct of_state ofs_text_1 = { OFIT_HANGER, VAL_INIT, "Choose the refpoint of the new text",
				create_text_1, clean_hanger, NULL, &ofs_start, &ofs_end, NULL, { NULL, NULL } };

struct of_state ofs_tex_text_1 = { OFIT_HANGER, VAL_INIT, "Choose the refpoint of the new TeX text",
				create_tex_text_1, clean_hanger, NULL, &ofs_start, &ofs_end, NULL, { NULL, NULL } };

static string noname_op;

static int tsort_on;
#define TSORT_ON { tsort_hanger_presort( factory.hanger_cursor, factory.tlo ); tsort_on = 1; }
#define TSORT_OFF { tsort_presort_end( factory.tlo ); tsort_on = 0; }

void factory_init(void)
{
	factory.operation_active = 0;
	factory.start_state = &ofs_select;
	factory.tlo = NULL;
	factory.group = NULL;
	factory.hanger_cursor = NULL;

	factory.snap_tol_pix = prop_lookup_uns(PROP_STORE_O(ps_global), string_tolerance, TOLERANCE_PIX);
	factory.snap_set = SNAP_HANGERS | SNAP_DEPENDENCY;

	factory_set_state( &ofs_select );
	factory.desc = noname_op = string_ref(string_lookup_ref("Anonymous operation"));
}
void factory_finish(void)
{
	if ( factory.desc) string_unref(factory.desc);
	string_unref( noname_op );
	if ( factory.group ) go_unref( &factory.group->go );
	if ( factory.tlo ) obj_unref( &factory.tlo->obj );
}

void factory_op_start( struct go_group * group, struct of_state * state, char * desc, int inside_trans )
{
	string s;
	ASSERT_GROUP(group);
	ASSERT(factory.desc);
	ASSERT(state && state->prev == &ofs_start );
	DBG("factory_op_start called for group %p, state %p\n", group, state );

	if ( factory.operation_active ) factory_op_finish();
	if ( factory.group ) go_unref( &factory.group->go );
	if ( factory.tlo ) obj_unref( &factory.tlo->obj );
	factory.group = group;
	factory.tlo = go_get_tlo( &group->go );
	go_ref( &group->go );
	obj_ref( &factory.tlo->obj );

	factory.start_state = state;
	factory.inside_trans = inside_trans;

	s = factory.desc;
	factory.desc = ( !desc ) ? string_ref(factory.desc) : string_lookup_ref(desc);
	string_unref(s);

	factory_set_state( state );
}

static void factory_set_state( struct of_state * state )
{
	DBG("factory_set_state, last state = %p, new state = %p\n", factory.state, state );
	factory.state = state;
	factory.state->prev_undo_item = (factory.tlo) ? factory.tlo->undo_present : NULL;

	if ( state != &ofs_end )
		for ( GSList * l = view_list; l; l = l->next )
			statusbar_msg( &((struct view*)l->data)->status, factory.state->description );

	if ( factory.state->input_type == OFIT_FUNC )
	{
		ASSERT( factory.state->input_func );
		DBG("calling input function %p\n", factory.state->input_func );
		factory.state->input_func( factory.state );
	}

	if ( state == &ofs_end ) factory_op_finish();
	else refresh_gui();

	//wait for input from gui
}

void factory_moving_step(struct geom_point po UNUSED)
{
	if (moving_mode != MOVE_FACTORY) return; //ignore unwanted events
	char err[TRANS_ERR_SIZE];
	volatile int errval = 0;
	TRANS_BEGIN_MAIN(factory.tlo, factory.desc->text, err)
	{
		if(hanger_mc_move( factory.hanger_cursor, &po ))
			tsort_presorted_recompute( factory.tlo );
	}
	TRANS_FAILED  //ignore
		errval = 1;
	TRANS_END;
	if (!errval) factory_merge_trans();
}

static void factory_inner_moving_start(struct geom_point po)
{
	ASSERT( moving_mode == MOVE_NONE && !factory.hanger_cursor );
	struct geom_point pi;
	struct view * vw = (struct view*)context.w;
	factory.hanger_cursor = hanger_new_mc(&po);
	hanger_ref( factory.hanger_cursor );
	TSORT_ON; // all newly created things must hang on the cursor hanger
	visualisation_coords_to_pixels( &(vw->vis), &po, &pi );
	view_moving_start( pi, vw, MOVE_FACTORY, GDK_CROSSHAIR );
}

static void factory_inner_moving_end(void)
{
	if ( moving_mode != MOVE_FACTORY ) { DBG("Warning: move mode not factory!\n"); return; }
	view_moving_end();
	TSORT_OFF;
	hanger_unref( factory.hanger_cursor );
	factory.hanger_cursor = NULL;
}

static int factory_merge_trans_from(struct of_state * state)
{
	ASSERT( factory.operation_active );
	struct undo_gui * ug, * start = state->prev_undo_item;
	// merge all new partial transactions into one
	ug = start ? trans_get_next_gui( factory.tlo, start ) : trans_get_first_gui( factory.tlo );
	if ( ug )
		while ( ug != factory.tlo->undo_present )
			ug = trans_merge_undo( factory.tlo, ug, factory.desc );
	return ug != NULL;
}

/** Merge all transactions of the current factory operation state -- i.e., after
 factory.state->prev_undo_item of the current state -- into one.
 Returns nonzero if there was at least one new transaction since that. */
int factory_merge_trans(void)
{ return factory_merge_trans_from( factory.state ); }

static void factory_op_finish( void )
{
	DBG("factory_op_finish called\n");
	ASSERT( factory.operation_active );

	//cleanup after all actions
	for ( struct of_state * ofs = factory.start_state; ofs != &ofs_end; ofs = ofs->next )
		if ( ofs->cleanup_func ) ofs->cleanup_func( ofs );

	if ( moving_mode != MOVE_NONE ) view_moving_end();
	ASSERT( moving_mode == MOVE_NONE );

	// merge all new partial transactions into one
	if ( factory_merge_trans_from( factory.start_state ) )
		trans_rename_undo( factory.tlo, factory.tlo->undo_present, factory.desc );
	factory.start_state->prev_undo_item = factory.tlo->undo_present;

	factory.operation_active = 0;
	factory_set_state( factory.start_state );
}

void factory_op_break( void )
{
	char s[TRANS_ERR_SIZE];
	DBG("factory_op_break called\n");
	if ( ! factory.operation_active ) return;

	//do the cleanup first - must be before undo because of tsort!
	for ( struct of_state * ofs = factory.state->prev; ofs != &ofs_start; ofs = ofs->prev )
		if ( ofs->cleanup_func) ofs->cleanup_func( ofs );

	// merge all new partial transactions into one
	if ( factory_merge_trans_from( factory.start_state ) )
	{
		if (!trans_undo( factory.tlo, s )) report_error_f("%s\n", s);
		trans_clear_one_redo( factory.tlo );
	}
	factory.operation_active = 0;
	factory_set_state( factory.start_state );
}

void on_step_back( struct cmd * cmd UNUSED, void * subj UNUSED ) { factory_op_step_back(); }
void on_break_op( struct cmd * cmd UNUSED, void * subj UNUSED ) { factory_op_break(); }

void factory_op_step_back(void)
{
	char s[TRANS_ERR_SIZE];
	DBG("factory_op_step_back called\n");

	if ( factory.state == factory.start_state ) { report_error("Cannot step back.\n"); return; }
	if ( factory.state->prev == factory.start_state ) { factory_op_break(); return; }

	struct geom_point po;  //if moving, save the last position
	if ( moving_mode == MOVE_FACTORY && factory.hanger_cursor ) po = hanger_get_position(factory.hanger_cursor);

	int tsort_active = tsort_on;
	if ( tsort_active ) TSORT_OFF;

	struct of_state * prev = factory.state->prev;
	if ( prev->cleanup_func ) prev->cleanup_func( prev );

	// undo all new partial transactions
	if ( factory_merge_trans_from( factory.state->prev ) )
	{
		if (!trans_undo( factory.tlo, s )) report_error_f("%s\n", s);
		trans_clear_one_redo( factory.tlo );
	}
	factory_set_state( prev );

	if ( tsort_active ) TSORT_ON;

	//restore the moving position
	if ( moving_mode == MOVE_FACTORY ) factory_moving_step(po);
}

static void reform_input_to_hanger( void )
{
	struct dist_pass_result * dpr = &factory.state->value;
	switch ( dpr->type )
	{
	case DIST_PASS_POINT:
		dpr->data.hanger.hanger = hanger_new_mc( &dpr->data.point.point );
		break;
	case DIST_PASS_GO:
	{
		if ( factory.snap_set & SNAP_DEPENDENCY )
		{
			//create a parametric point
			DBG("Factory: creating a parametric point at %p, time %.3f\n", dpr->data.go.go, dpr->data.go.geom.time );
			struct go_parametric_point * gp;
			gp = go_parametric_point_new( dpr->data.go.go, dpr->data.go.geom.time );
			go_ref( go_link( &gp->go, factory.group, NULL ) );
			factory.state->tmp_go[OFR_HANGER_OWNER] = &gp->go;
			dpr->data.hanger.hanger = go_get_hanger( &gp->go, 0 );
		} else
			dpr->data.hanger.hanger = hanger_new_mc( &dpr->data.go.geom.point );
		break;
	}
	case DIST_PASS_INTERSECTION:
	{
		if ( factory.snap_set & SNAP_DEPENDENCY )
		{
			//create an intersection
			DBG("Factory: creating an intersection of %p and %p, param = %.3f\n", dpr->data.intersection.go1,
				dpr->data.intersection.go2, dpr->data.intersection.geom.time1 );
			struct go_intersection * gp = go_intersection_new( dpr->data.intersection.go1,
				dpr->data.intersection.go2, dpr->data.intersection.geom.time1 );
			go_ref( go_link( &gp->go, factory.group, NULL ) );
			factory.state->tmp_go[OFR_HANGER_OWNER] = &gp->go;
			dpr->data.hanger.hanger = go_get_hanger( &gp->go, 0 );
		} else
			dpr->data.hanger.hanger = hanger_new_mc( &dpr->data.intersection.geom.point );
		break;
	}
	case DIST_PASS_HANGER:
		if (!( factory.snap_set & SNAP_DEPENDENCY ))
		{
			struct geom_point pt = hanger_get_position( dpr->data.hanger.hanger );
			dpr->data.hanger.hanger = hanger_new_mc( &pt );
		}
		break;
	default: BUG();
	}
	ASSERT( dpr->data.hanger.hanger );
	hanger_ref( dpr->data.hanger.hanger );
	dpr->type = DIST_PASS_HANGER;
}

void factory_op_step( void )
{
	char err[TRANS_ERR_SIZE];
	volatile int success = 1;
	DBG("factory_op_step\n");
	if ( !factory.operation_active )
	{
		factory.operation_active = 1; //operation becomes active after the first step
		// update the pointer in case that someone has changed undo history since the last mode switch
		factory.start_state->prev_undo_item = factory.tlo->undo_present;
	}

	if ( (factory.state->action_func)  && factory.inside_trans || factory.state->input_type != factory.state->value.type )
	{
		TRANS_BEGIN_MAIN( factory.tlo, factory.desc->text, err )
			// Some not yet existing hangers may be passed as something else
			if ( (factory.state->input_type == OFIT_HANGER) ) reform_input_to_hanger();

			factory.state->action_func( factory.state );
		TRANS_FAILED
			success = 0;
			report_error_f("Invalid thing chosen: %s Please try again.\n", err );
			if ( factory.hanger_cursor && !tsort_on )
				TSORT_ON; //the trans_fail came while tsort was temporarily switched off
			if ( factory.state->cleanup_func ) factory.state->cleanup_func( factory.state );
		TRANS_END

		if ( success )
		{
			factory_merge_trans_from( factory.state ); //leave just one item after the state
			factory_set_state( factory.state->next ); //all succeeded, go to next state
		}
		else if ( factory.state->input_type == OFIT_FUNC )
		{
			report_error("Input function result invalid. Process cancelled.\n");
			factory_op_break();
		}
	}
	else
	{
		if ( factory.state->action_func ) factory.state->action_func( factory.state );
		factory_set_state( factory.state->next );
	}
}

struct go * factory_get_created_go(void)
{
	if ( !factory.operation_active ) return NULL;
	if ( factory.start_state->next->tmp_go[OFR_DISPLAY] ) return factory.start_state->next->tmp_go[OFR_DISPLAY];
	if ( factory.start_state->tmp_go[OFR_DISPLAY] ) return factory.start_state->tmp_go[OFR_DISPLAY];
	return NULL;
}
/**** CREATION OF GO ****/

void snap_point( struct geom_point * pos, struct dist_pass_result * dpr, struct geom_transform2 * grid, struct obj_tlo * tlo, real maxdist )
{
	volatile struct geom_point minp;
	volatile real mindist = 0.0, dist;
	volatile int isnotset = 1;

	struct dist_pass_result info;
	struct center_pass cp;
	char err[TRANS_ERR_SIZE];
	int valid;
	volatile uns pass_flags = 0;
	if ( factory.snap_set & SNAP_HANGERS ) pass_flags |= DIST_PASS_HANGERS;
	if ( factory.snap_set & SNAP_LINES ) pass_flags |= DIST_PASS_GOS;
	if ( factory.snap_set & SNAP_INTERS ) pass_flags |= DIST_PASS_INTERSECTIONS;

	if ( pass_flags )
	{
		TRANS_BEGIN_MAIN( tlo, "Snap the point", err )
			center_pass_init(&cp, CENTER_PASS_FLAGS | pass_flags, NULL, pos, maxdist, tlo);
			while ( valid = center_pass_next( &cp, &info ) )
			{
				switch (info.type)
				{
				case DIST_PASS_GO:
					if (tsort_is_go_presorted(info.data.go.go)) continue;
					break;
				case DIST_PASS_HANGER:
					if (!hanger_is_position(info.data.hanger.hanger)) continue;
					if (tsort_is_go_presorted(hanger_get_go(info.data.hanger.hanger))) continue;
					break;
				case DIST_PASS_INTERSECTION:
					if (tsort_is_go_presorted(info.data.intersection.go1) ||
						tsort_is_go_presorted(info.data.intersection.go2)) continue;
					break;
				default:;
				}
				break; // if none of the bad cases matches, we have found a good one
			}
			if (valid)
			{
				isnotset = 0;
				switch (info.type)
				{
				case DIST_PASS_GO:
					minp = info.data.go.geom.point;
					mindist = info.data.go.geom.dist;
					DBG("Snapping to go %p, time = %.3f\n", info.data.go.go, info.data.go.geom.time );
					break;
				case DIST_PASS_HANGER:
					minp = hanger_get_position(info.data.hanger.hanger);
					mindist = info.data.hanger.dist;
					DBG("Snapping to hanger %p\n", info.data.hanger.hanger );
					break;
				case DIST_PASS_INTERSECTION:
					minp = info.data.intersection.geom.point;
					mindist = info.data.intersection.dist;
					DBG("Snapping to intersection of %p and %p\n", info.data.intersection.go1, info.data.intersection.go2);
					break;
				default:;
				}
			}
			center_pass_destroy(&cp);
		TRANS_FAILED
			report_error_f("%s\n", err);
		TRANS_END
	}

	if ( factory.snap_set & SNAP_GRID ){
		struct geom_point gridpt;
		int res = geom_point_transform( pos, &grid->inverted, &gridpt );
		if ( res == 0 )
		{
			gridpt.x = floor( gridpt.x + 0.5 );
			gridpt.y = floor( gridpt.y + 0.5 );
			res = geom_point_transform( &gridpt, &grid->primary, &gridpt );
		} else
			report_error("snapping to grid: geom_transform failed\n");

		if ( res == 0 )
		{
			dist = hypot( pos->x - gridpt.x, pos->y - gridpt.y );
			if ( isnotset && dist < maxdist || !isnotset && dist < mindist )
			{
				mindist = dist; minp = gridpt;
				isnotset = 0;
				info.type = DIST_PASS_POINT;
				info.data.point.point = gridpt; info.data.point.dist = dist;
				DBG("Snapping to grid point %.3f, %.3f\n", gridpt.x, gridpt.y );
			}
		} else
			report_error("snapping to grid: geom_transform back failed\n");
	}

	// If no suitable result has been found, fall back to the original position.
	if ( isnotset )
	{
		mindist = 0.0; minp = *pos;
		info.type = DIST_PASS_POINT; info.data.point.point = minp;
		info.data.point.dist = 0.0;
	}
	DBG("Resulting point [%.3f, %.3f] in dist = %.3f\n", minp.x, minp.y, mindist);
	if (dpr) *dpr = info;
	*pos = minp;
}

void snap_to_go( struct geom_point * pos, struct dist_pass_result * dpr, struct obj_tlo * tlo, real maxdist )
{
	struct dist_pass_result info;
	struct center_pass cp;
	char err[TRANS_ERR_SIZE];
	int valid;
	TRANS_BEGIN_MAIN( tlo, "Find the nearest go", err )
		center_pass_init(&cp, CENTER_PASS_FLAGS | DIST_PASS_GOS, NULL, pos, maxdist, tlo);
		while ( (valid = center_pass_next( &cp, &info )) && tsort_is_go_presorted(info.data.go.go) );
		if (valid)
		{
			struct geom_point ap = info.data.go.geom.point;
			DBG("Snapped point %.3f, %.3f to %.3f, %.3f (go %p), time %.3f\n", pos->x, pos->y, ap.x, ap.y,
				info.data.go.go, info.data.go.geom.time );
			*pos = ap;
		}
		else
		{
			DBG("No near go found.\n");
			info.type = DIST_PASS_POINT;
			info.data.point.point = *pos; info.data.point.dist = 0.0;
		}
		if (dpr) *dpr = info;
		center_pass_destroy(&cp);
	TRANS_FAILED
		report_error_f("%s\n", err);
	TRANS_END
}

void snap_to_anchor( struct geom_point * pos, struct dist_pass_result * dpr, struct obj_tlo * tlo, real maxdist )
{
	struct dist_pass_result info;
	struct center_pass cp;
	char err[TRANS_ERR_SIZE];
	int valid;
	TRANS_BEGIN_MAIN( tlo, "Find the nearest anchor", err )
		center_pass_init(&cp, CENTER_PASS_FLAGS | DIST_PASS_ANCHORS, NULL, pos, maxdist, tlo);
		while ( (valid = center_pass_next( &cp, &info )) &&
			(tsort_is_go_presorted(anchor_get_go(info.data.anchor.anchor)) ||
			 !go_is_selected(anchor_get_go(info.data.anchor.anchor)) ||
			 !hanger_is_position(info.data.anchor.anchor->h))); // FIXME: is this necessary?
		if (valid)
		{
			struct geom_point ap = hanger_get_position(info.data.anchor.anchor->h);
			DBG("Snapped point %.3f, %.3f to %.3f, %.3f (anchor %p) over distance %.3f\n",
				pos->x, pos->y, ap.x, ap.y, info.data.anchor.anchor, info.data.anchor.dist );
			*pos = ap;
		}
		else
		{
			DBG("No anchor found.\n");
			info.type = DIST_PASS_POINT;
			info.data.point.point = *pos; info.data.point.dist = 0.0;
		}
		if (dpr) *dpr = info;
		center_pass_destroy(&cp);
	TRANS_FAILED
		report_error_f("%s\n", err);
	TRANS_END
}

void clean_hanger( struct of_state * state )
{
	if ( state->tmp_go[OFR_HANGER_OWNER] )
	{
		// If this is cleanup, the go has been unlinked already by undo
		go_unref( state->tmp_go[OFR_HANGER_OWNER] );
		state->tmp_go[OFR_HANGER_OWNER] = NULL;
	}
	if ( state->tmp_go[OFR_DISPLAY] )
	{
		go_unref( state->tmp_go[OFR_DISPLAY] );
		state->tmp_go[OFR_DISPLAY] = NULL;
	}
	if ( state->value.type == OFIT_HANGER )
	{
		if ( state->value.data.hanger.hanger ) hanger_unref( state->value.data.hanger.hanger );
	}
	else state->value.type = OFIT_HANGER;
	state->value.data.hanger.hanger = NULL;
}
void clean_hanger_move( struct of_state * state ) { clean_hanger(state); clean_move(state); }
void clean_move( struct of_state * state UNUSED ) { if ( moving_mode == MOVE_FACTORY ) factory_inner_moving_end(); }
void clean_go( struct of_state * state )
{
	if ( state->value.type == OFIT_GO )
	{ if ( state->value.data.go.go ) go_unref(state->value.data.go.go); }
	else state->value.type = OFIT_GO;

	state->value.data.go.go = NULL;
}
static void clean_anchor_move( struct of_state * state )
{
	go_unref(anchor_get_go(state->value.data.anchor.anchor));
	clean_move(state);
}

void fix_anchor_2( struct of_state * state )
{
	ASSERT( state->prev->tmp_go[OFR_DISPLAY] );
	anchor_rehang( go_get_anchor(state->prev->tmp_go[OFR_DISPLAY], 1), state->value.data.hanger.hanger );
}
void fix_anchor_3( struct of_state * state )
{
	ASSERT( state->prev->prev->tmp_go[OFR_DISPLAY] );
	anchor_rehang( go_get_anchor(state->prev->prev->tmp_go[OFR_DISPLAY], 2), state->value.data.hanger.hanger );
}
void fix_anchor_4( struct of_state * state )
{
	ASSERT( state->prev->prev->prev->tmp_go[OFR_DISPLAY] );
	anchor_rehang( go_get_anchor(state->prev->prev->prev->tmp_go[OFR_DISPLAY], 3), state->value.data.hanger.hanger );
}
static void perturb_cursor(void)
{
	struct geom_point po = hanger_get_position(factory.hanger_cursor);
	po.x += factory.snap_tolerance;
	po.y += factory.snap_tolerance;
	hanger_mc_move( factory.hanger_cursor, &po );
}

/* transformation */

void do_transform_1( struct of_state * s UNUSED )
{
	ASSERT(factory.group);
	tsort_selected_flag_presort( factory.group, factory.tlo );
}

void do_transform_2( struct of_state * s UNUSED ) { }

static void clean_transform( struct of_state * s UNUSED )
{
	view_moving_end();
	tsort_presort_end( factory.tlo );
}

/* Rehanging of anchors */

void do_rehang_1( struct of_state * s UNUSED )
{
	struct anchor * a = ofs_rehang_1.value.data.anchor.anchor;
	struct go * go = anchor_get_go(a);
	go_ref(go);
	factory_inner_moving_start( hanger_get_position(a->h) );
	tsort_presort_end( factory.tlo ); // FIXME: Volat tsort_presort na prazdnou mnozinu (na nove vytvorene mysi nic nevisi) a vzdy hned tsort zrusit mi pripada divne.
	tsort_go_presort( go, factory.tlo );
	anchor_rehang( a, factory.hanger_cursor );
	tsort_presorted_recompute( factory.tlo ); // Pokud je mys na stejnem miste jako puvodni hanger, tak je to mozna zbytecne.
}

void do_rehang_2( struct of_state * s UNUSED )
{
	anchor_rehang( ofs_rehang_1.value.data.anchor.anchor, ofs_rehang_2.value.data.hanger.hanger );
}

#define MS(_type) \
	struct _type * g; \
	struct hanger * h = state->value.data.hanger.hanger; \
	factory_inner_moving_start( hanger_get_position(h) ); \
	struct hanger * hc = factory.hanger_cursor

#define ML \
	state->tmp_go[OFR_DISPLAY] = &g->go; \
	go_link( go_ref( &g->go ), factory.group, NULL ); \
	gui_prop_recycle( go2o(&g->go) ); \
	change_context( context.w, factory.group, go2o(&g->go) )

#define MLT \
	state->tmp_go[OFR_DISPLAY] = &g->go; \
	TSORT_OFF; \
	go_link( go_ref( &g->go ), factory.group, NULL ); \
	TSORT_ON; \
	gui_prop_recycle( go2o(&g->go) ); \
	change_context( context.w, factory.group, go2o(&g->go) )

/* Point */
void create_point( struct of_state * state )
{
	struct go_point * g = go_point_new( ofs_point.value.data.hanger.hanger );
	ML;
}

void create_decorator_pt( struct of_state * state )
{
	struct go_decorator_point * g = go_decorator_point_new( state->value.data.hanger.hanger, 0, 2.0, 0.0 );
	ML;
}
static void create_intersection( struct of_state * state UNUSED )
{
	struct go_intersection * g = go_intersection_new( ofs_point_inters.value.data.go.go,
		ofs_point_inters_2.value.data.go.go, ofs_point_inters.value.data.go.geom.time );
	ML;
}

/* Segment */
void create_segment_1( struct of_state * state )
{
	MS(go_segment);
	g = go_segment_new( h, hc );
	MLT;
}

/* Cubic Bezier */
void create_bez_cub_1( struct of_state * state )
{
	MS(go_bezier);
	g = go_bezier_new_cubic( h, 1.0, hc, 1.0, hc, 1.0, hc, 1.0 );
	MLT;
}

/* Quadratic Bezier */
void create_bez_quad_1( struct of_state * state )
{
	MS(go_bezier);
	g = go_bezier_new_quadratic( h, 1.0, hc, 1.0, hc, 1.0 );
	MLT;
}

/* Arc by three points */
void create_arc_3p_1( struct of_state * state UNUSED )
{
	MS(go_elarc);
	perturb_cursor();
	g = go_elarc_new_3ecc( h, hc, hc, 0.0, 0.0, GEOM_CONIC_ARC_CCW );
	MLT;
}
/* Circle/circular arc by center and radius */
void create_arc_center_radius_1( struct of_state * state )
{
	struct go_elarc * g = go_elarc_new_xyr( state->value.data.hanger.hanger, 0.0, 10.0, 10.0, 0.0,
		(state == &ofs_arc_center_radius_1) ? 2.0 : R_2PI, GEOM_CONIC_ARC_START_DIF );
	ML;
}
/* Circle by three points */
void create_circ_3p_1( struct of_state * state UNUSED )
{
	MS(go_elarc);
	//perturb_cursor();
	g = go_elarc_new_3ecc( h, hc, hc, 0.0, 0.0, GEOM_CONIC_ARC_POINT_ENTIRE );
	MLT;
}

/* Circle by center and point */
void create_circ_center_pt_1( struct of_state * state UNUSED )
{
	MS(go_elarc);
	g = go_elarc_new_xy1ecc( h, hc, 0.0, 0.0, GEOM_CONIC_ARC_POINT_ENTIRE );
	MLT;
}

/* Smallest ellipse by 3 points */
void create_ellipse_3p_smallest_2( struct of_state * state UNUSED )
{
	TSORT_OFF;
	go_unlink( state->prev->tmp_go[OFR_DISPLAY] );//unlink the old go
	perturb_cursor();
	struct go_elarc * g = go_elarc_new_3small( state->prev->value.data.hanger.hanger,
		state->value.data.hanger.hanger, factory.hanger_cursor, GEOM_CONIC_ARC_POINT_ENTIRE );
//	go_link( go_ref( &g->go ), factory.group, NULL );
//	change_context( context.w, factory.group, go2o(&g->go) );
	ML;
	TSORT_ON;
}
/* Ellipse by 3 points, rotation and eccentricity */
void create_ellipse_3p_ecc_1( struct of_state * state UNUSED )
{
	MS(go_elarc);
	perturb_cursor();
	g = go_elarc_new_3ecc( h, hc, hc, 0, 0.5, GEOM_CONIC_ARC_POINT_ENTIRE );
	MLT;
}
/* Ellipse/elliptic arc by center and radiii */
void create_ellarc_center_radii_1( struct of_state * state )
{
	struct go_elarc * g = go_elarc_new_xyr( state->value.data.hanger.hanger, 0.0, 20.0, 10.0, 0.0,
		(state == &ofs_ellarc_center_radii_1) ? 2.0 : R_2PI, GEOM_CONIC_ARC_START_DIF );
	ML;
}
/* Ellipse by center, point, rotation and eccentricity */
void create_ellipse_xy1ecc_1( struct of_state * state UNUSED )
{
	MS(go_elarc);
	perturb_cursor();
	g = go_elarc_new_xy1ecc( h, hc, 0.0, 0.5, GEOM_CONIC_ARC_POINT_ENTIRE );
	MLT;
}
/* Ellipse by foci and point */
void create_ellipse_segment( struct of_state * state )
{
	MS(go_segment);
	g = go_segment_new( h, hc );
	state->tmp_go[OFR_DISPLAY] = &g->go;
	TSORT_OFF;
	go_link( go_ref( &g->go ), factory.group, NULL );
	TSORT_ON;
}

void create_ellipse_foci_pt_2( struct of_state * state )
{
	TSORT_OFF;
	go_unlink( state->prev->tmp_go[OFR_DISPLAY] );//unlink the old go
	struct go_elarc * g = go_elarc_new_foci( state->prev->value.data.hanger.hanger,
		state->value.data.hanger.hanger, factory.hanger_cursor, GEOM_CONIC_ARC_START_ENTIRE );
	ML;
	TSORT_ON;
}
void fix_ellipse_anchor_2( struct of_state * state )
{
	struct go_elarc * c = (struct go_elarc*)state->prev->tmp_go[OFR_DISPLAY];
	anchor_rehang( go_get_anchor(&c->go, 2), state->value.data.hanger.hanger );
}

/* Text and TeX text */
void create_text_1( struct of_state * state )
{
	struct go_text * g = go_text_new( state->value.data.hanger.hanger, 0.5, 0.5, 0.0, 0.0,
		string_lookup(" "), fonts_get_default(), 10); //FIXME prazdny text, ne mezera
	ML;
	text_edit_create( &g->go );
}
void create_tex_text_1( struct of_state * state )
{
	struct go_tex_text * g = go_tex_text_new( state->value.data.hanger.hanger,
		0.5, 0.5, 0.0, 0.0, string_lookup("") );
	ML;
	text_edit_create( &g->go );
}

/****************  Random things **************/

static inline real random_coord(void)
{ return ((random () * 8.0) / RAND_MAX) - 4; }

static inline struct hanger* random_hanger(void)
{ return hanger_new_mc_xy( random_coord(), random_coord() ); }

static void create_random_point( struct go_group * group, char * err )
{
	struct go_point *pt;

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Create a random point", err )
		pt = go_point_new( random_hanger() );
		go_link( &pt->go, group, NULL );
	TRANS_FAILED
		report_error("Cannot add a random point\n");
	TRANS_END
}

static void create_random_bezier_quad( struct go_group * group, char * err )
{
	struct go_bezier * gs;

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Create a random quadratic bezier", err )
		gs = go_bezier_new_quadratic( random_hanger(), 1, random_hanger(), 1, random_hanger(), 1 );
		go_link( &gs->go, group, NULL );
	TRANS_FAILED
		report_error("Cannot add a random bezier\n");
	TRANS_END
}

static void create_random_bezier_cubic( struct go_group * group, char * err )
{
	struct go_bezier * gs;

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Create a random cubic bezier", err )
		gs = go_bezier_new_cubic( random_hanger(), 1, random_hanger(), 1, random_hanger(), 1, random_hanger(), 1 );
		go_link( &gs->go, group, NULL );
	TRANS_FAILED
		report_error("Cannot add a random bezier\n");
	TRANS_END
}

static void create_random_segment( struct go_group * group, char * err )
{
	struct go_segment * gs;

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Create a random segment", err )
		gs = go_segment_new( random_hanger(), random_hanger() );
		go_link( &gs->go, group, NULL );
	TRANS_FAILED
		report_error("Cannot add a random segment\n");
	TRANS_END
}

#define RANDOM_THING(f,n) \
void f( struct cmd * cmd UNUSED, void * subj ) \
{ \
	struct go * grp; \
	char err[TRANS_ERR_SIZE]; \
	if ( subj ) \
	{ \
		grp = void2go(subj); \
		ASSERT_GROUP(grp); \
		n((struct go_group*)grp, err); \
	} else WALK_GROUP_SELECTED_GO(tlo_get_tlg(context.tlo), grp) n((struct go_group*)grp, err); \
}

RANDOM_THING(on_random_point, create_random_point)
RANDOM_THING(on_random_bezier_quad, create_random_bezier_quad)
RANDOM_THING(on_random_bezier_cubic, create_random_bezier_cubic)
RANDOM_THING(on_random_segment, create_random_segment)
