#undef LOCAL_DEBUG
#include "lib/lib.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <gtk/gtk.h>
#ifdef USE_LIBPAPER
#include <paper.h>
#endif

#include "kernel/kernel.h"
#include "gui/main.h"
#include "export/export.h"
#include "import/import.h"
#include "gui/dialogs.h"
#include "gui/properties.h"
#include "gui/cmdmgr.h"

#define POINT2MM	(25.4/72)

/*** IMPORT DIALOGS ****/
#define SVG_EXT "*.svg"


/*** EXPORT DIALOGS ****/
#define FALLBACK_A4_WIDTH	210
#define FALLBACK_A4_HEIGHT	297

char last_open_folder[PATH_MAX];
char last_save_folder[PATH_MAX];

#ifdef USE_LIBPAPER
static int last_paper;
#endif

static GtkWidget * create_paper_combobox( GtkSpinButton * spinw, GtkSpinButton * spinh );
static void fill_paper_sizes( GtkComboBox * combobox, GtkSpinButton * spinw );

static GtkWidget * create_paper_combobox( GtkSpinButton * spinw, GtkSpinButton * spinh )
{
	GtkComboBox * cb = GTK_COMBO_BOX( gtk_combo_box_new_text() );
#ifdef USE_LIBPAPER
	const struct paper *p;
	for ( p = paperfirst(); p; p = papernext(p) ) gtk_combo_box_append_text( cb, papername(p) );
	gtk_combo_box_set_active( cb, last_paper );
#else
	gtk_combo_box_append_text( cb, "a4" );
	gtk_combo_box_set_active( cb, 0 );
#endif
	g_object_set_data( G_OBJECT(spinw), "other", spinh );
	g_signal_connect( G_OBJECT(cb), "changed", G_CALLBACK(fill_paper_sizes), spinw );
	fill_paper_sizes( cb, spinw );

	return GTK_WIDGET(cb);
}
#ifdef USE_LIBPAPER
static void fill_paper_sizes( GtkComboBox * combobox, GtkSpinButton * spinw )
{
	GtkSpinButton * spinh = GTK_SPIN_BUTTON( g_object_get_data( G_OBJECT(spinw), "other" ));
	const struct paper * p = paperinfo(gtk_combo_box_get_active_text(combobox));
	gtk_spin_button_set_value( spinw, paperpswidth(p) * POINT2MM );
	gtk_spin_button_set_value( spinh, paperpsheight(p) * POINT2MM );
}
#else
static void fill_paper_sizes( GtkComboBox * combobox UNUSED, GtkSpinButton * spinw )
{
	GtkSpinButton * spinh = GTK_SPIN_BUTTON( g_object_get_data( G_OBJECT(spinw), "other" ));
	gtk_spin_button_set_value( spinw, FALLBACK_A4_WIDTH );
	gtk_spin_button_set_value( spinh, FALLBACK_A4_HEIGHT );
}
#endif

void save_folder( const char * filename, char * dest )
{
	int i = strlen(filename) - 1;
	for ( ; i && filename[i] != '/'; i-- ); //find the last '/'
	for ( dest[i+1] = '\0'; i >= 0; dest[i] = filename[i--] );
	DBG( "extracted folder path:%s\n", dest );
}

void on_svg_file_import( struct cmd *cmd UNUSED, void *subj )
{
	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_TLO( subj );

	OPEN_DLG_START( "Import SVG file ...", NULL, NULL )
	{
		if ( !import_svg (filenames[i], subj) )
			statusbar_msg( &wnd_univ_browser.status, "File %s imported.", filenames[i] );
	}
	OPEN_DLG_END;
}

void on_import_ipe5( struct cmd * cmd UNUSED, void * subj )
{
	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_TLO( subj );

	OPEN_DLG_START( "Import Ipe 5 file ...", NULL, NULL )
	{
		if ( !import_tlo_ipe5(filenames[i], subj) )
			statusbar_msg( &wnd_univ_browser.status, "File %s imported.", filenames[i] );
	}
	OPEN_DLG_END;
}

void on_export_pdf( struct cmd * cmd UNUSED, void * volatile subj )
{
	char buf[TRANS_ERR_SIZE];
	FILE * file;

	GtkWidget * table = gtk_table_new( 2, 2, FALSE );

	GtkWidget * include_fonts = gtk_check_button_new_with_mnemonic( "Include font data" );
	gtk_table_attach( GTK_TABLE(table), include_fonts, 0, 2, 0, 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(include_fonts),
		PROP_STORE_GET_UNS(ps_global, "pdf-export-include-fonts", TRUE ));

	gtk_table_attach (GTK_TABLE (table), gtk_label_new("Compression level:"),
		0, 1, 1, 2, GTK_EXPAND | GTK_FILL, 0, 0, 0);
	GtkComboBox * c_box = GTK_COMBO_BOX(gtk_combo_box_new_text());
	gtk_combo_box_append_text( c_box, "No compression" );
	gtk_combo_box_append_text( c_box, "Fast compression" );
	gtk_combo_box_append_text( c_box, "Best compression" );
	gtk_combo_box_set_active( c_box, PROP_STORE_GET_UNS(ps_global, "pdf-export-compression-level", 0) );
	gtk_table_attach (GTK_TABLE (table), GTK_WIDGET(c_box), 1, 2, 1, 2, GTK_EXPAND | GTK_FILL, 0, 0, 0);

	gtk_widget_show_all( table );

	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_DOC( subj );
	SUGGEST_FILENAME( subj, "pdf", nn );

	SAVE_DLG_START( "Export PDF file ...", table, nn )
	{
		if ( !(file = fopen( filename, "w" )) )
			report_error_f("open file failed: %s\n", strerror(errno) );
		else
		{
			uns incl_fnt = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(include_fonts));
			uns cl = gtk_combo_box_get_active(c_box);
			TRANS_BEGIN_ANONYMOUS(buf)
				export_pdf( (struct obj_doc*)subj, file, !incl_fnt, cl);
			TRANS_FAILED
				report_error_f("export failed: %s\n", buf );
			TRANS_END

			fclose( file );
			statusbar_msg(&wnd_univ_browser.status, "File written successfully.");

			PROP_STORE_SET_UNS(ps_global, "pdf-export-include-fonts", incl_fnt );
			PROP_STORE_SET_UNS(ps_global, "pdf-export-compression-level", cl );
		}
	}
	SAVE_DLG_END;
}

void on_export_ps( struct cmd * cmd UNUSED, void * volatile subj )
{
	FILE * file;
	GtkWidget *table;
	GtkWidget *spin_height, *spin_width, *combo_format;
	GtkWidget *label_height, *label_width, *label_format;
	GtkWidget *check_fit;
	GtkWidget *include_fonts;

	table = gtk_table_new (5, 2, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER (table), 10);

	label_format = gtk_label_new ("Set sizes to standard page format:");
	gtk_table_attach (GTK_TABLE (table), label_format, 0, 1, 0, 1, GTK_FILL, 0, 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label_format), 0, 0.5);

	label_width = gtk_label_new ("Page width (in mm):");
	gtk_table_attach (GTK_TABLE (table), label_width, 0, 1, 1, 2, GTK_FILL, 0, 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label_width), 0, 0.5);

	label_height = gtk_label_new ("Page height (in mm):");
	gtk_table_attach (GTK_TABLE (table), label_height, 0, 1, 2, 3, GTK_FILL, 0, 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label_height), 0, 0.5);

	spin_width = gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (210, 0, 1000, 1, 10, 10)), 1, 0);
	gtk_table_attach (GTK_TABLE (table), spin_width, 1, 2, 1, 2, GTK_FILL, 0, 0, 0);

	spin_height = gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (297, 0, 1000, 1, 10, 10)), 1, 0);
	gtk_table_attach (GTK_TABLE (table), spin_height, 1, 2, 2, 3, GTK_FILL, 0, 0, 0);

	combo_format = create_paper_combobox( GTK_SPIN_BUTTON(spin_height), GTK_SPIN_BUTTON(spin_width) );
	gtk_table_attach (GTK_TABLE (table), combo_format, 1, 2, 0, 1, GTK_FILL, 0, 0, 0);

	check_fit = gtk_check_button_new_with_mnemonic ("Fit to page");
	gtk_table_attach (GTK_TABLE (table), check_fit, 0, 2, 3, 4, GTK_EXPAND | GTK_FILL, 0, 0, 0);

	include_fonts = gtk_check_button_new_with_mnemonic ("Include font data");
	gtk_table_attach (GTK_TABLE (table), include_fonts, 0, 2, 4, 5, GTK_EXPAND | GTK_FILL, 0, 0, 0);
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(include_fonts), TRUE );

	gtk_widget_show_all(table);

	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_DOC( subj );
	SUGGEST_FILENAME( subj, "ps", nn );

	SAVE_DLG_START( "Export PS file ...", table, nn )
	{
		if ( !(file = fopen( filename, "w" )) )
			report_error_f("open file failed: %s\n", strerror(errno) );
		else {
			int width, height, fit, no_fonts;

			width = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (spin_width));
			height = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (spin_height));
			fit = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check_fit));
			no_fonts = !gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (include_fonts));

			char buf[TRANS_ERR_SIZE];
			TRANS_BEGIN_ANONYMOUS(buf)
				export_ps( subj, file, width, height, fit, no_fonts );
			TRANS_FAILED
				report_error_f("export failed: %s\n", buf );
			TRANS_END

			fclose( file );
			statusbar_msg(&wnd_univ_browser.status, "File written successfully.");
		}
	}
	SAVE_DLG_END;
}

void on_export_eps( struct cmd * cmd UNUSED, void * volatile subj )
{
	FILE * file;
	GtkWidget *include_fonts;

	include_fonts = gtk_check_button_new_with_mnemonic ("Include font data");
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(include_fonts), TRUE );
	gtk_widget_show (include_fonts);

	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_TLO( subj );
	SUGGEST_FILENAME( subj, "eps", nn );

	SAVE_DLG_START( "Export EPS file ...", include_fonts, nn )
	{
		if ( !(file = fopen( filename, "w" )) )
			report_error_f("open file failed: %s\n", strerror(errno) );
		else
		{
			int no_fonts UNUSED = !gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (include_fonts));
			char buf[TRANS_ERR_SIZE];
			TRANS_BEGIN_ANONYMOUS(buf)
				export_eps( subj, file, no_fonts );
			TRANS_FAILED
				report_error_f("export failed: %s\n", buf );
			TRANS_END

			fclose( file );
			statusbar_msg(&wnd_univ_browser.status, "File written successfully.");
		}
	}
	SAVE_DLG_END;
}

void on_export_svg( struct cmd * cmd UNUSED, void * volatile subj )
{
	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_TLO( subj );
	SUGGEST_FILENAME(subj, "svg", nn );

	SAVE_DLG_START("Export SVG file ...", NULL, nn)
	{
		int result = 1;
		char buf[TRANS_ERR_SIZE];
		TRANS_BEGIN_ANONYMOUS(buf)
			export_svg( subj, filename, &result );
		TRANS_FAILED
			report_error_f("export failed: %s\n", buf );
		TRANS_END
		if (result)
			statusbar_msg(&wnd_univ_browser.status, "File written successfully.");
	}
	SAVE_DLG_END;
}

/*** OPTIONS WINDOW ****/

#define GLOBAL_SETTINGS_CNT 14
struct wnd_settings wnd_settings;
static void destroy_settings_window( GtkWidget * widget, gpointer data );
static void settings_add_prop( string key, GtkTable * table, int * counter, int * row, char * desc );
static void sync_settings_window(struct o *o, struct obj_hook * h );
static void create_settings_window(void);

void on_show_settings_window( struct cmd * cmd, void * subj UNUSED )
{
	if ( wnd_settings.window )
	{
		gtk_widget_destroy( wnd_settings.window );
		wnd_settings.window = 0;
		cmd->state &= ~VCS_CHECKED;
	}
	else
	{
		create_settings_window();
		cmd->state |= VCS_CHECKED;
	}
}

static void destroy_settings_window( GtkWidget * widget UNUSED, gpointer data UNUSED )
{
	wnd_settings.window = NULL;
	xfree(wnd_settings.pi);
	wnd_settings.pi = NULL;
	obj_hook_remove(&wnd_settings.oh);
	refresh_gui();
}

static void settings_add_prop( string key, GtkTable * table, int * counter, int * row, char * desc )
{
	ASSERT(*counter < GLOBAL_SETTINGS_CNT);
	struct prop_item * pi = wnd_settings.pi + *counter;
	prop_item_init( pi, obj2o(obj_universum), key );

	GtkWidget * label = gtk_label_new(desc);
	g_object_set( G_OBJECT(label), "xalign", 0.0, NULL );
	gtk_table_attach( table, label, 0, 1, *row, *row+1, GTK_FILL, GTK_FILL, 5, 0 );
	if (pi->unit_edit)
	{
		gtk_table_attach( table, pi->value_edit, 1, 2, *row, *row+1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0 );
		gtk_table_attach( table, pi->unit_edit, 2, 3, *row, *row+1, GTK_FILL, GTK_FILL, 0, 0 );
	} else
		gtk_table_attach( table, pi->value_edit, 1, 3, *row, *row+1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0 );
	(*counter)++;
	(*row)++;
}

static void sync_settings_window(struct o *o, struct obj_hook * h UNUSED)
{
	if ( o != obj2o(obj_universum) ) return;
	int counter = 0;
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_tolerance ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_grid_x1 ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_grid_y1 ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_grid_x2 ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_grid_y2 ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_rulerx_res ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_rulery_res ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_drag_image ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_editor ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_editor_in_term ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_browser ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_browser_in_term ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_terminal ) );
	prop_sync( wnd_settings.pi + counter++, prop_find( o, string_maxhist ) );
}

static void create_settings_window(void)
{
	wnd_settings.pi = xmalloc_zero(GLOBAL_SETTINGS_CNT * sizeof(struct prop_item));
	int counter = 0;
	int row = 0;

	wnd_settings.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title( GTK_WINDOW(wnd_settings.window), "Global settings" );
	g_signal_connect( G_OBJECT(wnd_settings.window), "delete_event", G_CALLBACK(gtk_widget_destroy), NULL );
	g_signal_connect( G_OBJECT(wnd_settings.window), "destroy", G_CALLBACK(destroy_settings_window), NULL);
	GtkWidget * vbox = gtk_vbox_new( FALSE, 0 );
	gtk_container_add( GTK_CONTAINER(wnd_settings.window), vbox );
	
	GtkWidget * frame = gtk_frame_new("Editor settings");
	gtk_container_set_border_width( GTK_CONTAINER(frame), 5 );
	gtk_box_pack_start( GTK_BOX(vbox), frame, FALSE, TRUE, 0 );
	GtkWidget * table = gtk_table_new( 1, 3, FALSE );
	gtk_container_set_border_width( GTK_CONTAINER(table), 5 );
	gtk_container_add( GTK_CONTAINER(frame), table );

	settings_add_prop( string_tolerance, GTK_TABLE(table), &counter, &row, "Snap tolerance (in pixels)" );
	settings_add_prop( string_grid_x1, GTK_TABLE(table), &counter, &row, "Grid x1" );
	settings_add_prop( string_grid_y1, GTK_TABLE(table), &counter, &row, "Grid y1" );
	settings_add_prop( string_grid_x2, GTK_TABLE(table), &counter, &row, "Grid x2" );
	settings_add_prop( string_grid_y2, GTK_TABLE(table), &counter, &row, "Grid y2" );
	settings_add_prop( string_rulerx_res, GTK_TABLE(table), &counter, &row, "Ruler x-resolution" );
	settings_add_prop( string_rulery_res, GTK_TABLE(table), &counter, &row, "Ruler y-resolution" );
	settings_add_prop( string_drag_image, GTK_TABLE(table), &counter, &row, "Panning: drag the image" );

	frame = gtk_frame_new("External programs");
	gtk_container_set_border_width( GTK_CONTAINER(frame), 5 );
	gtk_box_pack_start( GTK_BOX(vbox), frame, FALSE, TRUE, 0 );
	table = gtk_table_new( 1, 3, FALSE );
	gtk_container_set_border_width( GTK_CONTAINER(table), 5 );
	gtk_container_add( GTK_CONTAINER(frame), table );
	row = 0;
	
	settings_add_prop( string_editor, GTK_TABLE(table), &counter, &row, "Your favourite text editor" );
	settings_add_prop( string_editor_in_term, GTK_TABLE(table), &counter, &row, "Run the text editor in a terminal?" );
	settings_add_prop( string_browser, GTK_TABLE(table), &counter, &row, "Your favourite WWW browser" );
	settings_add_prop( string_browser_in_term, GTK_TABLE(table), &counter, &row, "Run the browser in a terminal?" );
	settings_add_prop( string_terminal, GTK_TABLE(table), &counter, &row, "Your favourite terminal" );
	
	frame = gtk_frame_new("Not implemented");
	gtk_container_set_border_width( GTK_CONTAINER(frame), 5 );
	gtk_box_pack_start( GTK_BOX(vbox), frame, FALSE, TRUE, 0 );
	table = gtk_table_new( 1, 3, FALSE );
	gtk_container_set_border_width( GTK_CONTAINER(table), 5 );
	gtk_container_add( GTK_CONTAINER(frame), table );
	row = 0;
	
	settings_add_prop( string_maxhist, GTK_TABLE(table), &counter, &row, "Maximum volume of undo history (in KB)" );

	wnd_settings.oh.data = &wnd_settings;
	wnd_settings.oh.obj_prop_change = sync_settings_window;
	obj_hook_add( &wnd_settings.oh );

	gtk_widget_show_all( GTK_WIDGET(wnd_settings.window) );
#ifdef VRR_POTEMKIN
	gtk_widget_hide(frame);
#endif
}

/*** ERROR WINDOW ***/

void create_err_window(GtkWidget *widget, const char *title UNUSED, const char *label, const char *button_label, ...)
{
	char msg[MAX_ERR_MSG_LENGTH];
	va_list ap;
	va_start(ap, button_label);
	vsnprintf(msg, MAX_ERR_MSG_LENGTH, label, ap);
	va_end(ap);

	GtkWidget * dlg = gtk_message_dialog_new( GTK_WINDOW(widget), GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, msg );
	gtk_dialog_run( GTK_DIALOG(dlg) );
	gtk_widget_destroy(dlg);
}
