#ifndef __VRR_DIALOGS_H__
#define __VRR_DIALOGS_H__

/** \file gui/dialogs.h
 Macros for open and save dialogs */

#include <sys/param.h>

extern char last_open_folder[PATH_MAX];
extern char last_save_folder[PATH_MAX];

void save_folder( const char * filename, char * dest );

#define OPEN_DLG_START(_title, _extra, _fname) \
	GtkFileSelection * dlg = GTK_FILE_SELECTION( gtk_file_selection_new(_title) );			\
													\
	if (_fname) gtk_file_selection_set_filename( dlg, _fname );					\
	else { if (last_open_folder[0]) gtk_file_selection_set_filename( dlg, last_open_folder ); }	\
	gtk_file_selection_set_select_multiple( dlg, TRUE );						\
													\
	if (_extra) gtk_box_pack_start( GTK_BOX(dlg->action_area), _extra, GTK_FILL, GTK_FILL, 0 );	\
													\
	if ( gtk_dialog_run( GTK_DIALOG(dlg) ) == GTK_RESPONSE_OK )					\
	{												\
		char ** filenames = gtk_file_selection_get_selections(dlg);				\
		save_folder( filenames[0], last_open_folder );						\
		for ( volatile int i = 0; filenames[i]; i++ )

#define OPEN_DLG_END \
		g_strfreev(filenames);			\
	}						\
	gtk_widget_destroy( GTK_WIDGET(dlg) )

#define SAVE_DLG_START(_title, _extra, _fname) \
	GtkFileSelection * dlg = GTK_FILE_SELECTION( gtk_file_selection_new(_title) );			\
													\
	if (_fname) gtk_file_selection_set_filename( dlg, _fname );					\
	else { if (last_save_folder[0]) gtk_file_selection_set_filename( dlg, last_save_folder ); }	\
	gtk_file_selection_set_select_multiple( dlg, FALSE );						\
													\
	if (_extra) gtk_box_pack_start( GTK_BOX(dlg->action_area), _extra, GTK_FILL, GTK_FILL, 0 );	\
													\
	if ( gtk_dialog_run( GTK_DIALOG(dlg) ) == GTK_RESPONSE_OK )					\
	{												\
		const char * filename = gtk_file_selection_get_filename(dlg);				\
		save_folder( filename, last_save_folder );

#define SAVE_DLG_END \
	}					\
	gtk_widget_destroy( GTK_WIDGET(dlg) )	\

#endif

#define SUGGEST_FILENAME(_o, _ext, _sugname) 							\
	const char * on = o_lookup_name((struct o*)_o);						\
	char _sugname[strlen(on) + strlen(last_save_folder) + strlen(_ext) + 2];		\
	sprintf( _sugname, "%s%s.%s", last_save_folder[0] ? last_save_folder : "", on, _ext )
