/** \file gui/glt_gui.c
  \brief GUILE-link - interface functions for kernel data types. */


#include "lib/lib.h"
#include <stdio.h>
#include <string.h>
#include <guile/gh.h>
#include "kernel/kernel.h"
#include "gui/main.h"
#include "gui/guileaccess.h"
#include "build/path.h"

#include "kernel/gl_trans.h"
#include "kernel/glt_common.h"
#include "kernel/glt_kernel.h"
#include "gui/glt_gui.h"

scm_t_bits window_tag;

uns is_window_proxy (SCM proxy)
{ return SCM_SMOB_PREDICATE(window_tag, proxy); }


static size_t window_proxy_free (SCM proxy)
{
    if (vptr (proxy) != NULL)
    {
	vptr (proxy) -> proxy = SCM_BOOL_F;
	SCM_SET_SMOB_DATA (proxy, NULL); // unneccessary
    }
    return (0);
}

SCM window2scm(struct window *p)
{
  if (p == NULL) return (SCM_BOOL_F);

  if (SCM_EQ_P (p -> proxy, SCM_BOOL_F))
    SCM_NEWSMOB (p -> proxy, window_tag, p);

  return (p -> proxy);
}

void window_proxy_clean (struct window *p)
{
    if (! SCM_EQ_P (p -> proxy, SCM_BOOL_F))
    {
	SCM_SET_SMOB_DATA (p -> proxy, NULL);
	p -> proxy = SCM_BOOL_F; // unneccessary
    }
}

static SCM
gl_window_p (SCM obj)
{ return (SCM_BOOL (is_window_proxy (obj))); }


#include "obj/gui/generated_bindings.h"

void
glt_gui_init(void)
{
  window_tag = scm_make_smob_type ("vrr-window", 0);
  scm_set_smob_free (window_tag, window_proxy_free);

  scm_c_define_gsubr ("window?", 1, 0, 0, gl_window_p);

  gui_generated_bindings_init ();
}



