#undef LOCAL_DEBUG
#include "lib/lib.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <locale.h>
#ifdef USE_LIBPAPER
#include <paper.h>
#endif

#include <gtk/gtk.h>

#include "kernel/kernel.h"

#include "gui/main.h"
#include "gui/cmdmgr.h"
#include "gui/visualisation.h"
#include "gui/glt_gui.h"
#include "gui/cmdmgr.h"
#include "gui/properties.h"
#include "gui/changeunits.h"
#include "gui/dialogs.h"
#include "vcl/vcl.h"

#include "build/path.h"

#define VRR_EXT		"*.vrr"

struct context context;

GtkWidget * main_window;
struct console *console_window;

int open_windows_cnt;
int main_window_open;
GList *global_units_glist;
GList *global_multiplier_glist;
GSList * wuh_list;

static int closing_all;

GdkPixbuf * icon_arrow_down, * icon_arrow_up, * icon_arrow_left, * icon_arrow_right;
GdkPixbuf * icon_pencil_on, * icon_modified;
GdkPixbuf * icon_rotation_left, * icon_rotation_right, * icon_zoom_in, * icon_zoom_out;

static int noname_counter;
static char noname_buf[20];

//Hook calbacks for handling selection changes
static void mh_init(void);
static void mh_finish(void);
static void mh_go_selected( struct o * o, struct go_hook * hook );
static void mh_go_unselected( struct o * o, struct go_hook * hook );
static void mh_obj_sel_changed( struct o * o, struct obj_hook * hook );
static void mh_obj_added( struct o * o, struct obj_hook * hook );
static void mh_obj_after_delete( struct o * o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook * hook UNUSED );
static void mh_go_after_delete( struct o * o, struct o * owner UNUSED, struct o * prev UNUSED, struct go_hook * gh UNUSED );
static void mh_add_go_hook( struct obj * obj );

static void resources_init(void);
static void resources_finish(void);

static void gui_save_document( struct obj * doc, const char * filename );

/******* CALLBACKS *********/

void on_go_delete( struct cmd * cmd UNUSED, void * subj )
{
	ASSERT_TLO(context.tlo);
	ASSERT_GROUP(context.group);
	char buf[TRANS_ERR_SIZE];
	struct go * go = (struct go*)subj;
	if (go) ASSERT_O(go);

	TRANS_BEGIN_MAIN(context.tlo, "Delete GO", buf)
		if (go) go_select(go);
		group_remove_rehang_selected_go(context.group);
	TRANS_FAILED
		report_error_f("Could not delete GO: %s\n", buf);
	TRANS_END
}

void on_go_enclose_sel( struct cmd * cmd UNUSED, void * subj UNUSED )
{
	ASSERT_TLO(context.tlo);
	ASSERT_GROUP(context.group);
	char s[TRANS_ERR_SIZE];

	TRANS_BEGIN_MAIN( context.tlo, "Add dependent to selection", s)
		//tlo_add_selection_down();  FIXME OLD
	TRANS_FAILED
		report_error_f("Could not add dependent to selection: %s.\n", s);
	TRANS_END
}

void on_obj_undo( struct cmd * cmd UNUSED, void * subj UNUSED )
{
	char s[TRANS_ERR_SIZE];
	if (!trans_undo(tlo_universum, s)) report_error_f("%s\n", s);
}

void on_obj_redo( struct cmd * cmd UNUSED, void * subj UNUSED )
{
	char s[TRANS_ERR_SIZE];
	if (!trans_redo(tlo_universum, s)) report_error_f("%s\n", s);
}

void on_tlo_undo( struct cmd * cmd UNUSED, void * subj )
{
	char s[TRANS_ERR_SIZE];
	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_TLO( (struct obj_tlo*)subj );
	if (!trans_undo( (struct obj_tlo*)subj, s )) report_error_f("%s\n", s);
}

void on_tlo_redo( struct cmd * cmd UNUSED, void * subj )
{
	char s[TRANS_ERR_SIZE];
	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_TLO( (struct obj_tlo*)subj );
	if (!trans_redo( (struct obj_tlo*)subj, s )) report_error_f("%s\n", s);
}

void on_copy( struct cmd * cmd UNUSED, void * subj )
{
	char s[TRANS_ERR_SIZE], ss[TRANS_ERR_SIZE];
	ASSERT_TLO(context.tlo);
	if ( context.tlo == tlo_clipboard ) return;

	TRANS_BEGIN_MAIN( context.tlo, "Copy", ss)
		if (subj) go_select(void2go(subj));
		if (!clipboard_copy( context.group, "Copy", s )) report_error_f("%s\n", s);
	TRANS_FAILED
		report_error_f("%s\n", ss);
	TRANS_END
}

void on_cut( struct cmd * cmd UNUSED, void * subj )
{
	char s[TRANS_ERR_SIZE];
	ASSERT_TLO(context.tlo);
	if ( context.tlo == tlo_clipboard ) return;

	TRANS_BEGIN_MAIN( context.tlo, "Cut", s)
		if (subj) go_select(void2go(subj));
		clipboard_cut( context.group );
	TRANS_FAILED
		report_error_f("%s\n", s);
	TRANS_END
}

void on_paste( struct cmd * cmd UNUSED, void * subj UNUSED )
{
	char s[TRANS_ERR_SIZE];
	ASSERT_TLO(context.tlo);
	if ( context.tlo != tlo_clipboard )
		TRANS_BEGIN_MAIN( context.tlo, "Cut", s)
			clipboard_paste( context.group, NULL );
		TRANS_FAILED
			report_error_f("%s\n", s);
		TRANS_END
}

void on_group_select_all( struct cmd * cmd UNUSED, void * subj )
{
	char s[TRANS_ERR_SIZE];
	struct go_group * volatile group;
	if ( subj ) group = (struct go_group*)subj;
	else group = (struct go_group*) group_get_first_selected_go( tlo_get_tlg(context.tlo) );
	ASSERT_GROUP(group);

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Select all", s );
		group_select_all_go(group);
	TRANS_FAILED
		die("Something bad happened during selection\n");
	TRANS_END
}

void on_group_select_none( struct cmd * cmd UNUSED, void * subj )
{
	char s[TRANS_ERR_SIZE];
	struct go_group * volatile group;
	if ( subj ) group = (struct go_group*)subj;
	else group = (struct go_group*) group_get_first_selected_go( tlo_get_tlg(context.tlo) );
	ASSERT_GROUP(group);

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Unselect all", s );
		group_unselect_all_go(group);
	TRANS_FAILED
		die("Something bad happened during selection\n");
	TRANS_END
}

void on_group_selected( struct cmd * cmd UNUSED, void * subj UNUSED )
{
	ASSERT(context.group);
	char s[TRANS_ERR_SIZE];
	struct go_group * group = context.group, * new_group;
	struct go * go;

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Group selected objects", s )
		new_group = go_group_new();
		go = group_get_first_selected_go(group);
		go = go ? group_get_prev_go(group, go) : NULL;
		go_link( &new_group->go, group, go );
		group_relink_selected_go( group, new_group, NULL );
		go_select(&new_group->go);
	TRANS_FAILED
		report_error_f("%s\n", s);
	TRANS_END
}
void on_group_ungroup( struct cmd * cmd UNUSED, void * subj )
{
	char s[TRANS_ERR_SIZE];
	struct go_group * volatile group, * up_group;
	if ( subj ) group = (struct go_group*)subj;
	else group = (struct go_group*) group_get_first_selected_go( tlo_get_tlg(context.tlo) );
	ASSERT_GROUP(group);

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Ungroup the group", s )
		if ( !(up_group = go_get_group(&group->go)) )
			trans_fail("Cannot ungroup the top group!\n");
		group_relink_all_go( group, up_group, &group->go );
		go_unlink(&group->go);
	TRANS_FAILED
		report_error_f("%s\n", s);
	TRANS_END
}
#define GET_ARG_GO \
	struct go * volatile go; \
	ASSERT(context.group); \
	if (subj) go = void2go(subj); else go = group_get_first_selected_go(context.group); \
	ASSERT_GO(go)

void on_move_up( struct cmd * cmd UNUSED, void * subj )
{
	GET_ARG_GO;
	struct go * volatile prev = group_get_prev_go(context.group, go);
	ASSERT(go);
	prev = group_get_prev_go(context.group, prev);
	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( go_get_tlo(go), "Move up", s )
		go_relink( go, context.group, prev );
	TRANS_FAILED
		report_error_f("%s\n", s);
	TRANS_END
}
void on_move_down( struct cmd * cmd UNUSED, void * subj )
{
	GET_ARG_GO;
	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( go_get_tlo(go), "Move down", s )
		go_relink( go, context.group, group_get_next_go(context.group, go) );
	TRANS_FAILED
		report_error_f("%s\n", s);
	TRANS_END
}
void on_move_front( struct cmd * cmd UNUSED, void * subj )
{
	GET_ARG_GO;
	if ( go == group_get_first_go(context.group) ) return;
	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( go_get_tlo(go), "Move front", s )
		go_relink( go, context.group, NULL );
	TRANS_FAILED
		report_error_f("%s\n", s);
	TRANS_END
}
void on_move_back( struct cmd * cmd UNUSED, void * subj )
{
	GET_ARG_GO;
	if ( go == group_get_last_go(context.group) ) return;
	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( go_get_tlo(go), "Move back", s )
		go_relink( go, context.group, group_get_last_go(context.group) );
	TRANS_FAILED
		report_error_f("%s\n", s);
	TRANS_END
}
void on_change_obj_prop( struct cmd *cmd UNUSED, void * subj )
{
	struct obj * obj;
	if ( subj ) obj = void2obj(subj);
	else obj = obj_get_first_selected();
	ASSERT_O( obj );

	prop_window_create(obj2o(obj));
}

void on_change_go_prop( struct cmd *cmd UNUSED, void * subj )
{
	struct go * go;
	if ( subj ) go = void2go(subj);
	else
	{
		ASSERT_GROUP(context.group);
		go = group_get_first_selected_go(context.group);
	}
	ASSERT_GO( go );

	prop_window_create(go2o(go));
}

void on_edit_text( struct cmd * cmd UNUSED, void * subj )
{
	ASSERT_GROUP(context.group);
	struct go * go = subj ? (struct go*)subj : group_get_first_selected_go( context.group );
	ASSERT_O(go);
	text_edit_create( go );
}

//JUMP
void on_snap_hangers( struct cmd * cmd UNUSED, void * subj UNUSED ){ factory.snap_set ^= SNAP_HANGERS; }
void on_snap_grid( struct cmd * cmd UNUSED, void * subj UNUSED )
{
	factory.snap_set ^= SNAP_GRID;
	for ( GSList * l = view_list; l; l = l->next ) view_update_grid( (struct view*)l->data );
}
void on_snap_lines( struct cmd * cmd UNUSED, void * subj UNUSED ){ factory.snap_set ^= SNAP_LINES; }
void on_snap_inters( struct cmd * cmd UNUSED, void * subj UNUSED ){ factory.snap_set ^= SNAP_INTERS; }
void on_snap_depend( struct cmd * cmd UNUSED, void * subj UNUSED ){ factory.snap_set ^= SNAP_DEPENDENCY; }

void on_ruler_resolution( struct cmd * cmd UNUSED, void * subj UNUSED)
{
//	DBG("CHANGING RULER RESOLUTION\n");
}

void on_dump( struct cmd * cmd UNUSED, void * subj UNUSED )
{
	obj_dump_subtree(obj_universum);
	if ( context.group )
		group_dump_subtree(context.group);
	if ( context.tlo )
		tlo_dump_tsort(context.tlo);
}

void on_file_new( struct cmd * cmd UNUSED, void * subj UNUSED )
{
	char buf[TRANS_ERR_SIZE];

	TRANS_BEGIN_MAIN(tlo_universum, "Create a new file", buf)
		struct obj_doc * doc = doc_new_named(get_new_noname());
		struct obj_tlo * tlo = tlo_new_named(get_new_page_noname(doc));
		obj_link( &tlo->obj, &doc->obj );
		obj_link( &doc->obj, obj_universum );
		doc_set_all_saved(doc);
	TRANS_FAILED
		report_error_f("Cannot create a new file: %s\n", buf);
	TRANS_END
}

static int prompt_save_file( struct obj_doc * doc )
{
	int result = 0;
	GtkWidget * q = gtk_message_dialog_new( NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
		"The document \"%s\" is unsaved. Do you wish to save it before closing it?", obj_lookup_name(&doc->obj) );
	gtk_dialog_add_buttons( GTK_DIALOG(q), "Save", GTK_RESPONSE_YES, "Cancel", GTK_RESPONSE_CANCEL,
		"Discard changes" , GTK_RESPONSE_NO, NULL );
	switch ( gtk_dialog_run(GTK_DIALOG(q)) )
	{
	case GTK_RESPONSE_NO: break;
	case GTK_RESPONSE_YES: result = 2; break;
	default: result = 1;
	}
	gtk_widget_destroy(GTK_WIDGET(q));
	if (result == 2) { result = 0; on_file_save(NULL, doc); }
	return result;
}

void on_file_close( struct cmd * cmd UNUSED, void * subj )
{
	char buf[TRANS_ERR_SIZE];
	if ( subj )
	{
		ASSERT_DOC( subj );
		if ( !doc_is_all_saved(subj) && prompt_save_file(subj) ) return;
		TRANS_BEGIN_MAIN( tlo_universum, "Close file", buf )
			obj_unlink( (struct obj*)subj );
		TRANS_FAILED
			report_error_f("Cannot close the file: %s\n", buf);
		TRANS_END
	} else {
		struct obj * obj;
		WALK_OBJ_SELECTED(obj)
			if ( obj->type == OT_DOCUMENT &&
				!doc_is_all_saved(obj2doc(obj)) && prompt_save_file(obj2doc(obj)) ) return;

		TRANS_BEGIN_MAIN(tlo_universum, "Close files", buf)
		  	obj_unlink_selected();
		TRANS_FAILED
			report_error_f("Cannot close all selected files: %s\n", buf);
		TRANS_END
	}
}

void on_file_open( struct cmd * cmd UNUSED, void * subj UNUSED )
{
	OPEN_DLG_START( "Open file ...", NULL, NULL )
	{
		char err[TRANS_ERR_SIZE];
		struct obj_doc * doc;
		TRANS_BEGIN_MAIN(tlo_universum, "Open a file", err)
			doc = (struct obj_doc*)doc_load(filenames[i]);
			ASSERT_DOC(doc);
			statusbar_msg(&wnd_univ_browser.status, "File %s loaded.", filenames[i]);
			prop_set_text( obj2o(&doc->obj), string_filename, filenames[i] );
			doc_set_all_saved(doc);
		TRANS_FAILED
			report_error_f("Cannot load the file: %s\n", err);
		TRANS_END
  	}
	OPEN_DLG_END;
}

static void gui_save_document( struct obj * doc, const char * filename )
{
	char err[TRANS_ERR_SIZE];
	ASSERT_DOC( doc );
	TRANS_BEGIN_MAIN(tlo_universum, "Save File", err)
		doc_save ( obj2doc(doc), filename);
		statusbar_msg(&wnd_univ_browser.status, "File %s saved.", filename);
		prop_set_text( obj2o(doc), string_filename, filename );
		doc_set_all_saved( obj2doc(doc) );
	TRANS_FAILED
		report_error_f("Cannot save the file: %s\n", err);
	TRANS_END
}

void on_file_save( struct cmd * cmd, void * subj )
{
	char err[TRANS_ERR_SIZE];
	struct obj * volatile obj;
	if ( subj ) obj = subj; else obj = obj_get_first_selected();

	TRANS_BEGIN_MAIN(tlo_universum, "Save File", err)
		const char * filename = prop_lookup_text( obj2o(obj), string_filename );
		if ( !filename || filename == string_file_unsaved->text ) on_file_save_as( cmd, obj );
		else gui_save_document( obj, filename );
	TRANS_FAILED
		report_error_f("Cannot save the file: %s\n", err);
	TRANS_END
}

void on_file_save_as( struct cmd * cmd UNUSED, void * subj )
{
	struct obj * obj;
	if ( subj ) obj = subj; else obj = obj_get_first_selected();
	ASSERT_DOC(obj);

	const char * on = prop_lookup_text( obj2o(obj), string_name );
	if (!on) BUG();
	char nn[strlen(on) + 10];
	sprintf( nn, "%s%s.vrr", last_save_folder[0] ? last_save_folder : "", on );

	SAVE_DLG_START("Save file as ... ", NULL, nn )
	{ gui_save_document( obj, filename ); }
	SAVE_DLG_END;
}

void on_file_exit( struct cmd * cmd UNUSED, void * subj UNUSED ) { main_quit(); }

void on_file_new_page( struct cmd * cmd UNUSED, void * subj )
{
	char buf[TRANS_ERR_SIZE];
	struct obj * volatile obj;
	if ( subj ) obj = subj; else obj = obj_get_first_selected();

	ASSERT_DOC( obj );
	TRANS_BEGIN_MAIN(tlo_universum, "Create new page", buf)
		struct obj_tlo * tlo = tlo_new_named( get_new_page_noname(obj2doc(obj)) );
		obj_link( &tlo->obj, obj );
	TRANS_FAILED
		report_error_f("Cannot create new page for the file: %s\n", buf);
	TRANS_END
}

void on_file_close_page( struct cmd * cmd UNUSED, void * subj )
{
	char buf[TRANS_ERR_SIZE];
	if ( subj )
	{
		ASSERT_TLO( subj );
		TRANS_BEGIN_MAIN(tlo_universum, "Close page", buf)
			obj_unlink( subj );
		TRANS_FAILED
			report_error_f("Cannot close the page: %s\n", buf);
		TRANS_END
	} else {
		TRANS_BEGIN_MAIN(tlo_universum, "Close pages", buf)
			obj_unlink_selected();
		TRANS_FAILED
			report_error_f("Cannot close all selected pages: %s\n", buf);
		TRANS_END
	}
}

void on_group_new_view( struct cmd * cmd UNUSED, void * subj )
{
	struct go * group;
	if ( subj ) {
		group = void2go(subj);
		ASSERT_GROUP(group);
		create_view((struct go_group*)group);
	}
	else WALK_GROUP_SELECTED_GO(tlo_get_tlg(context.tlo), group) create_view((struct go_group*)group);
}

/**** basic transforms ****/

static void on_transform_basic( const struct geom_transform *t, void * subj, const char *desc )
{
	ASSERT_TLO(context.tlo);
	ASSERT(desc);
	struct go * volatile go = NULL;
	if (subj)
	{
		ASSERT_O(subj);
		go = void2go(subj);
	}
	struct geom_transform2 tt;
	int errval;
	tt.primary = *t;
	if (errval = geom_transform_invert(&tt.primary, &tt.inverted))
	{ report_error_f("Numeric error in transformation\n"); return; }

	char errs[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( context.tlo, desc, errs )
		if (go) go_select(go);
		tsort_selected_transform(context.group, &tt);
	TRANS_FAILED
		report_error_f( "%s\n", errs );
	TRANS_END;
}

void on_rotate90( struct cmd * cmd UNUSED, void * subj ) { on_transform_basic( &geom_transform_rot90, subj, "Rotate by 90 degrees" ); }
void on_rotate180( struct cmd * cmd UNUSED, void * subj ) { on_transform_basic( &geom_transform_rot180, subj, "Rotate by 180 degrees" ); }
void on_rotate270( struct cmd * cmd UNUSED, void * subj ) { on_transform_basic( &geom_transform_rot270, subj, "Rotate by 270 degrees" ); }
void on_flip_vertical( struct cmd * cmd UNUSED, void * subj ) { on_transform_basic( &geom_transform_vflip, subj, "Flip vertically" ); }
void on_flip_horizontal( struct cmd * cmd UNUSED, void * subj ) { on_transform_basic( &geom_transform_hflip, subj, "Flip horizontally" ); }

/**** windows *****/

void on_show_main_window( struct cmd * cmd, void * subj UNUSED )
{
	if ( main_window_open ){
		gtk_widget_hide( main_window );
		open_windows_cnt--;
		main_window_open--;
		cmd->state &= ~VCS_CHECKED;
	// it can be called only from view => at least one window is open
	// => it's not necessary to test if open_windows_cnt > 0
	} else {
		gtk_widget_show( main_window );
		open_windows_cnt++;
		main_window_open++;
		cmd->state |= VCS_CHECKED;
	}
}

void on_show_property_window (struct cmd *cmd, void *subj UNUSED)
{
	if (!context_property_window)
	{
		prop_window_create(NULL);
		cmd->state |= VCS_CHECKED;
	}
	else
	{
		gtk_widget_destroy(context_property_window->window);
		cmd->state &= ~VCS_CHECKED;
	}
}

void on_show_console_window( struct cmd * cmd, void * subj UNUSED )
{
	if (!console_window)
	{
		console_create();
	}
	else if ( console_window->visible )
	{
		gtk_widget_hide( console_window->window );
		console_window->visible = 0;
		cmd->state &= ~VCS_CHECKED;
	}
	else
	{
		gtk_widget_show( console_window->window );
		console_window->visible = 1;
		cmd->state |= VCS_CHECKED;
	}
}

void on_show_universe_browser( struct cmd * cmd, void * subj UNUSED )
{
	if (!(wnd_univ_browser.window)){
		create_universe_browser();
		cmd->state |= VCS_CHECKED;
		return;
	}
	if ( wnd_univ_browser.visible ){
		gtk_widget_hide( wnd_univ_browser.window );
		wnd_univ_browser.visible = 0;
		cmd->state &= ~VCS_CHECKED;
	} else {
		gtk_widget_show( wnd_univ_browser.window );
		wnd_univ_browser.visible = 1;
		cmd->state |= VCS_CHECKED;
	}
}

void on_show_undo_history( struct cmd * cmd UNUSED, void * subj UNUSED )
{
	create_wnd_undo_history(tlo_universum);
}

void on_view_clipboard( struct cmd * cmd UNUSED, void * subj UNUSED ){ create_view( tlo_get_tlg(tlo_clipboard) ); }


void on_undo_history( struct cmd *cmd UNUSED, void * subj )
{
	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_TLO( subj );
	create_wnd_undo_history( subj );
}

static void open_in_browser(const char * url)
{
	const char * browser = prop_lookup_text( obj2o(obj_universum), string_browser );
	if (!browser) { report_error("Please set the WWW browser in the Settings first!\n"); return; }
	uns run_in_term = prop_lookup_uns( obj2o(obj_universum), string_browser_in_term, 2 );
	if (run_in_term == 2) { report_error("Please set if to run the browser in a terminal first!\n"); return; }
	const char * term = NULL;
	if (run_in_term)
	{
		term = prop_lookup_text( obj2o(obj_universum), string_terminal );
		if (!term) { report_error("Please set the terminal in the Settings first!\n"); return; }
	}

	switch ( fork() )
	{
	case -1: report_error("Cannot fork!\n"); break;
	case 0:
	{
		if (run_in_term)
			execlp( term, term, "-e", browser, url, NULL );
		else
			execlp( browser, browser, url, NULL );

		report_error_f("Could not run the browser %s on %s: %s\n", browser, url, strerror(errno));
		exit(1);
	}
	default: ;
	}
}

void on_open_www(struct cmd * cmd UNUSED, void * subj UNUSED) { open_in_browser(VRR_HOMEPAGE); }
void on_help(struct cmd * cmd UNUSED, void * subj UNUSED) { open_in_browser(DOC_PATH"index.html"); }

void on_kerntest( struct cmd * cmd UNUSED, void * subj UNUSED ) { }


void main_quit(void)
{
	struct obj_doc * doc;
	WALK_UNIVERSUM_DOC(doc)
		if ( !doc_is_all_saved(doc) && prompt_save_file(doc) ) return;

	closing_all = 1;
	DBG("Close the UB window if open\n");
	if ( wnd_univ_browser.window )
		gtk_widget_destroy( wnd_univ_browser.window );
	DBG("Destroying the UB menu\n");
	if ( wnd_univ_browser.menu )
		gtk_object_sink( GTK_OBJECT(wnd_univ_browser.menu) );
	DBG("Close all open views\n");
	while ( view_list )
		gtk_widget_destroy( ((struct view*)view_list->data)->window );
	DBG("Destroying the view menu\n");
	if ( view_menu )
		gtk_object_sink( GTK_OBJECT(view_menu) );
	DBG("Close all undo history windows if open\n");
	while ( wuh_list )
		gtk_widget_destroy( ((struct wnd_undo_history*)wuh_list->data)->window );
	DBG("Close all text edit windows if open\n");
	while ( text_edit_list )
		gtk_widget_destroy( ((struct text_edit*)text_edit_list->data)->window );
	DBG("Close settings window if open\n");
	if ( wnd_settings.window ) gtk_widget_destroy( wnd_settings.window );
	DBG("Freeing font list\n");
	fonts_list_free();

	gui_prop_finish();

	DBG("Quit\n");
	gtk_main_quit();
}

char * get_new_noname(void)
{
	sprintf( noname_buf, "Untitled %d", ++noname_counter );
	return noname_buf;
}
char * get_new_page_noname( struct obj_doc * doc )
{
	uns cnt = 1;
	for ( struct obj * ooo = obj_get_first_son( &doc->obj ); ooo; ooo = obj_get_next_son(ooo) ) cnt++;
	sprintf( noname_buf, "Page %u", cnt );
	return noname_buf;
}

void refresh_gui(void)
{
	if ( closing_all ) return;
	main_window_refresh_menu();
	main_window_refresh_toolbar();
	ub_refresh_menu();
	view_refresh_menu();
	for ( GSList * l = view_list; l; l = l->next )
	{
		struct view * vw = (struct view*)l->data;
		view_refresh_toolbar(vw);
		if ( vw->tlo == context.tlo ) gtk_widget_show( vw->image_on );
		else gtk_widget_hide( vw->image_on );
		if ( tlo_is_saved(vw->tlo) ) gtk_widget_hide( vw->image_modified );
		else gtk_widget_show( vw->image_modified );
		transform_update(vw);
	}
	for ( GSList * l = wuh_list; l; l = l->next )
		wuh_refresh_toolbar( (struct wnd_undo_history*)l->data );
}

GdkPixbuf * icon_load( const char * filename )
{
	GError * err = NULL;
	GdkPixbuf * pb;
	if (!(pb = gdk_pixbuf_new_from_file(filename, &err)))
	{
		report_error_f("Error loading icon: %s\n", err->message);
		g_error_free(err);
	}
	return pb;
}

GtkWidget * image_get_from_icon( GdkPixbuf * icon )
{
	if ( icon ) return gtk_image_new_from_pixbuf(icon);
	else return gtk_image_new_from_stock("gtk-missing-image", GTK_ICON_SIZE_MENU);
}

static void resources_init(void)
{
	icon_arrow_up = icon_load(LIB_PATH"up.png");
	icon_arrow_down = icon_load(LIB_PATH"down.png");
	icon_arrow_left = icon_load(LIB_PATH"left.png");
	icon_arrow_right = icon_load(LIB_PATH"right.png");
	icon_pencil_on = icon_load(LIB_PATH"pencil-on.png");
	icon_modified = icon_load(LIB_PATH"modified.png");
	icon_rotation_left = icon_load(LIB_PATH"rotation_left.png");
	icon_rotation_right = icon_load(LIB_PATH"rotation_right.png");
	icon_zoom_out = icon_load(LIB_PATH"zoom_out.png");
	icon_zoom_in = icon_load(LIB_PATH"zoom_in.png");
}
static void resources_finish(void) { }

int main( int argc, char * * argv )
{
	gtk_disable_setlocale();
	setlocale (LC_CTYPE, "");
	gtk_init( &argc, &argv ); //read GTK options first

	kernel_init(argc, argv, 1);
	glt_gui_init();
	vcl_init();

	resources_init();
	mh_init();
	command_manager_init();
	console_init ();
#ifdef USE_LIBPAPER
	paperinit ();
#endif

	open_windows_cnt = 0;
	view_list = wuh_list = NULL;
	prop_window_list = NULL;
	text_edit_list = NULL;
	wnd_univ_browser.menu = view_menu = NULL;
	wnd_plugins.visible = 0;
	wnd_settings.window = NULL;

	main_window_create();
	gui_prop_init();
	factory_init();
	trans_clear_undo(tlo_universum); //clear the initial transactions
	refresh_gui();

	gtk_main();

	factory_finish();
	mh_finish();
	command_manager_finish();
	resources_finish();
	kernel_finish();
#ifdef USE_LIBPAPER
	paperdone();
#endif

	return 0;
}

void report_error(const char * s)
{ err(":-(  %s", s); gdk_beep(); }

void report_error_f(const char * fmt, ... )
{
	va_list ap;
        char buf[TRANS_ERR_SIZE];
        va_start( ap, fmt );
        int cnt = vsnprintf( buf, TRANS_ERR_SIZE, fmt, ap );
        va_end( ap );
        if (unlikely( cnt < 0 || cnt >= TRANS_ERR_SIZE))
		report_error("Error while reporting an error: too long a message.\n");
	else
		report_error( buf );
}

// cuts white chars form beginning and end of the string
char *str_trim(char *str)
{
	char *ptr=str + strlen(str);

	while((ptr > str) && isspace(*(ptr-1)))
		ptr--;
	if(ptr > str)
		*ptr = 0;
	ptr = str;
	while(*ptr && isspace(*ptr))
		ptr++;
	memmove(str, ptr, strlen(ptr) + 1);

	return str;
}

/* Main GUI hooks */
void mh_add_go_hook( struct obj * obj )
{
	struct go_hook * gh = xmalloc_zero(sizeof(struct go_hook));
	gh->data = (void *)&mh_init;
	gh->go_unlink = mh_go_after_delete;
	gh->go_selected = mh_go_selected;
	gh->go_unselected = mh_go_unselected;
	DBG("Main hooks: Adding a GO hook to object %p\n", obj );
	go_hook_add(obj2tlo(obj), gh);
}

void mh_init(void)
{
	struct obj_hook * oh = xmalloc_zero(sizeof(struct obj_hook));
	oh->data = (void *)&mh_init;
	oh->obj_link = mh_obj_added;
	oh->obj_unlink = mh_obj_after_delete;
	oh->obj_selected = oh->obj_unselected = mh_obj_sel_changed;
	DBG("Main hooks: adding the obj hook\n");
	obj_hook_add(oh);

	struct obj * ooo;
	WALK_OBJ_RECURSIVE( obj_universum, ooo ) if ( ooo->type == OT_TLO ) mh_add_go_hook(ooo);
}
void mh_finish(void)
{
	DBG("Main hooks: removing the obj hook\n");
	obj_hook_remove_by_data((void *)&mh_init);

	struct obj* ooo;
	struct go_hook * h;
	WALK_OBJ_RECURSIVE( obj_universum, ooo )
		if ( ooo->type == OT_TLO )
		{
			DBG("Main hooks: Removing a GO hook from %p\n", ooo );
			if ( h = go_hook_remove_by_data( obj2tlo(ooo), (void *)&mh_init )) xfree(h); else BUG();
		}
}
void mh_go_selected( struct o * o, struct go_hook * hook UNUSED )
{
	struct go * go = o2go(o);
	if ( go_get_group(go) != context.group ) return;

	if ( ++context.sel_cnt[go->type] == 1 ) context.sel_mask |= GOT_TO_VTM(go->type);
	context.sel_cnt_all++;
}
void mh_go_unselected( struct o * o, struct go_hook * hook UNUSED )
{
	struct go * go = o2go(o);
	if ( go_get_group(go) != context.group ) return;

	ASSERT( context.sel_cnt[go->type] > 0  &&  context.sel_cnt_all > 0 );
	if ( --context.sel_cnt[go->type] == 0 ) context.sel_mask &= ~(GOT_TO_VTM(go->type));
	context.sel_cnt_all--;
}
void mh_obj_sel_changed( struct o * o UNUSED, struct obj_hook * hook UNUSED ) { context.msel_valid = 0; }

void mh_obj_added( struct o * o, struct obj_hook * hook UNUSED )
{
	DBG("Main hooks: Object %p added\n", o );
	struct obj * oo, * obj = o2obj(o);
	WALK_OBJ_RECURSIVE( obj, oo ) if ( oo->type == OT_TLO ) mh_add_go_hook(oo);

	if ( obj->type == OT_DOCUMENT && (oo = obj_get_first_son(obj)) ) create_view(tlo_get_tlg(obj2tlo(oo)));
	if ( obj->type == OT_TLO ) create_view(tlo_get_tlg(obj2tlo(obj)));
}
void mh_obj_after_delete( struct o * o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook * hook UNUSED )
{
	struct obj * oo, * obj = o2obj(o);
	void * h;
	DBG("Main hooks: Object %p after delete\n", o );
	DBG("Main hooks: Removing an OBJ hook from %p\n", o );
	WALK_OBJ_RECURSIVE( obj, oo )
		if ( oo->type == OT_TLO )
		{
			DBG("Main hooks: Removing a GO hook from %p\n", oo);
			if ( h = go_hook_remove_by_data( obj2tlo(oo), (void *)&mh_init )) xfree(h); else BUG();
		}

	if ( context.tlo && o_is_ancestor( obj2o(obj), obj2o(&context.tlo->obj) ))
		change_context( context.w, NULL, NULL );
}
void mh_go_after_delete( struct o * o, struct o * owner UNUSED, struct o * prev UNUSED, struct go_hook * gh UNUSED )
{
	if ( o == context.o ) change_context( context.w, context.group, NULL );
	else if ( o == go2o(&context.group->go) ) change_context( context.w, NULL, NULL );
}
