#undef LOCAL_DEBUG
#include "lib/lib.h"
#include <stdio.h>
#include <string.h>

#include <gtk/gtk.h>

#include "kernel/kernel.h"
#include "gui/properties.h"
#include "gui/selectcolor.h"
#include "gui/glt_gui.h"
#include "gui/dialogs.h"

/* FIXME:

   - okno na centralni editaci jednotek
   - zakladani jednotky: zadat zvlast typ a subtyp, vsechny moznosti, co jsou v kernelu?
   - delete -- tak, jak je ted? Nebo nejak umoznit selektovat jednotky?

   - merge properties pro selekci -- ano, zobrazovat prunik + hromadne nastaveni

*/

enum { PW_NONEMPTY = 1, PW_CONTEXT = 2, PW_MERGED = 4 };

enum { PWT_BUG, PWT_SPIN_UNS, PWT_SPIN_REAL, PWT_COMBO, PWT_CHECKBOX, PWT_ENTRY, PWT_FUNC };

#define COLOR_INNER_TO_GTK(x) ((x) << 8)
#define COLOR_GTK_TO_INNER(x) ((x) >> 8)
#define PROP_EPSILON 0.0001

struct pst_data
{
	uns widget_type;
	union
	{
		struct { gdouble lower, upper, step, page; } spin;
		struct { uns max; string * strings; } combo;
		struct { uns dummy; } nothing;
		struct {
			void (*create_edit_func)( struct prop_item * pi );
			void (*update_edit_func)( struct prop_item * pi, struct prop * prop );
		} func;
	} data;
};

#define FN(_subtype,_what) _subtype##_edit_##_what
#define CREATE_F(_subtype)		static void FN(_subtype,create)( struct prop_item * pi )
#define UPDATE_F(_subtype,_u,_uu)	static void FN(_subtype,update)( struct prop_item * pi _u, struct prop * prop _uu )
#define CALLB_F(_subtype	)	static void FN(_subtype,callback)( GtkWidget * widget UNUSED, struct prop_item * pi )
#define GEN_FH(x) CREATE_F(x); UPDATE_F(x,,)
#define GEN_FN(x) FN(x,create), FN(x,update)

GEN_FH(color);
GEN_FH(font);
CALLB_F(color);
CALLB_F(font);
static struct pst_data pst_uns[PTU_MAX] =
{
	[PTU_UNSPECIFIED] = { PWT_SPIN_UNS, { .spin = { 0, REAL_MAX, 1.0, 10.0 }}},
	[PTU_COLOR] = { PWT_FUNC, { .func = { GEN_FN(color) }}},
	[PTU_FONT] = { PWT_FUNC, { .func = { GEN_FN(font) }}},
	[PTU_CONIC_TYPE_0P] = { PWT_COMBO, { .combo = { GEOM_CONIC_ARC_0P, string_conic_type }}},
	[PTU_CONIC_TYPE_1P] = { PWT_COMBO, { .combo = { GEOM_CONIC_ARC_1P, string_conic_type }}},
	[PTU_CONIC_TYPE_2P] = { PWT_COMBO, { .combo = { GEOM_CONIC_ARC_2P, string_conic_type }}},
	[PTU_CONIC_TYPE_3P] = { PWT_COMBO, { .combo = { GEOM_CONIC_ARC_3P, string_conic_type }}},
	[PTU_BOOLEAN] = { PWT_CHECKBOX, { .nothing = {0}}},
	[PTU_CAP_STYLE] = { PWT_COMBO, { .combo = { PSC_MAX - 1, string_stroke_cap_type }}},
};

GEN_FH(ptr);
static struct pst_data pst_ptr[PTP_MAX] =
{
	[PTP_UNSPECIFIED] = { PWT_FUNC, { .func = { GEN_FN(ptr) }}},
	[PTP_TRANSFORM] = { PWT_FUNC, { .func = { GEN_FN(ptr) }}},
	[PTP_TEX_PROCESS] = { PWT_FUNC, { .func = { GEN_FN(ptr) }}},
};

GEN_FH(filename);
CALLB_F(filename);
GEN_FH(largetext);
CALLB_F(largetext);
static struct pst_data pst_string[PTS_MAX] =
{
	[PTS_UNSPECIFIED] = { PWT_ENTRY, { .nothing = {0}}},
	[PTS_FILE_NAME] = { PWT_FUNC, { .func = { GEN_FN(filename) }}},
	[PTS_LARGE_TEXT] = { PWT_FUNC, { .func = { GEN_FN(largetext) }}},
};

static struct pst_data pst_real[PTR_MAX] =
{
	[PTR_UNSPECIFIED] = { PWT_SPIN_REAL, { .spin = { -REAL_MAX, REAL_MAX, 1.0, 10.0 }}},
	[PTR_COORDINATE] = { PWT_SPIN_REAL, { .spin = { -REAL_MAX, REAL_MAX, 1.0, 10.0 }}},
	[PTR_NON_NEGATIVE] = { PWT_SPIN_REAL, { .spin = { 0.0, REAL_MAX, 1.0, 10.0 }}},
	[PTR_ANGLE_PI] = { PWT_SPIN_REAL, { .spin = { 0.0, R_PI, R_PI/12, R_PI/4 }}},
	[PTR_ANGLE_2PI] = { PWT_SPIN_REAL, { .spin = { 0.0, R_2PI, R_PI/12, R_PI/4 }}},
	[PTR_ANGLE_4PI] = { PWT_SPIN_REAL, { .spin = { -R_2PI, R_2PI, R_PI/12, R_PI/4 }}},
	[PTR_REFERENCE] = { PWT_SPIN_REAL, { .spin = { 0.0, 1.0, 0.1, 0.2 }}}
};

static struct pst_data * pst_info[PT_MAX]  = { pst_uns, pst_ptr, pst_string, pst_real };

//TODO: update lists
static GtkListStore * pq_none_list, * pq_len_list, * pq_angle_list, * pq_ref_list;
static GtkListStore ** pq_lists[PQ_MAX] = { &pq_none_list, &pq_len_list, &pq_angle_list, &pq_ref_list };

GSList * prop_window_list;

/** The one context window, if exists */
struct prop_window * context_property_window = NULL;

static void prop_window_destroy( GtkWidget * w UNUSED, struct prop_window * pw );
static void pw_empty( struct prop_window * pw );
static void pw_fill( struct prop_window * pw, struct o * o );
static void prop_value_changed( GtkWidget * w, struct prop_item * pi );
static void prop_unit_changed( GtkWidget * w, struct prop_item * pi );

/*static void add_new_property( GtkWidget * w UNUSED, struct prop_window * pw );
static void delete_property( GtkWidget * w UNUSED, struct prop_window * pw );
static void manage_units( GtkWidget * w UNUSED, struct prop_window * pw );*/

static void prop_go_unlink(struct o *o, struct o * owner UNUSED, struct o * prev UNUSED, struct go_hook * h);
static void prop_go_create(struct o *o, struct prop* p, struct go_hook * h);
static void prop_go_change(struct o *o, struct go_hook * h);
static void prop_go_delete(struct o *o, struct prop* p, struct go_hook * h);
static void prop_obj_unlink(struct o *o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook * h);
static void prop_obj_create(struct o *o, struct prop* p, struct obj_hook * h);
static void prop_obj_change(struct o *o, struct obj_hook * h);
static void prop_obj_delete(struct o *o, struct prop* p, struct obj_hook * h);
static gboolean entry_focus_out( GtkWidget * w, GdkEventFocus * event UNUSED, struct prop_item * pi );

#define PROP_MULT ((prop_subtype2quantity[prop->type][prop->subtype] == PQ_NONE) ? 1.0 : prop_get_unit_multiplier(pi->o,prop))

#define REF(_name, _string) string_##_name = string_lookup_ref(_string)

static string string_cm, string_m, string_pi, string_inch, string_pt;
string GUI_STRINGS;

static int setting_range;
PROP_STORE_DEFINE(ps_global);
static PROP_STORE_DEFINE(ps_recycler);

void prop_store_set( PROP_STORE_DEFINE(ps), const char * name, uns type, prop_value pv )
{
	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( PROP_STORE_TLO(ps), "Store property", buf )
		prop_create( PROP_STORE_O(ps), string_lookup(name), type, 0, PROP_UNIT_MAX, pv );
	TRANS_FAILED
		DBG("Could not store value %s\n", name);
	TRANS_END
}

static void set_default_setting( string key, uns type, uns subtype, prop_value pv )
{
	struct o * o = obj2o(obj_universum);
	struct prop * prop = prop_find(o, key );
	if ( prop && (prop->type != type || prop->subtype != subtype ) )
	{
		prop_delete(o, prop);
		prop = NULL;
	}
	if ( !prop )
		prop_create(o, key, type, subtype, PROP_UNIT_MAX, pv );
}

void gui_prop_init(void)
{
	REF(tolerance, "snap-tolerance");
	REF(editor, "text-editor");
	REF(editor_in_term, "run-text-editor-in-terminal");
	REF(browser, "www-browser");
	REF(browser_in_term, "run-www-browser-in-terminal");
	REF(terminal, "terminal");
	REF(grid_x1, "grid-x1");
	REF(grid_y1, "grid-y1");
	REF(grid_x2, "grid-x2");
	REF(grid_y2, "grid-y2");
	REF(rulerx_res, "ruler-x-res");
	REF(rulery_res, "ruler-y-res");
	REF(drag_image, "drag-image");

	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Initialize properties", buf )
		//pokusne pridani jednotek
		//TODO: ukladat nekam do nastaveni
		unit_set( REF(cm, "cm"), 10.0, PQ_LENGTH );
		unit_set( REF(m, "m"), 1000.0, PQ_LENGTH );
		unit_set( REF(inch, "inch"), 25.4, PQ_LENGTH );
		unit_set( REF(pt, "pt"), 25.4/72, PQ_LENGTH );
		unit_set( REF(pi, "Pi"), R_PI, PQ_ANGLE );

		PROP_STORE_NEW(ps_recycler);
		PROP_STORE_NEW(ps_global);

		//set defaults for unloaded settings
		set_default_setting(string_tolerance, PT_UNS, PTU_UNSPECIFIED, (prop_value){ .u = TOLERANCE_PIX });
		set_default_setting(string_editor, PT_STRING, PTS_UNSPECIFIED, (prop_value){ .s = string_lookup(EXTERNAL_TEXT_EDITOR) });
		set_default_setting(string_editor_in_term, PT_UNS, PTU_BOOLEAN, (prop_value){ .u = TEXT_EDITOR_RUN_IN_TERM });
		set_default_setting(string_browser, PT_STRING, PTS_UNSPECIFIED, (prop_value){ .s = string_lookup(WWW_BROWSER) });
		set_default_setting(string_browser_in_term, PT_UNS, PTU_BOOLEAN, (prop_value){ .u = WWW_BROWSER_RUN_IN_TERM });
		set_default_setting(string_terminal, PT_STRING, PTS_UNSPECIFIED, (prop_value){ .s = string_lookup(FAVOURITE_TERMINAL) });
		set_default_setting(string_grid_x1, PT_REAL, PTR_COORDINATE, (prop_value){ .r = GRID_X1 });
		set_default_setting(string_grid_y1, PT_REAL, PTR_COORDINATE, (prop_value){ .r = GRID_Y1 });
		set_default_setting(string_grid_x2, PT_REAL, PTR_COORDINATE, (prop_value){ .r = GRID_X2 });
		set_default_setting(string_grid_y2, PT_REAL, PTR_COORDINATE, (prop_value){ .r = GRID_Y2 });
		set_default_setting(string_rulerx_res, PT_REAL, PTR_COORDINATE, (prop_value){ .r = RULERX_RES });
		set_default_setting(string_rulery_res, PT_REAL, PTR_COORDINATE, (prop_value){ .r = RULERY_RES });
		set_default_setting(string_drag_image, PT_UNS, PTU_BOOLEAN, (prop_value){ .u = PANNING_DRAG_IMAGE });

	TRANS_FAILED
	TRANS_END

	int i;
	for ( i = 0; i < PQ_MAX; i++ )
		*(pq_lists[i]) = GTK_LIST_STORE(gtk_list_store_new( 1, G_TYPE_STRING ));

	//fill the unit lists
	for ( i = 0; i < PROP_UNIT_MAX; i++ )
		if ( unit_array[i].used )
		{
			GtkTreeIter ti;
			GtkListStore * li = *(pq_lists[unit_array[i].quantity]);
			gtk_list_store_append( li, &ti );
			gtk_list_store_set( li, &ti, 0, unit_array[i].name->text, -1 );
		}
}

void gui_prop_finish(void)
{
	while ( prop_window_list )
		gtk_widget_destroy( ((struct prop_window*)prop_window_list->data)->window );
	if ( context_property_window )
		gtk_widget_destroy(context_property_window->window);

	//TODO: presunout pokusne jednotky na lepsi misto
	string_unref(string_cm);
	string_unref(string_m);
	string_unref(string_pi);
	string_unref(string_inch);
	string_unref(string_pt);

	string array[] = { GUI_STRINGS };
	for(int i = sizeof(array) / sizeof(string) - 1; i >= 0; i--)
		string_unref(array[i]);

	PROP_STORE_DESTROY(ps_recycler);
	PROP_STORE_DESTROY(ps_global);
}

static inline int prop_is_recyclable( struct prop * prop )
{ return prop->key == string_fill_color || prop->key == string_stroke_color ||
	prop->key == string_stroke_width || prop->key == string_stroke_cap; }

void gui_prop_recycle( struct o * o )
{
	DBG("Recycle properties for o %p\n", o );
	struct prop * volatile prop;
	struct prop * rec_prop;
	struct o * or = PROP_STORE_O(ps_recycler);
	uns unit;
	char buf[TRANS_ERR_SIZE];
	WALK_PROP(o, prop)
	// check if the ps_recycler has the same property (including the same type and subtype)
	// if so, change the o's property accordingly
		if ( prop_is_recyclable(prop) && (rec_prop = prop_find(or, prop->key)) &&
			rec_prop->type == prop->type && rec_prop->subtype == prop->subtype )
		{
			// the change can fail for various reasons (e.g. if the value is invalid for this o)
			DBG("Recycle property %s\n", prop->key->text );
			TRANS_BEGIN(buf)
				prop_set( o, prop->key, prop_get_value(or, rec_prop) );
				unit = prop_get_unit( or, rec_prop );
				if (unit != PROP_UNIT_MAX) prop_set_unit( o, prop, unit );
			TRANS_FAILED
				DBG("Could not recycle property %s: %s\n", prop->key->text, buf);
			TRANS_END
		}
}

static void gui_prop_remember( struct o * o, struct prop * prop )
{
	if (!prop_is_recyclable(prop)) return;

	struct prop * rec_prop;
	struct o * or = PROP_STORE_O(ps_recycler);
	char buf[TRANS_ERR_SIZE];
	// we want to rewrite the old property and delete it if incompatible
	TRANS_BEGIN_MAIN( PROP_STORE_TLO(ps_recycler), "Remember the property for further use", buf )
		if ( (rec_prop = prop_find(or, prop->key)) && (rec_prop->type != prop->type || rec_prop->subtype != prop->subtype) )
		{
			DBG("Property %s remembered with different type/subtype, delete\n", prop->key->text);
			prop_delete( or, rec_prop );
		}
		DBG("Remember property %s\n", prop->key->text );
		prop_create( or, prop->key, prop->type, prop->subtype, prop_get_unit(o, prop), prop_get_value(o, prop));
	TRANS_FAILED
		DBG("Could not remember property %s: %s\n", prop->key->text, buf);
	TRANS_END
}

void prop_window_create( struct o * o )
{
	if ( !o && context_property_window )
	{
		DBG("Context property window already exists, will not be created.\n");
		gtk_window_present( GTK_WINDOW(context_property_window->window) );
		return;
	}

	struct prop_window * pw = xmalloc_zero(sizeof(struct prop_window));
	pw->type = WT_PROPERTIES;
	pw->proxy = SCM_BOOL_F;
	pw->flags = 0;
	pw->items = NULL;
	if ( !o )
	{ pw->flags |= PW_CONTEXT; context_property_window = pw; }

	pw->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_size_request(pw->window, 300, 300); //FIXME at se okno rozsiruje podle potreby
	GtkWidget * vbox = gtk_vbox_new( FALSE, 5 );
	gtk_container_add( GTK_CONTAINER(pw->window), vbox );
	g_signal_connect( G_OBJECT(pw->window), "destroy", G_CALLBACK(prop_window_destroy), pw );

	GtkWidget * scw = gtk_scrolled_window_new(NULL, NULL);
	gtk_box_pack_start( GTK_BOX(vbox), scw, TRUE, TRUE, 5 );
	pw->table = GTK_TABLE(gtk_table_new(1, 3, FALSE));
	GtkWidget * viewport = gtk_viewport_new(NULL, NULL);
	gtk_container_add( GTK_CONTAINER(viewport), GTK_WIDGET(pw->table) );
	gtk_container_add( GTK_CONTAINER(scw), GTK_WIDGET(viewport) );
	g_object_set( G_OBJECT(viewport), "shadow-type", GTK_SHADOW_NONE, NULL );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(scw), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC );

	GtkWidget * hbutbox = gtk_hbutton_box_new();
	gtk_button_box_set_layout( GTK_BUTTON_BOX(hbutbox), GTK_BUTTONBOX_SPREAD );
	gtk_box_pack_start( GTK_BOX(vbox), hbutbox, FALSE, TRUE, 0 );

	/*GtkWidget * button;
	gtk_box_pack_start( GTK_BOX(hbutbox), pw->button_add = gtk_button_new_with_label( "Add ..." ), FALSE, FALSE, 0 );
	g_signal_connect( G_OBJECT(pw->button_add), "clicked", G_CALLBACK(add_new_property), pw );
	gtk_box_pack_start( GTK_BOX(hbutbox), pw->button_delete = gtk_button_new_with_label( "Delete" ), FALSE, FALSE, 0 );
	g_signal_connect( G_OBJECT(pw->button_delete), "clicked", G_CALLBACK(delete_property), pw );
	gtk_box_pack_start( GTK_BOX(hbutbox), button = gtk_button_new_with_label( "Units ..." ), FALSE, FALSE, 0 );
	g_signal_connect( G_OBJECT(button), "clicked", G_CALLBACK(manage_units), pw );*/

	pw->status.statusbar = gtk_statusbar_new();
	pw->status.context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(pw->status.statusbar), "Properties");
	gtk_box_pack_start( GTK_BOX(vbox), pw->status.statusbar, FALSE, FALSE, 0 );

	prop_window_list = g_slist_prepend( prop_window_list, (gpointer)pw );
	pw_fill( pw, o ? o : context.o );

	gtk_widget_show_all(pw->window);

	pw->go_hook.data = pw;
	pw->go_hook.go_unlink = prop_go_unlink;
	pw->go_hook.go_prop_create = prop_go_create;
	pw->go_hook.go_prop_delete = prop_go_delete;
	pw->go_hook.go_prop_change = prop_go_change;

	pw->obj_hook.data = pw;
	pw->obj_hook.obj_unlink = prop_obj_unlink;
	pw->obj_hook.obj_prop_create = prop_obj_create;
	pw->obj_hook.obj_prop_delete = prop_obj_delete;
	pw->obj_hook.obj_prop_change = prop_obj_change;
}

static void prop_window_destroy( GtkWidget * w UNUSED, struct prop_window * pw )
{
	pw_empty(pw);
	if ( pw->flags & PW_CONTEXT ) context_property_window = NULL;
	prop_window_list = g_slist_remove( prop_window_list, (gpointer)pw );
	window_proxy_clean( (struct window*)pw );
	xfree(pw);
}
static void prop_widget_destroy( GtkWidget * w, gpointer data UNUSED ){ gtk_widget_destroy(w); }
static void pw_empty( struct prop_window * pw )
{
	ASSERT(pw);
	//destroy all prop_items of the widgets
	gtk_container_foreach( GTK_CONTAINER(pw->table), prop_widget_destroy, NULL );
	if (!pw->o) return;

	obj_hook_remove(&pw->obj_hook);
	if ( pw->tlo )
	{
		go_hook_remove(pw->tlo, &pw->go_hook);
		pw->tlo = NULL;
	}
	gtk_table_resize( GTK_TABLE(pw->table), 1, 3 );
	pw->flags &= ~PW_NONEMPTY;
	xfree(pw->items);
	pw->items = NULL;
	pw->o = NULL;
}

static void pw_set_window_title( struct prop_window * pw )
{
	if ( pw->flags & PW_NONEMPTY )
	{
		string s = o_get_name(pw->o);
		char * name = (s == string_noname) ? "" : s->text;
		const char * tn = ((pw->o->kind == T_OBJ) ? string_obj_type : string_go_type)[pw->o->type]->text;
		char buf[strlen(name) + strlen(tn) + 25];
		sprintf( buf, "Properties of %s %s", tn, name );
		gtk_window_set_title( GTK_WINDOW(pw->window), buf );
	} else gtk_window_set_title( GTK_WINDOW(pw->window), "Properties" );
}

static void pw_fill( struct prop_window * pw, struct o * o )
{
	ASSERT(pw);

	pw->o = o;
	if ( !o ) { pw->flags &= ~PW_NONEMPTY; pw->prop_count = 0; } else pw->flags |= PW_NONEMPTY;
	pw_set_window_title(pw);
	if ( !o )
	{
		gtk_table_attach( GTK_TABLE(pw->table), gtk_label_new("no object to display"), 0, 3, 0, 1,
			GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0 );
		gtk_widget_show_all( GTK_WIDGET(pw->table) );
		return;
	}

	obj_hook_add(&pw->obj_hook);
	if ( o->kind == T_GO ) go_hook_add(pw->tlo = go_get_tlo(o2go(o)), &pw->go_hook);
	else pw->tlo = NULL;

	pw->prop_count = prop_get_count(o);
	pw->items = xmalloc_zero(sizeof(struct prop_item) * pw->prop_count);
	gtk_table_resize( GTK_TABLE(pw->table), pw->prop_count, 3 );

	struct prop * prop;
	struct prop_item * pi;
	uns counter = 0;
	WALK_PROP(o, prop)
	{
		ASSERT(prop && prop->key);
		pi = pw->items + counter;
		pi->o = o;
		pi->key = prop->key;
		prop_value_edit_create( pi, prop );
		prop_unit_edit_create( pi, prop );
		prop_sync( pi, prop );

		GtkWidget * label = gtk_label_new(prop->key->text);
		g_object_set( G_OBJECT(label), "xalign", 0.0, NULL );
		gtk_table_attach( pw->table, label, 0, 1, counter, counter+1, GTK_FILL, GTK_FILL, 5, 0 );
		if (pi->unit_edit)
		{
			gtk_table_attach( pw->table, pi->value_edit, 1, 2, counter, counter+1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0 );
			gtk_table_attach( pw->table, pi->unit_edit, 2, 3, counter, counter+1, GTK_FILL, GTK_FILL, 0, 0 );
		} else
		gtk_table_attach( pw->table, pi->value_edit, 1, 3, counter, counter+1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0 );
		counter++;
	}
	gtk_widget_show_all( GTK_WIDGET(pw->table) );
}

void prop_item_init( struct prop_item * pi, struct o * o, string key )
{
	struct prop * prop = prop_find( o, key );
	ASSERT(prop);
	pi->o = o;
	pi->key = key;
	prop_value_edit_create( pi, prop );
	prop_unit_edit_create( pi, prop );
	prop_sync( pi, prop );
}

void prop_value_edit_create( struct prop_item * pi, struct prop * prop )
{
	ASSERT( prop && prop->key );
	struct pst_data data = pst_info[prop->type][prop->subtype];

	switch( data.widget_type )
	{
	case PWT_BUG: BUG(); break;
	case PWT_SPIN_UNS:
		pi->value_edit = gtk_spin_button_new_with_range( data.data.spin.lower, data.data.spin.upper, data.data.spin.step );
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(pi->value_edit), 0.0 );
		gtk_spin_button_set_digits(GTK_SPIN_BUTTON(pi->value_edit), 0 );
		//g_object_set( G_OBJECT(pi->value_edit), "xalign", 1.0, NULL );
		g_signal_connect( G_OBJECT(pi->value_edit), "value-changed", G_CALLBACK(prop_value_changed), pi);
		break;
	case PWT_SPIN_REAL:
		pi->value_edit = gtk_spin_button_new_with_range( data.data.spin.lower, data.data.spin.upper, data.data.spin.step );
		gtk_spin_button_set_digits(GTK_SPIN_BUTTON(pi->value_edit), 4 );
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(pi->value_edit), 0.0 );
		//g_object_set( G_OBJECT(pi->value_edit), "xalign", 1.0, NULL );
		g_signal_connect( G_OBJECT(pi->value_edit), "value-changed", G_CALLBACK(prop_value_changed), pi);
		break;
	case PWT_COMBO:
		pi->value_edit = gtk_combo_box_new_text();
		for ( uns i = 0; i <= data.data.combo.max; i++ )
			gtk_combo_box_append_text( GTK_COMBO_BOX(pi->value_edit), data.data.combo.strings[i]->text );
		g_signal_connect( G_OBJECT(pi->value_edit), "changed", G_CALLBACK(prop_value_changed), pi );
		break;
	case PWT_CHECKBOX:
		pi->value_edit = gtk_check_button_new();
		g_signal_connect( G_OBJECT(pi->value_edit), "toggled", G_CALLBACK(prop_value_changed), pi);
		break;
	case PWT_ENTRY:
		pi->value_edit = gtk_entry_new();
		g_signal_connect(G_OBJECT(pi->value_edit), "focus-out-event", G_CALLBACK(entry_focus_out), pi);
		g_signal_connect(G_OBJECT(pi->value_edit), "activate", G_CALLBACK(prop_value_changed), pi);
		break;
	case PWT_FUNC:
		data.data.func.create_edit_func( pi );
		break;
	}
	ASSERT(pi->value_edit);
}

static gdouble tame_number( real x, real mult, int * was_bad )
{
	DBG("Taming number %f with mult %f ", x, mult );
	x *= mult;
	if (isinf(x)) { x = signbit(x) ? -REAL_MAX : REAL_MAX; if (was_bad) *was_bad = 1; }
	else { if (was_bad) *was_bad = 0; }
	DBG(" = %f\n", (gdouble)x );
	return (gdouble)x; //FIXME: staci to takhle?
}

static real untame_number( gdouble y, real mult, int * was_bad )
{
	DBG("Untaming number %f with mult %f ", y, mult );
	real x = (real)y * mult;
	if (isinf(x)) { x = signbit(x) ? -REAL_MAX : REAL_MAX; if (was_bad) *was_bad = 1; }
	else { if (was_bad) *was_bad = 0; }
	DBG(" = %f\n", x );
	return x;
}

#define TRANS_PROP_CHANGE(_val) \
{ \
	char buf[TRANS_ERR_SIZE]; \
	char transname[strlen(pi->key->text) + 25]; \
	sprintf( transname, "Change the %s property", pi->key->text ); \
	factory_op_break(); \
	TRANS_BEGIN_MAIN( pi->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(pi->o)), transname, buf) \
		DBG("\n*** Changing the value of property %s\n", pi->key->text ); \
		prop_set( pi->o, pi->key, _val ); \
		gui_prop_remember( pi->o, prop ); \
	TRANS_FAILED \
		report_error_f("Could not change property: %s\n", buf); \
		prop_sync(pi, prop); /*return to the original value*/ \
	TRANS_END \
}

#define FIND_PROP \
	ASSERT( pi && pi->o && pi->key && pi->value_edit ); \
	struct prop *prop; \
	if (!(prop = prop_find( pi->o, pi->key ))) \
	{ report_error_f("Property which should be changed does not exist.\n"); return; }

static void prop_value_changed( GtkWidget * w UNUSED, struct prop_item * pi )
{
	if (setting_range) return; // prevent unwanted changes when the original value was clamped due to range change
	FIND_PROP;
	struct pst_data data = pst_info[prop->type][prop->subtype];
	volatile prop_value new_value;
	volatile string s = NULL;

	switch ( data.widget_type )
	{
	case PWT_BUG: BUG(); break;
	case PWT_SPIN_UNS:
		new_value.u = (uns)gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(pi->value_edit));
		break;
	case PWT_SPIN_REAL:
	{
		int bad;
		new_value.r = untame_number(gtk_spin_button_get_value(GTK_SPIN_BUTTON(pi->value_edit)), PROP_MULT, &bad );
		//if (bad) statusbar_msg(&pi->pw->status, "Too big value in property %s, changed to 0.", pi.prop->key->text);
		break;
	}
	case PWT_COMBO:
	{
		int active = gtk_combo_box_get_active(GTK_COMBO_BOX(pi->value_edit));
		if (active == -1) new_value.u = prop_get_value(pi->o, prop).u;
		else new_value.u = (uns)active;
		break;
	}
	case PWT_CHECKBOX:
		new_value.u = (uns)gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pi->value_edit));
		break;
	case PWT_ENTRY:
		s = new_value.s = string_lookup_ref(gtk_entry_get_text(GTK_ENTRY(pi->value_edit)));
		break;
	default: BUG();
	}

	real diff;
	if ( (prop->type == PT_REAL &&
			(diff = fabs(new_value.r - prop_get_value(pi->o,prop).r)) > untame_number( PROP_EPSILON, PROP_MULT, NULL)) ||
		(prop->type != PT_REAL && !prop_value_compare( prop->type, prop_get_value(pi->o, prop), new_value )))
	{
		if (prop->type == PT_REAL) { DBG("Difference is %f\n", diff); }
		TRANS_PROP_CHANGE(new_value);
	}

	if (s) string_unref(s);
}

void prop_sync( struct prop_item * pi, struct prop * prop )
{
	ASSERT( pi && prop && pi->o && pi->key && pi->key == prop->key && pi->value_edit );
	//I suppose that the count of properties and the order of the list has not
	//changed without letting me know

	struct pst_data data = pst_info[prop->type][prop->subtype];
	prop_value new_value = prop_get_value( pi->o, prop );

	switch ( data.widget_type )
	{
	case PWT_BUG: BUG(); break;
	case PWT_SPIN_UNS:
		if (new_value.u != (uns)gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(pi->value_edit)))
			gtk_spin_button_set_value(GTK_SPIN_BUTTON(pi->value_edit), (gdouble)new_value.u );
		break;
	case PWT_SPIN_REAL:
	{
		int bad;
		real mult = 1.0 / PROP_MULT;
		DBG("Property %s has unit %d, 1/mult = %f \n", pi->key->text, prop_get_unit(pi->o, prop), mult );
		gdouble x = tame_number(new_value.r, mult, &bad);
		//if (bad) statusbar_msg(&pi->pw->status, "Too big value in property %s, changed to 0.", pi.prop->key->text);
		gdouble diff = fabs(x - gtk_spin_button_get_value(GTK_SPIN_BUTTON(pi->value_edit)));
		DBG("Difference is %f\n", diff );
		if ( diff > PROP_EPSILON)
		{
			gtk_spin_button_set_value(GTK_SPIN_BUTTON(pi->value_edit), x );
			DBG("Setting the value of %s to %f\n", pi->key->text, x );
		}
		//set range
		x = tame_number(data.data.spin.step, mult, NULL);
		gdouble inc;
		gtk_spin_button_get_increments(GTK_SPIN_BUTTON(pi->value_edit), &inc, NULL);
		if ( fabs(x - inc) > PROP_EPSILON )
		{
			setting_range = 1;
			gtk_spin_button_set_range(GTK_SPIN_BUTTON(pi->value_edit),
				tame_number(data.data.spin.lower, mult, NULL), tame_number(data.data.spin.upper, mult, NULL) );
			gtk_spin_button_set_increments(GTK_SPIN_BUTTON(pi->value_edit),
				x, tame_number(data.data.spin.page, mult, NULL) );
			setting_range = 0;
		}
		break;
	}
	case PWT_COMBO:
		if ((int)new_value.u != gtk_combo_box_get_active(GTK_COMBO_BOX(pi->value_edit)))
			gtk_combo_box_set_active(GTK_COMBO_BOX(pi->value_edit), new_value.u);
		break;
	case PWT_CHECKBOX:
		if (new_value.u != (uns)gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pi->value_edit)))
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pi->value_edit), new_value.u > 0 );
		break;
	case PWT_ENTRY:
		if (strcmp(new_value.s->text, gtk_entry_get_text(GTK_ENTRY(pi->value_edit))))
			gtk_entry_set_text( GTK_ENTRY(pi->value_edit), new_value.s->text );
		break;
	case PWT_FUNC:
		data.data.func.update_edit_func(pi, prop);
		break;
	}
	gtk_widget_set_sensitive( pi->value_edit, !(prop->flags & PTF_READ_ONLY));
	if ( pi->unit_edit )
	{
		// find out the index of the unit in our combobox
		int index = -1;
		uns q = prop_subtype2quantity[prop->type][prop->subtype];
		for ( int i = prop_get_unit(pi->o, prop); i >= 0; i-- )
			if ( unit_array[i].used && unit_array[i].quantity == q ) index++;

		DBG("Translated unit %u to item %d of quantity %u\n", prop_get_unit(pi->o, prop), index, q );
		if ( index == -1 ) BUG(); //there is no chosen item
		if ( index != gtk_combo_box_get_active( GTK_COMBO_BOX(pi->unit_edit) ))
			gtk_combo_box_set_active( GTK_COMBO_BOX(pi->unit_edit), index );
	}
}

void prop_unit_edit_create( struct prop_item * pi, struct prop * prop )
{
	uns quantity = prop_subtype2quantity[prop->type][prop->subtype];
	if (quantity == PQ_NONE)
	{ pi->unit_edit = NULL; return; }

	pi->unit_edit = gtk_combo_box_new_with_model(GTK_TREE_MODEL(*(pq_lists[quantity])));
	GtkCellRenderer * cr = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start( GTK_CELL_LAYOUT(pi->unit_edit), cr, FALSE );
	gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(pi->unit_edit), cr, "text", 0 );

	//gtk_combo_box_set_active( GTK_COMBO_BOX(pi->unit_edit), prop_get_unit( pi->o, prop ) );
	g_signal_connect( G_OBJECT(pi->unit_edit), "changed", G_CALLBACK(prop_unit_changed), pi );
}

void prop_unit_changed( GtkWidget * w UNUSED, struct prop_item * pi )
{
	ASSERT( pi && pi->o && pi->key && pi->unit_edit );
	struct prop *prop;
	if (!(prop = prop_find( pi->o, pi->key )))
	{ report_error_f("Property whose unit should be changed does not exist.\n"); return; }

	//get the unit name from the widget and get the appropriate unit number
	int index = gtk_combo_box_get_active( GTK_COMBO_BOX(pi->unit_edit) );
	if ( index == -1 ) return;

	volatile uns unit = 0;
	uns q = prop_subtype2quantity[prop->type][prop->subtype];
	for ( index++; unit < PROP_UNIT_MAX; unit++ )
		if ( unit_array[unit].used && unit_array[unit].quantity == q ) if (--index == 0) break;
	DBG("Translated item %d of quantity %u to unit %u\n", gtk_combo_box_get_active( GTK_COMBO_BOX(pi->unit_edit)), q, unit );
	if (index) BUG(); //there is no chosen unit

	if ( unit != prop_get_unit(pi->o, prop) )
	{
		factory_op_break();
		char buf[TRANS_ERR_SIZE];
		TRANS_BEGIN_MAIN( pi->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(pi->o)), "Change unit", buf)
			DBG("\n*** Set unit of property %s to %d\n", pi->key->text, unit );
			prop_set_unit( pi->o, prop, unit );
			gui_prop_remember( pi->o, prop );
		TRANS_FAILED
			report_error_f("Could not change property unit: %s\n", buf);
			prop_sync(pi, prop); //return to the original value
		TRANS_END
		prop_sync(pi, prop); //FIXME: nemelo by byt potreba, mel by se volat hook
	}
}

/*static void add_new_property( GtkWidget * w UNUSED, struct prop_window * pw UNUSED )
{
	//TODO: open the Add Property Dialog
}
static void delete_property( GtkWidget * w UNUSED, struct prop_window * pw UNUSED )
{
	//TODO: delete the selected properties
}
static void manage_units( GtkWidget * w UNUSED, struct prop_window * pw UNUSED )
{
	//TODO: open the unit window
}*/

void prop_window_context_changed(struct o * o)
{
	if ( context_property_window && o != context_property_window->o )
	{
		pw_empty( context_property_window );
		pw_fill( context_property_window, o );
	}
}
static void pw_unset(struct o *o, struct prop_window * pw )
{
	if (o_is_ancestor(o, pw->o))
	{
		if ( pw->flags & PW_CONTEXT ) { pw_empty(pw); pw_fill(pw, NULL); } else gtk_widget_destroy(pw->window);
	}
}
static inline void pw_refill( struct o * o, struct prop_window * pw )
{ if ( o == pw->o ) { pw_empty(pw); pw_fill(pw, o); } }

static void pw_sync(struct o *o, struct prop_window * pw)
{
	if (o == pw->o)
	{
		struct prop * prop;
		uns counter = 0;
		WALK_PROP(o, prop) prop_sync( pw->items + counter++, prop );
		pw_set_window_title(pw);
	}
}
static void prop_go_unlink(struct o *o, struct o * owner UNUSED, struct o * prev UNUSED, struct go_hook * h)
{ pw_unset(o,(struct prop_window*)h->data); }
static void prop_obj_unlink(struct o *o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook * h)
{ pw_unset(o,(struct prop_window*)h->data); }
static void prop_go_create(struct o *o, struct prop* p UNUSED, struct go_hook * h)
{ pw_refill(o,(struct prop_window*)h->data); }
static void prop_go_delete(struct o *o, struct prop* p UNUSED, struct go_hook * h)
{ pw_refill(o,(struct prop_window*)h->data); }
static void prop_obj_create(struct o *o, struct prop* p UNUSED, struct obj_hook * h)
{ pw_refill(o,(struct prop_window*)h->data); }
static void prop_obj_delete(struct o *o, struct prop* p UNUSED, struct obj_hook * h)
{ pw_refill(o,(struct prop_window*)h->data); }
static void prop_go_change(struct o *o, struct go_hook * h) { pw_sync(o,(struct prop_window*)h->data);}
static void prop_obj_change(struct o *o, struct obj_hook * h) { pw_sync(o,(struct prop_window*)h->data);}

static gboolean entry_focus_out( GtkWidget * w, GdkEventFocus * event UNUSED, struct prop_item * pi )
{ prop_value_changed(w, pi); return FALSE; }

CREATE_F(color)
{
	GtkWidget * w = pi->value_edit = gtk_button_new();
	GtkWidget * child = gtk_drawing_area_new();
	gtk_widget_set_size_request( child, -1, 10 );
	gtk_container_add( GTK_CONTAINER(pi->value_edit), child );
	g_signal_connect( G_OBJECT(w), "clicked", G_CALLBACK(color_edit_callback), pi );
}

UPDATE_F(color,,)
{
	GdkColor color;
	uns r, g, b, alpha;
	uns2rgba_uns( prop_get_value(pi->o, prop).u, &r, &g, &b, &alpha);
	color.red = COLOR_INNER_TO_GTK(r); color.green = COLOR_INNER_TO_GTK(g);
	color.blue = COLOR_INNER_TO_GTK(b);
	GtkWidget * child = gtk_bin_get_child(GTK_BIN(pi->value_edit));
	gtk_widget_modify_bg( child, GTK_STATE_NORMAL, &color );
	gtk_widget_modify_bg( child, GTK_STATE_ACTIVE, &color );
	gtk_widget_modify_bg( child, GTK_STATE_PRELIGHT, &color );
	gtk_widget_modify_bg( child, GTK_STATE_SELECTED, &color );
	gtk_widget_modify_bg( child, GTK_STATE_INSENSITIVE, &color );
}
CALLB_F(color)
{
	FIND_PROP;
	GtkWidget *color_selector = wnd_select_color_new();
	GtkWidget *dialog = gtk_dialog_new_with_buttons("Color selection", NULL,
			GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
			GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
	gtk_container_add (GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), color_selector);
	gtk_widget_show_all(dialog);
	gtk_window_resize(GTK_WINDOW(dialog), 400, 300);

	uns r, g, b, a;
	uns2rgba_uns( prop_get_value(pi->o, prop).u, &r, &g, &b, &a );
	wnd_select_color_set_rgb((WndSelectColor *) color_selector, (int)r, (int)g, (int)b );
	wnd_select_color_set_alpha((WndSelectColor *) color_selector, (int)a );

	if ( gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT )
	{
		int na, nr, ng, nb;
		wnd_select_color_get_rgb((WndSelectColor*) color_selector, &nr, &ng, &nb );
		wnd_select_color_get_alpha((WndSelectColor *) color_selector, &na );
		prop_value new_val;
		new_val.u = rgba_uns2uns( nr, ng, nb, na );
		TRANS_PROP_CHANGE(new_val);
	}
	gtk_widget_destroy(dialog);
}

CREATE_F(font)
{
	pi->value_edit = gtk_button_new_with_label("Change font ...");
	gtk_button_set_relief(GTK_BUTTON(pi->value_edit), GTK_RELIEF_NONE);
	g_signal_connect(G_OBJECT(pi->value_edit), "clicked", G_CALLBACK(font_edit_callback), pi);
}
UPDATE_F(font,UNUSED,UNUSED) { }
CALLB_F(font) { text_edit_create( o2go(pi->o) ); /*FIXME: co kdyz to neni go_text?*/  }

CREATE_F(filename)
{
	pi->value_edit = gtk_button_new_with_label("");
	gtk_button_set_relief(GTK_BUTTON(pi->value_edit), GTK_RELIEF_NONE);
	g_signal_connect(G_OBJECT(pi->value_edit), "clicked", G_CALLBACK(filename_edit_callback), pi);
}
UPDATE_F(filename,,)
{ gtk_button_set_label( GTK_BUTTON(pi->value_edit), prop_get_value(pi->o, prop).s->text ); }

CALLB_F(filename)
{
	FIND_PROP;
	//TODO: set the initial filaname
	SAVE_DLG_START( "Choose the filename ...", NULL, prop_get_value(pi->o, prop).s->text )
	{
		prop_value new_val;
		new_val.s = string_lookup_ref(filename);
		TRANS_PROP_CHANGE(new_val);
		string_unref(new_val.s);
	}
	SAVE_DLG_END;
}

CREATE_F(largetext)
{
	pi->value_edit = gtk_button_new_with_label("Edit text ...");
	gtk_button_set_relief (GTK_BUTTON(pi->value_edit), GTK_RELIEF_NONE);
	g_signal_connect (G_OBJECT (pi->value_edit), "clicked", G_CALLBACK(largetext_edit_callback), pi);
}
UPDATE_F(largetext,UNUSED,UNUSED) {}
CALLB_F(largetext) { text_edit_create( o2go(pi->o) ); /*FIXME: co kdyz to neni go_text ani go_tex_text?*/  }

CREATE_F(ptr)
{
	pi->value_edit = gtk_label_new("");
	gtk_label_set_justify( GTK_LABEL(pi->value_edit), GTK_JUSTIFY_LEFT );
}
UPDATE_F(ptr,,)
{
	char pstr[100];
	sprintf(pstr, "%p", prop_get_value(pi->o, prop).v);
	gtk_label_set_text( GTK_LABEL(pi->value_edit), pstr );
}
