#ifndef __VRR_PROPERTIES_H__
#define __VRR_PROPERTIES_H__

#include "gui/main.h"

/**
 \file gui/properties.h
 The property window showing properties of a certain object, property editors etc.
*/

struct prop_window
{
	WIN_O_MAGIC;
	GtkWidget * window;	/**< the GTK window -- should be in WIN_O_MAGIC */
	struct o * o;	/**< the object whose properties are displayed in the window */
	struct obj_tlo * tlo; /**< the parent tlo, if the object displayed is a go */
	struct go_hook go_hook;
	struct obj_hook obj_hook;
	int redraw; /** should gtk widgets be redrawn? */
	GtkTable *table; /**< table for properties */
	GtkWidget *button_add; /** button for adding new properties */
	GtkWidget *button_delete; /** button for deleting an existing property */
	struct status_data status; /**< status bar and context_id */
	uns prop_count;		/** the count of displayed properties */
	struct prop_item * items; /**< The array of prop_item structures for all displayed properties */
	uns flags;
};

extern GSList * prop_window_list;

extern struct prop_window * context_property_window;

struct prop_item
{
	struct o * o;
	string key;
	GtkWidget * value_edit;
	GtkWidget * unit_edit;
};

/** Do some initialization: initialize the unit lists etc. */
void gui_prop_init(void);

/** Free the global unit lists. */
void gui_prop_finish(void);

/** Create and open the property window.
 \param o A pointer to the object whose properties are to be displayed; or NULL,
 if the craeted window should show properties of context objects. */
void prop_window_create(struct o * o);

/** A callback to inform the property window that the context has changed.
 * \param o The new context o. */
void prop_window_context_changed(struct o * o);

/** Create a widget to edit the property.
 \param pi THe structure filled with data needed to create the widget.
 Sets the pointer to thw widget editor. */
void prop_value_edit_create( struct prop_item * pi, struct prop * prop );

/** Create a widget to edit the property unit.
 \param pi THe structure filled with data needed to create the widget.
 Sets the pointer to thw widget editor. */
void prop_unit_edit_create( struct prop_item * pi, struct prop * prop );

/** Initialize the property item -- set the given values, create editor
 widgets and synchronize them with the current values. */
void prop_item_init( struct prop_item * pi, struct o * o, string key );

/** Sychronize the property widgets in the prop_item struct with
 the current value of property prop. If the struct pi contains nonzero
 pointer to a unit editor, it will be synchronized, too. */
void prop_sync( struct prop_item * pi, struct prop * prop );

#define PROP_STORE_DEFINE(_id) struct obj_tlo * _id
#define PROP_STORE_NEW(_id) _id = tlo_new(); obj_ref(&_id->obj)
#define PROP_STORE_DESTROY(_id) obj_unref(&_id->obj); _id = NULL
#define PROP_STORE_O(_id) go2o(&tlo_get_tlg(_id)->go)
#define PROP_STORE_TLO(_id) _id

#define PROP_STORE_GET(_id, _name, _type, _t, _default) \
prop_named_lookup( PROP_STORE_O(_id), _name, _type, (prop_value){ ._t = _default })._t

#define PROP_STORE_SET(_id, _name, _type, _t, _val) \
{ prop_value v = { ._t = _val }; prop_store_set( _id, _name, _type, v); }

#define PROP_STORE_GET_UNS(_id, _name, _default) PROP_STORE_GET(_id, _name, PT_UNS, u, _default)
#define PROP_STORE_SET_UNS(_id, _name, _val) PROP_STORE_SET(_id, _name, PT_UNS, u, _val)

void prop_store_set( PROP_STORE_DEFINE(ps), const char * name, uns type, prop_value pv );

extern PROP_STORE_DEFINE(ps_global);

/** Go through the saved properties and if there are some that
 match, change their values in o accordingly.
 Must be called in a transaction.  */
void gui_prop_recycle( struct o * o );

#define GUI_STRINGS \
	string_tolerance, string_editor, string_editor_in_term, string_browser, string_browser_in_term, \
	string_terminal, string_grid_x1, string_grid_x2, string_grid_y1, string_grid_y2, \
	string_rulerx_res, string_rulery_res, string_drag_image

extern string GUI_STRINGS;
#endif
