#include "lib/lib.h"
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <values.h>

#include <gtk/gtk.h>

#include "geomlib/types.h"
#include "kernel/kernel.h"
#include "gui/propsettings.h"
#include "gui/selectcolor.h"

#define GTK_COLOR_MULTIPLIER	256	// sizeof(GdkColor.red) / sizeof(vrrcolor.red)
#define PROP_UNS_DEFAULT ((prop_value){.u = 0})
#define PROP_REAL_DEFAULT ((prop_value){.r = 0})
#define PROP_PTR_DEFAULT ((prop_value){.v = NULL})
#define PROP_TEXT_DEFAULT ((prop_value){.s = string_noname})		// This isn't the best string.

#define NEW_UNIT	"New..."

struct prop_settings *property_window = NULL;
GList *property_list = NULL;
GList *prop_window_list;

static const uns go_elarc_type_max[GOST_ELARC_MAX] = {
};

static void add_property_to_table(struct prop *node, struct prop_settings *ps);
static void change_string_property(GtkWidget *entry, struct prop_item *pi);
static void set_fg_color(GtkWidget *widget, GdkColor *color);
static void reset_fg_color(GtkWidget *widget);
static void fill_property_window(struct prop_settings *ps, struct prop *p);
static gboolean expose_event_callback (GtkWidget *widget, GdkEventExpose *event UNUSED, gpointer data UNUSED);

static void property_list_init(void)
{
	property_list = g_list_append(property_list, "a) Integer");
	property_list = g_list_append(property_list, "b) Color");
	property_list = g_list_append(property_list, "c) Font");
	property_list = g_list_append(property_list, "d) String");
	property_list = g_list_append(property_list, "e) Long text");
	property_list = g_list_append(property_list, "f) Filename");
	property_list = g_list_append(property_list, "g) Real - unspec");
	property_list = g_list_append(property_list, "h) Real - coord");
	property_list = g_list_append(property_list, "i) Real - length");
	property_list = g_list_append(property_list, "j) Real - angle");
	property_list = g_list_append(property_list, "k) Real - parameter");
	property_list = g_list_append(property_list, "l) Pointer");
}

static void clean_property_window(struct prop_settings *ps)
{
	struct prop* node;
	struct prop_item *pi;

	while((node = slist_head(&ps->valuelist)))
	{
		pi = (struct prop_item *) node;
		string_unref(pi->p.key);
		if(pi->glist)
		{
//			g_list_free(pi->glist);
			pi->glist = NULL;
		}
		slist_remove(&ps->valuelist, &node->n);
		xfree(pi);
	}
	gtk_widget_destroy(ps->table);
	ps->table = gtk_table_new (0, 2, FALSE);
	gtk_widget_show (ps->table);
	gtk_container_add (GTK_CONTAINER (ps->viewport), ps->table);
	gtk_container_set_border_width(GTK_CONTAINER (ps->table), 6);

	gtk_widget_set_sensitive(ps->button_add, FALSE);
	gtk_widget_set_sensitive(ps->button_delete, FALSE);
}

static void refresh_property_window(struct prop_settings *ps, struct prop *p)
{
	clean_property_window(ps);
	fill_property_window(ps, p);
}

void destroy (GtkWidget *widget UNUSED, struct prop_settings *ps)
{
	clean_property_window(ps);
	gtk_widget_destroy(ps->window);
	xfree(ps->title);

	obj_hook_remove( &ps->oh );
	if (ps->o)
	{
		if (ps->o->kind == T_GO) go_hook_remove( go_get_tlo(o2go(ps->o)), &ps->gh );
		o_unref(ps->o);
	}
	if (ps == property_window)
	{
		property_window = NULL;
		refresh_gui();
	}
	if(ps->ap)
		gtk_widget_destroy(ps->ap->window);
	if(ps->dp)
		gtk_widget_destroy(ps->dp->window);
	prop_window_list = g_list_remove(prop_window_list, ps);
	xfree(ps);
}

static void button_close_ps_pressed(GtkWidget *widget UNUSED, struct prop_settings *ps)
{
	if(ps == property_window)
	{
		property_window = NULL;
		refresh_gui();
	}
	gtk_widget_destroy(GTK_WIDGET(ps->window));
}

static void change_property_color_value(struct prop_item *pi, int value)
{
	struct prop_settings *ps = pi->ps;
	struct prop *p;
	char buf[TRANS_ERR_SIZE];

	p = prop_find(ps->o, pi->p.key);
	if(!p)
	{
		report_error_f("Property which should be changed doesn't exist.\n");
		return;
	}

	factory_op_break();
	TRANS_BEGIN_MAIN(ps->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(ps->o)), "Changing the properties", buf)
		ASSERT(p->type == PT_UNS);
		prop_set_uns(ps->o, pi->p.key, (uintptr_t) value);
	TRANS_FAILED
		statusbar_msg(&ps->status, buf);
		report_error_f("Couldn't change property of the object: %s\n", buf);
	TRANS_END
}

static void on_color_clicked (GtkWidget *widget UNUSED, struct prop_item *pi)
{
	int result;
	GtkWidget *area;
	int alpha, newcolor;
	unsigned int u;
	uns r, g, b, a;
	GdkColor color;

	GtkWidget *color_selector = wnd_select_color_new();
	GtkWidget *dialog = gtk_dialog_new_with_buttons("Color selection",
			GTK_WINDOW(pi->ps->window),
			GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
			GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
			GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);

	gtk_container_add (GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), color_selector);
	gtk_widget_show_all(dialog);
	gtk_window_set_default_size (GTK_WINDOW (dialog), 400, 300);
	gtk_window_resize(GTK_WINDOW(dialog), 400, 300);
	u = prop_get_value (pi->ps->o, (struct prop *) (pi->node)).u;
	uns2rgba_uns(u, &r, &g, &b, &a);
	wnd_select_color_set_rgb((WndSelectColor *) color_selector, r, g, b);
	wnd_select_color_set_alpha((WndSelectColor *) color_selector, a);
	result = gtk_dialog_run(GTK_DIALOG(dialog));

	switch(result)
	{
		case GTK_RESPONSE_NONE:
		case GTK_RESPONSE_REJECT:
			break;
		case GTK_RESPONSE_ACCEPT:
			color.pixel = 0;
			wnd_select_color_get_rgb((WndSelectColor *)color_selector,
					(int *)&color.red, (int *)&color.green,
					(int *)&color.blue);
			wnd_select_color_get_alpha((WndSelectColor *)color_selector,
					&alpha);
			area = (GtkWidget *) gtk_object_get_data((GtkObject *)widget, "area");
			newcolor = rgba_uns2uns(color.red, color.green, color.blue, alpha);
			color.red *= GTK_COLOR_MULTIPLIER;
			color.green *= GTK_COLOR_MULTIPLIER;
			color.blue *= GTK_COLOR_MULTIPLIER;
			reset_fg_color(area);
			set_fg_color(area, &color);
			expose_event_callback(area, 0, 0);
			pi->ps->redraw--;
			change_property_color_value(pi, newcolor);
			pi->ps->redraw++;
			break;
		default:
			report_error_f("Bad color selection dialog response.\n");
	}
	gtk_widget_destroy(dialog);
}

static void open_text_editor (GtkWidget *widget UNUSED, struct prop_item *pi)
{
	struct prop_settings *ps = pi->ps;

	u8 type = ((struct o*) ps->o)->type;
	if (type != GOT_TEXT && type != GOT_TEX_TEXT)
		return;
	text_edit_create ((struct go *) ps->o, type == GOT_TEX_TEXT, NULL);
}

static void filename_value_changed(GtkWidget *widget, struct prop_item *pi)
{
	struct prop_settings *ps = pi->ps;
	if(!ps->redraw)
		return;
	else
		ps->redraw--;

	GtkWidget *dialog = gtk_file_chooser_dialog_new("File selection",
			GTK_WINDOW(ps->window),
			GTK_FILE_CHOOSER_ACTION_OPEN,
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
			NULL );

	statusbar_clear(&ps->status);
	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
	{
		const char *filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
		// check new values
		char buf[TRANS_ERR_SIZE];
		struct prop *p;
		int trans_ok = 1;

		if(!(*filename))
		{
			statusbar_msg(&ps->status, "Bad filename!");
			ps->redraw++;
			return;
		}

		p = prop_find(ps->o, pi->p.key);
		if(!p)
		{
			ps->redraw++;
			report_error_f("Property which should be changed doesn't exist.\n");
			return;
		}

		factory_op_break();
		TRANS_BEGIN_MAIN(ps->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(ps->o)), "Changing the properties", buf)
			ASSERT(p->type == PT_STRING);
			prop_set_text(ps->o, pi->p.key, filename);
		TRANS_FAILED
			trans_ok = 0;
			statusbar_msg(&ps->status, buf);
			report_error_f("Couldn't change property of the object: %s\n", buf);
		TRANS_END

		if (trans_ok)
			gtk_button_set_label(GTK_BUTTON(widget), filename);
		ps->redraw++;
	}
	gtk_widget_destroy (dialog);
}

static gboolean entry_focus_lost(GtkWidget *entry, GdkEvent *event UNUSED,
		struct prop_item *pi)
{
	change_string_property(entry, pi);
	return FALSE;
}

static void entry_value_changed(GtkWidget *entry, struct prop_item *pi)
{
	change_string_property(entry, pi);
}

static void change_string_property(GtkWidget *entry, struct prop_item *pi)
{
	// check new values
	char *value;
	char buf[TRANS_ERR_SIZE];
	struct prop_settings *ps = pi->ps;
	struct prop *p;

	if(!ps->redraw)
		return;
	else
		ps->redraw--;

	value = (char *) gtk_entry_get_text(GTK_ENTRY(entry));
	if(!(*value))
	{
		statusbar_msg(&ps->status, "Enter the property value.");
		ps->redraw++;
		return;
	}

	p = prop_find(ps->o, pi->p.key);
	if(!p)
	{
		ps->redraw++;
		report_error_f("Property which should be changed doesn't exist.\n");
		return;
	}

	factory_op_break();
	TRANS_BEGIN_MAIN(ps->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(ps->o)), "Changing the properties", buf)
		ASSERT(p->type == PT_STRING);
		prop_set_text(ps->o, pi->p.key, value);
	TRANS_FAILED
		statusbar_msg(&ps->status, buf);
		report_error_f("Couldn't change property of the object: %s\n", buf);
	TRANS_END

	ps->redraw++;
}

static gboolean expose_event_callback (GtkWidget *widget, GdkEventExpose *event UNUSED, gpointer data UNUSED)
{
	gdk_draw_rectangle(widget->window,
		widget->style->fg_gc[GTK_WIDGET_STATE (widget)], TRUE, 0, 0,
		widget->allocation.width, widget->allocation.height);

	return TRUE;
}

static void new_unit_ok_pressed(GtkWidget *widget UNUSED, struct new_unit *nu)
{
	uns unit;
	string unit_name;
	char *unit_multiplier;
	char *name;
	char rtext[128];
	real mult, rvalue;
	char buf[TRANS_ERR_SIZE];
	struct prop_item *pi = nu->pi;
	int trans_ok = 0;

	name = (char *) gtk_entry_get_text(GTK_ENTRY (nu->entry_name));
	if(!name[0])
	{
		statusbar_msg(&nu->status, "You haven't entered the unit name!");
		return;
	}

	TRANS_BEGIN_MAIN(pi->ps->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(pi->ps->o)), "Looking up the string", buf)
		unit_name = string_lookup(gtk_entry_get_text(GTK_ENTRY(nu->entry_name)));
		unit = unit_find(unit_name);
		if(unit != PROP_UNIT_MAX)
			statusbar_msg(&nu->status, "Unit with this name already exists!");
		else
		{
			unit_multiplier = (char *) gtk_entry_get_text(GTK_ENTRY(nu->entry_multiplier));
#ifdef FLOAT_REAL
			if(!sscanf(unit_multiplier, "%f", &mult))
#else
			if(!sscanf(unit_multiplier, "%lf", &mult))
#endif
				statusbar_msg(&nu->status, "Bad multiplier!");
			else
			{
				unit = unit_set(unit_name, 1.0 / mult, prop_subtype2quantity[pi->p.type][pi->p.subtype]);
				prop_set_unit(pi->ps->o, pi->node, unit);
				trans_ok = 1;
			}
		}
	TRANS_FAILED
		statusbar_msg(&nu->status, buf);
		report_error_f("Couldn't change unit of the object: %s\n", buf);
	TRANS_END

	if(trans_ok)
	{
		pi->glist = g_list_prepend(pi->glist, unit_array[unit].name->text);
		gtk_combo_set_popdown_strings(GTK_COMBO(pi->unitsw), pi->glist);
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(pi->unitsw)->entry), name);
		rvalue = prop_get_value(pi->ps->o, pi->node).r * prop_get_unit_multiplier(pi->ps->o, pi->node);
		if(!isfinite(rvalue))
		{
			rvalue = 0;
			statusbar_msg(&nu->status, "Too big value in property %s, changed to 0.", pi->node->key->text);
		}
		sprintf(rtext, "%g", rvalue);
		gtk_entry_set_text(GTK_ENTRY(pi->valuew), rtext);
		gtk_widget_destroy(nu->window);
	}
}

static void new_unit_destroy (GtkWidget *widget UNUSED, struct new_unit *nu)
{
	int unit;
	struct prop_item *pi = nu->pi;

	unit = prop_get_unit(pi->ps->o, pi->node);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(pi->unitsw)->entry), unit_array[unit].name->text);

	xfree(nu);
}

static void new_unit_cancel_pressed(GtkWidget *widget UNUSED, struct new_unit *nu)
{
	gtk_widget_destroy(nu->window);
}

static void new_unit_clear_status (GtkWidget *widget UNUSED, struct new_unit *nu)
{
	statusbar_clear (&nu->status);
}

static void create_new_unit_window(struct prop_item *pi)
{
	GtkTooltips *tooltips;
	GtkWidget *button_ok, *button_cancel;
	GtkWidget *hbox;
	GtkWidget *label_name, *label_multiplier;
	GtkWidget *table;
	struct new_unit *nu = xmalloc(sizeof(struct new_unit));

	nu->pi = pi;
	tooltips = gtk_tooltips_new();

	nu->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_container_set_border_width (GTK_CONTAINER(nu->window), 6);
	gtk_window_set_title (GTK_WINDOW (nu->window), "Creating new unit...");
	gtk_window_set_modal (GTK_WINDOW (nu->window), TRUE);
	gtk_window_set_destroy_with_parent (GTK_WINDOW (nu->window), TRUE);

	table = gtk_table_new (4, 2, FALSE);
	gtk_widget_show (table);
	gtk_container_add (GTK_CONTAINER (nu->window), table);
	gtk_table_set_row_spacings (GTK_TABLE (table), 10);
	gtk_table_set_col_spacings (GTK_TABLE (table), 5);

	label_name = gtk_label_new ("Name:");
	gtk_widget_show (label_name);
	gtk_table_attach (GTK_TABLE (table), label_name, 0, 1, 0, 1,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_label_set_justify (GTK_LABEL (label_name), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label_name), 0, 0.5);

	nu->entry_name = gtk_entry_new ();
	gtk_widget_show (nu->entry_name);
	gtk_table_attach (GTK_TABLE (table), nu->entry_name, 1, 2, 0, 1,
			(GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, nu->entry_name, "Enter the new unit's name.", NULL);

	label_multiplier = gtk_label_new ("Multiplier:");
	gtk_widget_show (label_multiplier);
	gtk_table_attach (GTK_TABLE (table), label_multiplier, 0, 1, 1, 2,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_label_set_justify (GTK_LABEL (label_multiplier), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label_multiplier), 0, 0.5);

	nu->entry_multiplier = gtk_entry_new ();
	gtk_widget_show (nu->entry_multiplier);
	gtk_table_attach (GTK_TABLE (table), nu->entry_multiplier, 1, 2, 1, 2,
			(GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, nu->entry_multiplier, "Enter the nuw unit's multiplier (real positive number).", NULL);

	hbox = gtk_hbox_new (TRUE, 0);
	gtk_widget_show (hbox);
	gtk_table_attach (GTK_TABLE (table), hbox, 0, 2, 2, 3,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (GTK_FILL), 0, 0);

	button_ok = gtk_button_new_from_stock ("gtk-ok");
	gtk_widget_show (button_ok);
	gtk_box_pack_start (GTK_BOX (hbox), button_ok, FALSE, FALSE, 0);
	gtk_widget_set_size_request (button_ok, 84, -1);
	gtk_tooltips_set_tip (tooltips, button_ok, "Create new unit with entered name and multiplier and close the window.", NULL);

	button_cancel = gtk_button_new_from_stock ("gtk-cancel");
	gtk_widget_show (button_cancel);
	gtk_box_pack_start (GTK_BOX (hbox), button_cancel, FALSE, FALSE, 0);
	gtk_widget_set_size_request (button_cancel, 84, -1);
	gtk_tooltips_set_tip (tooltips, button_cancel, "Close the window without creating a new unit.", NULL);

	nu->status.statusbar = gtk_statusbar_new();
	nu->status.context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(nu->status.statusbar), "New unit");
	gtk_widget_show(nu->status.statusbar);
	gtk_table_attach(GTK_TABLE (table), nu->status.statusbar, 0, 2, 3, 4,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);

	g_signal_connect (G_OBJECT(button_cancel), "clicked",
			G_CALLBACK (new_unit_cancel_pressed), nu);
	g_signal_connect (G_OBJECT(button_ok), "clicked",
			G_CALLBACK (new_unit_ok_pressed), nu);
	g_signal_connect (G_OBJECT(nu->window), "destroy",
			G_CALLBACK (new_unit_destroy), nu);
	g_signal_connect (G_OBJECT(nu->entry_multiplier), "changed",
			G_CALLBACK (new_unit_clear_status), nu);
	g_signal_connect (G_OBJECT(nu->entry_name), "changed",
			G_CALLBACK (new_unit_clear_status), nu);

	gtk_widget_show(nu->window);
}

static void unit_changed(GtkWidget *entry, struct prop_item *pi)
{
	char *value;
	struct prop_settings *ps = pi->ps;

	if(!ps->redraw)
		return;
	else
		ps->redraw--;

	statusbar_clear(&ps->status);
	value = (char *) gtk_entry_get_text(GTK_ENTRY (entry));
	if(!strcmp(value, ""))
	{
		ps->redraw++;
		return;
	}
	if (!strcmp(value, NEW_UNIT))
		create_new_unit_window(pi);
	else
	{
		int unit;
		real rvalue;
		char rtext[32];

		unit = unit_named_find(value);
		if(unit == PROP_UNIT_MAX)
		{
			ps->redraw++;
			return;
		}

		prop_set_unit(pi->ps->o, pi->node, unit);

		rvalue = prop_get_value(pi->ps->o, pi->node).r * prop_get_unit_multiplier(pi->ps->o, pi->node);

		sprintf(rtext, "%g", (double) rvalue);
		gtk_entry_set_text(GTK_ENTRY(pi->valuew), rtext);
	}
	ps->redraw++;
}

static void conic_combo_changed (GtkEntry *entry, struct prop_item *pi)
{
	char *value;
	char buf[TRANS_ERR_SIZE];
	struct prop_settings *ps = pi->ps;
	struct prop *p;
	unsigned int i;

	if (!ps->redraw)
		return;
	else
		ps->redraw--;

	statusbar_clear (&ps->status);


	value = (char *) gtk_entry_get_text (entry);
	for (i = 0; i < go_elarc_type_max[ps->o->subtype]; i++)
		if (!strcmp (value, string_conic_type[i]->text))
			break;
	if (i == go_elarc_type_max[ps->o->subtype])
	{
		statusbar_msg (&ps->status, "Couldn't change the subtype.");
		ps->redraw++;
		return;
	}

	p = prop_find (ps->o, pi->p.key);
	if (!p)
	{
		report_error_f("Property which should be changed doesn't exist.\n");
		ps->redraw++;
	}
	factory_op_break ();
	TRANS_BEGIN_MAIN (ps->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(ps->o)), "Changing the subtype property", buf)
		ASSERT (p->type == PT_UNS);
		prop_set_uns(ps->o, pi->p.key, (uintptr_t) i);
	TRANS_FAILED
		statusbar_msg (&ps->status, buf);
		report_error_f("Couldnt change property of the object: %s\n", buf);
	TRANS_END
		ps->redraw++;
}

static void spin_button_uint_value_changed(GtkWidget *spin, struct prop_item *pi)
{
	int value;
	char buf[TRANS_ERR_SIZE];
	struct prop_settings *ps = pi->ps;
	struct prop *p;

	if(!ps->redraw)
		return;
	else
		ps->redraw--;

	statusbar_clear(&ps->status);
	value = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spin));

	p = prop_find(ps->o, pi->p.key);
	if(!p)
	{
		report_error_f("Property which should be changed doesn't exist.\n");
		ps->redraw++;
	}

	factory_op_break();
	TRANS_BEGIN_MAIN(ps->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(ps->o)), "Changing the properties", buf)
		ASSERT(p->type == PT_UNS);
		prop_set_uns(ps->o, pi->p.key, (uintptr_t) value);
	TRANS_FAILED
		statusbar_msg(&ps->status, buf);
		report_error_f("Couldn't change property of the object: %s\n", buf);
	TRANS_END
		ps->redraw++;
}

static void spin_button_real_value_changed(GtkWidget *spin, struct prop_item *pi)
{
	real value;
	char buf[TRANS_ERR_SIZE];
	struct prop_settings *ps = pi->ps;
	struct prop *p;

	if (!ps->redraw)
		return;
	ps->redraw--;
	statusbar_clear(&ps->status);
	value = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(spin));

	p = prop_find(ps->o, pi->p.key);
	if(!p)
	{
		ps->redraw++;
		report_error_f("Property which should be changed doesn't exist.\n");
		return;
	}

	factory_op_break();
	TRANS_BEGIN_MAIN(ps->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(ps->o)), "Changing the properties", buf)
		ASSERT(p->type == PT_REAL);
		prop_set_real_unit(ps->o, pi->p.key, value);
	TRANS_FAILED
		statusbar_msg(&ps->status, buf);
		report_error_f("Couldn't change property of the object: %s\n", buf);
	TRANS_END

	ps->redraw++;
}

static void delete_property_delete_pressed(GtkWidget *widget UNUSED, struct delete_property *dp)
{
	char *name;
	char buf[TRANS_ERR_SIZE];
	int trans_ok = 1;

	name = (char *) gtk_entry_get_text(GTK_ENTRY(dp->entry));
	if (!(*name))
	{
		statusbar_msg(&dp->status, "Bad property name. Couldn't delete property.");
		return;
	}

	TRANS_BEGIN_MAIN(dp->ps->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(dp->ps->o)), "Deleting property", buf)
		prop_find_delete(dp->ps->o, string_lookup(name));
	TRANS_FAILED
		trans_ok = 0;
		statusbar_msg(&dp->status, buf);
		report_error(buf);
	TRANS_END

	if (trans_ok)
		gtk_widget_destroy(dp->window);
}

static void add_property_ok_pressed (GtkWidget *widget UNUSED, struct add_property *ap)
{
	// check new values
	char *name;
	char type[33];
	string str;
	int trans_ok = 1;
	char buf[TRANS_ERR_SIZE];
	struct o * o = ap->ps->o;

	name = (char *) gtk_entry_get_text(GTK_ENTRY(ap->entry_name));
	if(!(*name))
	{
		statusbar_msg(&ap->status, "Enter the property name.");
		return;
	}

	str = string_lookup(name);
	if (prop_find (ap->ps->o, str))
	{
		statusbar_msg(&ap->status, "Property with this name already exists.");
		return;
	}

	factory_op_break();
	TRANS_BEGIN_MAIN(ap->ps->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(ap->ps->o)), "Changing the properties", buf)
		strncpy(type, gtk_entry_get_text(GTK_ENTRY(ap->entry_type)), 32); 	// Why?
		switch(type[0])		// This should be better: switch(gtk_entry_get_text(GTK_ENTRY(ap->entry_type)[0])
		{
			case 'a':	// Integer
				prop_create(o, str, PT_UNS, PTU_UNSPECIFIED, PROP_UNIT_MAX, PROP_UNS_DEFAULT);
				break;
			case 'b':	// Color
				prop_create(o, str, PT_UNS, PTU_COLOR, PROP_UNIT_MAX, PROP_UNS_DEFAULT);
				break;
			case 'c':	// Font
				prop_create(o, str, PT_UNS, PTU_FONT, PROP_UNIT_MAX, PROP_UNS_DEFAULT);
				break;
			case 'd':	// String
				prop_create(o, str, PT_STRING, PTS_UNSPECIFIED, PROP_UNIT_MAX, PROP_TEXT_DEFAULT);
				break;
			case 'e':	// Long text
				prop_create(o, str, PT_STRING, PTS_LARGE_TEXT, PROP_UNIT_MAX, PROP_TEXT_DEFAULT);
				break;
			case 'f': 	// Filename
				prop_create(o, str, PT_STRING, PTS_FILE_NAME, PROP_UNIT_MAX, PROP_TEXT_DEFAULT);
				break;
			case 'g':
				prop_create(o, str, PT_REAL, PTR_UNSPECIFIED, PROP_UNIT_MAX, PROP_REAL_DEFAULT);
				break;
			case 'h':
				prop_create(o, str, PT_REAL, PTR_COORDINATE, PROP_UNIT_MAX, PROP_REAL_DEFAULT);
				break;
			case 'i':
				prop_create(o, str, PT_REAL, PTR_NON_NEGATIVE, PROP_UNIT_MAX, PROP_REAL_DEFAULT);
				break;
			case 'j':
				prop_create(o, str, PT_REAL, PTR_ANGLE_4PI, PROP_UNIT_MAX, PROP_REAL_DEFAULT);
				break;
			case 'k':
				prop_create(o, str, PT_REAL, PTR_REFERENCE, PROP_UNIT_MAX, PROP_REAL_DEFAULT);
				break;
			case 'l':	// Pointer, will not be allowed (it's just for debugging
						// => there are no controls whether this value is correct)
				prop_create(o, str, PT_POINTER, PTS_UNSPECIFIED, PROP_UNIT_MAX, PROP_PTR_DEFAULT);
				break;
			default:
				trans_fail("Bad type name while creating new property!!!");
		}
	TRANS_FAILED
		trans_ok = 0;
		statusbar_msg(&ap->status, buf);
		report_error(buf);
	TRANS_END

	if (trans_ok)
		gtk_widget_destroy(ap->window);
}

static void delete_property_destroy (GtkWidget *widget UNUSED, struct delete_property *dp)
{
	dp->ps->dp = NULL;
//	if(dp->glist)
//		g_list_free(dp->glist);
	xfree(dp);
}

static void delete_property_cancel_pressed (GtkWidget *widget UNUSED, struct delete_property *dp)
{
	gtk_widget_destroy(dp->window);
}

static void add_property_destroy (GtkWidget *widget UNUSED, struct add_property *ap)
{
	ap->ps->ap = NULL;
	xfree(ap);
}

static void add_property_cancel_pressed (GtkWidget *widget UNUSED, struct add_property *ap)
{
	gtk_widget_destroy(ap->window);
}

static void create_delete_property_window(GtkWidget *widget UNUSED,
		struct prop_settings *ps)
{
	struct prop *node;
	GtkWidget *button_delete, *button_cancel;
	GtkWidget *combo, *label;
	GtkWidget *table;
	GtkWidget *hbox;
	const char * sname = o_lookup_name (ps->o);
	char title[strlen(sname) + 32];
	GtkTooltips *tooltips;
	struct delete_property *dp = xmalloc(sizeof(struct delete_property));

	if(ps->dp)
	{
		gtk_window_present(GTK_WINDOW(ps->dp->window));
		return;
	}
	dp->glist = NULL;
	WALK_PROP(ps->o, node)
	{
		if (!((PTF_VIRTUAL | PTF_READ_ONLY) & node->flags))
			dp->glist = g_list_append(dp->glist, node->key->text);
	}
	if (!dp->glist)
	{
		statusbar_msg(&ps->status, "No properties can be deleted.");
		return;
	}

	dp->ps = ps;
	ps->dp = dp;

	tooltips = gtk_tooltips_new();

	dp->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	sprintf(title, "Delete property (%s)", sname);
	gtk_window_set_title(GTK_WINDOW(dp->window), title);

	table = gtk_table_new (3, 2, FALSE);
	gtk_widget_show(table);
	gtk_container_add (GTK_CONTAINER (dp->window), table);
	gtk_table_set_row_spacings(GTK_TABLE(table), 5);
	gtk_table_set_col_spacings(GTK_TABLE(table), 5);

	label = gtk_label_new("Name");
	gtk_widget_show(label);
	gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_tooltips_set_tip(tooltips, label, "Choose the name of property you want to delete.", NULL);
	gtk_table_attach(GTK_TABLE (table), label, 0, 1, 0, 1,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 10);

	combo = gtk_combo_new();
	gtk_combo_set_popdown_strings(GTK_COMBO(combo), dp->glist);
	gtk_combo_disable_activate(GTK_COMBO(combo));
	gtk_entry_set_editable(GTK_ENTRY((GTK_COMBO(combo))->entry), FALSE);
	gtk_widget_show(combo);
	gtk_table_attach(GTK_TABLE (table), combo, 1, 2, 0, 1,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 10);

	dp->entry = GTK_COMBO (combo)->entry;
	gtk_widget_show(dp->entry);
	gtk_tooltips_set_tip(tooltips, dp->entry, "Choose the name of property you want to delete.", NULL);

	hbox = gtk_hbox_new (FALSE, 10);
	gtk_widget_show (hbox);
	gtk_table_attach(GTK_TABLE (table), hbox, 0, 2, 1, 2,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 3, 6);

	button_cancel = gtk_button_new_from_stock("gtk-cancel");
	gtk_widget_show (button_cancel);
	gtk_box_pack_start (GTK_BOX(hbox), button_cancel, TRUE, TRUE, 20);
	gtk_tooltips_set_tip(tooltips, button_cancel, "Close the window without deleting a property.", NULL);

	button_delete = gtk_button_new_from_stock ("gtk-delete");
	gtk_widget_show (button_delete);
	gtk_box_pack_start (GTK_BOX (hbox), button_delete, TRUE, TRUE, 20);
	gtk_widget_set_size_request (button_delete, 75, 0);
	gtk_tooltips_set_tip(tooltips, button_delete, "Delete a chosen property and close the window.", NULL);

	dp->status.statusbar = gtk_statusbar_new();
	dp->status.context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(dp->status.statusbar), "Delete property");
	gtk_widget_show(dp->status.statusbar);
	gtk_table_attach(GTK_TABLE (table), dp->status.statusbar, 0, 2, 2, 3,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);

	g_signal_connect (G_OBJECT(button_cancel), "clicked",
			G_CALLBACK(delete_property_cancel_pressed), dp);
	g_signal_connect (G_OBJECT(button_delete), "clicked",
			G_CALLBACK(delete_property_delete_pressed), dp);
	g_signal_connect (G_OBJECT(dp->window), "destroy",
			G_CALLBACK(delete_property_destroy), dp);

	gtk_widget_show(dp->window);
}

static void create_add_new_property_window(GtkWidget *widget UNUSED,
		struct prop_settings *ps)
{
	GtkWidget *table;
	GtkWidget *label_type, *label_name;
	GtkWidget *button_ok, *button_cancel;
	GtkWidget *combo;
	const char * sname = o_lookup_name (ps->o);
	char title[strlen(sname) + 32];
	GtkTooltips *tooltips;
	GList *glist;
	struct add_property *ap = xmalloc(sizeof(struct add_property));
	GtkWidget *hbox;

	if(ps->ap)
	{
		gtk_window_present(GTK_WINDOW(ps->ap->window));
		return;
	}
	if (!property_list)
		property_list_init();
	glist = property_list;

	// related prop_settings
	ap->ps = ps;
	ps->ap = ap;

	tooltips = gtk_tooltips_new();

	ap->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	sprintf(title, "Add new property (%s)", sname);
	gtk_window_set_title(GTK_WINDOW(ap->window), title);

	table = gtk_table_new (4, 2, FALSE);
	gtk_widget_show(table);
	gtk_container_add (GTK_CONTAINER (ap->window), table);
	gtk_table_set_row_spacings(GTK_TABLE(table), 8);
	gtk_table_set_col_spacings(GTK_TABLE(table), 10);

	label_type = gtk_label_new("Type");
	gtk_widget_show(label_type);
	gtk_table_attach(GTK_TABLE(table), label_type, 0, 1, 0, 1,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 10);
	gtk_label_set_justify (GTK_LABEL (label_type), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label_type), 0, 0.5);

	gtk_tooltips_set_tip(tooltips, label_type, "Choose the property type.", NULL);

	label_name = gtk_label_new("Name");
	gtk_widget_show(label_name);
	gtk_table_attach(GTK_TABLE(table), label_name, 0, 1, 1, 2,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);
	gtk_label_set_justify (GTK_LABEL(label_name), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label_name), 0, 0.5);
	gtk_tooltips_set_tip(tooltips, label_name, "Enter the property name.", NULL);

	combo = gtk_combo_new();
	gtk_combo_set_popdown_strings(GTK_COMBO(combo), glist);
	gtk_combo_disable_activate(GTK_COMBO(combo));
	gtk_entry_set_editable(GTK_ENTRY((GTK_COMBO(combo))->entry), FALSE);
	gtk_widget_show(combo);
	gtk_table_attach(GTK_TABLE(table), combo, 1, 2, 0, 1,
			(GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
			(GtkAttachOptions) (0), 0, 10);

	ap->entry_type = GTK_COMBO (combo)->entry;
	gtk_widget_show(ap->entry_type);
	gtk_tooltips_set_tip(tooltips, ap->entry_type, "Choose the property type.", NULL);

	ap->entry_name = gtk_entry_new();
	gtk_table_attach (GTK_TABLE (table), ap->entry_name, 1, 2, 1, 2,
			(GtkAttachOptions) (GTK_FILL | GTK_EXPAND),
			(GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip(tooltips, ap->entry_name, "Enter the property name.", NULL);

	hbox = gtk_hbox_new (FALSE, 10);
	gtk_widget_show (hbox);
	gtk_table_attach(GTK_TABLE (table), hbox, 0, 2, 2, 3,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 3, 6);

	button_cancel = gtk_button_new_from_stock("gtk-cancel");
	gtk_widget_show (button_cancel);
	gtk_box_pack_start (GTK_BOX(hbox), button_cancel, TRUE, TRUE, 20);
	gtk_tooltips_set_tip(tooltips, button_cancel, "Close the window without adding a new property.", NULL);

	button_ok = gtk_button_new_from_stock ("gtk-ok");
	gtk_widget_show (button_ok);
	gtk_box_pack_start (GTK_BOX (hbox), button_ok, TRUE, TRUE, 20);
	gtk_widget_set_size_request (button_ok, 75, 0);
	gtk_tooltips_set_tip(tooltips, button_ok, "Add new property to the object and close this window.", NULL);

	ap->status.statusbar = gtk_statusbar_new();
	ap->status.context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(ap->status.statusbar), "Add property");
	gtk_table_attach(GTK_TABLE (table), ap->status.statusbar, 0, 2, 3, 4,
			(GtkAttachOptions) (GTK_FILL),
			(GtkAttachOptions) (0), 0, 0);

	g_signal_connect (G_OBJECT(button_cancel), "clicked",
			G_CALLBACK(add_property_cancel_pressed), ap);
	g_signal_connect (G_OBJECT(button_ok), "clicked",
			G_CALLBACK(add_property_ok_pressed), ap);
	g_signal_connect (G_OBJECT(ap->window), "destroy",
			G_CALLBACK(add_property_destroy), ap);

	gtk_widget_show_all(ap->window);
}

static void reset_fg_color(GtkWidget *widget)
{
	gtk_widget_modify_fg(widget, GTK_STATE_NORMAL, NULL);
	gtk_widget_modify_fg(widget, GTK_STATE_ACTIVE, NULL);
	gtk_widget_modify_fg(widget, GTK_STATE_PRELIGHT, NULL);
	gtk_widget_modify_fg(widget, GTK_STATE_SELECTED, NULL);
}

static void set_fg_color(GtkWidget *widget, GdkColor *color)
{
	gtk_widget_modify_fg(widget, GTK_STATE_NORMAL, color);
	gtk_widget_modify_fg(widget, GTK_STATE_ACTIVE, color);
	gtk_widget_modify_fg(widget, GTK_STATE_PRELIGHT, color);
	gtk_widget_modify_fg(widget, GTK_STATE_SELECTED, color);
}

static void delete_property (GtkWidget *widget, struct prop_settings *ps)
{
	create_delete_property_window(widget, ps);
}

static void add_new_property (GtkWidget *widget, struct prop_settings *ps)
{
	statusbar_clear(&ps->status);
	create_add_new_property_window(widget, ps);
}

static void obj_unlink_notice(struct o* o, uns rank UNUSED, struct obj_hook *oh)
{
	struct prop_settings *ps = (struct prop_settings *)(oh->data);

	if (!o_is_ancestor(o, ps->o)) return;

	if(ps == property_window)
	{
		redraw_property_window(0);
	}
	else
		gtk_widget_destroy(GTK_WIDGET(ps->window));
}

static void obj_prop_created(struct o *o, struct prop *objprop, struct obj_hook *hook)
{
	struct prop_settings *ps = (struct prop_settings *)(hook->data);

	if (o == ps->o) add_property_to_table(objprop, ps);
}

static void prop_changed(struct o* o UNUSED, struct prop_settings *ps)
{
	refresh_property_window(ps, 0);
}

static void obj_prop_changed(struct o *o, struct obj_hook *hook)
{
	struct prop_settings *ps = (struct prop_settings *)(hook->data);
	if (o == ps->o) prop_changed(o, ps);
}

static void obj_prop_deleted(struct o *o, struct prop *objprop, struct obj_hook *hook)
{
	struct prop_settings *ps = (struct prop_settings *)(hook->data);
	if (o == ps->o) refresh_property_window(ps, objprop);
}

static void go_prop_created(struct o *gro, struct prop *goprop, struct go_hook *hook)
{
	struct prop_settings *ps = (struct prop_settings *)(hook->data);

	if (ps->o != gro)
		return;

	add_property_to_table(goprop, ps);
}

static void go_prop_changed(struct o *gro, struct go_hook *hook)
{
	prop_changed(gro, (struct prop_settings *)hook->data);
}

static void go_prop_deleted(struct o *gro, struct prop *goprop, struct go_hook *hook)
{
	struct prop_settings *ps = (struct prop_settings *)(hook->data);

	if(ps->o != gro)
		return;

	refresh_property_window(ps, goprop);
}

static void go_unlink_notice(struct o* o, uns rank UNUSED, struct go_hook *gh)
{
	struct prop_settings *ps = (struct prop_settings *)(gh->data);

	if (ps->o != o)
		return;

	if(ps == property_window)
	{
		redraw_property_window(0);
	}
	else
		gtk_widget_destroy(GTK_WIDGET(ps->window));
}

static void add_property_to_table(struct prop *node, struct prop_settings *ps)
{
	GtkWidget *label;
	struct prop_item *pi = (struct prop_item *)xmalloc(sizeof(struct prop_item));
	GtkWidget *propbox;
	int unit;
	char pstr[128];
	unsigned quantity;
	real rvalue;

	ASSERT( node->key );

	pi->ps = ps;
	propbox = gtk_hbox_new (FALSE, 0);
	gtk_widget_show (propbox);
	pi->index = ps->cnt++;
	pi->glist = NULL;
	gtk_table_resize(GTK_TABLE(ps->table), ps->cnt, 2);

	// create a label and appropriate component for value
	label = gtk_label_new(node->key->text);
	gtk_widget_show(label);
	gtk_table_attach (GTK_TABLE (ps->table), label, 0, 1, pi->index,
			pi->index + 1, (GtkAttachOptions) (GTK_FILL | GTK_EXPAND),
			(GtkAttachOptions) (0), 0, 0);
	gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);

	switch(node->type)
	{
		case PT_UNS:	// spin, 0 -> UINT_MAX
			// number of digits is limited to 6 (now.. :))
			switch(node->subtype)
			{
				case PTU_UNSPECIFIED:
					pi->valuew = gtk_spin_button_new(
							GTK_ADJUSTMENT(gtk_adjustment_new(prop_get_value(ps->o, node).u, 0,
							UINT_MAX, 1, 10, 10)), 1, 6);
					gtk_spin_button_set_digits(GTK_SPIN_BUTTON (pi->valuew), 0);
					g_signal_connect(G_OBJECT(pi->valuew), "value-changed",
							G_CALLBACK(spin_button_uint_value_changed), pi);

					break;
				case PTU_COLOR:
				{
					GdkColor color;
					uns r, g, b, alpha;
					unsigned u;
					GtkWidget *area;

					color.pixel = 0;
					u = prop_get_value (ps->o, node).u;
					uns2rgba_uns(u, &r, &g, &b, &alpha);
					color.red = r * GTK_COLOR_MULTIPLIER;
					color.green = g * GTK_COLOR_MULTIPLIER;
					color.blue = b * GTK_COLOR_MULTIPLIER;

					area = gtk_drawing_area_new();
					pi->valuew = gtk_button_new();
					gtk_object_set_data(GTK_OBJECT (pi->valuew), "area", area);
					gtk_button_set_relief (GTK_BUTTON(pi->valuew), GTK_RELIEF_NONE);
					gtk_widget_set_size_request(pi->valuew, -1, 20);
					gtk_container_add(GTK_CONTAINER(pi->valuew), area);
					set_fg_color (area, &color);

					g_signal_connect (G_OBJECT (area),"expose_event",
							G_CALLBACK (expose_event_callback), NULL);
					g_signal_connect (G_OBJECT (pi->valuew), "clicked", G_CALLBACK(on_color_clicked), pi);
					gtk_widget_show(area);
					break;
				}
				case PTU_CONIC_TYPE:
				{
					GList *list = NULL;
					unsigned i;
					int u;

					if (ps->o->type != GOT_ELARC)
					{
						report_error_f("Bad attribute subtype in object %p (%d)\n", ps->o, node->subtype);
						return;
					}

					u = prop_get_value (ps->o, node).u;
					pi->valuew = gtk_combo_new ();
					gtk_entry_set_editable (GTK_ENTRY ((GTK_COMBO(pi->valuew))->entry), FALSE);
					for (i = 0; i < go_elarc_type_max[ps->o->subtype]; i++)
						list = g_list_append (list, string_conic_type[i]->text);
					gtk_combo_set_popdown_strings (GTK_COMBO (pi->valuew), list);
					gtk_entry_set_text (GTK_ENTRY ((GTK_COMBO (pi->valuew))->entry), string_conic_type[u]->text);
					g_signal_connect (G_OBJECT (GTK_COMBO(pi->valuew)->entry), "changed", G_CALLBACK(conic_combo_changed), pi);
					break;
				}
				case PTU_FONT:
				{
					pi->valuew = gtk_button_new_with_label("Change font");
					gtk_button_set_relief(GTK_BUTTON(pi->valuew), GTK_RELIEF_NONE);
					g_signal_connect(G_OBJECT(pi->valuew), "clicked",
							G_CALLBACK(open_text_editor), pi);
					gtk_widget_show(pi->valuew);
					break;
				}
				default:
					report_error_f("Bad attribute subtype in object %p (%d)\n", ps->o, node->subtype);
					return;
					// bad subtype
			}
			break;
		case PT_STRING:	// edit
			switch(node->subtype)
			{
				case PTS_UNSPECIFIED:
					pi->valuew = gtk_entry_new();
					gtk_entry_set_text(GTK_ENTRY(pi->valuew),
							prop_get_value(ps->o, node).s->text);
					g_signal_connect(G_OBJECT(pi->valuew), "focus-out-event",
							G_CALLBACK(entry_focus_lost), pi);
					g_signal_connect(G_OBJECT(pi->valuew), "activate",
							G_CALLBACK(entry_value_changed), pi);
					break;
				case PTS_LARGE_TEXT:
					pi->valuew = gtk_button_new_with_label("Edit text ...");
					gtk_button_set_relief (GTK_BUTTON(pi->valuew), GTK_RELIEF_NONE);
					g_signal_connect (G_OBJECT (pi->valuew), "clicked",
							G_CALLBACK(open_text_editor), pi);
					break;
				case PTS_FILE_NAME:
					pi->valuew = gtk_button_new_with_label(prop_get_value(ps->o, node).s->text);
					gtk_button_set_relief(GTK_BUTTON(pi->valuew), GTK_RELIEF_NONE);
					g_signal_connect(G_OBJECT(pi->valuew), "clicked",
							G_CALLBACK(filename_value_changed), pi);
					break;
				default:
					// bad subtype
					report_error_f("Bad attribute subtypetype in object %p (%d)\n", ps->o, node->subtype);
					return;
			}
			break;
		case PT_POINTER: // ????? (what should be shown?)
			sprintf(pstr, "%p", prop_get_value(ps->o, node).v);
			pi->valuew = gtk_label_new(pstr);
			gtk_label_set_justify(GTK_LABEL (label), GTK_JUSTIFY_LEFT);
			break;
		case PT_REAL:
			rvalue = prop_get_value(ps->o, node).r * prop_get_unit_multiplier(ps->o, node);
			if(!isfinite(rvalue))
			{
				rvalue = 0;
				statusbar_msg(&ps->status, "Too big value in property %s, changed to 0.", node->key->text);
			}
			switch (node->subtype)
			{
				case PTR_UNSPECIFIED:
				case PTR_COORDINATE:
					pi->valuew = gtk_spin_button_new(
							GTK_ADJUSTMENT(gtk_adjustment_new(rvalue, -MAXFLOAT,
							MAXFLOAT, 0.1, 1, 1)), 0.1, 4);
					break;
				case PTR_NON_NEGATIVE:
					pi->valuew = gtk_spin_button_new(
							GTK_ADJUSTMENT(gtk_adjustment_new(rvalue, 0, MAXFLOAT, 0.1, 1, 1)), 0.1, 4);
					break;
				case PTR_ANGLE_PI:
					pi->valuew = gtk_spin_button_new(
							GTK_ADJUSTMENT(gtk_adjustment_new(rvalue, 0, R_PI, R_PI/12, R_PI/4, R_PI/4)), R_PI/12, 4);
					break;
				case PTR_ANGLE_2PI:
					pi->valuew = gtk_spin_button_new(
							GTK_ADJUSTMENT(gtk_adjustment_new(rvalue, 0, R_2PI, R_PI/12, R_PI/4, R_PI/4)), R_PI/12, 4);
					break;
				case PTR_ANGLE_4PI:
					pi->valuew = gtk_spin_button_new(
							GTK_ADJUSTMENT(gtk_adjustment_new(rvalue, -R_2PI, R_2PI, R_PI/12, R_PI/4, R_PI/4)), R_PI/12, 4);
					break;
				case PTR_REFERENCE:
					pi->valuew = gtk_spin_button_new(
							GTK_ADJUSTMENT(gtk_adjustment_new(rvalue, 0, 1, 0.05, 0.1, 0.1)), 0.05, 6);
					break;
				default:
					report_error_f("Bad attribute subtype in object %p (%d)\n",ps->o, node->subtype);
					return;
			}
			// add units, if needed

			quantity = prop_subtype2quantity[PT_REAL][node->subtype];
			if (quantity != PQ_NONE)
			{
				GtkWidget *entry;
				int i;

				pi->unitsw = gtk_combo_new();
				unit = prop_get_unit(ps->o, node);
				for (i = 0; i < PROP_UNIT_MAX; i++)
					if (unit_array[i].used && unit_array[i].quantity == quantity)
						pi->glist = g_list_append (pi->glist, unit_array[i].name->text);
				pi->glist = g_list_append (pi->glist, NEW_UNIT);
				entry = (GTK_COMBO (pi->unitsw))->entry;
				gtk_combo_set_popdown_strings(GTK_COMBO (pi->unitsw), pi->glist);
				gtk_entry_set_editable(GTK_ENTRY(entry), FALSE);
				gtk_widget_show(pi->unitsw);
				gtk_widget_set_size_request(pi->valuew, 30, -1);
				gtk_widget_set_size_request(pi->unitsw, 100, -1);
				gtk_entry_set_text(GTK_ENTRY(entry), unit_array[unit].name->text);
				gtk_box_pack_end (GTK_BOX (propbox), pi->unitsw, FALSE, FALSE, 0);
				g_signal_connect(G_OBJECT(entry), "changed",
						G_CALLBACK(unit_changed), pi);
			}

//			gtk_spin_button_set_digits(GTK_SPIN_BUTTON (pi->valuew), 4);
			g_signal_connect(G_OBJECT(pi->valuew), "value-changed",
					G_CALLBACK(spin_button_real_value_changed), pi);
			break;
		default:
			report_error_f("Bad attribute type in object %p (%d)\n", ps->o, node->type);
			return;
	}

	if (node->flags & PTF_READ_ONLY)
		gtk_widget_set_sensitive(pi->valuew, FALSE);

	gtk_box_pack_start (GTK_BOX (propbox), pi->valuew, TRUE, TRUE, 0);
	gtk_table_attach(GTK_TABLE(ps->table), propbox, 1, 2,
			pi->index, pi->index+1, (GtkAttachOptions) (GTK_FILL | GTK_EXPAND),
			(GtkAttachOptions) (GTK_FILL), 0, 0);
	pi->p.key = string_ref(node->key);
	pi->p.type = node->type;
	pi->p.flags = node->flags;
	pi->p.subtype = node->subtype;
	pi->node = node;

	slist_add(&ps->valuelist, &pi->p.n);

	gtk_widget_show(pi->valuew);
}

static void create_window(struct prop_settings *ps)
{
	GtkWidget *vbox_main, *hbox_buttons;
	GtkWidget *button_close;
	GtkWidget *scrolledwindow;

	ps->redraw = 2;
	ps->cnt = 0;

	ps->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(ps->window), ps->title);
	gtk_window_set_position(GTK_WINDOW(ps->window), GTK_WIN_POS_MOUSE);
	gtk_window_set_destroy_with_parent(GTK_WINDOW(ps->window), FALSE);
	gtk_window_set_default_size (GTK_WINDOW (ps->window), 300, 270);
	gtk_widget_set_size_request (ps->window, 300, 270);

	ps->status.statusbar = gtk_statusbar_new();
	ps->status.context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(ps->status.statusbar), "Properties");

	vbox_main = gtk_vbox_new (FALSE, 5);
	gtk_widget_show (vbox_main);
	gtk_container_add (GTK_CONTAINER (ps->window),vbox_main);

	scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_show (scrolledwindow);
	gtk_box_pack_start (GTK_BOX (vbox_main), scrolledwindow, TRUE, TRUE, 0);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

	ps->viewport = gtk_viewport_new (NULL, NULL);
	gtk_widget_show (ps->viewport);
	gtk_viewport_set_shadow_type(GTK_VIEWPORT(ps->viewport), GTK_SHADOW_NONE);
	gtk_container_add (GTK_CONTAINER (scrolledwindow), ps->viewport);

	ps->table = gtk_table_new (0, 2, FALSE);
	gtk_widget_show (ps->table);
	gtk_container_add (GTK_CONTAINER (ps->viewport), ps->table);
	gtk_container_set_border_width(GTK_CONTAINER (ps->table), 6);

	hbox_buttons = gtk_hbox_new (FALSE, 9);
	gtk_widget_show (hbox_buttons);
	gtk_box_pack_start (GTK_BOX (vbox_main), hbox_buttons, FALSE, FALSE, 0);

	gtk_box_pack_start (GTK_BOX (vbox_main), ps->status.statusbar, FALSE, FALSE, 0);
	gtk_widget_show(ps->status.statusbar);

	ps->button_add = gtk_button_new_from_stock("gtk-add");
	gtk_widget_set_sensitive (ps->button_add, FALSE);
	gtk_box_pack_start (GTK_BOX (hbox_buttons), ps->button_add, TRUE, FALSE, 0);
	gtk_widget_set_size_request(ps->button_add, 84, -1);
	gtk_tooltips_set_tip(ps->tooltips, ps->button_add, "Add new property.", NULL);

	ps->button_delete = gtk_button_new_with_label("Delete ...");
	gtk_widget_set_sensitive (ps->button_delete, FALSE);
	gtk_box_pack_start (GTK_BOX (hbox_buttons), ps->button_delete, TRUE, FALSE, 0);
	gtk_widget_set_size_request (ps->button_delete, 84, -1);
	gtk_tooltips_set_tip(ps->tooltips, ps->button_delete, "Delete chosen property.", NULL);

	button_close = gtk_button_new_from_stock ("gtk-close");
	gtk_widget_show (button_close);
	gtk_box_pack_start (GTK_BOX (hbox_buttons), button_close, TRUE, FALSE, 0);
	gtk_widget_set_size_request(button_close, 84, -1);
	gtk_tooltips_set_tip(ps->tooltips, button_close, "Close window without saving changes.", NULL);

	g_signal_connect (G_OBJECT(button_close), "clicked",
			G_CALLBACK(button_close_ps_pressed), ps);
	g_signal_connect_after (G_OBJECT(ps->button_add), "clicked",
			G_CALLBACK(add_new_property), ps);
	g_signal_connect_after (G_OBJECT(ps->button_delete), "clicked",
			G_CALLBACK(delete_property), ps);
	g_signal_connect (G_OBJECT(ps->window), "destroy",
			G_CALLBACK(destroy), ps);

	gtk_widget_grab_focus (button_close);
	gtk_widget_show_all(ps->window);
}

static void remove_hooks(void)
{
	obj_hook_remove( &property_window->oh );
	if ( property_window->o && property_window->o->kind == T_GO ) go_hook_remove( go_get_tlo(o2go(property_window->o)), &property_window->gh );
}

static void add_hooks(struct prop_settings *ps, struct o* o)
{
	obj_hook_fill(&ps->oh, ps, NULL, obj_unlink_notice, NULL, NULL, obj_prop_created, obj_prop_changed, obj_prop_deleted);
	obj_hook_add(&ps->oh);

	go_hook_fill(&ps->gh, ps, NULL, go_unlink_notice, NULL, NULL, NULL, go_prop_created, go_prop_changed, go_prop_deleted, NULL);
	if ( o->kind == T_GO ) go_hook_add( go_get_tlo(o2go(o)), &ps->gh );
}

static void fill_property_window(struct prop_settings *ps, struct prop *p)
{
	struct prop *node;

	WALK_PROP(ps->o, node)
		if (node != p)
			add_property_to_table(node, ps);

	gtk_widget_set_sensitive(ps->button_add, TRUE);
	gtk_widget_set_sensitive(ps->button_delete, TRUE);
}

void redraw_property_window(struct o *o)
{
	if(property_window->o)
	{
		clean_property_window(property_window);
		remove_hooks();
		o_unref(property_window->o);
		property_window->o = NULL;
	}
	if(o)
	{
		ASSERT_O(o);
		property_window->o = o_ref(o);
		add_hooks(property_window, o);
		fill_property_window(property_window, 0);
		statusbar_clear(&property_window->status);
	}
	else
		statusbar_msg(&property_window->status, "No context available.");
}

/*
 Opens the window of object properties. If any change submitted, it must
 be shown in all views containing this object (!).
 \param title 	Title of the window (e.g. GO_NAME properites [ord])
 \param o		Pointer to the object, whose properties will be shown
*/
void create_context_prop_window(const char *title, struct o *o)
{
	property_window = xmalloc(sizeof(struct prop_settings));
	property_window->tooltips = gtk_tooltips_new();
	property_window->title = xstrdup(title);
	property_window->ap = NULL;
	property_window->dp = NULL;
	property_window->o = NULL;
	create_window(property_window);
	slist_init(&property_window->valuelist);
	if(!o)
	{
		statusbar_msg(&property_window->status, "No context available.");
		return;
	}
	ASSERT_O(o);

	property_window->o = o_ref(o);
	fill_property_window(property_window, 0);
	add_hooks(property_window, o);
}

/*
 Opens the window of object properties. If any change submitted, it must
 be shown in all views containing this object (!).
 \param title 	Title of the window (e.g. GO_NAME properites [ord])
 \param o		Pointer to the object, whose properties will be shown
*/
struct prop_settings *create_prop_window(const char *title, struct o *o)
{
	struct prop_settings *ps = xmalloc(sizeof(struct prop_settings));

	ps->tooltips = gtk_tooltips_new();
	ps->title = xstrdup(title);
	ps->ap = NULL;
	ps->dp = NULL;
	create_window(ps);

	if(!o)
		return ps;
	ASSERT_O(o);

	ps->o = o_ref(o);
	slist_init(&ps->valuelist);
	fill_property_window(ps, 0);
	add_hooks(ps, o);
	prop_window_list = g_list_append (prop_window_list, ps);

	return ps;
}
