#ifndef _GUI_PROP_SETTINGS_
#define _GUI_PROP_SETTINGS_

/**
 \file propsettings.h
 This window shows all defined properties of selected go.
*/

#include "lib/lib.h"
#include <gtk/gtk.h>
#include "kernel/kernel.h"
#include "gui/main.h"

extern struct prop_settings *property_window;
extern GList * prop_window_list; // list of opened property windows

/**
  Data for window "Add new property".
*/
struct add_property
{
	GtkWidget *window;	/**< pointer to add property window */
	GtkWidget *entry_type;	/**< pointer to the entry Type (of property) */
	GtkWidget *entry_name;	/**< pointer to the entry Name (of property) */
	struct prop_settings *ps;	/**< pointer to the prop_settings structure */
	struct status_data status;	/**< status for statusline */
};

/**
  Data for window "Delete property".
*/
struct delete_property
{
	GtkWidget *window;	/** pointer to delete property window */
	GtkWidget *entry; /** pointer to the entry of property name */
	struct prop_settings *ps; /**< pointer to the prop_settings structure */
	struct status_data status; /**< status for statusline */
	GList *glist;	/** the list of property names */
};

/**
 Structure which handles property settings window.
*/
struct prop_settings
{
	GtkWidget *window;	/**< pointer to window of property settings */
	struct slist valuelist;	/**< property values */
	struct o *o;	/**< object, whose properties will be changed */
	struct go_hook gh; /**< hooks for go */
	struct obj_hook oh; /**< hooks for objects */
	char *title;	/** title of the window */
	GtkTooltips *tooltips; /** window tooltips */
	int redraw; /** should gtk widgets be redrawn? */
	struct add_property *ap; /** pointer on add_property_window data */
	struct delete_property *dp; /** pointer on delete_property_window data */
	GtkWidget *table; /**< table for properties */
	GtkWidget *viewport; /**< scrolled window viewport containing all the properties */
	GtkWidget *button_add; /** button for adding new properties */
	GtkWidget *button_delete; /** button for deleting an existing property */
	struct status_data status; /**< status bar and context_id */
	int cnt;	/** number of rows of table */
};

/**
 Structure which handles the property and related widget (this widget contains
 the property value..)
 */
struct prop_item
{
	struct prop p;	/** property */
	GtkWidget *valuew;   /**< widget with property value */
	GtkWidget *unitsw; /**< units combobox */
	int index;		/**< index of row in table */
	struct prop_settings *ps; /** property settings */
	struct prop *node; /** pointer to the property node */
	GList *glist; /** list of possible units */
};

struct new_unit
{
	GtkWidget *window;	/**< pointer to window of property settings */
	GtkWidget *entry_name;	/**< entry for unit's name setting */
	GtkWidget *entry_multiplier;	/**< entry for unit's multiplier setting */
	struct prop_item *pi;	/**< property item */
	struct status_data status; /**< status bar and context_id */
};

/**
 * Changes the context property window's content.
 */
void redraw_property_window(struct o *o);

/**
 * Destroys whole property window.
 */
void destroy(GtkWidget *widget UNUSED, struct prop_settings *ps);

/**
 Opens the window of object properties. If any change submitted, it must
 be shown in all views containing this object (!).
 \param title	Title of the window (e.g. GO_NAME properties [ord]).
 \param o		Pointer to the object, whose properties will be shown
*/
struct prop_settings *create_prop_window(const char *title, struct o *o);

/**
 Opens the context window of object properties.
 \param title	Title of the window (e.g. GO_NAME properties [ord]).
 \param o		Pointer to the object, whose properties will be shown
*/
void create_context_prop_window(const char *title, struct o *o);

#endif
