#undef LOCAL_DEBUG
#include "lib/lib.h"
#include <stdio.h>

#include "kernel/kernel.h"

#include <gtk/gtk.h>
#include "gui/main.h"
#include "gui/cmdmgr.h"
#include "gui/glt_gui.h"

GdkColor UB_EVEN_ROW_COLOR = { 0, 0xFFFF, 0xFFFF, 0xFFFF };
GdkColor UB_ODD_ROW_COLOR = { 0, 0xFF00, 0xFFFF, 0xFFF0 };

#define GTK_TYPE_VRR_STORE		(gtk_vrr_store_get_type())
#define GTK_VRR_STORE(obj)		(G_TYPE_CHECK_INSTANCE_CAST ((obj), GTK_TYPE_VRR_STORE, GtkVrrStore))
#define GTK_VRR_STORE_CLASS(klass)	(G_TYPE_CHECK_CLASS_CAST((klass), GTK_TYPE_VRR_STORE, GtkVrrStoreClass))
#define GTK_IS_VRR_STORE(obj)		(G_TYPE_CHECK_INSTANCE_TYPE((obj), GTK_TYPE_VRR_STORE))
#define GTK_IS_VRR_STORE_CLASS(klass)	(G_TYPE_CHECK_CLASS_TYPE((klass), GTK_TYPE_VRR_STORE))
#define GTK_VRR_STORE_GET_CLASS(obj)	(G_TYPE_INSTANCE_GET_CLASS((obj), GTK_TYPE_VRR_STORE, GtkVrrStoreClass))

enum vrr_store_columns{
	VSC_TITLE,
	VSC_TYPE,
	VSC_ADDRESS,
	VSC_BG_COLOR,
	VSC_ICON,
	VSC_COUNT
};

typedef struct GtkVrrStore GtkVrrStore;
typedef struct GtkVrrStoreClass	GtkVrrStoreClass;

/** The structure providing the mediator between GtkTreeView and our internal kernel structures.*/
struct GtkVrrStore
{
	GObject parent; /**< The parent GTK object. */
	gint stamp; /**< A random naumber used to identify iters. */
	gint n_columns; /**< Column count. */
};

struct GtkVrrStoreClass
{
	GObjectClass parent_class;
};

/** A function used (very often) by the GTK. */
static GType gtk_vrr_store_get_type(void);
static GtkVrrStore * gtk_vrr_store_new(void);

static void vrr_store_object_changed( GtkVrrStore * store, struct o * o );
static void vrr_store_object_deleted( GtkVrrStore * store, struct o * father, int son_number );
static void vrr_store_object_new( GtkVrrStore * store, struct o * o );

/* The Universe Browser :) */

struct wnd_univ_browser wnd_univ_browser;

static void ub_obj_changed( struct o * o, struct obj_hook * hook UNUSED );
static void ub_obj_after_delete( struct o * o, struct o * father, struct o * prev, struct obj_hook * hook UNUSED );
static void ub_obj_added( struct o * o, struct obj_hook * hook UNUSED );

static void ub_go_changed( struct o * o, struct go_hook * hook UNUSED );
static void ub_go_after_delete( struct o * o, struct o * owner, struct o * prev, struct go_hook * hook UNUSED );
static void ub_go_added( struct o * o, struct go_hook * hook UNUSED );
static void ub_go_relink( struct o * o, struct go_group *old_group, struct go * prev, struct go_hook* h UNUSED );

static void create_universe_view(void);


static gboolean on_button_press( GtkTreeView * tree_view, GdkEventButton *event, gpointer user_data );

static gboolean on_destroy_ub( GtkWindow * w UNUSED, gpointer data UNUSED )
{
	DBG("Closing universe browser\n");

	obj_hook_remove_by_data(&wnd_univ_browser);

	struct obj* ooo;
	struct go_hook * h;
	WALK_OBJ_RECURSIVE( obj_universum, ooo )
		if ( ooo->type == OT_TLO )
		{
			DBG("UB: Removing a GO hook from %p, type %d\n", ooo, ooo->type );
			if ( h = go_hook_remove_by_data( obj2tlo(ooo), &wnd_univ_browser )) xfree(h); else BUG();
		}

	wnd_univ_browser.visible = 0;
	window_proxy_clean( (struct window*)&wnd_univ_browser );
	wnd_univ_browser.window = NULL;
	refresh_gui();
	return TRUE;
}

/* Create the browser and fill it with experimental data. */
void create_universe_browser(void){

	GtkWidget * sw;
	GtkWidget * vbox, * hbox, *wnd;
	GtkTargetEntry entry;

	wnd_univ_browser.window = wnd = gtk_window_new( GTK_WINDOW_TOPLEVEL );
	wnd_univ_browser.type = WT_UNIVERSE_BROWSER;
	wnd_univ_browser.proxy = SCM_BOOL_F;

	gtk_window_set_title( GTK_WINDOW( wnd ), "Universe browser" );
	gtk_window_set_default_size( GTK_WINDOW( wnd ), 400, 300 );
	gtk_window_add_accel_group( GTK_WINDOW(wnd), get_command_accel_group());

	vbox = gtk_vbox_new( FALSE, 0 );
	sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start( GTK_BOX( vbox ), sw, TRUE, TRUE, 0 );

	/* Create the widget before adding it */
	create_universe_view();
	gtk_container_add( GTK_CONTAINER( sw ), wnd_univ_browser.tree_view );

	hbox = gtk_hbutton_box_new();
	gtk_box_pack_start( GTK_BOX( vbox ), hbox, FALSE, FALSE, 0 );

	wnd_univ_browser.status.statusbar = gtk_statusbar_new();
	gtk_box_pack_start( GTK_BOX( vbox ), wnd_univ_browser.status.statusbar,
			FALSE, TRUE, 0 );

	wnd_univ_browser.status.context_id = gtk_statusbar_get_context_id(
			GTK_STATUSBAR(wnd_univ_browser.status.statusbar),
			"Universe browser");

	/* Initialize DND */
	entry.target = "Vrr Object";
	entry.flags = GTK_TARGET_SAME_APP;
	gtk_drag_source_set( GTK_WIDGET( wnd_univ_browser.tree_view ),
			GDK_BUTTON1_MASK, &entry, 1, GDK_ACTION_COPY );

	g_signal_connect( G_OBJECT( wnd_univ_browser.tree_view ), "button-press-event",
			G_CALLBACK( on_button_press ), NULL );
	/* Closing the window */
//	g_signal_connect_swapped( G_OBJECT( wnd_univ_browser.close_button ),
//			"clicked", G_CALLBACK( gtk_widget_destroy ),
//			G_OBJECT( wnd_univ_browser.window ) );
	g_signal_connect( G_OBJECT( wnd_univ_browser.window ),
			"delete_event", G_CALLBACK( gtk_widget_destroy ), NULL );
	/* Destroying */
	g_signal_connect( G_OBJECT( wnd_univ_browser.window ),
			"destroy", G_CALLBACK( on_destroy_ub ),
			NULL );

	gtk_container_add( GTK_CONTAINER( wnd ), vbox );
	gtk_widget_show_all( GTK_WIDGET(wnd) );

	wnd_univ_browser.visible = 1;

	DBG("UB: adding an OBJ hook\n" );
	obj_hook_add( obj_hook_fill( xmalloc_zero(sizeof(struct obj_hook)), &wnd_univ_browser,
		ub_obj_added, ub_obj_after_delete, ub_obj_changed, ub_obj_changed, NULL,
		ub_obj_changed, NULL) );

	struct obj* ooo;
	WALK_OBJ_RECURSIVE( obj_universum, ooo )
		if ( ooo->type == OT_TLO )
		{
			DBG("UB: adding a GO hook to %p\n", ooo );
			go_hook_add( obj2tlo(ooo), go_hook_fill( xmalloc_zero(sizeof(struct go_hook)),
				&wnd_univ_browser, ub_go_added, ub_go_after_delete, ub_go_relink, ub_go_changed, ub_go_changed,
				NULL, ub_go_changed, NULL, NULL));
		}
}

void ub_obj_changed( struct o * o, struct obj_hook * hook UNUSED )
{
	DBG("UB: object %p changed\n", o );
	ASSERT ( o_is_linked(o) );
	vrr_store_object_changed( wnd_univ_browser.tree_store, o );
}

void ub_go_changed( struct o * o, struct go_hook * hook UNUSED )
{
	DBG("UB: GO %p changed\n", o );
	ASSERT( o_is_linked(o) );
	vrr_store_object_changed( wnd_univ_browser.tree_store, o );
}

void ub_obj_after_delete( struct o * o, struct o * father, struct o * prev, struct obj_hook * hook UNUSED )
{
	struct obj * obj = o2obj(o), * ooo;
	struct go_hook * h;
	DBG("UB: object %p after delete\n", o );

	WALK_OBJ_RECURSIVE( obj, ooo )
		if ( ooo->type == OT_TLO )
		{
			DBG("UB: remove GO hook from %p\n", ooo );
			if ( h = go_hook_remove_by_data( obj2tlo(ooo), &wnd_univ_browser )) xfree(h); else BUG();
		}

	vrr_store_object_deleted( wnd_univ_browser.tree_store, father, obj_get_rank_son((struct obj*)prev) + 1 );
}

void ub_go_after_delete( struct o * o UNUSED, struct o * owner, struct o * prev, struct go_hook * hook UNUSED )
{
	struct go_group * gr = (struct go_group*)owner;
	struct obj_tlo * tlo = go_get_tlo(&gr->go);
	DBG("UB: go deleted from group %p, prev = %p, rank=%u\n", owner, prev, group_get_rank_go(gr, (struct go*)prev) + 1 );
	// if the owner group of the go is the toplevel group of the tlo, then pass the tlo as father;
	// otherwise, pass the owner group
	vrr_store_object_deleted( wnd_univ_browser.tree_store, ( tlo_get_tlg(tlo) == gr ) ? obj2o(&tlo->obj) : go2o(&gr->go),
		group_get_rank_go(gr, (struct go*)prev) + 1 );
}

void ub_obj_added( struct o * o, struct obj_hook * hook UNUSED )
{
	DBG("UB: obj %p added, type %d\n", o, o->type );
	ASSERT( o_is_linked(o) );

	struct obj* ooo, * obj = o2obj(o);
	WALK_OBJ_RECURSIVE( obj, ooo )
	{
		vrr_store_object_new( wnd_univ_browser.tree_store, obj2o(ooo) );
		if ( ooo->type == OT_TLO )
		{
			DBG("UB: adding a GO hook to %p\n", ooo );
			go_hook_add( obj2tlo(ooo), go_hook_fill( xmalloc_zero(sizeof(struct go_hook)),
				&wnd_univ_browser, ub_go_added, ub_go_after_delete, ub_go_relink, ub_go_changed, ub_go_changed,
				NULL, ub_go_changed, NULL, NULL));
		}
	}
}

void ub_go_added( struct o * o, struct go_hook * hook UNUSED )
{
	struct go * go = o2go(o), * go2;
	DBG("UB: go %p added\n", o );
	ASSERT( o_is_linked(o) );
	WALK_GO_RECURSIVE( go, go2 ) vrr_store_object_new( wnd_univ_browser.tree_store, go2o(go2) );
}

void ub_go_relink( struct o * o, struct go_group *old_group, struct go * prev, struct go_hook* h UNUSED )
{
	struct obj_tlo * tlo = go_get_tlo(&old_group->go);
	uns rank = group_get_rank_go(old_group, prev) + 1;
	// if the owner group of the go is the toplevel group of the tlo, then pass the tlo as father;
	// otherwise, pass the owner group
	DBG("UB: go %p relinked\n", o );
	DBG("UB: go relinked from group %p, prev = %p, rank=%u\n", old_group, prev, rank );
	// if the go was relinked into the same group, the rank of 'prev' might have increased
	if ( old_group == go_get_group(o2go(o)) )
		if ( rank > group_get_rank_go(old_group, o2go(o)) + 1 ) rank--;

	vrr_store_object_deleted( wnd_univ_browser.tree_store,
		( tlo_get_tlg(tlo) == old_group ) ? obj2o(&tlo->obj) : go2o(&old_group->go), rank );
	vrr_store_object_new( wnd_univ_browser.tree_store, o );
}

void create_universe_view(void){

	GtkCellRenderer * renderer; //, * ren2;
	GtkTreeViewColumn * column;

	DBG("Create the vrr tree store\n");
	wnd_univ_browser.tree_store = gtk_vrr_store_new();
	/* create the view */
	wnd_univ_browser.tree_view = gtk_tree_view_new_with_model(
			GTK_TREE_MODEL( wnd_univ_browser.tree_store) );
	g_object_set( G_OBJECT( wnd_univ_browser.tree_view ), "headers-clickable", TRUE, NULL );
	g_object_set( G_OBJECT( wnd_univ_browser.tree_view ), "expander-column",
			VSC_TITLE, NULL );
	/* Set alternating row colors */
	gtk_tree_view_set_rules_hint( GTK_TREE_VIEW( wnd_univ_browser.tree_view ), TRUE );

	/* NAME column */
	column = gtk_tree_view_column_new();
	g_object_set( G_OBJECT(column), "title", "Name", "max-width", 250, NULL );
	renderer = gtk_cell_renderer_text_new(); /* The text renderer */
	g_object_set( G_OBJECT( renderer ), "weight", 1000, NULL );
	gtk_tree_view_column_pack_start( column, renderer, TRUE );
	gtk_tree_view_column_set_attributes( column, renderer, "text", VSC_TITLE,
			"background", VSC_BG_COLOR, NULL );
	/* Allow sorting */
	gtk_tree_view_append_column( GTK_TREE_VIEW(wnd_univ_browser.tree_view),
			column );

	/* TYPE column */
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(
			"Type", renderer, "text", VSC_TYPE, "background", VSC_BG_COLOR, NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW(wnd_univ_browser.tree_view), column );

#ifndef VRR_POTEMKIN
	/* ADDRESS COLUMN */
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(
			"Address", renderer, "text", VSC_ADDRESS, NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW(wnd_univ_browser.tree_view), column );
#endif

	/* Selection handling */
	gtk_tree_selection_set_mode( gtk_tree_view_get_selection(GTK_TREE_VIEW( wnd_univ_browser.tree_view )), GTK_SELECTION_NONE );
}
void ub_refresh_menu(void)
{
	if ( !(wnd_univ_browser.menu) || !sync_menu( wnd_univ_browser.menu, VCL_UB_MENU ) ){
		struct cmddata cd = { NULL, NULL };
                DBG("UB menu needs rebuilding!\n");
                if (wnd_univ_browser.menu) gtk_object_sink( GTK_OBJECT(wnd_univ_browser.menu) );
                wnd_univ_browser.menu = create_menu( VCL_UB_MENU, cd );
                DBG("Rebuild OK\n");
        }
}

void modify_meta_selection( struct obj * obj, GdkEventButton * event )
{
	ASSERT( obj && event );
	if ( event->button != 1 ) return;
	char err[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Selection", err )

		if ( !(event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK)) && !(obj->flags & OF_SELECTED) )
			obj_unselect_all();

		if ( !(event->state & GDK_CONTROL_MASK) )
			obj_select(obj);

		if ( event->state & GDK_CONTROL_MASK && !(event->state & GDK_SHIFT_MASK) )
			obj_unselect(obj);
	TRANS_FAILED
		die("Something bad happened during selection: %s\n", err);
	TRANS_END
}

#define WALK_AND_DO_ENCLOSED(_xxx,_what) \
	if ( rect ) \
	{ \
		TLO_RECT_ENCLOSE_QUERY_BEGIN( _xxx, tlo, rect, gg ) \
			if ( go_get_group(gg) == group ) _what(gg); \
			else \
			{ \
				fit = 1; \
				if (grp = go_get_owner_and_member(gg, group)) \
/* TODO: but if the group is not enclosed, its next enclosed go invokes the same walk -- flag the group?? */ \
				{ \
					WALK_GO_RECURSIVE( grp, go ) \
						if (go->type != GOT_GROUP && !go_rect_enclose( go, rect )) { fit = 0; break; } \
					if (fit) _what(grp); \
				} \
			} \
		TLO_RECT_ENCLOSE_QUERY_END(_xxx); \
	} else { if (!group || go_get_group(go) == group) _what(go); }

void modify_selection( struct go * volatile go, struct obj_tlo * tlo, GdkEventButton * event, struct geom_rectangle * rect, struct go_group * group )
{
	ASSERT( tlo && (rect || go) && event );
	if ( event->button != 1 ) return;
	char err[TRANS_ERR_SIZE];
	int fit;
	struct go * grp;

	TRANS_BEGIN_MAIN( tlo, "Selection", err )

		if ( !(event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK)) && (!go || !go_is_selected(go)) )
			if ( group ) group_unselect_all_go(group);

		if ( !(event->state & GDK_CONTROL_MASK) )
			WALK_AND_DO_ENCLOSED(xxx,go_select);

		if ( event->state & GDK_CONTROL_MASK && !(event->state & GDK_SHIFT_MASK) )
			WALK_AND_DO_ENCLOSED(yyy,go_unselect);

	TRANS_FAILED
		die("Something bad happened during selection: %s\n", err);
	TRANS_END
}

static gboolean on_button_press( GtkTreeView * tree_view, GdkEventButton *event, gpointer user_data UNUSED )
{
	GtkTreePath * tp;
	GtkTreeViewColumn * twc;
	GtkTreeIter ti;
	GtkTreeModel * tm = GTK_TREE_MODEL(wnd_univ_browser.tree_store);

	factory_op_break();

	/* Let the TreeView process the event first */
	/* Do not edit this unless you know what you are doing! */
	GTK_WIDGET_GET_CLASS(tree_view)->button_press_event( GTK_WIDGET(tree_view), event );

	gtk_tree_view_get_cursor( tree_view, &tp, &twc );
	//DBG( "Tree path is %s", gtk_tree_path_to_string(tp) );

	if (likely( tp && gtk_tree_model_get_iter( tm, &ti, tp ))) {
		//the user clicked a row
		struct o * o = (struct o*)ti.user_data;
		if ( o->kind == T_GO )
		{
			struct go * go = o2go(o);
			modify_selection( go, go_get_tlo(go), event, NULL, go_get_group(go) );
			change_context( (struct window*)&wnd_univ_browser, go_get_group(go), o );
		}
		else
		{
			modify_meta_selection( o2obj(o), event );
			change_context( (struct window*)&wnd_univ_browser,
				( o->type == OT_TLO ) ? tlo_get_tlg(obj2tlo(o2obj(o))) : NULL, o );
		}
	} else {
		change_context( (struct window*)&wnd_univ_browser, NULL, NULL );
	}
	refresh_gui();

	if ( event->button == 3 ){
		gtk_menu_popup( GTK_MENU(wnd_univ_browser.menu), NULL, NULL, NULL, NULL, event->button,
				gtk_get_current_event_time() );
		return TRUE;
	}
	return TRUE;
}


/**** THE IMPLEMENTATION OF A GTK WIDGET TO SHOW OUR KERNEL STRUCTURES ****/

#define VALID_ITER(iter, vrr_store) (iter && iter->user_data)

GType column_types[] = { G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER };
gint n_columns = 4;

static GdkPixbuf * pixbuf = NULL;

static GObjectClass *parent_class = NULL;
/* initializations and finalization */
static void gtk_vrr_store_init( GtkVrrStore *vrr_store );
static void gtk_vrr_store_class_init( GtkVrrStoreClass *vrr_store_class );
static void gtk_vrr_store_tree_model_init( GtkTreeModelIface *iface );
static void gtk_vrr_store_finalize( GObject *object );

/* implementation of GtkTreeModelIface */
static GtkTreeModelFlags gtk_vrr_store_get_flags( GtkTreeModel *tree_model );
static gint gtk_vrr_store_get_n_columns( GtkTreeModel *tree_model);
static GType gtk_vrr_store_get_column_type( GtkTreeModel *tree_model, gint index);
static gboolean	gtk_vrr_store_get_iter( GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreePath *path);
static GtkTreePath *gtk_vrr_store_get_path( GtkTreeModel *tree_model, GtkTreeIter *iter);
static void gtk_vrr_store_get_value( GtkTreeModel *tree_model, GtkTreeIter *iter, gint column, GValue *value);
static gboolean	gtk_vrr_store_iter_next( GtkTreeModel *tree_model, GtkTreeIter *iter);
static gboolean	gtk_vrr_store_iter_children( GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *parent);
static gboolean gtk_vrr_store_iter_has_child( GtkTreeModel *tree_model, GtkTreeIter *iter);
static gint gtk_vrr_store_iter_n_children( GtkTreeModel *tree_model, GtkTreeIter *iter);
static gboolean gtk_vrr_store_iter_nth_child( GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *parent, gint n);
static gboolean	gtk_vrr_store_iter_parent( GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *child);

GType gtk_vrr_store_get_type(void)
{
	static GType vrr_store_type = 0;

	if (!vrr_store_type){
		static const GTypeInfo vrr_store_info =
		{ sizeof (GtkVrrStoreClass), NULL, NULL, (GClassInitFunc) gtk_vrr_store_class_init,
		NULL, NULL, sizeof (GtkVrrStore), 0, (GInstanceInitFunc) gtk_vrr_store_init, NULL };

		static const GInterfaceInfo tree_model_info =
		{ (GInterfaceInitFunc) gtk_vrr_store_tree_model_init, NULL, NULL };

		vrr_store_type = g_type_register_static (G_TYPE_OBJECT, "GtkVrrStore", &vrr_store_info, 0);

		g_type_add_interface_static (vrr_store_type, GTK_TYPE_TREE_MODEL, &tree_model_info);
	}
	return vrr_store_type;
}

static void gtk_vrr_store_class_init (GtkVrrStoreClass *class)
{
	GObjectClass *object_class;
	parent_class = g_type_class_peek_parent (class);
	object_class = (GObjectClass *) class;
	object_class->finalize = gtk_vrr_store_finalize;
}

static void gtk_vrr_store_tree_model_init (GtkTreeModelIface *iface)
{
	iface->get_flags = gtk_vrr_store_get_flags;
	iface->get_n_columns = gtk_vrr_store_get_n_columns;
	iface->get_column_type = gtk_vrr_store_get_column_type;
	iface->get_iter = gtk_vrr_store_get_iter;
	iface->get_path = gtk_vrr_store_get_path;
	iface->get_value = gtk_vrr_store_get_value;
	iface->iter_next = gtk_vrr_store_iter_next;
	iface->iter_children = gtk_vrr_store_iter_children;
	iface->iter_has_child = gtk_vrr_store_iter_has_child;
	iface->iter_n_children = gtk_vrr_store_iter_n_children;
	iface->iter_nth_child = gtk_vrr_store_iter_nth_child;
	iface->iter_parent = gtk_vrr_store_iter_parent;
}

static void gtk_vrr_store_init (GtkVrrStore *vrr_store)
{
	do vrr_store->stamp = g_random_int ();
	while (vrr_store->stamp == 0);
	vrr_store->n_columns = n_columns;
}

GtkVrrStore * gtk_vrr_store_new(void)
{ return GTK_VRR_STORE (g_object_new (GTK_TYPE_VRR_STORE, NULL)); }

/***********************************************************************/

static void gtk_vrr_store_finalize (GObject *object)
{ (* parent_class->finalize) (object); }

static GtkTreeModelFlags gtk_vrr_store_get_flags (GtkTreeModel *tree_model)
{
	DBG("gtk_vrr_store_get_flags called\n");
	g_return_val_if_fail (GTK_IS_VRR_STORE (tree_model), 0);
	return GTK_TREE_MODEL_ITERS_PERSIST; /* it means that the structure is branching */
}

static gint gtk_vrr_store_get_n_columns (GtkTreeModel *tree_model)
{
	//DBG("gtk_vrr_store_get_n_columns called\n");
	GtkVrrStore *vrr_store = (GtkVrrStore *) tree_model;
	g_return_val_if_fail (GTK_IS_VRR_STORE (tree_model), 0);
	return vrr_store->n_columns;
}

static GType gtk_vrr_store_get_column_type (GtkTreeModel *tree_model, gint index)
{
	DBG("gtk_vrr_store_get_column type with index %d called\n", index);
	g_return_val_if_fail (GTK_IS_VRR_STORE (tree_model), G_TYPE_INVALID);
	g_return_val_if_fail (index < GTK_VRR_STORE(tree_model)->n_columns &&
				index >= 0, G_TYPE_INVALID);
	/* determine what type the column is */
	return column_types[index];
}

static gboolean gtk_vrr_store_get_iter(GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreePath *path)
{
	DBG("gtk_vrr_store_get_iter called ");
	GtkVrrStore *vrr_store = (GtkVrrStore *) tree_model;
	g_return_val_if_fail (GTK_IS_VRR_STORE (vrr_store), FALSE);

	struct obj * obj;
	gint *indices;
	int depth, i = 0;

#ifdef LOCAL_DEBUG
	char * p = gtk_tree_path_to_string(path);
	DBG("for path = %s\n", p );
	g_free(p);
#endif
	indices = gtk_tree_path_get_indices (path);
	depth = gtk_tree_path_get_depth (path);
	g_return_val_if_fail (depth > 0, FALSE);

	iter->stamp = vrr_store->stamp;

	for ( obj = obj_universum; obj->type != OT_TLO && i < depth; )
		if ( !(obj = obj_get_nth_first_son(obj, indices[i++]) )) return FALSE;

	if ( i == depth ) iter->user_data = (gpointer)obj;
	else
	{
		struct go_group * group = tlo_get_tlg(obj2tlo(obj));
		struct go * go;
		for ( ; ; )
		{
			go = group_get_first_go(group);
			for ( int j = indices[i++]; j && go; j-- )
				if ( !(go = group_get_next_go(group, go)) ) return FALSE;

			if ( i == depth )
			{ iter->user_data = (gpointer)go; break; }

			if ( go->type == GOT_GROUP ) group = (struct go_group*)go;
			else return FALSE;
		}
	}
	DBG("found an object %p for that path\n", iter->user_data);
	return TRUE;
}

static GtkTreePath * gtk_vrr_store_get_path(GtkTreeModel *tree_model, GtkTreeIter  *iter)
{
	g_return_val_if_fail (GTK_IS_VRR_STORE (tree_model), NULL);
	g_return_val_if_fail (iter, NULL);
	g_return_val_if_fail (iter->user_data, NULL);
	g_return_val_if_fail (iter->stamp == GTK_VRR_STORE (tree_model)->stamp, NULL);
	DBG("gtk_vrr_store_get_path called for o = %p\n", iter->user_data);

	GtkTreePath *path = gtk_tree_path_new();

	struct o * o = void2o(iter->user_data);
	struct obj * obj;
	if ( o->kind == T_GO )
	{
		struct go * go = o2go(o);
		struct go_group * parent_group = tlo_get_tlg(go_get_tlo(go));
		struct go_group * group;
		for ( ;; )
		{
			group = go_get_group(go);
			if (!group) BUG();
			gtk_tree_path_prepend_index( path, group_get_rank_go(group, go) );
			if ( group == parent_group ) break;
			else go = &group->go;
		}
		obj = &go_get_tlo(go)->obj;
	}
	else obj = o2obj(o);

	for ( ; obj && obj != obj_universum; obj = obj->father )
		gtk_tree_path_prepend_index( path, obj_get_rank_son(obj) );
#ifdef LOCAL_DEBUG
	char * p = gtk_tree_path_to_string(path);
	DBG("resulting path = %s\n", p );
	g_free(p);
#endif
	return path;
}

/* Get the value at given position */
static void gtk_vrr_store_get_value(GtkTreeModel *tree_model, GtkTreeIter *iter, gint column, GValue *value)
{
	DBG("gtk_vrr_store_get_value called\n");
	g_return_if_fail (GTK_IS_VRR_STORE (tree_model));
	g_return_if_fail (iter && iter->user_data);
	g_return_if_fail (iter->stamp == GTK_VRR_STORE (tree_model)->stamp);
	g_return_if_fail (column < GTK_VRR_STORE (tree_model)->n_columns);

	struct o * o = void2o(iter->user_data);
	g_value_init( value, column_types[column] );

	switch ( column ){
		case VSC_TITLE:
		{
			string n = o_get_name(o);
			g_value_set_string( value, (n == string_noname) ? "" : n->text ); break;
		}
		case VSC_TYPE:
			if ( o->kind == T_OBJ ) g_value_set_string( value, string_obj_type[o->type]->text );
			else g_value_set_string( value, string_go_type[o->type]->text );
			break;
		case VSC_ADDRESS:
		{
			char buf[20];
			sprintf( buf, "%p", (void *)o );
			g_value_set_string( value, buf );
			break;
		}
		case VSC_BG_COLOR:
			if ( o->kind == T_GO ){
				if ( go_is_selected(o2go(o)) ) g_value_set_string( value, "lightblue" );
				else g_value_reset( value );
			} else {
				if ( obj_is_selected(o2obj(o)) ) g_value_set_string( value, "lightpink" );
				else g_value_reset( value );
			}
			break;
		case VSC_ICON:
			g_value_set_pointer( value, pixbuf ); break;
		default: BUG();
	}
}

/* Move the iter to next node */
static gboolean gtk_vrr_store_iter_next (GtkTreeModel *tree_model, GtkTreeIter *iter)
{
	g_return_val_if_fail (iter, FALSE);
	g_return_val_if_fail (iter->user_data, FALSE);
	g_return_val_if_fail (iter->stamp == GTK_VRR_STORE (tree_model)->stamp, FALSE);
	DBG("gtk_vrr_store_iter_next called for o = %p\n", iter->user_data);

	struct o * o = void2o(iter->user_data);

	if ( o->kind == T_OBJ )
		iter->user_data = (gpointer)obj_get_next_son(o2obj(o));

	else
	{
		struct go * go = o2go(o);
		iter->user_data = (gpointer)group_get_next_go(go_get_group(go), go);
	}
	DBG("Has next: %s\n", iter->user_data ? "yes" : "no" );
	return iter->user_data != NULL;
}

/* Move the iter to child */
static gboolean gtk_vrr_store_iter_children (GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *parent)
{
	g_return_val_if_fail (!parent || parent->user_data, FALSE);
	g_return_val_if_fail (!parent || parent->stamp == GTK_VRR_STORE (tree_model)->stamp, FALSE);
	DBG("gtk_vrr_store_iter_children called for o %p\n", parent->user_data);

	iter->stamp = parent->stamp;
	struct o * o = void2o(parent->user_data);
	if ( o->kind == T_OBJ )
	{
		if ( o->type == OT_TLO ) iter->user_data = (gpointer)group_get_first_go(tlo_get_tlg(obj2tlo(o2obj(o))));
		else iter->user_data = (gpointer)obj_get_first_son( o2obj(o) );
	} else {
		if ( o->type == GOT_GROUP ) iter->user_data = (gpointer)group_get_first_go((struct go_group*)o);
		else iter->user_data = NULL;
	}

	DBG("any children: %s\n", iter->user_data ? "yes" : "no" );
	return iter->user_data != NULL;
}
/* If the iter has a child */
static gboolean gtk_vrr_store_iter_has_child (GtkTreeModel *tree_model, GtkTreeIter  *iter)
{
	DBG("gtk_vrr_store_iter_has_child called\n");
	g_return_val_if_fail (GTK_IS_VRR_STORE (tree_model), FALSE);
	g_return_val_if_fail (iter->stamp == GTK_VRR_STORE (tree_model)->stamp, FALSE);
	g_return_val_if_fail (iter->user_data, FALSE);

	struct o * o = void2o(iter->user_data);

	if ( o->kind == T_OBJ )
	{
		if ( o->type == OT_TLO ) return group_get_first_go(tlo_get_tlg(obj2tlo(o2obj(o)))) != NULL;
		else return obj_get_first_son(o2obj(o)) != NULL;
	}
	else return o->type == GOT_GROUP && group_get_first_go((struct go_group*)o);
}
/* Probably the number of childen of given node */
static gint gtk_vrr_store_iter_n_children (GtkTreeModel *tree_model, GtkTreeIter  *iter)
{
	DBG("gtk_vrr_store_iter_n_children called\n");
	g_return_val_if_fail (GTK_IS_VRR_STORE (tree_model), 0);
	g_return_val_if_fail (!iter || iter->user_data, FALSE);
	struct o  * o = void2o(iter->user_data);
	struct go_group * group = NULL;
	int cnt = 0;

	if ( o->kind == T_OBJ && o->type == OT_TLO ) group = tlo_get_tlg(obj2tlo(o2obj((o))));
	if ( o->kind == T_GO && o->type == GOT_GROUP ) group = (struct go_group*)group;

	if ( group )
		for ( struct go * g = group_get_first_go(group); g; g = group_get_next_go(group, g) ) cnt++;

	else if ( o->kind  == T_OBJ )
		for ( struct obj * ooo = obj_get_first_son( o2obj(o) ); ooo; ooo = obj_get_next_son(ooo) ) cnt++;

	DBG("the count of children: %d\n", cnt );
	return cnt;
}

/* Move to the nth child */
static gboolean gtk_vrr_store_iter_nth_child (GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *parent, gint n)
{
	DBG("gtk_vrr_store_iter_nth_child called\n");
	g_return_val_if_fail (GTK_IS_VRR_STORE (tree_model), FALSE);

	struct o * o;
	if ( !parent || !(parent->user_data) ) o = obj2o(obj_universum);
	else o = void2o(parent->user_data);

	struct go_group * group = NULL;
	iter->stamp = GTK_VRR_STORE(tree_model)->stamp;
	iter->user_data = NULL;

	if ( o->kind == T_OBJ && o->type == OT_TLO ) group = tlo_get_tlg(obj2tlo(o2obj(o)));
	if ( o->kind == T_GO && o->type == GOT_GROUP ) group = (struct go_group*)group;

	if ( group )
	{
		struct go * g;
		for ( g = group_get_first_go(group); g && n; group_get_next_go(group, g) ) n--;
		iter->user_data = (gpointer)g;
	}
	else if ( o->kind  == T_OBJ )
	{
		struct obj * ooo;
		for ( ooo = obj_get_first_son( o2obj(o) ); ooo && n; obj_get_next_son(ooo) ) n--;
		iter->user_data = (gpointer)ooo;
	}
	return iter->user_data != NULL && n == 0;
}

/* Move to parent */
static gboolean gtk_vrr_store_iter_parent (GtkTreeModel *tree_model, GtkTreeIter  *iter,
			    GtkTreeIter  *child)
{
	DBG("gtk_vrr_store_iter_parent called\n");
	g_return_val_if_fail (iter, FALSE);
	g_return_val_if_fail (child, FALSE);
	g_return_val_if_fail (child->user_data, FALSE);
	g_return_val_if_fail (child->stamp == (GTK_VRR_STORE(tree_model))->stamp, FALSE);

	struct o * o = void2o(child->user_data);
	iter->stamp = GTK_VRR_STORE(tree_model)->stamp;
	if ( o->kind == T_GO )
	{
		struct obj_tlo * tlo = go_get_tlo(o2go(o));
		if ( tlo_get_tlg(tlo) == go_get_group(o2go(o)) )
			iter->user_data = (gpointer)tlo; //we must not return a toplevel group, return the tlo instead
		else
			iter->user_data = o_get_ancestor(o);

	} else iter->user_data = o_get_ancestor(o);

	return iter->user_data != NULL;
}

void vrr_store_object_changed( GtkVrrStore * store, struct o * o )
{
	GtkTreeIter iter;
	iter.stamp = store->stamp;
	iter.user_data = (gpointer)o;
	DBG("object changed\n");
	g_signal_emit_by_name( (gpointer)store, "row-changed",  NULL, &iter, NULL );
}

void vrr_store_object_deleted( GtkVrrStore * store, struct o * father, int son_number )
{
	GtkTreeIter iter;
	GtkTreePath * path;
	iter.stamp = store->stamp;
	iter.user_data = (gpointer)father;
	DBG("object deleted from father %p, rank %d\n", father, son_number );
	path = gtk_vrr_store_get_path( GTK_TREE_MODEL(store), &iter ); //get the father's path
	gtk_tree_path_append_index( path, son_number );
	g_signal_emit_by_name( (gpointer)store, "row-deleted", path, NULL );

	if ( father->kind == T_OBJ && !obj_get_first_son(o2obj(father)) ||
		father->kind == T_GO && father->type == GOT_GROUP && !group_get_first_go(go2group(o2go(father))))
		// the last son will be deleted
		g_signal_emit_by_name( (gpointer)store, "row-has-child-toggled", NULL, &iter, NULL );

	gtk_tree_path_free( path );
}

void vrr_store_object_new( GtkVrrStore * store, struct o * o )
{
	GtkTreeIter iter;
	iter.stamp = store->stamp;
	iter.user_data = (gpointer)o;
	DBG("object added\n");
	g_signal_emit_by_name( (gpointer)store, "row-inserted",  NULL, &iter, NULL );

/*	if ( !obj_get_younger_brother(obj_get_oldest_son(obj->father))){
		iter.user_data = (gpointer)(obj->father); //his father has one child only, so it must have been added now
		g_signal_emit_by_name( (gpointer)store, "row-has-child-toggled", NULL, &iter, NULL );
	}*/
}

