#undef LOCAL_DEBUG
#include "lib/lib.h"

// TODO zvetsovani okna

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <ctype.h>

#include "kernel/kernel.h"
#include "kernel/go.h"

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "gui/main.h"
#include "gui/cmdmgr.h"
#include "gui/ruler.h"
#include "gui/changeunits.h"
#include "gui/visualisation.h"
#include "gui/properties.h"
#include "gui/glt_gui.h"

#include "vcl/vcl.h"
#include "geomlib/geomlib.h"

#define PAGE_INCREMENT_X 5
#define PAGE_INCREMENT_Y 5
#define MOVE_INTERVAL    20

enum transform_display_state
{
	TDM_HIDDEN = 0, TDM_VISIBLE = 1, TDM_NONEMPTY = 1 << 1, TDM_WIDE = 1 << 2, TDM_HIGH = 1 << 3, TDM_SHIFT = 1 << 4,
	TDM_SHAPE_MASK = TDM_SHIFT - 1, TDM_RECT = (TDM_WIDE | TDM_HIGH)
};

GtkWidget * view_menu;
GSList * view_list;

static void set_zoom_label( struct status_data * status, real zoom, real rotation );
static void set_view_as_active( struct view * vw );
static void ensure_op(void);

static gboolean button_press_callback(GtkWidget *widget UNUSED, GdkEventButton *event, struct view *vw);
static gboolean button_release_callback(GtkWidget *widget UNUSED, GdkEventButton *event, struct view *vw);
static gboolean key_press_callback(GtkWidget *widget UNUSED, GdkEventKey *event, struct view * vw);
static void transform_recalc( struct view * vw );
static void transform_redisplay( struct view * vw );
static void transform_hide( struct view * vw );
static void view_sel_changed(struct o *o, struct go_hook* h);
static void view_sel_changed_2(struct go *go, const struct changed_data_generic *data UNUSED, struct go_hook* h);
static void view_obj_before_delete( struct o* o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook *hook );
static void view_go_before_delete( struct o* o, struct o* owner UNUSED, struct o* prev UNUSED, struct go_hook *hook );
static void view_obj_prop_change( struct o * o, struct obj_hook *hook );

static void create_view_navigator(struct view * vw);
static void view_nav_update( struct view_navigator * vwn, real zoom, real rotation );
static void view_nav_redraw( struct view_navigator * vwn );
static void view_nav_resize( GtkWidget *w UNUSED, GdkRectangle *a UNUSED, struct view_navigator *vwn );
static void view_nav_value_changed( GtkWidget * w UNUSED, struct view_navigator * vwn );
static gboolean view_nav_motion(GtkWidget *widget UNUSED, GdkEventMotion *event, struct view_navigator *vwn);
static gboolean view_nav_button_press(GtkWidget *widget UNUSED, GdkEventButton *event, struct view_navigator *vwn);
static int change_in_progress;

static struct
{
	int events_pending;
	int last_pix_x, last_pix_y;
	struct geom_point start_coords, last_coords, last_vcoords;
	struct view * last_view, * first_view;
	guint32 last_time, prev_time;
	gulong hook_id;
	guint signal_id;
	GdkEventButton press_event;
} moving;
enum moving_mode moving_mode;

static gboolean button_release_anywhere( GSignalInvocationHint *ihint UNUSED, guint n UNUSED, const GValue *p UNUSED, gpointer d UNUSED );
static gboolean process_last_me( void * u UNUSED );
static void calculate_selection( struct view * vw );
static struct vcl_path * tmp_rect;

enum tf_device
{
	TFD_NONE,
	TFD_MOVE,
	TFD_ROTATE,
	TFD_SCALE_EQ,
	TFD_SCALE_X,
	TFD_SCALE_Y,
	TFD_MOVE_ORIGIN,
	TFD_SKEW,
	TFD_AXIS,
	TFD_CNT
};
static GdkCursorType tf_device_cursor[TFD_CNT] = { GDK_LEFT_PTR, GDK_FLEUR, GDK_EXCHANGE, GDK_SIZING,
	GDK_SB_H_DOUBLE_ARROW, GDK_SB_V_DOUBLE_ARROW, GDK_FLEUR, GDK_DIAMOND_CROSS, GDK_FLEUR };

static enum tf_device transform_device;

void view_obj_before_delete( struct o* o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook *hook )
{
	struct view * view = (struct view*)hook->data;
	if ( o_is_ancestor( o, go2o(&view->group->go) )) gtk_widget_destroy( view->window );
}

static void view_go_before_delete( struct o* o, struct o* owner UNUSED, struct o* prev UNUSED, struct go_hook *hook )
{
	struct view * view = (struct view*)hook->data;
	if ( o_is_ancestor( o, go2o(&view->group->go) )) gtk_widget_destroy( view->window );

}
static void view_destroy( GtkWidget *widget UNUSED, struct view *vw )
{
        visualisation_finish (&(vw -> vis));
	if ( vw->navigator ) gtk_widget_destroy( vw->navigator->window );

	go_hook_remove( vw->tlo, &vw->go_hook );
	obj_hook_remove( &vw->obj_hook );
	obj_unref( &vw->tlo->obj );
	expand_unref( vw->tlo );
	go_unref( &vw->group->go );
	//remove the view from the list of all views
	view_list = g_slist_remove( view_list, (gpointer)vw );

	//change_context( NULL, NULL, NULL ); //this will be changed by another view
	window_proxy_clean ((struct window *) vw);
	xfree(vw);
	open_windows_cnt--;
	if(!open_windows_cnt) main_quit ();
	//refresh_gui();
}

static void view_recompute_tolerance( struct view * vw )
{
	struct geom_point pi0, pit, po0, pot;
	pi0.x = 0; pi0.y = 0;
	pit.x = factory.snap_tol_pix; pit.y = 0;
	visualisation_pixels_to_coords( &vw->vis, &pi0, &po0 );
	visualisation_pixels_to_coords( &vw->vis, &pit, &pot );
	vw->tolerance = hypot( po0.x - pot.x, po0.y - pot.y );
	DBG("Recomputed tolerance, %d pixels = %.3f coords.\n", factory.snap_tol_pix, vw->tolerance );
}

static void update_rulers (struct view *vw)
{
	struct geom_vector px, py;
	struct geom_transform t = geom_transform_identity;

	vcl_inverted_apply_matrix (vw -> vis . area, &t);
	px . dx = 1; px . dy = 0;
	py . dx = 0; py . dy = 1;
	geom_vector_transform (&px, &t, &px);
	geom_vector_transform (&py, &t, &py);

	ruler_set_resolution (RULER (vw -> hruler),  1 / hypot (px . dx, px . dy) *
		prop_lookup_real( obj2o(obj_universum), string_rulerx_res, RULERX_RES) );
	ruler_set_resolution (RULER (vw -> vruler), -1 / hypot (py . dx, py . dy) *
		prop_lookup_real( obj2o(obj_universum), string_rulery_res, RULERY_RES) );
}

static void visualisation_changed_cb (struct view *vw)
{
	GtkWidget *visw = vw -> vis . canvas -> gtk -> widget;
	gint width = visw -> allocation . width;
	gint height = visw -> allocation . height;
	struct geom_rectangle bbox1, bbox2, bbox3;
	struct geom_point pi, po, ux, zero;
	GtkAdjustment *adj;
	real xzoom;
	real xrot;

	struct geom_transform t = geom_transform_identity;
        vcl_primary_apply_matrix (vw -> vis . area, &t);

        pi . x = pi . y = 0;
        geom_point_transform (&pi, &t, &po);

	visualisation_get_visible_bbox (&(vw -> vis), &bbox1);
	geom_rectangle_transform (&bbox1, &t, &bbox2);
	bbox1 . ll . x = bbox1 . ur . x = width / 2.0;
	bbox1 . ll . y = bbox1 . ur . y = height / 2.0;
	geom_rectangle_merge (&bbox1, &bbox2, &bbox3);

	adj = vw -> hadj;
	adj -> lower = bbox3 . ll . x - width / 2.0 - po . x;
	adj -> upper = bbox3 . ur . x + width / 2.0 - po . x;
	adj -> value = - po . x;
	gtk_widget_queue_draw (vw -> hscrollbar);
	ruler_set_offset (RULER (vw -> hruler), po . x);

	adj = vw -> vadj;
	adj -> lower = bbox3 . ll . y - height / 2.0 - po . y;
	adj -> upper = bbox3 . ur . y + height / 2.0 - po . y;
	adj -> value = - po . y;
	gtk_widget_queue_draw (vw -> vscrollbar);
	ruler_set_offset (RULER (vw -> vruler), po . y);

	//calculate the zoom and rotation
	ux.x = 1;
	ux.y = zero.x = zero.y = 0;
	visualisation_coords_to_vcoords( &vw->vis, &ux, &ux );
	visualisation_coords_to_vcoords( &vw->vis, &zero, &zero );
	xzoom = hypot( ux.x - zero.x , ux.y - zero.y ) * 100.0;
	xrot = (real)atan2( (double)(ux.y - zero.y), (double)(ux.x - zero.x) ) / R_PI * 180.0;
	set_zoom_label( &vw->status, xzoom, xrot );
	if ( vw->navigator )
	{ view_nav_update( vw->navigator, xzoom, xrot ); view_nav_redraw( vw->navigator ); }

	update_rulers (vw);
}

static void view_size_allocate (GtkWidget *widget UNUSED, GdkRectangle *allocation UNUSED, struct view *vw)
{
	GtkWidget *visw = vw->vis.canvas->gtk->widget;
	GtkAdjustment *adj;
	gint height = visw -> allocation . height;
	gint width = visw -> allocation . width;
	real delta;

	if (vw -> state == 0)
	  {
	    visualisation_changed_cb (vw);
	    vw -> state = 1;
	  }

	adj = vw -> hadj;
	delta = width - adj -> page_size;

	adj -> page_size = width;
	adj -> lower -= delta / 2;
	adj -> value -= delta / 2;
	adj -> upper += delta / 2;

	if (! IS_ZERO (delta))
	  {
	    gtk_widget_queue_draw (vw -> hscrollbar);
	    ruler_set_offset (RULER (vw -> hruler), - adj -> value);
	  }

	adj = vw -> vadj;
	delta = height - adj -> page_size;

	adj -> page_size = height;
	adj -> lower -= delta / 2;
	adj -> value -= delta / 2;
	adj -> upper += delta / 2;

	if (! IS_ZERO (delta))
	  {
	    gtk_widget_queue_draw (vw -> vscrollbar);
	    ruler_set_offset (RULER (vw -> vruler), - adj -> value);
	  }
	if ( vw->navigator ) view_nav_redraw( vw->navigator );
}


static int visualisation_fit_box( struct visualisation * vis, struct geom_rectangle bbox )
{
	if ( bbox.ur.x <= bbox.ll.x && bbox.ur.y <= bbox.ll.y )
		return 1;

	GtkWidget *visw = vis->canvas->gtk->widget;
	gint width = visw->allocation.width;
	gint height = visw->allocation.height;
	struct geom_rectangle vis_rect;
	real xratio, yratio;

	vis_rect.ll.x = vis_rect.ll.y = 0;
	vis_rect.ur.x = width - 1;
	vis_rect.ur.y = -height + 1;
	visualisation_pixels_to_coords( vis, &vis_rect.ll, &vis_rect.ll );
	visualisation_pixels_to_coords( vis, &vis_rect.ur, &vis_rect.ur );

	xratio = ( vis_rect.ur.x - vis_rect.ll.x ) / ( bbox.ur.x - bbox.ll.x );
	yratio = ( vis_rect.ur.y - vis_rect.ll.y ) / ( bbox.ur.y - bbox.ll.y );
	xratio = MIN( xratio, yratio );

	if (likely( xratio > 0 )) visualisation_scale( vis, xratio, xratio ); //FIXME: this should not be necessary!
	visualisation_center( vis, (bbox.ur.x + bbox.ll.x) * 0.5, (bbox.ur.y + bbox.ll.y) * 0.5 );
	return 0;
}

void on_view_fit_selected( struct cmd * cmd UNUSED, void * subj )
{
	ASSERT(subj && ((struct window*)subj)->type == WT_VIEW);
	struct view * vw = (struct view*)subj;
	struct geom_rectangle t;
	visualisation_get_selected_bbox( &vw->vis, &t );
	visualisation_fit_box( &vw->vis, t ); // TODO: nejaka hezka hlaska, kdyz selze
}

void on_view_fit_all( struct cmd * cmd UNUSED, void * subj )
{
	ASSERT(subj && ((struct window*)subj)->type == WT_VIEW);
	struct view * vw = (struct view*)subj;
	struct geom_rectangle t;
	visualisation_get_visible_bbox( &vw->vis, &t );
	visualisation_fit_box( &vw->vis, t ); // TODO: nejaka hezka hlaska, kdyz selze
}

static void view_zoom_in( GtkWidget * w UNUSED, struct view * vw )
{
	visualisation_scale( &vw->vis, 1.25, 1.25 );
	view_recompute_tolerance(vw);
}
void on_view_zoom_in( struct cmd * cmd UNUSED, void * subj )
{
	ASSERT(subj && ((struct window*)subj)->type == WT_VIEW);
	view_zoom_in( NULL, (struct view*)subj );
}
static void view_zoom_out( GtkWidget * w UNUSED, struct view * vw )
{
	visualisation_scale( &vw->vis, 0.8, 0.8 );
	view_recompute_tolerance(vw);
}
void on_view_zoom_out( struct cmd * cmd UNUSED, void * subj )
{
	ASSERT(subj && ((struct window*)subj)->type == WT_VIEW);
	view_zoom_out( NULL, (struct view*)subj );
}
static void view_zoom_reset( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_set_orientation( &vw->vis, 0.0, 0.0, 1.0, 1.0, 0.0 ); }

void on_view_zoom_reset( struct cmd * cmd UNUSED, void * subj )
{
	ASSERT(subj && ((struct window*)subj)->type == WT_VIEW);
	view_zoom_reset( NULL, (struct view*)subj );
}
static void view_move_up( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_move( &vw->vis, 0, -PAGE_INCREMENT_Y ); }

static void view_move_down( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_move( &vw->vis, 0, PAGE_INCREMENT_Y ); }

static void view_move_left( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_move( &vw->vis, PAGE_INCREMENT_X, 0 ); }

static void view_move_right( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_move ( &vw->vis, -PAGE_INCREMENT_X, 0 ); }

static void view_rotate_plus( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_rotate( &vw->vis, 0.2 ); }

static void view_rotate_minus( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_rotate( &vw->vis, -0.2 ); }

static void button_navigate_pressed(GtkWidget *button UNUSED, struct view *vw)
{
	if ( vw->navigator ) gtk_window_present( GTK_WINDOW(vw->navigator->window) );
	else create_view_navigator(vw);
}

static gboolean scroll_callback(GtkWidget *widget UNUSED, GdkEventScroll *event, struct view *vw)
{
	switch ( event->direction )
	{
		case GDK_SCROLL_UP:	view_zoom_in( NULL, vw ); break;
		case GDK_SCROLL_DOWN:	view_zoom_out( NULL, vw ); break;
		default: break;
	}
	return TRUE;
}
static void scrollbar_value_changed (GtkAdjustment *x UNUSED, struct view *vw)
{
	visualisation_absolut_move_px (&(vw -> vis), - vw -> hadj -> value,
				       - vw -> vadj -> value);
}

/* This is where Santiago's original transform tool begins */

static real tmp_hypot (real x, real y, struct geom_point p)
{
  x -= p . x;
  y -= p . y;
  return (x * x + y * y);
}

static int get_near_tf_point (struct view *vw, real x, real y)
{
  real old_delta = vw -> tolerance * vw -> tolerance;
  real new_delta;
  int retval = 0;

  if ((vw -> vis . flags & VIS_FLAGS_TF_MOVE) &&
      ((new_delta = tmp_hypot (x, y, vw -> vis . transform_points [0])) < old_delta))
    {
      retval = 1;
      old_delta = new_delta;
    }
  else if ((vw -> vis . flags & VIS_FLAGS_TF_SECOND) &&
	   ((new_delta = tmp_hypot (x, y, vw -> vis . transform_points [1])) < old_delta))
    {
      retval = (vw -> vis . flags & VIS_FLAGS_TF_RESIZE) ? 2 : 3;
      old_delta = new_delta;
    }
  else if ((vw -> vis . flags & VIS_FLAGS_TF_SKEW) &&
	   ((new_delta = tmp_hypot (x, y, vw -> vis . transform_points [2])) < old_delta))
    {
      retval = 4;
      old_delta = new_delta;
    }

  return (retval);
}

static int add_new_tf_point (struct view *vw, real x, real y)
{
  if (! (vw -> vis . flags & VIS_FLAGS_TF_MOVE))
    {
      visualisation_set_tf_move (&(vw -> vis), x, y);
      return 1;
    }
  else if (! (vw -> vis . flags & VIS_FLAGS_TF_SECOND))
    {
      visualisation_set_tf_resize (&(vw -> vis), x, y);
      return 2;
    }
  else // if (! (vw -> vis . flags & VIS_FLAGS_TF_SKEW))
    {
      visualisation_set_tf_skew (&(vw -> vis), x, y);
      return 4;
    }
}

static int compute_transformed_points (struct view *vw, struct geom_transform *t, struct geom_point *pts)
{
  int i = 0;
  if (vw -> vis . flags & VIS_FLAGS_TF_MOVE)
    i = i || geom_point_transform (vw -> vis . transform_points + 0, t, pts + 0);
  if (vw -> vis . flags & VIS_FLAGS_TF_SECOND)
    i = i || geom_point_transform (vw -> vis . transform_points + 1, t, pts + 1);
  if (vw -> vis . flags & VIS_FLAGS_TF_SKEW)
    i = i || geom_point_transform (vw -> vis . transform_points + 2, t, pts + 2);

  return (i);
}

static void update_transformed_points (struct view *vw, struct geom_point *pts)
{
  if (vw -> vis . flags & VIS_FLAGS_TF_MOVE)
    visualisation_set_tf_move (&(vw -> vis), pts [0] . x, pts [0] . y);
  if (vw -> vis . flags & VIS_FLAGS_TF_RESIZE)
    visualisation_set_tf_resize (&(vw -> vis), pts [1] . x, pts [1] . y);
  if (vw -> vis . flags & VIS_FLAGS_TF_ROTATE)
    visualisation_set_tf_rotate (&(vw -> vis), pts [1] . x, pts [1] . y);
  if (vw -> vis . flags & VIS_FLAGS_TF_SKEW)
    visualisation_set_tf_skew (&(vw -> vis), pts [2] . x, pts [2] . y);
}



static int transform_mode_1 (struct geom_transform2 *t, const struct geom_point *pf,
			     const struct geom_point *pt)
{ return geom_transform2_set_move (t, pt -> x - pf -> x, pt -> y - pf -> y); }

static int transform_mode_2 (struct geom_transform2 *t, const struct geom_point *p1,
			     const struct geom_point *pf, const struct geom_point *pt)
{
  int i = 0;
  i = i || geom_transform2_set_identity (t);
  i = i || geom_transform2_translate (t, - p1 -> x, - p1 -> y);
  i = i || geom_transform2_scale (t, (pt -> x - p1 -> x) / (pf -> x - p1 -> x),
				 (pt -> y - p1 -> y) / (pf -> y - p1 -> y));
  i = i || geom_transform2_translate (t, p1 -> x,  p1 -> y);
  return (i ? -1 : 0);
}

static int transform_mode_3 (struct geom_transform2 *t, const struct geom_point *p1,
			     const struct geom_point *pf, const struct geom_point *pt)
{
  int i = 0;
  double pfx = pf -> x - p1 -> x;
  double pfy = pf -> y - p1 -> y;
  double ptx = pt -> x - p1 -> x;
  double pty = pt -> y - p1 -> y;
  double scale = hypot (ptx, pty) / hypot (pfx, pfy);
  double angle = atan2 (pty, ptx) - atan2 (pfy, pfx);
  i = i || geom_transform2_set_identity (t);
  i = i || geom_transform2_translate (t, - p1 -> x, - p1 -> y);
  i = i || geom_transform2_scale (t, scale, scale);
  i = i || geom_transform2_rotate (t, angle);
  i = i || geom_transform2_translate (t, p1 -> x,  p1 -> y);
  return (i ? -1 : 0);
}

static int transform_mode_4 (struct geom_transform2 *t,
			     const struct geom_point *p1, const struct geom_point *p2,
			     const struct geom_point *pf, const struct geom_point *pt)
{ return (geom_transform2_set_skew (t, p1 -> x, p1 -> y, p2 -> x, p2 -> y,
				    pf -> x, pf -> y, pt -> x, pt -> y)); }

/* This is where Santiago's original transform tool ends */

#define SNAP_DEVICE(d,a,b) tmpt.x = (a); tmpt.y = (b);			\
	dist = hypot( pt->x - tmpt.x, pt->y - tmpt.y );			\
	if ( dist < maxdist && ( mindist < 0 || mindist > dist ))	\
	{ dist = mindist; dev = (d); dev_pt = tmpt;}

static enum tf_device snap_to_tf_device( struct view * vw, struct geom_point * pt, int mod1, int mod2 UNUSED )
{
        struct geom_rectangle bb;
	enum tf_device dev = TFD_NONE;
	struct geom_point dev_pt;
	struct geom_point tmpt;
	real dist, mindist = -1.0, maxdist = vw->tolerance;
        visualisation_get_selected_bbox( &vw->vis, &bb );
	if ( !(vw->transform_state & TDM_NONEMPTY) ) return TFD_NONE;

	if ( vw->transform_state & TDM_NONEMPTY )
	{
		// origin
		SNAP_DEVICE( mod1 ? TFD_MOVE_ORIGIN : TFD_MOVE, vw->transform_origin.x, vw->transform_origin.y );
	}
	if ( (vw->transform_state & TDM_RECT) == TDM_RECT )
	{
		// scale/skew at midpoints of the edges
		SNAP_DEVICE( mod1 ? TFD_SKEW : TFD_SCALE_X, bb.ur.x, (bb.ll.y + bb.ur.y)/2 );
		SNAP_DEVICE( mod1 ? TFD_SKEW : TFD_SCALE_X, bb.ll.x, (bb.ll.y + bb.ur.y)/2 );
		SNAP_DEVICE( mod1 ? TFD_SKEW : TFD_SCALE_Y, (bb.ll.x + bb.ur.x)/2, bb.ll.y );
		SNAP_DEVICE( mod1 ? TFD_SKEW : TFD_SCALE_Y, (bb.ll.x + bb.ur.x)/2, bb.ur.y );
	}
	if ( vw->transform_state & TDM_RECT )
	{
		// scale equally or rotate at corners
		SNAP_DEVICE( mod1 ? TFD_ROTATE : TFD_SCALE_EQ, bb.ll.x, bb.ll.y );
		SNAP_DEVICE( mod1 ? TFD_ROTATE : TFD_SCALE_EQ, bb.ur.x, bb.ur.y );
		SNAP_DEVICE( mod1 ? TFD_ROTATE : TFD_SCALE_EQ, bb.ll.x, bb.ur.y );
		SNAP_DEVICE( mod1 ? TFD_ROTATE : TFD_SCALE_EQ, bb.ur.x, bb.ll.y );
	}
	if ( mod1 && (vw->transform_state & TDM_RECT) == TDM_RECT )
	{
		SNAP_DEVICE( TFD_AXIS, vw->axis.x, vw->axis.y ); //edit axis
	}
	if ( dev == TFD_NONE && // vertical bbox lines
		(fabs(bb.ll.x - pt->x) < maxdist || fabs(bb.ur.x - pt->x) < maxdist) &&
		(pt->y < bb.ur.y + maxdist) && (pt->y > bb.ll.y - maxdist) )
	{
		dev = TFD_MOVE; dev_pt = *pt;
	}
	if ( dev == TFD_NONE && // horizontal bbox lines
		(fabs(bb.ll.y - pt->y) < maxdist || fabs(bb.ur.y - pt->y) < maxdist) &&
		(pt->x < bb.ur.x + maxdist) && (pt->x > bb.ll.x - maxdist) )
	{
		dev = TFD_MOVE; dev_pt = *pt;
	}
	if ( dev != TFD_NONE ) *pt = dev_pt;
	return dev;
}

static int process_transform_step_x( struct view * vw, struct geom_point po )
{
	struct geom_transform2 t;
	volatile int errval;
	struct geom_point zero = vw->transform_origin;
	struct geom_point prev = moving.last_coords;
	switch ( transform_device ){
		case TFD_NONE:
			BUG(); break;
		case TFD_MOVE:
			errval = geom_transform2_set_move( &t, po.x - prev.x, po.y - prev.y );
			break;
		case TFD_SCALE_X:
			errval = geom_transform2_set_scale_center( &t, zero.x, zero.y, (po.x - zero.x) / (prev.x - zero.x), 1.0 );
			break;
		case TFD_SCALE_Y:
			errval = geom_transform2_set_scale_center( &t, zero.x, zero.y, 1.0, (po.y - zero.y) / (prev.y - zero.y) );
			break;
		case TFD_SCALE_EQ:
		{
			real r = hypot(po.x - zero.x, po.y - zero.y) / hypot(prev.x - zero.x, prev.y - zero.y);
			errval = geom_transform2_set_scale_center( &t, zero.x, zero.y, r, r );
			break;
		}
		case TFD_ROTATE:
		{
			real angle = atan2( prev.x - zero.x, prev.y - zero.y ) - atan2( po.x - zero.x, po.y - zero.y );
			errval = geom_transform2_set_rotate_center( &t, zero.x, zero.y, angle );
			break;
		}
		case TFD_SKEW:
		{
			errval = geom_transform2_set_skew_normal( &t, zero.x, zero.y, vw->axis.x, vw->axis.y,
				prev.x, prev.y, po.x, po.y );
			break;
		}
		case TFD_MOVE_ORIGIN:
			vw->axis.x += po.x - vw->transform_origin.x;
			vw->axis.y += po.y - vw->transform_origin.y;
			vw->transform_origin = po;
			visualisation_set_center_gadget( &vw->vis, vw->transform_origin.x, vw->transform_origin.y );
			visualisation_set_xaxis_gadget( &vw->vis, vw->axis.x, vw->axis.y );
			return 0;
		case TFD_AXIS:
			vw->axis = po;
			visualisation_set_xaxis_gadget( &vw->vis, vw->axis.x, vw->axis.y );
			return 0;
		default: BUG();
	}
	if (errval == 0)
  	{
    		char errs[TRANS_ERR_SIZE];
		TRANS_BEGIN_MAIN(vw->tlo, factory.desc->text, errs)
			tsort_presorted_transform(&t);
		TRANS_FAILED
			err( "%s\n", errs );
			errval = 1;
		TRANS_END;
		if ( !errval )
		{
			// transform the center and axes accordingly (but usually they are fixpoints)
			struct geom_point dxa = { vw->axis.x - vw->transform_origin.x, vw->axis.y - vw->transform_origin.y };
			errval = geom_point_transform( &vw->transform_origin, &t.primary, &vw->transform_origin );
			if ( !errval )
			{
				vw->axis.x = vw->transform_origin.x + dxa.x;
				vw->axis.y = vw->transform_origin.y + dxa.y;
				visualisation_set_center_gadget( &vw->vis, vw->transform_origin.x, vw->transform_origin.y );
				visualisation_set_xaxis_gadget( &vw->vis, vw->axis.x, vw->axis.y );
			} else { DBG("Warning: transformation of gadgets failed\n"); }

			factory_merge_trans();
			moving.last_coords = po;
		}
	}
	else err( "Numeric error in transformation\n" );
	return !errval;
}

static int process_transform_step( struct view * vw, struct geom_point po )
{
	struct geom_transform2 t;
	volatile int errval;

	switch (vw -> tf_drag_on)
	{
	case 1: errval = transform_mode_1 (&t, vw -> vis . transform_points + 0, &po);
		break;
	case 2: errval = transform_mode_2 (&t, vw -> vis . transform_points + 0,
					       vw -> vis . transform_points + 1, &po);
   		break;
  	case 3: errval = transform_mode_3 (&t, vw -> vis . transform_points + 0,
					       vw -> vis . transform_points + 1, &po);
    		break;
  	case 4: errval = transform_mode_4 (&t, vw -> vis . transform_points + 0,
					       vw -> vis . transform_points + 1,
					       vw -> vis . transform_points + 2, &po);
		break;
	default: BUG ();
	}

	if (errval == 0)
  	{
    		char errs[TRANS_ERR_SIZE];

		TRANS_BEGIN_MAIN (vw -> tlo, factory.desc->text, errs)
		{
			tsort_presorted_transform(&t);

			struct geom_point p [3];
			int i = compute_transformed_points (vw, &(t . primary), p);
			if (i != 0) trans_fail ("Numeric error in transformation");
			update_transformed_points (vw, p);
		}
		TRANS_FAILED
			err( "%s\n", errs );
			errval = 1;
		TRANS_END;
		if ( !errval ) factory_merge_trans();
	}
	else err( "Numeric error in transformation\n" );
	return !errval;
}

static gboolean process_last_me( void * u UNUSED )
{
	if ( moving_mode == MOVE_NONE ) { DBG("Warning: process_last_me called when not moving!\n"); return TRUE; }
	ASSERT( moving.last_view );
	if ( !moving.events_pending || ( moving.last_time > moving.prev_time && moving.last_time - moving.prev_time < MOVE_INTERVAL ) )
		return TRUE; //not yet
	moving.events_pending = 0;

	struct geom_point pi = { moving.last_pix_x, moving.last_pix_y };
	struct geom_point po, vp;
	visualisation_pixels_to_coords( &moving.last_view->vis, &pi, &po );

	switch ( moving_mode )
	{
		case MOVE_FACTORY: factory_moving_step(po); break;
		case MOVE_TRANSFORM: process_transform_step_x(moving.last_view, po); break;
		case MOVE_TRANSFORM_2: process_transform_step(moving.last_view, po); break;
		case MOVE_PANNING:
		case MOVE_PANNING_IMG:
		case MOVE_PANNING_NAV:
			if ( moving.last_view != moving.first_view ) break;
			if ( moving_mode == MOVE_PANNING_NAV )
			{
				visualisation_pixels_to_coords( &moving.last_view->navigator->vis, &pi, &vp );
				visualisation_coords_to_vcoords( &moving.last_view->vis, &vp, &vp );
			}
			else
				visualisation_pixels_to_vcoords( &moving.last_view->vis, &pi, &vp );
			if ( moving_mode == MOVE_PANNING_IMG )
				visualisation_move( &moving.last_view->vis, vp.x - moving.last_vcoords.x, vp.y - moving.last_vcoords.y );
			else
				visualisation_move( &moving.last_view->vis, moving.last_vcoords.x - vp.x, moving.last_vcoords.y - vp.y );
			moving.last_vcoords = vp;
			break;
		case MOVE_SELECT:
		{
			moving.last_coords = po;
			struct geom_point corner = { moving.last_coords.x, moving.start_coords.y };
			visualisation_gui_rectangle_update( &moving.first_view->vis, &tmp_rect,
				moving.last_coords, corner, moving.start_coords );
			break;
		}
		default:;
	}
	return TRUE;
}

void view_moving_start( struct geom_point pix, struct view * vw, enum moving_mode mode, GdkCursorType cursor_type )
{
	ASSERT( vw->navigator || mode != MOVE_PANNING_NAV );
	ASSERT( moving_mode == MOVE_NONE  && mode != MOVE_NONE );
	moving.last_time = moving.prev_time = gtk_get_current_event_time();
	moving.last_pix_x = pix.x; moving.last_pix_y = pix.y;
	moving.events_pending = 0;
	moving.start_coords = moving.last_coords;
	moving.first_view = moving.last_view = vw;
	moving_mode = mode;

	g_idle_add( (GSourceFunc)process_last_me, moving.first_view );

	GdkCursor * cursor = gdk_cursor_new( cursor_type );
	switch ( mode )
	{
	case MOVE_PANNING:
	case MOVE_PANNING_IMG:
		gdk_window_set_cursor( vw->vis.canvas->gtk->widget->window, cursor );
		break;
	case MOVE_PANNING_NAV: break;
	default:
		for ( GSList * l = view_list; l; l = l->next )
		{
			struct view * vvv = (struct view*)l->data;
			if ( vvv->tlo == vw->tlo ) gdk_window_set_cursor( vvv->vis.canvas->gtk->widget->window, cursor );
		}
	}
	gdk_cursor_unref( cursor );

	// this is necessary so that no other widget steals our release event that we are waiting for
	moving.hook_id = g_signal_add_emission_hook( moving.signal_id = g_signal_lookup("button-release-event", GTK_TYPE_WIDGET),
				0, button_release_anywhere, NULL, NULL );
}

void view_moving_end(void)
{
	if ( moving.events_pending )
	{
		moving.prev_time = 0;
		process_last_me(NULL); //force the processing of the last one
		/* FIXME should be the coordinates of button release instead of this hack! */
	}
	if ( moving.first_view->tf_drag_on ) moving.first_view->tf_drag_on = 0;
	//unset the cursor for all relevant views
	switch ( moving_mode )
	{
	case MOVE_PANNING:
	case MOVE_PANNING_IMG:
		gdk_window_set_cursor( moving.first_view->vis.canvas->gtk->widget->window, NULL );
		break;
	case MOVE_PANNING_NAV: break;
	default:
		for ( GSList * l = view_list; l; l = l->next )
		{
			struct view * vvv = (struct view*)l->data;
			if ( vvv->tlo == moving.first_view->tlo ) gdk_window_set_cursor( vvv->vis.canvas->gtk->widget->window, NULL );
		}
	}
	g_idle_remove_by_data(moving.first_view);
	g_signal_remove_emission_hook( moving.signal_id, moving.hook_id );
	moving_mode = MOVE_NONE;
	//moving.first_view = moving.last_view = NULL;
}

static gboolean motion_notify_callback(GtkWidget *widget UNUSED, GdkEventMotion *event, struct view *vw)
{
	vert_ruler_motion_notify(vw->vruler, event);
	horiz_ruler_motion_notify(vw->hruler, event);
	if ( moving_mode != MOVE_NONE && vw->tlo == moving.first_view->tlo )
	{
		moving.last_pix_x = event->x; moving.last_pix_y = event->y;
		moving.last_view = vw;
		moving.last_time = event->time;
		moving.events_pending = 1;
	}
	return TRUE;
}

static gboolean button_release_anywhere( GSignalInvocationHint *ihint UNUSED, guint n UNUSED, const GValue *p UNUSED, gpointer d UNUSED )
{ button_release_callback( NULL, NULL, NULL ); return TRUE; }

static gboolean button_release_callback(GtkWidget *widget UNUSED, GdkEventButton *event UNUSED, struct view *vw UNUSED)
{
	switch ( moving_mode )
	{
		case MOVE_FACTORY: return TRUE; // the click has been handled after the 'press' callback
		case MOVE_TRANSFORM:
		case MOVE_TRANSFORM_2: factory_op_step(); break;
		case MOVE_PANNING:
		case MOVE_PANNING_IMG:
		case MOVE_PANNING_NAV: view_moving_end(); break;
		case MOVE_SELECT:
			view_moving_end();
			visualisation_gui_rectangle_destroy( &moving.first_view->vis, &tmp_rect );
			calculate_selection(moving.first_view);
			refresh_gui();
			break;
		default:;
	}
	return TRUE;
}

static gboolean ruler_button_press_callback (GtkWidget *widget UNUSED, GdkEventButton *event, struct view * vw)
{
	set_view_as_active(vw);

	if ( event->button == 3 ) //open menu
	{
		view_refresh_menu();
		gtk_menu_popup( GTK_MENU(view_menu), NULL, NULL, NULL, NULL, event->button, gtk_get_current_event_time() );
	}
	return TRUE;
}

static gboolean button_press_callback(GtkWidget *widget UNUSED, GdkEventButton *event, struct view *vw)
{
	struct geom_point pi = {event -> x, event -> y};
	struct geom_point po;
	moving.press_event = *event;

	visualisation_pixels_to_coords (&(vw -> vis), &pi, &po);
	DBG("%f %f\n", po . x, po . y);
	set_view_as_active(vw);

	if ( event->button == 1 ) switch ( factory.state->input_type ) {
	case OFIT_HANGER:
		ensure_op();
		snap_point( &po, &factory.state->value, &vw->grid, context.tlo, factory.snap_tolerance );
		factory_op_step();
		break;
	case OFIT_ANCHOR:
	{
		if ( moving_mode != MOVE_NONE ) return TRUE; // prevent from misbehaviour in case of a double-click
		struct dist_pass_result dpr;
		ensure_op();
		snap_to_anchor( &po, &dpr, context.tlo, factory.snap_tolerance );
		if ( dpr.type == DIST_PASS_ANCHOR )
		{
			factory.state->value = dpr;
			factory_op_step();
		}
		else
		{
			//select
			moving.last_coords = po;
			visualisation_gui_rectangle_update( &vw->vis, &tmp_rect, po, po, po );
			view_moving_start( pi, vw, MOVE_SELECT, GDK_TOP_LEFT_ARROW );
		}
		break;
	}
	case OFIT_POINT:
		ensure_op();
		factory.state->value.type = OFIT_POINT;
		factory.state->value.data.point.point = po;
		factory_op_step();
		break;
	case OFIT_GO:
	{
		struct dist_pass_result dpr;
		ensure_op();
		snap_to_go( &po, &dpr, context.tlo, factory.snap_tolerance );
		if ( dpr.type == DIST_PASS_GO )
		{
			go_ref(dpr.data.go.go);
			factory.state->value = dpr;
			factory_op_step();
		}
		break;
	}
	case OFIT_SEL:
	{
		if ( moving_mode != MOVE_NONE ) return TRUE; // prevent from misbehaviour in case of a double-click
		ensure_op();
		moving.last_coords = po;
		if ( transform_device = snap_to_tf_device( vw, &po, event->state & GDK_SHIFT_MASK, event->state & GDK_CONTROL_MASK ) )
		{
			factory_op_step();
			view_moving_start( pi, vw, MOVE_TRANSFORM, tf_device_cursor[transform_device] );
		} else {
			//(rectangular) selection
			visualisation_gui_rectangle_update( &vw->vis, &tmp_rect, po, po, po );
			view_moving_start( pi, vw, MOVE_SELECT, GDK_TOP_LEFT_ARROW );
		}
		break;
	}
	case OFIT_TF:
	{
		if ( moving_mode != MOVE_NONE ) return TRUE; // prevent from misbehaviour in case of a double-click
		ensure_op();
		if ( event->state & GDK_SHIFT_MASK ) //edit transformation points
		{
			DBG("Edit transformation point\n");
			switch (get_near_tf_point (vw, po . x, po . y))
			{
			case 0:
				add_new_tf_point (vw, po . x, po . y);
				break;
			case 1:
		 		visualisation_unset_tf_move (&(vw -> vis));
				break;
			case 2:
				// because user clicked on resize img, then there are
				// valid data in visualisation . transform_points [1]
				visualisation_set_tf_rotate (&(vw -> vis), vw -> vis . transform_points [1] . x,
								vw -> vis . transform_points [1] . y);
		   		break;
			case 3:
				visualisation_unset_tf_second (&(vw -> vis));
				break;
			case 4:
				visualisation_unset_tf_skew (&(vw -> vis));
				break;
			}
		}
		else
		{
			if ( vw -> tf_drag_on = get_near_tf_point (vw, po . x, po . y) )
			{
				factory_op_step();
				view_moving_start( pi, vw, MOVE_TRANSFORM_2, GDK_FLEUR );
			}
		}
		break;
	}
	default:;
	}
	if ( moving_mode != MOVE_NONE ) return TRUE; // for sure

	if ( event->button == 2 )
	{
		if ( event->state & GDK_CONTROL_MASK )
		{
			snap_point( &po, NULL, &vw->grid, context.tlo, factory.snap_tolerance );
			visualisation_center( &vw->vis, po.x, po.y );
		} else {
			visualisation_pixels_to_vcoords( &vw->vis, &pi, &moving.last_vcoords );
			uns pan_img = prop_lookup_uns( obj2o(obj_universum), string_drag_image, PANNING_DRAG_IMAGE );
			view_moving_start( pi, vw, pan_img ? MOVE_PANNING_IMG : MOVE_PANNING, GDK_HAND2 );
		}
	}

	if ( event->button == 3 ) //open menu
	{
		view_refresh_menu();
		gtk_menu_popup( GTK_MENU(view_menu), NULL, NULL, NULL, NULL, event->button, gtk_get_current_event_time() );
	}
	return TRUE;
}

static void calculate_selection( struct view * vw )
{
	if ( fabs(moving.last_coords.x - moving.start_coords.x) <= factory.snap_tolerance ||
		fabs(moving.last_coords.y - moving.start_coords.y) <= factory.snap_tolerance )
	{
		// individual selection
		struct dist_pass_result dpr;
		struct geom_point po = moving.last_coords;
		snap_to_go( &po, &dpr, context.tlo, factory.snap_tolerance );
		if ( dpr.type == DIST_PASS_GO )
		{
			modify_selection( dpr.data.go.go, go_get_tlo(dpr.data.go.go), &moving.press_event, NULL, vw->group );
			change_context( (struct window*)vw, vw->group, go2o(go_get_owner_and_member(dpr.data.go.go, vw->group)) );
		} else {
			char err[TRANS_ERR_SIZE];
			TRANS_BEGIN_MAIN( context.tlo, "Unselect all", err );
				group_unselect_all_go(context.group);
			TRANS_FAILED
	       			die("Unselect failed with %s", err);
			TRANS_END
		}
	} else {
		//rectangular selection
		struct geom_rectangle rect;
		if ( moving.last_coords.x < moving.start_coords.x )
		{ rect.ll.x = moving.last_coords.x; rect.ur.x = moving.start_coords.x; }
		else
		{ rect.ll.x = moving.start_coords.x; rect.ur.x = moving.last_coords.x; }
		if ( moving.last_coords.y < moving.start_coords.y )
		{ rect.ll.y = moving.last_coords.y; rect.ur.y = moving.start_coords.y; }
		else
		{ rect.ll.y = moving.start_coords.y; rect.ur.y = moving.last_coords.y; }

		modify_selection( NULL, context.tlo, &moving.press_event, &rect, vw->group );
		change_context( (struct window*)vw, vw->group, (struct o*)group_get_first_selected_go(vw->group) );
	}
}

static gboolean key_press_callback(GtkWidget *widget UNUSED, GdkEventKey *event, struct view * vw)
{
	if ( vw->tlo != factory.tlo || factory.start_state != &ofs_select ) return FALSE;

	if ( event->keyval == GDK_Shift_L || event->keyval == GDK_Shift_R )
	{
		if ( event->type == GDK_KEY_PRESS ) vw->transform_state |= TDM_SHIFT;
		else vw->transform_state &= ~TDM_SHIFT;
		transform_redisplay(vw);
	}
	return FALSE;
}

void statusbar_msg(struct status_data *status, const char * fmt, ... )
{
	if (!GTK_IS_STATUSBAR(status->statusbar) || !fmt) return;
	va_list ap;
        char buf[TRANS_ERR_SIZE];
        va_start( ap, fmt );
        int cnt = vsnprintf( buf, TRANS_ERR_SIZE, fmt, ap );
        va_end( ap );

	gtk_statusbar_pop(GTK_STATUSBAR(status->statusbar), status->context_id);
	if(unlikely( cnt < 0 || cnt >= TRANS_ERR_SIZE ))
		report_error("Statusbar message too long.\n");
	else
		if (cnt) gtk_statusbar_push(GTK_STATUSBAR(status->statusbar), status->context_id, buf);
}

void statusbar_clear (struct status_data *status)
{
	if (!GTK_IS_STATUSBAR (status->statusbar))
			return;
	gtk_statusbar_pop (GTK_STATUSBAR (status->statusbar), status->context_id);
	gtk_statusbar_push (GTK_STATUSBAR (status->statusbar), status->context_id, " ");
}

static void set_view_as_active( struct view * vw )
{
	change_context( (struct window*)vw, vw->group,
			(context.o && context.o->kind == T_GO && go_get_group(o2go(context.o)) == vw->group) ?
			context.o : NULL );

	factory.snap_tolerance = vw->tolerance;
}
static void ensure_op(void)
{
	if ( factory.group != context.group )
	{
		if ( factory.operation_active ) factory_op_break();
		factory_op_start( context.group, factory.start_state, factory.desc->text, factory.inside_trans );
	}
}

static void set_zoom_label( struct status_data * status, real zoom, real rotation )
{
	char buf[TRANS_ERR_SIZE];
	if (!GTK_IS_LABEL(status->zoom_bar)) return;
	sprintf( buf, " %.1f%%, %.1f* ", zoom, rotation );
	gtk_label_set_text( GTK_LABEL(status->zoom_bar), buf );
}

static void set_window_title(GtkWindow *window, struct obj_tlo * tlo)
{
	const char * fn, *objname, *fathername;
	char *buf;

	objname = obj_lookup_name (&tlo->obj);
	if ( tlo->obj.father->type == OT_ZOMBIE )
	{ gtk_window_set_title(window, objname); return; }

	fathername = obj_lookup_name (tlo->obj.father);
	if (fn = prop_lookup_text( obj2o(tlo->obj.father), string_filename))
	{
		buf = alloca (strlen(fn) + strlen(objname) + strlen(fathername) + 8);
		sprintf(buf, "%s of %s (%s)", objname, fathername, fn);
	}
	else
	{
		buf = alloca (strlen(objname) + strlen(fathername) + 5);
		sprintf(buf, "%s of %s", objname, fathername);
	}
	gtk_window_set_title (window, buf);
}

// create view window
struct view *create_view(struct go_group * group)
{
	GtkWidget *table;
	GtkWidget *button_tl, *button_tr, *button_br;
	GtkWidget *button_tl_icon, *button_tr_icon, *button_br_icon;
	GtkWidget *vscrollbarbox, *hscrollbarbox, *hbox;
	GtkWidget *button_left, *button_right, *button_up, *button_down;
	GtkWidget *button_left_icon, *button_right_icon, *button_up_icon, *button_down_icon;
	GtkRange *range;
	struct view *vw = xmalloc_zero(sizeof(struct view));

	vw->type = WT_VIEW;
	vw->proxy = SCM_BOOL_F;

	vw->state = 0;
	vw->tf_drag_on = 0;
	vw->group = group;
	vw->tlo = go_get_tlo(&group->go);
	vw->navigator = NULL;
	obj_ref(&vw->tlo->obj);
	expand_ref(vw->tlo);
	go_ref(&vw->group->go);

	vw->obj_hook.data = vw;
	vw->obj_hook.obj_unlink = view_obj_before_delete;
	vw->obj_hook.obj_prop_change = view_obj_prop_change;
	obj_hook_add( &vw->obj_hook );

	vw->go_hook.data = vw;
	vw->go_hook.go_selected = vw->go_hook.go_unselected = view_sel_changed;
	vw->go_hook.go_changed = view_sel_changed_2;
	vw->go_hook.go_unlink = view_go_before_delete;
	go_hook_add( vw->tlo, &vw->go_hook );

	open_windows_cnt++;

	vw->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	set_window_title(GTK_WINDOW(vw->window), vw->tlo);
	gtk_widget_set_size_request (vw->window, 500, -1);
	gtk_window_add_accel_group( GTK_WINDOW( vw->window ), get_command_accel_group() );
	g_signal_connect (G_OBJECT (vw->window), "destroy", G_CALLBACK (view_destroy), vw);
	g_signal_connect_after (G_OBJECT (vw->window), "size_allocate", G_CALLBACK(view_size_allocate), vw);

	gtk_container_add(GTK_CONTAINER(vw->window), vw->table = table = gtk_table_new(4, 5, FALSE) );

	vw->toolbar = NULL;
	vw->toolbar2 = NULL;
	view_refresh_toolbar(vw);

	// vertical ruler
	vw->vruler = vert_ruler_new();
	gtk_table_attach(GTK_TABLE(table), vw->vruler, 2, 3, 1, 2, GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
	g_signal_connect(G_OBJECT (vw->vruler), "button-press-event", G_CALLBACK (ruler_button_press_callback), vw);

	// box for vertical scrollbar and its buttons
	vscrollbarbox = gtk_vbox_new(FALSE,0);

	// button up
	button_up = gtk_button_new();
	button_up_icon = image_get_from_icon(icon_arrow_up);
	gtk_container_add(GTK_CONTAINER(button_up), button_up_icon);

	gtk_box_pack_start(GTK_BOX(vscrollbarbox), button_up, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(button_up), "clicked", G_CALLBACK(view_move_up), vw);

	// vertical scrollbar
	// TODO change adj's values
	vw->vadj = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 0, 1, 1, 1));
	vw->vscrollbar = gtk_vscrollbar_new(vw->vadj);
	gtk_box_pack_start(GTK_BOX(vscrollbarbox), vw->vscrollbar, TRUE, TRUE, 0);

	gtk_table_attach(GTK_TABLE(table), vscrollbarbox, 4, 5, 1, 2, GTK_FILL, GTK_FILL, 0, 0);

	range = GTK_RANGE(vw->vscrollbar);
	range->has_stepper_a = range->has_stepper_d = FALSE;

	// button down
	button_down = gtk_button_new();
	button_down_icon = image_get_from_icon(icon_arrow_down);
	gtk_container_add(GTK_CONTAINER(button_down), button_down_icon);

	gtk_box_pack_start(GTK_BOX(vscrollbarbox), button_down, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(button_down), "clicked", G_CALLBACK(view_move_down), vw);

	// horizontal ruler
	vw->hruler = horiz_ruler_new();
	gtk_table_attach (GTK_TABLE(table), vw->hruler, 3, 4, 0, 1, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
	g_signal_connect(G_OBJECT (vw->hruler), "button-press-event", G_CALLBACK (ruler_button_press_callback), vw);

	// box for horizontal scrollbar and ints buttons
	hscrollbarbox = gtk_hbox_new(FALSE, 0);

	// button left
	button_left = gtk_button_new();
	button_left_icon = image_get_from_icon(icon_arrow_left);
	gtk_container_add(GTK_CONTAINER(button_left), button_left_icon);

	gtk_box_pack_start(GTK_BOX(hscrollbarbox), button_left, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(button_left), "clicked", G_CALLBACK(view_move_left), vw);

	// horizontal scrollbar
	// TODO pagesize
	vw->hadj = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 0, 1, 1, 1));
	vw->hscrollbar = gtk_hscrollbar_new(vw->hadj);
	gtk_box_pack_start(GTK_BOX(hscrollbarbox), vw->hscrollbar, TRUE, TRUE, 0);

	range = GTK_RANGE(vw->hscrollbar);
	range->has_stepper_a = range->has_stepper_d = FALSE;

	// button right
	button_right = gtk_button_new();
	button_right_icon = image_get_from_icon(icon_arrow_right);
	gtk_container_add(GTK_CONTAINER(button_right), button_right_icon);

	gtk_box_pack_start(GTK_BOX(hscrollbarbox), button_right, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(button_right), "clicked", G_CALLBACK(view_move_right), vw);


	gtk_table_attach (GTK_TABLE(table), hscrollbarbox, 3, 4, 2, 3, GTK_FILL, GTK_FILL, 0, 0);

	g_signal_connect (G_OBJECT(vw->vscrollbar), "value-changed",
			G_CALLBACK(scrollbar_value_changed), vw);
	g_signal_connect (G_OBJECT(vw->hscrollbar), "value-changed",
			G_CALLBACK(scrollbar_value_changed), vw);

	// button (top, right)
	button_tr = gtk_button_new();
	button_tr_icon = image_get_from_icon(icon_rotation_right);
	gtk_container_add(GTK_CONTAINER(button_tr), button_tr_icon);
	gtk_table_attach(GTK_TABLE(table), button_tr, 4, 5, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
	g_signal_connect(G_OBJECT(button_tr), "clicked", G_CALLBACK(view_rotate_minus), vw);

	// button (top, left)
	button_tl = gtk_button_new();
	button_tl_icon = image_get_from_icon(icon_rotation_left);
	gtk_container_add(GTK_CONTAINER(button_tl), button_tl_icon);
	gtk_table_attach(GTK_TABLE(table), button_tl, 2, 3, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
	g_signal_connect(G_OBJECT(button_tl), "clicked", G_CALLBACK(view_rotate_plus), vw);

	// button (bottom, right)
	button_br = gtk_button_new();
	button_br_icon = image_get_from_icon(icon_zoom_out);
	gtk_container_add(GTK_CONTAINER(button_br), button_br_icon);
	gtk_table_attach(GTK_TABLE(table), button_br, 4, 5, 2, 3, GTK_FILL, GTK_FILL, 0, 0);
	g_signal_connect(G_OBJECT(button_br), "clicked", G_CALLBACK(button_navigate_pressed), vw);

	/*// button (bottom, left)
	button_bl = gtk_button_new();
	button_bl_icon = image_get_from_icon(icon_zoom_in);
	gtk_container_add(GTK_CONTAINER(button_bl), button_bl_icon);
	gtk_table_attach(GTK_TABLE(table), button_bl, 2, 3, 2, 3, GTK_FILL, GTK_FILL, 0, 0);
	g_signal_connect(G_OBJECT(button_bl), "clicked", G_CALLBACK(view_zoom_out), vw);*/

	// status bars
	gtk_table_attach( GTK_TABLE(table), hbox = gtk_hbox_new( FALSE, 0 ), 0, 5, 3, 4, GTK_FILL, GTK_FILL, 0, 0 );
	gtk_box_pack_start( GTK_BOX(hbox), vw->status.statusbar = gtk_statusbar_new(), TRUE, TRUE, 0 );
	gtk_box_pack_start( GTK_BOX(hbox), vw->image_on = gtk_image_new_from_pixbuf( icon_pencil_on ), FALSE, TRUE, 0 );
	gtk_box_pack_start( GTK_BOX(hbox), vw->image_modified = gtk_image_new_from_pixbuf( icon_modified ), FALSE, TRUE, 0 );
	gtk_box_pack_start( GTK_BOX(hbox), vw->status.zoom_bar = gtk_label_new(""), FALSE, TRUE, 0 );
	vw->status.context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(vw->status.statusbar), "View created.");
	gtk_statusbar_set_has_resize_grip( GTK_STATUSBAR(vw->status.statusbar), FALSE );

	//add the view into the list of all views
	view_list = g_slist_prepend( view_list, (gpointer)vw );

	gtk_table_attach(GTK_TABLE(table),
			 visualisation_init (&(vw -> vis), vw->group,
					     (simple_cb) visualisation_changed_cb, vw),
			 3, 4, 1, 2, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
	gtk_widget_add_events(vw -> vis . canvas -> gtk -> widget, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK);
	g_signal_connect(G_OBJECT (vw -> vis . canvas -> gtk -> widget), "motion-notify-event", G_CALLBACK (motion_notify_callback), vw);
	g_signal_connect(G_OBJECT (vw -> vis . canvas -> gtk -> widget), "button-press-event", G_CALLBACK (button_press_callback), vw);
	g_signal_connect(G_OBJECT (vw -> vis . canvas -> gtk -> widget), "button-release-event", G_CALLBACK (button_release_callback), vw);
	g_signal_connect(G_OBJECT (vw->window), "key-press-event", G_CALLBACK (key_press_callback), vw);
	g_signal_connect(G_OBJECT (vw->window), "key-release-event", G_CALLBACK (key_press_callback), vw);
	g_signal_connect(G_OBJECT (vw -> vis . canvas -> gtk -> widget), "scroll-event", G_CALLBACK (scroll_callback), vw);

	gtk_widget_show_all (vw->window);
	gtk_widget_set_size_request (vw->window, -1, -1);
	gtk_window_set_resizable (GTK_WINDOW (vw->window), TRUE);

	view_recompute_tolerance(vw);
        range = GTK_RANGE(vw->vscrollbar);
        range->has_stepper_a = range->has_stepper_d = FALSE;
        gtk_widget_show(vw->vscrollbar);

        range = GTK_RANGE(vw->hscrollbar);
        range->has_stepper_a = range->has_stepper_d = FALSE;
        gtk_widget_show(vw->hscrollbar);

	visualisation_changed_cb(vw); //initialize the rulers and scrollbars
	on_view_fit_all( NULL, vw );
	change_context( (struct window*)vw, vw->group, NULL );
	int err = geom_transform2_set_basis( &vw->grid, 0.0, 0.0, GRID_X1, GRID_X2, GRID_Y1, GRID_Y2 );
	ASSERT(!err);
	view_update_grid(vw);
	transform_update(vw);
	return vw;
}

static void view_obj_prop_change( struct o * o, struct obj_hook *hook )
{
	struct view * vw = (struct view*)hook->data;
	if ( o_is_ancestor( o, obj2o(&vw->tlo->obj)) )
		set_window_title( GTK_WINDOW(vw->window), vw->tlo );

	if ( o == obj2o(obj_universum) )
	{
		// snap tolerance
		uns tolerance = prop_lookup_uns( o, string_tolerance, TOLERANCE_PIX );
		if ( tolerance != factory.snap_tol_pix )
		{
			factory.snap_tol_pix = tolerance;
			if ( context.w->type == WT_VIEW ) view_recompute_tolerance( (struct view*)context.w );
		}
		// grid and rulers
		for ( GSList * l = view_list; l; l = l->next )
		{
			view_update_grid( (struct view*)l->data );
			update_rulers( (struct view*)l->data );
		}
	}
}

void view_update_grid( struct view * vw )
{
	struct o * o = obj2o(obj_universum);
	if ( factory.snap_set & SNAP_GRID )
	{
		real x1 = prop_lookup_real( o, string_grid_x1, GRID_X1 );
		real y1 = prop_lookup_real( o, string_grid_y1, GRID_Y1 );
		real x2 = prop_lookup_real( o, string_grid_x2, GRID_X2 );
		real y2 = prop_lookup_real( o, string_grid_y2, GRID_Y2 );
	
		if ( x1 != vw->grid.primary.coef[0][0] || y1 != vw->grid.primary.coef[0][1] ||
			x2 != vw->grid.primary.coef[1][0] || y2 != vw->grid.primary.coef[1][1] )
		{
			struct geom_transform2 orig_grid = vw->grid;
			if (geom_transform2_set_basis( &vw->grid, 0.0, 0.0, x1, x2, y1, y2 ))
			{
				DBG("Erroneous grid %.3f, %.3f, %.3f, %.3f\n", x1, y1, x2, y2 );
				if (geom_transform2_set_basis( &vw->grid, 0.0, 0.0, x1, -y1, y1, x1 ))
				{
					DBG("Erroneous grid %.3f, %.3f, %.3f, %.3f\n", x1, y1, -y1, x1 );
					vw->grid = orig_grid;
				}
			}
		}
		visualisation_set_grid( &vw->vis, &vw->grid );
	}
	else visualisation_unset_grid( &vw->vis );
}

void view_sel_changed(struct o *o UNUSED, struct go_hook* h)
{ transform_recalc( (struct view*)h->data ); }

void view_sel_changed_2(struct go *go, const struct changed_data_generic *data UNUSED, struct go_hook* h)
{
	if ( moving_mode == MOVE_NONE && (go->flags & GOF_SELECTED) )
		transform_recalc( (struct view*)h->data );
}

void view_refresh_toolbar(struct view * vw)
{
	if ( !(vw->toolbar) || !sync_toolbar( vw->toolbar, VCL_VIEW_TBAR ) ){
		struct cmddata cd = { (struct window*)vw, vw->group };
		DBG("View toolbar needs rebuilding!\n");
		if ( vw->toolbar ) gtk_container_remove( GTK_CONTAINER(vw->table), vw->toolbar );
		gtk_widget_show( vw->toolbar = create_toolbar( VCL_VIEW_TBAR, 1, cd ) );// create toolbar
		gtk_table_attach(GTK_TABLE(vw->table), vw->toolbar, 0, 1, 0, 3, GTK_FILL, GTK_FILL, 0, 0);
		DBG("Rebuild OK\n");
	}
	if ( !(vw->toolbar2) || !sync_toolbar( vw->toolbar2, VCL_VIEW_TBAR2 ) ){
		struct cmddata cd = { (struct window*)vw, vw->group };
		DBG("View toolbar2 needs rebuilding!\n");
		if ( vw->toolbar2 ) gtk_container_remove( GTK_CONTAINER(vw->table), vw->toolbar2 );
		gtk_widget_show( vw->toolbar2 = create_toolbar( VCL_VIEW_TBAR2, 1, cd ) );// create toolbar
		gtk_table_attach(GTK_TABLE(vw->table), vw->toolbar2, 1, 2, 0, 3, GTK_FILL, GTK_FILL, 0, 0);
		DBG("Rebuild OK\n");
	}
}

void view_refresh_menu(void)
{
	if ( !(view_menu) || !sync_menu( view_menu, VCL_VIEW_MENU ) ){
		struct cmddata cd = { NULL, NULL };
		DBG("View menu needs rebuilding!\n");
		if ( view_menu ) gtk_object_sink( GTK_OBJECT(view_menu) );
		view_menu = create_menu( VCL_VIEW_MENU, cd );
		DBG("Rebuild OK\n");
	}
}
/** Show or hide the transform tool according to the current factory state
  in the given view */
void transform_update( struct view *vw )
{
	if ( factory.state == &ofs_select || factory.state == &ofs_transform_x )
	{
		if (!(vw->transform_state & TDM_SHAPE_MASK))
		{ vw->transform_state = TDM_VISIBLE; transform_recalc(vw); }
	}
	else
	{
		if ( vw->transform_state & TDM_SHAPE_MASK )
		{ transform_hide(vw); vw->transform_state = TDM_HIDDEN; }
	}

	if ( factory.state == &ofs_rehang_1 || factory.state == &ofs_rehang_2 )
	{
		visualisation_set_show_anchors_mode( &vw->vis, VIS_SHOW_SELECTED, NULL );
		visualisation_set_show_hangers_mode( &vw->vis, VIS_SHOW_ALL, NULL );
	}
	else
	{
		struct go * g; //show anchors for the currently craeted go
		if ( factory.tlo == vw->tlo && (g = factory_get_created_go()))
			visualisation_set_show_anchors_mode( &vw->vis, VIS_SHOW_SPECIFIC, g );
		else 	visualisation_set_show_anchors_mode( &vw->vis, VIS_SHOW_NONE, NULL );
		visualisation_set_show_hangers_mode( &vw->vis, VIS_SHOW_NONE, NULL );
	}
}
static void transform_hide( struct view * vw )
{
	int i;
	for ( i=0; i<=22; i++ ) visualisation_set_gadget_visible( &vw->vis, i, 0 );
}
/** Recalculate the transform tool coordinates of the center and axis point and
 update the gadgets in the view */
static void transform_recalc( struct view * vw )
{
	if ( vw->transform_state == TDM_HIDDEN ) return;

	struct geom_rectangle bbox;
	uns new_mode;
	visualisation_get_selected_bbox( &vw->vis, &bbox );
	if ( bbox.ll.x > bbox.ur.x ) //empty bbox
	{
		vw->transform_state = TDM_VISIBLE;
		transform_hide(vw);
		return;
	}
	else new_mode = TDM_NONEMPTY;
	if ( bbox.ur.x - bbox.ll.x > vw->tolerance ) new_mode |= TDM_WIDE;
	if ( bbox.ur.y - bbox.ll.y > vw->tolerance ) new_mode |= TDM_HIGH;

	vw->transform_origin.x = (bbox.ll.x + bbox.ur.x)/2;
	vw->transform_origin.y = (bbox.ll.y + bbox.ur.y)/2;
	vw->axis.x = 2 * bbox.ur.x - vw->transform_origin.x;
	vw->axis.y = vw->transform_origin.y;
	visualisation_set_center_gadget( &vw->vis, vw->transform_origin.x, vw->transform_origin.y );
	visualisation_set_xaxis_gadget( &vw->vis, vw->axis.x, vw->axis.y );
	if ( (new_mode & TDM_SHAPE_MASK) != (vw->transform_state & TDM_SHAPE_MASK) )
	{
		vw->transform_state = vw->transform_state & TDM_SHIFT | new_mode;
		transform_redisplay(vw);
	}
}
/** Show the gadgets of the transform tool in the particular mode according to vw->transform_state */
static void transform_redisplay( struct view * vw )
{
	int i = (vw->transform_state & TDM_SHIFT) ? 0 : 1;
	for ( ; i<=17; i+=2 ) visualisation_set_gadget_visible( &vw->vis, i, 0 );

	i = (vw->transform_state & TDM_SHIFT) ? 1 : 0;
	visualisation_set_gadget_visible( &vw->vis, 8, vw->transform_state & TDM_NONEMPTY );
	visualisation_set_gadget_visible( &vw->vis, 18, vw->transform_state & TDM_NONEMPTY ); // the center points

	visualisation_set_gadget_visible( &vw->vis, i, vw->transform_state & TDM_WIDE ); // two opposite corners
	visualisation_set_gadget_visible( &vw->vis, i + 16, vw->transform_state & TDM_WIDE );

	visualisation_set_gadget_visible( &vw->vis, i + 4, vw->transform_state & TDM_HIGH ); // the other two opposite corners
	visualisation_set_gadget_visible( &vw->vis, i + 12, vw->transform_state & TDM_HIGH );

	visualisation_set_gadget_visible( &vw->vis, i + 2, (vw->transform_state & TDM_RECT) == TDM_RECT ); // the edge midpoints
	visualisation_set_gadget_visible( &vw->vis, i + 6, (vw->transform_state & TDM_RECT) == TDM_RECT );
	visualisation_set_gadget_visible( &vw->vis, i + 10, (vw->transform_state & TDM_RECT) == TDM_RECT );
	visualisation_set_gadget_visible( &vw->vis, i + 14, (vw->transform_state & TDM_RECT) == TDM_RECT );

	int axis_visible = ((vw->transform_state & TDM_RECT) == TDM_RECT) && (vw->transform_state & TDM_SHIFT);
	visualisation_set_gadget_visible( &vw->vis, 19, axis_visible );
	visualisation_set_gadget_visible( &vw->vis, 21, axis_visible );
}
/* VIEW NAVIGATOR WINDOW */

static void destroy_view_navigator( GtkWidget * w UNUSED, struct view_navigator * vwn)
{
	visualisation_finish(&vwn->vis);
	vwn->parent_view->navigator = NULL;
	xfree(vwn);
}
static void create_view_navigator(struct view * vw)
{
	GtkWidget * table, * hbutbox, * button;
	struct view_navigator * vwn = xmalloc_zero(sizeof(struct view_navigator));
	vwn->parent_view = vw;
	vw->navigator = vwn;
	vwn->box = NULL;

	vwn->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	set_window_title( GTK_WINDOW(vwn->window), vw->tlo );
	gtk_window_add_accel_group( GTK_WINDOW(vwn->window), get_command_accel_group() );
	g_signal_connect( G_OBJECT(vwn->window), "destroy", G_CALLBACK(destroy_view_navigator), vwn);

	gtk_container_add( GTK_CONTAINER(vwn->window), table = gtk_table_new( 3, 4, FALSE ) );

	gtk_table_attach( GTK_TABLE(table), visualisation_init( &vwn->vis, vw->group, NULL, NULL ),
			0, 3, 0, 1, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 5, 5 );
	gtk_widget_set_size_request( GTK_WIDGET(vwn->vis.canvas->gtk->widget), 100, 100 );
	g_signal_connect_after( G_OBJECT(vwn->vis.canvas->gtk->widget), "size_allocate", G_CALLBACK(view_nav_resize), vwn);
	gtk_widget_add_events( vwn->vis.canvas->gtk->widget, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK );
	g_signal_connect( G_OBJECT( vwn->vis.canvas->gtk->widget), "motion-notify-event", G_CALLBACK(view_nav_motion), vwn);
	g_signal_connect( G_OBJECT( vwn->vis.canvas->gtk->widget), "button-press-event", G_CALLBACK(view_nav_button_press), vwn);
	g_signal_connect( G_OBJECT( vwn->vis.canvas->gtk->widget), "scroll-event", G_CALLBACK (scroll_callback), vw);

	gtk_table_attach( GTK_TABLE(table), gtk_label_new( "Zoom:"), 0, 1, 1, 2, 0, 0, 5, 5 );
	vwn->zoom_entry = GTK_SPIN_BUTTON(button = gtk_spin_button_new_with_range( 5, 1000, 50 ));
	g_object_set( G_OBJECT(vwn->zoom_entry), "climb-rate", 1.0, "digits", 1, "numeric", TRUE, "xalign", 1.0, NULL );
	gtk_table_attach( GTK_TABLE(table), button, 1, 2, 1, 2, GTK_FILL, 0, 5, 5 );
	g_signal_connect( G_OBJECT(vwn->zoom_entry), "value-changed", G_CALLBACK(view_nav_value_changed), vwn);
	gtk_table_attach( GTK_TABLE(table), gtk_label_new( "%"), 2, 3, 1, 2, 0, 0, 5, 5 );

	gtk_table_attach( GTK_TABLE(table), gtk_label_new( "Rotation:"), 0, 1, 2, 3, 0, 0, 5, 5 );
	vwn->rot_entry = GTK_SPIN_BUTTON(button = gtk_spin_button_new_with_range( -190, 190, 10 ));
	g_object_set( G_OBJECT(vwn->rot_entry), "digits", 1, "numeric", TRUE, "xalign", 1.0, NULL );
	gtk_table_attach( GTK_TABLE(table), button, 1, 2, 2, 3, GTK_FILL, 0, 5, 5 );
	g_signal_connect( G_OBJECT(vwn->rot_entry), "value-changed", G_CALLBACK(view_nav_value_changed), vwn);
	gtk_table_attach( GTK_TABLE(table), gtk_label_new( "deg"), 2, 3, 2, 3, 0, 0, 5, 5 );

	gtk_table_attach( GTK_TABLE(table), hbutbox = gtk_hbutton_box_new(), 0, 3, 3, 4, 0, 0, 5, 5 );
	gtk_box_pack_start( GTK_BOX(hbutbox), button = gtk_button_new_with_label("Reset"), FALSE, TRUE, 0 );
	g_signal_connect( G_OBJECT(button), "clicked", G_CALLBACK(view_zoom_reset), vw );

	visualisation_set_show_anchors_mode( &vwn->vis, VIS_SHOW_NONE, NULL );
	visualisation_set_show_hangers_mode( &vwn->vis, VIS_SHOW_NONE, NULL );
	
	change_in_progress = 1;
	gtk_widget_show_all( vwn->window );
	change_in_progress = 0;

	GdkCursor * cursor = gdk_cursor_new( GDK_FLEUR );
	gdk_window_set_cursor( vwn->vis.canvas->gtk->widget->window, cursor );
	gdk_cursor_unref( cursor );

	//get parent zoom and rotation
	struct geom_point ux, zero;
	ux.x = 1;
	ux.y = zero.x = zero.y = 0;
	visualisation_coords_to_vcoords( &vw->vis, &ux, &ux );
	visualisation_coords_to_vcoords( &vw->vis, &zero, &zero );
	view_nav_update( vwn, hypot( ux.x - zero.x , ux.y - zero.y ) * 100.0, //zoom
		(real)atan2( (double)(ux.y - zero.y), (double)(ux.x - zero.x) ) / R_PI * 180.0 ); //rotation
}

void view_nav_redraw( struct view_navigator * vwn )
{
	struct geom_rectangle t;
	struct geom_point a, b, c;
	GtkWidget * visw = vwn->parent_view->vis.canvas->gtk->widget;
	int width = visw->allocation.width;
	int height = visw->allocation.height;

	visualisation_get_visible_bbox( &vwn->vis, &t );
	if ( t.ur.x > t.ll.x || t.ur.y > t.ll.y ) visualisation_fit_box( &vwn->vis, t );

	a.x = a.y = 0;
	b.x = width; b.y = 0;
	c.x = width; c.y = height;
	visualisation_pixels_to_coords( &vwn->parent_view->vis, &a, &a );
	visualisation_pixels_to_coords( &vwn->parent_view->vis, &b, &b );
	visualisation_pixels_to_coords( &vwn->parent_view->vis, &c, &c );

	visualisation_gui_rectangle_update( &vwn->vis, &vwn->box, a, b, c );
}
void view_nav_value_changed( GtkWidget * w UNUSED, struct view_navigator * vwn UNUSED )
{
	if ( change_in_progress ) return;

	real zoom = gtk_spin_button_get_value_as_float(vwn->zoom_entry);
	real rot = gtk_spin_button_get_value_as_float(vwn->rot_entry);
	//this is necessary so that it does not diverge due to numeric errors
	if ( fabs( zoom - vwn->zoom ) < 0.01 && fabs( rot - vwn->rotation ) < 0.01 ) return;

	struct geom_point center = { 0.0, 0.0 };
	visualisation_vcoords_to_coords( &vwn->parent_view->vis, &center, &center );
	zoom /= 100.0; rot = rot * R_PI / 180.0;
	DBG("Setting zoom = %.3f, rotation = %.3f\n", zoom, rot );
	visualisation_set_orientation( &vwn->parent_view->vis, center.x, center.y, zoom, zoom, rot );
}
void view_nav_update( struct view_navigator * vwn, real zoom, real rotation )
{
	change_in_progress = 1;
	gtk_spin_button_set_value( vwn->zoom_entry, vwn->zoom = zoom );
	gtk_spin_button_set_value( vwn->rot_entry, vwn->rotation = rotation );
	change_in_progress = 0;
}
void view_nav_resize( GtkWidget *w UNUSED, GdkRectangle *a UNUSED, struct view_navigator *vwn )
{ view_nav_redraw( vwn ); }

static gboolean view_nav_motion(GtkWidget *widget UNUSED, GdkEventMotion *event, struct view_navigator *vwn)
{
	if ( moving_mode == MOVE_PANNING_NAV && moving.first_view == vwn->parent_view )
	{
		moving.last_pix_x = event->x; moving.last_pix_y = event->y;
		moving.last_view = vwn->parent_view;
		moving.last_time = event->time;
		moving.events_pending = 1;
	}
	return TRUE;
}
static gboolean view_nav_button_press(GtkWidget *widget UNUSED, GdkEventButton *event, struct view_navigator *vwn)
{
        struct geom_point pi = { event->x, event->y };
	struct geom_point po;
	visualisation_pixels_to_coords( &vwn->vis, &pi, &po );

	if ( event->state & GDK_CONTROL_MASK ) visualisation_center( &vwn->parent_view->vis, po.x, po.y );
	else
	{
		visualisation_coords_to_vcoords( &vwn->parent_view->vis, &po, &moving.last_vcoords );
		view_moving_start( pi, vwn->parent_view, MOVE_PANNING_NAV, GDK_FLEUR );
	}
	return TRUE;
}
