#include "lib/lib.h"
#include <gtk/gtk.h>
#include "gui/viewsettings.h"
#include "gui/ruler.h"
#include "gui/main.h"

static void destroy (GtkWidget *widget UNUSED, gpointer data)
{
	xfree(data);
}

static void rulers_set_zoom(struct view *lw UNUSED, int zoom UNUSED)
{
  //	vert_ruler_set_zoom(RULER (lw->vruler), zoom);
  //	horiz_ruler_set_zoom(RULER (lw->hruler), zoom);
}

static void spin_rotation_changed(GtkWidget *spin, struct view_settings *vs)
{
	int rot = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spin));
	while(rot < 0)
		rot += 360;
	while(rot > 359)
		rot -= 360;

	*(vs->rotation) = rot;
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin), rot);
}

static void button_apply_vs_pressed(GtkWidget *widget, struct view_settings *vs)
{
	*(vs->zoom) = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(vs->spin_zoom));
	rulers_set_zoom(vs->lw, *(vs->zoom));

	*(vs->value) = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(vs->spin_value));

	*(vs->rotation) = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(vs->spin_rotation));

	gtk_widget_destroy(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW));
	xfree(vs);
}

static void button_cancel_vs_pressed(GtkWidget *widget, gpointer data)
{
	xfree(data);
	gtk_widget_destroy(gtk_widget_get_ancestor(widget, GTK_TYPE_WINDOW));
}

struct view_settings *create_view_settings(struct view *lw)
{
  GtkWidget *table;
  GtkWidget *label_percent_zoom;
  GtkWidget *label_percent_value;
  GtkWidget *label_degrees;
  GtkWidget *label_zoom;
  GtkWidget *label_value;
  GtkWidget *label_rotation;
  GtkWidget *hbox;
  GtkWidget *button_apply;
  GtkWidget *button_cancel;
  GtkTooltips *tooltips;
  struct view_settings *vs = xmalloc(sizeof(struct view_settings));

  vs->lw = lw;
  vs->zoom = 0; // &(vs->lw->zoom);
  vs->value = 0; // &(vs->lw->value);
    vs->rotation = 0; // &(vs->lw->rotation);

  tooltips = gtk_tooltips_new ();

  vs->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (vs->window), "View settings");
  gtk_window_set_position (GTK_WINDOW (vs->window), GTK_WIN_POS_MOUSE);
  gtk_window_set_resizable (GTK_WINDOW (vs->window), FALSE);
  gtk_window_set_destroy_with_parent (GTK_WINDOW (vs->window), TRUE);

  table = gtk_table_new (4, 3, FALSE);
  gtk_widget_show (table);
  gtk_container_add (GTK_CONTAINER (vs->window), table);
  gtk_container_set_border_width (GTK_CONTAINER (table), 11);
  gtk_table_set_row_spacings (GTK_TABLE (table), 9);
  gtk_table_set_col_spacings (GTK_TABLE (table), 14);

  label_percent_zoom = gtk_label_new ("%");
  gtk_widget_show (label_percent_zoom);
  gtk_table_attach (GTK_TABLE (table), label_percent_zoom, 2, 3, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label_percent_zoom), GTK_JUSTIFY_LEFT);
  gtk_misc_set_alignment (GTK_MISC (label_percent_zoom), 0, 0.5);

  label_percent_value = gtk_label_new ("%");
  gtk_widget_show (label_percent_value);
  gtk_table_attach (GTK_TABLE (table), label_percent_value, 2, 3, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label_percent_value), GTK_JUSTIFY_LEFT);
  gtk_misc_set_alignment (GTK_MISC (label_percent_value), 0, 0.5);

  label_degrees = gtk_label_new ("degrees");
  gtk_widget_show (label_degrees);
  gtk_table_attach (GTK_TABLE (table), label_degrees, 2, 3, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label_degrees), GTK_JUSTIFY_LEFT);
  gtk_misc_set_alignment (GTK_MISC (label_degrees), 0, 0.5);

  label_zoom = gtk_label_new ("Zoom:");
  gtk_widget_show (label_zoom);
  gtk_table_attach (GTK_TABLE (table), label_zoom, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label_zoom), GTK_JUSTIFY_LEFT);
  gtk_misc_set_alignment (GTK_MISC (label_zoom), 0, 0.5);

  label_value = gtk_label_new ("Value:");
  gtk_widget_show (label_value);
  gtk_table_attach (GTK_TABLE (table), label_value, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label_value), GTK_JUSTIFY_LEFT);
  gtk_misc_set_alignment (GTK_MISC (label_value), 0, 0.5);

  label_rotation = gtk_label_new ("Rotation:");
  gtk_widget_show (label_rotation);
  gtk_table_attach (GTK_TABLE (table), label_rotation, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_label_set_justify (GTK_LABEL (label_rotation), GTK_JUSTIFY_LEFT);
  gtk_misc_set_alignment (GTK_MISC (label_rotation), 0, 0.5);

  vs->spin_zoom_adj = gtk_adjustment_new (*(vs->zoom), 20, 1000, 10, 25, 25);
  vs->spin_zoom = gtk_spin_button_new (GTK_ADJUSTMENT (vs->spin_zoom_adj), 10, 0);
  gtk_widget_show (vs->spin_zoom);
  gtk_table_attach (GTK_TABLE (table), vs->spin_zoom, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_widget_set_size_request (vs->spin_zoom, 70, -1);
  gtk_tooltips_set_tip (tooltips, vs->spin_zoom, "Sets zoom of the view in percents.", NULL);
  gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (vs->spin_zoom), GTK_UPDATE_IF_VALID);
  gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (vs->spin_zoom), TRUE);

  vs->spin_value_adj = gtk_adjustment_new (*(vs->value), 1, 200, 1, 10, 10);
  vs->spin_value = gtk_spin_button_new (GTK_ADJUSTMENT (vs->spin_value_adj), 1, 0);
  gtk_widget_show (vs->spin_value);
  gtk_table_attach (GTK_TABLE (table), vs->spin_value, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_widget_set_size_request (vs->spin_value, 70, -1);
  gtk_tooltips_set_tip (tooltips, vs->spin_value, "Changes the value of view.", NULL);
  gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (vs->spin_value), GTK_UPDATE_IF_VALID);
  gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (vs->spin_value), TRUE);

  vs->spin_rotation_adj = gtk_adjustment_new (*(vs->rotation), -1000, 1000, 90, 90, 90);
  vs->spin_rotation = gtk_spin_button_new (GTK_ADJUSTMENT (vs->spin_rotation_adj), 90, 0);
  gtk_widget_show (vs->spin_rotation);
  gtk_table_attach (GTK_TABLE (table), vs->spin_rotation, 1, 2, 2, 3,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_widget_set_size_request (vs->spin_rotation, 70, -1);
  gtk_tooltips_set_tip (tooltips, vs->spin_rotation, "Sets rotation of the view. The value is in degrees.", NULL);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox);
  gtk_table_attach (GTK_TABLE (table), hbox, 0, 3, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 5);

  button_apply = gtk_button_new_from_stock ("gtk-apply");
  gtk_widget_show (button_apply);
  gtk_box_pack_start (GTK_BOX (hbox), button_apply, TRUE, FALSE, 0);
  gtk_tooltips_set_tip (tooltips, button_apply, "Save changes and close the window.", NULL);

  button_cancel = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (button_cancel);
  gtk_box_pack_start (GTK_BOX (hbox), button_cancel, TRUE, FALSE, 0);
  gtk_tooltips_set_tip (tooltips, button_cancel, "Close window without saving changes.", NULL);

  g_signal_connect (G_OBJECT (vs->spin_rotation), "value-changed",
		  G_CALLBACK(spin_rotation_changed), vs);
  g_signal_connect (G_OBJECT (button_apply), "clicked",
		  G_CALLBACK(button_apply_vs_pressed), vs);
  g_signal_connect (G_OBJECT (button_cancel), "clicked",
		  G_CALLBACK(button_cancel_vs_pressed), vs);
  g_signal_connect (G_OBJECT(vs->window), "destroy",
		  G_CALLBACK(destroy), 0);

  gtk_widget_grab_focus (button_apply);
  gtk_widget_show_all(vs->window);

  return vs;
}

