#undef LOCAL_DEBUG
#include "lib/lib.h"

#include <stdio.h>
#include <stdlib.h>

#include "kernel/kernel.h"
#include "kernel/property.h"
#include "kernel/go.h"
#include "gui/main.h"
#include "gui/visualisation.h"

#include "vcl/vcl.h"
#include "geomlib/geomlib.h"
#include "font/font.h"
#include "import/import.h"

static void move_trans (struct vcl_affinity *vt, real x, real y);

static struct  vcl_property *
make_line_style (UNUSED uns color)
{
  struct vcl_property * prop = vcl_property_new ();
  vcl_property_set_stroke_color (prop, color);
  return (prop);
}

static inline int geom_rtree_box_equal(struct geom_rectangle *r1, struct geom_rectangle *r2)
{
	return (r1->ll.x == r2->ll.x && r1->ll.y == r2->ll.y &&
			r1->ur.x == r2->ur.x && r1->ur.y == r2->ur.y);
}

static inline void
update_ha_show (struct visualisation *v)
{ vcl_le_area_flush (v -> ha_show); }


// Hooks for VCL le_area

static inline void call_changed_cb (struct visualisation *v)
{ if (v -> changed_cb) v -> changed_cb (v -> changed_data); }


static void *
make_anchor_image (struct geom_point c)
{
  real dx = 1.0;
  real dy = -0.5;
  real dy2 = -1.5;

  struct vcl_property *prop = make_line_style (rgba_uns2uns (0, 128, 0, 255));
  struct vcl_path * path = vcl_path_new (NULL);

  struct geom_bezier bz;
  struct geom_path * p = &(path -> embedded);
  GEOM_ITEM_CREATE (p, path);
  GEOM_ITEM_CREATE (&bz, bezier);
  geom_bezier_fill_linear_nonrational (&bz, c . x, c . y - dy2, c . x + dx, c . y + dy);
  geom_path_append_dup (p, &bz);
  geom_bezier_fill_linear_nonrational (&bz, c . x + dx, c . y + dy, c . x - dx, c . y + dy);
  geom_path_append_dup (p, &bz);
  geom_bezier_fill_linear_nonrational (&bz, c . x - dx, c . y + dy, c . x, c . y - dy2);
  geom_path_append_dup (p, &bz);
  geom_bezier_destroy (&bz);
  vcl_path_source_changed (path);

  vcl_enclosure_set_child (prop, path);
  return (prop);
}

static void *
make_hanger_image (struct geom_point c)
{
  real dx = 1.0;
  real dy = 0.5;
  real dy2 = 1.5;

  struct vcl_property *prop = make_line_style (rgba_uns2uns (0, 0, 255, 255));
  struct vcl_path * path = vcl_path_new (NULL);

  struct geom_bezier bz;
  struct geom_path * p;
  p = &(path -> embedded);
  GEOM_ITEM_CREATE (p, path);
  GEOM_ITEM_CREATE (&bz, bezier);
  geom_bezier_fill_linear_nonrational (&bz, c . x, c . y - dy2, c . x + dx, c . y + dy);
  geom_path_append_dup (p, &bz);
  geom_bezier_fill_linear_nonrational (&bz, c . x + dx, c . y + dy, c . x - dx, c . y + dy);
  geom_path_append_dup (p, &bz);
  geom_bezier_fill_linear_nonrational (&bz, c . x - dx, c . y + dy, c . x, c . y - dy2);
  geom_path_append_dup (p, &bz);
  geom_bezier_destroy (&bz);
  vcl_path_source_changed (path);

  vcl_enclosure_set_child (prop, path);
  return (prop);
}

static uns
show_p (uns mode, struct go * stored, struct go *actual)
{
    switch (mode)
    {
	case VIS_SHOW_NONE:
	    return (0);
	case VIS_SHOW_ALL:
	    return (1);
	case VIS_SHOW_SELECTED:
	    return (go_is_selected (actual));
	case VIS_SHOW_SPECIFIC:
	    return (actual == stored);
        default:
	    BUG();
    }
    return (0);
}

static inline uns
show_anchors_p (struct visualisation *v, struct go * go)
{ return show_p (v -> anchor_mode, v -> anchor_specific, go); }

static inline uns
show_hangers_p (struct visualisation *v, struct go * go)
{ return show_p (v -> hanger_mode, v -> hanger_specific, go); }


static void
bbox_of_child (void * child_id UNUSED, struct vcl_le_area * questioner UNUSED,
	       struct geom_rectangle *bbox, void * data UNUSED)
{
  struct go * go = child_id;
  const struct geom_rectangle *r = go_get_expand_bbox (go_get_tlo(go), go);

  if (r) *bbox = *r;
  else geom_rectangle_set_empty (bbox);
  /* geom_rectangle_dump (bbox); */
}

static void *
expand_child (void * child_id, struct vcl_le_area * questioner UNUSED, void * data UNUSED)
{
  struct go * go = (struct go *) child_id;
  struct visualisation *v = data;
  struct vcl_property * prop = vcl_property_new ();
  void * inner;
  void * obj;

  switch (go -> type)
    {
    case GOT_TEXT:
      {
	struct go_text * gotxt = (struct go_text *) child_id;

	struct geom_transform2 t;
	go_text_get_transformation(gotxt, &t.primary);
	geom_transform_invert (&(t . primary), &(t . inverted));

	obj = vcl_affinity_new ();
	vcl_affinity_set (obj, &t);
	inner = vcl_text_layout_new (1, strdup (gotxt -> text -> text),
				      gotxt ->font_id, gotxt -> font_size, NULL);

	vcl_enclosure_set_child (obj, inner);
	break;
      }
    case GOT_TEX_TEXT:
      {
	struct go_tex_text * gotxt = (struct go_tex_text *) child_id;

	uns size = gotxt -> process -> glyphs_cnt * sizeof (struct tex_glyph);
	struct tex_glyph *gl = xmalloc (size);
	memcpy (gl, gotxt -> process -> glyphs -> data, size);

	struct geom_transform2 t;
	go_tex_text_get_transformation(gotxt, &t.primary);
	geom_transform_invert (&(t . primary), &(t . inverted));

	obj = vcl_affinity_new ();
	vcl_affinity_set (obj, &t);
	inner = vcl_tex_layout_new (1, gl, gotxt -> process -> glyphs_cnt, &(gotxt -> process -> bbox));
	vcl_enclosure_set_child (obj, inner);
	break;
      }
    case GOT_POINT:
      {
	struct go_point * go = (struct go_point *) child_id;
	obj = vcl_segment_new (hanger_get_position (go -> a [0] . h),
			       hanger_get_position (go -> a [0] . h));
	break;
      }
    case GOT_SEGMENT:
      {
	struct go_segment * go = (struct go_segment *) child_id;
	obj = vcl_segment_new (hanger_get_position (go -> a [0] . h),
				  hanger_get_position (go -> a [1] . h));
	break;
      }
    case GOT_BEZIER:
      {
	struct go_bezier * go = (struct go_bezier *) child_id;
	struct vcl_path * p = vcl_path_new (NULL);

	geom_group_add_head (&(p -> embedded), geom_item_duplicate(&(go -> geom)));
	obj = p;
	break;
      }
    default:
      {
	struct vcl_path * p = vcl_path_new (NULL);

	// const struct geom_rectangle *bbox = go_get_expand_bbox (go_get_tlo(go), go);
	//	const struct geom_path *path = go_path_expand_ref (go_get_tlo (go), go);
	const struct geom_path *path = go_get_expand_path (go_get_tlo (go), go);
	if (path != NULL) geom_path_copy (&(p -> embedded), (void *)path);
	//	vcl_path_source_changed (p);

	obj = p;
	break;
      }
    }

  vcl_enclosure_set_child (prop, obj);
  
  struct prop *c;
  
  u32 fc, sc;
  real sw;
  
  fc = v -> default_fill_color;
  sc = v -> default_stroke_color;
  sw = v -> default_stroke_width;
  
  c = prop_find (go2o(go), string_fill_color);
  if (c && (c -> type == PT_UNS) && (c -> subtype == PTU_COLOR))
    fc = prop_get_value ((struct o *) go, c) . u;

  c = prop_find (go2o(go), string_stroke_color);
  if (c && (c -> type == PT_UNS) && (c -> subtype == PTU_COLOR))
    sc = prop_get_value ((struct o *) go, c) . u;

  c = prop_find (go2o(go), string_stroke_width);
  if (c && (c -> type == PT_REAL) && (c -> subtype == PTR_NON_NEGATIVE))
    sw = prop_get_value ((struct o *) go, c) . r;
    
  if (go_is_selected (go))
    {
      fc = rgba_uns_set_a_uns (v -> selected_fill_color,   uns2a_uns (fc));
      sc = rgba_uns_set_a_uns (v -> selected_stroke_color, uns2a_uns (sc));
      sw = sw + v -> selected_stroke_addition;
    }
  
  vcl_property_set_fill_color (prop, fc);
  vcl_property_set_stroke_color (prop, sc);
  vcl_property_set_stroke_width (prop, sw);

  return (prop);
}

static void
get_children (struct vcl_le_area * questioner, UNUSED uns backwards,
	      UNUSED const struct geom_rectangle * bbox, void * data)
{
  struct visualisation *v = data;
  struct go * go;

  DBG ("CHILDREN IN BBOX Q %f %f %f %f\n", bbox -> ll . x, bbox -> ll . y,
       bbox -> ur . x, bbox -> ur . y);

  WALK_GO_RECURSIVE_WITHOUT_ROOT ((struct go *) (v -> grp), go)
    { //  FIXME should be returned only if it is in bbox
      vcl_le_area_answer_child_in_bbox (questioner, go);
    }
}

static void
ha_bbox_of_child (void * child_id, struct vcl_le_area * questioner UNUSED,
	       struct geom_rectangle *bbox, void * data)
{
  struct go * go = child_id;
  const struct geom_rectangle *r = go_get_expand_bbox (go_get_tlo(go), go);
  struct visualisation *v = data;

  // coords to vcoords transformation
  if (r)
  {
    geom_rectangle_transform (r, &(vcl_affinity_get (v -> zoom) -> primary), bbox);
    geom_rectangle_enlarge (bbox, 5, bbox);
  }
  else geom_rectangle_set_empty (bbox);
}

static void *
ha_expand_child (void * child_id, struct vcl_le_area * questioner UNUSED, void * data)
{
  struct visualisation *v = data;

  struct go * go = (struct go *) child_id;
  void *retval = NULL;

  int sa = show_anchors_p(v, go);
  int sh = show_hangers_p(v, go);

  if (sa || sh)
  {
    struct vcl_group * g = vcl_group_new ();

    if (sa)
    {
      struct anchor * a;
      int i;

      WALK_GO_ANCHOR( go, a, i )
	{
	  if (!hanger_is_position(a->h)) continue;
	  struct geom_point pt = hanger_get_position(a->h);
	  visualisation_coords_to_vcoords (v, &pt, &pt);
	  vcl_placement_add_child_on_top (g, make_anchor_image (pt));
	}
    }

    if (sh)
    {
        struct hanger *h;

        WALK_GO_HANGER_BEGIN (go, h)
	{
	    if (!hanger_is_position(h)) continue;
	    struct geom_point pt = hanger_get_position(h);
	    visualisation_coords_to_vcoords (v, &pt, &pt);
	    vcl_placement_add_child_on_top (g, make_hanger_image (pt));
	}
	WALK_GO_HANGER_END;
    }

    retval = g;
  }

  return (retval);
}

static void
ha_get_children (struct vcl_le_area * questioner, UNUSED uns backwards,
		      UNUSED const struct geom_rectangle * bbox, void * data)
{
  struct visualisation * v = data;

  if ((v -> anchor_mode == VIS_SHOW_ALL) ||
      (v -> anchor_mode == VIS_SHOW_SELECTED) ||
      (v -> hanger_mode == VIS_SHOW_ALL) ||
      (v -> hanger_mode == VIS_SHOW_SELECTED))
  {
    struct go * go;
    // FIXME in final version transform bbox from vcoords to coords and enlarge
    WALK_GO_RECURSIVE_WITHOUT_ROOT ((struct go *) (v -> grp), go)
    { //  FIXME should be returned only if it is in bbox
      vcl_le_area_answer_child_in_bbox (questioner, go);
    }
  }
  else
  {
    if (v -> anchor_specific) vcl_le_area_answer_child_in_bbox (questioner, v -> anchor_specific);
    if (v -> hanger_specific) vcl_le_area_answer_child_in_bbox (questioner, v -> hanger_specific);
  }
}

// Supporting function

static void
recompute_selected (struct go_group * grp, struct geom_rectangle *rect)
{
  struct go * go;
  struct obj_tlo *tlo = go_get_tlo(&grp->go);

  geom_rectangle_set_empty (rect);

  // FIXME any checks for return value?
  WALK_GROUP_SELECTED_GO (grp, go)
    geom_rectangle_merge (go_get_expand_bbox(tlo, go), rect, rect);
}

static void
update_selector (struct visualisation *v, struct geom_rectangle *old_bbox)
{
  struct geom_rectangle bbox;
  struct geom_path *p = &(v -> selector -> embedded);
  struct geom_bezier b;

  geom_rectangle_transform (&(v -> selected_bbox), &(vcl_affinity_get (v -> zoom) -> primary), &bbox);

  DBG ("SELECTOR %f %f %f %f\n", bbox . ll . x, bbox . ll . y,
       bbox . ur . x, bbox . ur . y);

  if (! geom_rectangle_is_empty (old_bbox))
    {
      v -> flags &= ~VIS_FLAGS_SELECTOR;
      vcl_placement_remove_child (v -> root, vcl_node_get_parent (v -> selector));
      int x = 1;
      for (int i = 0; i < 23; i ++)
	{
	  if (v -> gadgets & x)
	    vcl_placement_remove_child (v -> root, v -> transform2_imgs [i]);
	  x = x << 1;
	}
    }

  geom_path_clear (p);

  GEOM_ITEM_CREATE (&b, bezier);
  geom_bezier_fill_linear_nonrational (&b, bbox . ll . x, bbox . ll . y,
				       bbox . ur . x, bbox . ll . y);
  geom_path_append_dup (p, &b);
  geom_bezier_fill_linear_nonrational (&b, bbox . ur . x, bbox . ll . y,
				       bbox . ur . x, bbox . ur . y);
  geom_path_append_dup (p, &b);
  geom_bezier_fill_linear_nonrational (&b, bbox . ur . x, bbox . ur . y,
				       bbox . ll . x, bbox . ur . y);
  geom_path_append_dup (p, &b);
  geom_bezier_fill_linear_nonrational (&b, bbox . ll . x, bbox . ur . y,
				       bbox . ll . x, bbox . ll . y);
  geom_path_append_dup (p, &b);
  geom_bezier_destroy(&b);

  struct geom_point center;
  center . x = (bbox . ll . x + bbox . ur . x) / 2;
  center . y = (bbox . ll . y + bbox . ur . y) / 2;

  move_trans (v -> transform2_imgs [0], bbox . ll . x, bbox . ll . y);
  move_trans (v -> transform2_imgs [1], bbox . ll . x, bbox . ll . y);
  move_trans (v -> transform2_imgs [2], center . x, bbox . ll . y);
  move_trans (v -> transform2_imgs [3], center . x, bbox . ll . y);
  move_trans (v -> transform2_imgs [4], bbox . ur . x, bbox . ll . y);
  move_trans (v -> transform2_imgs [5], bbox . ur . x, bbox . ll . y);

  move_trans (v -> transform2_imgs [6], bbox . ll . x, center. y);
  move_trans (v -> transform2_imgs [7], bbox . ll . x, center. y);
  move_trans (v -> transform2_imgs [8], center . x, center. y);
  move_trans (v -> transform2_imgs [9], center . x, center. y);
  move_trans (v -> transform2_imgs [10], bbox . ur . x, center . y);
  move_trans (v -> transform2_imgs [11], bbox . ur . x, center . y);

  move_trans (v -> transform2_imgs [12], bbox . ll . x, bbox . ur . y);
  move_trans (v -> transform2_imgs [13], bbox . ll . x, bbox . ur . y);
  move_trans (v -> transform2_imgs [14], center . x, bbox . ur . y);
  move_trans (v -> transform2_imgs [15], center . x, bbox . ur . y);
  move_trans (v -> transform2_imgs [16], bbox . ur . x, bbox . ur . y);
  move_trans (v -> transform2_imgs [17], bbox . ur . x, bbox . ur . y);

  vcl_path_source_changed (v -> selector);

  if (! geom_rectangle_is_empty (&bbox))
    {
      v -> flags |= VIS_FLAGS_SELECTOR;
      vcl_placement_add_child_on_top (v  -> root, vcl_node_get_parent (v -> selector));
      int x = 1;
      for (int i = 0; i < 23; i ++)
	{
	  if (v -> gadgets & x)
	    vcl_placement_add_child_on_top (v -> root, v -> transform2_imgs [i]);
	  x = x << 1;
	}
    }

}

static void
update_axis (struct visualisation *v)
{
  struct geom_point c;
  struct geom_point x;
  struct geom_point y;
  struct geom_bezier bz;
  struct geom_path *p;

  visualisation_coords_to_vcoords (v, &(v -> g_center), &c);
  visualisation_coords_to_vcoords (v, &(v -> g_xaxis), &x);
  visualisation_coords_to_vcoords (v, &(v -> g_yaxis), &y);

  move_trans (v -> transform2_imgs [18], c . x, c . y);
  move_trans (v -> transform2_imgs [19], x . x, x . y);
  move_trans (v -> transform2_imgs [20], y . x, y . y);

  p = &(((struct vcl_path *) (vcl_enclosure_get_child (v -> transform2_imgs [21]))) -> embedded);
  geom_path_clear (p);
  GEOM_ITEM_CREATE (&bz, bezier);
  geom_bezier_fill_linear_nonrational (&bz, c . x, c . y, x . x, x . y);
  geom_path_append_dup (p, &bz);
  vcl_path_source_changed ((struct vcl_path *) (vcl_enclosure_get_child (v -> transform2_imgs [21])));

  p = &(((struct vcl_path *) (vcl_enclosure_get_child (v -> transform2_imgs [22]))) -> embedded);
  geom_path_clear (p);
  geom_bezier_fill_linear_nonrational (&bz, c . x, c . y, y . x, y . y);
  geom_path_append_dup (p, &bz);
  geom_bezier_destroy(&bz);
  vcl_path_source_changed ((struct vcl_path *) (vcl_enclosure_get_child (v -> transform2_imgs [22])));
}

// Kernel go hooks

static void go_unselected (struct o *o, struct go_hook* h);


static void
go_linked (struct o *o, struct go_hook* h)
{
  struct go * go = (struct go *) o;
  struct visualisation *v = ((struct visualisation *) (h -> data));
  struct geom_rectangle old_global_bbox, new_global_bbox;

  visualisation_get_visible_bbox (v, &old_global_bbox);

  vcl_le_area_child_appeared_notification (v -> area, go);
  vcl_le_area_child_appeared_notification (v -> ha_show, go);

  visualisation_get_visible_bbox (v, &new_global_bbox);
  if (! geom_rtree_box_equal (&old_global_bbox, &new_global_bbox)) call_changed_cb (v);
}

static void
go_unlinked (struct o *o, struct o *group UNUSED, struct o *prev UNUSED, struct go_hook* h)

{
  struct go * go = (struct go *) o;
  struct visualisation *v = ((struct visualisation *) (h -> data));
  struct geom_rectangle old_global_bbox, new_global_bbox;

  visualisation_get_visible_bbox (v, &old_global_bbox);

//go_unselected (o, h);
  vcl_le_area_child_disappeared_notification (v -> area, go);
  vcl_le_area_child_disappeared_notification (v -> ha_show, go);

  visualisation_get_visible_bbox (v, &new_global_bbox);
  if (! geom_rtree_box_equal (&old_global_bbox, &new_global_bbox)) call_changed_cb (v);
}

// FIXME - this probably should be done cleverly
static void
go_relinked (struct o *o, struct go_group *old_group, struct go * old_prev, struct go_hook* h)
{
  go_unlinked (o, (struct o*)old_group, (struct o*)old_prev, h);
  go_linked (o, h);
}

static void
go_selected (struct o *o, struct go_hook* h)
{
  struct go * go = (struct go *) o;
  struct visualisation *v = ((struct visualisation *) (h -> data));
  const struct geom_rectangle *bbox = go_get_expand_bbox(go_get_tlo(go), go);
  struct geom_rectangle old = v -> selected_bbox;
  if (go_get_group(go) != v->grp) return;

  geom_rectangle_merge (bbox, &(v -> selected_bbox), &(v -> selected_bbox));
  update_selector (v, &old);

  vcl_le_area_child_changed_notification (v -> area, go, bbox, bbox);

  if ((v -> hanger_mode == VIS_SHOW_SELECTED) || (v -> anchor_mode == VIS_SHOW_SELECTED))
    update_ha_show (v);
}

static void
go_unselected (struct o *o, struct go_hook* h)
{
  struct go * go = (struct go *) o;
  struct visualisation *v = ((struct visualisation *) (h -> data));
  const struct geom_rectangle *bbox = go_get_expand_bbox (go_get_tlo (go), go);
  if (go_get_group (go) != v->grp) return;

  if (bbox && geom_rectangle_is_adjoin (&(v -> selected_bbox), bbox))
    {
      struct geom_rectangle old = v -> selected_bbox;
      recompute_selected (v -> grp, &(v -> selected_bbox));
      update_selector (v, &old);
    }

  vcl_le_area_child_changed_notification (v -> area, go, bbox, bbox);
}

static void
go_changed (struct go *go, const struct geom_rectangle *old_bbox, const struct geom_rectangle *new_bbox, struct go_hook* h)
{
  struct visualisation *v = ((struct visualisation *) (h -> data));
  struct geom_rectangle old_global_bbox, new_global_bbox;
  if ( !o_is_ancestor(go2o(&v->grp->go), go2o(go)) ) return;

  visualisation_get_visible_bbox (v, &old_global_bbox);

  if ((go -> flags & GOF_SELECTED))
//    geom_rectangle_is_adjoin (&(v -> selected_bbox), old_bbox))
    {
      struct geom_rectangle old_selected = v -> selected_bbox;
      recompute_selected (v -> grp, &(v -> selected_bbox));
      update_selector (v, &old_selected);
    }


  vcl_le_area_child_changed_notification (v -> area, go, old_bbox, new_bbox);
  vcl_le_area_child_changed_notification (v -> ha_show, go, old_bbox, new_bbox);

  visualisation_get_visible_bbox (v, &new_global_bbox);
  if (! geom_rtree_box_equal (&old_global_bbox, &new_global_bbox)) call_changed_cb (v);
}

static void
go_altered (struct go *go, const struct geom_rectangle *where, struct go_hook* h)
{
  struct visualisation *v = ((struct visualisation *) (h -> data));
  // if ( go_get_group(go) != v->grp ) return;
  vcl_le_area_child_altered_notification (v -> area, go, where);
  vcl_le_area_child_altered_notification (v -> ha_show, go, where);
}

static void
go_transformed (struct go *go, const struct geom_rectangle *old_bbox, const struct geom_rectangle *new_bbox, const struct geom_transform2 *t, struct go_hook* h)
{
  struct visualisation *v = ((struct visualisation *) (h -> data));
  struct geom_rectangle old_global_bbox, new_global_bbox;
  visualisation_get_visible_bbox (v, &old_global_bbox);

  if ((go -> flags & GOF_SELECTED))
//      geom_rectangle_is_adjoin (&(v -> selected_bbox), old_bbox))
    {
      struct geom_rectangle old_selected = v -> selected_bbox;
      recompute_selected (v -> grp, &(v -> selected_bbox));
      update_selector (v, &old_selected);
    }

  vcl_le_area_child_transformed_notification (v -> area, go, old_bbox, new_bbox, t);
  vcl_le_area_child_changed_notification (v -> ha_show, go, old_bbox, new_bbox);

  visualisation_get_visible_bbox (v, &new_global_bbox);
  if (! geom_rtree_box_equal (&old_global_bbox, &new_global_bbox)) call_changed_cb (v);
}

static void
go_changed_dispatch (struct go *go, const struct changed_data_generic *d, struct go_hook* h)
{
  if (go -> type == GOT_GROUP) return;

  switch (d -> kind)
    {
    case CK_CHANGED:
      go_changed (go, ((const struct changed_data *) d) -> old_bbox,
		  ((const struct changed_data *) d) -> new_bbox, h);
      break;
    case CK_ALTERED:
      go_altered (go, ((const struct altered_data *) d) -> where, h);
      break;
    case CK_TRANSFORMED:
      go_transformed (go, ((const struct transformed_data *) d) -> old_bbox,
		      ((const struct transformed_data *) d) -> new_bbox,
		      ((const struct transformed_data *) d) -> transformation, h);
      break;
    }
}

// Functions

static void
move_trans_vc (struct vcl_affinity *vt, struct visualisation * v, struct geom_point dst)
{
  struct geom_point c;
  struct geom_transform2 t;
  visualisation_coords_to_vcoords (v, &dst, &c);
  geom_transform2_set_move (&t, c . x, c . y);
  vcl_affinity_set (vt, &t);
}

static void
move_trans (struct vcl_affinity *vt, real x, real y)
{
  struct geom_transform2 t;
  geom_transform2_set_move (&t, x, y);
  vcl_affinity_set (vt, &t);
}

static struct vcl_affinity *
make_cursor (uns r, uns g, uns b)
{
  struct vcl_affinity *tf = vcl_affinity_new ();
  struct vcl_group *gr = vcl_group_new ();
  struct vcl_property *prop1 = make_line_style (rgba_uns2uns (r, g, b, 255));
  struct vcl_path *path1 = vcl_path_new (NULL);
  struct vcl_property *prop2 = make_line_style (rgba_uns2uns (r, g, b, 255));
  struct vcl_path *path2 = vcl_path_new (NULL);

  struct geom_path *p;
  struct geom_bezier bz;

  GEOM_ITEM_CREATE (&bz, bezier);

  p = &(path1 -> embedded);
  GEOM_ITEM_CREATE (p, path);
  geom_bezier_fill_linear_nonrational (&bz, -1, -1, 1, 1);
  geom_path_append_dup (p, &bz);
  vcl_path_source_changed (path1);

  p = &(path2 -> embedded);
  GEOM_ITEM_CREATE(p, path);
  geom_bezier_fill_linear_nonrational (&bz, -1, 1, 1, -1);
  geom_path_append_dup (p, &bz);
  vcl_path_source_changed (path2);

  vcl_enclosure_set_child (prop1, path1);
  vcl_enclosure_set_child (prop2, path2);
  vcl_placement_add_child_on_top (gr, prop1);
  vcl_placement_add_child_on_top (gr, prop2);
  vcl_enclosure_set_child (tf, gr);

  geom_bezier_destroy(&bz);
  return (tf);
}

static struct vcl_affinity *
make_rect (real w, uns r, uns g, uns b)
{
  struct vcl_affinity *tf = vcl_affinity_new ();
  struct vcl_property *prop = make_line_style (rgba_uns2uns (r, g, b, 255));
  struct vcl_path *path = vcl_path_new (NULL);

  struct geom_path *p;
  struct geom_bezier bz;
  p = &(path -> embedded);
  GEOM_ITEM_CREATE (p, path);
  GEOM_ITEM_CREATE (&bz, bezier);
  geom_bezier_fill_linear_nonrational (&bz, -w, -w, -w, w);
  geom_path_append_dup (p, &bz);
  geom_bezier_fill_linear_nonrational (&bz, -w, w, w, w);
  geom_path_append_dup (p, &bz);
  geom_bezier_fill_linear_nonrational (&bz, w, w, w, -w);
  geom_path_append_dup (p, &bz);
  geom_bezier_fill_linear_nonrational (&bz, w, -w, -w, -w);
  geom_path_append_dup (p, &bz);
  vcl_path_source_changed (path);
  geom_bezier_destroy(&bz);

  vcl_enclosure_set_child (prop, path);
  vcl_enclosure_set_child (tf, prop);

  return (tf);
}

GtkWidget *
visualisation_init (struct visualisation * v, struct go_group * grp,
		    simple_cb changed_cb, void * changed_data)
{
  struct go * go;

  v -> flags = 0;
  v -> gadgets = 0;
  v -> grp = grp;

  go_ref ((struct go *) grp);
  expand_ref (go_get_tlo ((struct go *) grp));

  v -> default_stroke_color = rgba_uns2uns (0, 0, 0, 255);
  v -> default_fill_color = rgba_uns2uns (0, 0, 0, 0);
  v -> default_stroke_width = 0;

  v -> selected_stroke_color = rgba_uns2uns (255, 0, 0, 255);
  v -> selected_fill_color = rgba_uns2uns (255, 255, 0, 255);
  v -> selected_stroke_addition = 2;

  v -> g_center . x = v -> g_center . y = 0;
  v -> g_xaxis . x = v -> g_yaxis . y = 1;
  v -> g_xaxis . y = v -> g_yaxis . x = 0;

  v -> anchor_mode = VIS_SHOW_ALL;
  v -> hanger_mode = VIS_SHOW_ALL;
  // v -> anchor_mode = VIS_SHOW_NONE;
  // v -> hanger_mode = VIS_SHOW_NONE;
  v -> anchor_specific = NULL;
  v -> hanger_specific = NULL;

  v -> selector = vcl_path_new (NULL);
  struct vcl_property *prop = make_line_style (rgba_uns2uns (255, 0, 0, 255));

  vcl_enclosure_set_child (prop, v -> selector);

  v -> snap_img = make_cursor (0, 0, 255);
  v -> transform_imgs [0] = make_cursor (255, 0, 255);
  v -> transform_imgs [2] = make_cursor (0, 0, 255);
  v -> transform_imgs [4] = make_cursor (255, 0, 0);
  v -> transform_imgs [6] = make_cursor (0, 255, 0);
  v -> transform_imgs [1] = NULL;
  v -> transform_imgs [3] = make_cursor (143, 143, 143);
  v -> transform_imgs [5] = make_cursor (143, 143, 143);
  v -> transform_imgs [7] = make_cursor (143, 143, 143);

  v -> transform2_imgs [0] = make_rect (1, 255, 0, 0);
  v -> transform2_imgs [1] = make_rect (1, 255, 0, 255);
  v -> transform2_imgs [2] = make_rect (1, 255, 0, 0);
  v -> transform2_imgs [3] = make_rect (1, 0, 0, 255);
  v -> transform2_imgs [4] = make_rect (1, 255, 0, 0);
  v -> transform2_imgs [5] = make_rect (1, 255, 0, 255);
  v -> transform2_imgs [6] = make_rect (1, 255, 0, 0);
  v -> transform2_imgs [7] = make_rect (1, 0, 0, 255);
  v -> transform2_imgs [8] = make_cursor (63, 63, 63);
  v -> transform2_imgs [9] = make_cursor (0, 255, 0);
  v -> transform2_imgs [10] = make_rect (1, 255, 0, 0);
  v -> transform2_imgs [11] = make_rect (1, 0, 0, 255);
  v -> transform2_imgs [12] = make_rect (1, 255, 0, 0);
  v -> transform2_imgs [13] = make_rect (1, 255, 0, 255);
  v -> transform2_imgs [14] = make_rect (1, 255, 0, 0);
  v -> transform2_imgs [15] = make_rect (1, 0, 0, 255);
  v -> transform2_imgs [16] = make_rect (1, 255, 0, 0);
  v -> transform2_imgs [17] = make_rect (1, 255, 0, 255);

  v -> transform2_imgs [18] = make_cursor (255, 0, 0);
  v -> transform2_imgs [19] = make_rect (1, 0, 0, 255);
  v -> transform2_imgs [20] = make_rect (1, 0, 255, 0);

  v -> transform2_imgs [21] = (struct vcl_affinity *) make_line_style (rgba_uns2uns (0, 0, 255, 255));
  v -> transform2_imgs [22] = (struct vcl_affinity *) make_line_style (rgba_uns2uns (0, 255, 0, 255));
  vcl_enclosure_set_child (v -> transform2_imgs [21], vcl_path_new (NULL));
  vcl_enclosure_set_child (v -> transform2_imgs [22], vcl_path_new (NULL));

  void * real_grid = vcl_grid_new ();
  v -> grid = vcl_affinity_new ();
  v -> canvas = vcl_canvas_new ();
  v -> root = vcl_group_new ();
  v -> internal_root = vcl_group_new ();
  v -> zoom = vcl_affinity_new ();
  v -> area = vcl_le_area_new (bbox_of_child, v,
				expand_child, v,
				get_children, v);
  v -> ha_show = vcl_le_area_new (ha_bbox_of_child, v,
				  ha_expand_child, v,
				  ha_get_children, v);

  vcl_canvas_setup (v -> canvas, VCL_CANVAS_REAL_UNITS | VCL_CANVAS_FLIP_Y, 0.5, 0.5, 1);

  vcl_enclosure_set_child (v -> grid, real_grid);
  vcl_placement_add_child_on_top (v -> internal_root, v -> area);
  vcl_enclosure_set_child (v -> zoom, v -> internal_root);
  vcl_placement_add_child_on_top (v -> root, v -> zoom);
  vcl_placement_add_child_on_top (v -> root, v -> ha_show);
  vcl_canvas_set_root (v -> canvas, v -> root);

  v -> changed_cb = changed_cb;
  v -> changed_data = changed_data;

  WALK_GO_RECURSIVE_WITHOUT_ROOT ((struct go *) (v -> grp), go)
    {
      vcl_le_area_child_appeared_notification (v -> area, go);
    }


  call_changed_cb (v);

  recompute_selected (grp, &(v -> selected_bbox));
  update_selector (v, NULL);

  go_hook_add (go_get_tlo ((struct go *) grp), go_hook_fill (&(v -> go_hook), v, go_linked, go_unlinked, go_relinked,
				  go_selected, go_unselected, NULL, NULL, NULL, go_changed_dispatch));

  return (v -> canvas -> gtk -> widget);
}


void
visualisation_finish (struct visualisation * v)
{
  vcl_object_destroy (v -> canvas);
  go_hook_remove(go_get_tlo ((struct go *) (v -> grp)), &(v -> go_hook));
  expand_unref (go_get_tlo ((struct go *) (v -> grp)));
  go_unref ((struct go *) (v -> grp));
}

static void
update_snap_point (struct visualisation *v)
{
  if (v -> flags & VIS_FLAGS_SNAP)
    move_trans_vc (v -> snap_img, v, v -> snap_point);
}

void
visualisation_set_snap (struct visualisation *v,
			real x, real y, UNUSED struct go * g1, UNUSED struct go * g2)
{
  v -> snap_point . x = x;
  v -> snap_point . y = y;
  move_trans_vc (v -> snap_img, v, v -> snap_point);

  if (! (v -> flags & VIS_FLAGS_SNAP))
    {
      vcl_placement_add_child_on_top (v -> root, v -> snap_img);
      v -> flags |= VIS_FLAGS_SNAP;
    }
}

void
visualisation_unset_snap (struct visualisation *v)
{
  if (v -> flags & VIS_FLAGS_SNAP)
    {
      vcl_placement_remove_child (v -> root, v -> snap_img);
      v -> flags &= ~VIS_FLAGS_SNAP;
    }
}

void
visualisation_set_gadget_visible (struct visualisation *v, int trans_id, int visible)
{
  int trans_flag = (1 << trans_id);
  if (visible)
  {
    if (! (v -> gadgets & trans_flag))
    {
      v -> gadgets |= trans_flag;

      if (v -> flags & VIS_FLAGS_SELECTOR)
        vcl_placement_add_child_on_top (v -> root, v -> transform2_imgs [trans_id]);
    }
  } else {
    if (v -> gadgets & trans_flag)
    {
      v -> gadgets &= ~trans_flag;

      if (v -> flags & VIS_FLAGS_SELECTOR)
	vcl_placement_remove_child (v -> root, v -> transform2_imgs [trans_id]);
    }
  }
}

void
visualisation_set_center_gadget (struct visualisation *v, real x, real y)
{
  v -> g_center . x = x;
  v -> g_center . y = y;
  update_axis (v);
}

void
visualisation_set_xaxis_gadget (struct visualisation *v, real x, real y)
{
  v -> g_xaxis . x = x;
  v -> g_xaxis . y = y;
  update_axis (v);
}

void
visualisation_set_yaxis_gadget (struct visualisation *v, real x, real y)
{
  v -> g_yaxis . x = x;
  v -> g_yaxis . y = y;
  update_axis (v);
}


static void
update_transform_points (struct visualisation *v)
{
  if (v -> flags & VIS_FLAGS_TF_MOVE)
    move_trans_vc (v -> transform_imgs [0], v, v -> transform_points [0]);
  if (v -> flags & VIS_FLAGS_TF_RESIZE)
    move_trans_vc (v -> transform_imgs [(v -> flags & VIS_FLAGS_TF_MOVE) ? 2 : 3],
		   v, v -> transform_points [1]);
  if (v -> flags & VIS_FLAGS_TF_ROTATE)
    move_trans_vc (v -> transform_imgs [(v -> flags & VIS_FLAGS_TF_MOVE) ? 4 : 5],
		    v, v -> transform_points [1]);
  if (v -> flags & VIS_FLAGS_TF_SKEW)
    move_trans_vc (v -> transform_imgs [((v -> flags & VIS_FLAGS_TF_MOVE) &&
					 (v -> flags & VIS_FLAGS_TF_SECOND)) ? 6 : 7],
		   v, v -> transform_points [2]);
}

static int
get_vv (struct visualisation *v)
{
  int i = 0;
  if (v -> flags & VIS_FLAGS_TF_MOVE)
    {
      i = 1;
      if (v -> flags & VIS_FLAGS_TF_RESIZE) i |= 4;
      if (v -> flags & VIS_FLAGS_TF_ROTATE) i |= 16;
      if (v -> flags & VIS_FLAGS_TF_SKEW) i |= ((i > 1) ? 64 : 128);
    }
  else
    {
      i = 0;
      if (v -> flags & VIS_FLAGS_TF_RESIZE) i |= 8;
      if (v -> flags & VIS_FLAGS_TF_ROTATE) i |= 32;
      if (v -> flags & VIS_FLAGS_TF_SKEW) i |= 128;
    }
  return (i);
}


static void
visualisation_change_tf (struct visualisation *v, int flag_on, int flag_off,
			 int ptc, real x, real y)
{
  int ovv = get_vv (v);
  v -> flags |= flag_on;
  v -> flags &= ~flag_off;
  int nvv = get_vv (v);
  int i, j;

  if (ptc >= 0)
    {
      v -> transform_points [ptc] . x = x;
      v -> transform_points [ptc] . y = y;
    }

  j = 1;
  for (i = 0; i < 8; i ++)
    {
      int pti = (i < 2) ? 0 : ((i < 6) ? 1 : 2);

      if ((ovv & j) && ! (nvv & j))
	  vcl_placement_remove_child (v -> root, v -> transform_imgs [i]);

      if ((nvv & j) && ((pti == ptc) || ! (ovv & j)))
	  move_trans_vc (v -> transform_imgs [i], v, v -> transform_points [pti]);

      if ((nvv & j) && ! (ovv & j))
	vcl_placement_add_child_on_top (v -> root, v -> transform_imgs [i]);

      j *= 2;
    }
}

void
visualisation_set_tf_move (struct visualisation *v, real x, real y)
{ visualisation_change_tf (v, VIS_FLAGS_TF_MOVE, 0, 0, x, y); }

void
visualisation_set_tf_resize (struct visualisation *v, real x, real y)
{ visualisation_change_tf (v, VIS_FLAGS_TF_RESIZE, VIS_FLAGS_TF_ROTATE, 1, x, y); }

void
visualisation_set_tf_rotate (struct visualisation *v, real x, real y)
{ visualisation_change_tf (v, VIS_FLAGS_TF_ROTATE, VIS_FLAGS_TF_RESIZE, 1, x, y); }

void
visualisation_set_tf_skew (struct visualisation *v, real x, real y)
{ visualisation_change_tf (v, VIS_FLAGS_TF_SKEW, 0, 2, x, y); }

void
visualisation_unset_tf_move (struct visualisation *v)
{ visualisation_change_tf (v, 0, VIS_FLAGS_TF_MOVE, -1, 0, 0); }

void
visualisation_unset_tf_second (struct visualisation *v)
{ visualisation_change_tf (v, 0, VIS_FLAGS_TF_SECOND, -1, 0, 0); }

void
visualisation_unset_tf_skew (struct visualisation *v)
{ visualisation_change_tf (v, 0, VIS_FLAGS_TF_SKEW, -1, 0, 0); }

void
visualisation_set_grid (struct visualisation *v, struct geom_transform2 *t)
{
  int was_on = (v -> flags & VIS_FLAGS_GRID);

  vcl_affinity_set (v -> grid, t);
  v -> flags |= VIS_FLAGS_GRID;

  if (! was_on)
    vcl_placement_add_child_on_top (v -> internal_root, v -> grid);
}

void
visualisation_unset_grid (struct visualisation *v)
{
  if (v -> flags & VIS_FLAGS_GRID)
    {
      vcl_placement_remove_child (v -> internal_root, v -> grid);
      v -> flags &= ~VIS_FLAGS_GRID;
    }
}

void
visualisation_set_show_anchors_mode (struct visualisation *v, uns mode, struct go * go)
{
    ASSERT ((mode == VIS_SHOW_SPECIFIC) || !go);
    if ((mode != v -> anchor_mode) || (go != v -> anchor_specific))
    {
        v -> anchor_mode = mode;
	v -> anchor_specific = go;

	update_ha_show (v);
    }
}

void
visualisation_set_show_hangers_mode (struct visualisation *v, uns mode, struct go * go)
{
    ASSERT ((mode == VIS_SHOW_SPECIFIC) || !go);
    if ((mode != v -> hanger_mode) || (go != v -> hanger_specific))
    {
        v -> hanger_mode = mode;
	v -> hanger_specific = go;

        update_ha_show (v);
    }
}

void
visualisation_coords_to_pixels (struct visualisation *v,
				const struct geom_point *pi,
				struct geom_point *po)
{
  *po = *pi; vcl_primary_apply_point (v -> area, po);
  DBG ("COORDS (%f %f) -> PIXELS (%f %f)\n", pi -> x, pi -> y, po -> x, po -> y);
}

void
visualisation_coords_to_vcoords (struct visualisation *v,
				const struct geom_point *pi,
				struct geom_point *po)
{
  geom_point_transform (pi, &(vcl_affinity_get (v -> zoom) -> primary), po);
  DBG ("COORDS (%f %f) -> VCOORDS (%f %f)\n", pi -> x, pi -> y, po -> x, po -> y);
}

void
visualisation_vcoords_to_coords (struct visualisation *v,
				const struct geom_point *pi,
				struct geom_point *po)
{
  geom_point_transform (pi, &(vcl_affinity_get (v -> zoom) -> inverted), po);
  DBG ("VCOORDS (%f %f) -> COORDS (%f %f)\n", pi -> x, pi -> y, po -> x, po -> y);
}


void
visualisation_pixels_to_coords (struct visualisation *v,
				const struct geom_point *pi,
				struct geom_point *po)
{
  *po = *pi; vcl_inverted_apply_point (v -> area, po);
  DBG ("PIXELS (%f %f) -> COORDS (%f %f)\n", pi -> x, pi -> y, po -> x, po -> y);
}


void
visualisation_pixels_to_vcoords (struct visualisation *v,
				const struct geom_point *pi,
				struct geom_point *po)
{
  *po = *pi; vcl_inverted_apply_point (v -> root, po);
  DBG ("PIXELS (%f %f) -> VCOORDS (%f %f)\n", pi -> x, pi -> y, po -> x, po -> y);
}


void
visualisation_get_visible_bbox (struct visualisation *v, struct geom_rectangle *bbox)
{ vcl_node_get_bbox (v -> area, bbox); }

void
visualisation_get_selected_bbox (struct visualisation *v, struct geom_rectangle *bbox)
{ *bbox = v -> selected_bbox; }

void
visualisation_move (struct visualisation *v, real dx, real dy)
{
  struct geom_transform2 t;
  int err = geom_transform2_set_move (&t, dx, dy);
  ASSERT (err == 0);
  visualisation_transform (v, &t);
}

void
visualisation_absolut_move (struct visualisation *v, real x, real y)
{
  struct geom_point pi, po;

  pi . x = pi . y = 0;
  visualisation_coords_to_vcoords (v, &pi, &po);
  visualisation_move (v, x - po . x, y - po . y);
}

void
visualisation_center (struct visualisation *v, real x, real y)
{
  struct geom_point pi = {x, y};
  struct geom_point po;

  visualisation_coords_to_vcoords (v, &pi, &po);
  visualisation_move (v, - po . x, - po . y);
}

void
visualisation_move_px (struct visualisation *v, real dx, real dy)
{
  visualisation_move (v, dx / 3.7, - (dy / 3.7)); // FIXME hack
}

void
visualisation_absolut_move_px (struct visualisation *v, real x, real y)
{
  struct geom_point pi, po;

  pi . x = x; pi . y = y;
  visualisation_pixels_to_vcoords (v, &pi, &po);
  visualisation_absolut_move (v, po . x, po . y);
}

/* alternative
void
visualisation_absolut_move_px (struct visualisation *v, real x, real y)
{
  struct geom_point pi, po;

  pi . x = pi . y = 0;
  visualisation_coords_to_pixels (v, &pi, &po);
  visualisation_move_px (v, x - po . x, y - po . y);
}
*/

void
visualisation_rotate (struct visualisation *v, real angle)
{
  struct geom_transform2 t;
  int err = geom_transform2_set_rotation (&t, angle);
  ASSERT (err == 0);
  visualisation_transform (v, &t);
}

void
visualisation_scale (struct visualisation *v, real scale_x, real scale_y)
{
  struct geom_transform2 t;
  int err = geom_transform2_set_scale (&t, scale_x, scale_y);
  ASSERT (err == 0);
  visualisation_transform (v, &t);
}


static void
visualisation_update (struct visualisation *v)
{
  update_selector (v, &(v -> selected_bbox));
  update_axis (v);
  update_snap_point (v);
  update_transform_points (v);
  update_ha_show (v);
  call_changed_cb (v);
}

void
visualisation_set_orientation (struct visualisation *v,
			       real x_center, real y_center,
			       real x_scale, real y_scale, real angle)
{
  struct geom_transform2 t;
  int err = 0;

  geom_transform2_set_identity (&t);
  err = err || geom_transform2_translate (&t, - x_center, - y_center);
  err = err || geom_transform2_scale (&t, x_scale, y_scale);
  err = err || geom_transform2_rotate (&t, angle);
  ASSERT (err == 0);

  vcl_affinity_set (v -> zoom, &t);
  visualisation_update (v);
}

void
visualisation_transform (struct visualisation *v, const struct geom_transform2 *t)
{
  vcl_affinity_outer_merge (v -> zoom, t);
  visualisation_update (v);
}

void
visualisation_gui_rectangle_update (struct visualisation *v, struct vcl_path ** r,
		struct geom_point a, struct geom_point b, struct geom_point c )
{
	ASSERT(v && r);
	if ( !(*r) )
	  {
	    struct vcl_property * prop = make_line_style (rgba_uns2uns (0, 255, 0, 255));
	    *r = vcl_path_new (NULL);
	    vcl_enclosure_set_child (prop, *r);
	  }
	else vcl_placement_remove_child( v -> root, vcl_node_get_parent (*r) );


	struct geom_path *p = &(*r)->embedded;
	//struct geom_segment s;
	struct geom_bezier s;
	geom_path_clear(p);

	visualisation_coords_to_vcoords( v, &a, &a );
	visualisation_coords_to_vcoords( v, &b, &b );
	visualisation_coords_to_vcoords( v, &c, &c );

	struct geom_point d;
	d.x = c.x + a.x - b.x; d.y = c.y + a.y - b.y;

	/*GEOM_ITEM_CREATE(&s, segment);
	geom_segment_set( &s, &a, &b ); geom_path_append_dup( p, &s );
	geom_segment_set( &s, &b, &c ); geom_path_append_dup( p, &s );
	geom_segment_set( &s, &c, &d ); geom_path_append_dup( p, &s );
	geom_segment_set( &s, &d, &a ); geom_path_append_dup( p, &s );
	geom_segment_destroy(&s);*/
	GEOM_ITEM_CREATE(&s, bezier);
	geom_bezier_fill_linear_nonrational( &s, a.x, a.y, b.x, b.y );
	geom_path_append_dup( p, &s );
	geom_bezier_fill_linear_nonrational( &s, b.x, b.y, c.x, c.y );
	geom_path_append_dup( p, &s );
	geom_bezier_fill_linear_nonrational( &s, c.x, c.y, d.x, d.y );
	geom_path_append_dup( p, &s );
	geom_bezier_fill_linear_nonrational( &s, d.x, d.y, a.x, a.y );
	geom_path_append_dup( p, &s );
	geom_bezier_destroy(&s);
	vcl_path_source_changed( *r );
	vcl_placement_add_child_on_top( v -> root, vcl_node_get_parent (*r) );
}

void
visualisation_gui_rectangle_destroy (struct visualisation *v, struct vcl_path ** r)
{
  ASSERT( v && r && *r );
  vcl_placement_remove_child (v -> root, vcl_node_get_parent (*r));
  vcl_object_destroy (vcl_node_get_parent (*r));
  *r = NULL;
}
