/** @file
 * 	VRR import routines.
 *
 * 	Routines to read in and decode information stored in
 * 	DVI file used by TeX.
 */

#ifndef _IMPORT_IMPORT_H_
#define _IMPORT_IMPORT_H_

#include "geomlib/types.h"
#include <stdio.h>

#define POINT2MM (25.4/72)

/**
 * DVI parser error codes.
 */
enum {
	DVI_EOF = -20,			///< EOF was encountered, used internally
	DVI_ERR_CORRUPTED,		///< DVI code seriously damaged
	DVI_ERR_UNDEF_FONT,		///< previously undefined font was used
	DVI_ERR_NO_FONT,		///< typeset character without loading any font
	DVI_ERR_VRRFONT,		///< error in VRR fontserver occured
	DVI_ERR_STACK_NONEMPTY,		///< DVI position stack should be empty
	DVI_ERR_STACK_UNDERFLOW,	///< DVI position stack underflow
	DVI_ERR_TFM_NOTFOUND,		///< TFM metrics for selected font was not found
	DVI_ERR_TFM_INVALID_HDR,	///< TFM header is corrupted
	DVI_ERR_TFM_CORRUPTED,		///< TFM file seriously damaged
	DVI_ERR_TFM_BAD_CHKSUM,		///< TFM checksum is invalid
};

/**
 * DVI error string.
 *
 * In case an error occurs and parsing fails, this buffer is set
 * to a message describing the error.
 */
extern char dvi_err_msgbuf[];

/**
 * One TeX glyph.
 */
struct tex_glyph {
	int is_rule;		///< if 1 this glyph represents TeX rule, size is stored in bbox
	int font_id;		///< font descriptor
	uns code;		///< character code
	real font_size;		///< in font_units
	real xpos, ypos;	///< position of reference point in TLO, in milimeters
	struct geom_rectangle bbox;	///< relative to reference point, in milimeters
};

struct garr;
/**
 * Parse DVI file and return discovered glyphs.
 *
 * @ai f file opened for reading
 * @ao glyphs_out parsed glyphs
 * @rv <0 error code on error
 * @rv glyphs number of parsed glyphs
 */
int import_dvi(FILE *f, struct garr *glyphs_out);

struct obj_tlo;

int import_ipe5(FILE *in);

#include "kernel/scheme_def.h"

SCHEME SINT import_tlo_ipe5(SSTRINGC filename, SO_obj_tlo out);

#endif


/*** SVG IMPORT ***/

/**
 * Parse svg file and imports graphical objects into the newly created tlo.
 *
 * @ai filename name of the file which will be imported
 * @ai tlo tlo for imported objects
 * @rv 0 file was imported successfully
 * @rv -1 an error occured
*/
int import_svg (char *filename, struct obj_tlo *tlo);
