#undef LOCAL_DEBUG

#include "lib/lib.h"
#include <stdio.h>
#include <libxml/xmlreader.h>
#include <ctype.h>
#include "kernel/kernel.h"
#include "import/import.h"
#include "font/font.h"

#define GET_REAL_ATTRIBUTE(val, attr) { char *tmp = xmlTextReaderGetAttribute (ptr, attr); if (!tmp) return; val = atof(tmp); free(tmp); }
#define GET_STR_ATTRIBUTE(val, attr) { val = xmlTextReaderGetAttribute (ptr, attr) ; if (!val) return; }
#define GET_ADDITIONAL_REAL_ATTRIBUTE(val, attr) { char *tmp = xmlTextReaderGetAttribute (ptr, attr); if (!tmp)  val = 0; else { val = atol (tmp); free (tmp); } }
#define GET_ADDITIONAL_STR_ATTRIBUTE(val, attr) { val = xmlTextReaderGetAttribute (ptr, attr); }
#define GET_ADDITIONAL_INT_ATTRIBUTE(val, attr) { char *tmp = xmlTextReaderGetAttribute (ptr, attr); if (!tmp) val = 0; else { val = atoi (tmp); free (tmp); } }
#define GET_STYLE(val) GET_ADDITIONAL_STR_ATTRIBUTE(val, style_str);
#define GET_COLOR_FROM_STYLE if (style) get_color(style + 6);
#define GET_COLOR_FROM_STROKE get_color(stroke);
#define NULL_COLORS_AND_RETURN 	{rc = gc = bc = 0; alpha = 255; return;}
#define GET_POINTS(x, y)	if (!str) return 0; str = get_points_from_str (str, x, y);

#define SVG_COLORS			16
#define UNDEFINED_VALUE		NAN
#define CUBIC_BEZIER		0
#define QUADRATIC_BEZIER	1
#define DEFAULT_SVG_FONT_SIZE	12
#define DEFAULT_SVG_FONT_FAMILY	"Verdana"

#define TRANSFORMX(x) (x * POINT2MM)
#define TRANSFORMLEN(x) (x * POINT2MM)
#define TRANSFORMY(y) ((height - y) * POINT2MM)

struct temp
{
	real x, y;
	int type;
};

static const char *color_strings[SVG_COLORS] = { "aqua", "black", "blue", "fuchsia", "gray", "green", "lime", "maroon", "navy", "olive", "purple", "red", "silver", "teal", "white", "yellow" };
static const unsigned int colors[SVG_COLORS][3] = { {0, 255, 255}, {0, 0, 0}, {0, 0, 255}, {255, 0, 255}, {128, 128, 128}, {0, 128, 0}, {0, 255, 0}, {128, 0, 0}, {0, 0, 128}, {128, 128, 0}, {128, 0, 128}, {255, 0, 0}, {192, 192, 192}, {0, 128, 128}, {255, 255, 255}, {255, 255, 0} };
static const char *svg = "svg";
static const char *rect = "rect", *ellipse = "ellipse", *circle = "circle",
	 *line = "line", *polygon = "polygon", *polyline = "polyline",
	 *path = "path", *text = "text";
static const char *width_str = "width", *height_str = "height", *style_str = "style",
	  *cx_str = "cx", *cy_str = "cy", *rx_str = "rx", *ry_str = "ry",
	  *r_str = "r", *d_str = "d", *points_str = "points", *x1_str = "x1",
	  *y1_str = "y1", *x2_str = "x2", *y2_str = "y2", *x_str = "x",
	  *y_str = "y", *fontsize_str = "font-size",
	  *fontfamily_str = "font-family", *stroke_str = "stroke";
static uns rc = 0, gc = 0, bc = 0, alpha = 255;
static real height = 0;

static void create_line (real x1, real y1, real x2, real y2);
static void create_cubic_bezier (real x1, real y1, real x2, real y2, real x3, real y3, real x4, real y4);
static void create_quadratic_bezier (real x1, real y1, real x2, real y2, real x3, real y3);
static void create_elarc (real x1, real y1, real x2, real y2, real rx, real ry, real rotation);

static void set_color (struct go *go)
{
	char buf[TRANS_ERR_SIZE];

	uns color = rgba_uns2uns (rc, gc, bc, alpha);
	DBG("color: %d, rgb: %d, %d, %d, alpha: %d\n", color, rc, gc, bc, alpha);
	TRANS_BEGIN(buf)
		prop_set_uns (go2o(go), string_stroke_color, (uintptr_t) color);
	TRANS_FAILED
		DBG(buf);
	TRANS_END
}

static char *skip_number (char *str)
{
	if(*str == '-' || *str == '+')
		str++;
	while (*str && (isdigit(*str) || (*str == '.')))
		str++;

	return str;
}

static char *skip_spaces (char *str)
{
	while (*str  && (isspace(*str) || (*str == ',')))
		str++;

	return str;
}

static char *get_points_from_str (char *str, real *x, real *y)
{
	if (!str)
		return 0;
	while (*str && isspace(*str))	// skip spaces on the beginning
		str++;
#ifdef FLOAT_REAL
	if (sscanf (str, "%f", x) < 1)
#else
	if (sscanf (str, "%lf", x) < 1)
#endif
		return 0;			
	str = skip_number(str);
	str = skip_spaces(str);
#ifdef FLOAT_REAL
	if (sscanf (str, "%f", y) < 1)
#else
	if (sscanf (str, "%lf", y) < 1)
#endif
		return (char *)(-1);	// ERROR
	str = skip_number(str);

	return str;
}

static char *get_point_from_str (char *str, real *x)
{
	if (!str)
		return 0;
	while (*str && isspace(*str)) // skip spaces on the beginning
		str++;
#ifdef FLOAT_REAL
	if (sscanf (str, "%f", x) < 1)
#else
	if (sscanf (str, "%lf", x) < 1)
#endif
		return 0;
	while (*str && (!isspace(*str))) // skip first number
		str++;

	return str;
}

static char *fill_one_point (char *str, real *x, real *y)
{
	GET_POINTS(x, y);

	return str;
}

#define CHECK_STR if ((str == (char *)(-1)) || (!str)) { *buf = 0; return 0; }
static char *fill_two_points (char **buf, real *x1, real *y1, real *x2, real *y2)
{
	char *str = *buf;

	GET_POINTS(x1, y1);
	CHECK_STR;
	GET_POINTS(x2, y2);
	CHECK_STR;

	*buf = str;
	return str;
}

static char *fill_three_points (char **buf, real *x1, real *y1, real *x2, real *y2, real *x3, real *y3)
{
	char *str = *buf;
	
	GET_POINTS(x1, y1);
	CHECK_STR;
	GET_POINTS(x2, y2);
	CHECK_STR;
	GET_POINTS(x3, y3);
	CHECK_STR;

	*buf = str;
	return str;
}
#undef CHECK_STR

static void get_fontsize_from_str (char *str, int *value)
{
	int i;

	if (!str)
	{
		*value = 0;
		return;
	}
   	i = strlen(str) - 10; // (10 == nr of chars in "font-size:")
	while (i > 0)
	{
		if (!strncmp (str, "font-size:", 10))
		{
			if (sscanf (str, "font-size:%d", value) < 1)
			{
				*value = 0;
		 		return;
			}
			else
				return;
		}
		i--;
		str++;
	}
	*value = 0;
}

static void get_fontfamily_from_str (char *str, char *value)
{
	int i;

	if (!str)
	{
		value = 0;
		return;
	}
   	i = strlen(str) - 12; // (12 == nr of chars in "font-family:")
	while (i > 0)
	{
		if (!strncmp (str, "font-family:", 12))
		{
			if (sscanf (str, "font-family:%s", value) < 1)
			{
				value = 0;
		 		return;
			}
			else
				return;
		}
		i--;
		str++;
	}
	value = 0;
}

static char *get_elarc_data (char **buf, real *rx, real *ry, real *rotation, real *flag1, real *flag2, real *x, real *y)
{
	char *str = *buf;
#define GET_REAL_FROM_STR(x) str = get_point_from_str (str, x); if (!str) return 0;
#define CHECK_STR if (!str) { *buf = 0; return 0; }
	GET_REAL_FROM_STR(rx);
	GET_REAL_FROM_STR(ry);
	CHECK_STR;
	GET_REAL_FROM_STR(rotation);
	CHECK_STR;
	GET_REAL_FROM_STR(flag1);
	CHECK_STR;
	GET_REAL_FROM_STR(flag2);
	CHECK_STR;
	GET_REAL_FROM_STR(x);
	CHECK_STR;
	GET_REAL_FROM_STR(y);
	CHECK_STR;
#undef CHECK_STR

	*buf = str;
	return str;
}

static char *get_path_from_str (char *str, real *firstx, real *firsty, real *lastx, real *lasty, struct temp *tp)
{
	real x, y, x1, y1, x2, y2;
	char *tmp, *ret;

	if (!str)
		return 0;
	while (*str && isspace(*str))
		str++;
	if (!(*str))
		return 0;
#define CHECK_PATH_BEGINNING if (*firstx == UNDEFINED_VALUE && *firsty == UNDEFINED_VALUE) { DBG("Trying to work with path without its initialisation"); return 0; };
#define CHECK_CURRENT_POINT if (*lastx == UNDEFINED_VALUE && *lasty == UNDEFINED_VALUE) { DBG("No current point set!"); return 0;}
#define INIT_PATH_BEGINNING *firstx = *lastx, *firsty = *lasty;
#define NULL_TMP_DATA tp->x = tp->y = tp->type = 0;
#define SET_TMP_DATA(xx, yy, ttype) { tp->x = xx; tp->y = yy; tp->type = ttype; }
#define CHECK_AND_GET_TMP_DATA(ttype)  if ((ttype != tp->type) || (tp->x == UNDEFINED_VALUE && tp->y == UNDEFINED_VALUE)) { x = *lastx; y = *lasty; } else { x1 = 2 * *lastx - tp->x; y1 = 2 * *lasty - tp-> y; }
	DBG("path: %s\n",str);
	switch(*str)
	{
		case 'M':	// moveto
		case 'm':	// moveto (relatively positioned)
			tmp = fill_one_point (str + 1, lastx, lasty);
			if (!str)
				return 0;
			INIT_PATH_BEGINNING;
			DBG("M %lf %lf\n", *lastx, *lasty);
			NULL_TMP_DATA;
			break;
		case 'L':	// lineto (absolutely positioned)
			CHECK_CURRENT_POINT;
			tmp = ret = fill_one_point (str + 1, &x, &y);
			while (ret)
			{
				DBG("L %lf %lf\n", x, y);
				create_line (*lastx, *lasty, x, y);
				*lastx = x;
				*lasty = y;
				tmp = ret;
				ret = fill_one_point(ret, &x, &y);
			}
			NULL_TMP_DATA;
			break;
		case 'l':	// lineto (relatively positioned)
			CHECK_CURRENT_POINT;
			tmp = ret = fill_one_point (str + 1, &x, &y);
			while (ret)
			{
				DBG("l %lf %lf\n", x, y);
				create_line (*lastx, *lasty, *lastx + x, *lasty + y);
				*lastx += x;
				*lasty += y;
				tmp = ret;
				ret = fill_one_point (ret, &x, &y);
			}
			NULL_TMP_DATA;
			break;
		case 'V':	// vertical lineto (absolutely positioned)
			CHECK_CURRENT_POINT;
			ret = get_point_from_str (str + 1, &y);
			while (ret)
			{
				DBG("V %lf\n", y);
				tmp = ret;
				create_line (*lastx, *lasty, *lastx, y);
				*lasty = y;
				ret = get_point_from_str (ret, &y);
			}
			NULL_TMP_DATA;
			break;
		case 'v':	// vertical lineto (relatively positioned)
			CHECK_CURRENT_POINT;
			ret = get_point_from_str (str + 1, &y);
			while (ret)
			{
				DBG("v %lf\n", y);
				tmp = ret;
				create_line (*lastx, *lasty, *lastx, *lasty + y);
				*lasty += y;
				ret = get_point_from_str (ret, &y);
			}
			NULL_TMP_DATA;
			break;
		case 'H': // horizontal lineto (absolutely positioned)
			CHECK_CURRENT_POINT;
			ret = get_point_from_str (str + 1, &x);
			while (ret)
			{
				DBG("H %lf\n", x);
				tmp = ret;
				create_line (*lastx, *lasty, x, *lasty);
				*lastx = x;
				ret = get_point_from_str (ret, &x);
			}
			NULL_TMP_DATA;
			break;
		case 'h': // horizontal lineto (relatively positioned)
			CHECK_CURRENT_POINT;
			ret = get_point_from_str (str + 1, &x);
			while (ret)
			{
				DBG("h %lf\n", x);
				tmp = ret;
				create_line (*lastx, *lasty, *lastx + x, *lasty);
				*lastx += x;
				ret = get_point_from_str (ret, &x);
			}
			NULL_TMP_DATA;
			break;
		case 'C':	// curveto (cubic bezier) (absolutely positioned)
			CHECK_CURRENT_POINT;
			tmp = str + 1;
			ret = fill_three_points (&tmp, &x1, &y1, &x2, &y2, &x, &y);
			while (ret)
			{
				DBG("C %lf %lf %lf %lf %lf %lf\n", x1, y1, x2, y2, x, y);
				create_cubic_bezier (*lastx, *lasty, x1, y1, x2, y2, x, y);
				SET_TMP_DATA(x2, y2, CUBIC_BEZIER);
				*lastx = x;
				*lasty = y;
				ret = fill_three_points (&tmp, &x1, &y1, &x2, &y2, &x, &y);
			}
			break;
		case 'c':	// curveto (cubic bezier) (relatively positioned)
			CHECK_CURRENT_POINT;
			tmp = str + 1;
			ret = fill_three_points (&tmp, &x1, &y1, &x2, &y2, &x, &y);
			while (ret)
			{
				DBG("c %lf %lf %lf %lf %lf %lf\n", x1, y1, x2, y2, x, y);
				create_cubic_bezier (*lastx, *lasty, *lastx + x1, *lasty + y1,
						*lastx + x2, *lasty + y2, *lastx + x, *lasty + y);
				SET_TMP_DATA(*lastx + x2, *lasty + y2, CUBIC_BEZIER);
				*lastx += x;
				*lasty += y;
				ret = fill_three_points (&tmp, &x1, &y1, &x2, &y2, &x, &y);
			}
			break;
		case 'S':	// smooth curveto (absolutely positioned)
			CHECK_CURRENT_POINT;
			tmp = str + 1;
			ret = fill_two_points (&tmp, &x2, &y2, &x, &y);
			while (ret)
			{
				DBG("S %lf %lf %lf %lf\n", x2, y2, x, y);
				CHECK_AND_GET_TMP_DATA(CUBIC_BEZIER);
				create_cubic_bezier (*lastx, *lasty, x1, y1, x2, y2, x, y);
				SET_TMP_DATA(x2, y2, CUBIC_BEZIER);
				*lastx = x;
				*lasty = y;
				ret = fill_two_points (&tmp, &x2, &y2, &x, &y);
			}
			break;
		case 's':	// smooth curveto (relatively positioned)
			CHECK_CURRENT_POINT;
			tmp = str + 1;
			ret = fill_two_points (&tmp, &x2, &y2, &x, &y);
			while (ret)
			{
				DBG("s %lf %lf %lf %lf\n", x2, y2, x, y);
				CHECK_AND_GET_TMP_DATA(CUBIC_BEZIER);
				create_cubic_bezier (*lastx, *lasty, x1, y1,
						*lastx + x2, *lasty + y2, *lastx + x, *lasty + y);
				SET_TMP_DATA(*lastx + x2, *lasty + y2, CUBIC_BEZIER);
				*lastx += x;
				*lasty += y;
				ret = fill_two_points (&tmp, &x2, &y2, &x, &y);
			}
			break;
		case 'Q':	// quadratic bezier curve (absolutely positioned)
			CHECK_CURRENT_POINT;
			tmp = str + 1;
			ret = fill_two_points (&tmp, &x1, &y1, &x, &y);
			while (ret)
			{
				DBG("Q %lf %lf %lf %lf\n", x1, y1, x, y);
				create_quadratic_bezier (*lastx, *lasty, x1, y1, x, y);
				SET_TMP_DATA(x1, y1, QUADRATIC_BEZIER);
				*lastx = x;
				*lasty = y;
				ret = fill_two_points (&tmp, &x1, &y1, &x, &y);
			}
			break;
		case 'q':	// quadratic bezier curve (relatively positioned)
			CHECK_CURRENT_POINT;
			tmp = str + 1;
			ret = fill_two_points (&tmp, &x1, &y1, &x, &y);
			while (ret)
			{
				DBG("q %lf %lf %lf %lf\n", x1, y1, x, y);
				create_quadratic_bezier (*lastx, *lasty,
						*lastx + x1, *lasty + y1, *lastx + x, *lasty + y);
				SET_TMP_DATA(*lastx + x1, *lasty + y1, QUADRATIC_BEZIER);
				*lastx += x;
				*lasty += y;
				ret = fill_two_points (&tmp, &x1, &y1, &x, &y);
			}
			break;
		case 'T':	// smooth quadratic bezier curveto (absolutely positioned)
			CHECK_CURRENT_POINT;
			tmp = ret = fill_one_point (str + 1, &x, &y);
			while (ret)
			{
				DBG("T %lf %lf\n", x, y);
				CHECK_AND_GET_TMP_DATA(QUADRATIC_BEZIER);
				create_quadratic_bezier (*lastx, *lasty, x1, y1, x, y);
				SET_TMP_DATA(*lastx, *lasty, QUADRATIC_BEZIER);
				*lastx = x;
				*lasty = y;
				tmp = ret;
				ret = fill_one_point (ret, &x, &y);
			}
			break;
		case 't':	// smooth quadratic bezier curveto (relatively positioned)
			CHECK_CURRENT_POINT;
			tmp = ret = fill_one_point (str + 1, &x, &y);
			while (ret)
			{
				DBG("t %lf %lf\n", x, y);
				CHECK_AND_GET_TMP_DATA(QUADRATIC_BEZIER);
				create_quadratic_bezier (*lastx, *lasty, x1, y1,
						*lastx + x, *lasty + y);
				SET_TMP_DATA(*lastx, *lasty, QUADRATIC_BEZIER);
				*lastx += x;
				*lasty += y;
				tmp = ret;
				ret = fill_one_point (ret, &x, &y);
			}
			break;
		case 'A':	// elliptical arc (absolutely positioned)
			{
				real rx, ry, rotation, flag1, flag2;

				CHECK_CURRENT_POINT;
				tmp = str + 1;
				tmp = ret = get_elarc_data (&tmp, &rx, &ry, &rotation,
						&flag1, &flag2, &x, &y);
				while (ret)
				{
					DBG("A %lf %lf %lf %lf %lf %lf %lf\n", rx, ry, rotation,
							flag1, flag2, x, y);
					create_elarc (*lastx, *lasty, x, y, rx, ry, rotation);
					*lastx = x;
					*lasty = y;
					ret = get_elarc_data (&tmp, &rx, &ry, &rotation,
							&flag1, &flag2, &x, &y);
				}
				NULL_TMP_DATA;
				break;
			}
		case 'a':	// elliptical arc (relatively positioned)
			{
				real rx, ry, rotation, flag1, flag2;

				CHECK_CURRENT_POINT;
				tmp = str + 1;
				tmp = ret = get_elarc_data (&tmp, &rx, &ry, &rotation, &flag1,
						&flag2, &x, &y);
				while (ret)
				{
					DBG("a %lf %lf %lf %lf %lf %lf %lf\n", rx, ry, rotation,
							flag1, flag2, x, y);
					create_elarc (*lastx, *lasty, *lastx + x, *lasty + y,
							rx, ry, rotation);
					*lastx += x;
					*lasty += y;
					ret = get_elarc_data (&tmp, &rx, &ry, &rotation, &flag1,
							&flag2, &x, &y);
				}
				NULL_TMP_DATA;
				break;
			}
		case 'Z':	// closepath
		case 'z':
			CHECK_PATH_BEGINNING;
			DBG("z - closing path : %lf %lf %lf %lf\n", *lastx, *lasty, *firstx, *firsty);
			create_line (*lastx, *lasty, *firstx, *firsty);
			INIT_PATH_BEGINNING;
			tmp = str + 1;
			break;
		default:	// default : lineto
			CHECK_CURRENT_POINT;
			tmp = ret = fill_one_point (str + 1, &x, &y);
			while (ret)
			{
				DBG("L %lf %lf\n", x, y);
				tmp = ret;
				create_line (*lastx, *lasty, x, y);
				*lastx = x;
				*lasty = y;
				ret = fill_one_point(ret, &x, &y);
			}
			NULL_TMP_DATA;
	}

	return tmp;
}

static void get_color(char *color)
{
	if (!color || !(*color))
		NULL_COLORS_AND_RETURN;
	if (!sscanf(color, "rgb(%d,%d,%d)", &rc, &gc, &bc))
	{
		if(!sscanf(color, "#%s", color))
		{
			int i;
			for (i = 0; i < SVG_COLORS; i++)
				if(!strcmp(color, color_strings[i]))
				{
					rc = colors[i][0];
					gc = colors[i][1];
					bc = colors[i][2];
					return;
				}
			if (!strcmp(color, "none")) // ???
				NULL_COLORS_AND_RETURN;
			if (i == SVG_COLORS)
				NULL_COLORS_AND_RETURN;
		}
		else
		{
			if (strlen(color) == 6)
			{
				sscanf(color + 4, "%x", &bc);
				color[4] = 0;
				sscanf(color + 2, "%x", &gc);
				color[2] = 0;
				sscanf(color, "%x", &rc);
				return;
			}
			if (strlen(color) == 3)
			{
				sscanf(color + 2, "%x", &bc);
				color[2] = 0;
				bc += bc << 4;
				sscanf(color + 1, "%x", &gc);
				gc += gc << 4;
				color[1] = 0;
				sscanf(color, "%x", &rc);
				rc += rc << 4;
				return;
			}
		}
	}
}

#define MC(index) hanger_new_mc_xy(TRANSFORMX(x##index), TRANSFORMY(y##index))

static void create_line (real x1, real y1, real x2, real y2)
{
	struct go_segment *gs;
	char buf[TRANS_ERR_SIZE];

	DBG("Creating line: %f %f %f %f\n", TRANSFORMX(x1), TRANSFORMY(y1), TRANSFORMX(x2), TRANSFORMY(y2));
	TRANS_BEGIN(buf)
		gs = go_segment_new (MC(1), MC(2));
		go_link (&gs->go, &trans_tlo->tlg->group, NULL);
		set_color (&gs->go);
	TRANS_FAILED
		DBG(buf);
	TRANS_END
}

static void create_cubic_bezier (real x1, real y1, real x2, real y2, real x3, real y3, real x4, real y4)
{
	struct go_bezier *gb;
	char buf[TRANS_ERR_SIZE];

	TRANS_BEGIN(buf)
		gb = go_bezier_new_cubic (MC(1), 1.0, MC(2), 1.0, MC(3), 1.0, MC(4), 1.0);
		go_link (&gb->go, &trans_tlo->tlg->group, NULL);
		set_color (&gb->go);
	TRANS_FAILED
		DBG(buf);
	TRANS_END
}

static void create_quadratic_bezier (real x1, real y1, real x2, real y2, real x3, real y3)
{
	struct go_bezier *gb;
	char buf[TRANS_ERR_SIZE];

	TRANS_BEGIN(buf)
		gb = go_bezier_new_quadratic (MC(1), 1.0, MC(2), 1.0, MC(3), 1.0);
		go_link (&gb->go, &trans_tlo->tlg->group, NULL);
		set_color (&gb->go);
	TRANS_FAILED
		DBG(buf);
	TRANS_END
}


static void create_elarc (real x1 UNUSED, real y1 UNUSED, real x2 UNUSED, real y2 UNUSED, real rx UNUSED, real ry UNUSED, real rotation UNUSED)
{
	DBG("Creating elarc..\n");
	// set color
}

static void import_rect (xmlTextReaderPtr ptr)
{
	real width, height;
	real x, y;
	char *style;

	GET_REAL_ATTRIBUTE(x, x_str);
	GET_REAL_ATTRIBUTE(y, y_str);
	GET_REAL_ATTRIBUTE(width, width_str);
	GET_REAL_ATTRIBUTE(height, height_str);
	GET_STYLE(style);
	DBG("style: %s\n", style);

	GET_COLOR_FROM_STYLE;
	DBG("Importing rectangle - width: %f, height: %f, style: %s, color: (%d, %d, %d)\n", width, height, style, rc, gc, bc);
	create_line (x, y, x + width, y);
	create_line (x + width, y, x + width, y + width);
	create_line (x + width, y + width, x, y + width);
	create_line (x, y + width, x, y);

	free (style);
}

static void import_ellipse (xmlTextReaderPtr ptr)
{
	real xc, yc, rx, ry;
	char *style;

	DBG("ELLIPSE\n");
	GET_REAL_ATTRIBUTE(xc, cx_str);
	GET_REAL_ATTRIBUTE(yc, cy_str);
	GET_REAL_ATTRIBUTE(rx, rx_str);
	GET_REAL_ATTRIBUTE(ry, ry_str);
	GET_STYLE(style);

	GET_COLOR_FROM_STYLE;
	DBG("Importing ellipse - cx: %f, cy: %f, rx: %f, ry: %f, style: %s, color: (%d, %d, %d)\n", xc, yc, rx, ry, style, rc, gc, bc);
	struct go_elarc *ge;
	char buf[TRANS_ERR_SIZE];

	TRANS_BEGIN(buf)
		ge = go_elarc_new_xyr (MC(c), 0, TRANSFORMLEN(rx), TRANSFORMLEN(ry), 0, 2 * R_PI, GEOM_CONIC_ARC_START_ENTIRE);
		go_link(&ge->go, &trans_tlo->tlg->group, NULL);
		set_color (&ge->go);
	TRANS_FAILED
		DBG(buf);
	TRANS_END

	free (style);
}

static void import_circle (xmlTextReaderPtr ptr)
{
	real xc, yc, r;
	char *style;

	GET_REAL_ATTRIBUTE(xc, cx_str);
	GET_REAL_ATTRIBUTE(yc, cy_str);
	GET_REAL_ATTRIBUTE(r, r_str);
	GET_STYLE(style);

	GET_COLOR_FROM_STYLE;
	DBG("Importing circle - cx: %f, cy: %f, style: %s, color: (%d, %d, %d)\n", xc, yc, style, rc, gc, bc);
	struct go_elarc *ge;
	char buf[TRANS_ERR_SIZE];

	TRANS_BEGIN(buf)
		ge = go_elarc_new_xyr (MC(c), 0, TRANSFORMLEN(r), TRANSFORMLEN(r), 0, 2 * R_PI, GEOM_CONIC_ARC_START_ENTIRE);
		go_link(&ge->go, &trans_tlo->tlg->group, NULL);
		if (style)
			set_color (&ge->go);
	TRANS_FAILED
		DBG(buf);
	TRANS_END

	free (style);
}

static void import_line (xmlTextReaderPtr ptr)
{
	real x1, y1, x2, y2;
	char *style;

	GET_REAL_ATTRIBUTE(x1, x1_str);
	GET_REAL_ATTRIBUTE(y1, y1_str);
	GET_REAL_ATTRIBUTE(x2, x2_str);
	GET_REAL_ATTRIBUTE(y2, y2_str);
	GET_STYLE(style);

	GET_COLOR_FROM_STYLE;
	DBG("Importing line - x1: %f, y1: %f, x2: %f, y2: %f, style: %s, color: (%d, %d, %d)\n", x1, y1, x2, y2, style, rc, gc, bc);
	create_line (x1, y1, x2, y2);

	free (style);
}

static void import_polygon (xmlTextReaderPtr ptr)
{
	char *style, *points, *tmp;
	real x1, y1, x2, y2;
	real x0, y0;

	GET_STR_ATTRIBUTE(points, points_str);
	GET_STYLE(style);
	GET_COLOR_FROM_STYLE;
	DBG("Importing polygon - points: %s, style: %s, color: (%d, %d, %d)\n", points, style, rc, gc, bc);
	tmp = fill_one_point (points, &x1, &y1);
	x0 = x1;
	y0 = y1;

	while (tmp = fill_one_point (tmp, &x2, &y2))
	{
		create_line (x1, y1, x2, y2);
		x1 = x2;
		y1 = y2;
	}
	create_line (x2, y2, x0, y0);

	free (points);
	free (style);
}

static void import_polyline (xmlTextReaderPtr ptr)
{
	char *style, *points, *tmp;
	real x1, y1, x2, y2;

	GET_STR_ATTRIBUTE(points, points_str);
	GET_STYLE(style);
	GET_COLOR_FROM_STYLE;
	DBG("Importing polyline - points: %s, style: %s\n, color: (%d, %d %d)\n", points, style, rc, gc, bc);
	tmp = fill_one_point (points, &x1, &y1);

	while (tmp = fill_one_point (tmp, &x2, &y2))
	{
		create_line (x1, y1, x2, y2);
		x1 = x2;
		y1 = y2;
	}

	free (points);
	free (style);
}

static void import_path (xmlTextReaderPtr ptr)
{
	char *d, *tmp, *stroke;
	real firstx, firsty, lastx, lasty;
	struct temp tp;

	firstx = firsty = lastx = lasty = UNDEFINED_VALUE;

	GET_STR_ATTRIBUTE(d, d_str);
	GET_ADDITIONAL_STR_ATTRIBUTE(stroke, stroke_str);
	GET_COLOR_FROM_STROKE;
	DBG("Importing path - d: %s\n", d);
	tmp = d;
	do
		tmp = get_path_from_str (tmp, &firstx, &firsty, &lastx, &lasty, &tp);
	while (tmp);

	free (d);
}

static void import_text (xmlTextReaderPtr ptr)
{
	real x = 0, y = 0;
	int fontsize = 0, font_id;
	char *fontfamily, *text, *style;
	char outbuf[1024];

	DBG("Importing text\n");
#define TEST_VALID(x, y) if  (!(x)) {DBG(y); return;}
	GET_REAL_ATTRIBUTE(x, x_str);
	GET_REAL_ATTRIBUTE(y, y_str);
	GET_STYLE(style);
	fontfamily = xmalloc (sizeof (style));
	get_fontfamily_from_str(style, fontfamily);
	if (!(*fontfamily))
	{
		xfree (fontfamily);
		fontfamily = 0;
	}
	get_fontsize_from_str(style, &fontsize);
	if (!fontfamily)
		GET_ADDITIONAL_STR_ATTRIBUTE(fontfamily, fontfamily_str);
	if (!fontsize)
		GET_ADDITIONAL_REAL_ATTRIBUTE(fontsize, fontsize_str);
	TEST_VALID(xmlTextReaderRead (ptr) == 1, "Can't read from xmlTextReader\n");
	TEST_VALID(xmlTextReaderNodeType (ptr) == XML_TEXT_NODE, "Bad node type (not text)\n");
	TEST_VALID(text = xstrdup((char *) xmlTextReaderConstValue (ptr)), "Couldn't read text string from <text></text>\n");

	if (!fontsize)
		fontsize = DEFAULT_SVG_FONT_SIZE;	// default value
	if (!fontfamily)
		fontfamily = DEFAULT_SVG_FONT_FAMILY;
	if (!font_search (fontfamily, NULL, outbuf, 1024))
	{
		DBG("Couldn't load fontid from fontfamily\n");
		return;
	}
	font_id = font_load_file (outbuf);
	TEST_VALID(font_id, "Couldn't get fontid from font file\n");

	struct go_text *gt;
	char buf[TRANS_ERR_SIZE];

	TRANS_BEGIN(buf);
		gt = go_text_new (MC(),
			0, 0,	// refp_x, refp_y
			0, 0,	// marg_x, marg_y
			string_lookup(text), font_id, fontsize);
		go_link (&gt->go, &trans_tlo->tlg->group, NULL);
		if (style)
			set_color (&gt->go);
	TRANS_FAILED
		DBG(buf);
	TRANS_END

	xfree (text);
	free (style);
	if (fontfamily != DEFAULT_SVG_FONT_FAMILY) // (is it text from const char?)
		free (fontfamily);
}

static void process_header (xmlTextReaderPtr ptr UNUSED)
{
	DBG("Processing header\n");
	GET_REAL_ATTRIBUTE(height, height_str);
}

static int parseDoc (char *docname)
{
	xmlTextReaderPtr ptr;
	xmlChar *tmp;

	ptr = xmlReaderForFile (docname, 0, 1 + 128);
	if (!ptr)
		return 0;
	while (xmlTextReaderRead (ptr) == 1)
	{
		if (xmlTextReaderNodeType (ptr) == XML_ELEMENT_NODE)
		{
#define TEST_IMPORT(str) else if (!strcmp (tmp, str)) import_##str(ptr)
#define TEST_SVG if (!strcmp (tmp, svg)) process_header(ptr)
			tmp = xmlTextReaderName (ptr);
			TEST_SVG;
			TEST_IMPORT(rect);
			TEST_IMPORT(ellipse);
			TEST_IMPORT(circle);
			TEST_IMPORT(line);
			TEST_IMPORT(polygon);
			TEST_IMPORT(polyline);
			TEST_IMPORT(path);
			TEST_IMPORT(text);
		}
	}
	xmlFreeTextReader(ptr);
	return 1;
}

int import_svg (char *filename, struct obj_tlo *tlo)
{
	char err_msg[TRANS_ERR_SIZE];
	int retval = 0;

	TRANS_BEGIN_MAIN (tlo, "SVG loading", err_msg)
		retval = parseDoc (filename);
	TRANS_FAILED
		err(err_msg);
		retval = -1;
	TRANS_END
	
	return retval;
}
