#include "kernel/internal.h"

void clipboard_init(void)
{
	char err[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN(tlo_universum, "Create clipboard", err)
		tlo_clipboard = tlo_new_named("clipboard");
		obj_ref(&tlo_clipboard->obj);
	TRANS_FAILED
		die("Cannot create clipboard.");
	TRANS_END
}

void clipboard_finish(void)
{
	obj_unref(&tlo_clipboard->obj);
}

static inline struct hanger* get_same_hanger(struct go *go1, struct hanger *h, struct go *go2)
	{ return go_get_hanger(go2, h - go_get_hanger(go1, 0)); }

static struct hanger* hcpy(struct hanger* orig)
{
	struct go *orig_go = hanger_get_go(orig), *copy_go = (struct go*)orig_go->temp_data;
	return copy_go ? get_same_hanger(orig_go, orig, copy_go) : hanger_new_mc_copy(orig);
}

#define CH(_i) hcpy(spec->a[_i].h)

#define COPY(_id, _call) case GO_TYPE(_id): { \
	UNUSED GO_STRUCT(_id) *spec = (GO_STRUCT(_id)*)go; \
	copy = &_call->go; \
	break; }

#define SUBCOPY(_id, _sid, _fce) case GO_SUBTYPE(_id, _sid): \
	copy = &_fce->go; \
	break;

static struct go* clipboard_copy_go(struct go *go)
{
	struct go *copy;
	struct prop *p;

	GO_TYPE_SWITCH(
		COPY(GROUP, go_group_new())
		COPY(POINT, go_point_new(CH(0)))
		COPY(SEGMENT, go_segment_new(CH(0), CH(1)))
		GO_SUBTYPE_SWITCH(BEZIER,
			SUBCOPY(BEZIER, CUBIC, go_bezier_new_cubic(CH(0), spec->geom.pt[0].w,  CH(1), spec->geom.pt[1].w, CH(2), spec->geom.pt[2].w, CH(3), spec->geom.pt[3].w))
			SUBCOPY(BEZIER, QUADRATIC, go_bezier_new_quadratic(CH(0), spec->geom.pt[0].w,  CH(1), spec->geom.pt[1].w, CH(2), spec->geom.pt[2].w))
		)
		GO_SUBTYPE_SWITCH(ELARC,
			SUBCOPY(ELARC, XYR, go_elarc_new_xyr(CH(0), spec->geom.rotation, spec->geom.a_radius, spec->geom.b_radius, spec->geom.start, spec->geom.dif, spec->conic))
			SUBCOPY(ELARC, FOCI, go_elarc_new_foci(CH(0), CH(1), CH(2), spec->conic))
			SUBCOPY(ELARC, 3SMALL, go_elarc_new_3small(CH(0), CH(1), CH(2), spec->conic))
			SUBCOPY(ELARC, 3ECC, go_elarc_new_3ecc(CH(0), CH(1), CH(2), spec->geom.rotation, spec->ecc, spec->conic))
			SUBCOPY(ELARC, XY1ECC, go_elarc_new_xy1ecc(CH(0), CH(1), spec->geom.rotation, spec->ecc, spec->conic))
		)
		COPY(PARPOINT, go_parametric_point_new_on_hanger(CH(0), spec->parameter, go_parametric_point_is_position_important(spec), spec->geom.point.x, spec->geom.point.y));
		COPY(INTERSECTION, go_intersection_new_on_hanger(CH(0), CH(1), spec->parameter, go_intersection_is_position_important(spec), spec->geom.point.x, spec->geom.point.y));
		COPY(TEXT, go_text_new(CH(0), spec->rpos.x, spec->rpos.y, spec->apos.x, spec->apos.y, spec->text, spec->font_id, spec->font_size))
		COPY(TEX, go_tex_text_new_process(CH(0), spec->rpos.x, spec->rpos.y, spec->apos.x, spec->apos.y, tex_process_ref(spec->process)))
		COPY(DECPOINT, go_decorator_point_new(CH(0), spec->n, spec->radius, spec->rotation))
		COPY(DIRECTION, go_direction_new(CH(0), spec->parameter, spec->n1, spec->n2, spec->radius, spec->rotation))
	)
		
	WALK_PROP(go2o(go), p)
		if(p->flags & PTF_SAVE)
			prop_create(go2o(copy), p->key, p->type, p->subtype,
					prop_subtype2quantity[p->type][p->subtype] == PQ_NONE ? 0 : prop_get_unit(go2o(go), p),
					prop_get_value(go2o(go), p));

	return copy;
}

void clipboard_copy_selected_go(struct go_group *source, struct go_group *target, struct go *after2)
{
	char buf[TRANS_ERR_SIZE];
	struct go *sel, *go, *copy;
	struct go_group *group;
	struct go* volatile after = after2;

	ASSERT_GROUP(source);
	ASSERT_GROUP(target);
	struct obj_tlo *source_tlo = go_get_tlo(&source->go);
	ASSERT(trans_tlo != source_tlo && trans_tlo == go_get_tlo(&target->go));
	ASSERT(!after || (after->kind == T_GO && go_get_group(after) == target));

	group_unselect_all_go(target);
	tsort_selected_presort(source, source_tlo);

	TRANS_BEGIN(buf)
		TSORT_NO_RECOMPUTE_BEGIN
			WALK_TSORT_GO(source_tlo, go)
				go->temp_data = (uintptr_t)clipboard_copy_go(go);

			WALK_GROUP_SELECTED_GO(source, sel)
			{
				if(sel->type == GOT_GROUP)
					die("Not implemented yet.\n");
				copy = (struct go*)sel->temp_data;
				go_link(copy, target, after);
				after = copy;
				go_select(copy);
				
				WALK_GO_RECURSIVE_WITHOUT_ROOT(sel, go)
				{
					copy = (struct go*)go->temp_data;
					group = (struct go_group*)go_get_group(go)->go.temp_data;
					go_link(copy, group, group_get_last_go(group));
					if(go_is_selected(go))
						go_select(copy);
				}
			}
		TSORT_NO_RECOMPUTE_END
	TRANS_FAILED
		trans_fail("Copying all selected go failed: %s", buf);
	TRANS_END

	tsort_presort_end(source_tlo);
}

uns clipboard_copy(struct go_group *source, const char* name_trans, char* error_string)
{
	volatile uns ret = 1;
	
	TRANS_BEGIN_MAIN(tlo_clipboard, name_trans, error_string)
		tlo_unlink_all_go();
		clipboard_copy_selected_go(source, tlo_get_tlg(tlo_clipboard), NULL);
	TRANS_FAILED
		ret = 0;
	TRANS_END

	return ret;
}

void clipboard_paste(struct go_group *target, struct go *after)
{
	clipboard_copy_selected_go(tlo_get_tlg(tlo_clipboard), target, after);
}

void clipboard_cut(struct go_group *source)
{
	char err[TRANS_ERR_SIZE];

	if(!clipboard_copy(tlo_get_tlg(trans_tlo), trans_get_last_gui(trans_tlo)->name_operation->text, err))
		trans_fail("Cannot copy selected objects into clipboard: %s", err);
		
	TRANS_BEGIN(err)
		group_remove_rehang_selected_go(source);
	TRANS_FAILED
		trans_undo(tlo_clipboard, err);
		trans_fail("Cannot remove selected go: %s", err);
	TRANS_END
}

