#include "kernel/internal.h"
#include <stdio.h>
#include <string.h>


/*
 * DUMP O
 */

UNUSED static void o_dump_short(struct o *o)
{
	ASSERT_O(o);
	if(o->kind == T_OBJ)
		msg("%s %s at %p", string_obj_type[o->type]->text, o_lookup_name(o), o);
	else
		msg("%s %s at %p", string_go_type[o->type]->text, o_lookup_name(o), o);
}

void o_dump(struct o *o)
{
	ASSERT_O(o);
	if(o->kind == T_OBJ)
		msg("Name %s, address %p, kind[%d] obj, type[%d] %s, subtype %d, ref_count %d.\n", o_lookup_name(o), o, o->kind, o->type, string_obj_type[o->type]->text, o->subtype, o->ref_count);
	else
		msg("Name %s, address %p, kind[%d] go, type[%d] %s, subtype[%d] %s, ref_count %d.\n", o_lookup_name(o), o, o->kind, o->type, string_go_type[o->type]->text, o->subtype, go_get_info(o2go(o))->name->text, o->ref_count);
}

void o_dump_prop(struct o *o)
{
	ASSERT_O(o);
	struct prop* p;
	char buf[PROP_DUMP_SIZE];
	msg("o_dump_prop:\n");
	WALK_PROP(o, p)
		msg("\t%s\n", prop_dump(buf, o, p));
}


/*
 * DUMP OBJ
 */

void obj_dump_father(struct obj* obj)
{
	ASSERT_OBJ(obj);
	int i;
	msg("obj_dump_father:\n");
	for(i = 1, obj = obj->father; obj; obj = obj->father, i++)
	{
		msg("\tfather %d: ", i);
		o_dump(obj2o(obj));
	}
}

void obj_dump_son(struct obj* obj)
{
	ASSERT_OBJ(obj);
	struct obj* son;
	msg("obj_dump_son: ");
	WALK_OBJ_SON(obj, son)
		msg("%s at %p, ", obj_lookup_name(son), son);
	msg("\n");
}

void obj_dump_subtree(struct obj* root)
{
	ASSERT_OBJ(root);
	struct obj *walk = root;
	char header[34], *pheader; // Maximum depth in the name-space tree is 10 (for this function).

	msg("\nobj_dump_subtree:\n");

	for(uns i = 0; i < 30; i += 3)
	{
		header[i] = ' ';
		header[i+1] = '|';
		header[i+2] = ' ';
	}
	header[30] = ' ';
	header[31] = '-';
	header[32] = ' ';
	header[33] = '\0';
	pheader = header + 33;

	for(;;)
	{
		msg("\t%s%s at %p", pheader, obj_lookup_name(walk), walk);
		if(walk->type == OT_TLO)
			msg(", tlo is %s", tlo_is_saved(obj2tlo(walk)) ? "saved" : "unsaved");
		else if(walk->type == OT_DOCUMENT)
			msg(", doc is %s", doc_is_saved(obj2doc(walk)) ? "saved" : "unsaved");
		msg("\n");

		if(obj_get_first_son(walk))
		{
			walk = obj_get_first_son(walk);
			pheader -= 3;
		}
		else
			for(;;)
			{
				if(walk == root)
					return;
				if(obj_get_next_son(walk))
				{
					walk = obj_get_next_son(walk);
					break;
				}
				else
				{
					walk = walk->father;
					pheader += 3;
				}
			}
	}
}

void obj_dump_flag(struct obj *obj)
{
	ASSERT_OBJ(obj);
	uns f = obj->flags;
	msg("obj_dump_flags: flags 0x%x, selected %s\n", f, (f & OF_SELECTED) ? "YES" : "NO");
}

void obj_dump_hook(struct obj* obj)
{
	ASSERT_OBJ(obj);
	struct go_hook *gh;
	struct trans_hook *th;
	msg("obj_dump_hook:\n");

	if(obj->type == OT_TLO)
	{
		struct obj_tlo *tlo = obj2tlo(obj);
		msg("\n\tgo_hooks: ");
		WALK_GO_HOOK(tlo, gh)
			msg("%p, ", gh);
		msg("\n\ttrans_hooks: ");
		WALK_TRANS_HOOK(tlo, th)
			msg("%p, ", th);
	}
	msg("\n");
}

void obj_dump_selected(void)
{
	struct obj* obj;
	msg("obj_dump_selected: ");
	WALK_OBJ_SELECTED(obj)
		msg("%s at %p, ", obj_lookup_name(obj), obj);
	msg("\n");
}

void tlo_dump_bbox(struct obj_tlo *tlo)
{
	ASSERT_TLO(tlo);
	struct geom_rectangle *r = &tlo->rtree.root->bbox;
	msg("tlo_dump_bbox: ");
	if(!geom_rectangle_valid(r))
		msg("! bbox invalid ! ");
	msg("ll.x = %f, ll.y = %f, ur.x = %f, ur.y = %f\n", r->ll.x, r->ll.y, r->ur.x, r->ur.y);
}

void obj_dump_all(struct obj* obj)
{
	ASSERT_OBJ(obj);
	msg("obj_dump_all: ");
	o_dump(obj2o(obj));
	obj_dump_father(obj);
	obj_dump_subtree(obj);
	o_dump_prop(obj2o(obj));
	obj_dump_flag(obj);
	if(obj->type == OT_TLO)
		tlo_dump_bbox(obj2tlo(obj));
}


/*
 * DUMP GO
 */

void go_dump_hanger(struct go *go)
{
	ASSERT(go->kind == T_GO);
	struct hanger *h;
	msg("go_dump_hanger: ");
	WALK_GO_HANGER_BEGIN(go, h)
		msg("%s at %p, ", h->info->name->text, h);
	WALK_GO_HANGER_END
	msg("\n");
}

void go_dump_flag(struct go* go)
{
	ASSERT_GO(go);
	uns f = go->flags;
	msg("go_dump_flags: flags 0x%x, selected %s, tsort %s\n", f, (f & GOF_SELECTED) ? "YES" : "NO", (f & GOF_TSORT) ? "YES" : "NO");
}

void go_dump_type(struct go* go)
{
	switch(go->type)
	{
#define DUMP(_id) case GO_TYPE(_id): GLUEE(GO_NAME(_id), _dump) ((GO_STRUCT(_id)*)go); break;
		GO_FOR_ALL_ID(DUMP)

		default:
			die("Function go_dump hasn't label for the go type %d - %s.\n", go->type, string_go_type[go->type]->text);
	}
}

void go_dump_all(struct go* go)
{
	ASSERT_GO(go);
	msg("go_dump_all: go = ");
	o_dump(go2o(go));
	o_dump_prop(go2o(go));
	go_dump_anchor(go);
	go_dump_hanger(go);
	go_dump_flag(go);
	go_dump_bbox(go);
	go_dump_type(go);
}

void go_dump_anchor(struct go *go)
{
	ASSERT_GO(go);
	msg("go_dump_anchor: \n");
	struct anchor *a;
	int i;
	WALK_GO_ANCHOR(go, a, i)
	{
		struct go *g = hanger_get_go(a->h);
		if(hanger_is_position(a->h))
		{
			struct geom_point gp = hanger_get_position(a->h);
			msg("\tanchor %i at %p, name %s, hanger %s at %p position [ %f, %f ], go %s at %p\n", -i, a, a->info->name->text, a->h->info->name->text, a->h, gp.x, gp.y, go_lookup_name(g), g);
		}
		else
			msg("\tanchor %i at %p, name %s, hanger %s at %p, go %s at %p\n", -i, a, a->info->name->text, a->h->info->name->text, a->h, go_lookup_name(g), g);
	}
}

void go_dump_bbox(struct go* go)
{
	ASSERT_GO(go);
	struct obj_tlo *tlo = go_get_tlo(go);
	msg("go_dump_bbox: ");
	if(tlo->expand)
	{
		const struct geom_rectangle *r = go_get_expand_bbox(tlo, go);
		if(r) {
			GEOM_DUMP_INLINE(geom_rectangle_dump(r));
		}
		else
			msg("go is invalid");
	}
	else
		msg("tlo haven't expansion");
	msg("\n");
}

void group_dump_go(struct go_group *group)
{
	ASSERT_GROUP(group);
	msg("group_dump_go: ");
	struct go* go;
	WALK_GROUP_GO(group, go)
		msg("%s at %p, ", go_lookup_name(go), go);
	msg("\n");
}

void group_dump_subtree(struct go_group *root)
{
	ASSERT_GROUP(root);
	msg("\ngroup_dump_subtree: ");
	o_dump(go2o(&root->go));
	struct go* go;
	struct go_group *owner;
	char header[3004], *pheader; // Maximum depth in the name-space tree is 1000 (for this function).

	for(uns i = 0; i < 3000; i += 3)
	{
		header[i] = ' ';
		header[i+1] = '|';
		header[i+2] = ' ';
	}
	header[3000] = ' ';
	header[3001] = '-';
	header[3002] = ' ';
	header[3003] = '\0';
	pheader = header + 3000;

	msg("\t%s%s at %p\n", pheader + 3, go_lookup_name(&root->go), root);
	WALK_GROUP_SMART(root, 1, go, owner,
			msg("\t%s%s at %p\n", pheader, go_lookup_name(&owner->go), owner); pheader -= 3,
			msg("\t%s%s at %p\n", pheader, go_lookup_name(go), go); ASSERT(go_get_group(go) == owner),
			pheader += 3);
}

void group_dump_selected_go(struct go_group *group)
{
	ASSERT_GROUP(group);
	msg("group_dump_selected_go: ");
	struct go* go;
	WALK_GROUP_SELECTED_GO(group, go)
		msg("%s at %p, ", go_lookup_name(go), go);
	msg("\n");
}

static void tlo_dump_tsort_go(struct go *go, void UNUSED *p)
{
	ASSERT_GO(go);
	struct hanger* h;
	struct anchor* a;
	struct go *g;
	int i;
	msg("%s at %p subtype %s, group %s at %p\n", go_lookup_name(go), go, go_get_info(go)->name->text,
			go_lookup_name(&go_get_group(go)->go), go_get_group(go));
	WALK_GO_ANCHOR(go, a, i)
	{
		h = a->h;
		g = hanger_get_go(h);
		msg("\tanchor %s at %p, hanger %s at %p, go %s at %p subtype %s\n", a->info->name->text, a,
				h->info->name->text, h, go_lookup_name(g), g, go_get_info(g)->name->text);
	}
	WALK_GO_HANGER_BEGIN(go, h)
	{
		msg("\thanger %s at %p\n", h->info->name->text, h);
		WALK_HANGER_ANCHOR(h, a)
		{
			g = anchor_get_go(a);
			msg("\t\tanchor %s at %p, go %s at %p subtype %s\n", a->info->name->text, a,
					go_lookup_name(g), g, go_get_info(g)->name->text);
		}
	}
	WALK_GO_HANGER_END
}

void tlo_dump_tsort(struct obj_tlo* tlo)
{
	ASSERT_TLO(tlo);
	msg("\ntlo_tsort_dump: ");
	o_dump(void2o(tlo));
	tsort_start(tlo);
	tsort_insert_group(&tlo->tlg->group, tlo);
	tsort_call(tlo, tlo_dump_tsort_go, NULL);
}

#define MSG_UNIT(_name) msg("\tunit_" #_name " = %u, %s\n", go->unit_##_name, go->unit_##_name == PROP_UNIT_MAX ? "default unit" : unit_array[go->unit_##_name].name->text)

void go_group_dump(struct go_group *go UNUSED)
{
	ASSERT_GROUP(go);
	msg("go_group_dump:\n");
}

void go_point_dump(struct go_point *go)
{
	ASSERT_GO_TYPE(go, POINT);
	msg("go_point_dump:\n");

	msg("\tx = { %f, %f }\n", go->geom.point.x, go->geom.point.y);

	MSG_UNIT(coordinate_x);
	MSG_UNIT(coordinate_y);
}

void go_segment_dump(struct go_segment *go)
{
	ASSERT_GO_TYPE(go, SEGMENT);
	msg("go_segment_dump:\n");

	msg("\tgeom.a = { %f, %f }\n", go->geom.a.x, go->geom.a.y);
	msg("\tgeom.b = { %f, %f }\n", go->geom.b.x, go->geom.b.y);

	MSG_UNIT(start_x);
	MSG_UNIT(start_y);
	MSG_UNIT(end_x);
	MSG_UNIT(end_y);
}

void go_bezier_dump(struct go_bezier *go)
{
	ASSERT_GO_TYPE(go, BEZIER);
	int i;
	msg("go_bezier_dump:\n");

	msg("\tgeom.flags = 0x%x\n", GEOM_INSTANCE_FLAGS(&go->geom));
	for(i = 0; i <= GEOM_BEZIER_DEGREE(&go->geom); i++)
		msg("\tgeom.pt[%d] = { %f, %f }, weight = %f\n", i, go->geom.pt[i].x, go->geom.pt[i].y, go->geom.pt[i].w);

	MSG_UNIT(control_0_x);
	MSG_UNIT(control_0_y);
	MSG_UNIT(control_1_x);
	MSG_UNIT(control_1_x);
	MSG_UNIT(control_2_y);
	MSG_UNIT(control_2_x);
	MSG_UNIT(control_3_x);
	MSG_UNIT(control_3_y);
	if(GEOM_BEZIER_DEGREE(&go->geom) == 4) /* Procpak? */
	{
		MSG_UNIT(control_0_y);
		MSG_UNIT(control_0_y);
	}
}

void go_elarc_dump(struct go_elarc *go)
{
	ASSERT_GO_TYPE(go, ELARC);
	msg("go_elarc_dump:\n");

	msg("\tgeom.center = { %f, %f }\n", go->geom.center.x, go->geom.center.y);
	msg("\tgeom.rotation = %f\n", go->geom.rotation);
	msg("\tgeom.a_radius = %f\n", go->geom.a_radius);
	msg("\tgeom.b_radius = %f\n", go->geom.b_radius);
	msg("\tgeom.start = %f\n", go->geom.start);
	msg("\tgeom.dif = %f\n", go->geom.dif);
	msg("\tecc = %f\n", go->ecc);
	msg("\tconic = %d, %s\n", go->conic, string_conic_type[go->conic]->text);

	MSG_UNIT(center_x);
	MSG_UNIT(center_y);
	MSG_UNIT(rotation);
	MSG_UNIT(a_radius);
	MSG_UNIT(b_radius);
	MSG_UNIT(start);
	MSG_UNIT(dif);
}

void go_parametric_point_dump(struct go_parametric_point *go)
{
	ASSERT_GO_TYPE(go, PARPOINT);
	msg("go_parametric_point_dump:\n");

	msg("\tx = { %f, %f }\n", go->geom.point.x, go->geom.point.y);
	msg("\tparameter = %f\n", go->parameter);

	MSG_UNIT(coordinate_x);
	MSG_UNIT(coordinate_y);
	MSG_UNIT(parameter);
}

void go_intersection_dump(struct go_intersection *go)
{
	ASSERT_GO_TYPE(go, INTERSECTION);
	msg("go_intersection_dump:\n");

	msg("\tx = { %f, %f }\n", go->geom.point.x, go->geom.point.y);
	msg("\tparameter = %f\n", go->parameter);

	MSG_UNIT(coordinate_x);
	MSG_UNIT(coordinate_y);
	MSG_UNIT(parameter);
}

void go_text_dump(struct go_text *go)
{
	ASSERT_GO_TYPE(go, TEXT);
	msg("go_text_dump:\n");

	msg("\tapos = [ %f, %f ]\n", go->apos.x, go->apos.y);
	msg("\trpos = [ %f, %f ]\n", go->rpos.x, go->rpos.y);
	msg("\ttransformation = "); GEOM_DUMP_INLINE(geom_transform_dump(go->transformation, 0)); msg("\n");
	msg("\tfont_id = %i\n", go->font_id);
	msg("\tfont_size = %f\n", go->font_size);
	msg("\ttext = %s\n", go->text->text);

	MSG_UNIT(rpos_x);
	MSG_UNIT(rpos_y);
	MSG_UNIT(apos_x);
	MSG_UNIT(apos_y);
	MSG_UNIT(coordinate_x);
	MSG_UNIT(coordinate_y);
	MSG_UNIT(size);
}

void go_tex_text_dump(struct go_tex_text *go)
{
	ASSERT_GO_TYPE(go, TEX);
	msg("go_tex_text_dump:\n");

	msg("\tapos = [ %f, %f ]\n", go->apos.x, go->apos.y);
	msg("\trpos = [ %f, %f ]\n", go->rpos.x, go->rpos.y);
	msg("\ttransformation = "); GEOM_DUMP_INLINE(geom_transform_dump(go->transformation, 0)); msg("\n");
	msg("\ttext = %s\n", go->process->text->text);

	MSG_UNIT(rpos_x);
	MSG_UNIT(rpos_y);
	MSG_UNIT(apos_x);
	MSG_UNIT(apos_y);
	MSG_UNIT(coordinate_x);
	MSG_UNIT(coordinate_y);
}

void go_decorator_point_dump(struct go_decorator_point *go)
{
	ASSERT_GO_TYPE(go, DECPOINT);
	msg("go_decorator_point_dump:\n");

	msg("\tx = { %f, %f }\n", go->x.x, go->x.y);
	msg("\tn = %u\n", go->n);
	msg("\tradius = %f\n", go->radius);
	msg("\trotation = %f\n", go->rotation);

	MSG_UNIT(coordinate_x);
	MSG_UNIT(coordinate_y);
	MSG_UNIT(radius);
	MSG_UNIT(rotation);
}

void go_direction_dump(struct go_direction *go)
{
	ASSERT_GO_TYPE(go, DIRECTION);
	msg("go_direction_dump:\n");

	msg("\tx = { %f, %f }\n", go->x.x, go->x.y);
	msg("\tn1 = %u\n", go->n1);
	msg("\tn2 = %u\n", go->n2);
	msg("\tradius = %f\n", go->radius);
	msg("\trotation = %f\n", go->rotation);

	MSG_UNIT(coordinate_x);
	MSG_UNIT(coordinate_y);
	MSG_UNIT(radius);
	MSG_UNIT(rotation);
}

/*
 * DUMP HANGER AND ANCHOR
 */

void hanger_dump(struct hanger* h)
{
	ASSERT_HANGER(h);
	struct geom_point gp = hanger_get_position(h);
	msg("hanger_dump: name %s, address %p, go %s at %p, position [%f, %f].\n", h->info->name->text, h, go_lookup_name(hanger_get_go(h)), hanger_get_go(h), gp.x, gp.y);
}

void hanger_dump_anchor(struct hanger *h)
{
	struct anchor* a;
	struct geom_point gp = hanger_get_position(h);
	msg("hanger_dump_anchor: name %s, address %p, go %s at %p, position [%f, %f].\n", h->info->name->text, h, go_lookup_name(hanger_get_go(h)), hanger_get_go(h), gp.x, gp.y);
	WALK_HANGER_ANCHOR(h, a)
		msg("\tname %s, address %p, go %s at %p\n.", a->info->name->text, a, go_lookup_name(anchor_get_go(a)), anchor_get_go(a));
}

void anchor_dump(struct anchor* a)
{
	ASSERT_ANCHOR(a);
	struct geom_point gp = hanger_get_position(a->h);
	msg("anchor_dump: name %s, address %p, hanger %s at %p position [ %f, %f ].", a->info->name->text, a, a->h->info->name->text, a->h, gp.x, gp.y);
	msg("a->go: ");
	o_dump(go2o(anchor_get_go(a)));
	msg("a->h->go: ");
	o_dump(go2o(hanger_get_go(a->h)));
}

void trans_dump_undo_kernel(struct undo_gui *ugo)
{
	struct undo_kernel* uk;
	char buf[PROP_DUMP_SIZE];
	WALK_UNDO_KERNEL(ugo, uk)
	{
		switch(uk->type)
		{
			case TRANS_OBJ_SELECT:
				msg("\tobj_select: obj %s at %p\n", obj_lookup_name(uk->u.obj_link.obj), uk->u.obj_link.obj);
				break;

			case TRANS_OBJ_UNSELECT:
				msg("\tobj_unselect: obj %s at %p\n", obj_lookup_name(uk->u.obj_link.obj), uk->u.obj_link.obj);
				break;

			case TRANS_OBJ_LINK:
				msg("\tobj_link: obj %s at %p, father %s at %p\n", obj_lookup_name(uk->u.obj_link.obj), uk->u.obj_link.obj, obj_lookup_name(uk->u.obj_link.father), uk->u.obj_link.father);
				break;

			case TRANS_OBJ_UNLINK:
				msg("\tobj_unlink: obj %s at %p, father %s at %p\n", obj_lookup_name(uk->u.obj_link.obj), uk->u.obj_link.obj, obj_lookup_name(uk->u.obj_link.father), uk->u.obj_link.father);
				break;

			case TRANS_GO_SELECT:
				msg("\tgo_select: go %s at %p\n", go_lookup_name(uk->u.go_link.go), uk->u.go_link.go);
				break;

			case TRANS_GO_UNSELECT:
				msg("\tgo_unselect: go %s at %p\n", go_lookup_name(uk->u.go_link.go), uk->u.go_link.go);
				break;

			case TRANS_GO_LINK:
				msg("\tgo_link: go %s at %p\n", go_lookup_name(uk->u.go_link.go), uk->u.go_link.go);
				break;

			case TRANS_GO_UNLINK:
				msg("\tgo_unlink: go %s at %p\n", go_lookup_name(uk->u.go_link.go), uk->u.go_link.go);
				break;

			case TRANS_ANCHOR_REHANG:
				msg("\tanchor_rehang: anchor %p, hanger %p\n",  uk->u.anchor_rehang.anchor, uk->u.anchor_rehang.hanger);
				break;

			case TRANS_MOUSE_CLICK_MOVE:
				msg("\tmouse click move: hanger %p, position [%f, %f]\n",  uk->u.mc_move.mc, uk->u.mc_move.gp.x, uk->u.mc_move.gp.y);
				break;

#define DUMP_PROP(_case) case _case: msg("\t" MKSTR(_case) ": o %s at %p, %s\n", o_lookup_name(uk->u.prop.o), uk->u.prop.o, prop_dump_value(buf, uk->u.prop.prop, uk->u.prop.value)); break;
			DUMP_PROP(TRANS_PROP_CHANGE)
			DUMP_PROP(TRANS_PROP_CREATE)
			DUMP_PROP(TRANS_PROP_DELETE)
			DUMP_PROP(TRANS_PROP_VIRTUAL)
			DUMP_PROP(TRANS_PROP_VIRTUAL_START_UNDO)
			DUMP_PROP(TRANS_PROP_VIRTUAL_END_UNDO)

			default:
				die("There in no label for this type");
		}
	}
}

void trans_dump_undo_gui(struct obj_tlo* tlo)
{
	ASSERT(tlo->obj.kind == T_OBJ && (tlo->obj.type == OT_TLO || tlo->obj.type == OT_UNIVERSUM));
	struct undo_gui *ug;
	WALK_UNDO_GUI(tlo, ug)
		msg(tlo->undo_present == ug ? "%s - topical command, " : "%s, ", ug->name_operation->text);
	msg("\n");
}

void trans_dump_undo(struct obj_tlo* tlo)
{
	ASSERT(tlo->obj.kind == T_OBJ && (tlo->obj.type == OT_TLO || tlo->obj.type == OT_UNIVERSUM));
	struct undo_gui *ug;
	WALK_UNDO_GUI(tlo, ug)
	{
		msg(tlo->undo_present == ug ? "%s - topical command\n" : "%s\n", ug->name_operation->text);
		if(ug != &tlo->undo_start)
			trans_dump_undo_kernel(ug);
	}
}


/*
 * DUMP PROPERTY
 */

#define PROP_DUMP_CASE(_case, _str, ...) case _case: \
	written = snprintf(buf, PROP_DUMP_SIZE - 1, "key %s, type[%d] %s, subtype[%d] %s, "_str, \
	prop->key->text, prop->type, string_prop_type[prop->type]->text, _case, #_case, __VA_ARGS__); break;

char* prop_dump_value(char buf[PROP_DUMP_SIZE], struct prop* prop, prop_value value)
{
	int written;
	switch(prop->type)
	{
		case PT_UNS:
			switch(prop->subtype)
			{
				PROP_DUMP_CASE(PTU_UNSPECIFIED, "%d", value.u)
				PROP_DUMP_CASE(PTU_COLOR, "red %d, green %d, blue %d, alpha %d", uns2r_uns(value.u), uns2g_uns(value.u), uns2b_uns(value.u), uns2a_uns(value.u))
				PROP_DUMP_CASE(PTU_FONT, "value %d", value.u); // FIXME: Nazev fontu.
				PROP_DUMP_CASE(PTU_CONIC_TYPE_0P, "value[%d] %s", value.u, string_conic_type[value.u]->text);
				PROP_DUMP_CASE(PTU_CONIC_TYPE_1P, "value[%d] %s", value.u, string_conic_type[value.u]->text);
				PROP_DUMP_CASE(PTU_CONIC_TYPE_2P, "value[%d] %s", value.u, string_conic_type[value.u]->text);
				PROP_DUMP_CASE(PTU_CONIC_TYPE_3P, "value[%d] %s", value.u, string_conic_type[value.u]->text);
				PROP_DUMP_CASE(PTU_BOOLEAN, "value[%d] %s", value.u, value.u ? "YES" : "NO");
				PROP_DUMP_CASE(PTU_CAP_STYLE, "value[%d] %s", value.u, string_stroke_cap_type[value.u]->text);

				default:
					die("Invalid property subtype\n");
			}
			break;

		case PT_POINTER:
			switch(prop->subtype)
			{
				PROP_DUMP_CASE(PTP_UNSPECIFIED, "value %p", value.v)
				PROP_DUMP_CASE(PTP_TRANSFORM, "value %p: flags=0x%x, Tx=%fx+%fy+%f, Ty=%fx+%fy+%f",
							(void*)value.t, value.t->flags,
							value.t->coef[0][0], value.t->coef[0][1], value.t->coef[0][2],
							value.t->coef[1][0], value.t->coef[1][1], value.t->coef[1][2])

				default:
					die("Invalid property subtype\n");
			}
			break;

		PROP_DUMP_CASE(PT_STRING, "value %s", value.s->text)
		PROP_DUMP_CASE(PT_REAL, "value %f", value.r)

		default:
			die("Invalid property type\n");
	}
	DANGEROUS(written <= PROP_DUMP_SIZE, "Error message is too long. Please, make constant PROP_DUMP_SIZE (kernel/internal.h) not be less than %d.\n", written);
	DANGEROUS(written >= 0, "Error message is too long. Please, make constant PROP_DUMP_SIZE (kernel/internal.h) larger.\n");
	return buf;
}

char* prop_dump(char buf[PROP_DUMP_SIZE], struct o *o, struct prop *prop)
{
	return prop_dump_value(buf, prop, prop_get_value(o, prop));
}

/*
void kernel_dump_memory(void)
{
#ifdef __GNUC__
	uns oc[OT_MAX], gc[GOT_MAX], hc = 0, ac = 0, mcc = 0, i, ugc = 0, ukc = 0;

	static void count_tlo(struct obj_tlo* tlo)
	{
		struct go* go;
		struct hanger* h;
		struct undo_gui* ugo;
		struct undo_kernel* uk;

		WALK_TLO_GO(tlo, go)
		{
			gc[go->type]++;
			ac += go->anchor_countlo;
			WALK_GO_HANGER(go, h)
				hc++;
		}
		WALK_GO_HANGER(&tlo->mouse_click, h)
			mcc++;
		WALK_UNDO_GUI(tlo, ugo)
		{
			ugc++;
			WALK_UNDO_KERNEL(ugobj, uk)
				ukc++;
		}
	}

	static void count_tree(struct obj* obj)
	{
		struct obj* obj;
		WALK_OBJ_DESCENDANTS(obj, obj)
		{
			oc[obj->type]++;
			if(obj->type == OT_TLO)
				count_tlo((struct obj_tlo*)obj);
		}
	}

	bzero(oc, sizeof(uns) * OT_MAX);
	bzero(gc, sizeof(uns) * GOT_MAX);

	count_tree(obj_universum);
	count_tree(obj_zombie);
	count_tlo(tlo_universum);
	count_tlo(tlo_zombie);

	for(i = 0; i < OT_MAX; i++)
		msg("%s:  %d * %dB = %dKB\n", string_obj_type[i]->text, 0, oc[i], 0);
	for(i = 0; i < GOT_MAX; i++)
		msg("%s:  %d * %dB = %dKB\n", string_go_type[i]->text, 0, gc[i], 0);
	msg("hanger:  %d * %dB = %dKB\n", hc, sizeof(struct hanger), sizeof(struct hanger) * hc / 1024);
	msg("anchor:  %d * %dB = %dKB\n", ac, sizeof(struct hanger), sizeof(struct anchor) * ac / 1024);
	msg("mouse click: %d * %dB = %dKB\n", mcc, sizeof(struct hanger),  sizeof(struct hanger) * mcc / 1024);
	msg("undo gui: %d * %dB = %dKB\n", ugc, sizeof(struct undo_gui),  sizeof(struct undo_gui) * ugc / 1024);
	msg("undo kernel: %d * %dB = %dKB\n", ukc, sizeof(struct undo_kernel),  sizeof(struct undo_kernel) * ukc / 1024);
#endif
}
*/
