/** \file kernel/glt_common.h
  \brief GUILE-link - interface functions for common types. */

#ifndef _KERNEL_GLT_COMMON_H_
#define _KERNEL_GLT_COMMON_H_

static inline void assert_SINT(SCM obj, unsigned int pos, const char *subr)
{ scm_num2int (obj, pos, subr); }

static inline void assert_SUNS(SCM obj, unsigned int pos, const char *subr)
{ scm_num2uint (obj, pos, subr); }

static inline void assert_SREAL(SCM obj, unsigned int pos, const char *subr)
{ scm_num2float (obj, pos, subr); }

static inline void assert_SSTRINGC(SCM obj, unsigned int pos, const char *subr)
{ SCM_ASSERT (SCM_STRINGP (obj), obj, pos, subr); }

static inline void assert_SSYMBOLC(SCM obj, unsigned int pos, const char *subr)
{ SCM_ASSERT (scm_symbol_p (obj), obj, pos, subr); }

static inline void assert_SBOOL(SCM obj, unsigned int pos, const char *subr)
{ SCM_ASSERT (SCM_BOOLP (obj), obj, pos, subr); }


static inline int in_SINT (SCM number)
{ return scm_num2int (number, SCM_ARGn, "num2int"); }

static inline uns in_SUNS (SCM number)
{ return scm_num2uint (number, SCM_ARGn, "num2uint"); }

static inline real in_SREAL(SCM number)
{ return scm_num2float (number, SCM_ARGn, "num2float"); }

static inline const char * in_SSTRINGC(SCM str)
{ return SCM_STRING_CHARS (str); }

static inline const char * in_SSYMBOLC(SCM str)
{ return SCM_SYMBOL_CHARS (str); }

static inline int in_SBOOL (SCM bool)
{ return SCM_NFALSEP (bool); }


static inline SCM out_SINT (int number)
{ return scm_int2num (number); }

static inline SCM out_SUNS (uns number)
{ return scm_uint2num (number); }

static inline SCM out_SREAL (real number)
{ return scm_float2num (number); }

static inline SCM out_SSTRINGC (const char *str)
{ return scm_makfrom0str (str); }

static inline SCM out_SSYMBOLC (const char *str)
{ return scm_str2symbol (str); }

static inline SCM out_SBOOL(int bool)
{ return SCM_BOOL (bool); }


#endif

