/** \file kernel/glt_kernel.h
  \brief GUILE-link - interface functions for kernel data types. */

#ifndef _KERNEL_GLT_KERNEL_H_
#define _KERNEL_GLT_KERNEL_H_


extern SCM symbol_obj;
extern SCM symbol_go;
extern SCM symbol_hanger;
extern SCM symbol_anchor;
extern SCM symbol_transaction_failed;
extern SCM symbol_condition_not_satisfied;

static inline struct o * optr (SCM obj) { return (void *) SCM_SMOB_DATA(obj); }
static inline struct hanger * hptr (SCM obj) { return (void *) SCM_SMOB_DATA(obj); }
static inline struct anchor * aptr (SCM obj) { return (void *) SCM_SMOB_DATA(obj); }

int is_o_proxy(SCM proxy);
int is_hanger_proxy(SCM proxy);
int is_anchor_proxy(SCM proxy);
int is_specific_proxy(SCM proxy, int kind, int type, int subtype);

static inline struct o * scm2o (SCM obj)
{ return is_o_proxy (obj) ? optr (obj) : NULL; }

static inline struct hanger * scm2hanger (SCM obj)
{ return is_hanger_proxy (obj) ? hptr (obj) : NULL; }

static inline struct anchor * scm2anchor (SCM obj)
{ return is_anchor_proxy (obj) ? aptr (obj) : NULL; }

SCM o2scm(struct o *p);
SCM hanger2scm(struct hanger *p);
SCM anchor2scm(struct anchor *p);

void condition_not_satisfied(const char *name, const char *dsc);


// FIXME: Vymyslet spravne podminky.
#define ASSERT_META_TRANS_SCHEME(name) if (! (trans_present && trans_tlo == tlo_universum)) condition_not_satisfied (name, "Function not called in meta transaction");
#define ASSERT_GO_TRANS_SCHEME(name) if (! (trans_present && trans_tlo && trans_tlo->obj.type == OT_TLO)) condition_not_satisfied (name, "Function not called in go transaction");
// #define ASSERT_O_TRANS_SCHEME(name, o) if (! (((o)->kind == T_OBJ) ? CONDITION_META_TRANS : CONDITION_CORRECT_GO(o2go(o)))) condition_not_satisfied (name, "Function not called in appropriate transaction");

#define ASSERT_O_TRANS_SCHEME(name, o)


#define ID_obj T_OBJ, -1
#define ID_obj_doc T_OBJ, OT_DOCUMENT
#define ID_obj_tlo T_OBJ, OT_TLO
#define ID_go T_GO, -1
#define ID_go_point T_GO, GOT_POINT
#define ID_go_segment T_GO, GOT_SEGMENT
#define ID_go_arc T_GO, GOT_ARC
#define ID_go_elarc T_GO, GOT_ELARC
#define ID_go_bezier T_GO, GOT_BEZIER
#define ID_go_text T_GO, GOT_TEXT
#define ID_go_tex_text T_GO, GOT_TEX_TEXT
#define ID_go_parametric_point T_GO, GOT_PARAMETRIC_POINT
#define ID_go_intersection T_GO, GOT_INTERSECTION
#define ID_go_decorator_point T_GO, GOT_DECORATOR_POINT
#define ID_go_direction T_GO, GOT_DIRECTION
#define ID_go_group T_GO, GOT_GROUP
#define ID_o -1, -1

static inline void assert_SO(SCM obj, unsigned int pos, const char *subr, uns kind, uns type)
{ SCM_ASSERT (SCM_FALSEP (obj) || is_specific_proxy (obj, kind, type, -1), obj, pos, subr); }

static inline void assert_SHANGER(SCM obj, unsigned int pos, const char *subr)
{ SCM_ASSERT (is_hanger_proxy (obj), obj, pos, subr); }

static inline void assert_SANCHOR(SCM obj, unsigned int pos, const char *subr)
{ SCM_ASSERT (is_anchor_proxy (obj), obj, pos, subr); }

static inline void assert_SSTRINGS(SCM obj, unsigned int pos, const char *subr)
{ SCM_ASSERT (SCM_STRINGP (obj), obj, pos, subr); }

static inline void assert_SSYMBOLS(SCM obj, unsigned int pos, const char *subr)
{ SCM_ASSERT (scm_symbol_p (obj), obj, pos, subr); }

static inline void * in_SO (SCM obj)
{ return scm2o (obj); }

static inline struct hanger * in_SHANGER (SCM obj)
{ return scm2hanger (obj); }

static inline struct anchor * in_SANCHOR (SCM obj)
{ return scm2anchor (obj); }

static inline string in_SSTRINGS (SCM str)
{ return string_lookup (SCM_STRING_CHARS (str)); }
// ^^^ FIXME allocate ref

static inline string in_SSYMBOLS (SCM str)
{ return string_lookup (SCM_SYMBOL_CHARS (str)); }
// ^^^ FIXME allocate ref


static inline SCM out_SO (void *ptr)
{ return o2scm (ptr); }

static inline SCM out_SHANGER (void *ptr)
{ return hanger2scm (ptr); }

static inline SCM out_SANCHOR (void *ptr)
{ return anchor2scm (ptr); }

static inline SCM out_SSTRINGS (string ptr)
{ return scm_makfrom0str (ptr -> text); }

static inline SCM out_SSYMBOLS (string ptr)
{ return scm_str2symbol (ptr -> text); }


#endif
