//#define LOCAL_DEBUG
#include "kernel/internal.h"

struct geom_point* point_transform(struct geom_point *p, const struct geom_transform *t)
{
	if(geom_point_transform(p, t, p) != 0)
		TRANS_FAIL(("Numeric error."), ("geom_point_transform failed\n"));
	return p;
}

#define RECOMPUTE(_id) case GO_TYPE(_id): GLUEE(GO_NAME(_id), _recompute) ((GO_STRUCT(_id)*)go); break;
#define SUBRECOMPUTE(_id, _sid) case GO_SUBTYPE(_id, _sid): GLUEE(GO_SUBNAME(_id, _sid), _recompute) (spec); break;

void go_recompute(struct go *go)
{
	DBG("GO recompute: %s at %p\n", go_lookup_name(go), go);
	ASSERT(go_can_work(go));
	GO_TYPE_SWITCH(
		RECOMPUTE(GROUP)
		RECOMPUTE(POINT)
		RECOMPUTE(SEGMENT)
		RECOMPUTE(BEZIER)
		GO_SUBTYPE_SWITCH(ELARC,
			SUBRECOMPUTE(ELARC, XYR)
			SUBRECOMPUTE(ELARC, FOCI)
			SUBRECOMPUTE(ELARC, 3SMALL)
			SUBRECOMPUTE(ELARC, 3ECC)
			SUBRECOMPUTE(ELARC, XY1ECC))
		RECOMPUTE(PARPOINT)
		RECOMPUTE(INTERSECTION)
		RECOMPUTE(TEXT)
		RECOMPUTE(TEX)
		RECOMPUTE(DECPOINT)
		RECOMPUTE(DIRECTION)
	)
	go_changed(go);
}

#define TRANSFORM(_id) case GO_TYPE(_id): GLUEE(GO_NAME(_id), _transform) ((GO_STRUCT(_id)*)go, t); break;
#define SUBTRANFORM(_id, _sid) case GO_SUBTYPE(_id, _sid): GLUEE(GO_SUBNAME(_id, _sid), _transform) (spec, t); break;

void go_transform(struct go *go, struct geom_transform2 *t2)
{
	ASSERT(go_can_work(go));
	DBG("GO transform: %s at %p\n", go_lookup_name(go), go);
	struct geom_transform *t = &(t2->primary);
	switch(go->type)
	{
		TRANSFORM(GROUP)
		TRANSFORM(POINT)
		TRANSFORM(SEGMENT)
		TRANSFORM(BEZIER)
		GO_SUBTYPE_SWITCH(ELARC,
			SUBTRANFORM(ELARC, XYR)
			SUBTRANFORM(ELARC, FOCI)
			SUBTRANFORM(ELARC, 3SMALL)
			SUBTRANFORM(ELARC, 3ECC)
			SUBTRANFORM(ELARC, XY1ECC))
		TRANSFORM(PARPOINT)
		TRANSFORM(INTERSECTION)
		TRANSFORM(TEXT)
		TRANSFORM(TEX)
		RECOMPUTE(DECPOINT)
		RECOMPUTE(DIRECTION)

		default:
		  die("Function go_transform hasn't label for the go type %d - %s.\n", go->type, string_go_type[go->type]->text);
	}
	go_transformed(go, t2);
}

void rectangle_normalize(struct geom_rectangle *gr)
{
#define RECTANGLE_NORMALIZE_SWAP(_c) if(gr->ll._c > gr->ur._c) { r = gr->ll._c; gr->ll._c = gr->ur._c; gr->ur._c = r; }
	real r;
	RECTANGLE_NORMALIZE_SWAP(x)
	RECTANGLE_NORMALIZE_SWAP(y)
}

int go_approx(struct geom_approx *approx, struct go *go, uns flags)
{
	const struct geom_path *path;

	if(go_get_info(go)->expand == ET_PATH)
		return (path = go_get_expand_path(go_get_tlo(go), go)) && geom_approx_path(approx, path, flags) == 0 ? 1 : 0;

	if(go->type == GOT_BEZIER)
		return geom_approx_bezier(approx, &((struct go_bezier*)go)->geom, flags) == 0 ? 1 : 0;

	die("Function go_approx hasn't label for the go type %d - %s.\n", go->type, string_go_type[go->type]->text);
}
