/** @file
 *
 *	VRR kernel - derived GO objects.
 *
 *	Expansion, initialization, kernel-level graphical operations
 */

#ifndef _KERNEL_GO_H_
#define _KERNEL_GO_H_


/*
 * Several functions which directly call geomlib.
 * See geomlib/item.h for description.
 */

/*
 * PARAMETRIZATIONS
 */

SCHEME static inline SREAL go_tlength(SO_go go)
	{ return geom_item_tlength(go2geom(go)); }

SCHEME static inline SREAL go_alength(SO_go go)
	{ return geom_item_alength(go2geom(go)); }

SCHEME static inline SREAL go_blength(SO_go go)
	{ return geom_item_blength(go2geom(go)); }

/* TIME <-> ATIME conversions */

SCHEME static inline SREAL go_time_to_atime(SO_go go, SREAL time)
	{ return geom_item_time_to_atime(go2geom(go), time); }

static inline int go_times_to_atimes(struct go *go, uns count, real *times, real *atimes)
	{ return geom_item_times_to_atimes(go2geom(go), count, times, atimes); }

SCHEME static inline SREAL go_atime_to_time(SO_go go, SREAL time)
	{ return geom_item_atime_to_time(go2geom(go), time); }

static inline int go_atimes_to_times(struct go *go, uns count, real *times, real *atimes)
	{ return geom_item_atimes_to_times(go2geom(go), count, times, atimes); }

/* TIME <-> RATIME conversions */

SCHEME static inline SREAL go_time_to_ratime(SO_go go, SREAL time)
	{ return geom_item_time_to_ratime(go2geom(go), time); }

static inline int go_times_to_ratimes(struct go *go, uns count, real *times, real *atimes)
	{ return geom_item_times_to_ratimes(go2geom(go), count, times, atimes); }

SCHEME static inline SREAL go_ratime_to_time(SO_go go, SREAL time)
	{ return geom_item_ratime_to_time(go2geom(go), time); }

static inline int go_ratimes_to_times(struct go *go, uns count, real *times, real *atimes)
	{ return geom_item_ratimes_to_times(go2geom(go), count, times, atimes); }

/* TIME <-> BTIME conversions */
	
SCHEME static inline SINT go_time_dif_btime(SO_go go)
	{ return geom_item_time_dif_btime(go2geom(go)); }

static inline real go_time_to_btime(struct go *go, struct geom_expansion *expansion, real time)
	{ return geom_item_time_to_btime(go2geom(go), expansion, time); }

static inline int go_times_to_btimes(struct go *go, struct geom_expansion *expansion, uns count, real *times, real *btimes)
	{ return geom_item_times_to_btimes(go2geom(go), expansion, count, times, btimes); }
	
static inline real go_btime_to_time(struct go *go, struct geom_expansion *expansion, real btime)
	{ return geom_item_btime_to_time(go2geom(go), expansion, btime); }

static inline int go_btimes_to_times(struct go *go, struct geom_expansion *expansion, uns count, real *btimes, real *times)
	{ return geom_item_btimes_to_times(go2geom(go), expansion, count, btimes, times); }

/*
 * CONVERSION TO OTHER TYPES OF CURVES
 */

static inline int go_expansion_append(struct go *go, struct geom_fpath *expansion)
	{ return geom_item_expansion_append(go2geom(go), expansion); }

static inline struct geom_expansion* go_expand(struct go *go)
	{ return geom_item_expand(go2geom(go)); }

static inline void go_expansion_unref(struct geom_expansion *ref)
	 { geom_expansion_unref(ref); }

static inline int go_to_bezier(struct go *go, struct geom_bezier *bezier)
	{ return geom_item_to_bezier(go2geom(go), bezier); }

/*
 * GEOMETRIC PROPERTIES
 */
	
SCHEME static inline SINT go_is_closed(SO_go go)
	{ return geom_item_is_closed(go2geom(go)); }

static inline int go_point_start(struct go *go, struct geom_point *result)
	{ return geom_item_point_start(go2geom(go), result); }

static inline int go_point_end(struct go *go, struct geom_point *result)
	{ return geom_item_point_end(go2geom(go), result); }

static inline int go_point_at_time(struct go *go, real time, struct geom_point *result)
	{ return geom_item_point_at_time(go2geom(go), time, result); }

static inline int go_derivation_start(struct go *go, struct geom_vector *result)
	{ return geom_item_derivation_start(go2geom(go), result); }

static inline int go_derivation_end(struct go *go, struct geom_vector *result)
	{ return geom_item_derivation_end(go2geom(go), result); }

static inline int go_derivation_at_time(struct go *go, real time, struct geom_vector *result)
	{ return geom_item_derivation_at_time(go2geom(go), time, result); }

static inline int go_bbox(struct go *go, struct geom_rectangle *bbox)
	{ return geom_item_bbox(go2geom(go), bbox); }

static inline int go_nearest_to_point(struct go *go, struct geom_point *point, uns flags, struct geom_nearest *nearest)
	{  return geom_item_nearest_to_point(go2geom(go), point, flags, nearest); }
	
static inline int go_horizontal_times(struct go *go, struct garr *result)
	{ return geom_item_horizontal_times(go2geom(go), result); }

static inline int go_intersections(struct go *go1, struct go *go2, uns flags, struct garr *result)
	{ return geom_items_intersections(go2geom(go1), go2geom(go2), flags, result); }


// TODO:
int go_approx(struct geom_approx *approx, struct go *go, uns flags);

/*
 * GROUP
 */

// Structure representing GO group and group's subtypes are defined in types.h.

SCHEME SO_go_group go_group_new(void) STRANS ;



/*
 * POINT
 */

struct go_point
{
	struct go go;
	struct geom_point_item geom;
	struct anchor a[1];
	struct hanger h[1];
	u8 unit_coordinate_x, unit_coordinate_y;
};

enum go_point_subtype
{
	GOST_POINT,
	GOST_POINT_MAX,
};

/** Create new point given by hanger. */
SCHEME SO_go_point go_point_new(SHANGER h) STRANS ;



/*
 * SEGMENT
 */

struct go_segment
{
	struct go go;
	struct geom_segment geom;
	struct anchor a[2]; // start point, end point
	struct hanger h[3]; // curve, start point, end point
	uns fill_color, stroke_color, stroke_cap;
	real stroke_width;
	u8 unit_stroke_width, unit_start_x, unit_start_y, unit_end_x, unit_end_y;
};

enum go_segment_subtype
{
	GOST_SEGMENT,
	GOST_SEGMENT_MAX,
};

/** Create new segment given by its endpoints. */
SCHEME SO_go_segment go_segment_new(SHANGER starting_point, SHANGER end_point) STRANS ;



/*
 * BEZIER
 */

struct go_bezier
{
	struct go go;
	struct geom_bezier geom;
	struct anchor a[4]; // control points
	struct hanger h[3]; // curve, start point, end point
	uns fill_color, stroke_color, stroke_cap;
	real stroke_width;
	u8 unit_stroke_width,
		unit_control_0_x, unit_control_0_y, unit_control_1_x, unit_control_1_y,
		unit_control_2_x, unit_control_2_y, unit_control_3_x, unit_control_3_y;
};

enum go_bezier_subtype
{
	GOST_BEZIER_QUADRATIC,
	GOST_BEZIER_CUBIC,
	GOST_BEZIER_MAX,
};

/** Create new cubic bezier curve. */
SCHEME SO_go_bezier go_bezier_new_cubic(SHANGER control_1, SREAL weight_1, SHANGER control_2, SREAL weight_2, SHANGER control_3, SREAL weight_3, SHANGER control_4, SREAL weight_4) STRANS ;

/** Create new quadratic bezier curve. */
SCHEME SO_go_bezier go_bezier_new_quadratic(SHANGER control_1, SREAL weight_1, SHANGER control_2, SREAL weight_2, SHANGER control_3, SREAL weight_3) STRANS ;



/*
 * ELLIPTIC ARC
 */

struct go_elarc
{
	struct go go;
	struct geom_elliptic_arc geom;
	struct anchor a[3]; // Hangers depend on subtype.
	struct hanger h[4]; // curve, start point, end point, center
	real ecc;
	uns conic;
	uns fill_color, stroke_color, stroke_cap;
	real stroke_width;
	u8 unit_stroke_width, unit_center_x, unit_center_y, unit_rotation, unit_a_radius, unit_b_radius, unit_start, unit_dif, unit_ecc;
};

enum go_elarc_subtype
{
	GOST_ELARC_XYR,
	GOST_ELARC_FOCI,
	GOST_ELARC_3SMALL,
	GOST_ELARC_3ECC,
	GOST_ELARC_XY1ECC,
	GOST_ELARC_MAX,
};

/** Create new elliptic arc given by centerpoint, two radii and start and difference angle. */
SCHEME SO_go_elarc go_elarc_new_xyr(SHANGER center, SREAL rotation, SREAL horizontal_radius, SREAL vertical_radius, SREAL start, SREAL dif, SUNS conic) STRANS ;

/** Create new ellipse given 2 foci and 1 point on circumference. */
SCHEME SO_go_elarc go_elarc_new_foci(SHANGER focus1, SHANGER focus2, SHANGER point, SUNS conic) STRANS ;

/** Create new ellipse given by 3 points on circumference with minimum area. */
SCHEME SO_go_elarc go_elarc_new_3small(SHANGER point1, SHANGER point2, SHANGER point3, SUNS conic) STRANS ;

/** Crete new ellipse given by 3 points on circumference, rotation and eccentricity. */
SCHEME SO_go_elarc go_elarc_new_3ecc(SHANGER point1, SHANGER point2, SHANGER point3, SREAL rotation, SREAL eccentricity, SUNS conic) STRANS ;

/** Crete new ellipse given by center, point on circumference, rotation and eccentricity. */
SCHEME SO_go_elarc go_elarc_new_xy1ecc(SHANGER center, SHANGER point, SREAL rotation, SREAL eccentricity, SUNS conic) STRANS ;



/*
 * PARAMETRIC POINT
 */

struct go_parametric_point
{
	struct go go;
	struct geom_point_item geom;
	real parameter;
	struct anchor a[1]; // curve
	struct hanger h[1]; // point
	u8 unit_coordinate_x, unit_coordinate_y, unit_parameter;
};

enum go_parametric_point_subtype
{
	GOST_PARAMETRIC_POINT,
	GOST_PARAMETRIC_POINT_MAX,
};

/** Create new parametric point on given GO. */
SCHEME SO_go_parametric_point go_parametric_point_new(SO_go go, SREAL parameter) STRANS ;

SCHEME SO_go_parametric_point go_parametric_point_new_on_hanger(SHANGER h, SREAL parameter, SBOOL position_important, SREAL x, SREAL y); STRANS

/** Check whether the parametric point hangs on curve. */
SCHEME static inline SBOOL go_parametric_point_is_valid(SO_go_parametric_point p)
	{ return hanger_get_go(p->a[0].h)->type != GOT_GROUP; }

/** Check whether the parametric point depents on its position. Necessary for copy and load/save. */
SCHEME static inline SBOOL go_parametric_point_is_position_important(SO_go_parametric_point p)
	{ return p->go.flags & GOF_PARPOINT_INVALID_X; }



/*
 * INTERSECTION POINT
 */

struct go_intersection
{
	struct go go;
	struct geom_point_item geom;
	real parameter;
	struct anchor a[2];
	struct hanger h[1];
	u8 unit_coordinate_x, unit_coordinate_y, unit_parameter;
};

enum go_intersection_subtype
{
	GOST_INTERSECTION,
	GOST_INTERSECTION_MAX,
};

SCHEME SO_go_intersection go_intersection_new(SO_go go1, SO_go go2, SREAL parameter) STRANS ;

SCHEME SO_go_intersection go_intersection_new_on_hanger(SHANGER h1, SHANGER h2, SREAL parameter, SBOOL position_important, SREAL x, SREAL y) STRANS ;

/** Check whether the intersection hangs on two intersected curves. */
SCHEME SBOOL go_intersection_is_valid(SO_go_intersection i);

/** Check whether the intersection depents on its position. Necessary for copy and load/save. */
SCHEME static inline SBOOL go_intersection_is_position_important(SO_go_intersection i)
	{ return i->go.flags & GOF_INTERSECTION_INVALID_X; }

/*
 * TEXT
 */

struct go_text
{
	struct go go;
	struct geom_point_item geom;
	struct anchor a[1];
	struct geom_point apos, rpos;
	struct geom_transform *transformation;
	int font_id;
	real font_size;
	string text;
	struct geom_rectangle string_bbox, // bbox stringu v LFT souradnicich
								 text_bbox; // bbox textu v TLO souradnicich
	u8 unit_rpos_x, unit_rpos_y, unit_apos_x, unit_apos_y, unit_coordinate_x, unit_coordinate_y, unit_size;
	struct hanger h[];
};

enum go_text_subtype
{
	GOST_TEXT,
	GOST_TEXT_MAX,
};

SCHEME SO_go_text go_text_new(SHANGER refpoint, SREAL rpos_x, SREAL rpos_y, SREAL apos_x, SREAL apos_y, SSTRINGS text, SINT fontid, SREAL fntsize) STRANS ;

static inline void go_text_get_string_bbox(struct go_text *t, struct geom_rectangle *bbox)
	{ *bbox = t->string_bbox; }

static inline void go_text_get_text_bbox(struct go_text *t, struct geom_rectangle *bbox)
	{ *bbox = t->text_bbox; }
	
static inline void go_text_get_visualisation_point(struct go_text *t, struct geom_point *gp)
	{ *gp = (struct geom_point) { t->transformation->coef[0][2], t->transformation->coef[1][2] }; }

static inline void go_text_get_transformation(struct go_text *t, struct geom_transform *tr)
	{ *tr = *t->transformation; }


/*
 * TEX
 */

struct tex_process
{
	uns ref_count;
	struct garr *glyphs;
	uns glyphs_cnt;			///< number of glyphs
	struct geom_rectangle bbox;	///< bbox is relative to position
	string text;
};

static inline struct tex_process* tex_process_ref(struct tex_process *tp)
	{ tp->ref_count++; return tp; }

static inline void tex_process_unref(struct tex_process *tp)
	{ tp->ref_count--; if(tp->ref_count) return; garr_destroy(tp->glyphs); string_unref(tp->text); xfree(tp); }

struct go_tex_text
{
	struct go go;
	struct geom_point_item geom;
	struct anchor a[1];
	struct geom_point apos, rpos;
	struct geom_transform *transformation;
	struct tex_process *process;
	struct geom_rectangle text_bbox; // bbox textu v TLO souradnicich
	u8 unit_rpos_x, unit_rpos_y, unit_apos_x, unit_apos_y, unit_coordinate_x, unit_coordinate_y;
	struct hanger h[];
};

enum go_tex_text_subtype
{
	GOST_TEX_TEXT,
	GOST_TEX_TEXT_MAX,
};

SCHEME SO_go_tex_text go_tex_text_new(SHANGER refpoint, SREAL rpos_x, SREAL rpos_y, SREAL apos_x, SREAL apos_y, SSTRINGS text) STRANS ;

struct go_tex_text* go_tex_text_new_process(struct hanger* ll, real refp_x, real refp_y, real marg_x, real marg_y, struct tex_process *process);

static inline void go_tex_text_get_text_bbox(struct go_tex_text *t, struct geom_rectangle *bbox)
	{ *bbox = t->text_bbox; }
	
static inline void go_tex_text_get_string_bbox(struct go_tex_text *t, struct geom_rectangle *bbox)
	{ *bbox = t->process->bbox; }

static inline void go_tex_text_get_visualisation_point(struct go_tex_text *t, struct geom_point *gp)
	{ *gp = (struct geom_point) { t->transformation->coef[0][2], t->transformation->coef[1][2] }; }

static inline void go_tex_text_get_transformation(struct go_tex_text *t, struct geom_transform *tr)
	{ *tr = *t->transformation; }


/*
 * DECORATOR_POINT
 */

#define GO_DECORATOR_POINT_MAX_VERTICES 100
	
struct go_decorator_point
{
	struct go go;
	struct geom_callback_item geom;
	struct anchor a[1];
	struct geom_point x;
	uns n;
	real radius, rotation;
	uns fill_color, stroke_color, stroke_cap;
	real stroke_width;
	u8 unit_stroke_width, unit_radius, unit_rotation, unit_coordinate_x, unit_coordinate_y;
	struct hanger h[];
};

enum go_decorator_point_subtype
{
	GOST_DECORATOR_POINT,
	GOST_DECORATOR_POINT_MAX,
};

SCHEME SO_go_decorator_point go_decorator_point_new(SHANGER point, SUNS n, SREAL radius, SREAL rotation) STRANS ;


/*
 * DIRECTION
 */

struct go_direction
{
	struct go go;
	struct geom_callback_item geom;
	struct anchor a[1];
	struct geom_point x;
	uns n1, n2;
	real radius, rotation, parameter;
	uns fill_color, stroke_color, stroke_cap;
	real stroke_width;
	u8 unit_stroke_width, unit_radius, unit_rotation, unit_coordinate_x, unit_coordinate_y;
	struct hanger h[];
};

enum go_direction_subtype
{
	GOST_DIRECTION,
	GOST_DIRECTION_MAX,
};

SCHEME SO_go_direction go_direction_new(SHANGER point, SREAL parameter, SUNS n1, SUNS n2, SREAL radius, SREAL rotation) STRANS ;

#endif
