#include "kernel/internal.h"

#define GO_ID BEZIER

#define GO_SUBID CUBIC



PROP_STYLES
PROP_ALL_WITHOUT_UNIT(weight_3, PSTRUCT(fill_color), PT_REAL, PTR_UNSPECIFIED, geom.pt[3].w, r,
		value.r > 0, (("Weight must be positive number."), ("Weight of control point 3 cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_REHANG(control_3, PSTRUCT(weight_3), geom.pt[3], 3, "Hanger of the fourth point must be a mouse click to change position.", 0)
PROP_ALL_WITHOUT_UNIT(weight_2, PSTRUCT(control_3_x), PT_REAL, PTR_UNSPECIFIED, geom.pt[2].w, r,
		value.r > 0, (("Weight must be positive number."), ("Weight of control point 2 cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_REHANG(control_2, PSTRUCT(weight_2), geom.pt[2], 2, "Hanger of the third control point must be a mouse click to change position.", 0)
PROP_ALL_WITHOUT_UNIT(weight_1, PSTRUCT(control_2_x), PT_REAL, PTR_UNSPECIFIED, geom.pt[1].w, r,
		value.r > 0, (("Weight must be positive number."), ("Weight of control point 1 cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_REHANG(control_1, PSTRUCT(weight_1), geom.pt[1], 1, "Hanger of the second control point must be a mouse click to change position.", 0)
PROP_ALL_WITHOUT_UNIT(weight_0, PSTRUCT(control_1_x), PT_REAL, PTR_UNSPECIFIED, geom.pt[0].w, r,
		value.r > 0, (("Weight must be positive number."), ("Weight of control point 0 cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_REHANG(control_0, PSTRUCT(weight_0), geom.pt[0], 0, "Hanger of the first control point must be a mouse click to change position.", 0)

PROP_VIRTUAL(PROP_STYLES_INIT,
		&PSTRUCT(control_0_x), "control-0-x", &PSTRUCT(control_0_y), "control-0-y", &PSTRUCT(weight_0), "weight-0",
		&PSTRUCT(control_1_x), "control-1-x", &PSTRUCT(control_1_y), "control-1-y", &PSTRUCT(weight_1), "weight-1",
		&PSTRUCT(control_2_x), "control-2-x", &PSTRUCT(control_2_y), "control-2-y", &PSTRUCT(weight_2), "weight-2",
		&PSTRUCT(control_3_x), "control-3-x", &PSTRUCT(control_3_y), "control-3-y", &PSTRUCT(weight_3), "weight-3")

INFO_ANCHOR_4(
		"controlpoint-1", QT_POINT, LT_AFFINITY,
		"controlpoint-2", QT_POINT, LT_AFFINITY,
		"controlpoint-3", QT_POINT, LT_AFFINITY,
		"controlpoint-4", QT_POINT, LT_AFFINITY)

INFO_HANGER_3(
		"curve", QT_CURVE, h[0], NULL, LT_AFFINITY,
		"controlpoint-1", QT_POINT, geom.pt[0], NULL, LT_AFFINITY,
		"controlpoint-4", QT_POINT, geom.pt[3], NULL, LT_AFFINITY)

struct go_bezier *go_bezier_new_cubic(
		struct hanger *h1, real w1, struct hanger *h2, real w2,
		struct hanger *h3, real w3, struct hanger *h4, real w4)
{
	struct go_bezier *b = xmalloc_zero(sizeof(struct go_bezier));
	go_init(&b->go, GOT_BEZIER, GOST_BEZIER_CUBIC, h1, h2, h3, h4);

	GEOM_BEZIER_DEGREE(&b->geom) = 3;
	b->geom.pt[0].w = w1;
	b->geom.pt[1].w = w2;
	b->geom.pt[2].w = w3;
	b->geom.pt[3].w = w4;

	b->fill_color = PROP_INIT_FILL_COLOR;
	b->stroke_color = PROP_INIT_STROKE_COLOR;
	b->stroke_width = PROP_INIT_STROKE_WIDTH;
	b->stroke_cap = PROP_INIT_STROKE_CAP;
	
	b->unit_stroke_width = b->unit_control_0_x = b->unit_control_0_y = b->unit_control_1_x = b->unit_control_1_y = b->unit_control_2_x = b->unit_control_2_y = b->unit_control_3_x = b->unit_control_3_y = PROP_UNIT_MAX;
	return b;
}

#undef GO_SUBID



#define GO_SUBID QUADRATIC

PROP_STYLES
PROP_ALL_WITHOUT_UNIT(weight_2, PSTRUCT(fill_color), PT_REAL, PTR_UNSPECIFIED, geom.pt[2].w, r,
		value.r > 0, (("Weight must be positive number."), ("Weight of control point 2 cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_REHANG(control_2, PSTRUCT(weight_2), geom.pt[2], 2, "Hanger of the third control point must be a mouse click to change position.", 0)
PROP_ALL_WITHOUT_UNIT(weight_1, PSTRUCT(control_2_x), PT_REAL, PTR_UNSPECIFIED, geom.pt[1].w, r,
		value.r > 0, (("Weight must be positive number."), ("Weight of control point 1 cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_REHANG(control_1, PSTRUCT(weight_1), geom.pt[1], 1, "Hanger of the second control point must be a mouse click to change position.", 0)
PROP_ALL_WITHOUT_UNIT(weight_0, PSTRUCT(control_1_x), PT_REAL, PTR_UNSPECIFIED, geom.pt[0].w, r,
		value.r > 0, (("Weight must be positive number."), ("Weight of control point 0 cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_REHANG(control_0, PSTRUCT(weight_0), geom.pt[0], 0, "Hanger of the first control point must be a mouse click to change position.", 0)

PROP_VIRTUAL(PROP_STYLES_INIT,
		&PSTRUCT(control_0_x), "control-0-x", &PSTRUCT(control_0_y), "control-0-y", &PSTRUCT(weight_0), "weight-0",
		&PSTRUCT(control_1_x), "control-1-x", &PSTRUCT(control_1_y), "control-1-y", &PSTRUCT(weight_1), "weight-1",
		&PSTRUCT(control_2_x), "control-2-x", &PSTRUCT(control_2_y), "control-2-y", &PSTRUCT(weight_2), "weight-2")

INFO_ANCHOR_3(
		"controlpoint-1", QT_POINT, LT_AFFINITY,
		"controlpoint-2", QT_POINT, LT_AFFINITY,
		"controlpoint-3", QT_POINT, LT_AFFINITY)

INFO_HANGER_3(
		"curve", QT_CURVE, h[0], NULL, LT_AFFINITY,
		"controlpoint-1", QT_POINT, geom.pt[0], NULL, LT_AFFINITY,
		"controlpoint-3", QT_POINT, geom.pt[2], NULL, LT_AFFINITY)

struct go_bezier *go_bezier_new_quadratic(struct hanger *h1, real w1, struct hanger *h2, real w2, struct hanger *h3, real w3)
{
	struct go_bezier *b = xmalloc_zero(sizeof(struct go_bezier));
	go_init(&b->go, GOT_BEZIER, GOST_BEZIER_QUADRATIC, h1, h2, h3);

	GEOM_BEZIER_DEGREE(&b->geom) = 2;
	b->geom.pt[0].w = w1;
	b->geom.pt[1].w = w2;
	b->geom.pt[2].w = w3;

	b->fill_color = PROP_INIT_FILL_COLOR;
	b->stroke_color = PROP_INIT_STROKE_COLOR;
	b->stroke_width = PROP_INIT_STROKE_WIDTH;
	b->stroke_cap = PROP_INIT_STROKE_CAP;
	
	b->unit_stroke_width = b->unit_control_0_x = b->unit_control_0_y = b->unit_control_1_x = b->unit_control_1_y = b->unit_control_2_x = b->unit_control_2_y = PROP_UNIT_MAX;
	return b;
}

#undef GO_SUBID

INFO_GO_2("bezier",
		CUBIC, "cubic-bezier", 3, 0, 1, 2, -1, 4, ET_NONE, control_0_x,
		QUADRATIC, "quadratic-bezier", 3, 0, 1, 2, -1, 3, ET_NONE, control_0_x)

void go_bezier_transform(struct go_bezier *b, struct geom_transform *t UNUSED)
{
	go_bezier_recompute(b);
}

void go_bezier_recompute(struct go_bezier *b)
{
	for(int i = 0; i <= GEOM_BEZIER_DEGREE(&b->geom); i++)
	{
		struct geom_point gp = hanger_get_position(b->a[i].h);
		b->geom.pt[i].x = gp.x;
		b->geom.pt[i].y = gp.y;
	}
	geom_bezier_after_change(&b->geom);
}

int go_bezier_bbox(struct go_bezier *b, struct geom_rectangle *bbox)
{
	geom_bezier_control_box(&b->geom, bbox);
	return 1;
}

