#include "kernel/internal.h"

#define GO_ID DECPOINT
#define GO_SUBID DECPOINT

PROP_STYLES
PROP_ALL_REHANG(coordinate, PSTRUCT(fill_color), x, 0, "Hanger of point must be a mouse click to change position.", 0)
PROP_ALL(radius, PSTRUCT(coordinate_x), PT_REAL, PTR_NON_NEGATIVE, radius, r,
		value.r >= 0, (("Invalid value of radius."), ("Radius cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL(rotation, PSTRUCT(radius), PT_REAL, PTR_ANGLE_2PI, rotation, r,
		1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(n, PSTRUCT(rotation), PT_UNS, PTU_UNSPECIFIED, n, u, 1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, 0, UINT_MAX)

PROP_VIRTUAL(PROP_STYLES_INIT,
		&PSTRUCT(coordinate_x), string_coordinate_x->text, &PSTRUCT(coordinate_y), string_coordinate_y->text,
		&PSTRUCT(radius), "radius", &PSTRUCT(rotation), "rotation", &PSTRUCT(n), "vertices")

INFO_ANCHOR_1("decorator-point", QT_POINT, LT_AFFINITY)
INFO_HANGER_0
INFO_GO_1("decorator-point",
		DECPOINT, "decorator-point", 0, -1, -1, -1, -1, 1, ET_PATH, n)

void go_decorator_point_recompute(struct go_decorator_point *d)
{
	d->x = hanger_get_position(d->a->h);
}

void go_decorator_point_transform(struct go_decorator_point *d, struct geom_transform *t UNUSED)
{
	go_decorator_point_recompute(d);
}

int go_decorator_point_path_and_bbox(struct go_decorator_point *d, struct geom_path *p, struct geom_rectangle *r)
{
/*
 * Pavle, dopis sem, prosim, expanzi na cestu.
 * d->x   stred
 * d->radius   vzdalenost stred - vrchol/obvod
 * d->rotation   nejaka rotace
 * d->n   pro n >= 3 ... pravidelny n-uhelnik
 *        pro n = 0 ...  kruznice
 *        pro n = 2 ... treba usecka
 *        pro n = 1 ... jestli Te napada neco pekneho ...
 */

 	/* FIXME:
	   takovy provizorni kod, jak prekopirovat expanzi do geom_path... tohle nebude potreba,
	   predpokladam, ze cesta p je inicializovana a prazdana a ze ji nekdo uvolni

	   pro d->geom se daji primo volat fce geom_item_cokoli

	   jeste bych mel udelat interface, jak zneplatnit geomlibovskou expanzi...
	   ted by to melo fungovat i bez toho - expanze se necachuji
	 */
 	struct geom_expansion *e = geom_item_expand(&d->geom);
	if (!e) trans_fail("Geomlib failed with message `%s' while expanding go-decorator-point.", geom_err_str(geom_errno));
	geom_path_copy(p, &e->path.path);
	geom_expansion_unref(e);

	geom_path_bbox(p, r);
	return 1;
}

static int go_dec_pt(struct go_decorator_point *go, struct geom_fpath *path, real d1, real a1, real d2, real a2)
{
	struct geom_bezier *ref;
	a1 = a1 * M_PI / 180;
	a2 = a2 * M_PI / 180;
	struct geom_point a = {go->x.x + d1 * go->radius * cos(a1 + go->rotation), go->x.y + d1 * go->radius * sin(a1 + go->rotation)};
	struct geom_point b = {go->x.x + d2 * go->radius * cos(a2 + go->rotation), go->x.y + d2 * go->radius * sin(a2 + go->rotation)};
	if (!geom_point_valid(&a) || !geom_point_valid(&b)) return 0;
	GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &a, &b);
	return 1;
}

static int go_decorator_point_callback(struct geom_callback_item *geom, struct geom_fpath *path)
{
	struct go_decorator_point *go = SKIP_BACK(struct go_decorator_point, geom, geom);
	struct geom_bezier *ref;

	/* Hidden feature :-) */
	if (go->n == 12345) {
		if (go_dec_pt(go, path, 1, -135, 1, 135) &&
			go_dec_pt(go, path, 1, 135, 1, -45) &&
			go_dec_pt(go, path, 1, -45, 1, 45) &&
			go_dec_pt(go, path, 1, 45, 1, 135) &&
			go_dec_pt(go, path, 1, 135, 1.5, 90) &&
			go_dec_pt(go, path, 1.5, 90, 1, 45) &&
			go_dec_pt(go, path, 1, 45, 1, -135) &&
			go_dec_pt(go, path, 1, -135, 1, -45))
		return 0;
		else return GEOM_ERR_NUMERIC;
	}

	/* FIXME: Jirko: spis omezit v prop_set a trans-failovat...*/
	if (go->n <= GO_DECORATOR_POINT_MAX_VERTICES)
		switch (go->n) {
		case 0:
			break;
		case 1:
			/* Single point */
			GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &go->x, &go->x);
			return 0;
		case 2: {
			/* Segment */
			struct geom_point p, q;
			p.x = go->radius * cos(go->rotation);
			p.y = go->radius * sin(go->rotation);
			q.x = go->x.x - p.x;
			q.y = go->x.y - p.y;
			p.x += go->x.x;
			p.y += go->x.y;
			if (!geom_point_valid(&p) || !geom_point_valid(&q)) return GEOM_ERR_NUMERIC;
			GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &p, &q);
			return 0; }
		default: {
			/* Polygon */
			uns i;
			struct geom_point p, q, s;
			real a = go->rotation;
			real a_inc = 2 * M_PI / go->n;
			s.x = go->radius * cos(a) + go->x.x;
			s.y = go->radius * sin(a) + go->x.y;
			if (!geom_point_valid(&s)) return GEOM_ERR_NUMERIC;
			for (i = 0, p = s; i < go->n; i++, p = q) {
				a += a_inc;
				q.x = go->radius * cos(a) + go->x.x;
				q.y = go->radius * sin(a) + go->x.y;
				if (!geom_point_valid(&q)) return GEOM_ERR_NUMERIC;
				GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &p, &q);
			}
			GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &p, &s);
			return 0; }
		}
	/* Circle */
	struct geom_elliptic_arc arc;
	int err = 0;
	GEOM_ITEM_INIT(&arc, elliptic_arc);
	geom_elliptic_arc_set_closed(&arc, &go->x, 0, go->radius, go->radius);
	err = geom_elliptic_arc_expansion_append(&arc, path);
	geom_elliptic_arc_destroy(&arc);
	return err;
}

struct go_decorator_point *go_decorator_point_new(struct hanger *point, uns n, real radius, real rotation)
{
	if(radius < 0)
		trans_fail("Invalid radius.");
	struct go_decorator_point *d = xmalloc_zero(sizeof(struct go_decorator_point));
	go_init(&d->go, GOT_DECORATOR_POINT, GOST_DECORATOR_POINT, point);
	geom_callback_item_create_set(&d->geom, &go_decorator_point_callback);
	d->radius = radius;
	d->rotation = rotation;
	d->n = n;
	d->fill_color = PROP_INIT_FILL_COLOR;
	d->stroke_color = PROP_INIT_STROKE_COLOR;
	d->stroke_width = PROP_INIT_STROKE_WIDTH;
	d->stroke_cap = PROP_INIT_STROKE_CAP;
	d->unit_stroke_width = d->unit_coordinate_x = d->unit_coordinate_y = d->unit_radius = d->unit_rotation = PROP_UNIT_MAX;
	return d;
}

#undef GO_ID
#undef GO_SUBID



#define GO_ID DIRECTION
#define GO_SUBID DIRECTION

PROP_STYLES
PROP_ALL_REHANG(coordinate, PSTRUCT(fill_color), x, 0, "Hanger of point must be a mouse click to change position.", 0)
PROP_ALL(radius, PSTRUCT(coordinate_x), PT_REAL, PTR_NON_NEGATIVE, radius, r,
		value.r >= 0, (("Invalid value of radius."), ("Radius cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL(rotation, PSTRUCT(radius), PT_REAL, PTR_ANGLE_2PI, rotation, r,
		1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(n2, PSTRUCT(rotation), PT_UNS, PTU_UNSPECIFIED, n2, u, 1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(n1, PSTRUCT(n2), PT_UNS, PTU_UNSPECIFIED, n1, u, 1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, 0, UINT_MAX)

PROP_VIRTUAL(PROP_STYLES_INIT,
		&PSTRUCT(coordinate_x), string_coordinate_x->text, &PSTRUCT(coordinate_y), string_coordinate_y->text,
		&PSTRUCT(radius), "radius", &PSTRUCT(rotation), "rotation", &PSTRUCT(n2), "vertices-2", &PSTRUCT(n1), "vertices-1")

INFO_ANCHOR_1("decorator-direction", QT_CURVE, LT_AFFINITY)
INFO_HANGER_0
INFO_GO_1("decorator-direction",
		DIRECTION, "decorator-direction", 0, -1, -1, -1, -1, 1, ET_PATH, n1)

void go_direction_recompute(struct go_direction *d)
{
	// FIXME: tady musi byt makro COMPUTE_POINT z go_fork.c
	if(!GEOM_ETEST(go_point_at_time(hanger_get_go(d->a->h), d->parameter, &d->x))) \
		TRANS_FAIL(("Numeric error."), ("go_point_at_time fail.\n")); \
}

void go_direction_transform(struct go_direction *d, struct geom_transform *t UNUSED)
{
	go_direction_recompute(d);
}

int go_direction_path_and_bbox(struct go_direction *d, struct geom_path *p, struct geom_rectangle *r)
{
	struct go *go = hanger_get_go(d->a->h);
	real direction;
	struct geom_vector vector;

	if(go->type == GOT_GROUP)
		direction = 0;
	else
	{
		go_derivation_at_time(go, d->parameter, &vector);
		direction = atan2(vector.dy, vector.dx);
	}
	direction += d->rotation;
	
/*
 * Pavle, dopis sem, prosim, expanzi na cestu. Mohlo by to byt zhruba takto:
 * start_point = d->x + d->size * (direction + d->angle)
 * end_point = d->x + d->size * (direction d->angle)
 * first_point = d->x + d->dist1 * direction
 * second_point = d->x + d->dist2 * direction
 * circle1 = circle_on_tree_points(start_point, first_point, end_point)
 * circle2 = circle_on_tree_points(end_point, second_point, start_point)
 * if(d->n1 == 0)
 * 	vloz do path expanzi circle1
 * else
 * 	predstav si (d->n1 + 1) bodu rovnomerne po circle1 tak, ze prvni, resp. posledni, bod je prave start_point, resp. end_point
 * 	vloz do path usecky spojujici sousedni body
 * to same pro druhou cast
 *
 * Asi vymyslis nejakou lepsi parametrizaci, tak ji klidne pouzij. Do struktury go_direction vloz promenne, ktere potrebujes.
 */

	geom_path_bbox(p, r);
	return 1;
}

static int go_direction_callback(struct geom_callback_item *geom, struct geom_fpath *path)
{
	struct go_direction *go = SKIP_BACK(struct go_direction, geom, geom);

	int err = 0;
	return err;
}

struct go_direction *go_direction_new(struct hanger *point, real parameter, uns n1, uns n2, real radius, real rotation)
{
	struct go_direction *d = xmalloc_zero(sizeof(struct go_direction));
	go_init(&d->go, GOT_DIRECTION, GOST_DIRECTION, point);
	geom_callback_item_create_set(&d->geom, &go_direction_callback);
	d->parameter = parameter;
	d->n1 = n1;
	d->n2 = n2;
	d->radius = radius;
	d->rotation = rotation;
	d->fill_color = PROP_INIT_FILL_COLOR;
	d->stroke_color = PROP_INIT_STROKE_COLOR;
	d->stroke_width = PROP_INIT_STROKE_WIDTH;
	d->stroke_cap = PROP_INIT_STROKE_CAP;
	d->unit_stroke_width = d->unit_coordinate_x = d->unit_coordinate_y = d->unit_radius = d->unit_rotation = PROP_UNIT_MAX;
	return d;
}

#undef GO_ID
#undef GO_SUBID

