//#define LOCAL_DEBUG
#include "kernel/internal.h"

#define GO_ID ELARC

const uns conic_max[GOST_ELARC_MAX] = {
	[GOST_ELARC_XYR] = GEOM_CONIC_ARC_0P,
	[GOST_ELARC_FOCI] = GEOM_CONIC_ARC_1P,
	[GOST_ELARC_3SMALL] = GEOM_CONIC_ARC_3P,
	[GOST_ELARC_3ECC] = GEOM_CONIC_ARC_3P,
	[GOST_ELARC_XY1ECC] = GEOM_CONIC_ARC_1P,
};

const uns start_undo_property_tab[GEOM_CONIC_ARC_MAX] = {
	[GEOM_CONIC_ARC_START_ENTIRE] = 1,
	[GEOM_CONIC_ARC_START_DIF] = 1,
};

const uns dif_undo_property_tab[GEOM_CONIC_ARC_MAX] = {
	[GEOM_CONIC_ARC_POINT_DIF] = 1,
	[GEOM_CONIC_ARC_START_DIF] = 1,
};

static inline int is_start_undo_property(uns conic) { return start_undo_property_tab[conic]; }
static inline int is_dif_undo_property(uns conic) { return dif_undo_property_tab[conic]; }
static inline int is_rotation_valid(real rot) { return rot >= 0 && rot <= R_PI; }
static inline int is_ecc_valid(real ecc) { return ecc >= 0 && ecc <= 1; }
static inline int is_start_valid(real start) { return start >= 0 && start <= R_2PI; }
static inline int is_dif_valid(real dif) { return dif >= -R_2PI && dif <= R_2PI; }
static inline int is_radii_valid(real a, real b) { return a >= b && b >= 0; }

static struct go_elarc *go_elarc_alloc(uns conic)
{
	struct go_elarc *a = xmalloc_zero(sizeof(struct go_elarc));
	a->fill_color = PROP_INIT_FILL_COLOR;
	a->stroke_color = PROP_INIT_STROKE_COLOR;
	a->stroke_width = PROP_INIT_STROKE_WIDTH;
	a->stroke_cap = PROP_INIT_STROKE_CAP;
	a->unit_center_x = a->unit_center_y =
		a->unit_rotation = a->unit_a_radius = a->unit_b_radius =
		a->unit_start = a->unit_dif = a->unit_ecc =
		a->unit_stroke_width = PROP_UNIT_MAX;
	a->conic = conic;
	if(is_start_undo_property(conic))
		a->go.flags |= GOF_ELARC_START;
	if(is_dif_undo_property(conic))
		a->go.flags |= GOF_ELARC_DIF;
	return a;
}

static quantity hanger_go_elarc_get_start(struct hanger *h)
{
	struct go_elarc *a = (struct go_elarc*)hanger_get_go(h);
	struct geom_point gp;
	ASSERT_GO_TYPE_ID(a);
	if(!GEOM_ETEST(geom_elliptic_arc_point_start(&a->geom, &gp)))
	{
		msg("Computing start point of %s fail. Returning position of anchor %s instead.\n", go_get_info(&a->go)->name->text, a->a->info->name->text);
		gp = hanger_get_position(a->a->h);
	}
	return (quantity){ .p = gp };
}

static quantity hanger_go_elarc_get_end(struct hanger *h)
{
	struct go_elarc *a = (struct go_elarc*)hanger_get_go(h);
	struct geom_point gp;
	ASSERT_GO_TYPE_ID(a);
	if(!GEOM_ETEST(geom_elliptic_arc_point_end(&a->geom, &gp)))
	{
		msg("Computing end point of %s fail. Returning position of anchor %s instead.\n", go_get_info(&a->go)->name->text, a->a->info->name->text);
		gp = hanger_get_position(a->a->h);
	}
	return (quantity){ .p = gp };
}

#define SET_CONIC \
static void PSET(conic)(struct o *o, prop_value value) \
{ \
	ASSERT_GO_SUBTYPE_ID(o); \
	struct go_elarc *e = (struct go_elarc*)o; \
	uns conic = value.u; \
 \
	if(conic > conic_max[GO_SUBTYPE_ID]) \
		trans_fail("Invalid arc type."); \
 \
	if(is_start_undo_property(conic) && !is_start_undo_property(e->conic)) \
		trans_new_undo_start(go2o(&e->go), &PSTRUCT(start), (prop_value) { .r = e->geom.start }); \
	else if(!is_start_undo_property(conic) && is_start_undo_property(e->conic)) \
		trans_new_undo_end(go2o(&e->go), &PSTRUCT(start), (prop_value) { .r = e->geom.start }); \
	if(is_dif_undo_property(conic) && !is_dif_undo_property(e->conic)) \
		trans_new_undo_start(go2o(&e->go), &PSTRUCT(dif), (prop_value) { .r = e->geom.dif }); \
	else if(!is_dif_undo_property(conic) && is_dif_undo_property(e->conic)) \
		trans_new_undo_end(go2o(&e->go), &PSTRUCT(dif), (prop_value) { .r = e->geom.dif }); \
 \
	trans_new_undo_prop_virtual(o, &PSTRUCT(conic), (prop_value){.u = e->conic}); \
	e->conic = conic; \
	tsort_dirty_recompute(&e->go); \
}

static inline real angle_transform(real a, struct geom_transform *t)
{
	real c = cosr(a), s = sinr(a),
		  x = t->coef[0][0] * c + t->coef[0][1] * s,
		  y = t->coef[1][0] * c + t->coef[1][1] * s;
	if(!GEOM_RTEST(x) || !GEOM_RTEST(y))
		TRANS_FAIL(("Numeric error."), ("Transformation angle fail."));
	return atan2r(y, x);
}

static inline real transform_determinant(struct geom_transform *t)
{
	real r = t->coef[0][0] * t->coef[1][1] - t->coef[1][0] * t->coef[0][1];
	if(!GEOM_RTEST(r))
		TRANS_FAIL(("Numeric error."), ("Invalid determinant of transformation."));
	return r;
}

static inline real sgn(real r)
{
	return r > 0 ? 1 : (r == 0 ? 0 : -1);
}

static real angle_dif_transform(real dif, struct geom_transform *t)
{
	return dif * sgn(transform_determinant(t));
}

#define DESTROY_TRANS_FAIL(_elarc, ...) do { geom_elliptic_arc_destroy(&result); TRANS_FAIL(__VA_ARGS__); } while(0)

#define CHANGE_UNDO(_var) if(a->geom._var != result._var) \
	{ trans_new_undo_prop_virtual(void2o(a), &PSTRUCT(_var), (prop_value){.r = a->geom._var}); a->geom._var = result._var; };

#define SET_A_RADIUS	\
static void PSET(a_radius) (struct o *o, prop_value value)	\
{	\
	ASSERT_GO_SUBTYPE_ID(o);	\
	ASSERT(isfinite(value.r));	\
	struct go_elarc *g = (struct go_elarc*)o;	\
	if(value.r < 0)	\
		trans_fail("Radius must be positive number.");	\
	trans_new_undo_prop_virtual(o, &PSTRUCT(a_radius), (prop_value){.r = g->geom.a_radius}); \
	g->geom.a_radius = value.r; \
	if(value.r < g->geom.b_radius) \
	{ \
		trans_new_undo_prop_virtual(o, &PSTRUCT(b_radius), (prop_value){.r = g->geom.b_radius}); \
		g->geom.b_radius = value.r; \
	} \
	RECOMPUTE_CHANGED;	\
}

#define SET_B_RADIUS	\
static void PSET(b_radius) (struct o *o, prop_value value)	\
{	\
	ASSERT_GO_SUBTYPE_ID(o);	\
	ASSERT(isfinite(value.r));	\
	struct go_elarc *g = (struct go_elarc*)o;	\
	if(value.r < 0)	\
		trans_fail("Radius must be positive number.");	\
	trans_new_undo_prop_virtual(o, &PSTRUCT(b_radius), (prop_value){.r = g->geom.b_radius}); \
	g->geom.b_radius = value.r; \
	if(value.r > g->geom.a_radius) \
	{ \
		trans_new_undo_prop_virtual(o, &PSTRUCT(a_radius), (prop_value){.r = g->geom.a_radius}); \
		g->geom.a_radius = value.r; \
	} \
	RECOMPUTE_CHANGED;	\
}

/*---------------------------------------------------------------*
 *                                                               *
 *  XYR: center, radii, rotation, start angle, difference angle  *
 *                                                               *
 *---------------------------------------------------------------*/

#define GO_SUBID XYR

PROP_STYLES
PROP_ALL_READ(ecc, PSTRUCT(fill_color), PT_REAL, PTR_REFERENCE, ecc, r, 0, 0)
PROP_ALL(dif, PSTRUCT(ecc), PT_REAL, PTR_ANGLE_4PI, geom.dif, r,
		is_dif_valid(value.r) && is_dif_undo_property(g->conic), (("Invalid value of difference angle."), ("Dif cannot be %f.", value.r)), CHANGED, LOW, 0, GOF_ELARC_DIF)
PROP_ALL(start, PSTRUCT(dif), PT_REAL, PTR_ANGLE_2PI, geom.start, r,
		is_start_valid(value.r) && is_start_undo_property(g->conic), (("Invalid value of start angle."), ("Start cannot be %f.", value.r)), CHANGED, LOW, 0, GOF_ELARC_START)
PROP_ALL_WITHOUT_SET(b_radius, PSTRUCT(start), PT_REAL, PTR_NON_NEGATIVE, geom.b_radius, r, LOW, 0, UINT_MAX)
PROP_ALL_WITHOUT_SET(a_radius, PSTRUCT(b_radius), PT_REAL, PTR_NON_NEGATIVE, geom.a_radius, r, LOW, 0, UINT_MAX)
PROP_ALL(rotation, PSTRUCT(a_radius), PT_REAL, PTR_ANGLE_PI, geom.rotation, r,
		is_rotation_valid(value.r), (("Invalid value of rotation."), ("Rotation cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_REHANG(center, PSTRUCT(rotation), geom.center, 0,
		"Hanger of the center of the elliptic arc must be a mouse click to change position.", 0)
PROP_ALL_WITHOUT_UNIT_AND_SET(conic, PSTRUCT(center_x), PT_UNS, PTU_CONIC_TYPE_0P, conic, u, LOW, 0, UINT_MAX)
SET_CONIC
SET_A_RADIUS
SET_B_RADIUS

PROP_VIRTUAL(PROP_STYLES_INIT,
		&PSTRUCT(conic), "conic", &PSTRUCT(center_x), "center-x", &PSTRUCT(center_y), "center-y", &PSTRUCT(rotation), "rotation",
		&PSTRUCT(a_radius), "a-radius", &PSTRUCT(b_radius), "b-radius", &PSTRUCT(start), "start", &PSTRUCT(dif), "dif", &PSTRUCT(ecc), "ecc")

INFO_ANCHOR_1("center", QT_POINT, LT_AFFINITY)

INFO_HANGER_4(
		"curve", QT_CURVE, h[0], NULL, LT_AFFINITY,
		"start", QT_POINT, h[1], hanger_go_elarc_get_start, LT_AFFINITY,
		"end", QT_POINT, h[2], hanger_go_elarc_get_end, LT_AFFINITY,
		"center", QT_POINT, geom.center, NULL, LT_AFFINITY)

void go_elarc_xyr_recompute(struct go_elarc *a)
{
	ASSERT(is_radii_valid(a->geom.a_radius, a->geom.b_radius) && is_start_valid(a->geom.start)
			&& is_dif_valid(a->geom.dif) && is_rotation_valid(a->geom.rotation));

	a->geom.center = hanger_get_position(a->a[0].h);
	geom_elliptic_arc_after_change(&a->geom);
	if(!GEOM_ETEST(geom_elliptic_arc_set_endpoints(&a->geom, NULL, NULL, NULL, a->geom.start, a->geom.dif, a->conic)))
		TRANS_FAIL(("Numeric error."), ("geom_elliptic_arc_set_endpoints fail."));
	if(!GEOM_RTEST(a->ecc = geom_elliptic_arc_eccentricity(&a->geom)))
		TRANS_FAIL(("Numeric error."), ("geom_elliptic_arc_eccentricity fail."));

/*	TODO: Polohy hangeru se pocitaji, kdyz jsou zapotrebi. Maji se ukladat ci kesovat?
	if(!GEOM_ETEST(geom_elliptic_arc_points(&a->geom, &a->h[1].position, &a->h[2].position)))
		trans_fail("Numeric error on computing terminal points."); */
}

void go_elarc_xyr_transform(struct go_elarc *a, struct geom_transform *t)
{
	struct geom_elliptic_arc result;
	GEOM_ITEM_INIT(&result, elliptic_arc);

	if(!GEOM_ETEST(geom_elliptic_arc_transform(&a->geom, t, &result)))
		DESTROY_TRANS_FAIL(result, ("Numeric error."), ("geom_elliptic_arc_transform fail."));

	CHANGE_UNDO(rotation);
	CHANGE_UNDO(a_radius);
	CHANGE_UNDO(b_radius);
	CHANGE_UNDO(start);
	CHANGE_UNDO(dif);

	geom_elliptic_arc_destroy(&result);
	go_elarc_xyr_recompute(a);
}

struct go_elarc *go_elarc_new_xyr(struct hanger *center, real rotation,
		real a_radius, real b_radius, real start, real dif, uns conic)
{
	if(conic > conic_max[GOST_ELARC_XYR])
		trans_fail("Invalid conic type.");
/*	if(!is_rotation_valid(rotation))
		trans_fail("Invalid value of rotation.");
	if(!is_radii_valid(a_radius, b_radius))
		trans_fail("Invalid values of radii.");
	if(!is_start_valid(start))
		trans_fail("Invalid value of starting angle.");
	if(!is_dif_valid(dif))
		trans_fail("Invalid value of angles difference."); */

	struct go_elarc *a = go_elarc_alloc(conic);
	go_init(&a->go, GOT_ELARC, GOST_ELARC_XYR, center);

	a->geom.rotation = fminmaxr(rotation, 0, R_PI);
	a->geom.b_radius = fmaxr(b_radius, 0);
	a->geom.a_radius = fmaxr(a_radius, a->geom.b_radius);
	a->geom.start = fminmaxr(start, 0, R_2PI);
	a->geom.dif = fminmaxr(dif, -R_2PI, R_2PI);
	return a;
}

#undef GO_SUBID


/*---------------------------------------------------------------*
 *                                                               *
 *         FOCI: both foci and point on circumference            *
 *                                                               *
 *---------------------------------------------------------------*/

#define GO_SUBID FOCI

PROP_STYLES
PROP_ALL_READ(ecc, PSTRUCT(fill_color), PT_REAL, PTR_REFERENCE, ecc, r, 0, 0)
PROP_ALL(dif, PSTRUCT(ecc), PT_REAL, PTR_ANGLE_4PI, geom.dif, r,
		is_dif_valid(value.r) && is_dif_undo_property(g->conic), (("Invalid value of difference angle."), ("Dif cannot be %f.", value.r)), CHANGED, LOW, 0, GOF_ELARC_DIF)
PROP_ALL(start, PSTRUCT(dif), PT_REAL, PTR_ANGLE_2PI, geom.start, r,
		is_start_valid(value.r) && is_start_undo_property(g->conic), (("Invalid value of start angle."), ("Start cannot be %f.", value.r)), CHANGED, LOW, 0, GOF_ELARC_START)
PROP_ALL_READ(b_radius, PSTRUCT(start), PT_REAL, PTR_NON_NEGATIVE, geom.b_radius, r, 0, 0)
PROP_ALL_READ(a_radius, PSTRUCT(b_radius), PT_REAL, PTR_NON_NEGATIVE, geom.a_radius, r, 0, 0)
PROP_ALL_READ(rotation, PSTRUCT(a_radius), PT_REAL, PTR_ANGLE_PI, geom.rotation, r, 0, 0)
PROP_ALL_READ(center_y, PSTRUCT(rotation), PT_REAL, PTR_COORDINATE, geom.center.y, r, 0, 0)
PROP_ALL_READ(center_x, PSTRUCT(center_y), PT_REAL, PTR_COORDINATE, geom.center.x, r, 0, 0)
PROP_ALL_WITHOUT_UNIT_AND_SET(conic, PSTRUCT(center_x), PT_UNS, PTU_CONIC_TYPE_1P, conic, u, LOW, 0, UINT_MAX)
SET_CONIC

PROP_VIRTUAL(PROP_STYLES_INIT,
		&PSTRUCT(conic), "conic", &PSTRUCT(center_x), "center-x", &PSTRUCT(center_y), "center-y", &PSTRUCT(rotation), "rotation",
		&PSTRUCT(a_radius), "a_radius", &PSTRUCT(b_radius), "b_radius", &PSTRUCT(start), "start", &PSTRUCT(dif), "dif", &PSTRUCT(ecc), "ecc")

INFO_ANCHOR_3(
		"focus-1", QT_POINT, LT_SIMILARITY,
		"focus-2", QT_POINT, LT_SIMILARITY,
		"point", QT_POINT, LT_AFFINITY)

INFO_HANGER_4(
		"curve", QT_CURVE, h[0], NULL, LT_AFFINITY,
		"start", QT_POINT, h[1], hanger_go_elarc_get_start, LT_AFFINITY,
		"end", QT_POINT, h[2], hanger_go_elarc_get_end, LT_AFFINITY,
		"center", QT_POINT, geom.center, NULL, LT_AFFINITY)

void go_elarc_foci_recompute(struct go_elarc *a)
{
	struct geom_point gp0 = hanger_get_position(a->a[0].h), gp1 = hanger_get_position(a->a[1].h), gp2 = hanger_get_position(a->a[2].h);
	// FIXME: Funkce geom_elliptic_arc_from_foci_point uzavre elipsu, cimz se ztrati hodnota dif, proto je nutne si ji zapamatovat a nasledne vratit. Podobne meni start.
	real dif = a->geom.dif, start = a->geom.start;
	if(!GEOM_ETEST(geom_elliptic_arc_from_foci_point(&a->geom, &gp0, &gp1, &gp2)))
		TRANS_FAIL(("Numeric error."), ("geom_elliptic_arc_from_foci_point fail."));
	if(!GEOM_ETEST(geom_elliptic_arc_set_endpoints(&a->geom, &gp2, NULL, NULL, start, dif, a->conic)))
		TRANS_FAIL(("Numeric error."), ("geom_elliptic_arc_set_endpoints fail."));
	if(!GEOM_RTEST(a->ecc = geom_elliptic_arc_eccentricity(&a->geom)))
		TRANS_FAIL(("Numeric error."), ("geom_elliptic_arc_eccentricity fail."));
}

void go_elarc_foci_transform(struct go_elarc *a, struct geom_transform *t)
{
	struct geom_elliptic_arc result;
	struct geom_point f1, f2;
	GEOM_ITEM_INIT(&result, elliptic_arc);
	if(!GEOM_ETEST(geom_elliptic_arc_transform(&a->geom, t, &result)))
		DESTROY_TRANS_FAIL(&result, ("Numeric error."), ("geom_elliptic_arc_transform fail."));
	if(!GEOM_ETEST(geom_elliptic_arc_foci(&result, &f1, &f2)))
		DESTROY_TRANS_FAIL(&result, ("Numeric error."), ("geom_elliptic_arc_foci fail."));
	DBG("Transform foci: first [ %f, %f ] -> [ %f, %f ], second [ %f, %f ] -> [ %f, %f ]\n", hanger_get_position(a->a[0].h).x, hanger_get_position(a->a[0].h).y, f1.x, f1.y, hanger_get_position(a->a[1].h).x, hanger_get_position(a->a[1].h).y, f2.x, f2.y);
	if(hanger_is_mc(a->a[0].h)) // FIXME: Prohazovani focusu
		anchor_rehang_mc_no_recompute(a->a, &f1);
	if(hanger_is_mc(a->a[1].h))
		anchor_rehang_mc_no_recompute(a->a+1, &f2);
	CHANGE_UNDO(start);
	CHANGE_UNDO(dif);
	geom_elliptic_arc_destroy(&result);
	go_elarc_foci_recompute(a);
}

struct go_elarc* go_elarc_new_foci(struct hanger *focus1, struct hanger *focus2, struct hanger *point, uns conic)
{
	if(conic > conic_max[GOST_ELARC_FOCI])
		trans_fail("Invalid conic type.");
	struct go_elarc *a = go_elarc_alloc(conic);
	go_init(&a->go, GOT_ELARC, GOST_ELARC_FOCI, focus1, focus2, point);
	return a;
}

#undef GO_SUBID


/*-----------------------------------------------------------------------*
 *                                                                       *
 *    3SMALL: the smallest ellipse with tree points on circumference     *
 *                                                                       *
 *-----------------------------------------------------------------------*/

#define GO_SUBID 3SMALL

PROP_STYLES
PROP_ALL_READ(ecc, PSTRUCT(fill_color), PT_REAL, PTR_REFERENCE, ecc, r, 0, 0)
PROP_ALL(dif, PSTRUCT(ecc), PT_REAL, PTR_ANGLE_4PI, geom.dif, r,
		is_dif_valid(value.r) && is_dif_undo_property(g->conic), (("Invalid value of difference angle."), ("Dif cannot be %f.", value.r)), CHANGED, LOW, 0, GOF_ELARC_DIF)
PROP_ALL(start, PSTRUCT(dif), PT_REAL, PTR_ANGLE_2PI, geom.start, r,
		is_start_valid(value.r) && is_start_undo_property(g->conic), (("Invalid value of start angle."), ("Start cannot be %f.", value.r)), CHANGED, LOW, 0, GOF_ELARC_START)
PROP_ALL_READ(b_radius, PSTRUCT(start), PT_REAL, PTR_NON_NEGATIVE, geom.b_radius, r, 0, 0)
PROP_ALL_READ(a_radius, PSTRUCT(b_radius), PT_REAL, PTR_NON_NEGATIVE, geom.a_radius, r, 0, 0)
PROP_ALL_READ(rotation, PSTRUCT(a_radius), PT_REAL, PTR_ANGLE_PI, geom.rotation, r, 0, 0)
PROP_ALL_READ(center_y, PSTRUCT(rotation), PT_REAL, PTR_COORDINATE, geom.center.y, r, 0, 0)
PROP_ALL_READ(center_x, PSTRUCT(center_y), PT_REAL, PTR_COORDINATE, geom.center.x, r, 0, 0)
PROP_ALL_WITHOUT_UNIT_AND_SET(conic, PSTRUCT(center_x), PT_UNS, PTU_CONIC_TYPE_3P, conic, u, LOW, 0, UINT_MAX)
SET_CONIC

PROP_VIRTUAL(PROP_STYLES_INIT,
		&PSTRUCT(conic), "conic", &PSTRUCT(center_x), "center-x", &PSTRUCT(center_y), "center-y", &PSTRUCT(rotation), "rotation",
		&PSTRUCT(a_radius), "a_radius", &PSTRUCT(b_radius), "b_radius", &PSTRUCT(start), "start", &PSTRUCT(dif), "dif", &PSTRUCT(ecc), "ecc")

INFO_ANCHOR_3(
		"point-1", QT_POINT, LT_AFFINITY,
		"point-2", QT_POINT, LT_AFFINITY,
		"point-3", QT_POINT, LT_AFFINITY)

INFO_HANGER_4(
		"curve", QT_CURVE, h[0], NULL, LT_AFFINITY,
		"start", QT_POINT, h[1], hanger_go_elarc_get_start, LT_AFFINITY,
		"end", QT_POINT, h[2], hanger_go_elarc_get_end, LT_AFFINITY,
		"center", QT_POINT, geom.center, NULL, LT_AFFINITY)


void go_elarc_3small_recompute(struct go_elarc *a)
{
	struct geom_point gp0 = hanger_get_position(a->a[0].h), gp1 = hanger_get_position(a->a[1].h), gp2 = hanger_get_position(a->a[2].h);
	// FIXME: Funkce geom_elliptic_arc_from_3_points_smallest uzavre elipsu, cimz se ztrati hodnota dif, proto je nutne si ji zapamatovat a nasledne vratit. Podobne meni start.
	real dif = a->geom.dif, start = a->geom.start;
	if(!GEOM_ETEST(geom_elliptic_arc_from_3_points_smallest(&a->geom, &gp0, &gp1, &gp2)))
		TRANS_FAIL(("Numeric error."), ("geom_elliptic_arc_from_3_points_smallest fail"));
	if(!GEOM_ETEST(geom_elliptic_arc_set_endpoints(&a->geom, &gp0, &gp2, &gp1, start, dif, a->conic)))
		TRANS_FAIL(("Numeric error."), ("geom_elliptic_arc_set_endpoints fail."));
	if(!GEOM_RTEST(a->ecc = geom_elliptic_arc_eccentricity(&a->geom)))
		TRANS_FAIL(("Numeric error."), ("geom_elliptic_arc_eccentricity fail."));
}

void go_elarc_3small_transform(struct go_elarc *a, struct geom_transform *t)
{
	real r;
	if(is_start_undo_property(a->conic))
	{
		r = angle_transform(a->geom.start, t);
		if(r != a->geom.start)
		{
			trans_new_undo_prop_virtual(void2o(a), &PSTRUCT(start), (prop_value) { .r = a->geom.start });
			a->geom.start = r;
		}
	}
	if(is_dif_undo_property(a->conic))
	{
		r = angle_dif_transform(a->geom.dif, t);
		if(r != a->geom.dif)
		{
			trans_new_undo_prop_virtual(void2o(a), &PSTRUCT(dif), (prop_value) { .r = a->geom.dif });
			a->geom.dif = r;
		}
	}
	go_elarc_3small_recompute(a);
}

struct go_elarc *go_elarc_new_3small(struct hanger *point1, struct hanger *point2, struct hanger *point3, uns conic)
{
	if(conic > conic_max[GOST_ELARC_3SMALL])
		trans_fail("Invalid conic type.");
	struct go_elarc *a = go_elarc_alloc(conic);
	go_init(&a->go, GOT_ELARC, GOST_ELARC_3SMALL, point1, point2, point3);
	return a;
}


#undef GO_SUBID


/*-----------------------------------------------------------------------*
 *                                                                       *
 *     3ECC: three points on circumference, rotation and eccentricity    *
 *                                                                       *
 *-----------------------------------------------------------------------*/

#define GO_SUBID 3ECC

PROP_STYLES
PROP_ALL(ecc, PSTRUCT(fill_color), PT_REAL, PTR_REFERENCE, ecc, r,
		is_ecc_valid(value.r), (("Invalid value of eccentricity."), ("Ecc cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL(dif, PSTRUCT(ecc), PT_REAL, PTR_ANGLE_4PI, geom.dif, r,
		is_dif_valid(value.r) && is_dif_undo_property(g->conic), (("Invalid value of difference angle."), ("Dif cannot be %f.", value.r)), CHANGED, LOW, 0, GOF_ELARC_DIF)
PROP_ALL(start, PSTRUCT(dif), PT_REAL, PTR_ANGLE_2PI, geom.start, r,
		is_start_valid(value.r) && is_start_undo_property(g->conic), (("Invalid value of start angle."), ("Start cannot be %f.", value.r)), CHANGED, LOW, 0, GOF_ELARC_START)
PROP_ALL_READ(b_radius, PSTRUCT(start), PT_REAL, PTR_NON_NEGATIVE, geom.b_radius, r, 0, 0)
PROP_ALL_READ(a_radius, PSTRUCT(b_radius), PT_REAL, PTR_NON_NEGATIVE, geom.a_radius, r, 0, 0)
PROP_ALL(rotation, PSTRUCT(a_radius), PT_REAL, PTR_ANGLE_PI, geom.rotation, r,
		is_rotation_valid(value.r), (("Invalid value of rotation."), ("Rotation cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_READ(center_y, PSTRUCT(rotation), PT_REAL, PTR_COORDINATE, geom.center.y, r, 0, 0)
PROP_ALL_READ(center_x, PSTRUCT(center_y), PT_REAL, PTR_COORDINATE, geom.center.x, r, 0, 0)
PROP_ALL_WITHOUT_UNIT_AND_SET(conic, PSTRUCT(center_x), PT_UNS, PTU_CONIC_TYPE_3P, conic, u, LOW, 0, UINT_MAX)
SET_CONIC

PROP_VIRTUAL(PROP_STYLES_INIT,
		&PSTRUCT(conic), "conic", &PSTRUCT(center_x), "center-x", &PSTRUCT(center_y), "center-y", &PSTRUCT(rotation), "rotation",
		&PSTRUCT(a_radius), "a_radius", &PSTRUCT(b_radius), "b_radius", &PSTRUCT(start), "start", &PSTRUCT(dif), "dif", &PSTRUCT(ecc), "ecc")

INFO_ANCHOR_3(
		"point-1", QT_POINT, LT_AFFINITY,
		"point-2", QT_POINT, LT_AFFINITY,
		"point-3", QT_POINT, LT_AFFINITY)

INFO_HANGER_4(
		"curve", QT_CURVE, h[0], NULL, LT_AFFINITY,
		"start", QT_POINT, h[1], hanger_go_elarc_get_start, LT_AFFINITY,
		"end", QT_POINT, h[2], hanger_go_elarc_get_end, LT_AFFINITY,
		"center", QT_POINT, geom.center, NULL, LT_AFFINITY)


void go_elarc_3ecc_recompute(struct go_elarc *a)
{
	struct geom_point gp0 = hanger_get_position(a->a[0].h), gp1 = hanger_get_position(a->a[1].h), gp2 = hanger_get_position(a->a[2].h);
	// FIXME: Funkce geom_elliptic_arc_from_3_points_rotation_eccentricity uzavre elipsu, cimz se ztrati hodnota dif, proto je nutne si ji zapamatovat a nasledne vratit. Podobne meni start.
	real dif = a->geom.dif, start = a->geom.start;
	if(!GEOM_ETEST(geom_elliptic_arc_from_3_points_rotation_eccentricity(&a->geom, &gp0, &gp1, &gp2, a->geom.rotation, a->ecc)))
		TRANS_FAIL(("Numeric error."), ("geom_elliptic_arc_from_3_points_smallest fail"));
	if(!GEOM_ETEST(geom_elliptic_arc_set_endpoints(&a->geom, &gp0, &gp2, &gp1, start, dif, a->conic)))
		TRANS_FAIL(("Numeric error."), ("geom_elliptic_arc_set_endpoints fail."));
}

// FIXME: Mozna by slo ztransformovat ecc nejak jednoduseji, ale to je na me moc slozita matika.
void go_elarc_3ecc_transform(struct go_elarc *a, struct geom_transform *t)
{
	real ecc;
	struct geom_elliptic_arc result;
	GEOM_ITEM_INIT(&result, elliptic_arc);

	if(!GEOM_ETEST(geom_elliptic_arc_transform(&a->geom, t, &result)))
		DESTROY_TRANS_FAIL(result, ("Numeric error."), ("geom_elliptic_arc_transform fail."));
	if(!GEOM_RTEST(ecc = geom_elliptic_arc_eccentricity(&result)))
		DESTROY_TRANS_FAIL(result, ("Numeric error."), ("geom_elliptic_arc_eccentricity fail."));

	CHANGE_UNDO(rotation);
	if(ecc != a->ecc)
	{
		trans_new_undo_prop_virtual(void2o(a), &PSTRUCT(ecc), (prop_value){.r = a->ecc});
		a->ecc = ecc;
	}
	
	if(is_start_undo_property(a->conic))
		CHANGE_UNDO(start);
	if(is_dif_undo_property(a->conic))
		CHANGE_UNDO(dif);

	geom_elliptic_arc_destroy(&result);
	go_elarc_3ecc_recompute(a);
}

struct go_elarc *go_elarc_new_3ecc(struct hanger *point1, struct hanger *point2, struct hanger *point3, real rot, real ecc, uns conic)
{
	if(conic > conic_max[GOST_ELARC_3ECC])
		trans_fail("Invalid conic type.");
	struct go_elarc *a = go_elarc_alloc(conic);
	go_init(&a->go, GOT_ELARC, GOST_ELARC_3ECC, point1, point2, point3);
	a->geom.rotation = rot;
	a->ecc = ecc;
	return a;
}


#undef GO_SUBID


/*-----------------------------------------------------------------------*
 *                                                                       *
 *  XY1ECC: center, point on circumference, rotation and eccentricity    *
 *                                                                       *
 *-----------------------------------------------------------------------*/

#define GO_SUBID XY1ECC

PROP_STYLES
PROP_ALL(ecc, PSTRUCT(fill_color), PT_REAL, PTR_REFERENCE, ecc, r,
		is_ecc_valid(value.r), (("Invalid value of eccentricity."), ("Ecc cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL(dif, PSTRUCT(ecc), PT_REAL, PTR_ANGLE_4PI, geom.dif, r,
		is_dif_valid(value.r) && is_dif_undo_property(g->conic), (("Invalid value of difference angle."), ("Dif cannot be %f.", value.r)), CHANGED, LOW, 0, GOF_ELARC_DIF)
PROP_ALL(start, PSTRUCT(dif), PT_REAL, PTR_ANGLE_2PI, geom.start, r,
		is_start_valid(value.r) && is_start_undo_property(g->conic), (("Invalid value of start angle."), ("Start cannot be %f.", value.r)), CHANGED, LOW, 0, GOF_ELARC_START)
PROP_ALL_READ(b_radius, PSTRUCT(start), PT_REAL, PTR_NON_NEGATIVE, geom.b_radius, r, 0, 0)
PROP_ALL_READ(a_radius, PSTRUCT(b_radius), PT_REAL, PTR_NON_NEGATIVE, geom.a_radius, r, 0, 0)
PROP_ALL(rotation, PSTRUCT(a_radius), PT_REAL, PTR_ANGLE_PI, geom.rotation, r,
		is_rotation_valid(value.r), (("Invalid value of rotation."), ("Rotation cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL_REHANG(center, PSTRUCT(rotation), geom.center, 0,
		"Hanger of the center of the elliptic arc must be a mouse click to change position.", 0)
PROP_ALL_WITHOUT_UNIT_AND_SET(conic, PSTRUCT(center_x), PT_UNS, PTU_CONIC_TYPE_3P, conic, u, LOW, 0, UINT_MAX)
SET_CONIC

PROP_VIRTUAL(PROP_STYLES_INIT,
		&PSTRUCT(conic), "conic", &PSTRUCT(center_x), "center-x", &PSTRUCT(center_y), "center-y", &PSTRUCT(rotation), "rotation",
		&PSTRUCT(a_radius), "a_radius", &PSTRUCT(b_radius), "b_radius", &PSTRUCT(start), "start", &PSTRUCT(dif), "dif", &PSTRUCT(ecc), "ecc")

INFO_ANCHOR_2(
		"center", QT_POINT, LT_AFFINITY,
		"point", QT_POINT, LT_AFFINITY)

INFO_HANGER_4(
		"curve", QT_CURVE, h[0], NULL, LT_AFFINITY,
		"start", QT_POINT, h[1], hanger_go_elarc_get_start, LT_AFFINITY,
		"end", QT_POINT, h[2], hanger_go_elarc_get_end, LT_AFFINITY,
		"center", QT_POINT, geom.center, NULL, LT_AFFINITY)


void go_elarc_xy1ecc_recompute(struct go_elarc *a)
{
	struct geom_point gp0 = hanger_get_position(a->a[0].h), gp1 = hanger_get_position(a->a[1].h);
	// FIXME: Funkce geom_elliptic_arc_from_center_point_rotation_eccentricity uzavre elipsu, cimz se ztrati hodnota dif, proto je nutne si ji zapamatovat a nasledne vratit. Podobne meni start.
	real dif = a->geom.dif, start = a->geom.start;
	if(!GEOM_ETEST(geom_elliptic_arc_from_center_point_rotation_eccentricity(&a->geom, &gp0, &gp1, a->geom.rotation, a->ecc)))
		TRANS_FAIL(("Numeric error."), ("geom_elliptic_arc_from_center_point_rotation_eccentricity fail"));
	if(!GEOM_ETEST(geom_elliptic_arc_set_endpoints(&a->geom, &gp1, NULL, NULL, start, dif, a->conic)))
		TRANS_FAIL(("Numeric error."), ("geom_elliptic_arc_set_endpoints fail."));
}

// FIXME: Tady by se taky hodila transformace excentricity
void go_elarc_xy1ecc_transform(struct go_elarc *a, struct geom_transform *t)
{
	real ecc;
	struct geom_elliptic_arc result;
	GEOM_ITEM_INIT(&result, elliptic_arc);

	if(!GEOM_ETEST(geom_elliptic_arc_transform(&a->geom, t, &result)))
		DESTROY_TRANS_FAIL(result, ("Numeric error."), ("geom_elliptic_arc_transform fail."));
	if(!GEOM_RTEST(ecc = geom_elliptic_arc_eccentricity(&result)))
		DESTROY_TRANS_FAIL(result, ("Numeric error."), ("geom_elliptic_arc_eccentricity fail."));

	CHANGE_UNDO(rotation);
	if(ecc != a->ecc)
	{
		trans_new_undo_prop_virtual(void2o(a), &PSTRUCT(ecc), (prop_value){.r = a->ecc});
		a->ecc = ecc;
	}
	
	if(is_start_undo_property(a->conic))
		CHANGE_UNDO(start);
	if(is_dif_undo_property(a->conic))
		CHANGE_UNDO(dif);

	geom_elliptic_arc_destroy(&result);
	go_elarc_xy1ecc_recompute(a);
}

struct go_elarc *go_elarc_new_xy1ecc(struct hanger *center, struct hanger *point, real rot, real ecc, uns conic)
{
	if(conic > conic_max[GOST_ELARC_XY1ECC])
		trans_fail("Invalid conic type.");
	struct go_elarc *a = go_elarc_alloc(conic);
	go_init(&a->go, GOT_ELARC, GOST_ELARC_XY1ECC, center, point);
	a->geom.rotation = rot;
	a->ecc = ecc;
	return a;
}


#undef GO_SUBID


/*----------------------------------------------------*
 *                                                    *
 *                    INITIALIZATION                  *
 *                                                    *
 *----------------------------------------------------*/

INFO_GO_5("elliptic-arc",
		XYR, "ellipse-by-center-2-radii-rotation-2-angles", 4, 0, 1, 2, 3, 1, ET_PATH, conic,
		FOCI, "ellipse-by-2-foci-point", 4, 0, 1, 2, 3, 3, ET_PATH, conic,
		3SMALL, "ellipse-by-3-points-smallest", 4, 0, 1, 2, 3, 3, ET_PATH, conic,
		3ECC, "ellipse-by-3-points-rotation-eccentricity", 4, 0, 1, 2, 3, 3, ET_PATH, conic,
		XY1ECC, "ellipse-by-center-point-rotation-eccentricity", 4, 0, 1, 2, 3, 2, ET_PATH, conic)

// FIXME: Can be faster.
int go_elarc_path_and_bbox(struct go_elarc *e, struct geom_path *path, struct geom_rectangle *bbox)
{
	struct go_info *gi = go_get_info(&e->go);
	struct geom_point gp;
	uns i;
	if(geom_old_expand(&e->geom, path) != 0 || geom_path_bbox(path, bbox) != 0)
		return 0;
	for(i = 1; i < gi->h_count; i++)
		geom_rectangle_merge_point(bbox, hanger_get_position_ptr(e->h + i, &gp), bbox);
	for(i = 1; i < gi->a_count; i++)
		geom_rectangle_merge_point(bbox, hanger_get_position_ptr(e->a[i].h, &gp), bbox);
	return 1;
}

