#include "kernel/internal.h"

#define GO_ID PARPOINT
#define GO_SUBID PARPOINT

static inline int is_parameter_valid(real p) { return p >= 0 && p <= 1; }
// static inline int is_parameter_valid(real p UNUSED) { return 1; }

PROP_ALL(coordinate_y, o_prop_name, PT_REAL, PTR_COORDINATE, geom.point.y, r,
		0, (("Read only property."), ("Bug, fix me.")), CHANGED, LOW, PTF_READ_ONLY, GOF_PARPOINT_INVALID_Y)
PROP_ALL(coordinate_x, PSTRUCT(coordinate_y), PT_REAL, PTR_COORDINATE, geom.point.x, r,
		0, (("Read only property."), ("Bug, fix me.")), CHANGED, LOW, PTF_READ_ONLY, GOF_PARPOINT_INVALID_X)
PROP_ALL(parameter, PSTRUCT(coordinate_x), PT_REAL, PTR_REFERENCE, parameter, r,
		is_parameter_valid(value.r), (("Invalid value of parameter."), ("Parameter cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)

PROP_VIRTUAL(&PSTRUCT(coordinate_x), "coordinate-x", &PSTRUCT(coordinate_y), "coordinate-y",
		&PSTRUCT(parameter), "parameter")

INFO_ANCHOR_1("curve", QT_CURVE, LT_AFFINITY)
INFO_HANGER_1("center", QT_POINT, geom.point, NULL, LT_AFFINITY)


#define COMPUTE_POINT(_go) \
do { \
	if(_go->type == GOT_GROUP) \
	{ \
		if(!(i->go.flags & GLUEE3(GOF_, GO_ID, _INVALID_X))) \
		{ \
			trans_new_undo_start(go2o(&i->go), &PSTRUCT(coordinate_x), (prop_value) { .r = i->geom.point.x }); \
			trans_new_undo_start(go2o(&i->go), &PSTRUCT(coordinate_y), (prop_value) { .r = i->geom.point.y }); \
		} \
		if(t) \
		{ \
			if(!GEOM_ETEST(geom_point_transform(&i->geom.point, t, &gp))) \
				TRANS_FAIL(("Numeric error."), ("geom_point_transform fail.\n")); \
			trans_new_undo_prop_virtual(go2o(&i->go), &PSTRUCT(coordinate_x), (prop_value) { .r = i->geom.point.x }); \
			trans_new_undo_prop_virtual(go2o(&i->go), &PSTRUCT(coordinate_y), (prop_value) { .r = i->geom.point.y }); \
			i->geom.point = gp; \
		} \
	} \
	else \
	{ \
		if(i->go.flags & GLUEE3(GOF_, GO_ID, _INVALID_X)) \
		{ \
			trans_new_undo_end(go2o(&i->go), &PSTRUCT(coordinate_x), (prop_value) { .r = i->geom.point.x }); \
			trans_new_undo_end(go2o(&i->go), &PSTRUCT(coordinate_y), (prop_value) { .r = i->geom.point.y }); \
		} \
		if(!GEOM_ETEST(go_point_at_time(_go, i->parameter, &gp))) \
			TRANS_FAIL(("Numeric error."), ("go_point_at_time fail.\n")); \
		i->geom.point = gp; \
	} \
} while(0)


static inline void compute_parametric_point(struct go_parametric_point *i, struct geom_transform *t)
{
	struct geom_point gp;
	struct go *go = hanger_get_go(i->a->h);
	COMPUTE_POINT(go);
}

void go_parametric_point_recompute(struct go_parametric_point *p)
{
	compute_parametric_point(p, NULL);
}

void go_parametric_point_transform(struct go_parametric_point *p, struct geom_transform *t)
{
	compute_parametric_point(p, t);
}

int go_parametric_point_bbox(struct go_parametric_point *p, struct geom_rectangle *r)
{
	*r = (struct geom_rectangle){ p->geom.point, p->geom.point };
	return 1;
}

struct go_parametric_point* go_parametric_point_new_on_hanger(struct hanger *h, real parameter, uns depend, real x, real y)
{
	ASSERT(hanger_is_curve(h));

	if(!is_parameter_valid(parameter))
		trans_fail("Invalid value of the parameter in a parametric point.");

	struct go_parametric_point *p = xmalloc_zero(sizeof(struct go_parametric_point));
	go_init(&p->go, GOT_PARAMETRIC_POINT, GOST_PARAMETRIC_POINT, h);
	p->parameter = parameter;
	if(depend)
	{
		p->geom.point = (struct geom_point) { x, y };
		p->go.flags |= GOF_PARPOINT_INVALID_X | GOF_PARPOINT_INVALID_Y;	
	}
	p->unit_coordinate_x = p->unit_coordinate_y = p->unit_parameter = PROP_UNIT_MAX;

	return p;
}

struct go_parametric_point* go_parametric_point_new(struct go* go, real parameter)
{
	int n = go_get_info(go)->h_curve;
	if(n == -1)
		trans_fail("Parametric point is not defined for a %s.", go_get_info(go)->name->text);

	return go_parametric_point_new_on_hanger(go_get_hanger(go, n), parameter, 0, 0, 0);
}

INFO_GO_1("parametric-point",
		PARPOINT, "parametric-point", 1, -1, -1, -1, 0, 1, ET_NONE, parameter)


#undef GO_ID
#undef GO_SUBID



#define GO_ID INTERSECTION
#define GO_SUBID INTERSECTION

PROP_ALL(coordinate_y, o_prop_name, PT_REAL, PTR_COORDINATE, geom.point.y, r,
		0, (("Read only property."), ("Bug, fix me.")), CHANGED, LOW, PTF_READ_ONLY, GOF_INTERSECTION_INVALID_Y)
PROP_ALL(coordinate_x, PSTRUCT(coordinate_y), PT_REAL, PTR_COORDINATE, geom.point.x, r,
		0, (("Read only property."), ("Bug, fix me.")), CHANGED, LOW, PTF_READ_ONLY, GOF_INTERSECTION_INVALID_X)
PROP_ALL(parameter, PSTRUCT(coordinate_x), PT_REAL, PTR_REFERENCE, parameter, r,
		is_parameter_valid(value.r), (("Invalid value of parameter."), ("Parameter cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)

PROP_VIRTUAL(&PSTRUCT(coordinate_x), "coordinate-x", &PSTRUCT(coordinate_y), "coordinate-y",
		&PSTRUCT(parameter), "parameter")

INFO_ANCHOR_2(
		"curve-1", QT_CURVE, LT_AFFINITY,
		"curve-2", QT_CURVE, LT_AFFINITY)
INFO_HANGER_1("center", QT_POINT, geom.point, NULL, LT_AFFINITY)

static int go_intersection_at(struct go *a, struct go *b, real near, struct geom_point *gp, real *param)
{
	garr g = garr_new(32, sizeof(struct geom_intersection));
	int i, count = go_intersections(a, b, GEOM_INTERSECTION_POINT | GEOM_INTERSECTION_TIME1, g);
	struct geom_intersection *inter = NULL, *inter2;
	real diff, diff2;
	if(count > 0)
	{
		inter = (struct geom_intersection *)garr_get(g, 0);
		diff = fabs(inter->time1 - near);
		for(i = 1; i < count; i++)
		{
			inter2 = (struct geom_intersection *)garr_get(g, i);
			diff2 = fabs(inter2->time1 - near);
			if (diff2 < diff) {
				inter = inter2;
				diff = diff2;
			}
		}
		*gp = inter->point;
	}
	garr_destroy(g);
	if(count <= 0)
		return 0;
	*param = inter->time1;
	return 1;
}

static void compute_intersection(struct go_intersection *i, struct geom_transform *t)
{
	struct geom_point gp;
	struct go *go1 = hanger_get_go(i->a[0].h), *go2 = hanger_get_go(i->a[1].h);
  	real param;

	if(go1->type == GOT_GROUP)
		COMPUTE_POINT(go2);

	else if(go2->type == GOT_GROUP)
		COMPUTE_POINT(go1);

	else if(go_intersection_at(go1, go2, i->parameter, &gp, &param))
	{
		if(i->go.flags & GOF_INTERSECTION_INVALID_X)
		{
			trans_new_undo_end(go2o(&i->go), &PSTRUCT(coordinate_x), (prop_value) { .r = i->geom.point.x });
			trans_new_undo_end(go2o(&i->go), &PSTRUCT(coordinate_y), (prop_value) { .r = i->geom.point.y });
		}
		i->geom.point = gp;
	}

	else if(!trans_is_undo_write_protected())
	{
		if(!(i->go.flags & GOF_INTERSECTION_INVALID_X))
		{
			trans_new_undo_start(go2o(&i->go), &PSTRUCT(coordinate_x), (prop_value) { .r = i->geom.point.x });
			trans_new_undo_start(go2o(&i->go), &PSTRUCT(coordinate_y), (prop_value) { .r = i->geom.point.y });
		}
		if(t)
		{
			if(!GEOM_ETEST(geom_point_transform(&i->geom.point, t, &gp)))
				TRANS_FAIL(("Numeric error."), ("geom_point_transform fail.\n"));
			trans_new_undo_prop_virtual(go2o(&i->go), &PSTRUCT(coordinate_x), (prop_value) { .r = i->geom.point.x });
			trans_new_undo_prop_virtual(go2o(&i->go), &PSTRUCT(coordinate_y), (prop_value) { .r = i->geom.point.y });
			i->geom.point = gp;
		}
	}
}

void go_intersection_recompute(struct go_intersection *i)
{
	compute_intersection(i, NULL);
}

void go_intersection_transform(struct go_intersection *i, struct geom_transform *t)
{
	compute_intersection(i, t);
}

int go_intersection_bbox(struct go_intersection *i, struct geom_rectangle *r)
{
	*r = (struct geom_rectangle){ i->geom.point, i->geom.point };
	return 1;
}

struct go_intersection* go_intersection_new_on_hanger(struct hanger *h1, struct hanger *h2, real parameter, uns depend, real x, real y)
{
	ASSERT(hanger_is_curve(h1) && hanger_is_curve(h2));

	if(!is_parameter_valid(parameter))
		trans_fail("Invalid value of the parameter in a intersection.");

	struct go_intersection *i = xmalloc_zero(sizeof(struct go_intersection));
	go_init(&i->go, GOT_INTERSECTION, GOST_INTERSECTION, h1, h2);
	i->parameter = parameter;
	if(depend)
	{
		i->geom.point = (struct geom_point) { x, y };
		i->go.flags |= GOF_INTERSECTION_INVALID_X | GOF_INTERSECTION_INVALID_Y;	
	}
	i->unit_coordinate_x = i->unit_coordinate_y = i->unit_parameter = PROP_UNIT_MAX;

	return i;
}

struct go_intersection* go_intersection_new(struct go *go1, struct go *go2, real parameter)
{
	int n1 = go_get_info(go1)->h_curve, n2 = go_get_info(go2)->h_curve;
	if(n1 == -1)
		trans_fail("Intersection is not defined for a %s.", go_get_info(go1)->name->text);
	if(n2 == -1)
		trans_fail("Intersection is not defined for a %s.", go_get_info(go2)->name->text);

	return go_intersection_new_on_hanger(go_get_hanger(go1, n1), go_get_hanger(go2, n2), parameter, 0, 0, 0);
}

uns go_intersection_is_valid(struct go_intersection* i)
{
	return !(i->go.flags & GOF_INTERSECTION_INVALID_X)
		&& hanger_get_go(i->a[0].h)->type != GOT_GROUP
		&& hanger_get_go(i->a[1].h)->type != GOT_GROUP;
}

INFO_GO_1("intersection",
		INTERSECTION, "intersection", 1, -1, -1, -1, 0, 2, ET_NONE, parameter)


#undef GO_ID
#undef GO_SUBID

