/*
 * VRR kernel - internal routines
 */

#ifndef _KERNEL_GO_INTERNAL_H_
#define _KERNEL_GO_INTERNAL_H_

#define GO_FOR_ALL_ID(_macro) _macro(GROUP) _macro(POINT) _macro(SEGMENT) _macro(BEZIER) _macro(ELARC) _macro(TEXT) _macro(TEX) \
	_macro(PARPOINT) _macro(INTERSECTION) _macro(DECPOINT) _macro(DIRECTION)

#define DECLARE_GO_BASIC(_id) \
void GLUEE(GO_NAME(_id), _init) (void); \
void GLUEE(GO_NAME(_id), _finish) (void); \
void GLUEE(GO_NAME(_id), _recompute) (GO_STRUCT(_id) *go); \
void GLUEE(GO_NAME(_id), _transform) (GO_STRUCT(_id) *p, struct geom_transform *t); \
int GLUEE(GO_NAME(_id), _bbox) (GO_STRUCT(_id) *p, struct geom_rectangle *bbox); \
int GLUEE(GO_NAME(_id), _path_and_bbox) (GO_STRUCT(_id) *p, struct geom_path *path, struct geom_rectangle *bbox); \
void GLUEE(GO_NAME(_id), _dump) (GO_STRUCT(_id) *p); \
extern struct go_info GLUEE(GO_NAME(_id), _info) [];

#define DECLARE_SUBGO_BASIC(_id, _sid) \
void GLUEE(GO_SUBNAME(_id, _sid), _recompute) (GO_STRUCT(_id) *go); \
void GLUEE(GO_SUBNAME(_id, _sid), _transform) (GO_STRUCT(_id) *p, struct geom_transform *t);

/*
 * go.c
 */

struct geom_point* point_transform(struct geom_point *p, const struct geom_transform *t);
void go_recompute(struct go *go);
void go_transform(struct go *go, struct geom_transform2 *t);
void go_get_rtree_bbox(struct go *go, struct geom_rectangle *bbox);
void rectangle_normalize(struct geom_rectangle *gr);


/*
 * go_group.c
 */

#define GO_NAME_GROUP go_group
#define GO_TYPE_GROUP GOT_GROUP
#define GO_STRUCT_GROUP go_group
#define GO_GEOM_GROUP group

#define GO_SUBTYPE_GROUP_GROUP GOST_GROUP
#define GO_SUBNAME_GROUP_GROUP go_group_group
#define GO_SUBTYPE_GROUP_TLG GOST_GROUP_TLG
#define GO_SUBNAME_GROUP_TLG go_group_tlg
#define GO_SUBTYPE_GROUP_TEMP GOST_GROUP_TEMP
#define GO_SUBNAME_GROUP_TEMP go_group_temp
#define GO_SUBTYPE_GROUP_ZOMBIE GOST_GROUP_ZOMBIE
#define GO_SUBNAME_GROUP_ZOMBIE go_group_zombie

	
/*
 * go_point.c
 */

#define GO_NAME_POINT go_point
#define GO_TYPE_POINT GOT_POINT
#define GO_STRUCT_POINT go_point
#define GO_GEOM_POINT point_item

#define GO_SUBNAME_POINT_POINT go_point
#define GO_SUBTYPE_POINT_POINT GOST_POINT
#define GO_SUBSTRUCT_POINT_POINT go_point

	
/*
 * go_seg.c
 */

#define GO_NAME_SEGMENT go_segment
#define GO_TYPE_SEGMENT GOT_SEGMENT
#define GO_STRUCT_SEGMENT go_segment
#define GO_GEOM_SEGMENT segment

#define GO_SUBNAME_SEGMENT_SEGMENT go_segment
#define GO_SUBTYPE_SEGMENT_SEGMENT GOST_SEGMENT
#define GO_SUBSTRUCT_SEGMENT_SEGMENT go_segment

	
/*
 * go_bez.c
 */

#define GO_NAME_BEZIER go_bezier
#define GO_TYPE_BEZIER GOT_BEZIER
#define GO_STRUCT_BEZIER go_bezier
#define GO_GEOM_BEZIER bezier

#define GO_SUBNAME_BEZIER_CUBIC go_bezier_cubic
#define GO_SUBTYPE_BEZIER_CUBIC GOST_BEZIER_CUBIC
#define GO_SUBSTRUCT_BEZIER_CUBIC go_bezier
DECLARE_SUBGO_BASIC(BEZIER, CUBIC)

#define GO_SUBNAME_BEZIER_QUADRATIC go_bezier_quadratic
#define GO_SUBTYPE_BEZIER_QUADRATIC GOST_BEZIER_QUADRATIC
#define GO_SUBSTRUCT_BEZIER_QUADRATIC go_bezier
DECLARE_SUBGO_BASIC(BEZIER, QUADRATIC)

	
/*
 * go_elarc.c
 */

#define GO_NAME_ELARC go_elarc
#define GO_TYPE_ELARC GOT_ELARC
#define GO_STRUCT_ELARC go_elarc
#define GO_GEOM_ELARC elliptic_arc

#define GO_SUBNAME_ELARC_XYR go_elarc_xyr
#define GO_SUBTYPE_ELARC_XYR GOST_ELARC_XYR
#define GO_SUBSTRUCT_ELARC_XYR go_elarc
DECLARE_SUBGO_BASIC(ELARC, XYR)

#define GO_SUBNAME_ELARC_FOCI go_elarc_foci
#define GO_SUBTYPE_ELARC_FOCI GOST_ELARC_FOCI
#define GO_SUBSTRUCT_ELARC_FOCI go_elarc
DECLARE_SUBGO_BASIC(ELARC, FOCI)

#define GO_SUBNAME_ELARC_3SMALL go_elarc_3small
#define GO_SUBTYPE_ELARC_3SMALL GOST_ELARC_3SMALL
#define GO_SUBSTRUCT_ELARC_3SMALL go_elarc
DECLARE_SUBGO_BASIC(ELARC, 3SMALL)

#define GO_SUBNAME_ELARC_3ECC go_elarc_3ecc
#define GO_SUBTYPE_ELARC_3ECC GOST_ELARC_3ECC
#define GO_SUBSTRUCT_ELARC_3ECC go_elarc
DECLARE_SUBGO_BASIC(ELARC, 3ECC)

#define GO_SUBNAME_ELARC_XY1ECC go_elarc_xy1ecc
#define GO_SUBTYPE_ELARC_XY1ECC GOST_ELARC_XY1ECC
#define GO_SUBSTRUCT_ELARC_XY1ECC go_elarc
DECLARE_SUBGO_BASIC(ELARC, XY1ECC)


/*
 * go_fork.c
 */

#define GO_NAME_PARPOINT go_parametric_point
#define GO_TYPE_PARPOINT GOT_PARAMETRIC_POINT
#define GO_STRUCT_PARPOINT go_parametric_point
#define GO_GEOM_PARPOINT point_item

#define GO_SUBNAME_PARPOINT_PARPOINT go_parametric_point
#define GO_SUBTYPE_PARPOINT_PARPOINT GOST_PARAMETRIC_POINT
#define GO_SUBSTRUCT_PARPOINT_PARPOINT go_parametric_point
	
#define GO_NAME_INTERSECTION go_intersection
#define GO_TYPE_INTERSECTION GOT_INTERSECTION
#define GO_STRUCT_INTERSECTION go_intersection
#define GO_GEOM_INTERSECTION point_item

#define GO_SUBNAME_INTERSECTION_INTERSECTION go_intersection
#define GO_SUBTYPE_INTERSECTION_INTERSECTION GOST_INTERSECTION
#define GO_SUBSTRUCT_INTERSECTION_INTERSECTION go_intersection
	

/*
 * go_text.c
 */
	
#define GO_NAME_TEXT go_text
#define GO_TYPE_TEXT GOT_TEXT
#define GO_STRUCT_TEXT go_text
#define GO_GEOM_TEXT point_item

#define GO_SUBNAME_TEXT_TEXT go_text
#define GO_SUBTYPE_TEXT_TEXT GOST_TEXT
#define GO_SUBSTRUCT_TEXT_TEXT go_text

	
#define GO_NAME_TEX go_tex_text
#define GO_TYPE_TEX GOT_TEX_TEXT
#define GO_STRUCT_TEX go_tex_text
#define GO_GEOM_TEX point_item

#define GO_SUBNAME_TEX_TEX go_tex_text
#define GO_SUBTYPE_TEX_TEX GOST_TEX_TEXT
#define GO_SUBSTRUCT_TEX_TEX go_tex_text


/*
 * go_decorator.c
 */

#define GO_NAME_DECPOINT go_decorator_point
#define GO_TYPE_DECPOINT GOT_DECORATOR_POINT
#define GO_STRUCT_DECPOINT go_decorator_point
#define GO_GEOM_DECPOINT callback_item

#define GO_SUBNAME_DECPOINT_DECPOINT go_decorator_point
#define GO_SUBTYPE_DECPOINT_DECPOINT GOST_DECORATOR_POINT
#define GO_SUBSTRUCT_DECPOINT_DECPOINT go_decorator_point

	
#define GO_NAME_DIRECTION go_direction
#define GO_TYPE_DIRECTION GOT_DIRECTION
#define GO_STRUCT_DIRECTION go_direction
#define GO_GEOM_DIRECTION callback_item

#define GO_SUBNAME_DIRECTION_DIRECTION go_direction
#define GO_SUBTYPE_DIRECTION_DIRECTION GOST_DIRECTION
#define GO_SUBSTRUCT_DIRECTION_DIRECTION go_direction



/*
 * common
 */

GO_FOR_ALL_ID(DECLARE_GO_BASIC)

#endif
