#include "kernel/internal.h"


#ifdef DEBUG_HOOK
#define DBG_HOOK(...) msg(__VA_ARGS__)
#else
#define DBG_HOOK(...) do { } while(0)
#endif

static slist obj_hooks;
static slist unit_hooks;


/* Call a hook _func of the obj _obj.  */
#define HOOK_OBJ_CALL(_func, ...) do { snode *_h, *_tmp; SLIST_WALK_DELSAFE(_h, obj_hooks, _tmp) { \
	struct obj_hook *_o = SKIP_BACK(struct obj_hook, n, _h); \
	if(_o->_func) _o->_func(__VA_ARGS__, _o); } } while(0)

/* dtto for go */
#define HOOK_GO_CALL(_func, ...) do { snode *_h, *_tmp; SLIST_WALK_DELSAFE(_h, trans_tlo->go_hooks, _tmp) { \
	struct go_hook *_g = SKIP_BACK(struct go_hook, n, _h); \
	if(_g->_func) _g->_func(__VA_ARGS__, _g); } } while(0)

/* dtto for trans */
#define HOOK_TRANS_CALL(_tlo, _func, ...) do { snode *_h, *_tmp; SLIST_WALK_DELSAFE(_h, _tlo->trans_hooks, _tmp) { \
	struct trans_hook *_t = SKIP_BACK(struct trans_hook, n, _h); \
	if(_t->_func) _t->_func(__VA_ARGS__, _t); } } while(0)

/* dtto for unit */
#define HOOK_UNIT_CALL(_func, ...) do { snode *_h, *_tmp; SLIST_WALK_DELSAFE(_h, unit_hooks, _tmp) { \
	struct unit_hook *_o = SKIP_BACK(struct unit_hook, n, _h); \
	if(_o->_func) _o->_func(__VA_ARGS__, _o); } } while(0)

#define WALK_OBJ_HOOK(_walk) for(_walk = (struct obj_hook*)(obj_hooks.head.next); _walk; _walk = (struct obj_hook*)_walk->n.next)
#define WALK_UNIT_HOOK(_walk) for(_walk = (struct unit_hook*)(unit_hooks.head.next); _walk; _walk = (struct unit_hook*)_walk->n.next)

void hook_init(void)
{
	slist_init(&obj_hooks);
	slist_init(&unit_hooks);
}

void hook_finish(void)
{
	struct unit_hook *uh;
	struct obj_hook *oh;
	WALK_OBJ_HOOK(oh)
		DANGEROUS(0, "Undeleted obj hook %p.\n", oh);
	WALK_UNIT_HOOK(uh)
		DANGEROUS(0, "Undeleted obj hook %p.\n", uh);
}

void hook_call_obj_link(struct obj* obj)
{
	DBG_HOOK("Hook call: obj link, object %s at %p, father %s at %p\n", obj_lookup_name(obj), obj, obj_lookup_name(obj->father), obj->father);
  	HOOK_OBJ_CALL(obj_link, obj2o(obj));
}

void hook_call_obj_unlink(struct obj* obj, struct obj* father, struct obj* prev)
{
	DBG_HOOK("Hook call: obj unlink, object %s at %p, father %s at %p\n", obj_lookup_name(obj), obj, obj_lookup_name(father), father);
	HOOK_OBJ_CALL(obj_unlink, obj2o(obj), obj2o(father), prev ? obj2o(prev) : NULL);
}

void hook_call_obj_selected(struct obj* obj)
{
	DBG_HOOK("Hook call: obj selected, object %s at %p\n", obj_lookup_name(obj), obj);
	HOOK_OBJ_CALL(obj_selected, obj2o(obj));
}

void hook_call_obj_unselected(struct obj* obj)
{
	DBG_HOOK("Hook call: obj unselected, object %s at %p\n", obj_lookup_name(obj), obj);
	HOOK_OBJ_CALL(obj_unselected, obj2o(obj));
}



void hook_call_go_link(struct go* go)
{
	DBG_HOOK("Hook call: go link, go %s at %p, group %s at %p, tlo %s at %p\n", go_lookup_name(go), go, go_lookup_name(&go_get_group(go)->go), go_get_group(go), obj_lookup_name(&go_get_root(go)->tlo->obj), go_get_root(go)->tlo);
	HOOK_GO_CALL(go_link, go2o(go));
}

void hook_call_go_unlink(struct go* go, struct go_group* owner, struct go *prev)
{
	DBG_HOOK("Hook call: go unlink, go %s at %p, group %s at %p, tlo %s at %p\n", go_lookup_name(go), go, go_lookup_name(&owner->go), owner, obj_lookup_name(&go_get_root(&owner->go)->tlo->obj), go_get_root(&owner->go)->tlo);
	HOOK_GO_CALL(go_unlink, go2o(go), void2o(owner), prev ? go2o(prev) : NULL);
}

void hook_call_go_relink(struct go* go, struct go_group *group, struct go *prev)
{
	DBG_HOOK("Hook call: go relink, go %s at %p, group %s at %p\n", go_lookup_name(go), go, go_lookup_name(&group->go), group);
	HOOK_GO_CALL(go_relink, go2o(go), group, prev);
}

void hook_call_go_selected(struct go* go)
{
	DBG_HOOK("Hook call: go selected, go %s at %p\n", go_lookup_name(go), go);
	HOOK_GO_CALL(go_selected, go2o(go));
}

void hook_call_go_unselected(struct go* go)
{
	DBG_HOOK("Hook call: go unselected, go %s at %p\n", go_lookup_name(go), go);
	HOOK_GO_CALL(go_unselected, go2o(go));
}

void hook_call_go_changed_generic(struct go* go, const struct changed_data_generic *d)
{
	HOOK_GO_CALL(go_changed, go, d);
}

#define PRINT_BBOX(b) (b)->ll.x, (b)->ll.y, (b)->ur.x, (b)->ur.x

void hook_call_go_changed(struct go* go, const struct geom_rectangle *old_bbox, const struct geom_rectangle *new_bbox)
{
	DBG_HOOK("Hook call: go changed, go %s at %p, old bbox %f %f %f %f, new bbox %f %f %f %f\n", go_lookup_name(go), go, PRINT_BBOX(old_bbox), PRINT_BBOX(new_bbox));
	struct changed_data d;

	d.kind = CK_CHANGED;
	d.old_bbox = old_bbox;
	d.new_bbox = new_bbox;
	hook_call_go_changed_generic(go, (void*) &d);
}

void hook_call_go_altered(struct go* go, const struct geom_rectangle *where)
{
	DBG_HOOK("Hook call: go altered, go %s at %p, bbox %f %f %f %f\n", go_lookup_name(go), go, PRINT_BBOX(where));
	struct altered_data d;

	d.kind = CK_ALTERED;
	d.where = where;
	hook_call_go_changed_generic (go, (void*) &d);
}

void hook_call_go_transformed(struct go* go, const struct geom_rectangle *old_bbox, const struct geom_rectangle *new_bbox, const struct geom_transform2 *t)
{
	DBG_HOOK("Hook call: go transformed, go %s at %p\n", go_lookup_name(go), go);
	struct transformed_data d;

	d.kind = CK_TRANSFORMED;
	d.old_bbox = old_bbox;
	d.new_bbox = new_bbox;
	d.transformation = t;
	hook_call_go_changed_generic (go, (void*) &d);
}


void hook_call_prop_create(struct o* o, struct prop* p)
{
	char buf[PROP_DUMP_SIZE] UNUSED;
	if(!o_is_linked(o))
		return;
	if(o->kind == T_OBJ)
	{
		DBG_HOOK("Hook call: obj property create, object %s at %p; property %s\n", o_lookup_name(o), o, prop_dump(buf, o, p));
		HOOK_OBJ_CALL(obj_prop_create, o, p);
	}
	else
	{
		DBG_HOOK("Hook call: go property create, go %s at %p; property %s\n", o_lookup_name(o), o, prop_dump(buf, o, p));
		HOOK_GO_CALL(go_prop_create, o, p);
	}
}

void hook_call_prop_change(struct o* o)
{
	if(!o_is_linked(o))
		return;
	if(o->kind == T_OBJ)
	{
		DBG_HOOK("Hook call: obj property change, object %s at %p\n", o_lookup_name(o), o);
		HOOK_OBJ_CALL(obj_prop_change, o);
	}
	else
	{
		DBG_HOOK("Hook call: go property change, go %s at %p\n", o_lookup_name(o), o);
		HOOK_GO_CALL(go_prop_change, o);
	}
}

void hook_call_prop_delete(struct o* o, struct prop* p)
{
	if(!o_is_linked(o))
		return;
	char buf[PROP_DUMP_SIZE] UNUSED;
	if(o->kind == T_OBJ)
	{
		struct obj* obj = o2obj(o);
		DBG_HOOK("Hook call: obj property delete, object %s at %p; property %s\n", obj_lookup_name(obj), obj, prop_dump(buf, o, p));
		HOOK_OBJ_CALL(obj_prop_delete, obj2o(obj), p);
	}
	else
	{
		struct go* go = o2go(o);
		DBG_HOOK("Hook call: go property delete, go %s at %p; property %s\n", go_lookup_name(go), go, prop_dump(buf, o, p));
		HOOK_GO_CALL(go_prop_delete, go2o(go), p);
	}
}



void hook_call_undo_new(struct obj_tlo* tlo, struct undo_gui* ug)
{
	DBG_HOOK("Hook call: undo new, object %s at %p, undo_gui %s at %p\n", obj_lookup_name(&tlo->obj), tlo, ug->name_operation->text, ug);
	HOOK_TRANS_CALL(tlo, undo_new, tlo, ug);
}

void hook_call_undo_delete(struct obj_tlo* tlo, struct undo_gui* ug, struct undo_gui* prev_item)
{
	DBG_HOOK("Hook call: undo delete, object %s at %p, undo_gui %s at %p\n", obj_lookup_name(&tlo->obj), tlo, ug->name_operation->text, ug);
	HOOK_TRANS_CALL(tlo, undo_delete, tlo, ug, prev_item);
}

void hook_call_undo_update(struct obj_tlo* tlo, struct undo_gui* ug)
{
	DBG_HOOK("Hook call: undo update, object %s at %p, undo_gui %s at %p\n", obj_lookup_name(&tlo->obj), tlo, ug->name_operation->text, ug);
	HOOK_TRANS_CALL(tlo, undo_update, tlo, ug);
}

void hook_call_undo(struct obj_tlo* tlo)
{
	DBG_HOOK("Hook call: undo, object %s at %p\n", obj_lookup_name(&tlo->obj), tlo);
	HOOK_TRANS_CALL(tlo, undo, tlo);
}

void hook_call_redo(struct obj_tlo* tlo)
{
	DBG_HOOK("Hook call: redo, object %s at %p\n", obj_lookup_name(&tlo->obj), tlo);
	HOOK_TRANS_CALL(tlo, redo, tlo);
}

void hook_call_unit_add(uns slot)
{
	DBG_HOOK("Hook call: unit add, slot %u, name %s, multiplier %f, quantity = %u, unused %u\n", slot, unit_array[slot].name->text, unit_array[slot].multiplier, unit_array[slot].quantity, unit_array[slot].used);
	HOOK_UNIT_CALL(add, slot);
}

void hook_call_unit_unused(uns slot)
{
	DBG_HOOK("Hook call: unit add, slot %u, name %s, multiplier %f, quantity = %u, unused %u\n", slot, unit_array[slot].name->text, unit_array[slot].multiplier, unit_array[slot].quantity, unit_array[slot].used);
	HOOK_UNIT_CALL(unused, slot);
}

void hook_call_unit_changed(uns slot)
{
	DBG_HOOK("Hook call: unit add, slot %u, name %s, multiplier %f, quantity = %u, unused %u\n", slot, unit_array[slot].name->text, unit_array[slot].multiplier, unit_array[slot].quantity, unit_array[slot].used);
	HOOK_UNIT_CALL(changed, slot);
}

void hook_call_unit_default(uns def)
{
	DBG_HOOK("Hook call: unit default change, name %s, multiplier %f, quantity = %d, unused %u\n", unit_array[unit_default[def]].name->text, unit_array[unit_default[def]].multiplier, def, unit_array[def].used);
	HOOK_UNIT_CALL(def, def);
}

struct obj_hook* obj_hook_add(struct obj_hook *hook)
{
	slist_add(&obj_hooks, &hook->n);
	return hook;
}

struct obj_hook* obj_hook_remove(struct obj_hook *hook)
{
	slist_remove(&obj_hooks, &hook->n);
	return hook;
}

struct obj_hook* obj_hook_remove_by_data(void *data)
{
	struct obj_hook *oh;
	for(snode* p = &obj_hooks.head; p; p = p->next)
		if(SKIP_BACK(struct obj_hook, n, p->next)->data == data)
		{
			oh = SKIP_BACK(struct obj_hook, n, p->next);
			slist_remove_after(p);
			return oh;
		}
	DANGEROUS(0, "There in no obj_hook with data %p\n", data);
	return NULL;
}

struct go_hook* go_hook_remove_by_data(struct obj_tlo *tlo, void * data)
{
	ASSERT_TLO(tlo);
	struct go_hook *gh;
	for(snode* p = &tlo->go_hooks.head; p; p = p->next)
		if(SKIP_BACK(struct go_hook, n, p->next)->data == data)
		{
			gh = SKIP_BACK(struct go_hook, n, p->next);
			slist_remove_after(p);
			return gh;
		}
	DANGEROUS(0, "Tlo %s at %p has no go_hook with data %p\n", obj_lookup_name(&tlo->obj), tlo, data);
	return NULL;
}

struct trans_hook* trans_hook_add(struct obj_tlo *tlo, struct trans_hook *hook)
{
	slist_add(&tlo->trans_hooks, &hook->n);
	return hook;
}

struct trans_hook* trans_hook_remove(struct obj_tlo *tlo, struct trans_hook *hook)
{
	UNUSED int ret = slist_remove_safe(&tlo->trans_hooks, &hook->n);
	ASSERT(ret);
	return hook;
}

struct trans_hook* trans_hook_remove_by_data(struct obj_tlo *tlo, void * data)
{
	ASSERT_TLO(tlo);
	struct trans_hook *th;
	for(snode* p = &tlo->trans_hooks.head; p; p = p->next)
		if(SKIP_BACK(struct trans_hook, n, p->next)->data == data)
		{
			th = SKIP_BACK(struct trans_hook, n, p->next);
			slist_remove_after(p);
			return th;
		}
	DANGEROUS(0, "Tlo %s at %p has no trans_hook with data %p\n", obj_lookup_name(&tlo->obj), tlo, data);
	return NULL;
}

struct unit_hook* unit_hook_add(struct unit_hook *hook)
{
	slist_add(&unit_hooks, &hook->n);
	return hook;
}

struct unit_hook* unit_hook_remove(struct unit_hook *hook)
{
	slist_remove(&unit_hooks, &hook->n);
	return hook;
}

struct unit_hook* unit_hook_remove_by_data(void *data)
{
	struct unit_hook *oh;
	for(snode* p = &unit_hooks.head; p; p = p->next)
		if(SKIP_BACK(struct unit_hook, n, p->next)->data == data)
		{
			oh = SKIP_BACK(struct unit_hook, n, p->next);
			slist_remove_after(p);
			return oh;
		}
	DANGEROUS(0, "There is no unit_hook with data %p\n", data);
	return NULL;
}

