#include "kernel/internal.h"

/*
 * Initialization and cleanup
 */

static void tlo_alloc_finish(struct obj_tlo *t)
{
#define FMALLOC_FINISH(_f) if(fmalloc_is_empty(t->_f)) fmalloc_finish(t->_f); else t->_f->call_if_empty = fmalloc_sfree_head;
	FMALLOC_FINISH(fname);
	FMALLOC_FINISH(fprop);
	FMALLOC_FINISH(fuk);
}

static struct obj* obj_init(struct obj *obj, uns type, struct obj *father)
{
	o_init((struct o*)obj, T_OBJ, type, 0, &o_prop_name);
	obj->flags = 0;
	clist_init(&obj->lson);
	obj->father = father;
	if(father)
		clist_add_tail(&father->lson, &obj->nson);
	return obj;
}

static void obj_finish(struct obj *obj)
{
	ASSERT_OBJ(obj);
	ASSERT(obj->father == obj_zombie && obj->ref_count == 0);	// The object's father must be the zombie otherwise this object has another reference.
	struct obj* son;

	clist_remove(&obj->nson);		//	Remove from zombie

	while((son = obj_get_first_son(obj)))
		obj_relink(son, obj_zombie, obj);

	switch(obj->type)
	{
		case OT_UNIVERSUM:
			die("Universum cannot be deleted by this function");

		case OT_DOCUMENT:
			break;

		case OT_TLO:
			{
				struct obj_tlo* tlo = obj2tlo(obj);
				struct go_hook* gh;
				struct trans_hook* th;
				struct go *go;
				struct anchor *a;
				int i;

				WALK_GO_HOOK(tlo, gh)
					 DANGEROUS(0, "There is an undeleted go hook %p in tlo %s at %p.\n", gh, obj_lookup_name(obj), obj);
				WALK_TRANS_HOOK(tlo, th)
					DANGEROUS(0, "There is an undeleted trans hook %p in tlo %s at %p.\n", th, obj_lookup_name(obj), obj);

				trans_delete_all(tlo);
				string_unref(tlo_universum->undo_start.name_operation);

				DANGEROUS(tlo->expand_ref_count == 0, "Somebody has %u reference to expansion table in tlo %s at %p.\n",
						tlo->expand_ref_count, obj_lookup_name(obj), obj);
				if(tlo->expand_ref_count > 0)
				{
					tlo->expand_ref_count = 1;
					expand_unref(tlo);
				}

				WALK_GO_RECURSIVE_WITHOUT_ROOT(tlg2go(tlo->tlg), go)
				{
					expand_remove(tlo, go);
					WALK_GO_ANCHOR(go, a, i)
						clist_remove(&a->nanchor);
				}

				tlo->tlg->tlo = NULL;
				go_insert(tlg2go(tlo->tlg), &tlo_zombie->tlg->group, NULL);
				go_unref(tlg2go(tlo->tlg));

				ASSERT(clist_empty(&tlo->temp->mc) && !group_get_first_go(&tlo->temp->group));
				o_finish(void2o(tlo->temp));

				tlo_alloc_finish(tlo);
			}
			break;

		default:
			die("There is no label for this type %d\n", obj->type);
	}

	o_finish(obj2o(obj));
}


static struct obj_tlo* tlo_init_trans_prop(struct obj_tlo *tlo)
{
	clist_init((clist*)&tlo->undo_start.ng);
	tlo->undo_present = &tlo->undo_start;
	tlo->undo_saved = &tlo->undo_start;
	tlo->undo_start.name_operation = string_lookup_ref("The begin of history.");
	slist_init(&tlo->trans_hooks);

	tlo->fname = smalloc(sizeof(struct fmalloc_head));
	tlo->fprop = smalloc(sizeof(struct fmalloc_head));
	tlo->fuk = smalloc(sizeof(struct fmalloc_head));
	fmalloc_init(tlo->fname, prop_hash_name_size);
	fmalloc_init(tlo->fprop, sizeof(struct prop));
	fmalloc_init(tlo->fuk, sizeof(struct undo_kernel));

	slist_init(&tlo->trans_hooks);
	slist_init(&tlo->go_hooks);

	flist_init(&tlo->tsort);
	return tlo;
}

void object_init(void)
{
	static struct go_tlg zombie;

	clist_init(&obj_lsel);

	obj_init(obj_universum, OT_UNIVERSUM, NULL);
	tlo_init_trans_prop(tlo_universum);
	prop_name_set_low(obj2o(obj_universum), string_lookup_ref("universum"), tlo_universum);

	obj_init(obj_temp, OT_TEMP, NULL);
	prop_name_set_low(obj2o(obj_temp), string_lookup_ref("temp"), tlo_universum);

	obj_init(&tlo_zombie->obj, OT_ZOMBIE, NULL);
	prop_name_set_low(void2o(tlo_zombie), string_lookup_ref("tlo_zombie"), tlo_universum);
	group_init(&zombie.group, GOST_GROUP_ZOMBIE);
	zombie.tlo = tlo_zombie;
	tlo_zombie->tlg = &zombie;
	prop_name_set_low(void2o(tlo_zombie->tlg), string_lookup_ref("tlo_zombie->tlg"), tlo_universum);
}

void object_finish(void)
{
	ASSERT(!trans_present && !trans_tlo);
	struct obj *obj;
	struct trans_hook *th;
	struct go *go;
	struct hanger_smart *hs;

	prop_finish(obj2o(obj_temp));
	DANGEROUS(obj_temp->ref_count == 1, "Somebody has %u reference to the temp.\n", obj_temp->ref_count - 1);

// Finish universum
	while((obj = obj_get_first_selected()))
		obj_unselect_low(obj);
	WALK_TRANS_HOOK(tlo_universum, th)
		DANGEROUS(0, "There is an undeleted trans hook %p in universum.\n", th);
	trans_delete_all(tlo_universum);
	string_unref(tlo_universum->undo_start.name_operation);
	while((obj = obj_get_first_son(obj_universum)))
		obj_relink(obj, obj_zombie, obj_universum);
	prop_finish(obj2o(obj_universum));
	DANGEROUS(obj_universum->ref_count == 1, "Somebody has %u referencies to the universum.\n", obj_universum->ref_count - 1);

// Finish zombie
	while((obj = obj_get_first_son(obj_zombie)))
	{
		ASSERT(obj->kind == T_OBJ && obj->father == obj_zombie);
		DANGEROUS(0, "Somebody has %d references to an object %s at %p.\n", obj->ref_count, obj_lookup_name(obj), obj);
		obj->ref_count = 0;
		obj_finish(obj);
	}
	while((go = group_get_first_go(&tlo_zombie->tlg->group)))
	{
		DANGEROUS(0, "Somebody has %d references to a go %s at %p.\n", go->ref_count, go_lookup_name(go), go);
		go->ref_count = 0;	// The zombie could be cleaned a little bit carefully.
		go_finish(go);
	}
	while((hs = tlg_get_first_hanger_smart(tlo_zombie->tlg)))
	{
		ASSERT(hs->go == tlg2go(tlo_zombie->tlg));
		DANGEROUS(0, "Somebody has %u reference to a mouse click %p.\n", hs->ref_count, hs);
		hs->ref_count = 1;
		hanger_unref(&hs->h);
	}
	prop_finish(void2o(tlo_zombie->tlg));
	DANGEROUS(tlo_zombie->tlg->group.go.ref_count == 1, "Somebody has %u reference to the zombie->tlg.\n", tlo_zombie->tlg->group.go.ref_count - 1);
	prop_finish(obj2o(obj_zombie));
	DANGEROUS(obj_zombie->ref_count == 1, "Somebody has %u reference to the zombie.\n", obj_zombie->ref_count - 1);

	tlo_alloc_finish(tlo_universum);
}


/*
 * Functions for creating objects.
 */

struct obj_doc* doc_new(void)
{
	ASSERT_META_TRANS;
	struct obj_doc *doc = xmalloc(sizeof(struct obj_doc));
	obj_init(&doc->obj, OT_DOCUMENT, obj_temp);
	prop_set_low(void2o(doc), string_ref(string_filename), PT_STRING, PTS_FILE_NAME, 0, 0, (prop_value){.s = string_ref(string_file_unsaved)});
	doc->saved = 1;
	doc->saved_start = doc->saved_end = NULL;
	return doc;
}

struct obj_doc* doc_new_named(const char *name)
{
	ASSERT_META_TRANS;
	struct obj_doc* doc = doc_new();
	prop_name_set_low(obj2o(&doc->obj), string_lookup_ref(name), tlo_universum);
	return doc;
}

struct obj_tlo* tlo_new(void)
{
	ASSERT_META_TRANS;
	struct obj_tlo *tlo = xmalloc(sizeof(struct obj_tlo));
	obj_init(&tlo->obj, OT_TLO, obj_temp);
	tlo_init_trans_prop(tlo);

	tlo->tlg = xmalloc_zero(sizeof(struct go_tlg));
	tlo->tlg->tlo = tlo;
	group_init(&tlo->tlg->group, GOST_GROUP_TLG);
	prop_name_set_low(void2o(tlo->tlg), string_lookup_ref("top-level-group"), tlo);

	tlo->temp = xmalloc_zero(sizeof(struct go_tlg));
	tlo->temp->tlo = tlo;
	group_init(&tlo->temp->group, GOST_GROUP_TEMP);
	prop_name_set_low(void2o(tlo->temp), string_lookup_ref("temp-group"), tlo);

	tlo->expand_ref_count = 0;
	tlo->expand = NULL;

	return tlo;
}

struct obj_tlo* tlo_new_named(const char *name)
{
	ASSERT_META_TRANS;
	struct obj_tlo* tlo = tlo_new();
	prop_name_set_low(obj2o(&tlo->obj), string_lookup_ref(name), tlo_universum);
	return tlo;
}


/*
 * Linking functions
 */

void obj_relink(struct obj *obj, struct obj *father, struct obj *old_father)
{
	ASSERT_OBJ(obj);
	ASSERT_OBJ(father);
	ASSERT(old_father == obj->father && father != old_father);
	ASSERT_KERNEL_SLOW(clist_is_node_member(&old_father->lson, &obj->nson));

#ifdef DEBUG_KERNEL_SLOW
	struct obj *walk;
	WALK_OBJ_RECURSIVE(obj, walk)
		ASSERT(!obj_is_selected(walk));
#endif

	struct obj *old_prev = obj_get_prev_son(obj);
	clist_remove(&obj->nson);
	clist_add_tail(&father->lson, &obj->nson);
	obj->father = father;
	
	if(obj_get_root(father) == obj_universum)
		hook_call_obj_link(obj);
	else if(obj_get_root(old_father) == obj_universum)
		hook_call_obj_unlink(obj, old_father, old_prev);
	
	if(old_father == obj_zombie)
		obj_ref(obj);
	else if(father == obj_zombie)
		obj_unref(obj);
}

struct obj* obj_link(struct obj* obj, struct obj* father)
{
	ASSERT_META_TRANS;
	ASSERT(obj_get_root(father) == obj_universum || obj_get_root(father) == obj_temp);
	DANGEROUS((father->type == OT_UNIVERSUM && obj->type == OT_DOCUMENT) || (father->type == OT_DOCUMENT && obj->type == OT_TLO),
		"Dangerous link object %s at %p type %s to father %s at %p type %s.\n",
		obj_lookup_name(obj), obj, string_obj_type[obj->type]->text,
		obj_lookup_name(father), father, string_obj_type[father->type]->text);

	TRANS_NEW_UNDO(TRANS_OBJ_LINK, obj_link, obj_ref(obj), obj_ref(father));
	obj_relink(obj, father, obj_temp);
	obj_changed(obj);
	return obj;
}

struct obj* obj_unlink(struct obj* obj)
{
	ASSERT_META_TRANS;
	ASSERT(obj_get_root(void2obj(obj)) == obj_universum && obj != obj_universum);
	ASSERT(obj->type != OT_TLO || !tlo_is_in_trans_stack(obj2tlo(obj)));

	struct obj* walk;
	WALK_OBJ_RECURSIVE(obj, walk)
		obj_unselect(walk);

	TRANS_NEW_UNDO(TRANS_OBJ_UNLINK, obj_link, obj_ref(obj), obj_ref(obj->father));
	obj_relink(obj, obj_temp, obj->father);
	obj_changed(obj);
	return obj;
}

void obj_unlink_selected(void)
{
	ASSERT_META_TRANS;
	struct obj *obj;
	while((obj = obj_get_first_selected()))
		obj_unlink(obj);
}


/*
 * Selection
 */

void obj_select_low(struct obj* obj)
{
	ASSERT(obj_can_work(obj) && !obj_is_selected(obj));
	obj->flags |= OF_SELECTED;
	clist_add_tail(&obj_lsel, &obj->nsel);
	obj_sel_count[obj->type]++;
	hook_call_obj_selected(obj);
	obj_changed(obj);
}

void obj_select(struct obj* obj)
{
	ASSERT(obj_can_work(obj));
	if(obj_is_selected(obj))
		return;
	obj_select_low(obj);
	TRANS_NEW_UNDO(TRANS_OBJ_SELECT, obj_select, obj_ref(obj));
	obj_changed(obj);
}

void o_select(struct o* o)
{
	if(o->kind == T_OBJ)
		obj_select(o2obj(o));
	else
		go_select(o2go(o));
}

void obj_unselect_low(struct obj* obj)
{
	ASSERT(obj_is_selected(obj));
	obj->flags &= ~OF_SELECTED;
	clist_remove(&obj->nsel);
	obj_sel_count[obj->type]--;
	hook_call_obj_unselected(obj);
}

void obj_unselect(struct obj* obj)
{
	ASSERT(obj_can_work(obj));
	if(!obj_is_selected(obj))
		return;
	obj_unselect_low(obj);
	TRANS_NEW_UNDO(TRANS_OBJ_UNSELECT, obj_select, obj_ref(obj));
}

void obj_unselect_all(void)
{
	ASSERT_META_TRANS;
	struct obj* walk;
	while((walk = obj_get_first_selected()))
		obj_unselect(walk);
}

void o_unselect(struct o* o)
{
	if(o->kind == T_OBJ)
		obj_unselect(o2obj(o));
	else
		go_unselect(o2go(o));
}

void o_unselect_all(void)
{
	struct obj* volatile obj;
	char err[TRANS_ERR_SIZE];

	TRANS_BEGIN_MAIN(tlo_universum, "Unselect all obj", err)
		obj_unselect_all();
	TRANS_FAILED
		die("Unselect cannot fail: %s\n", err);
	TRANS_END

	WALK_OBJ_RECURSIVE(obj_universum, obj)
		if(obj->type == OT_TLO)
		{
			TRANS_BEGIN_MAIN(obj2tlo(obj), "Unselect all go", err)
				tlo_unselect_all_go();
			TRANS_FAILED
				die("Unselect cannot fail: %s\n", err);
			TRANS_END
		}
}


/*
 * Movement
 */

struct obj* obj_recursive_next(const struct obj* root, struct obj* walk)
{
	ASSERT(root->kind == T_OBJ && walk->kind == T_OBJ);
	struct obj* o;
	if((o = obj_get_first_son(walk)))
		return o;
	else
		for(;;)
		{
			if(walk == root)
				return NULL;
			if((o = obj_get_next_son(walk)))
				return o;
			walk = walk->father;
		}
}

struct o* o_get_ancestor(struct o* o)
{
	if(o->kind == T_OBJ)
		return (struct o*)o2obj(o)->father;
	struct go* go = &go_get_group(o2go(o))->go;
	return go ? go2o(go) : void2o(go_get_tlo(o2go(o)));
}


/*
 * Miscellaneous
 */

static inline struct undo_kernel* doc_is_changed_by_undo(struct obj_doc *doc, struct undo_gui *ug, uns type)
{
	struct undo_kernel *uk, *u;
	WALK_UNDO_KERNEL(ug, uk)
		switch(uk->type)
		{
			case TRANS_OBJ_SELECT:
			case TRANS_OBJ_UNSELECT:
				if(obj_is_ancestor(&doc->obj, uk->u.obj_select.obj))
					goto FIND;
				break;
				
			case TRANS_OBJ_LINK:
			case TRANS_OBJ_UNLINK:
				if(obj_is_ancestor(&doc->obj, uk->u.obj_link.obj))
					goto FIND;
				break;
				
			case TRANS_PROP_CREATE:
			case TRANS_PROP_CHANGE:
			case TRANS_PROP_DELETE:
			case TRANS_PROP_VIRTUAL:
				if(obj_is_ancestor(&doc->obj, o2obj(uk->u.prop.o)))
					goto FIND;
				break;

			case TRANS_DOC_SAVED_START:
			case TRANS_DOC_SAVED_END:
				break;

			default:
				BUG();
		}
	return NULL;

FIND:
	u = fmalloc(tlo_universum->fuk);
	u->type = type;
	u->u.doc_save.doc = doc;
	clist_insert_after(&u->nk, &uk->nk);
	return u;
}

void doc_set_saved(struct obj_doc *doc)
{
	struct undo_gui *ug;
	
	if(doc->saved)
		return;
	if(doc->saved_start)
		trans_command_kernel_delete(doc->saved_start);
	if(doc->saved_end)
		trans_command_kernel_delete(doc->saved_end);
	doc->saved = 1;

	for(ug = trans_get_next_gui(tlo_universum, tlo_universum->undo_present); ug; ug = trans_get_next_gui(tlo_universum, ug))
		if((doc->saved_end = doc_is_changed_by_undo(doc, ug, TRANS_DOC_SAVED_END)))
			break;
	for(ug = tlo_universum->undo_present; ug; ug = trans_get_prev_gui(tlo_universum, ug))
		if((doc->saved_start = doc_is_changed_by_undo(doc, ug, TRANS_DOC_SAVED_START)))
			break;
}

void obj_changed(struct obj *obj)
{
	for(; obj; obj = obj->father)
		if(obj->type == OT_DOCUMENT)
		{
			struct obj_doc *doc = obj2doc(obj);
			if(doc->saved)
			{
				ASSERT(!doc->saved_end);
				TRANS_NEW_UNDO(TRANS_DOC_SAVED_END, doc_save, doc);
				doc->saved_end = trans_get_last_kernel(trans_get_last_gui(tlo_universum));
				doc->saved = 0;
			}
		}
}

uns doc_is_all_saved(struct obj_doc *doc)
{
	struct obj_tlo *tlo;
	if(!doc_is_saved(doc))
		return 0;
	WALK_DOC_TLO(doc, tlo)
		if(!tlo_is_saved(tlo))
			return 0;
	return 1;
}
		
void doc_set_all_saved(struct obj_doc *doc)
{
	doc_set_saved(doc);
	struct obj_tlo *tlo;
	WALK_DOC_TLO(doc, tlo)
		tlo_set_saved(tlo);
}



/*
 * Referencing
 */

#undef obj_ref
struct obj* obj_ref(struct obj* obj DEBUG_REF_O_ARGS)
{
	ASSERT_OBJ(obj);
	ASSERT(obj->ref_count > 0);
	obj->ref_count++;
	DBG_O("  Increase reference: obj %s at %p, new ref_count %d, file %s, line %d, fce %s\n", obj_lookup_name(obj), obj, obj->ref_count, file, line, fce);
	return obj;
}

#undef obj_unref
void obj_unref(struct obj* obj DEBUG_REF_O_ARGS)
{
	ASSERT_OBJ(obj);
	ASSERT(obj->ref_count > 0);
	obj->ref_count--;
	DBG_O("  Decrease reference: obj %s at %p, new ref_count %d, file %s, line %d, fce %s\n", obj_lookup_name(obj), obj, obj->ref_count, file, line, fce);

	if(obj->ref_count == 0)
		obj_finish(obj);
}

#undef o_ref
struct o* o_ref(struct o *o DEBUG_REF_O_ARGS)
{
	ASSERT_O(o);
	ASSERT(o->ref_count > 0);
	o->ref_count++;
	if(o->kind == T_OBJ)
		DBG_O("  Increase reference: obj %s at %p, new ref_count %d, file %s, line %d, fce %s\n", obj_lookup_name(o2obj(o)), o, o->ref_count, file, line, fce);
	else
		DBG_O("  Increase reference: go %s at %p, new ref_count %d, file %s, line %d, fce %s\n", go_lookup_name(o2go(o)), o, o->ref_count, file, line, fce);
	return o;
}

#undef o_unref
void o_unref(struct o *o DEBUG_REF_O_ARGS)
{
	ASSERT_O(o);
	ASSERT(o->ref_count > 0);
	o->ref_count--;
	if(o->kind == T_OBJ)
		DBG_O("  Decrease reference: obj %s at %p, new ref_count %d, file %s, line %d, fce %s\n", obj_lookup_name(o2obj(o)), o, o->ref_count, file, line, fce);
	else
		DBG_O("  Decrease reference: go %s at %p, new ref_count %d, file %s, line %d, fce %s\n", go_lookup_name(o2go(o)), o, o->ref_count, file, line, fce);

	if(o->ref_count == 0)
	{
		if(o->kind == T_OBJ)
			obj_finish(o2obj(o));
		else
			go_finish(o2go(o));
	}
}
