//#define LOCAL_DEBUG
#include "kernel/internal.h"
#include "lib/hashfunc.h"


/* The last item is used for searching with holdback. */
struct prop_unit_struct unit_array[PROP_UNIT_MAX + 1] =
{
	{ NULL, 1, PQ_LENGTH, 1 },
	{ NULL, 1, PQ_ANGLE, 1 },
	{ NULL, 1, PQ_REFERENCE, 1 },
};

uns unit_default[PQ_MAX] =
{
	[PQ_NONE] = PROP_UNIT_MAX,
	[PQ_LENGTH] = 0,
	[PQ_ANGLE] = 1,
	[PQ_REFERENCE] = 2,
};

uns prop_subtype2quantity[PT_MAX][PROP_SUBTYPE_MAX] =
{
	[PT_UNS] = { PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, PQ_NONE, },
	[PT_POINTER] = { PQ_NONE, PQ_NONE, },
	[PT_STRING] = { PQ_NONE, PQ_NONE, PQ_NONE, },
	[PT_REAL] =
	{
		[PTR_UNSPECIFIED] = PQ_NONE,
		[PTR_COORDINATE] = PQ_LENGTH,
		[PTR_NON_NEGATIVE] = PQ_LENGTH,
		[PTR_ANGLE_PI] = PQ_ANGLE,
		[PTR_ANGLE_2PI] = PQ_ANGLE,
		[PTR_ANGLE_4PI] = PQ_ANGLE,
		[PTR_REFERENCE] = PQ_REFERENCE,
	},
};

static struct pn_o_bucket* pn_new;

struct prop_name
{
	struct o* o;
	string name;
};


/*
 * Hash table name -> o
 */

#define HASH_NODE struct prop_name
#define HASH_PREFIX(x) pn_name_##x
#define HASH_KEY_COMPLEX(x) x name
#define HASH_KEY_DECL string name
#define HASH_WANT_CLEANUP
#define HASH_WANT_NEW
#define HASH_WANT_FIND
#define HASH_WANT_REMOVE

#define HASH_GIVE_HASHFN
static uns pn_name_hash(string name)
	{ return hash_int((intptr_t)name); }

#define HASH_GIVE_EQ
static inline int pn_name_eq(string name1, string name2)
	{ return name1 == name2; }

#define HASH_GIVE_INIT_KEY
static inline void pn_name_init_key(struct prop_name* pn, string name)
	{ pn->name = name; }

#define HASH_GIVE_ALLOC
inline void* pn_name_alloc(uns size);
static inline void pn_name_free(void *p UNUSED) { }

#include "lib/sht.h"


/*
 * Hash table o -> name
 */

struct prop_o  // FIXME: ugly hack
{
	struct pn_name_bucket *next;
	uns hash;
	struct o* o;
	string name;
};

#define HASH_NODE struct prop_o
//#define HASH_NODE struct pn_name_bucket
//#define HASH_KEY_ATOMIC o
#define HASH_PREFIX(x) pn_o_##x
#define HASH_KEY_COMPLEX(x) x o
#define HASH_KEY_DECL struct o* o
#define HASH_WANT_CLEANUP
#define HASH_WANT_NEW
#define HASH_WANT_FIND
#define HASH_WANT_REMOVE

#define HASH_GIVE_HASHFN
static uns pn_o_hash(struct o* o)
	{ return hash_int((intptr_t)o); }

#define HASH_GIVE_EQ
static inline int pn_o_eq(struct o* o1, struct o* o2)
{
	ASSERT_O(o1);
	ASSERT_O(o2);
	return o1 == o2;
}

#define HASH_GIVE_INIT_KEY
static inline void pn_o_init_key(struct prop_o* po, struct o* o)
{
	ASSERT_O(o);
	po->o = o;
}

#define HASH_GIVE_ALLOC
static inline void* pn_o_alloc(uns size UNUSED)
	{ return pn_new; }
static inline void pn_o_free(void *p UNUSED) { }

#include "lib/sht.h"


/* request for initialization fixed-size allocator */
const uns prop_hash_name_size = sizeof(struct pn_o_bucket);

inline void* pn_name_alloc(uns size UNUSED)
	{ return &pn_new->n; }


/*
 * type-cast functions
 */
static inline struct prop_o* name2o(struct prop_name* pn)
	{ return SKIP_BACK(struct prop_o, o, &pn->o); }

static inline struct prop_name* o2name(struct prop_o* po)
	{ return SKIP_BACK(struct prop_name, o, &po->o); }

static inline struct obj_tlo* o_get_allocator(struct o* o UNUSED)
	{ ASSERT(trans_present); return trans_tlo; }


/*
 * Inicialization and clean-up functions
 */
void property_init(void)
{
	unit_array[0].name = string_lookup_ref("mm");
	unit_array[1].name = string_lookup_ref("radian");
	unit_array[2].name = string_lookup_ref("parameter");
	pn_name_init();
	pn_o_init();
}

void property_finish(void)
{
	DANGEROUS(pn_o_table.hash_count == 0 && pn_name_table.hash_count == 0, "Error in property hash table\n");
#ifdef DEBUG_LOCAL
	prop_hash_dump();
#endif
	pn_name_cleanup();
	pn_o_cleanup();

	for(uns i = 0; i < PROP_UNIT_MAX; i++)
		if(unit_array[i].name)
			string_unref(unit_array[i].name);
}

void prop_finish(struct o *o)
{
	ASSERT_KERNEL(o->prop.head.next);
	struct prop *p;
	while((p = prop_get_first(o)) && !(p->flags & PTF_VIRTUAL))
	{
		string_unref(p->key);
		if(p->type == PT_STRING)
			string_unref(p->value.s);
		slist_remove(&o->prop, &p->n);
		ffree(p);
	}
	struct prop_o* po = pn_o_find(o);
	if(po)
	{
		pn_o_remove(po);
		pn_name_remove(o2name(po));
		string_unref(po->name);
		ffree(SKIP_BACK(struct pn_o_bucket, n, po));
	}
#ifdef DEBUG_KERNEL
	o->prop.head.next = NULL;
#endif
}


/*
 * Create property
 */

void prop_create(struct o *o, string key, uns type, uns subtype, uns unit, prop_value value)
{
	ASSERT(type != PT_REAL || isfinite (value.r));
	struct prop* p = prop_find(o, key);
	struct undo_kernel *u;

	if(o->kind == T_OBJ)
		ASSERT_META_TRANS;
	else
		ASSERT_CORRECT_GO(o2go(o));

	if(p)
	{
		if(type != p->type || subtype != subtype)
			trans_fail("Cannot create property: different type or subtype");
		if(p->flags & PTF_VIRTUAL) 										// change a virtual property
		{
			if(prop_subtype2quantity[type][subtype] != PQ_NONE)
				prop2virt(p)->set_unit(o, unit);
			if(!prop_value_compare(type, value, prop2virt(p)->get(o)))
				prop2virt(p)->set(o, value);
			if(o->kind == T_OBJ)
				obj_changed(o2obj(o));
			return;
		}
		else																		// change a non-virtual property
		{
			p->unit = unit;
			if(prop_value_compare(type, p->value, value))
				return;
  			u = trans_new_undo(TRANS_PROP_CHANGE);
			u->u.prop.value = p->value;									// If type is PT_STRING then reference of value is moved from property to undo.
			p->value = value;
			hook_call_prop_change(o);
		}
	}
	else																			// create a new propety
	{
  		u = trans_new_undo(TRANS_PROP_CREATE);
		p = prop_set_low(o, key, type, subtype, PTF_SAVE, unit, value);
		string_ref(key);														// Reference for property
		p->flags |= PTF_UNDO_CREATE;
		hook_call_prop_create(o, p);
	}

	if(type == PT_STRING)
		string_ref(value.s);												// Reference for property

	u->u.prop.o = o_ref(o);						// Reference for undo
	u->u.prop.prop = p;
	if(o->kind == T_OBJ)
		obj_changed(o2obj(o));
}

void prop_load(struct o *o, const char* key, uns type, uns subtype, prop_value value)
{
	ASSERT(trans_present);
	ASSERT(type != PT_REAL || isfinite (value.r));
	string s = string_lookup(key);
	struct prop *p = prop_find(o, s);

	if(p)
	{
		if (type != p->type || subtype != p->subtype)
			trans_fail("Invalid type or subtype.");	// FIXME

		if (p->flags & PTF_VIRTUAL)
			prop2virt(p)->set(o, value);
	}
	else
	{
		p = prop_set_low(o, s, type, subtype, PTF_SAVE, PROP_UNIT_MAX, value);
		if(type == PT_STRING)
			string_ref(value.s);	// Reference for property
		string_ref(s);			// Reference for property
	}
}


/*
 * Properties searching.
 */

struct prop* prop_find(struct o *o, string key)
{
	struct prop* p;
	WALK_PROP(o, p)
		if(p->key == key)
			return p;
	return NULL;
}

prop_value prop_lookup(struct o *o, string key, uns type, prop_value err)
{
	struct prop *p = prop_find(o, key);
	if(!p || p->type != type)
		return err;
	return prop_get_value(o, p);
}


prop_value prop_named_lookup(struct o *o, const char* key, uns type, prop_value err)
{
	string s = string_find(key);
	return s ? prop_lookup(o, s, type, err) : err;
}

prop_value prop_get_value(struct o *o, struct prop *p)
{
	return (p->flags & PTF_VIRTUAL) ? prop2virt(p)->get(o) : p->value;
}

void prop_set_text(struct o *o, string key, const char* value)
{
	prop_set(o, key, (prop_value){.s=string_lookup(value)});
}

const char* prop_lookup_text(struct o *o, string key)
{
	string ret = prop_lookup(o, key, PT_STRING, (prop_value){.s=NULL}).s;
	return ret ? ret->text : NULL;
}

real prop_lookup_real_unit(struct o *o, string key, real err)
{
	struct prop *p = prop_find(o, key);

	if(!p || p->type != PT_REAL || prop_subtype2quantity[p->type][p->subtype] == PQ_NONE)
		return err;

	real value = prop_get_value(o, p).r / prop_get_unit_multiplier(o, p);
	if(!isfinite(value))
		return err;  // FIXME: What may be returned now?

	return value;
}

real prop_named_lookup_real_unit(struct o *o, const char* key, real err)
{
	string s = string_find(key);
	return s ? prop_lookup_real_unit(o, s, err) : err;
}


/*
 * Change value of a property.
 */

struct prop* prop_set_low(struct o *o, string key, uns type, uns subtype, uns flags, uns unit, prop_value value)
{
	ASSERT(!prop_find(o, key) && !(flags & PTF_VIRTUAL));
	struct prop* p = fmalloc(o_get_allocator(o)->fprop);
	p->key = key;
	p->type = type;
	p->subtype = subtype;
	p->flags = flags;
	p->unit = unit;
	p->value = value;
	slist_add(&o->prop, &p->n);
	return p;
}

void prop_set(struct o *o, string key, prop_value value)
{
	struct prop* p = prop_find(o, key);
	struct undo_kernel *u;

	if(!p)
		trans_fail("Property %s does not exist.", key->text);

	ASSERT(p->type != PT_REAL || isfinite (value.r));
	if(o->kind == T_OBJ)
		ASSERT_META_TRANS;
	else
		ASSERT_CORRECT_GO(o2go(o));

	if(p->flags & PTF_VIRTUAL) 		// change a virtual property
	{
		if(!prop_value_compare(p->type, value, prop2virt(p)->get(o)))
			prop2virt(p)->set(o, value);
	}
	else					// change a non-virtual property
	{
		if(prop_value_compare(p->type, p->value, value))
			return;
  		u = trans_new_undo(TRANS_PROP_CHANGE);
		u->u.prop.value = p->value;		// If type is PT_STRING then reference of value is moved from property to undo.
		p->value = value;
		if(p->type == PT_STRING)
			string_ref(value.s);		// Reference for property
		u->u.prop.o = o_ref(o);						// Reference for undo
		u->u.prop.prop = p;
		hook_call_prop_change(o);
	}
	if(o->kind == T_OBJ)
		obj_changed(o2obj(o));
}

void prop_set_real_unit(struct o *o, string key, real value)
{
	struct prop *p = prop_find(o, key);

	if(!p)
		trans_fail("Property %s does not exist.", key->text);
	if(p->type != PT_REAL || prop_subtype2quantity[p->type][p->subtype] == PQ_NONE)
		trans_fail("Invalid type of a property.");

	value *= prop_get_unit_multiplier(o, p);
	if(!isfinite(value))
		trans_fail("Numeric error occured.");

	prop_set(o, key, (prop_value){.r = value});
}

void prop_named_set_real_unit(struct o *o, const char* key, real value)
{
	string s = string_find(key);
	if(!s)
		trans_fail("Property %s does not exist.", key);

	prop_set_real_unit(o, s, value);
}


/*
 * Properties deletion.
 */

void prop_delete(struct o *o, struct prop* p)
{
	ASSERT(!(p->flags & PTF_VIRTUAL));

	if(o->kind == T_OBJ)
		ASSERT_META_TRANS;
	else
		ASSERT_CORRECT_GO(o2go(o));

	slist_remove(&o->prop, &p->n);
	p->n.next = NULL;
	hook_call_prop_delete(o, p);

	struct undo_kernel *u = trans_new_undo(TRANS_PROP_DELETE);
	u->u.prop.o = o_ref(o);							// Reference for undo
	u->u.prop.prop = p;
	p->flags |= PTF_UNDO_DELETE;	
	if(o->kind == T_OBJ)
		obj_changed(o2obj(o));
}

uns prop_find_delete(struct o *o, string key)
{
	struct prop* p = prop_find(o, key);
	if(!p || (p->flags & PTF_VIRTUAL))
		return 0;

	prop_delete(o, p);
	return 1;
}


/*
 * Property unit functions
 */

uns unit_find(string name)
{
	uns i;
	unit_array[PROP_UNIT_MAX].name = name;
	for(i = 0; unit_array[i].name != name; i++);
	return i;
}

uns unit_named_find (const char *name)
{
	string s = string_find(name);
	return s ? unit_find(s) : PROP_UNIT_MAX;
}

uns prop_get_unit(struct o *o, struct prop *prop)
{
	if((prop->type != PT_REAL) || (prop_subtype2quantity[prop->type][prop->subtype] == PQ_NONE))
		return PROP_UNIT_MAX;
	uns u = (prop->flags & PTF_VIRTUAL) ? prop2virt(prop)->get_unit(o) : prop->unit;
	if(u == PROP_UNIT_MAX)
		u = unit_default[prop_subtype2quantity[prop->type][prop->subtype]];
	return u;
}

real prop_get_unit_multiplier(struct o *o, struct prop *prop)
{
	uns u = prop_get_unit(o, prop);
	return u == PROP_UNIT_MAX ? 1 : unit_array[u].multiplier;
}

void prop_set_unit(struct o *o, struct prop *prop, uns unit)
{
	ASSERT(prop->type == PT_REAL && prop_subtype2quantity[prop->type][prop->subtype] != PQ_NONE && unit <= PROP_UNIT_MAX);
	if(prop->flags & PTF_VIRTUAL)
		prop2virt(prop)->set_unit(o, unit);
	else
		prop->unit = unit;
}

uns unit_set(string name, real mult, uns quantity)
{
	ASSERT(name && trans_present && quantity != PQ_NONE && quantity < PQ_MAX);
	if(mult < EQ_EPSILON || !isfinite(mult))
		trans_fail("Too small or invalid multiplier.");
	int changed = 1;

	uns i = unit_find(name);
	if(i == PROP_UNIT_MAX)
	{
		for(i = 0; i < PROP_UNIT_MAX && unit_array[i].used; i++);
		if(i == PROP_UNIT_MAX)
			trans_fail("Too many units.");
		changed = 0;
	}

	if(unit_array[i].name)
		string_unref(unit_array[i].name);
	unit_array[i].name = string_ref(name);
	unit_array[i].multiplier = mult;
	unit_array[i].quantity = quantity;
	unit_array[i].used = 1;

	if(changed)
		hook_call_unit_add(i);
	else
		hook_call_unit_changed(i);
	
	return i;
}

void unit_unused(uns index)
{
	ASSERT(index < PROP_UNIT_MAX && trans_present);
	if(unit_default[unit_array[index].quantity] != index)
		trans_fail("This unit is used as default unit.");
  	unit_array[index].used = 0;
	hook_call_unit_unused(index);
}

void unit_default_set(uns quantity, uns index)
{
	ASSERT(quantity != PQ_NONE && quantity < PQ_MAX && index < PROP_UNIT_MAX && trans_present);
	if(!unit_array[index].used)
		trans_fail("This unit isn't used.");
	unit_default[quantity] = index;
	hook_call_unit_default(quantity);
}


/*
 * Property name functions
 */

void prop_name_set_low(struct o* o, string name, struct obj_tlo *alloc)
{
	ASSERT(name != string_noname && !pn_o_find(o));
	pn_new = fmalloc(alloc->fname);
	pn_o_new(o);
	pn_name_new(name);
	pn_new = NULL;
}

prop_value o_get_prop_name(struct o *o)
{
	ASSERT_O(o);
	struct prop_o * po = pn_o_find(o);
	return (prop_value){.s = po ? po->name : string_noname};
}

static void o_set_prop_name_low(struct o *o, prop_value name)
{
	DBG("o_set_prop_name_low, o %p, name %s\n", o, name.s->text);
	struct prop_o *po = pn_o_find(o);

	if(name.s == string_noname)
	{
		ASSERT(po);		// This function is not called if old and new name is same.
		pn_o_remove(po);
		pn_name_remove(o2name(po));
		string_unref(po->name);
		ffree(SKIP_BACK(struct pn_o_bucket, n, po));
		return;
	}
	string_ref(name.s);
	if(po)
	{
		ASSERT(po->name != name.s);
		string_unref(po->name);
		pn_name_remove(o2name(po));
		pn_new = SKIP_BACK(struct pn_o_bucket, n, po);
	}
	else
	{
		pn_new = fmalloc(o_get_allocator(o)->fname);
		pn_o_new(o);
	}
	pn_name_new(name.s);
	pn_new->n.name = name.s;
	pn_new = NULL;
}

static void o_set_prop_name(struct o *o, prop_value name)
{
	trans_new_undo_prop_virtual(o, &o_prop_name, o_get_prop_name(o));
	o_set_prop_name_low(o, name);
	hook_call_prop_change(o);
}

struct prop_virtual o_prop_name = {
	{ { NULL }, NULL, {0}, PT_STRING, PTS_UNSPECIFIED, PTF_VIRTUAL | PTF_SAVE, 0 },
	UINT_MAX, o_get_prop_name, o_set_prop_name, o_set_prop_name_low,
	prop_unit_get_bug, prop_unit_set_bug
};

struct o* o_lookup(string name)
{
	struct prop_name *pn = pn_name_find(name);
	if(pn)
		return pn->o;

	DBG("o_lookup: %s not found\n", name->text);
	return NULL;
}

struct o* o_named_lookup(const char* name)
{
	string str = string_find(name);
	if(str)
		return o_lookup(str);

	DBG("o_named_lookup: %s not found\n", name);
	return NULL;
}

void o_set_name(struct o* o, const char* str)
{
	prop_set(o, string_name, (prop_value){ .s = string_lookup(str) });
}


/*
 * Miscellaneous functions for properties
 */

uns prop_value_compare(uns prop_type, prop_value v1, prop_value v2)
{
	switch(prop_type)
	{
		case PT_UNS:
			return v1.u == v2.u;
		case PT_POINTER:
			return v1.v == v2.v;
		case PT_STRING:
			return v1.s == v2.s;
		case PT_REAL:
			return v1.r == v2.r;
		default:
			die("Invalid property type.");
	}
}


/*
 * Virtual property functions.
 */

void prop_virtual_init(uns start, ...)
{
	struct prop_virtual *v;
	va_list ap;
	va_start(ap, start);
	while((v = va_arg(ap, struct prop_virtual*)))
	{
		ASSERT(v->p.key == NULL);
		v->p.key = string_lookup_ref(va_arg(ap, char*));
	}
	va_end(ap);
}

void prop_virtual_finish(uns start, ...)
{
	struct prop_virtual *v;
	va_list ap;
	va_start(ap, start);
	for(; (v = va_arg(ap, struct prop_virtual*)); va_arg(ap, char*))
		string_unref(v->p.key);
	va_end(ap);
}


/*
 * Other functions
 */

static UNUSED void prop_hash_dump(void)
{
	msg("Property name hash table\n");
	HASH_FOR_ALL(pn_name, walk)
		msg("\to %p, name %s\n", walk->o, walk->name->text);
		ASSERT_O(walk->o);
	HASH_END_FOR;
	msg("Property o hash table\n");
	HASH_FOR_ALL(pn_o, walk)
		msg("\to %p, name %s\n", walk->o, walk->name->text);
		ASSERT_O(walk->o);
	HASH_END_FOR;
}

