#ifndef _KERNEL_PROPERTY_H_
#define _KERNEL_PROPERTY_H_


/* TODO: implement properties somehow better */
typedef union prop_value
{
	string s;
	uns u;
	void* v;
	real r;
	struct geom_transform* t;
} prop_value;

struct prop
{
	struct snode n;
	string key;
	prop_value value;
	u8 type, subtype, flags, unit;
};

struct prop_unit_struct
{
	string name;
	real multiplier;
	uns quantity;
	uns used;
};


enum prop_type
{
SCONST	PT_UNS,
SCONST	PT_POINTER,
SCONST	PT_STRING,
SCONST	PT_REAL,
SCONST	PT_MAX,
};

enum prop_subtype_uns
{
SCONST	PTU_UNSPECIFIED,
SCONST	PTU_COLOR,
SCONST	PTU_FONT,
// allowed values are 0 .. GEOM_CONIC_ARC_iP for PTU_CONIC_TYPE_iP
SCONST	PTU_CONIC_TYPE_0P,
SCONST	PTU_CONIC_TYPE_1P,
SCONST	PTU_CONIC_TYPE_2P,
SCONST	PTU_CONIC_TYPE_3P,
SCONST	PTU_BOOLEAN,
SCONST	PTU_CAP_STYLE,
SCONST	PTU_MAX,
};

enum prop_subtype_pointer
{
SCONST	PTP_UNSPECIFIED,
SCONST	PTP_TRANSFORM,
	PTP_TEX_PROCESS,	// internal
SCONST	PTP_MAX,
};

enum prop_subtype_string
{
SCONST	PTS_UNSPECIFIED,
SCONST	PTS_FILE_NAME,
SCONST	PTS_LARGE_TEXT,
SCONST	PTS_MAX,
};

enum prop_subtype_real
{
SCONST	PTR_UNSPECIFIED,
SCONST	PTR_COORDINATE,
SCONST	PTR_NON_NEGATIVE,
SCONST	PTR_ANGLE_PI,
SCONST	PTR_ANGLE_2PI,
SCONST	PTR_ANGLE_4PI,
SCONST	PTR_REFERENCE,
SCONST	PTR_MAX,
};

enum prop_flags
{
SCONST	PTF_VIRTUAL = 0x1,
SCONST	PTF_READ_ONLY = 0x2,
SCONST	PTF_SAVE = 0x4,
	PTF_UNDO_CREATE = 0x40,	// An internal flag.
	PTF_UNDO_DELETE = 0x80,	// An internal flag.
};

enum prop_quantity
{
SCONST	PQ_NONE,
SCONST	PQ_LENGTH,
SCONST	PQ_ANGLE,
SCONST	PQ_REFERENCE,
SCONST	PQ_MAX,
};

enum prop_cap_style
{
SCONST	PSC_BUTT,
SCONST	PSC_ROUND,
SCONST	PSC_PROJECTING,
SCONST	PSC_MAX,
};

#define PROP_SUBTYPE_MAX MAX( MAX(PTU_MAX, PTP_MAX), MAX(PTR_MAX, PTS_MAX) )

#define PROP_UNIT_MAX 255

extern struct prop_unit_struct unit_array[];
extern uns prop_subtype2quantity[][PROP_SUBTYPE_MAX];
extern uns unit_default[];


/*
 * Create a new property.
 */

void prop_create(struct o *o, string key, uns type, uns subtype, uns unit, prop_value value);
void prop_load(struct o *o, const char* key, uns type, uns subtype, prop_value value);


/*
 * Properties searching.
 * Each property is uniquely determined by its name
 */

struct prop* prop_find(struct o *o, string key);
prop_value prop_lookup(struct o *o, string key, uns type, prop_value err);
prop_value prop_named_lookup(struct o *o, const char* key, uns type, prop_value err);

prop_value prop_get_value(struct o *o, struct prop *p);

static inline uns prop_lookup_uns(struct o *o, string key, uns err)
	{ return prop_lookup(o, key, PT_UNS, (prop_value){.u=err}).u; }
static inline real prop_lookup_real(struct o *o, string key, real err)
	{ return prop_lookup(o, key, PT_REAL, (prop_value){.r=err}).r; }
const char* prop_lookup_text(struct o *o, string key);
real prop_lookup_real_unit(struct o *o, string key, real err);
real prop_named_lookup_real_unit(struct o *o, const char* key, real err);


/*
 * Change value of a property which allready exist.
 */

void prop_set(struct o *o, string key, prop_value value);

static inline void prop_set_uns(struct o *o, string key, uns value)
	{ prop_set(o, key, (prop_value){.u=value}); }
static inline void prop_set_real(struct o *o, string key, real value)
	{ prop_set(o, key, (prop_value){.r=value}); }
void prop_set_text(struct o *o, string key, const char* value);
void prop_set_real_unit(struct o *o, string key, real value);
void prop_named_set_real_unit(struct o *o, const char* key, real value);


/*
 * Properties deletion
 * Property cannot be virtual.
 */

void prop_delete(struct o *o, struct prop* prop);
uns prop_find_delete(struct o *o, string key);


/*
 * Movement PROP function
 */

static inline struct prop* prop_get_first(struct o *o)
	{ struct snode* n = o->prop.head.next; return n ? SKIP_BACK(struct prop, n, n) : NULL; }
static inline struct prop* prop_get_next(struct prop* prop)
	{ struct snode* n = prop->n.next; return n ? SKIP_BACK(struct prop, n, n) : NULL; }
#define WALK_PROP(_o, _walk) for(_walk = prop_get_first(_o); _walk; _walk = prop_get_next(_walk))
static inline uns prop_get_count(struct o *o)
	{ uns n = 0; struct prop* p; WALK_PROP(o, p) n++; return n; }


/*
 * Property unit functions
 */

uns unit_find(string name);
uns unit_named_find(const char *name);

uns prop_get_unit(struct o *o, struct prop* prop);
real prop_get_unit_multiplier(struct o *o, struct prop *prop);
void prop_set_unit(struct o *o, struct prop* prop, uns unit);

uns unit_set(string name, real mult, uns quantity);
void unit_default_set(uns quantity, uns index);
void unit_unused(uns index);


/*
 * Property name functions
 */

prop_value o_get_prop_name(struct o *o);
static inline string o_get_name(struct o *o)
	{ return o_get_prop_name(o).s; }
static inline const char* o_lookup_name(struct o *o)
	{ return o_get_prop_name(o).s->text; }
static inline string obj_get_name(struct obj *obj)
	{ return o_get_prop_name((struct o*)obj).s; }
static inline const char* obj_lookup_name(struct obj *obj)
	{ return o_get_prop_name((struct o*)obj).s->text; }
static inline string go_get_name(struct go *go)
	{ return o_get_prop_name((struct o*)go).s; }
static inline const char* go_lookup_name(struct go *go)
	{ return o_get_prop_name((struct o*)go).s->text; }

void o_set_name(struct o* o, const char* str);

struct o* o_lookup(string name);
SCHEME SO_o o_named_lookup(SSTRINGC name);


/*
 * Functions for working with color
 */

SCHEME static inline SUNS rgba_uns2uns(SUNS r, SUNS g, SUNS b, SUNS a)
	{ return (r << 24) | (g << 16) | (b << 8) | a; }
static inline void uns2rgba_uns(uns u, uns *r, uns *g, uns *b, uns *a)
	{ *r = (u>>24)&255; *g = (u>>16)&255; *b = (u>>8)&255; *a = u&255; }

static inline uns rgba_real2uns(real r, real g, real b, real a)
	{ ASSERT(r>=0 && r<=1 && g>=0 && g<=1 && b>=0 && b<=1 && a>=0 && a<=1); return (((uns)(r*255)) << 24) | (((uns)(g*255)) << 16) | (((uns)(b*255)) << 8) | ((uns)(a*255)); }
static inline void uns2rgba_real(uns u, real *r, real *g, real *b, real *a)
	{ *r = ((real)((u>>24)&255))/255; *g = ((real)((u>>16)&255))/255; *b = ((real)((u>>8)&255))/255; *a = ((real)(u&255))/255; }

SCHEME static inline SUNS uns2r_uns(SUNS u)
        { return ((u>>24)&255); }
SCHEME static inline SUNS uns2g_uns(SUNS u)
        { return ((u>>16)&255); }
SCHEME static inline SUNS uns2b_uns(SUNS u)
        { return ((u>>8)&255); }
SCHEME static inline SUNS uns2a_uns(SUNS u)
        { return (u&255); }

SCHEME static inline SUNS rgba_uns_set_r_uns(SUNS c, SUNS r) { return ((c & 0x00FFFFFF) | (r << 24)); }
SCHEME static inline SUNS rgba_uns_set_g_uns(SUNS c, SUNS g) { return ((c & 0xFF00FFFF) | (g << 16)); }
SCHEME static inline SUNS rgba_uns_set_b_uns(SUNS c, SUNS b) { return ((c & 0xFFFF00FF) | (b <<  8)); }
SCHEME static inline SUNS rgba_uns_set_a_uns(SUNS c, SUNS a) { return ((c & 0xFFFFFF00) | a); }

/*
 * Properties hooks
 */

struct unit_hook
{
	snode n;
	void* data;
	void (*add) (uns slot, struct unit_hook *h);		// New unit was added into slot.
	void (*unused) (uns slot, struct unit_hook *h);			// Unit will not be used more.
	void (*changed) (uns slot, struct unit_hook *h);		// Value or quantity of unit was changed.
	void (*def) (uns quantity, struct unit_hook *h);		// Default unit of quantity was changed.
};

struct unit_hook* unit_hook_add(struct unit_hook *hook);
struct unit_hook* unit_hook_remove(struct unit_hook *hook);
struct unit_hook* unit_hook_remove_by_data(void *data);

/*
 * Miscellaneous functions for properties
 */

uns prop_value_compare(uns prop_type, prop_value v1, prop_value v2);

#define PROP_INIT_FILL_COLOR 0	// black: red 0, green 0, blue 0, alpha 0
#define PROP_INIT_STROKE_COLOR 255	// black: red 0, green 0, blue 0, alpha 255
#define PROP_INIT_STROKE_CAP PSC_BUTT
#define PROP_INIT_STROKE_WIDTH 0


/*
 * Dump property.
 */

#define PROP_DUMP_SIZE 256
char* prop_dump_value(char buf[PROP_DUMP_SIZE], struct prop* prop, prop_value value);
char* prop_dump(char buf[PROP_DUMP_SIZE], struct o *o, struct prop* prop);


#endif
