
(define (call-with-error-pack thunk)
  (catch #t 
	 (lambda () (cons #t (thunk)))
	 (lambda args (cons #f args))))

(define (output . items)
  (for-each display items)
  (newline))

(define (run-test-line . items)
  (let run-test-case ((fn (car items))
		      (name (cadr items))
		      (args '())
		      (rest (cddr items)))
    (let ((result (call-with-error-pack (lambda () (apply fn args)))))
      (if (car result)
	  (begin
	    (output "success: " name)
	    (if (not (null? rest))
		(run-test-case (car rest) (cadr rest) (cdr result) (cddr rest))))
	  (begin
	     (output "FAIL: " name " error: " (cdr result)))))))

(define (test-make-tlo)
  (trans-main obj-universum
	      (lambda ()
		(let ((doc (doc-new))
		      (tlo (tlo-new)))
		  (obj-link tlo doc)
		  (obj-link doc obj-universum)
		  (list doc tlo)))))

(define (test-make-point doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-point (coords 12 12))
		(list doc tlo))))

(define (test-make-segment doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-segment (coords 10 10) (coords 20 30))
		(list doc tlo))))

(define (test-make-quadratic doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-bezier/2 (coords 10 10) (coords -5 40) (coords 20 30))
		(list doc tlo))))


(define (test-make-cubic doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-bezier/3 (coords 10 10) (coords -5 40) (coords 40 0) (coords 20 30))
		(list doc tlo))))

(define (test-make-e/c2rr doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-ellipse/c2rr (coords -10 10) 50 20 0.2)
		(list doc tlo))))

(define (test-make-e/2fp doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-ellipse/2fp (coords -15 10) (coords 15 15) (coords 5 0))
		(list doc tlo))))

(define (test-make-e/3ps doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-ellipse/3ps (coords -15 10) (coords 15 15) (coords 5 0))
		(list doc tlo))))

(define (test-make-e/3pre doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-ellipse/3pre (coords -15 10) (coords 15 15) (coords 5 0) 0.7 0.6)
		(list doc tlo))))

(define (test-make-e/cpre doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-ellipse/cpre (coords -10 -5) (coords 15 20) 0.2 0.6)
		(list doc tlo))))

(define (test-make-text doc tlo)
  (trans-main tlo
	      (lambda ()
		(make-text (coords 25 25) "Hello Hello!" "/usr/X11R6/lib/X11/fonts/Type1/c0648bt_.pfb" 10)
		(list doc tlo))))

(define (test-make-tex doc tlo)
  (trans-main tlo
	      (lambda ()
		(make-tex-text (coords 30 30) "$\\alpha + 2^\\beta$")
		(list doc tlo))))

(define (test-make-decorator doc tlo)
  (trans-main tlo
	      (lambda ()
		(make-decorator-point (coords 17 13) 5 10 0.2)
		(list doc tlo))))

(define (test-make-deps-1 doc tlo)
  (trans-main tlo
	      (lambda ()
		(let ((s1 (make-segment (coords 5 5) (coords 15 15)))
		      (s2 (make-segment (coords 25 5) (coords 35 15))))
		  (make-segment (hget-endpoint-1 s1) (hget-endpoint-2 s2)))
		(list doc tlo))))

(define (test-make-deps-2 doc tlo)
  (trans-main tlo
	      (lambda ()
		(let ((s1 (make-segment (coords 5 25) (coords 15 35)))
		      (s2 (make-segment (coords 25 25) (coords 35 35))))
		  (make-segment (hget-center (make-parametric-point s1 0.3))
				(hget-center (make-parametric-point s2 0.6))))
		(list doc tlo))))

(define (test-make-deps-3 doc tlo)
  (trans-main tlo
	      (lambda ()
		(let ((s1 (make-segment (coords 0 0) (coords -50 -50)))
		      (s2 (make-segment (coords -50 0) (coords 0 -50))))
		  (make-segment (hget-center (make-intersection s1 s2 0.5))
				(coords -25 0)))
		(list doc tlo))))

(define (test-save-doc doc tlo)
  (save-doc doc "file.vrr")
  (list doc))

(define (test-load-doc)
  (list (load-doc "file.vrr")))

(define (test-unlink-doc doc)
  (trans-main obj-universum
	      (lambda ()
		(obj-unlink doc)))
  '())

(define (run-tests)  
(run-test-line test-make-tlo      "make-tlo"
	       test-make-point    "make-point"
	       test-save-doc      "save-point"
	       test-unlink-doc    "unlink-doc"
	       test-load-doc      "load-point")

(run-test-line test-make-tlo      "make-tlo"
	       test-make-segment  "make-segment"
	       test-save-doc      "save-segment"
	       test-unlink-doc    "unlink-doc"
	       test-load-doc      "load-segment")

(run-test-line test-make-tlo       "make-tlo"
	       test-make-quadratic "make-bezier-2"
	       test-save-doc       "save-bezier-2"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-bezier-2")

(run-test-line test-make-tlo       "make-tlo"
	       test-make-cubic     "make-bezier-3"
	       test-save-doc       "save-bezier-3"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-bezier-3")

(run-test-line test-make-tlo       "make-tlo"
	       test-make-e/c2rr    "make-ellipse-c2rr"
	       test-save-doc       "save-ellipse-c2rr"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-ellipse-c2rr")

(run-test-line test-make-tlo       "make-tlo"
	       test-make-e/2fp     "make-ellipse-2fp"
	       test-save-doc       "save-ellipse-2fp"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-ellipse-2fp")

(run-test-line test-make-tlo       "make-tlo"
	       test-make-e/3ps    "make-ellipse-3ps"
	       test-save-doc       "save-ellipse-3ps"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-ellipse-3ps")

(run-test-line test-make-tlo       "make-tlo"
	       test-make-e/3pre    "make-ellipse-3pre"
	       test-save-doc       "save-ellipse-3pre"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-ellipse-3pre")

(run-test-line test-make-tlo       "make-tlo"
	       test-make-e/cpre    "make-ellipse-cpre"
	       test-save-doc       "save-ellipse-cpre"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-ellipse-cpre")


(run-test-line test-make-tlo       "make-tlo"
	       test-make-text      "make-text"
	       test-save-doc       "save-text"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-text")


(run-test-line test-make-tlo       "make-tlo"
	       test-make-tex       "make-tex-text"
	       test-save-doc       "save-tex-text"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-tex-text")

(run-test-line test-make-tlo       "make-tlo"
	       test-make-decorator "make-decorator"
	       test-save-doc       "save-decorator"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-decorator")

(run-test-line test-make-tlo       "make-tlo"
	       test-make-deps-1    "make-dependent-segments-endpoints"
	       test-save-doc       "save-dependent-segments-endpoints"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-dependent-segments-endpoints")

(run-test-line test-make-tlo       "make-tlo"
	       test-make-deps-2    "make-dependent-segments-parametric-points"
	       test-save-doc       "save-dependent-segments-parametric-points"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-dependent-segments-parametric-points")

(run-test-line test-make-tlo       "make-tlo"
	       test-make-deps-3    "make-dependent-segments-intersections"
	       test-save-doc       "save-dependent-segments-intersections"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-dependent-segments-intersections"))
	       
(run-tests)	     

