#include "kernel/internal.h"
#include <stdio.h>
#include <string.h>

enum
{
	TT_MAIN,
	TT_SUB,
	TT_SUBMAIN,
	TT_ANONYMOUS,
	TT_INTERNAL,
};

#define TRANS_BEGIN_INTERNAL(_tlo, _es)				\
do {								\
		jmp_buf _es##jb;				\
		trans_begin_internal(_tlo, _es, &_es##jb);	\
		if(!setjmp(_es##jb))				\
		{	do	{

inline struct undo_gui* trans_get_first_gui(struct obj_tlo* t)
	{ return &t->undo_start; }
inline struct undo_gui* trans_get_next_gui(struct obj_tlo* t, struct undo_gui *ug)
	{ ug = SKIP_BACK(struct undo_gui, ng, ug->ng.next); return ug == &t->undo_start ? NULL : ug; }
inline struct undo_gui* trans_get_last_gui(struct obj_tlo* t)
	{ return SKIP_BACK(struct undo_gui, ng, t->undo_start.ng.prev); }
inline struct undo_gui* trans_get_prev_gui(struct obj_tlo* t, struct undo_gui *ug)
	{ return ug == &t->undo_start ? NULL : SKIP_BACK(struct undo_gui, ng, ug->ng.prev); }
inline uns trans_get_gui_rank(struct obj_tlo* t, struct undo_gui *ug)
	{ uns n = 0; while((ug = trans_get_prev_gui(t, ug))) n++; return n; }

struct trans* tlo_is_in_trans_stack(struct obj_tlo* tlo)
{
	struct trans *t;
	if(trans_tlo == tlo)
		return trans_present;
	WALK_TRANS_STACK(t)
		if(t->tlo == tlo)
			return t->prev;
	return NULL;
}

// Undo history cannot be changing when trans_fail or trans_undo or trans_redo is working.
int trans_is_undo_write_protected(void)
{
	for(struct trans* t = trans_present; t; t = t->prev)
		if(t->type == TT_INTERNAL)
			return 1;
	return 0;
}

// Used only in debugging macros.
static UNUSED uns trans_get_depth_of_stack(void)
{
	struct trans* t;
	uns u = 0;
	WALK_TRANS_STACK(t)
		u++;
	return u;
}

static inline void trans_fifo_update(struct go* go)
{
	if(!go->temp_data && go_get_tlo(go)->obj.type != OT_ZOMBIE)
		tsort_dirty(go);
}

static void trans_command_undo(struct undo_kernel* u, uns direction)
{
	ASSERT(direction == TRANS_UNDO || direction == TRANS_REDO);
	switch(u->type | direction)
	{
		case TRANS_OBJ_LINK | TRANS_UNDO:
		case TRANS_OBJ_UNLINK | TRANS_REDO:
			obj_relink(u->u.obj_link.obj, obj_zombie, u->u.obj_link.father);
			break;

		case TRANS_OBJ_UNLINK | TRANS_UNDO:
		case TRANS_OBJ_LINK | TRANS_REDO:
			obj_relink(u->u.obj_link.obj, u->u.obj_link.father, obj_zombie);
			break;

		case TRANS_OBJ_SELECT | TRANS_UNDO:
		case TRANS_OBJ_UNSELECT | TRANS_REDO:
			obj_unselect_low(u->u.obj_link.obj);
			break;

		case TRANS_OBJ_UNSELECT | TRANS_UNDO:
		case TRANS_OBJ_SELECT | TRANS_REDO:
			obj_select_low(u->u.obj_link.obj);
			break;

		case TRANS_GO_SELECT | TRANS_UNDO:
		case TRANS_GO_UNSELECT | TRANS_REDO:
			go_unselect_low(u->u.go_link.go);
			break;

		case TRANS_GO_UNSELECT | TRANS_UNDO:
		case TRANS_GO_SELECT | TRANS_REDO:
			go_select_low(u->u.go_link.go);
			break;

		case TRANS_GO_LINK | TRANS_UNDO:
		case TRANS_GO_UNLINK | TRANS_REDO:
			go_relink_low(u->u.go_link.go, &tlo_zombie->tlg->group, NULL);
			break;

		case TRANS_GO_UNLINK | TRANS_UNDO:
		case TRANS_GO_LINK | TRANS_REDO:
			go_relink_low(u->u.go_link.go, u->u.go_link.group, u->u.go_link.after);
			break;
			

		case TRANS_GO_RELINK | TRANS_UNDO:
		case TRANS_GO_RELINK | TRANS_REDO:
			go_relink_low(u->u.go_link.go, u->u.go_link.group, u->u.go_link.after);
			break;

		case TRANS_ANCHOR_REHANG | TRANS_UNDO:
		case TRANS_ANCHOR_REHANG | TRANS_REDO:
		{
			struct hanger* h = u->u.anchor_rehang.hanger;
			struct anchor* a = u->u.anchor_rehang.anchor;
			ASSERT(go_can_work(hanger_get_go(h)) && go_can_work(anchor_get_go(a)));

			clist_remove(&a->nanchor);
			u->u.anchor_rehang.hanger = a->h;
			a->h = h;		// Referencies of both hangers are exchanged between the anchor and undo.
			clist_add_tail(&h->lanchor, &a->nanchor);

			trans_fifo_update(anchor_get_go(a));

			break;
		}

		case TRANS_MOUSE_CLICK_MOVE | TRANS_UNDO:
		case TRANS_MOUSE_CLICK_MOVE | TRANS_REDO:
		{
			struct anchor *a;
			struct hanger_smart *mc = u->u.mc_move.mc;

			struct geom_point gp = mc->q.p;
			mc->q.p = u->u.mc_move.gp;
			u->u.mc_move.gp = gp;

			WALK_HANGER_ANCHOR(&mc->h, a)
				trans_fifo_update(anchor_get_go(a));

			break;
		}

		case TRANS_PROP_CREATE | TRANS_UNDO:
		case TRANS_PROP_DELETE | TRANS_REDO:
		{
			struct o *o = u->u.prop.o;
			struct prop *p = u->u.prop.prop;
			ASSERT(p && prop_find(o, p->key) == p && !(p->flags & PTF_VIRTUAL));

			slist_remove(&o->prop, &p->n);
			p->n.next = NULL;
			hook_call_prop_delete(o, p);
			break;
		}

		case TRANS_PROP_VIRTUAL_END_UNDO | TRANS_UNDO:
		case TRANS_PROP_VIRTUAL_START_UNDO | TRANS_REDO:
		{
			uns mask = prop2virt(u->u.prop.prop)->undo_mask;
			ASSERT(mask != 0 && mask != UINT_MAX && !(mask & u->u.prop.o->flags));
			u->u.prop.o->flags |= mask;	
			goto CHANGE_PROP;
		}
			
		case TRANS_PROP_VIRTUAL_START_UNDO | TRANS_UNDO:
		case TRANS_PROP_VIRTUAL_END_UNDO | TRANS_REDO:
		{
			uns mask = prop2virt(u->u.prop.prop)->undo_mask;
			ASSERT(mask != 0 && mask != UINT_MAX && (mask & u->u.prop.o->flags));
			u->u.prop.o->flags &= ~mask;	
			goto CHANGE_PROP;
		}
		
		case TRANS_PROP_VIRTUAL | TRANS_UNDO:
		case TRANS_PROP_VIRTUAL | TRANS_REDO:
CHANGE_PROP:
		{
			struct o *o = u->u.prop.o;
			struct prop* p = u->u.prop.prop;
			ASSERT(p && (p->flags & PTF_VIRTUAL));
			prop_value value = prop2virt(p)->get(o);

			if(p->type == PT_STRING)
			{
				string_ref(value.s);
				string_unref(u->u.prop.value.s);
			}
			prop2virt(p)->set_low(o, u->u.prop.value);
			u->u.prop.value = value;

			if(o->kind == T_GO)
				trans_fifo_update(o2go(o));
			break;
		}

		case TRANS_PROP_CHANGE | TRANS_UNDO:
		case TRANS_PROP_CHANGE | TRANS_REDO:
		{
			struct o *o = u->u.prop.o;
			struct prop* p = u->u.prop.prop;
			ASSERT(p && !(p->flags & PTF_VIRTUAL) && prop_find(o, p->key) == p && !prop_value_compare(p->type, p->value, u->u.prop.value));

			prop_value value = p->value;
			p->value = u->u.prop.value;
			u->u.prop.value = value;

			if(o->kind == T_GO)
				trans_fifo_update(o2go(o));
			break;
		}

		case TRANS_PROP_DELETE | TRANS_UNDO:
		case TRANS_PROP_CREATE | TRANS_REDO:
		{
			struct o *o = u->u.prop.o;
			struct prop* p = u->u.prop.prop;
			ASSERT(p && !prop_find(o, p->key) && !(p->flags & PTF_VIRTUAL));

			slist_add(&o->prop, &p->n);
			hook_call_prop_create(o, p);
			if(o->kind == T_GO)
				trans_fifo_update(o2go(o));
			break;
		}

		case TRANS_DOC_SAVED_START	| TRANS_UNDO:
		case TRANS_DOC_SAVED_END | TRANS_REDO:
			u->u.doc_save.doc->saved = 0;
			break;

		case TRANS_DOC_SAVED_END	| TRANS_UNDO:
		case TRANS_DOC_SAVED_START | TRANS_REDO:
			u->u.doc_save.doc->saved = 1;
			break;

		default:
			die("There is no label for trans %u.", u->type);
	}
}

void trans_command_kernel_delete(struct undo_kernel* uk)
{
	clist_remove(&uk->nk);

	switch(uk->type)
	{
		case TRANS_OBJ_LINK:
		case TRANS_OBJ_UNLINK:
			obj_unref(uk->u.obj_link.obj);
			obj_unref(uk->u.obj_link.father);
			break;

		case TRANS_OBJ_SELECT:
		case TRANS_OBJ_UNSELECT:
			obj_unref(uk->u.obj_link.obj);
			break;

		case TRANS_GO_SELECT:
		case TRANS_GO_UNSELECT:
			go_unref(uk->u.go_select.go);
			break;

		case TRANS_GO_LINK:
		case TRANS_GO_UNLINK:
		case TRANS_GO_RELINK:
			go_unref(uk->u.go_link.go);
			break;

		case TRANS_ANCHOR_REHANG:
			go_unref(anchor_get_go(uk->u.anchor_rehang.anchor));
			hanger_unref(uk->u.anchor_rehang.hanger);
			break;

		case TRANS_MOUSE_CLICK_MOVE:
			hanger_unref(&uk->u.mc_move.mc->h);
			break;

		case TRANS_PROP_CHANGE:
		case TRANS_PROP_VIRTUAL:
		case TRANS_PROP_VIRTUAL_START_UNDO:
		case TRANS_PROP_VIRTUAL_END_UNDO:
			if(uk->u.prop.prop->type == PT_POINTER)
				switch(uk->u.prop.prop->subtype)
				{
				case PTP_TRANSFORM: xfree(uk->u.prop.value.t); break;
				case PTP_TEX_PROCESS: tex_process_unref(uk->u.prop.value.v); break;
				}
			else if(uk->u.prop.prop->type == PT_STRING)
				string_unref(uk->u.prop.value.s);
			o_unref(uk->u.prop.o);
			break;

		case TRANS_PROP_CREATE:
		{
			struct prop* p = uk->u.prop.prop;
			ASSERT((p->flags & PTF_UNDO_CREATE) && !(p->flags & PTF_VIRTUAL));
			o_unref(uk->u.prop.o);
			if(p->n.next || (p->flags & PTF_UNDO_DELETE))
				p->flags &= ~PTF_UNDO_CREATE;
			else
			{
				string_unref(p->key);
				if(p->type == PT_STRING)
					string_unref(p->value.s);
				sfree(p);
			}
			break;
		}

		case TRANS_PROP_DELETE:
		{
			struct prop* p = uk->u.prop.prop;
			ASSERT((p->flags & PTF_UNDO_DELETE) && !(p->flags & PTF_VIRTUAL));
			o_unref(uk->u.prop.o);
			if(p->n.next || (p->flags & PTF_UNDO_CREATE))
				p->flags &= ~PTF_UNDO_DELETE;
			else
			{
				string_unref(p->key);
				if(p->type == PT_STRING)
					string_unref(p->value.s);
				sfree(p);
			}
			break;
		}

		case TRANS_DOC_SAVED_START:
			uk->u.doc_save.doc->saved_start = NULL;
			break;

		case TRANS_DOC_SAVED_END:
			uk->u.doc_save.doc->saved_end = NULL;
			break;

		default:
			die("There is no label for trans type %u", uk->type);
	}

	ffree(uk);
}

static void trans_command_gui_delete(struct obj_tlo *tlo, struct undo_gui* ug)
{
	struct undo_kernel *uk;
	while((uk = trans_get_first_kernel(ug)))
		trans_command_kernel_delete(uk);
	clist_remove(&ug->ng);
	if(tlo)
		hook_call_undo_delete(tlo, ug, trans_get_prev_gui(tlo, ug));
	string_unref(ug->name_operation);
	sfree(ug);
}

void trans_delete_all(struct obj_tlo *tlo)
{
	struct undo_gui *ug;
	while((ug = trans_get_next_gui(tlo, &tlo->undo_start)))
		trans_command_gui_delete(NULL, ug);
}

static void trans_begin(char *error_string, jmp_buf *jb, uns type)
{
	struct trans *t = smalloc(sizeof(struct trans));
	t->prev = trans_present;
	trans_present = t;
	t->type = type;
	t->es = error_string;
	t->jb = jb;
	t->log = NULL;
	t->tlo = trans_tlo;
	t->uk = NULL;
}

void trans_begin_main(struct obj_tlo* tlo, const char *name, char* error_string, jmp_buf* jb)
{
	ASSERT(tlo && tlo->obj.kind == T_OBJ && (tlo->obj.type == OT_TLO || tlo->obj.type == OT_UNIVERSUM));
	ASSERT(!trans_is_undo_write_protected());
	ASSERT(!trans_present || trans_present->type != TT_INTERNAL);
	DBG_TRANS("Main trans begin: depth %u, name %s, tlo %s at %p\n", trans_get_depth_of_stack() + 1, name, obj_lookup_name(&tlo->obj), tlo);

	struct trans *tx = tlo_is_in_trans_stack(tlo);
	struct undo_gui *ug;

	while(ug = trans_get_last_gui(tlo), ug != tlo->undo_present)
	{
		hook_call_undo_delete(tlo, ug, trans_get_prev_gui(tlo, ug));
		trans_command_gui_delete(NULL, ug);
	}

	trans_begin(error_string, jb, tx ? TT_SUBMAIN : TT_MAIN);
	trans_tlo = tlo;

	if(tx)
		trans_present->uk = trans_get_last_kernel(trans_get_last_gui(trans_tlo));
	else
	{
		ug = smalloc(sizeof(struct undo_gui));
		clist_init(&ug->lk);
		ug->name_operation = string_lookup_ref(name);
		trans_present->uk = NULL;
		clist_add_tail((clist*)&tlo->undo_start.ng, &ug->ng);
		tlo->undo_present = ug;
	}
}

void trans_begin_sub(char* error_string, jmp_buf* jb)
{
	ASSERT(trans_present && trans_tlo);
	ASSERT((1<<trans_present->type) & (1<<TT_MAIN | 1<<TT_SUBMAIN | 1<<TT_SUB | 1<<TT_INTERNAL));
	DBG_TRANS("Trans begin: depth %u\n", trans_get_depth_of_stack() + 1);
	trans_begin(error_string, jb, TT_SUB);
	trans_present->uk = trans_get_last_kernel(trans_get_last_gui(trans_tlo));
}

void trans_begin_anonymous(char* error_string, jmp_buf* jb)
{
	DBG_TRANS("Trans begin anonymous: depth %u\n", trans_get_depth_of_stack() + 1);
	trans_begin(error_string, jb, TT_ANONYMOUS);
	trans_tlo = NULL;
}

static void trans_begin_internal(struct obj_tlo* tlo, char* error_string, jmp_buf* jb)
{
	DBG_TRANS("Trans begin internal: depth %u\n", trans_get_depth_of_stack() + 1);
	trans_begin(error_string, jb, TT_INTERNAL);
	trans_tlo = tlo;
	if(tlo != tlo_universum && !(tlo->obj.flags & OF_TSORT_ACTIVE))
		tsort_start(tlo);
}

void trans_end(void)
{
	DBG_TRANS("Trans end: depth %u\n", trans_get_depth_of_stack());
	ASSERT(trans_present);
	struct go *go;
	struct hanger_smart *hs;

	if(trans_present->type == TT_MAIN)
	{
		if(trans_tlo == tlo_universum)
		{
			struct obj *obj;
			while((obj = obj_get_first_son(obj_temp)))
				obj_relink(obj, obj_zombie, obj_temp);
		}
		else
		{
			if(trans_tlo->obj.flags & OF_TSORT_ACTIVE)
				tsort_recompute(trans_tlo);

			while((hs = tlg_get_first_hanger_smart(trans_tlo->temp)))
			{
				ASSERT(hs->go == tlg2go(trans_tlo->temp) && hs->ref_count == 0 && clist_empty(&hs->h.lanchor));
				clist_remove(&hs->nhanger);
				xfree(hs);
			}

			while((go = group_get_first_go(&trans_tlo->temp->group)))
			{
				ASSERT(go->kind == T_GO && go_get_group(go) == &trans_tlo->temp->group && !go_is_selected(go));

#ifdef DEBUG_KERNEL_SLOW
struct go *g, *dep;
				WALK_GO_RECURSIVE(go, g)
					WALK_GO_DEPENDENT_BEGIN(g, dep)
							ASSERT(go_get_root(dep) == tlo_zombie->tlg || go_get_root(dep) == trans_tlo->temp);
					WALK_GO_DEPENDENT_END
#endif

				go_relink_low(go, &tlo_zombie->tlg->group, NULL);
			}
		}

		ASSERT(trans_get_last_gui(trans_tlo) == trans_tlo->undo_present && trans_tlo->undo_present != &trans_tlo->undo_start);
		if(trans_get_first_kernel(trans_tlo->undo_present))
			hook_call_undo_new(trans_tlo, trans_tlo->undo_present);
		else
		{
			clist_remove(&trans_tlo->undo_present->ng);
			string_unref(trans_tlo->undo_present->name_operation);
			sfree(trans_tlo->undo_present);
			trans_tlo->undo_present = trans_get_last_gui(trans_tlo);
		}
		ASSERT(trans_tlo->undo_present == trans_get_last_gui(trans_tlo));
		ASSERT(trans_tlo->undo_present == &trans_tlo->undo_start || trans_get_first_kernel(trans_tlo->undo_present));
	}

	trans_tlo = trans_present->tlo;

	struct trans* t = trans_present;
	struct trans_log* l;
	while((l = t->log))
	{
		t->log = t->log->prev;
		sfree(l);
	}
	trans_present = trans_present->prev;
	sfree(t);

	if(!trans_present)
		string_trans_end();

	ASSERT(trans_present || !trans_tlo);
}

void trans_fail(const char* format, ...)
{
	ASSERT(trans_present);

	jmp_buf* jb = trans_present->jb;

	va_list ap;
	va_start(ap, format);
	int written = vsnprintf(trans_present->es, TRANS_ERR_SIZE - 1, format, ap);
	va_end(ap);
	DBG_TRANS("Trans fail: depth %u, error string \"%s\"\n", trans_get_depth_of_stack(), trans_present->es);
	if(unlikely(written > TRANS_ERR_SIZE))
		msg("Error message is too long. Please, make constant TRANS_ERR_SIZE (kernel/kernel.h) not be less than %d.\n", written);
	if(unlikely(written < 0))
		msg("Error message is too long. Please, make constant TRANS_ERR_SIZE (kernel/kernel.h) larger.\n");

	struct trans_log* l;
	while((l = trans_present->log))
	{
		trans_present->log = trans_present->log->prev;
		l->func(l->arg);
		sfree(l);
	}

	if((1<<trans_present->type) & (1<<TT_MAIN | 1<<TT_SUBMAIN | 1<<TT_SUB))
	{
		char buf[TRANS_ERR_SIZE];
		struct undo_gui *ug = trans_get_last_gui(trans_tlo);
		ASSERT(ug == trans_tlo->undo_present);
		struct undo_kernel *uk, *stop = trans_present->uk;

		if(trans_tlo != tlo_universum)
			tsort_trans_fail(trans_tlo);

		DBG_TRANS("Fail: ");
		TRANS_BEGIN_INTERNAL(trans_tlo, buf)
			while(uk = trans_get_last_kernel(ug), uk != stop)
			{
				trans_command_undo(uk, TRANS_UNDO);
				trans_command_kernel_delete(uk);
			}
			DBG_TRANS("Fail: ");
		TRANS_FAILED
			die("Trans double fail: the first error \"%s\", the second error \"%s\".", trans_present->es, buf);
		TRANS_END
	}

	trans_end();

	longjmp(*jb, 1);
	die("longjmp may leave this function");
}

struct undo_kernel* trans_new_undo(uns trans_type)
{
	ASSERT(trans_present && trans_tlo);
	ASSERT((1<<trans_present->type) & (1<<TT_MAIN | 1<<TT_SUBMAIN | 1<<TT_SUB));
	struct undo_kernel* u = fmalloc(trans_tlo->fuk);
	u->type = trans_type;
	clist_add_tail(&trans_tlo->undo_present->lk, &u->nk);
	return u;
}

void trans_new_undo_prop_virtual(struct o *o, struct prop_virtual *prop, prop_value value)
{
	TRANS_NEW_UNDO(TRANS_PROP_VIRTUAL, prop, o_ref(o), &prop->p, value);
	if(prop->p.type == PT_STRING)
		string_ref(value.s);
}

void trans_new_undo_start(struct o *o, struct prop_virtual *prop, prop_value value)
{
	ASSERT(prop->undo_mask != 0 && prop->undo_mask != UINT_MAX && !(o->flags & prop->undo_mask));
	TRANS_NEW_UNDO(TRANS_PROP_VIRTUAL_START_UNDO, prop, o_ref(o), &prop->p, value);
	if(prop->p.type == PT_STRING)
		string_ref(value.s);
	o->flags |= prop->undo_mask;	
}

void trans_new_undo_end(struct o *o, struct prop_virtual *prop, prop_value value)
{
	ASSERT(prop->undo_mask != 0 && prop->undo_mask != UINT_MAX && (o->flags & prop->undo_mask));
	TRANS_NEW_UNDO(TRANS_PROP_VIRTUAL_END_UNDO, prop, o_ref(o), &prop->p, value);
	if(prop->p.type == PT_STRING)
		string_ref(value.s);
	o->flags &= ~prop->undo_mask;	
}

uns trans_undo(struct obj_tlo* tlo, char *error_string)
{
	ASSERT(tlo && tlo->obj.kind == T_OBJ && (tlo->obj.type == OT_TLO || tlo->obj.type == OT_UNIVERSUM));
	ASSERT(!tlo_is_in_trans_stack(tlo));

	struct undo_kernel* u = NULL;
	volatile uns ret = 1;

	if(tlo->undo_present == &tlo->undo_start)
	{
		DBG_TRANS("Trans undo: The begin of history, tlo %s at %p.\n", obj_lookup_name(&tlo->obj), tlo);
		strcpy(error_string, "This is the begin of history.");
		return 0;
	}

	DBG_TRANS("Undo: ");
	TRANS_BEGIN_INTERNAL(tlo, error_string)

		for(u = trans_get_last_kernel(tlo->undo_present); u; u = trans_get_prev_kernel(tlo->undo_present, u))
			trans_command_undo(u, TRANS_UNDO);

		if(tlo != tlo_universum)
			tsort_recompute(tlo);
		tlo->undo_present = trans_get_prev_gui(tlo, tlo->undo_present);
		hook_call_undo(tlo);

		DBG_TRANS("Undo: ");
	TRANS_FAILED
		char buf[TRANS_ERR_SIZE];

		DBG_TRANS("Undo fail: ");
		TRANS_BEGIN_INTERNAL(tlo, buf)
			for(; u; u = trans_get_prev_kernel(tlo->undo_present, u))
				trans_command_undo(u, TRANS_REDO);
			DBG_TRANS("Undo fail: ");
		TRANS_FAILED
			die("Undo double fail: the first error \"%s\", the second error \"%s\".", error_string, buf);
		TRANS_END

		ret = 0;
	TRANS_END

	return ret;
}

uns trans_redo(struct obj_tlo* tlo, char *error_string)
{
	ASSERT(tlo && tlo->obj.kind == T_OBJ && (tlo->obj.type == OT_TLO || tlo->obj.type == OT_UNIVERSUM));
	ASSERT(!tlo_is_in_trans_stack(tlo));

	struct undo_kernel* u = NULL;
	volatile uns ret = 1;

	if(!trans_get_next_gui(tlo, tlo->undo_present))
	{
		DBG_TRANS("Trans redo: The end of history, tlo %s at %p.\n", obj_lookup_name(&tlo->obj), tlo);
		strcpy(error_string, "This is the end of history.");
		return 0;
	}
	tlo->undo_present = trans_get_next_gui(tlo, tlo->undo_present);

	DBG_TRANS("Redo: ");
	TRANS_BEGIN_INTERNAL(tlo, error_string)

		for(u = trans_get_first_kernel(tlo->undo_present); u; u = trans_get_next_kernel(tlo->undo_present, u))
			trans_command_undo(u, TRANS_REDO);

		if(tlo != tlo_universum)
			tsort_recompute(tlo);
		hook_call_redo(tlo);

		DBG_TRANS("Redo: ");
	TRANS_FAILED
		char buf[TRANS_ERR_SIZE];

		DBG_TRANS("Redo fail: ");
		TRANS_BEGIN_INTERNAL(tlo, buf)
			for(; u; u = trans_get_prev_kernel(tlo->undo_present, u))
				trans_command_undo(u, TRANS_UNDO);
			DBG_TRANS("Redo fail: ");
		TRANS_FAILED
			die("Redo double fail: the first error \"%s\", the second error \"%s\".", error_string, buf);
		TRANS_END

		tlo->undo_present = trans_get_prev_gui(tlo, tlo->undo_present);
		ret = 0;
	TRANS_END

	return ret;
}

void trans_log(void (*func)(uintptr_t arg), uintptr_t arg)
{
	ASSERT(trans_present);		// This function must be called in trans.
	struct trans_log *l = smalloc(sizeof(struct trans_log));
	l->func = func;
	l->arg = arg;
	l->prev = trans_present->log;
	trans_present->log = l;
}

// FIXME: ulozenost souboru ...
void trans_clear_undo(struct obj_tlo* tlo)
{
	ASSERT(tlo->obj.kind == T_OBJ && (tlo->obj.type == OT_TLO || tlo->obj.type == OT_UNIVERSUM) && !tlo_is_in_trans_stack(tlo));
	struct undo_gui* ug;
	while(ug = tlo->undo_present, ug != &tlo->undo_start)
	{
		tlo->undo_present = trans_get_prev_gui(tlo, ug);
		trans_command_gui_delete(tlo, ug);
	}
}

void trans_clear_one_redo(struct obj_tlo* tlo)
{
	ASSERT(tlo->obj.kind == T_OBJ && (tlo->obj.type == OT_TLO || tlo->obj.type == OT_UNIVERSUM) && !tlo_is_in_trans_stack(tlo));
	struct undo_gui* ug = trans_get_last_gui(tlo);
	ASSERT(ug != tlo->undo_present);
	if(ug == tlo->undo_saved)
		tlo->undo_saved = NULL;
	trans_command_gui_delete(tlo, ug);
}

void trans_clear_all_redo(struct obj_tlo* tlo)
{
	ASSERT(tlo->obj.kind == T_OBJ && (tlo->obj.type == OT_TLO || tlo->obj.type == OT_UNIVERSUM) && !tlo_is_in_trans_stack(tlo));
	struct undo_gui* ug;
	while(ug = trans_get_last_gui(tlo), ug != tlo->undo_present)
	{
		if(ug == tlo->undo_saved)
			tlo->undo_saved = NULL;
		trans_command_gui_delete(tlo, ug);
	}
}

struct undo_gui* trans_merge_undo(struct obj_tlo *tlo, struct undo_gui *first, string name)
{
	ASSERT(tlo->obj.kind == T_OBJ && (tlo->obj.type == OT_TLO || tlo->obj.type == OT_UNIVERSUM) && !tlo_is_in_trans_stack(tlo));
	ASSERT(clist_is_node_member((clist*)&tlo->undo_start.ng, &first->ng));
	struct undo_gui *second = trans_get_next_gui(tlo, first);
	ASSERT(second && tlo->undo_present != first);	// The first cannot be the last item of a history and present cannot be between the first and the second.

	if(first == tlo->undo_saved)
		tlo->undo_saved = NULL;
	
	string_ref(name);
	hook_call_undo_delete(tlo, first, trans_get_prev_gui(tlo, first));
	clist_remove(&first->ng);
	string_unref(second->name_operation);
	string_unref(first->name_operation);
	first->name_operation = name;
	hook_call_undo_update(tlo, second);
	
	struct undo_kernel *first_last = trans_get_last_kernel(first),
			*second_first = trans_get_first_kernel(second),
			*first_first = trans_get_first_kernel(first);
	first_last->nk.next = &second_first->nk;
	second_first->nk.prev = &first_last->nk;
	first_first->nk.prev = &second->lk.head;
	second->lk.head.next = &first_first->nk;

	sfree(first);
	return second;
}

void trans_rename_undo(struct obj_tlo *tlo, struct undo_gui *ug, string name)
{
	ASSERT(tlo->obj.kind == T_OBJ && (tlo->obj.type == OT_TLO || tlo->obj.type == OT_UNIVERSUM) && !tlo_is_in_trans_stack(tlo));
	ASSERT(clist_is_node_member((clist*)&tlo->undo_start.ng, &ug->ng));

	if(name != ug->name_operation) 	// Gui mostly calls this function such that name is same as ug->name_operation.
	{
		string_ref(name);
		string_unref(ug->name_operation);
		ug->name_operation = name;
		hook_call_undo_update(tlo, ug);
	}
}
