//#define LOCAL_DEBUG
#include "kernel/internal.h"

#define WALK_TSORT_DEPENDENT(_go, _walk, _command) \
	WALK_GO_DEPENDENT_BEGIN(_go, _walk) _command; WALK_GO_DEPENDENT_END \

/* #define WALK_TSORT_DEPENDENT(_go, _walk, _command) do { ASSERT(go_get_group(_go)); \
	WALK_GO_DEPENDENT_BEGIN(_go, _walk) _command; WALK_GO_DEPENDENT_END \
	_walk = &go_get_group(_go)->go; if(_walk->subtype != GOST_GROUP_TLG) do { _command; } while(0);  \
	} while(0) */

static inline int tsort_is_active(struct obj_tlo *tlo)
	{ ASSERT_TLO(tlo); return tlo->obj.flags & OF_TSORT_ACTIVE; }
static inline int tsort_is_dirty(struct obj_tlo *tlo)
	{ return tlo->obj.flags & OF_TSORT_DIRTY; }
static inline int tsort_is_presort(struct obj_tlo *tlo)
	{ return tlo->obj.flags & OF_TSORT_PRESORT; }
static inline int tsort_is_go_flag(struct go *go)
	{ return go->flags & GOF_TSORT; }

#define TSORT_STATUS(_tlo, _d, _p) ( tsort_is_active(_tlo) && \
	( _d != -1 ? (_d == (tsort_is_dirty(_tlo) == OF_TSORT_DIRTY)) : 1 ) && \
	( _p != -1 ? (_p == (tsort_is_presort(_tlo) == OF_TSORT_PRESORT)) : 1 ))
	
/*
 * Debugging
 */

#define ASSERT_INITIALIZATION(_tlo, _tlg) WALK_GO_RECURSIVE(tlg2go(_tlo->_tlg), go) ASSERT(go->temp_data == 0 && !tsort_is_go_presorted(go) && !tsort_is_go_flag(go));

#ifdef DEBUG_KERNEL
static void assert_tsort_pasive(struct obj_tlo *tlo)
{
	ASSERT(!tsort_is_active(tlo) && !tsort_is_dirty(tlo) && !tsort_is_presort(tlo) && flist_is_initial(&tlo->tsort));

#ifdef DEBUG_KERNEL_SLOW
	struct go* go;
	ASSERT_INITIALIZATION(tlo, tlg);
	ASSERT_INITIALIZATION(tlo, temp);
	ASSERT_INITIALIZATION(tlo_zombie, tlg);
#endif
}

#else
static inline void assert_tsort_pasive(struct obj_tlo *tlo UNUSED) { }
#endif


/*
 * Callback
 */

static void tsort_void_recompute(struct go *go, void UNUSED *v)
{
	go_recompute(go);
}


/*
 * Initialization and cleanup
 */

static void tsort_cleanup(struct obj_tlo *tlo)
{
	struct go *go, *tlg = &tlo_get_tlg(tlo)->go;
	WALK_GO_RECURSIVE(tlg, go)
	{
		go->temp_data = 0;
		go->nfifo.next = NULL;
		go->flags &= ~GOF_TSORT;
	}
	flist_init(&tlo->tsort);
}

void tsort_start(struct obj_tlo *tlo)
{
	DBG_TSORT("tsort active start: tlo %s at %p\n", obj_lookup_name(&tlo->obj), tlo);
	assert_tsort_pasive(tlo);
	tlo->obj.flags |= OF_TSORT_ACTIVE;
}

static inline void tsort_end(struct obj_tlo *tlo)
{
	ASSERT(TSORT_STATUS(tlo, 0, -1));
	
	if(tsort_is_dirty(tlo))
	{
		DANGEROUS(0, "Tsort of tlo %s at %p is finishing but it has dirty gos.\n", obj_lookup_name(&tlo->obj), tlo);
		tsort_call_recompute(tlo);
	}
	
	DANGEROUS(!tsort_is_presort(tlo), "Tsort of tlo %s at %p is finishing but it is presorted.\n", obj_lookup_name(&tlo->obj), tlo);
	
	if(!flist_empty(&tlo->tsort))	
	{
		DANGEROUS(0, "Tsort of tlo %s at %p is finishing but it has gos in list.\n", obj_lookup_name(&tlo->obj), tlo);
		tsort_cleanup(tlo);
	}
	if(tsort_is_presort(tlo))
		DBG_TSORT("tsort presort end: tlo %s at %p\n", obj_lookup_name(&tlo->obj), tlo);
	DBG_TSORT("tsort active end: tlo %s at %p\n", obj_lookup_name(&tlo->obj), tlo);
	tlo->obj.flags &= ~(OF_TSORT_ACTIVE | OF_TSORT_PRESORT);
	assert_tsort_pasive(tlo);
}

void tsort_presort_end(struct obj_tlo *tlo)
{
	ASSERT(TSORT_STATUS(tlo, -1, 1));
	struct snode* n;
	struct go *go;

	if(tsort_is_dirty(tlo))
		tsort_presorted_recompute(tlo);
	
	for(n = tlo->tsort.head.next; n != &tlo->tsort.head; )
	{
		go = nfifo2go(n, tlo);
		n = n->next;
		go->nfifo.next = NULL;
		go->flags &= ~GOF_TSORT;
		go->temp_data = 0;
	}
	flist_init(&tlo->tsort);
	
	DBG_TSORT("tsort presort end: tlo %s at %p\n", obj_lookup_name(&tlo->obj), tlo);
	DBG_TSORT("tsort active end: tlo %s at %p\n", obj_lookup_name(&tlo->obj), tlo);
	tlo->obj.flags &= ~(OF_TSORT_ACTIVE | OF_TSORT_PRESORT);
	assert_tsort_pasive(tlo);
}


/*
 * Insert go into tsort list.
 */

void tsort_insert(struct go *go, struct obj_tlo *tlo)
{
	ASSERT(TSORT_STATUS(tlo, -1, -1) && assert_go_in_tlo(go, tlo));
	if(!tsort_is_go_presorted(go) && go->type != GOT_GROUP)
	{
		if(tsort_is_go_presorted(go))
		{
			DANGEROUS(0, "Inserting into presorted tlo %s at %p.\n", obj_lookup_name(&tlo->obj), tlo);
			tsort_unsort(tlo);
		}
		DBG("tsort insert: go %s at %p, temp_data %u\n", go_lookup_name(go), go, go->temp_data);
		flist_add_tail(&tlo->tsort, &go->nfifo);
	}
}

void tsort_insert_flag(struct go *go, struct obj_tlo *tlo)
{
	if(go->type != GOT_GROUP)
	{
		tsort_insert(go, tlo);
		go->flags |= GOF_TSORT;
	}
}

void tsort_insert_group(struct go_group *group, struct obj_tlo *tlo)
{
	ASSERT(TSORT_STATUS(tlo, -1, -1) && assert_go_in_tlo(group2go(group), tlo));
	struct go *go;
	WALK_GO_RECURSIVE_WITHOUT_ROOT(&group->go, go)
		tsort_insert(go, tlo);
}

void tsort_insert_group_flag(struct go_group *group, struct obj_tlo *tlo)
{
	ASSERT(TSORT_STATUS(tlo, -1, -1) && assert_go_in_tlo(group2go(group), tlo));
	struct go *go;
	WALK_GO_RECURSIVE_WITHOUT_ROOT(&group->go, go)
		tsort_insert_flag(go, tlo);
}

void tsort_insert_selected(struct go_group *group, struct obj_tlo *tlo)
{
	ASSERT(TSORT_STATUS(tlo, -1, -1) && assert_go_in_tlo(group2go(group), tlo));
	struct go *go, *g;
	WALK_GROUP_SELECTED_GO(group, go)
		WALK_GO_RECURSIVE(go, g)
			tsort_insert(g, tlo);
}

void tsort_insert_selected_flag(struct go_group *group, struct obj_tlo *tlo)
{
	ASSERT(TSORT_STATUS(tlo, -1, -1) && assert_go_in_tlo(group2go(group), tlo));
	struct go *go, *g;
	WALK_GROUP_SELECTED_GO(group, go)
		WALK_GO_RECURSIVE(go, g)
			tsort_insert_flag(g, tlo);
}

void tsort_insert_hanger(struct hanger *h, struct obj_tlo *tlo)
{
	ASSERT(TSORT_STATUS(tlo, -1, -1) && assert_go_in_tlo(hanger_get_go(h), tlo));
	struct anchor *a;
	WALK_HANGER_ANCHOR(h, a)
		tsort_insert(anchor_get_go(a), tlo);
}

void tsort_dirty(struct go *go)
{
	ASSERT(TSORT_STATUS(trans_tlo, -1, -1) && assert_go_in_tlo(go, trans_tlo));
	if(&go_get_root(go)->group != tlo_get_tlg(trans_tlo))
		return;
	tsort_insert(go, trans_tlo);
	if(!tsort_is_dirty(trans_tlo))
		DBG_TSORT("tsort dirty start: tlo %s at %p\n", obj_lookup_name(&trans_tlo->obj), trans_tlo);
  	trans_tlo->obj.flags |= OF_TSORT_DIRTY;
}
	
void tsort_dirty_hanger(struct hanger *h)
{
	ASSERT(TSORT_STATUS(trans_tlo, -1, -1) && assert_go_in_tlo(hanger_get_go(h), trans_tlo));
	struct anchor *a;
	WALK_HANGER_ANCHOR(h, a)
		tsort_dirty(anchor_get_go(a));
}

/*
 * Remove go from tsort list.
 */

void tsort_remove(struct go *go, struct obj_tlo *tlo)
{
	ASSERT(go_get_tlo(go) == tlo || go_get_tlo(go) == tlo_zombie);
	if(go->nfifo.next)
	{
		ASSERT(TSORT_STATUS(tlo, -1, 0));
		DBG("tsort remove: go %s at %p, temp_data %u\n", go_lookup_name(go), go, go->temp_data);
		flist_remove(&tlo->tsort, &go->nfifo);
		go->flags &= ~GOF_TSORT;
	}
}


/*
 * Topological sort
 */

static uns tsort_prepare(struct obj_tlo *tlo)
{
	struct snode* n, *prev, *mark = tlo->tsort.tail;
	struct go *go;
	uns outside_list = 0;

	for(n = tlo->tsort.head.next; n != &tlo->tsort.head; n = n->next)
		WALK_TSORT_DEPENDENT(nfifo2go(n, tlo), go,
			if(!go->nfifo.next)
			{
				flist_add_tail(&tlo->tsort, &go->nfifo);
				outside_list++;
			}
			go->temp_data++;
			DBG("tsort increase: go %s at %p, temp_data %u\n", go_lookup_name(go), go, go->temp_data);
		);

	tlo->tsort.tail = mark;
	mark->next = &tlo->tsort.head;

	for(prev = &tlo->tsort.head; prev->next != &tlo->tsort.head; )
	{
		go = nfifo2go(prev->next, tlo);
		if(go->temp_data)
		{
			flist_remove_after(&tlo->tsort, prev);
			outside_list++;
			DBG("tsort remove: go %s at %p, temp_data %u\n", go_lookup_name(go), go, go->temp_data);
		}
		else
			prev = prev->next;
	}
	
	return outside_list;
}

void tsort_call(struct obj_tlo *tlo, void (*cb) (struct go*, void*), void *data)
{
	DBG("tsort call: tlo %s at %p\n", obj_lookup_name(&tlo->obj), tlo);
	ASSERT(TSORT_STATUS(tlo, -1, -1));
	struct go *go, *dep;
	uns outside_list = 0;

	// This is only heuristic.
	if(tsort_is_presort(tlo))
		WALK_TSORT_GO(tlo, go)
			cb(go, data);	
	else
	{
		outside_list = tsort_prepare(tlo);
		while(!flist_empty(&tlo->tsort))
		{
			go = nfifo2go(flist_head(&tlo->tsort), tlo);

			DBG("tsort callback and remove: go %s at %p, temp_data %u\n", go_lookup_name(go), go, go->temp_data);
			cb(go, data);
			flist_remove_head(&tlo->tsort); // This must be after calling cb. If cb is recompute then go must stay in tsort list to be recomputed.

			WALK_TSORT_DEPENDENT(go, dep,
				dep->temp_data--;
				DBG("tsort decrease: go %s at %p, temp_data %u, hangs on %s at %p\n", go_lookup_name(dep), dep, dep->temp_data, go_lookup_name(go), go);
				if(dep->temp_data == 0)
				{
					flist_add_tail(&tlo->tsort, &dep->nfifo);
					outside_list--;
					DBG("tsort insert: go %s at %p, temp_data %u\n", go_lookup_name(dep), dep, dep->temp_data);
				}
			);
		}
	}
	if(outside_list)
	{
		DANGEROUS(!outside_list, "Dependence graph contains circle with %d gos.\n", outside_list);
		tsort_cleanup(tlo);
		trans_fail("Cirkular dependence.");
	}
	if(tsort_void_recompute == cb)
	{
		DBG_TSORT("tsort dirty end: tlo %s at %p\n", obj_lookup_name(&tlo->obj), tlo);
		tlo->obj.flags &= ~OF_TSORT_DIRTY;
	}
	tsort_end(tlo);
}

void tsort_presort(struct obj_tlo *tlo)
{
	DBG_TSORT("tsort presort start: tlo %s at %p\n", obj_lookup_name(&tlo->obj), tlo);
	ASSERT(TSORT_STATUS(tlo, -1, 0));
	struct go *go, *dep;

	uns outside_list = tsort_prepare(tlo);
	tlo->obj.flags |= OF_TSORT_PRESORT;

	WALK_TSORT_GO(tlo, go)
		WALK_TSORT_DEPENDENT(go, dep,
			dep->temp_data--;
			DBG("tsort decrease: go %s at %p, temp_data %u\n", go_lookup_name(dep), dep, dep->temp_data);
			if(dep->temp_data == 0)
			{
				outside_list--;
				DBG("tsort insert: go %s at %p, temp_data %u\n", go_lookup_name(dep), dep, dep->temp_data);
				flist_add_tail(&tlo->tsort, &dep->nfifo);
			}
		);
	if(outside_list)
	{
		DANGEROUS(0, "Dependence graph contains circle with %d gos.\n", outside_list);
		tsort_cleanup(tlo);
		trans_fail("Circular dependence.");
	}
}

void tsort_trans_fail(struct obj_tlo *tlo)
{
	if(!tsort_is_active(tlo) || tsort_is_presort(tlo))
		return;

	DBG_TSORT("tsort trans fail: tlo %s at %p, trans_tlo %s at %p\n",
			obj_lookup_name(&tlo->obj), tlo, obj_lookup_name(&tlo->obj), tlo);
	
	struct go *go, *dep;

	WALK_TSORT_GO(tlo, go)
	{
		WALK_TSORT_DEPENDENT(go, dep,
			dep->temp_data--;
			DBG("tsort decrease: go %s at %p, temp_data %u, hangs on %s at %p\n", go_lookup_name(dep), dep, dep->temp_data, go_lookup_name(go), go);
			if(dep->temp_data == 0)
			{
				flist_add_tail(&tlo->tsort, &dep->nfifo);
				DBG("tsort insert: go %s at %p, temp_data %u\n", go_lookup_name(dep), dep, dep->temp_data);
			}
		);
	}
}


/*
 * Shortcuts for tsort_start, tsort_insert_xxx, tsort_call
 */

void tsort_go_call(struct go *go, struct obj_tlo *tlo, void (*cb) (struct go*, void*), void *data)
{
	ASSERT(assert_go_in_tlo(go, tlo) && !tsort_is_active(tlo));
	if(go_has_dependent(go))
	{
		tsort_start(tlo);
		tsort_insert(go, tlo);
		tsort_call(tlo, cb, data);
	}
	else  // This is only heuristic.
		cb(go, data);
}

void tsort_hanger_call(struct hanger *h, struct obj_tlo *tlo, void (*cb) (struct go*, void*), void *data)
{
	ASSERT(assert_go_in_tlo(hanger_get_go(h), tlo) && !tsort_is_active(tlo));
	struct anchor* a = hanger_get_first_anchor(h);

	if(!a)
		return;
	if(!hanger_get_next_anchor(a))
		tsort_go_call(anchor_get_go(a), tlo, cb, data);
	else
	{
		tsort_start(tlo);
		WALK_HANGER_ANCHOR(h, a)
			tsort_insert(anchor_get_go(a), tlo);
		tsort_call(tlo, cb, data);
	}
}


/*
 * Transformation
 */

static inline int type_transform(uns type, int sim)
	{ return ((type == LT_AFFINITY) || (type == LT_SIMILARITY && sim)); }

void tsort_presorted_transform(struct geom_transform2 *t)
{
	ASSERT(TSORT_STATUS(trans_tlo, 0, 1));
	int i, hpos, sim = geom_transform_is_similar(&t->primary);
	struct anchor *a;
	struct hanger *h;
	struct geom_point gp;
	struct go *go, *g;

	WALK_TSORT_GO(trans_tlo, go)
	{
		if(tsort_is_go_flag(go))
		{
			WALK_GO_ANCHOR(go, a, i)
			{
				h = a->h;
				g = hanger_get_go(h);
				if((hpos = hanger_is_position(h)))
					gp = hanger_get_position(h);
				if(hanger_is_mc(h))
				{
					if(hanger_is_position(h) && a->info->linear == LT_AFFINITY)	// All non-affinity anchors hanging on mouse-click must be transformed by go's transformation function.
						anchor_rehang_mc_no_recompute(a, point_transform(&gp, &t->primary));
				}
				else if(!type_transform(a->info->linear, sim) // bad dependent between go and anchor
					  	|| !type_transform(h->info->linear, sim)  // bad dependent between hanger and go
						|| !(tsort_is_go_flag(g) // If go isn't transformed hanger must be in the fix point.
							|| (hpos && geom_transform_is_point_fixed(hanger_get_position_ptr(h, &gp), &t->primary))))
					TRANS_FAIL(("Cannot transform."),
							("go %s at %p flags %d, anchor %s, hanger %s, go %s at %p flags %d, similar %d, fixed point %d.",
							 go_lookup_name(go), go, go->flags, a->info->name->text, h->info->name->text,
							 go_lookup_name(g), g, g->flags, geom_transform_is_similar(&t->primary),
							 hpos ? geom_transform_is_point_fixed(&gp, &t->primary) : -1));
			}
			go_transform(go, t);
		}
		else
			go_recompute(go);
	}

	DBG_TSORT("tsort dirty end: tlo %s at %p\n", obj_lookup_name(&trans_tlo->obj), trans_tlo);
	trans_tlo->obj.flags &= ~OF_TSORT_DIRTY;
}


/*
 * Recomputing
 */

void tsort_call_recompute(struct obj_tlo *tlo)
{
	ASSERT(TSORT_STATUS(tlo, -1, -1));
	if(!tsort_is_dirty(tlo))
		DBG_TSORT("Tsort: Recompute is not requested.\n");
	tsort_call(tlo, tsort_void_recompute, NULL);
}

void tsort_presorted_recompute(struct obj_tlo *tlo)
{
	ASSERT(TSORT_STATUS(tlo, -1, 1));
	if(!tsort_is_dirty(tlo))
		DBG_TSORT("Tsort: Recompute is not requested.\n");
	struct go *go;

	if(!tsort_is_dirty(tlo))
		return;

	WALK_TSORT_GO(tlo, go)
		go_recompute(go);

	DBG_TSORT("tsort dirty end: tlo %s at %p\n", obj_lookup_name(&tlo->obj), tlo);
	tlo->obj.flags &= ~OF_TSORT_DIRTY;
}

void tsort_recompute(struct obj_tlo *tlo)
{
	ASSERT(TSORT_STATUS(tlo, -1, -1));
	if(tsort_is_presort(tlo))
		tsort_presorted_recompute(tlo);
	else
		tsort_call_recompute(tlo);
}


/*
 * Miscellaneous
 */

void tsort_unsort(struct obj_tlo *tlo)
{
	ASSERT(TSORT_STATUS(tlo, -1, 1));
	DBG_TSORT("tsort presort end: tlo %s at %p\n", obj_lookup_name(&tlo->obj), tlo);
	tlo->obj.flags &= ~OF_TSORT_PRESORT;
}

