#ifndef _KERNEL_TSORT_H_
#define _KERNEL_TSORT_H_


/*
 * Function tsort_start must be call the first.
 */

void tsort_start(struct obj_tlo *tlo);


/*
 * Then insert gos into tsort list.
 */

void tsort_insert(struct go *go, struct obj_tlo *tlo);
void tsort_insert_group(struct go_group *group, struct obj_tlo *tlo);
void tsort_insert_selected(struct go_group *group, struct obj_tlo *tlo);
void tsort_insert_hanger(struct hanger *h, struct obj_tlo *tlo);


/*
 * Some special operation makes differnce between marked na unmarked gos.
 * For example: transformation.
 */

void tsort_insert_flag(struct go *go, struct obj_tlo *tlo);
void tsort_insert_group_flag(struct go_group *group, struct obj_tlo *tlo);
void tsort_insert_selected_flag(struct go_group *group, struct obj_tlo *tlo);


/*
 * When all necessary gos are inserted two functions can be called.
 *
 * The first, tsort_call, is called if tsort is used once.
 * When this function finish tsort list is empty.
 */

void tsort_call(struct obj_tlo *tlo, void (*cb) (struct go*, void*), void *data);


/*
 * Shortcuts for tsort_start, tsort_insert, tsort_call.
 */

void tsort_go_call(struct go *go, struct obj_tlo *tlo, void (*cb) (struct go*, void*), void *data);
void tsort_hanger_call(struct hanger *h, struct obj_tlo *tlo, void (*cb) (struct go*, void*), void *data);


/*
 * The second, tsort_presort, presorts given gos.
 */

void tsort_presort(struct obj_tlo *tlo);


/*
 * Shortcuts for tsort_start, tsort_insert, tsort_presort.
 */

static inline void tsort_hanger_presort(struct hanger *h, struct obj_tlo *tlo)
	{ tsort_start(tlo); tsort_insert_hanger(h, tlo); tsort_presort(tlo); }
	
static inline void tsort_go_presort(struct go *go, struct obj_tlo *tlo)
	{ tsort_start(tlo); tsort_insert(go, tlo); tsort_presort(tlo); }

static inline void tsort_selected_presort(struct go_group *group, struct obj_tlo *tlo)
	{ tsort_start(tlo); tsort_insert_selected(group, tlo); tsort_presort(tlo); }

static inline void tsort_selected_flag_presort(struct go_group *group, struct obj_tlo *tlo)
	{ tsort_start(tlo); tsort_insert_selected_flag(group, tlo); tsort_presort(tlo); }


/*
 * Presorted list can be used arbitrary times.
 */

static inline struct go* nfifo2go(struct snode *n, struct obj_tlo *tlo)
	{ return assert_go_in_tlo(SKIP_BACK(struct go, nfifo, n), tlo); }
	
static inline struct go* tsort_get_first_go(struct obj_tlo *tlo)
	{ return tlo->tsort.head.next != &tlo->tsort.head ? nfifo2go(tlo->tsort.head.next, tlo) : NULL; }
	
static inline struct go* tsort_get_next_go(struct go *go, struct obj_tlo *tlo)
	{ return go->nfifo.next != &tlo->tsort.head ? nfifo2go(go->nfifo.next, tlo) : NULL; }
	
#define WALK_TSORT_GO(_tlo, _walk) for(_walk = tsort_get_first_go(_tlo); _walk; _walk = tsort_get_next_go(_walk, _tlo))


/*
 * In the end function tsort_presort_end must be called.
 */

void tsort_presort_end(struct obj_tlo *tlo);


/*
 * Transformation
 */

void tsort_presorted_transform(struct geom_transform2 *t);

static inline void tsort_selected_transform(struct go_group *group, struct geom_transform2 *t)
	{ tsort_selected_flag_presort(group, trans_tlo); tsort_presorted_transform(t); tsort_presort_end(trans_tlo); }


/*
 * Tsort is also used for suspense recomputing.
 */

void tsort_call_recompute(struct obj_tlo *tlo);
void tsort_presorted_recompute(struct obj_tlo *tlo);
void tsort_recompute(struct obj_tlo *tlo);

#define TSORT_NO_RECOMPUTE_BEGIN do { \
	ASSERT_TLO(trans_tlo); \
	int _tsort_active = trans_tlo->obj.flags & OF_TSORT_ACTIVE; \
	if(!_tsort_active) tsort_start(trans_tlo); \

#define TSORT_NO_RECOMPUTE_END \
	if(!_tsort_active) tsort_recompute(trans_tlo); \
	} while(0);


/*
 * Miscellaneous
 */

static inline int tsort_is_go_presorted(struct go *go)
	{ ASSERT_GO(go); return go->nfifo.next ? 1 : 0; }

void tsort_unsort(struct obj_tlo *tlo);

#endif
