#ifndef _LIB_AVLTREE_H
#define _LIB_AVLTREE_H

/*
 *	Example:
 *
 *	#include avltree.h
 *
 *	struct mynode {
 *		struct avlnode avlnode;
 *		int key;
 *	}
 *
 *	#define AVLTREE_TYPE int
 *	#define AVLTREE_KEY(x) ((mynode *)x)->key
 *	#include avltreegen.h
 *
 *	int main()
 *	{
 *		struct avltree tree;
 *		struct mynode node;
 *
 *		avltree_init(&tree);
 *		node.key = 100;
 *		avltree_insert(&tree, &node.avlnode);
 *
 *		...
 *	}
 */


/** AVL trees height limit */
#define AVLTREE_MAXHEIGHT 64

#define AVLTREE_FUNC_FIND     0x01
#define AVLTREE_FUNC_FIND_GE  0x02
#define AVLTREE_FUNC_FIND_LE  0x04
#define AVLTREE_FUNC_INSERT   0x08
#define AVLTREE_FUNC_ASSERT   0x10

#define AVLTREE_FUNC_ALL      0x1f

/** avltree_node structure */
typedef struct avlnode {
	struct avlnode *parent, *left, *right;
	u8 height;
} avlnode;


/** avltree_tree structure */
typedef struct avltree {
	struct avlnode *root;
} avltree;


static inline void avltree_init(struct avltree *tree)
{ tree->root = NULL; }

static inline int avltree_empty(struct avltree *tree)
{ return tree->root == NULL; }

static inline struct avlnode *avltree_head(struct avltree *tree)
{
    struct avlnode *node;
    if (!(node = tree->root)) return NULL;
    while (node->left) node = node->left;
    return node;
}

static inline struct avlnode *avltree_tail(struct avltree *tree)
{
    struct avlnode *node;
    if (!(node = tree->root)) return NULL;
    while (node->right) node = node->right;
    return node;
}

static inline struct avlnode *avltree_prev(struct avlnode *node)
{
    struct avlnode *p;
	if (node->left) for (node = node->left; node->right; node = node->right);
	else while ((node = (p = node)->parent) && node->left == p);
	return node;
}

static inline struct avlnode *avltree_next(struct avlnode *node)
{
    struct avlnode *p;
	if (node->right) for (node = node->right; node->left; node = node->left);
	else while ((node = (p = node)->parent) && node->right == p);
	return node;
}

void avltree_node_rebalance(struct avltree *tree, struct avlnode *node);

void avltree_add_head(struct avltree *tree, struct avlnode *node);

void avltree_add_tail(struct avltree *tree, struct avlnode *node);

void avltree_insert_before(struct avltree *tree, struct avlnode *node, struct avlnode *before);

void avltree_insert_after(struct avltree *tree, struct avlnode *node, struct avlnode *after);

void avltree_remove(struct avltree *tree, struct avlnode *node);

void avltree_remove_head(struct avltree *tree);

void avltree_remove_tail(struct avltree *tree);

/*struct avlnode *avltree_common_parent(struct avlnode *node1, struct avlnode *node2);

int avltree_order_compare(struct avlnode *node1, struct avlnode *node2);

void avltree_order_sort(uns n, struct avlnode **nodes);*/

#define AVLTREE_WALK(tree,n) for (n=(void*)avltree_head(&(tree));n;n=(void*)avltree_next((void*)n))
#define AVLTREE_WALK_DELSAFE(tree,n,tmp) for (n=(void *)avltree_head(&(tree));n&&((tmp=(void*)avltree_next((void *)n))||1);n=(void *)tmp)

#endif

