/*
 *
 *   This is not a normal header file, it's a generator of AVL tree
 *   routines. Each time you include it with parameters set in the
 *   corresponding preprocesor macros, it generates routines code
 *   with the parameters given.
 *
 *   Supported parameters:
 *
 *   AVLTREE_PREFIX(x)     - add a name prefix
 *                           (used on all generated global names)
 *   AVLTREE_FUNC          - add keywords to generated functions
 *   AVLTREE_TYPE          - data type of key entry
 *   AVLTREE_KEY(x)        - returns key of node x
 *   AVLTREE_NODUP         - if defined, no duplicite keys are allowed
 *   AVLTREE_LT(x,y)       - true, if key x is lower than key y
 *   AVLTREE_CMP(x,y)      - compare result as integer value
 *   AVLTREE_EXTRA_ARGS    - extra arguments for generated functions
 *                           (trey are always visible in all macros above)
 */

/*
 * TODO:
 *   zprovoznit EXTRA_ARGS v assertu
 */

#ifndef AVLTREE_PREFIX
#define AVLTREE_PREFIX(x) avltree_##x
#endif

#ifndef AVLTREE_FUNC
#define AVLTREE_FUNC AVLTREE_FUNC_ALL
#endif

#ifndef AVLTREE_FUNC_MOD
#define AVLTREE_FUNC_MOD static
#endif

#ifndef AVLTREE_TYPE
#define AVLTREE_TYPE (struct avlnode *)
#endif

#ifndef AVLTREE_KEY
#define AVLTREE_KEY(x) (x)
#endif

#ifndef AVLTREE_LT
#ifndef AVLTREE_CMP
#define AVLTREE_LT(x,y) ((x)<(y))
#else
#define AVLTREE_LT(x,y) (AVLTREE_CMP((x),(y))<0)
#endif
#endif

#ifndef AVLTREE_EXTRA_ARGS
#define AVLTREE_EXTRA_ARGS
#endif

/* Returns first node in tree with equal key or NULL, if no such node exists. */

#if AVLTREE_FUNC & AVLTREE_FUNC_FIND
AVLTREE_FUNC_MOD struct avlnode *AVLTREE_PREFIX(find)
	(struct avltree *tree, AVLTREE_TYPE key AVLTREE_EXTRA_ARGS)
{
	struct avlnode *node, *next;
#ifdef AVLTREE_CMP
	int cmp;
#endif
	ASSERT(tree);
	node = tree->root;
	while (1)
	{
		if (!node)
			return NULL;
#ifdef AVLTREE_CMP
		if ((cmp = AVLTREE_CMP(key, AVLTREE_KEY(node))) < 0)
			node = node->left;
		else if (cmp > 0)
			node = node->right;
#else
		if (AVLTREE_LT(key, AVLTREE_KEY(node)))
			node = node->left;
		else if (AVLTREE_LT(AVLTREE_KEY(node), key))
			node = node->right;
#endif
		else
			break;
	}
#ifndef AVLTREE_NODUP
	while ((next = node->left) && !AVLTREE_LT(AVLTREE_KEY(next), key))
		node = next;
#endif
	return node;
}
#endif

/* Returns first node in tree with greater or equal key or NULL, if no such node exists. */

#if AVLTREE_FUNC & AVLTREE_FUNC_FIND_GE
AVLTREE_FUNC_MOD struct avlnode *AVLTREE_PREFIX(find_ge)
	(struct avltree *tree, AVLTREE_TYPE key AVLTREE_EXTRA_ARGS)
{
	struct avlnode *node, *next, *save = NULL;
#ifdef AVLTREE_CMP
	int cmp;
#endif
	ASSERT(tree);
	node = tree->root;
	while (1)
	{
		if (!node)
			return save;
#ifdef AVLTREE_CMP
		if ((cmp = AVLTREE_CMP(key, AVLTREE_KEY(node))) < 0)
			node = (save = node)->left;
		else if (cmp > 0)
			node = node->right;
#else
		if (AVLTREE_LT(key, AVLTREE_KEY(node)))
			node = (save = node)->left;
		else if (AVLTREE_LT(AVLTREE_KEY(node), key))
			node = node->right;
#endif
		else
			break;
	}
#ifndef AVLTREE_NODUP
	while ((next = node->left) && !AVLTREE_LT(AVLTREE_KEY(next), key))
		node = next;
#endif
	return node;
}
#endif

/* Returns last node in tree with lower or equal key or NULL, if no such node exists. */

#if AVLTREE_FUNC & AVLTREE_FUNC_FIND_LE
AVLTREE_FUNC_MOD struct avlnode *AVLTREE_PREFIX(find_le)
	(struct avltree *tree, AVLTREE_TYPE key AVLTREE_EXTRA_ARGS)
{
	struct avlnode *node, *next, *save = NULL;
#ifdef AVLTREE_CMP
	int cmp;
#endif
	ASSERT(tree);
	node = tree->root;
	while (1)
	{
		if (!node)
			return save;
#ifdef AVLTREE_CMP
		if ((cmp = AVLTREE_CMP(key, AVLTREE_KEY(node))) < 0)
			node = node->left;
		else if (cmp > 0)
			node = (save = node)->right;
#else
		if (AVLTREE_LT(key, AVLTREE_KEY(node)))
			node = node->left;
		else if (AVLTREE_LT(AVLTREE_KEY(node), key))
			node = (save = node)->right;
#endif
		else
			break;
	}
#ifndef AVLTREE_NODUP
	while ((next = node->right) && !AVLTREE_LT(key, AVLTREE_KEY(next)))
		node = next;
#endif
	return node;
}
#endif

/* avltree_insert */

#if AVLTREE_FUNC & AVLTREE_FUNC_INSERT
AVLTREE_FUNC_MOD void AVLTREE_PREFIX(insert)
	(struct avltree *tree, struct avlnode *node AVLTREE_EXTRA_ARGS)
{
	struct avlnode *cur, **next;
	ASSERT(tree && node);

	node->left = node->right = NULL;
	node->height = 1;
	if (!tree->root)
	{
		node->parent = NULL;
		tree->root = node;
		return;
	}
	next = &tree->root;
	do
	{
		cur = *next;
		if (AVLTREE_LT(AVLTREE_KEY(node), AVLTREE_KEY(cur)))
			next = &cur->left;
		else
			next = &cur->right;
	}
	while (*next);
	node->parent = cur;
	*next = node;
	avltree_node_rebalance(tree, cur);
}
#endif

/* avltree_assertnode */

#if AVLTREE_FUNC & AVLTREE_FUNC_ASSERT
AVLTREE_FUNC_MOD void AVLTREE_PREFIX(assertnode)
	(struct avlnode *node, struct avlnode *parent, int level)
{
	int lh, rh;
	if (!node)
		return;
	/* test tree height limit */
	ASSERT(level < AVLTREE_MAXHEIGHT);
	/* test parent pointer */
	ASSERT(parent == node->parent);
	/* left and right subtrees could not be identical */
	ASSERT(!node->left || !node->right || node->left != node->right);
	/* test for valid height */
	lh = node->left ? node->left->height : 0;
	rh = node->right ? node->right->height : 0;
	ASSERT(node->height == MAX(lh, rh) + 1);
#ifdef AVLTREE_NODUP
	/* test for (left key < node key < right key)*/
	if (node->left)
		ASSERT(AVLTREE_LT(AVLTREE_KEY(node->left), AVLTREE_KEY(node)));
	if (node->right)
		ASSERT(AVLTREE_LT(AVLTREE_KEY(node), AVLTREE_KEY(node->right)));
#else
	/* test for (left key <= node key <= right key) */
	if (node->left)
		ASSERT(!AVLTREE_LT(AVLTREE_KEY(node), AVLTREE_KEY(node->left)));
	if (node->right)
		ASSERT(!AVLTREE_LT(AVLTREE_KEY(node->right), AVLTREE_KEY(node)));
#endif
	/* test subtrees */
	AVLTREE_PREFIX(assertnode)(node->left, node, level + 1);
	AVLTREE_PREFIX(assertnode)(node->right, node, level + 1);
}

/* avltree_assert */

AVLTREE_FUNC_MOD void AVLTREE_PREFIX(assert) (struct avltree *tree)
{
	ASSERT(tree);
	AVLTREE_PREFIX(assertnode)(tree->root, NULL, 0);
}
#endif

#undef AVLTREE_PREFIX
#undef AVLTREE_FUNC
#undef AVLTREE_FUNC_MOD
#undef AVLTREE_TYPE
#undef AVLTREE_KEY
#undef AVLTREE_NODUP
#undef AVLTREE_EQ
#undef AVLTREE_LT

