/* FIXME: Dabelska implementace B-Tree... under construction :-) */

#ifndef BTREE_PREFIX
#define BTREE_PREFIX(x) btree_##x
#endif

#define BTREE_(x) BTREE_PREFIX(x)

/* btree_item_t */

#ifndef BTREE_ITEM_T
#define BTREE_ITEM_T struct BTREE_(item_t)
BTREE_ITEM_T { BTREE_ITEM_T *child, *next, *mixed; };
static inline BTREE_ITEM_T *BTREE_(item_child) (BTREE_ITEM_T *item) { return item->child; }
static inline BTREE_ITEM_T *BTREE_(item_next) (BTREE_ITEM_T *item) { return item->next; }
static inline BTREE_ITEM_T *BTREE_(item_mixed) (BTREE_ITEM_T *item) { return item->mixed; }
static inline uns BTREE_(item_flags) (BTREE_ITEM_T *node) { return item->flags; }
static inline void BTREE_(item_child_set) (BTREE_ITEM_T *item, BTREE_ITEM_T *value) { item->child = value; }
static inline void BTREE_(item_next_set) (BTREE_ITEM_T *item, BTREE_ITEM_T *value) { item->next = value; }
static inline void BTREE_(item_mixed_set) (BTREE_ITEM_T *item, BTREE_ITEM_T *value) { item->mixed = value; }
static inline void BTREE_(item_flags_set) (BTREE_ITEM_T *item, uns value)  { item->flags = value; }
#endif
typedef BTREE_ITEM_T BTREE_(item_t);

#define BTREE_FLAGS_PARENT      0x01
#define BTREE_FLAGS_LAST        0x02
#define BTREE_FLAGS_PARENT_LAST 0x04
#define BTREE_FLAGS_ROOT        0x08
#define BTREE_FLAGS_FIRST       0x10

/* btree_t */

#ifndef BTREE_T
#define BTREE_T struct BTREE_(t)
BTREE_T { BTREE_ITEM_T *root; };
static inline BTREE_ITEM_T *BTREE_(root) (BTREE_T *tree) { return tree->root; }
static inline void BTREE_(root_set) (BTREE_T *tree, BTREE_ITEM_T *value) { tree->root = value; }
#endif
typedef BTREE_T BTREE_(t);

/* btree_item_parent */

static inline BTREE_ITEM_T *BTREE_(item_parent) (BTREE_ITEM_T *item)
{
#ifdef BTREE_MIXED_PREV
	uns flags = BTREE_(item_flags)(item);
	if (flags & BTREE_FLAGS_PARENT);
	else if (flags & BTREE_FLAGS_LAST) {
		item = BTREE_(item_mixed)(item);
		if (BTREE_(item_flags)(item) & BTREE_FLAGS_PARENT == 0)
			item = BTREE_(item_next(item))
	}
	else
		item = BTREE_(item_next)(item);
#endif
	return BTREE_(item_mixed)(item);
}

/* btree_item_prev */

static inline BTREE_ITEM_T *BTREE_(item_prev) (BTREE_ITEM_T *item)
{
	uns flags = BTREE_(item_flags)(item);
#ifdef BTREE_MIXED_PREV
	if (flags & BTREE_FLAGS_PARENT)
		return BTREE_(item_mixed)(BTREE_(item_next)(item));
	else {
		item = BTREE_(item_mixed)(item);
		if (flags & BTREE_FLAGS_PARENT)
			return BTREE_(item_next)(item);
		else
			return item;
	}
#else
	BTREE_ITEM_T *p, *q;
	if (flags & BTREE_FLAGS_PARENT_LAST)
		p = BTREE_(item_next)(BTREE_(item_mixed)(item));
	else
		p = BTREE_(item_child)(BTREE_(item_mixed)(item));
	while ((q = BTREE_(item_next)(p)) != item)
		p = q;
	return p;
#endif
}

/* btree_node_last */

static inline BTREE_ITEM_T *BTREE_(node_last) (BTREE_ITEM_T *node)
{
#ifdef BTREE_MIXED_PREV
	uns flags = BTREE_(item_flags)(node);
	if (flags & BTREE_FLAGS_LAST)
		return node;
	if (flags & BTREE_FLAGS_PARENT)
		node = BTREE_(item_next)(node);
	return BTREE_(item_mixed)(node);
#else
	while (BTREE_(item_flags)(node) & BTREE_FLAGS_LAST == 0)
		node = BTREE_(item_next)(node);
#endif
}

/* btree_subtree_head */

static inline BTREE_ITEM_T *BTREE_(subtree_head) (BTREE_ITEM_T *node)
{
	BTREE_ITEM_T *p;
	while (p = BTREE_(node_child)(node))
		node = p;
	return node;
}

/* btree_subtree_tail */

static inline BTREE_ITEM_T *BTREE_(subtree_tail) (BTREE_ITEM_T *node)
{
	BTREE_ITEM_T *p;
	while (node = BTREE_(node_next)(p = BTREE_(node_last)(node)));
	return p;
}

/* btree_head */

static inline BTREE_ITEM_T *BTREE_(head) (BTREE_T *tree)
{
	BTREE_ITEM_T *p;
	return (p = BTREE_(root)(tree)) ? BTREE_(subtree_head)(p) : NULL;
}

/* btree_tail */

static inline BTREE_ITEM_T *BTREE_(tail) (BTREE_T *tree)
{
	BTREE_ITEM_T *p;
	return (p = BTREE_(root)(tree)) ? BTREE_(subtree_tail)(p) : NULL;
}

/* btree_next */

static inline BTREE_ITEM_T *BTREE_(next) (BTREE_ITEM_T *item)
{
	BTREE_ITEM_T *p;
	uns flags;
	p = BTREE_(item_next)(item);
	if ((flags = BTREE_(item_flags)(item)) & BTREE_FLAGS_LAST == 0 || p)
		while (p) p = BTREE_(item_child)(item = p);
	else {
		while (flags & BTREE_FLAGS_PARENT_LAST)
			flags = BTREE_(item_flags)(node = BTREE_(item_parent)(node));
		if (flags & BTREE_FLAGS_ROOT)
			return NULL;
	}
	return item;
}

/* btree_prev */

static inline BTREE_ITEM_T *BTREE_(prev) (BTREE_ITEM_T *item)
{
	BTREE_ITEM_T *p;
	uns flags;
	p = BTREE_(item_child)(item);
	if (p)
		return BTREE_(subtree_tail)(p);
	while ((flags = BTREE_(item_flags)(item)) & BTREE_FLAGS_FIRST) {
		if (flags & BTREE_FLAGS_ROOT)
			return NULL;
		item = BTREE_(item_parent)(item);
	}
	return BTREE_(item_prev)(item);
}

/* btree_node_merge */

// FIXME
/*static inline void BTREE_(node_merge) (BTREE_NODE_T *left_node)
{
	BTREE_NODE_T *parent_item, *right_node;
	parent_item = BTREE_(item_parent)(left_node);
}*/


#undef BTREE_
#undef BTREE_PREFIX
#undef BTREE_NODE_T
#undef BTREE_T
#undef BTREE_MIXED_PREV

