#include "lib/lib.h"
#include "lib/cache.h"
#include <stdlib.h>

#define EXAMPLE_COUNT 1000
#define EXAMPLE_SIZE 1000

struct example {
	struct cache_node node;
	uns index, refcount;
};
struct example *example_table[EXAMPLE_COUNT];

static void excache_callback(struct example *);
static struct cache_user excache_user;

#define CACHE_PREFIX(x) excache_##x
#define CACHE_TYPE struct example
#define CACHE_NODE_SIZE(n) sizeof(*n)
#include "lib/gcache.h"

static void excache_callback(struct example *node)
{
	msg(" [Force remove %03d]", node->index);
	excache_node_dec(node);
	example_table[node->index] = NULL;
	xfree(node);
};

static struct example *example_ref(uns index)
{
	struct example *node = example_table[index];

	if (node) {
		if (node->refcount == 0) {
			msg(" [Remove %03d]", index);
			excache_remove(node);
		}
		node->refcount++;
	}
	else {
		node = xmalloc(sizeof(struct example));
		node->refcount = 1;
		node->index = index;
	}
	return node;
}

static void example_unref(struct example *node)
{
	node->refcount--;
	if (node->refcount == 0) {
		msg(" [Insert %03d]", node->index);
		excache_add(node);
	}
}

int main(void)
{
	uns i, j;
	struct example *e;
	
	msg("Initializing cache structure.\n");
	cache_init(&global_cache);

	msg("Initializing cache user.\n");
	excache_init();

	msg("Setting limits.\n");
	cache_set_limits(&global_cache, 800, 0, CACHE_UNLIMITED_COUNT_MAX);

	for (i = 0; i < 100; i++) {
		j = rand() % EXAMPLE_COUNT;
		msg("Access to %03d", j);
		e = example_ref(j);
		ASSERT(e->index == j);
		example_unref(e);
		msg("\n");
		if (i == 50) {
			msg("Changing limits.");
			cache_set_limits(&global_cache, 500, 0, CACHE_UNLIMITED_COUNT_MAX);
			msg("\n");
		}
	}
	
	msg("Destroying cache structure.");
	cache_destroy(&global_cache);
	msg("\n");

	msg("All done, see you soon...\n");
	return 0;
}

