#include "lib/lib.h"

typedef struct splay_node {
	struct splay_node *left, *right, *back;
} splay_node;

typedef splay_node *splay_tree;

#define SPLAY_INIT(s) do{ (s) = NULL; } while(0)

#define SPLAY_FIND(s,n,cmp) do { SPLAY_SPLAY((s),(n),cmp); } while(0)
#define SPLAY_GET(s) (s)

#define SPLAY_MIN(s) do { SPLAY_SPLAY((s),NULL,-1); } while(0)
#define SPLAY_MAX(s) do { SPLAY_SPLAY((s),NULL,1); } while(0)

#define SPLAY_INSERT(s,nd,cmp) do {				\
	splay_node *_n = (nd);					\
								\
	if (!(s)) {						\
		(s) = _n;					\
		_n->left = _n->right = _n->back = NULL;		\
	}							\
	else {							\
		SPLAY_SPLAY((s),_n,cmp);			\
		if (cmp((s), _n)<0) {				\
			_n->back = NULL;			\
			_n->left = (s)->left;			\
			_n->right = (s);			\
			(s)->left = NULL;			\
			(s)->back = _n;				\
			(s) = _n;				\
		}						\
		else if (cmp((s), _n)>0) {			\
			_n->back = NULL;			\
			_n->right = (s)->right;			\
			_n->left = (s);				\
			(s)->right = NULL;			\
			(s)->back = _n;				\
			(s) = _n;				\
		}						\
	}							\
} while(0)

#define SPLAY_DELETE(s,n,cmp) do {					\
	if (s) {							\
		SPLAY_SPLAY((s),(n),cmp);				\
		if (cmp((s),(n)) == 0) {				\
			if (!(s)->left) {				\
				(s) = (s)->right;			\
				(s)->back = NULL;			\
			}						\
			else {						\
				SPLAY_SPLAY((s)->left, (n), cmp);	\
				(s)->left->right = (s)->right;		\
				(s) = (s)->left;			\
				(s)->right->back = (s);			\
			}						\
		}							\
	}								\
} while(0)

static inline void splay_rot_right(splay_node *t)
{
	splay_node *l = t->left;

	if (t->back) {
		if (t->back->left == t)
			t->back->left = l;
		else if (t->back->right == t)
			t->back->right = l;
		else BUG();
	}
	t->left = l->right;
	if (t->left)
		t->left->back = l;
	l->right = t;
	l->back = t->back;
	t->back = l;
}

static inline void splay_rot_left(splay_node *t)
{
	splay_node *r = t->right;

	if (t->back) {
		if (t->back->left == t)
			t->back->left = r;
		else if (t->back->right == t)
			t->back->right = r;
		else BUG();
	}
	t->right = r->left;
	if (t->right)
		t->right->back = r;
	r->left = t;
	r->back = t->back;
	t->back = r;
}

static inline void splay_rot_double(splay_node *p)
{
	splay_node *f = p->back, *gf = p->back->back;

	if (gf->back) {
		if (gf->back->left == gf)
			gf->back->left = p;
		else
			gf->back->right = p;
	}
	p->back = gf->back;
	if (f == gf->left) {
		f->right = p->left;
		if (p->left)
			p->left->back = f;
		gf->left = p->right;
		if (p->right)
			p->right->back = gf;
		p->left = f;
		p->right = gf;
		f->back = gf->back = p;
	}
	else if (f == gf->right) {
		f->left = p->right;
		if (p->right)
			p->right->back = f;
		gf->right = p->left;
		if (p->left)
			p->left->back = gf;
		p->left = gf;
		p->right = f;
		f->back = gf->back = p;
	}
	else BUG();
}

#define SPLAY_SPLAY(s,nd,cmp) do {									\
	splay_node *__p = (s), *__n = (nd), *_b, *_bb;							\
	int r;												\
													\
	while ((r=cmp(__p,__n)) != 0) {									\
		if (r>0) {										\
			if (!__p->right)								\
				break;									\
			__p = __p->right;								\
		}											\
		if (r<0) {										\
			if (!__p->left)									\
				break;									\
			__p = __p->left;								\
		}											\
	}												\
													\
	while (__p->back) {										\
		if (!__p->back->back) {									\
			if (__p->back->left == __p)							\
				splay_rot_right(__p->back);						\
			else 										\
				splay_rot_left(__p->back);						\
		}											\
		else {											\
			_b = __p->back;									\
			_bb = __p->back->back;								\
			if (__p == _b->left && _b == _bb->left) {		\
				splay_rot_right(_bb);					\
				splay_rot_right(_b);						\
			}										\
			else if (__p == __p->back->right && __p->back == __p->back->back->right) {	\
				splay_rot_left(_bb);					\
				splay_rot_left(_b);						\
			}										\
			else										\
				splay_rot_double(__p);							\
		}											\
	}												\
	(s) = __p;											\
} while(0)
