#define LOCAL_DEBUG

#include "lib/lib.h"
#include "geomlib/fpath.h"
#include "kernel/kernel.h"
#include "kernel/plugin.h"
#include <stdlib.h>

/* Prototypes */

u32 plugin_start(void);
void tsunami(union plugin_arg *);
void hilbert_curve(union plugin_arg *);
void random_pastes(union plugin_arg *);
void show_horizontal_points(union plugin_arg *);
void test_solve_y(union plugin_arg *);
union plugin_arg plus(union plugin_arg *arg);


u32 plugin_start(void)
{
	plugin_function_register("tsunami", NULL, PLUGIN_T_VOID, 2, PLUGIN_T_REAL, PLUGIN_T_REAL);
	plugin_function_register("hilbert_curve", NULL, PLUGIN_T_VOID, 2, PLUGIN_T_OBJ_TLO, PLUGIN_T_INT);
	plugin_function_register("random_pastes", NULL, PLUGIN_T_VOID, 2, PLUGIN_T_OBJ_TLO, PLUGIN_T_INT);
	plugin_function_register("show_horizontal_points", NULL, PLUGIN_T_VOID, 1, PLUGIN_T_OBJ_TLO);
	plugin_function_register("test_solve_y", NULL, PLUGIN_T_VOID, 2, PLUGIN_T_OBJ_TLO, PLUGIN_T_REAL);
	plugin_function_register("plus", "adds two integer numbers", PLUGIN_T_INT, 2, PLUGIN_T_INT, PLUGIN_T_INT);

	return 0;
}

union plugin_arg plus(union plugin_arg *arg)
{
	return ((union plugin_arg){.t_int = arg[0].t_int + arg[1].t_int});
}

static void random_vect(struct geom_vector *v)
{
	v->dx = ((double)(rand() % 1000)) / 500 - 1;
	v->dy = ((double)(rand() % 1000)) / 500 - 1;
}

//void tsunami(real step, real amp)
void tsunami(union plugin_arg *arg)
{
	UNUSED real step = arg[0].t_real;
	UNUSED real amp = arg[1].t_real;

	UNUSED uns count, i, j;
	struct obj_tlo *tlo;
	char err_msg_main[TRANS_ERR_SIZE];
	struct go *go;
	const struct geom_path *path;
	UNUSED struct geom_bezier *bezier, *b;
	struct geom_vector v;
	UNUSED double len, *t, *s;
	struct hanger *h[4];
	struct go_bezier *go_bez;
	struct geom_fpath bez_list;

	// FIXME: Osklive, funkce by mela dostavat tlo jako argument. Nebo by mela dostavat i grupu.
	tlo = obj2tlo(obj_get_first_son(obj_get_first_son(obj_universum)));

	GEOM_INSTANCE_CREATE(&bez_list, fpath, geom_fpath_create);
	TRANS_BEGIN_MAIN(tlo, "Tsunami", err_msg_main)

		// FIXME: osklive zachazeni s geomlibovskyma instancema
		DBG("Walking old objects.\n");
		expand_ref(tlo);
		random_vect(&v);
		WALK_GO_RECURSIVE(&tlo->tlg->group.go, go) {
			DBG("Walking GO at %p\n", go);
			if (go_get_info(go)->expand != ET_PATH || !(path = go_get_expand_path(tlo, go)))
				continue;
			GEOM_GROUP_WALK((void *)path, bezier) {
				len = geom_bezier_alength(bezier);
				count = (uns)(len / step) + 1;
				DBG("Expanding bezier to %d parts.\n", count);
				t = xmalloc((count + 1) * sizeof(double));
				s = xmalloc((count + 1) * sizeof(double));
				for (i = 0; i <= count; i++)
					t[i] = len * (((double)i) / count);
				for (i = 0; i <= count; i++)
					s[i] = geom_bezier_atime_to_time(bezier, t[i]);
				for (i = 0; i < count; i++)	{
					GEOM_FPATH_APPEND_NEW_CREATE(b, &bez_list, bezier, geom_bezier_create);
					geom_bezier_part(bezier, s[i], s[i + 1], b);
					geom_bezier_elevate(b, GEOM_BEZIER_CUBIC, b);
					for (j = 0; j <= 3; j++)
						b->pt[j].w = 1;
					GEOM_INSTANCE_FLAGS(b) |= GEOM_BEZIER_NONRATIONAL;
					geom_bezier_after_change(b);
					for (j = 0; j < 4; j++) {
						b->pt[j].x += v.dx * amp;
						b->pt[j].y += v.dy * amp;
						if (j == 1)
							random_vect(&v);
					}
				}
				xfree(t);
				xfree(s);
			} 
		}
		DBG("Removing old objects.\n");
		tlo_unlink_all_go();
		geom_fpath_finish(&bez_list);
		GEOM_GROUP_WALK(&bez_list, b) {
			for (j = 0; j < 4; j++)
				h[j] = hanger_new_mc_xy(b->pt[j].x, b->pt[j].y);
			go_bez = go_bezier_new_cubic(h[0], 1, h[1], 1, h[2], 1, h[3], 1);
			go_link(&go_bez->go, &tlo->tlg->group, NULL);
		}
		expand_unref(tlo);

	TRANS_FAILED
	TRANS_END
	GEOM_INSTANCE_DESTROY(&bez_list);
}



/* HILBERT CURVE */

static struct geom_point hilbert_curve_prev;

static void hilbert_curve_step(real x0, real y0, real xis, real xjs, real yis, real yjs, int n)
{
	if (n <= 0) {
		struct geom_point point;

		point.x = x0 + 0.5 * (xis + yis);
		point.y = y0 + 0.5 * (xjs + yjs);
		if (likely(isfinite(hilbert_curve_prev.x)))
			go_link(&go_segment_new(hanger_new_mc(&hilbert_curve_prev), hanger_new_mc(&point))->go, &trans_tlo->tlg->group, NULL);
		hilbert_curve_prev = point;

	}
	else {
		hilbert_curve_step(x0, y0, yis/2, yjs/2, xis/2, xjs/2, n-1);
		hilbert_curve_step(x0+xis/2, y0+xjs/2 ,xis/2, xjs/2, yis/2, yjs/2, n-1);
		hilbert_curve_step(x0+xis/2+yis/2, y0+xjs/2+yjs/2, xis/2, xjs/2, yis/2, yjs/2,n-1);
		hilbert_curve_step(x0+xis/2+yis, y0+xjs/2+yjs, -yis/2,-yjs/2, -xis/2, -xjs/2,n-1);
	}
}

//void hilbert_curve(struct obj_tlo *tlo, int n)
void hilbert_curve(union plugin_arg *arg)
{
	struct obj_tlo *tlo = arg[0].ptr;
	int n = arg[1].t_int;

	char err_msg_main[TRANS_ERR_SIZE];

	TRANS_BEGIN_MAIN(tlo, "Hilbert curve", err_msg_main);

	if (n > 8)
		trans_fail("Too large N.");

	hilbert_curve_prev.x = INFINITY;
	hilbert_curve_step(-50, -50, 100, 0, 0, 100, n);

	TRANS_FAILED;
	msg("Hilbert curve failed: %s\n", err_msg_main);
	TRANS_END;
}

/* RANDOM PASTES */

UNUSED static real random_number(real min, real max, real dist UNUSED)
{
	int i = rand() % 1000000000;
	real r = min + i * 0.000000001 * (max - min);
	if (!GEOM_RTEST(r))
		trans_fail("Numeric error.");
	return r;
}

//void random_pastes(struct obj_tlo *tlo, int n)
void random_pastes(union plugin_arg *arg UNUSED)
{
// FIXME: Clipboard isn't implemented yet.
#if 0
	struct obj_tlo *tlo = arg[0].ptr;
	int n = arg[1].t_int;

	char err_msg_main[TRANS_ERR_SIZE];
	real move_x, move_y, scale, rotation;
	struct geom_transform2 t, t2;
	int i;

	TRANS_BEGIN_MAIN(tlo, "Random pastes", err_msg_main);

	for (i = 0; i < n; i++) {
		tlo_clipboard_paste();

		move_x = random_number(-100, 100, 0);
		move_y = random_number(-100, 100, 0);
		scale = random_number(0.5, 2, 0);
		rotation = random_number(0, R_2PI, 0);

		if (!(
			GEOM_ETEST(geom_transform2_set_scale(&t, scale, scale)) &&
			GEOM_ETEST(geom_transform2_set_move(&t2, move_x, move_y)) &&
			GEOM_ETEST(geom_transform2_merge(&t, &t2, &t)) &&
			GEOM_ETEST(geom_transform2_set_rotation(&t2, rotation)) &&
			GEOM_ETEST(geom_transform2_merge(&t, &t2, &t))
			))
			trans_fail("Numeric error.");

		tlo_transform_selected_go(tlo, &t);
	}

	TRANS_FAILED;
	msg("Random pastes failed: %s\n", err_msg_main);
	TRANS_END;
#endif
}


/* SHOW HORIZONTAL POINTS */

//void show_horizontal_points(struct obj_tlo *tlo)
void show_horizontal_points(union plugin_arg *arg)
{
	struct obj_tlo *tlo = arg[0].ptr;

	char err_msg_main[TRANS_ERR_SIZE];
	struct go *go;
//	uns type;
//	int i, n;
	struct garr *garr;
//	struct geom_point point;

	garr = garr_new(64, sizeof(real));
	TRANS_BEGIN_MAIN(tlo, "Show horizontal points", err_msg_main);
	// FIXME: Funkce by mela dostavat grupy, protoze podle grupy se rozhoduje, co je zaselektene.
	WALK_GROUP_SELECTED_GO(&tlo->tlg->group, go) {

// FIXME: az se implementuje propojeni cestami a grupami
#if 0
		if (type = go_to_curve_type(go)) {
			if (!GEOM_ETEST(n = geom_curve_horizontal_times(type, go_to_curve_ptr(go), garr)))
				trans_fail("Numeric error.");
			for (i = 0; i < n; i++) {
				go_point_at_time(go, *(real *)garr_get(garr, i), &point);
				go_link(&(go_point_new(0, hanger_new_mouse_click(&point)))->go);
			}
		}
#endif
	}
	TRANS_FAILED;
	msg("Transaction failed: %s\n", err_msg_main);
	TRANS_END;
	garr_destroy(garr);
}


/* TEST SOLVE Y */

//void test_solve_y(struct obj_tlo *tlo, real y)
void test_solve_y(union plugin_arg *arg)
{
	struct obj_tlo *tlo = arg[0].ptr;
//	real y = arg[1].t_real;

	char err_msg_main[TRANS_ERR_SIZE];
	struct go *go;
//	uns type;
//	int n;
//	struct geom_curve_point result;

	TRANS_BEGIN_MAIN(tlo, "Test of solve Y", err_msg_main);
	// FIXME: Funkce by mela dostavat grupy, protoze podle grupy se rozhoduje, co je zaselektene.
	WALK_GROUP_SELECTED_GO(&tlo->tlg->group, go) {
// FIXME: az se implementuje propojeni cestami a grupami
#if 0
		if (type = go_to_curve_type(go)) {
			n = geom_curve_solve_y(type, go_to_curve_ptr(go), 0, 1, y, GEOM_CURVE_POINT_ALL, &result);
			if (!GEOM_ETEST(n))
				trans_fail("Numeric error.");
			if (n > 0)
				go_link(&(go_point_new(0, hanger_new_mouse_click(&result.point)))->go);
		}
#endif
	}
	TRANS_FAILED;
	msg("Transaction failed: %s\n", err_msg_main);
	TRANS_END;
}


