#undef LOCAL_DEBUG

#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "import/dvi.h"
#include "kernel/kernel.h"
#include "kernel/font.h"

vdo_vtable *vcl_vt_chargroup;

static inline struct geom_transform *
to_parent_coords (struct vcl_chargroup * obj)
{
  return (&(obj -> coords . primary));
}

static inline struct geom_transform *
from_parent_coords (struct vcl_chargroup * obj)
{
  return (&(obj -> coords . inverted));
}

static inline int
want_redraw (struct vcl_chargroup * obj)
{
  return  (obj -> parent != NULL);
}

static void
recompute_bbox (struct vcl_chargroup * obj)
{
  geom_rectangle_set_empty (&(obj -> lc_bbox));

  /*
  for (i = 0; i < obj -> exp -> tex_glyphs -> size; i ++)
    {
      struct tex_glyph *gi = (struct tex_glyph *)
	garr_get (obj -> exp -> tex_glyphs, i);

      geom_rectangle_merge (&(obj -> lc_bbox), &(gi -> bbox), &(obj -> lc_bbox));
    }
  */
  
  obj -> lc_bbox . ll . x = -200;
  obj -> lc_bbox . ll . y = -200;
  obj -> lc_bbox . ur . x = 200;
  obj -> lc_bbox . ur . y = 200;

  geom_rectangle_transform (&(obj -> lc_bbox), to_parent_coords (obj), &(obj -> pc_bbox));
}

static void
vcl_chargroup_destroy (struct vcl_chargroup * obj)
{
  /* FIXME check placement */
  expansion_unref (obj -> exp);
  if (obj -> parent != NULL) vcl_placement_remove_child (obj -> parent, obj);
  xfree (obj);
}

static void
vcl_chargroup_set_data (struct vcl_chargroup * obj, int data)
{
  obj -> data = data;
}

static int
vcl_chargroup_get_data (struct vcl_chargroup * obj)
{
  return (obj -> data);
}

static void
vcl_chargroup_set_parent (struct vcl_chargroup * obj, void * parent)
{
  obj -> parent = parent;
}

static void *
vcl_chargroup_get_parent (struct vcl_chargroup * obj)
{
  return (obj -> parent);
}

static void
vcl_chargroup_get_bbox (struct vcl_chargroup * obj, struct geom_rectangle * bbox)
{
  *bbox = obj -> pc_bbox;
}

static void
vcl_chargroup_get_center (struct vcl_chargroup * obj, struct geom_point * p)
{
  p -> x = obj -> coords . primary . coef [0][2];
  p -> y = obj -> coords . primary . coef [1][2];
}

static void
vcl_chargroup_transform (struct vcl_chargroup * obj, struct geom_transform2 * trans)
{
  struct geom_rectangle old_bbox = obj -> pc_bbox;
  geom_transform2_merge (&(obj -> coords), trans, &(obj -> coords));

  geom_rectangle_transform (&(obj -> lc_bbox), to_parent_coords (obj), &(obj -> pc_bbox));
  if (want_redraw (obj))
    vcl_container_child_changed (obj -> parent, obj, &old_bbox, &(obj -> pc_bbox));
}

#define SWAP(tmp, a, b) (tmp = a, a = b, b = tmp)

static void
vcl_chargroup_draw (struct vcl_chargroup * obj, void * context,
		   const struct geom_rectangle * wanted_bbox) 
{
  struct vcl_hack * ct = context;
  if (want_redraw (obj) &&
      geom_rectangle_have_intersection (&(obj -> pc_bbox), wanted_bbox))
    switch (obj -> exp -> type)
      {
      case GOT_TEXT:
	{
	  struct render_ctl ctl;
	  // FIXME bug - no free

	  hack_apply_transform (context, &(obj -> coords));

	  ctl . fnt_id =  obj -> font_id;
	  ctl . fnt_size = 5.0;
	  ctl . pix_per_unit = 3;
	  ctl . transform = ct -> gt;
	  ctl . bitmap = NULL;
	  ctl . bsize = 0;
	  
	  if (font_render_string (&ctl, obj -> exp -> text . text -> text) >= 0)
	    {
	      int tx, ty, tmp;
	      DBG ("String render offsets: %d %d\n", ctl . xoff, ctl . yoff);

	      /* swap lines to change orientation of bitmap */
	      for (ty = 0; ty < (ctl . height / 2); ty ++)
		for (tx = 0; tx < ctl . width; tx ++)
		  {
		    SWAP (tmp, ctl . bitmap [tx + ty * ctl . width],
			  ctl . bitmap [tx + (ctl . height - ty - 1) * ctl . width]);
		  }
		  
	      gdk_draw_gray_image (ct -> win, ct -> gc, ctl . xoff, ctl . yoff,
				   ctl . width, ctl . height, GDK_RGB_DITHER_NORMAL,
				   ctl . bitmap, ctl . width);
	    }
	  else DBG("Font rendering failed");

	  hack_unapply_transform (context, &(obj -> coords));
	}
	break;
      case GOT_TEX_TEXT:
	{
	  int px, py, fx, fy;
	  uns i;
	  byte *bmp = NULL;

	  struct geom_rectangle new_bbox, glyph_bbox;
	  struct geom_transform *to_pixels, offset;
	  struct geom_point p;

	DBG ("tttd\n");	  
	  geom_rectangle_transform (wanted_bbox, from_parent_coords (obj), &new_bbox);
	  hack_apply_transform (context, &(obj -> coords));

	  to_pixels = &(ct -> gt);
	  
	  for (i = 0; i < obj -> exp -> tex_text . glyphs_cnt; i ++)
	    {
	      struct tex_glyph *gi = (struct tex_glyph *)
		garr_get (obj -> exp -> tex_text . glyphs, i);
	      
	      if (geom_rectangle_have_intersection (&(gi->bbox), &new_bbox))
		  {
		  /*
		   * FIXME: renderovani a kresleni glyphu
		   */
		    struct render_ctl g;
		    bzero(&g, sizeof(g));
		    g.fnt_id = gi->font_id;
		    g.fnt_size = gi->font_size;
		    g.pix_per_unit = 5;
		    g.transform = geom_transform_identity;
		    font_render_glyph(&g, gi->code);
		    p.x = obj->exp->tex_text.position.x + gi->xpos;
		    p.y = obj->exp->tex_text.position.y + gi->ypos;
		    DBG("%f %f\n", p.x, p.y);
		    gdk_draw_gray_image(ct->win, ct->gc, p.x + 300, p.y + 300, g.width, g.height, GDK_RGB_DITHER_NORMAL, g.bitmap, g.width);
		    xfree(g.bitmap);
#if 0
		    font_set_size (gi -> font_id, gi -> font_size, 3.0); // FIXME
		    font_load_char (gi -> font_id, gi -> code);

		    geom_transform_set_move (&offset, gi -> bbox . ll . x, gi -> bbox . ll . y);
		    font_glyph_transform (gi -> font_id, &offset);
		    font_glyph_transform (gi -> font_id, to_pixels);
		    font_get_glyph_bbox (gi -> font_id, &glyph_bbox);
		    geom_transform_set_move (&offset, - round (glyph_bbox . ll . x),
					     - round (glyph_bbox . ll . y));
		    font_glyph_transform (gi -> font_id, &offset);
		    if ((bmp == NULL) ||
			(px < (glyph_bbox . ur . x - glyph_bbox . ll . x)) ||
			(py < (glyph_bbox . ur . y - glyph_bbox . ll . y)))
		      {
			px = (glyph_bbox . ur . x - glyph_bbox . ll . x);
			py = (glyph_bbox . ur . y - glyph_bbox . ll . y);
			if (bmp != NULL) xfree (bmp);
			bmp = xmalloc (px * py);			
		      }

		    memset (bmp, 0, px * py); 
		    font_render_glyph (gi -> font_id, bmp, px, py, &fx, &fy);
		    geom_point_transform (&(gi -> bbox . ll), to_pixels, &p);
		    gdk_draw_gray_image (ct -> win, ct -> gc, p . x, p . y - py,
					 fx, fy, GDK_RGB_DITHER_NORMAL, bmp, px);
#endif
		  }
	    }

	  hack_unapply_transform (context, &(obj -> coords));
	  hack_draw_bbox (context, &(obj -> pc_bbox));
	}
	break;
      }
}


struct vcl_chargroup *
vcl_chargroup_new (struct expansion * exp, real x_refpnt, real y_refpnt, int font_id)
{
  MALLOCATED (struct vcl_chargroup, obj);
  vdo_object_init (obj, vcl_vt_chargroup);

  geom_transform2_set_move (&(obj -> coords), x_refpnt, y_refpnt);

  obj -> exp = exp;
  obj -> font_id = font_id;
  recompute_bbox (obj);
  return (obj);
}

void
vcl_chargroup_source_changed (struct vcl_chargroup * obj)
{
  struct geom_rectangle old_bbox;
  old_bbox = obj -> pc_bbox;
  recompute_bbox (obj);

  if (want_redraw (obj))
    vcl_container_child_changed (obj -> parent, obj, &old_bbox, &(obj -> pc_bbox));
}

void
vcl_chargroup_change_source (struct vcl_chargroup * obj, struct expansion * exp)
{
  expansion_unref (obj -> exp);
  obj -> exp = exp;
  vcl_chargroup_source_changed (obj);
}

void
vcl_chargroup_main (void)
{
  vcl_vt_chargroup = vdo_class_new ("Chargroup");

  vdo_class_add_interface (vcl_vt_chargroup, vdo_if_object);
  vdo_class_add_method_li (vdo_mid_object_destroy, vcl_chargroup_destroy);

  vdo_class_add_interface (vcl_vt_chargroup, vcl_if_item);
  vdo_class_add_method_li (vcl_mid_item_set_data, vcl_chargroup_set_data);
  vdo_class_add_method_li (vcl_mid_item_get_data, vcl_chargroup_get_data);
  vdo_class_add_method_li (vcl_mid_item_set_parent, vcl_chargroup_set_parent);
  vdo_class_add_method_li (vcl_mid_item_get_parent, vcl_chargroup_get_parent);
  vdo_class_add_method_li (vcl_mid_item_get_bbox, vcl_chargroup_get_bbox);
  vdo_class_add_method_li (vcl_mid_item_get_center, vcl_chargroup_get_center);
  vdo_class_add_method_li (vcl_mid_item_transform, vcl_chargroup_transform);
  vdo_class_add_method_li (vcl_mid_item_draw, vcl_chargroup_draw);
}
