#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vdo_vtable *vcl_vt_grid;

static inline struct geom_transform *
to_parent_coords (struct vcl_grid * obj)
{
  return (&(obj -> coords . primary));
}

static inline struct geom_transform *
from_parent_coords (struct vcl_grid * obj)
{
  return (&(obj -> coords . inverted));
}

static inline int
want_redraw (struct vcl_grid * obj)
{
  return  (obj -> parent != NULL);
}

static void
vcl_grid_destroy (struct vcl_grid * obj)
{
  /* FIXME check placement */
  if (obj -> parent != NULL) vcl_placement_remove_child (obj -> parent, obj);
  xfree (obj);
}

static void
vcl_grid_set_data (struct vcl_grid * obj, int data)
{
  obj -> data = data;
}

static int
vcl_grid_get_data (struct vcl_grid * obj)
{
  return (obj -> data);
}

static void
vcl_grid_set_parent (struct vcl_grid * obj, void * parent)
{
  obj -> parent = parent;
}

static void *
vcl_grid_get_parent (struct vcl_grid * obj)
{
  return (obj -> parent);
}

static void
vcl_grid_get_bbox (struct vcl_grid * obj UNUSED, struct geom_rectangle * bbox)
{
  // FIXME +/- infinity ?
  bbox -> ll . x = -1000000;
  bbox -> ll . y = -1000000;
  bbox -> ur . x = +1000000;
  bbox -> ur . y = +1000000;
}

static void
vcl_grid_get_center (struct vcl_grid * obj, struct geom_point * p)
{
  p -> x = 0; 
  p -> y = 0; 
  geom_point_transform (p, to_parent_coords (obj), p);
}


static void
vcl_grid_transform (struct vcl_grid * obj, struct geom_transform2 * trans)
{
  geom_transform2_merge (&(obj -> coords), trans, &(obj -> coords));
  if (want_redraw (obj))
    {
      struct geom_rectangle bbox; 
      vcl_grid_get_bbox (obj, &bbox);
      vcl_container_child_altered (obj -> parent, obj, &bbox);
    }
}

static double square_distance (struct geom_point *p1, struct geom_point *p2)
{
  double dx = p1 -> x - p2 -> x;
  double dy = p1 -> y - p2 -> y;
  return (dx * dx + dy * dy);
}

static void
vcl_grid_draw (struct vcl_grid * obj, void * context,
		   const struct geom_rectangle * wanted_bbox)
{
  if (want_redraw (obj))
    {
      int draw = 1;
      struct geom_rectangle int_bbox;
      struct geom_point p;

      geom_rectangle_transform (wanted_bbox, from_parent_coords (obj), &int_bbox);

      hack_set_color (context, 0x7FFF, 0x7FFF, 0x7FFF);
      hack_apply_transform (context, &(obj -> coords));      

      p = int_bbox . ll;

      {
	struct geom_point pi = int_bbox . ll;
	struct geom_point po1, po2, po3;
	geom_point_transform (&pi, &(((struct vcl_hack *) context) -> gt), &po1);
	pi . x += 1;
	geom_point_transform (&pi, &(((struct vcl_hack *) context) -> gt), &po2);
	pi = int_bbox . ll;
	pi . y += 1;
	geom_point_transform (&pi, &(((struct vcl_hack *) context) -> gt), &po3);
	if ((square_distance (&po1, &po2) <= 1) || (square_distance (&po1, &po3) <= 1))
	  draw = 0;
      }

      // FIXME check underflow
      if (draw)
	for (p . x = ceil (int_bbox . ll . x); p . x <= int_bbox . ur . x; p . x = p . x + 1)
	  for (p . y = ceil (int_bbox . ll . y); p . y <= int_bbox . ur . y; p . y = p . y + 1)
	    hack_draw_point (context, &p);

      hack_unapply_transform (context, &(obj -> coords));
    }
}

struct vcl_grid *
vcl_grid_new (const struct geom_transform2 * trans)
{
  MALLOCATED (struct vcl_grid, obj);
  vdo_object_init (obj, vcl_vt_grid);
  if (trans == NULL) geom_transform2_set_identity (&(obj -> coords));
  else obj -> coords = *trans;

  return (obj);
}

void
vcl_grid_set_transformation (struct vcl_grid * obj, const struct geom_transform2 * trans)
{
  obj -> coords = *trans;

  if (want_redraw (obj))
    {
      struct geom_rectangle bbox; 
      vcl_grid_get_bbox (obj, &bbox);
      vcl_container_child_altered (obj -> parent, obj, &bbox);
    }
}

void
vcl_grid_main (void)
{
  vcl_vt_grid = vdo_class_new ("Grid");

  vdo_class_add_interface (vcl_vt_grid, vdo_if_object);
  vdo_class_add_method_li (vdo_mid_object_destroy, vcl_grid_destroy);

  vdo_class_add_interface (vcl_vt_grid, vcl_if_item);
  vdo_class_add_method_li (vcl_mid_item_set_data, vcl_grid_set_data);
  vdo_class_add_method_li (vcl_mid_item_get_data, vcl_grid_get_data);
  vdo_class_add_method_li (vcl_mid_item_set_parent, vcl_grid_set_parent);
  vdo_class_add_method_li (vcl_mid_item_get_parent, vcl_grid_get_parent);
  vdo_class_add_method_li (vcl_mid_item_get_bbox, vcl_grid_get_bbox);
  vdo_class_add_method_li (vcl_mid_item_get_center, vcl_grid_get_center);
  vdo_class_add_method_li (vcl_mid_item_transform, vcl_grid_transform);
  vdo_class_add_method_li (vcl_mid_item_draw, vcl_grid_draw);
}
