#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

static void
compute_point (struct geom_transform *gt, GdkPoint *r, xxxnum_t x, xxxnum_t y)
{
  struct geom_point pi, po;
  pi . x = x;
  pi . y = y;
  geom_point_transform (&pi, gt, &po);
  //  r -> x = nearbyint (po . x);
  //  r -> y = nearbyint (po . y);
  if ((po . x > 100000) || (po . y > 100000) ||
      (po . x < -100000) || (po . y < -100000))
    {
      r -> x = r -> y = 100000;
    }
  else
    {
      r -> x = po . x;
      r -> y = po . y;
    }
}

struct vcl_hack *
make_hack_context (GtkWidget *wg)
{
  MALLOCATED (struct vcl_hack, context);
  context -> win = wg -> window;
  context -> gc = gdk_gc_new (context -> win);
  gdk_gc_copy (context -> gc, gtk_widget_get_style (wg) -> fg_gc [0]);
  context -> gc2 = gtk_widget_get_style (wg) -> fg_gc [4];
  context -> gt = geom_transform_identity;
  return (context);
}

void
hack_set_color (struct vcl_hack *context, int r, int g, int b)
{
  GdkColor clr = {0, r, g, b};
  gdk_gc_set_rgb_fg_color (context -> gc, &clr);
}

void
hack_draw_image (struct vcl_hack *context, xxxnum_t x1, xxxnum_t y1, xxxnum_t x2, xxxnum_t y2)
{
  GdkPoint pts [9];
  
  compute_point (&(context -> gt), pts + 0, x1, y1);
  compute_point (&(context -> gt), pts + 1, x2, y2);
  compute_point (&(context -> gt), pts + 2, x1, y2);
  compute_point (&(context -> gt), pts + 3, (x1 + x2) / 2.0, y2 + (y2 - y1));
  compute_point (&(context -> gt), pts + 4, x2, y2);
  compute_point (&(context -> gt), pts + 5, x2, y1);
  compute_point (&(context -> gt), pts + 6, x1, y1);
  compute_point (&(context -> gt), pts + 7, x1, y2);
  compute_point (&(context -> gt), pts + 8, x2, y1);

  gdk_draw_lines (context -> win, context -> gc, pts, 9);
}

void
hack_draw_bbox (struct vcl_hack *context, const struct geom_rectangle *r)
{
  GdkPoint pts [5];

  xxxnum_t x1 = r -> ll . x;
  xxxnum_t y1 = r -> ll . y;
  xxxnum_t x2 = r -> ur . x;
  xxxnum_t y2 = r -> ur . y;

  compute_point (&(context -> gt), pts + 0, x1, y1);
  compute_point (&(context -> gt), pts + 1, x1, y2);
  compute_point (&(context -> gt), pts + 2, x2, y2);
  compute_point (&(context -> gt), pts + 3, x2, y1);
  compute_point (&(context -> gt), pts + 4, x1, y1);

  gdk_draw_lines (context -> win, context -> gc2, pts, 5);
}


static void
intersect_v (struct geom_point *p, double dx, double dy, 
	     double at, double lb, double ub,
	     struct geom_point *ints, double * int_pars, int * int_count)
{
  double par = (at - p -> x) / dx;
  double dst = p -> y + par * dy;
  if ((lb <= dst) && (ub >= dst) && (par >=0) && (par <= 1))
    {
      DBG ("%f %f %f (%f %f %f %f %f)\n",lb, dst, ub, p-> x, p->y, dx, dy, par);
      ints [*int_count] . x = at;
      ints [*int_count] . y = dst;
      int_pars [*int_count] = par;
      (*int_count) ++;
    }
}

static void
intersect_h (struct geom_point *p, double dx, double dy, 
	     double at, double lb, double ub,
	     struct geom_point *ints, double * int_pars, int * int_count)
{
  double par = (at - p -> y) / dy;
  double dst = p -> x + par * dx;
  if ((lb <= dst) && (ub >= dst) && (par >=0) && (par <= 1))
    {
      DBG ("%f %f %f (%f %f %f %f %f)\n",lb, dst, ub, p-> x, p->y, dx, dy, par);
      ints [*int_count] . x = dst;
      ints [*int_count] . y = at;
      int_pars [*int_count] = par;
      (*int_count) ++;
    }
}

static int
geom_segment_crop (struct geom_point *p1, struct geom_point *p2,
		   const struct geom_rectangle *crop,
		   struct geom_point *d1, struct geom_point *d2)
{
  int int_count = 0;
  struct geom_point ints [2];
  double int_pars [2];
  double dx, dy;
  int in1, in2;

  int p1x = (p1 -> x < crop -> ll . x) ? 0 : ((p1 -> x > crop -> ur . x) ? 2 : 1);
  int p1y = (p1 -> y < crop -> ll . y) ? 0 : ((p1 -> y > crop -> ur . y) ? 2 : 1);
  int p2x = (p2 -> x < crop -> ll . x) ? 0 : ((p2 -> x > crop -> ur . x) ? 2 : 1);
  int p2y = (p2 -> y < crop -> ll . y) ? 0 : ((p2 -> y > crop -> ur . y) ? 2 : 1);

  DBG ("DRAW %d %d %d %d\n", p1x, p1y, p2x, p2y);

  if (((p1x == 0) && (p2x == 0)) || ((p1x == 2) && (p2x == 2)) ||
      ((p1y == 0) && (p2y == 0)) || ((p1y == 2) && (p2y == 2)))
    // Everything is cropped
    return (1);

  in1 = (p1x == 1) && (p1y == 1); 
  in2 = (p2x == 1) && (p2y == 1); 

  if (in1 && in2)
    { // Everything in, no crop
      *d1 = *p1;
      *d2 = *p2;
      return (0);
    }

  dx = p2 -> x - p1 -> x;
  dy = p2 -> y - p1 -> y;

  DBG ("DX %f %f %f\n", p1 -> x, p2 -> x, dx);
  DBG ("DY %f %f %f\n", p1 -> y, p2 -> y, dy);

  if (! IS_ZERO (dx))
    {
      intersect_v (p1, dx, dy, crop -> ll . x, crop -> ll . y, crop -> ur . y,
		 ints, int_pars, &int_count);
      intersect_v (p1, dx, dy, crop -> ur . x, crop -> ll . y, crop -> ur . y,
		 ints, int_pars, &int_count);
    }

  if (! IS_ZERO (dy))
    {
      if (int_count < 2)
	intersect_h (p1, dx, dy, crop -> ll . y, crop -> ll . x, crop -> ur . x,
		     ints, int_pars, &int_count);

      if (int_count < 2)
	intersect_h (p1, dx, dy, crop -> ur . y, crop -> ll . x, crop -> ur . x,
		     ints, int_pars, &int_count);
    }

  if (in1 && (! in2))
    { // Leaving rectangle
      DBG ("LEAVE %d (%f %f) (%f %f)\n", int_count, ints [0] . x, ints [0] . y,
	      ints [1] . x, ints [1] . y);
      if (int_count != 1) return (-1);
      *d1 = *p1;
      *d2 = ints [0];
      return (0);
    }

  if ((! in1) && in2)
    { // Entering rectangle
      DBG ("ENTER %d\n", int_count);
      if (int_count != 1) return (-1);
      *d1 = ints [0];
      *d2 = *p2;
      return (0);
    }

  if (int_count == 2)
    { // Full crop
      if (int_pars [0] < int_pars [1])
	{
	  *d1 = ints [0];
	  *d2 = ints [1];
	}
      else
	{
	  *d1 = ints [1];
	  *d2 = ints [0];
	}

      return (0);
    }
  else return ((int_count == 0) ? 1 : -1);
}

static inline int
compute_visible_part (double x1, double y1, double x2, double y2,
		      GdkPoint *dst1, GdkPoint *dst2)
{
  static const struct geom_rectangle cr = {{-500, -500}, {2000, 2000}};
  struct geom_point p1 = {x1, y1};
  struct geom_point p2 = {x2, y2};
  struct geom_point d1, d2;
  int retval = geom_segment_crop (&p1, &p2, &cr, &d1, &d2);
  dst1 -> x = d1 . x;
  dst1 -> y = d1 . y;
  dst2 -> x = d2 . x;
  dst2 -> y = d2 . y;
  return (! retval);
}

static inline void
store_pt (struct vdo_growing_array *point_buff, GdkPoint **dst, GdkPoint *src)
{
  *dst = vdo_ga_allocate (point_buff, 1);
  **dst = *src;
}

void
hack_draw_path (struct vcl_hack *context, const struct geom_path *p)
{
  static double *buff = NULL;
  static int buff_size = 0;
  static struct vdo_growing_array *point_buff = NULL;
  GdkPoint *op = NULL;
  struct geom_bezier dst;
  unsigned int i;

  if (point_buff == NULL) point_buff = vdo_ga_new (16, sizeof (GdkPoint));
  else vdo_ga_flush (point_buff);

  for (i = 0; i < p -> obj_count; i ++)
    {
      int pts, j;
      GdkPoint p1, p2;

      geom_bezier_transform (&(GEOM_PATH_NTH_BEZIER(p, i)), &(context -> gt), &dst);
      geom_bezier_approximate (&dst, 0.1, &buff, &buff_size, &pts);

      for (j = 0; j < (pts - 1); j ++)
	{
	  if (compute_visible_part (buff [j * 2], buff [j * 2 + 1],
				    buff [j * 2 + 2], buff [j * 2 + 3],
				    &p1, &p2))
	    {
	      if ((op != NULL) && ((op -> x != p1 . x) || (op -> y != p1 . y)))
		{
		  gdk_draw_lines (context -> win, context -> gc,
				  point_buff -> array, point_buff -> used);
		  gdk_draw_points (context -> win, context -> gc,
				   point_buff -> array, point_buff -> used);
		  vdo_ga_flush (point_buff);
		  op = NULL;
		}
	      if (op == NULL) store_pt (point_buff, &op, &p1);
	      store_pt (point_buff, &op, &p2);
	      
	    }
	}
    }

  if (op != NULL)
    {
      gdk_draw_lines (context -> win, context -> gc, point_buff -> array, point_buff -> used);
      gdk_draw_points (context -> win, context -> gc, point_buff -> array, point_buff -> used);
    }
}

void
hack_draw_point (struct vcl_hack *context, const struct geom_point *p)
{
  GdkPoint pt;
  compute_point (&(context -> gt), &pt, p -> x, p -> y);
  gdk_draw_point (context -> win, context -> gc, pt . x, pt . y);
}


void
hack_apply_transform (struct vcl_hack *context, struct geom_transform2 *t)
{
  geom_transform_merge (&(t -> primary), &(context -> gt), &(context -> gt));
}

void
hack_unapply_transform (struct vcl_hack *context, struct geom_transform2 *t)
{
  geom_transform_merge (&(t -> inverted), &(context -> gt), &(context -> gt));
}

