#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vdo_vtable *vcl_vt_testrect;

static const struct geom_rectangle onerect = {{0, 0}, {1, 2}};

static inline struct geom_transform *
to_parent_coords (struct vcl_testrect * obj)
{
  return (&(obj -> coords . primary));
}

static inline struct geom_transform *
from_parent_coords (struct vcl_testrect * obj)
{
  return (&(obj -> coords . inverted));
}

static inline int
want_redraw (struct vcl_testrect * obj)
{
  return  (obj -> parent != NULL);
}


static void
vcl_testrect_destroy (struct vcl_testrect * obj)
{
  /* FIXME check placement */
  if (obj -> parent != NULL) vcl_placement_remove_child (obj -> parent, obj);
  xfree (obj);
}

static void
vcl_testrect_set_data (struct vcl_testrect * obj, int data)
{
  obj -> data = data;
}

static int
vcl_testrect_get_data (struct vcl_testrect * obj)
{
  return (obj -> data);
}

static void
vcl_testrect_set_parent (struct vcl_testrect * obj, void * parent)
{
  obj -> parent = parent;
}

static void *
vcl_testrect_get_parent (struct vcl_testrect * obj)
{
  return (obj -> parent);
}

static void
vcl_testrect_get_bbox (struct vcl_testrect * obj, struct geom_rectangle * bbox)
{
  *bbox = obj -> bbox;
}

static void
vcl_testrect_get_center (struct vcl_testrect * obj, struct geom_point * p)
{
  p -> x = obj -> coords . primary . coef [0][2];
  p -> y = obj -> coords . primary . coef [1][2];
}

static void
vcl_testrect_transform (struct vcl_testrect *obj, const struct geom_transform2 * trans)
{
  struct geom_rectangle old_bbox = obj -> bbox;
  geom_transform2_merge (&(obj -> coords), trans, &(obj -> coords));

  geom_rectangle_transform (&onerect, to_parent_coords (obj), &(obj -> bbox));
  if (want_redraw (obj))
    vcl_container_child_changed (obj -> parent, obj, &old_bbox, &(obj -> bbox));

}

static void
vcl_testrect_draw (struct vcl_testrect * obj, void * context,
		   const struct geom_rectangle * wanted_bbox) 
{
  if (! want_redraw (obj)) return;
  if (geom_rectangle_have_intersection (&(obj -> bbox), wanted_bbox))
    {
      hack_draw_bbox (context, &(obj -> bbox));
      hack_apply_transform (context, &(obj -> coords));
      hack_draw_image (context, 0, 0, 1, 1);
      hack_unapply_transform (context, &(obj -> coords));
    }
}

struct vcl_testrect *
vcl_testrect_new (void)
{
  MALLOCATED (struct vcl_testrect, obj);
  vdo_object_init (obj, vcl_vt_testrect);
  
  geom_transform2_set_identity (&(obj -> coords));
  obj -> bbox = onerect;
  return (obj);
}

void
vcl_testrect_main (void)
{
  vcl_vt_testrect = vdo_class_new ("TestRect");

  vdo_class_add_interface (vcl_vt_testrect, vdo_if_object);
  vdo_class_add_method_li (vdo_mid_object_destroy, vcl_testrect_destroy);

  vdo_class_add_interface (vcl_vt_testrect, vcl_if_item);
  vdo_class_add_method_li (vcl_mid_item_set_data, vcl_testrect_set_data);
  vdo_class_add_method_li (vcl_mid_item_get_data, vcl_testrect_get_data);
  vdo_class_add_method_li (vcl_mid_item_set_parent, vcl_testrect_set_parent);
  vdo_class_add_method_li (vcl_mid_item_get_parent, vcl_testrect_get_parent);
  vdo_class_add_method_li (vcl_mid_item_get_bbox, vcl_testrect_get_bbox);
  vdo_class_add_method_li (vcl_mid_item_get_center, vcl_testrect_get_center);
  vdo_class_add_method_li (vcl_mid_item_transform, vcl_testrect_transform);
  vdo_class_add_method_li (vcl_mid_item_draw, vcl_testrect_draw);
}

