#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>


#define COMPOSE_OP(X) \
void									\
vcl_compose_ ## X ## _ll (struct vcl_pixbuf *src1,                      \
			  struct vcl_pixbuf *src2,                      \
			  struct vcl_pixbuf *dst)			\
{									\
  ASSERT ((src1 -> channels == src2 -> channels) &&			\
	  (src1 -> channels == dst -> channels));			\
  ASSERT ((src1 -> depth == src2 -> depth) && 				\
	  (src1 -> depth == dst -> depth) &&				\
	  (src1 -> depth == 1));					\
  ASSERT ((src1 -> width == src2 -> width) &&				\
	  (src1 -> width <= dst -> width));				\
  ASSERT ((src1 -> height == src2 -> height) &&				\
	  (src1 -> height <= dst -> height));				\
									\
  u8 *src1_data = src1 -> data;						\
  u8 *src2_data = src2 -> data;						\
  u8 *dst_data  = dst  -> data;						\
  u8 *src1_stop = src1 -> data + src1 -> height * src1 -> rowstride; 	\
									\
  if (src1 -> channels == 1)						\
    while (src1_data < src1_stop)					\
      {									\
	compose_ ## X ## _1c_1d (src1_data, src2_data, dst_data,	\
				 src1 -> width);			\
	src1_data += src1 -> rowstride;					\
	src2_data += src2 -> rowstride;					\
	dst_data += dst -> rowstride;					\
      }									\
  else if (src1 -> channels == 4)					\
    while (src1_data < src1_stop)					\
      {									\
	compose_ ## X ## _4c_1d (src1_data, src2_data, dst_data,	\
				 src1 -> width, src1 -> alpha, 		\
				 src2 -> alpha);			\
	src1_data += src1 -> rowstride;					\
	src2_data += src2 -> rowstride;					\
	dst_data += dst -> rowstride;					\
      }									\
  else 									\
    while (src1_data < src1_stop)					\
      {									\
	compose_ ## X ## _uc_1d (src1_data, src2_data, dst_data,	\
				 src1 -> width, src1 -> alpha,		\
				 src2 -> alpha, src1 -> channels);	\
	src1_data += src1 -> rowstride;					\
	src2_data += src2 -> rowstride;					\
	dst_data += dst -> rowstride;					\
      }									\
}




/* compose operation OVER */

static inline void
compose_over_1c_1d (u8 *src1, u8 *src2, u8 *dst, uns width)
{
  for (int i = 0; i < width; i ++)
    dst [i] = src1 [i] + src2 [i] - MULT_CHAR (src1 [i], src2 [i]);
}

static inline void
compose_over_4c_1d (u8 *src1, u8 *src2, u8 *dst, uns width, uns a1, uns a2)
{
  for (int i = 0; i < (width * 4); i ++)
    dst [i] = src1 [i] + MULT_CHAR (src2 [i], 255 - src1 [(i & ~3) + a1]);
}

static inline void
compose_over_uc_1d (u8 *src1, u8 *src2, u8 *dst, uns width, uns a1, uns a2, uns channels)
{
  int a_count = 0;

  for (int i = 0; i < (width * channels); i ++)
    {
      dst [i] = src1 [i] + MULT_CHAR (src2 [i], 255 - src1 [a1]);

      a_count ++;
      if (a_count == channels)
	{
	  a_count == 0;
	  a1 += channels;
	}
    }
}

COMPOSE_OP(over);


/* compose operation IN */

static inline void
compose_in_1c_1d (u8 *src1, u8 *src2, u8 *dst, uns width)
{
  for (int i = 0; i < width; i ++)
    dst [i] = MULT_CHAR (src1 [i], src2 [i]);
}

static inline void
compose_in_4c_1d (u8 *src1, u8 *src2, u8 *dst, uns width, uns a1, uns a2)
{
  for (int i = 0; i < (width * 4); i ++)
    dst [i] = MULT_CHAR (src1 [i], src2 [(i & ~3) + a2]);
}

static inline void
compose_in_uc_1d (u8 *src1, u8 *src2, u8 *dst, uns width, uns a1, uns a2, uns channels)
{
  int a_count = 0;

  for (int i = 0; i < (width * channels); i ++)
    {
      dst [i] = MULT_CHAR (src1 [i], src2 [a2]);

      a_count ++;
      if (a_count == channels)
	{
	  a_count == 0;
	  a2 += channels;
	}
    }
}

COMPOSE_OP(in);


/* compose operation OUT */

static inline void
compose_out_1c_1d (u8 *src1, u8 *src2, u8 *dst, uns width)
{
  for (int i = 0; i < width; i ++)
    dst [i] = MULT_CHAR (src1 [i], 255 - src2 [i]);
}

static inline void
compose_out_4c_1d (u8 *src1, u8 *src2, u8 *dst, uns width, uns a1, uns a2)
{
  for (int i = 0; i < (width * 4); i ++)
    dst [i] = MULT_CHAR (src1 [i], 255 - src2 [(i & ~3) + a2]);
}

static inline void
compose_out_uc_1d (u8 *src1, u8 *src2, u8 *dst, uns width, uns a1, uns a2, uns channels)
{
  int a_index = a2;
  int a_count = 0;

  for (int i = 0; i < (width * channels); i ++)
    {
      dst [i] = MULT_CHAR (src1 [i], 255 - src2 [a_index]);

      a_count ++;
      if (a_count == channels)
	{
	  a_count == 0;
	  a_index += channels;
	}
    }
}

COMPOSE_OP(out);


/* compose operation ATOP */

/*
void render_atop_1c_8u_8u (struct vcl_compose * obj, unsigned char *buffer, int y, int x1, int x2)
missing - copy buffer b
*/

static inline void
compose_atop_4c_1d (u8 *src1, u8 *src2, u8 *dst, uns width, uns a1, uns a2))
{
  for (int i = 0; i < (width * 4); i ++)
    dst [i] = MULT_CHAR (src1 [i], src2 [(i & ~3) + a2])
      + MULT_CHAR (src2 [i], 255 - src1 [(i & ~3) + a1]);
}

static inline void
compose_atop_uc_1d (u8 *src1, u8 *src2, u8 *dst, uns width, uns a1, uns a2, uns channels)
{
  int a_count = 0;

  for (int i = 0; i < (width * channels); i ++)
    {
      dst [i] = MULT_CHAR (src1 [i], src2 [a2])
	+ MULT_CHAR (src2 [i], 255 - src1 [a1]);

      a_count ++;
      if (a_count == channels)
	{
	  a_count == 0;
	  a1 += channels;
	  a2 += channels;
	}
    }
}

COMPOSE_OP(atop);


/* compose operation XOR */

static inline void
compose_xor_1c_1d (u8 *src1, u8 *src2, u8 *dst, uns width)
{
  for (int i = 0; i < width; i ++)
    dst [i] = src1 [i] + src2 [i]
      - 2 * MULT_CHAR (src1 [i], src2 [i]);
}

static inline void
compose_xor_4c_1d (u8 *src1, u8 *src2, u8 *dst, uns width, uns a1, uns a2)
{
  for (int i = 0; i < (width * 4); i ++)
    dst [i] = MULT_CHAR (src1 [i], 255 - src2 [(i & ~3) + a2])
      + MULT_CHAR (src2 [i], 255 - src1 [(i & ~3) + a1]);
}

static inline void
compose_xor_uc_1d (u8 *src1, u8 *src2, u8 *dst, uns width, uns a1, uns a2, uns channels)
{
  int a_count = 0;

  for (int i = 0; i < (width * channels); i ++)
    { 
      dst [i] = MULT_CHAR (src1 [i], 255 - src2 [a2])
	+ MULT_CHAR (src2 [i], 255 - src1 [a1]);

      dst [i] = MULT_CHAR (src1 [i], 255 - src2 [a_index]);

      a_count ++;
      if (a_count == channels)
	{
	  a_count == 0;
	  a1 += channels;
	  a2 += channels;
	}
    }
}

COMPOSE_OP(xor);


/* dissolve operation */

static inline void
dissolve_1c_1c_1d (u8 *src, u8 *alpha, u8 *dst, uns width)
{
  for (int i = 0; i < width; i ++)
    dst [i] = MULT_CHAR (src [i], alpha [i]);
}

static inline void
dissolve_4c_1c_1d (u8 *src, u8 *alpha, u8 *dst, uns width)
{
  for (int i = 0; i < (width * 4); i ++)
    dst [i] = MULT_CHAR (src [i], alpha [(i >> 2)]);
}

static inline void
dissolve_uc_uc_1d (u8 *src, u8 *alpha, u8 *dst, uns width, uns c1, uns c2)
{
  int a_count = 0;

  for (int i = 0; i < (width * channels); i ++)
    {
      dst [i] = MULT_CHAR (src [i], alpha [alpha]);

      a_count ++;
      if (a_count == c1)
	{
	  a_count == 0;
	  alpha += c2;
	}
    }
}

void
vcl_dissolve_ll (struct vcl_pixbuf *src, struct vcl_pixbuf *alpha,	
		 struct vcl_pixbuf *dst)
{
  ASSERT ((src -> channels == dst -> channels));
  ASSERT ((src1 -> depth == src2 -> depth) &&
	  (src1 -> depth == dst -> depth) &&
	  (src1 -> depth == 1));
  ASSERT ((src -> width == alpha -> width) &&
	  (src -> width <= dst -> width));
  ASSERT ((src -> height == alpha -> height) &&
	  (src -> height <= dst -> height));

  u8 *src_data = src -> data;
  u8 *alpha_data = alpha -> data + alpha -> alpha;
  u8 *dst_data  = dst -> data;
  u8 *src_stop = src -> data + src -> height * src -> rowstride;

  if ((src -> channels == 1) && (alpha -> channels == 1))
    while (src_data < src_stop)
      {
	dissolve_1c_1c_1d (src_data, alpha_data, dst_data, src -> width);
	src_data += src -> rowstride;
	alpha_data += alpha -> rowstride;
	dst_data += dst -> rowstride;
      }
  else if ((src -> channels == 4) && (alpha -> channels == 1))
    while (src_data < src_stop)
      {
	dissolve_4c_1c_1d (src_data, alpha_data, dst_data, src -> width);
	src_data += src -> rowstride;
	alpha_data += alpha -> rowstride;
	dst_data += dst -> rowstride;
      }
  else
    while (src_data < src_stop)
      {
	dissolve_uc_uc_1d (src_data, alpha_data, dst_data, src -> width,
			   src -> channels, alpha -> channels);
	src_data += src -> rowstride;
	alpha_data += alpha -> rowstride;
	dst_data += dst -> rowstride;
      }
}
