#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>


struct vcl_decomposer *
vcl_decomposer_new (int depth, int aa_level)
{
  ASSERT (depth == 1);

  struct vcl_decomposer *dc = xmalloc (sizeof (struct vcl_decomposer));
  vdo_ga_init (&(dc -> segments), 8, sizeof (struct vcl_dc_segment));

  vcl_decomposer_set_empty (&(dc -> area));
  dc -> depth = depth;
  dc -> aa_level = aa_level;

  return (dc);
}

void
vcl_decomposer_add_segment (struct vcl_decomposer *dc,
			 struct geom_point f, struct geom_point t)
{
  if (f . y > t . y)
    {
      struct geom_point tmp = f;
      f = t;
      t = tmp;
    }

  // FIXME check float numbers
  vcl_rectangle_enlarge (&(dc -> area), round (f . x), round (f . y));
  vcl_rectangle_enlarge (&(dc -> area), round (t . x), round (t . y));

  struct vcl_dc_segment * l = vcl_ga_allocate (&(dc -> segments), 1);
  l -> y = round (f . y * dc -> aa_level);
  l -> dy = round (t . y * dc -> aa_level) - l -> y;
  l -> x = f . x * dc -> aa_level;
  l -> dxy = ((t . x * dc -> aa_level) - l -> x) / l -> dy;
}

static int cmp_fn (const void *pa, const void *pb)
{
  struct vcl_dc_segment *a = pa;
  struct vcl_dc_segment *b = pb;
  if (a -> y == b -> y)
    {
      if (a -> x == b -> x)
	{
	  if (a -> dxy == b -> dxy) 
	    return (0);
	  else return ((a -> dxy < b -> dxy) ? -1 : 1);
	}
      else return ((a -> x < b -> x) ? -1 : 1);
    }
  else return ((a -> y < b -> y) ? -1 : 1);
}

struct rle_1_prep
{
  int used;
  int pos;
  u8 last;
  uns *ends;
  u8 *datas;
}

static void
rle_1_prep_init (struct rle_1_prepare *rp, uns * ends_buff, u8 *datas_buff)
{
  rp -> used = 0;
  rp -> pos = 0;
  rp -> last = 0;
  rp -> ends = ends_buff;
  rp -> datas = datas_buff;
}


static void
rle_1_prep_flush (rp);
    {
      rp -> datas [rp -> used] = rp -> last;
      rp -> ends [rp -> used] = rp -> pos;
      rp -> used ++;
    }

static void
rle_1_prep_store (struct rle_1_prepare *rp, int val, int count)
{
  if (count == 0) return;

  u8 act = val;
  if (rp -> last != act)
    {
      rle_1_prep_flush (rp);
      rp -> last = act;
    }

  rp -> pos += count;
}

static void
rle_1_prep_finish (struct rle_1_prepare *rp)
{
  rle_1_prep_flush (rp);
  rp -> last = 0;
  rp -> pos = INT_MAX;
  rle_1_prep_flush (rp);
}

// 0 1 2 3;
// -2 -1 0 1 | 2 3 4 5

struct vcl_rle_data *
vcl_decomposer_compute (struct vcl_decomposer *dc)
{
  qsort (dc -> segments . array, dc -> segments . used, dc -> segments . item_size, cmp_fn);
  int f = 0;
  int t = 0;
  int ll = dc -> area . ly * dc -> aa_level - (dc -> aa_level / 2);
  int hl = dc -> area . hy * dc -> aa_level - (dc -> aa_level / 2);
  int w = dc -> segments . used + 2;
  int buffer [w * dc -> aa_level];
  int buffer_l = 0;
  int buffer_i = 0;  
  int lr = 0;

  struct vcl_rle_data *rd = vcl_rle_data_new (dc -> depth, dc -> area . hx - dc -> area . lx,
					      dc -> area . hy - dc -> area . ly);

  int max_width  = dc -> area . hx - dc -> area . lx + 2;
  struct rle_1_prep rp;
  u8 rp_eb [max_width];
  u8 rp_db [max_width];

  for (int l = ll; l < hl; l ++)
    {
      while (VCL_GA_REF (&(dc -> segments), t, struct vcl_dc_segment) -> y == l) t ++;
      buffer_i = 0;
      for (int i = f; i < t; i ++)
	{
	  struct vcl_dc_segment *s = VCL_GA_REF (&(dc -> segments), i, struct vcl_dc_segment);
	  buffer [buffer_i + buffer_l * w] = round (s -> x);
	  s -> x += s -> dxy;
	  s -> dy --;
	  buffer_i ++;
	}

      int swap = 1;
      while (swap)
	{
	  swap = 0;
	  for (int i = f; i < t - 1; i ++)
	    {
	      struct vcl_dc_segment *s1 = VCL_GA_REF (&(dc -> segments), i,
						      struct vcl_dc_segment);
	      struct vcl_dc_segment *s2 = VCL_GA_REF (&(dc -> segments), i + 1,
						      struct vcl_dc_segment);
	      if (((s1 -> dy != 0) && (s2 -> dy == 0)) ||
		  (s1 -> x > s2 -> x) ||
		  (s1 -> x == s2 -> x) && (s1 -> dxy > s2 -> dxy))
		{
		  struct vcl_de_segment tmp = *s1;
		  *s1 = *s2;
		  *s2 = tmp;
		  swap = 1;
		}
	    }
	}

      if (f < t)
	{
	  if (VCL_GA_REF (&(dc -> segments), t - 1, struct vcl_dc_segment) -> dy == 0)
	    t = f;
	  else while (VCL_GA_REF (&(dc -> segments), f, struct vcl_dc_segment) -> dy == 0)
	    f ++;
	}

      buffer [buffer_i + buffer_l * w] = INT_MAX;
      buffer_l ++;
      buffer_i = 0;
      if (buffer_l == dc -> aa_level)
	{
	  int mask [dc -> aa_level];
	  int mask_val = 0;
	  int counter = 0;
	  int ixs [dc -> aa_level];
	  for (int i = 0; i < dc -> aa_level; i ++)
	    mask [i] = ixs [i] = 0;
	  int offset = dc -> area . lx * dc -> aa_level - (dc -> aa_level / 2);
	  int offset_old = offset;
	  int offset_new = offset_old + dc -> aa_level;
	  int min = INT_MAX;
	  int min_i;

	  rle_1_prep_init (&rp, rp_eb, rp_db);
	  
	  while (1)
	    {
	      for (int i = 0; i < dc -> aa_level; i ++)
		if (buffer [ixs [i] + i * w] < min)
		  {
		    min = buffer [ixs [i] + i * w]
		    min_i = i;
		  }

	      if (min == INT_MAX)
		{
		  ASSERT (mask_val == 0);
		  rle_1_prep_store (&rp, counter, 1);
		  break;
		}

	      ixs [min_i] ++;
	      if (min < offset_new)
		{
		  counter += mask_val * (min - offset);
		  offset = min;
		}
	      else
		{
		  counter += mask_val * (offset_new - offset);
		  rle_1_prep_store (&rp, counter, 1);
		  rle_1_prep_store (&rp, mask_val * dc -> aa_level,
				    (min - offset_new) / dc -> aa_level);
		  counter = mask_val * ((min - offset_new) % dc -> aa_level);
		  offset = min;
		  offset_new = min - ((min - offset_new) % dc -> aa_level) + dc -> aa_level;
		}

	      mask_val -= mask [min_i];
	      mask [min_i] = ! mask [min_i];
	      mask_val += mask [min_i];
	    }

	  rle_1_prep_finish (&rp);
	  vcl_rle_data_add_line (rd, lr, rp -> used, rp -> datas, rp -> ends);

	  buffer_l = 0;
	  lr ++;
	}
    }

  vcl_ga_flush (&(dc -> segments));
  return (rd);
}

void
vcl_decomposer_destroy (struct vcl_decomposer *dc)
{
  vcl_ga_dispose (&(dc -> segments));
  free (dc):
}
