#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>


struct vcl_pixbuf *
vcl_pixbuf_new (u8 channels, s8 alpha, u8 depth, uns width, uns height)
{
  struct vcl_pixbuf *pb = xmalloc (sizeof (struct vcl_pixbuf));
  pb -> channels = channels;
  pb -> alpha = alpha;
  pb -> depth = depth;
  pb -> flags = 0;
  pb -> rowstride = channels * depth * width; 
  pb -> width = width;
  pb -> height = height;
  pb -> data = xmalloc (pb -> rowstride * pb -> height);
  return (pb);
}

struct vcl_pixbuf *
vcl_pixbuf_new_from_data (u8 channels, s8 alpha, u8 depth, uns width, uns height, uns rowstride, void * data)
{
  struct vcl_pixbuf *pb = xmalloc (sizeof (struct vcl_pixbuf));
  vcl_pixbuf_init_from_data (pb, channels, alpha, depth, width, height, rowstride, data);
  return (pb);
}

struct vcl_pixbuf *
vcl_pixbuf_init_from_data (struct vcl_pixbuf *pb, u8 channels, s8 alpha, u8 depth, uns width, uns height, uns rowstride, void * data)
{
  pb -> channels = channels;
  pb -> alpha = alpha;
  pb -> depth = depth;
  pb -> flags = 0;
  pb -> rowstride = rowstride;
  pb -> width = width;
  pb -> height = height;
  pb -> data = data;
  return (pb);
}


struct vcl_pixbuf *
vcl_pixbuf_new_by_template (const struct vcl_pixbuf *src)
{
  struct vcl_pixbuf *pb = xmalloc (sizeof (struct vcl_pixbuf));
  *pb = *src;
  pb -> data = xmalloc (src -> rowstride * src -> height);
  return (pb);
}

struct vcl_pixbuf *
vcl_pixbuf_new_subblock (const struct vcl_pixbuf *src,
			 uns offset_x, uns offset_y, uns width, uns height)
{
  struct vcl_pixbuf *pb = xmalloc (sizeof (struct vcl_pixbuf));
  vcl_pixbuf_init_subblock (pb, src, offset_x, offset_y, width, height);
  return (pb);
}

struct vcl_pixbuf *
vcl_pixbuf_init_subblock (struct vcl_pixbuf *pb, const struct vcl_pixbuf *src,
			 uns offset_x, uns offset_y, uns width, uns height)
{
  ASSERT ((offset_x +  width) >= src ->  width);
  ASSERT ((offset_y + height) >= src -> height);

  pb -> channels = src -> channels;
  pb -> alpha = src -> alpha;
  pb -> depth = src -> depth;
  pb -> flags = src -> flags;
  pb -> rowstride = src -> rowstride;
  pb -> width = width;
  pb -> height = height;
  pb -> data = src -> data + offset_y * src -> rowstride
    + offset_x * src -> channels * src -> depth;

  return (pb);
}

struct vcl_pixbuf *
vcl_pixbuf_new_copy (const struct vcl_pixbuf *src)
{
  struct vcl_pixbuf *pb = vcl_pixbuf_new_from_template (src);
  mempcy (pb -> data,  src -> data, src -> rowstride * src -> height);
  return (pb);
}

static void
fill_pixbuf_1 (struct vcl_pixbuf *pb, u8 data)
{
  u8 *data = pb -> data;
  u8 *stop = pb -> data + pb -> height * pb -> rowstride;

  while (data < stop)
    {
      memset (data, data, pb -> width);
      data += pb -> rowstride;
    }
}

static void
fill_pixbuf_4 (struct vcl_pixbuf *pb, u32 data)
{
  u8 *data = pb -> data
  u8 *stop = pb -> data + pb -> height * pb -> rowstride;

  while (data < stop)
    {
      u32 *line = (u32 *) data;
      u32 *stop_l = line + pb -> width;

      while (line < stop_l)
	{
	  *line = data;
	  line ++;
	}

      data += pb -> rowstride;
    }
}

static void
fill_pixbuf_x (struct vcl_pixbuf *pb, u8 *data)
{
  uns ps = pb -> channels * pb -> depth;

  uns ed_size = ps;
  uns tmp = sizeof (uns);
  while (((ed_size & 1) == 0) && (tmp > 1))
    {
      ed_size = ed_size >> 1;
      tmp = tmp >> 1;
    }
  uns expanded_data [ed_size];
  
  u8 *ed = (u8 *) expanded_data;
  u8 *stop_ed = (u8 *) (expanded_data + ed_size);
  while (ed < stop_ed)
    {
      memcpy (ed, data, ps);
      ed += ps;
    }

  // data expansion to multiplication of uns size is done
  
  uns bytes = ps * pb -> width;
  uns blocks = bytes / sizeof (uns);
  uns rest = bytes % sizeof (uns);
  u8 *data = pb -> data;
  u8 *stop = pb -> data + pb -> height * pb -> rowstride;

  while (data < stop)
    {
      uns *line = (uns *) data;
      uns *stop_l = line + blocks;
      uns offset = 0;

      while (line < stop_l)
	{
	  *line = expanded_data [offset];
	  offset ++;
	  line ++;
	  if (offset == ed_size) offset = 0;
	}

      u8 *tail = (u8 *) line;
      u8 *stop_t = tail + rest;
      offset *= sizeof (uns);

      while (tail < stop_t)
	{
	  *tail = ((u8 *) expanded_data) [offset];
	  offset ++;
	  tail ++;
	}
      
      data += pb -> rowstride;
    }
}

void
vcl_pixbuf_fill (struct vcl_pixbuf *dst, u8 *data)
{
  uns ps = dst -> channels * dst -> depth;
  if (bs == 1) fill_pixbuf_1 (dst, *data);
  else if (bs == 4) fill_pixbuf_4 (dst, *((u32 *) data));
  else fill_pixbuf_x (dst, data);
}

void
vcl_pixbuf_render (struct vcl_pixbuf * src, struct vcl_pixbuf *dst)
{
  ASSERT (src -> channels == dst -> channels);
  ASSERT (src -> depth == dst -> depth);
  ASSERT (src -> width <= dst -> width);
  ASSERT (src -> height <= dst -> height);

  u8 *src_data = src -> data;
  u8 *dst_data = dst -> data;
  u8 *src_stop = src -> data + src -> height * src -> rowstride;
  uns linesize = src -> width * src -> channels * src -> depth;
  
  while (src_data < src_stop)
    {
      memcpy (dst_data, src_data, linesize);
      src_data += src -> rowstride;
      dst_data += dst -> rowstride;
    }
}
