#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>


struct vcl_rle_data *
vcl_rle_data_new (u8 depth, uns width, uns height)
{
  struct vcl_rle_data *rd = xmalloc (sizeof (struct vcl_rle_data));

  rd -> depth = depth;
  rd -> width = width;
  rd -> height = height;
  rd -> offset_x = 0;
  rd -> offset_y = 0;
  rd -> lines = xmalloc_zero (sizeof (struct vcl_rle_data) * height);

  return (rle);
}

struct vcl_rle_data *
vcl_rle_data_new_subblock (const struct vcl_rle_data *src, uns offset_x, uns offset_y, 
			   uns width, uns height);
{
  struct vcl_rle_data *rd = xmalloc (sizeof (struct vcl_rle_data));
  vcl_rle_data_init_subblock (rd, src, offset_x, offset_y, width, height);
  return (rd);
}


struct vcl_rle_data *
vcl_rle_data_init_subblock (struct vcl_rle_data *rd, const struct vcl_rle_data *src,
			    uns offset_x, uns offset_y, uns width, uns height)
{
  ASSERT ((offset_x +  width) >= src ->  width);
  ASSERT ((offset_y + height) >= src -> height);

  rd -> depth = src -> depth;
  rd -> width = width;
  rd -> height = height;
  rd -> offset_x = src -> offset_x + offset_x;
  rd -> offset_y = src -> offset_y + offset_y;
  rd -> lines = src -> lines;

  return (rd);
}

void
vcl_rle_data_add_line (struct vcl_rle_data *rd, uns lineno, uns width,
		       const u8* datas, const uns* ends)
{
  struct vcl_rle_line *l = rd -> lines + lineno + rd -> offset_y;

  l -> datas = xmalloc (width * sizeof (u8) * rd -> depth);
  memcpy (l -> datas, datas, width * sizeof (u8) * rd -> depth);

  l -> ends = xmalloc (width * sizeof (uns));
  memcpy (l -> ends, ends, width * sizeof (uns));

  l -> width = width;
}


static void
render_line_1 (struct vcl_rle_data * rle, u8 * buffer, int y, int x1, int x2)
{
  uns *ends = rle -> lines [y] . ends;
  u8 *datas = rle -> lines [y] . datas;
  int i = 0;
  int old = x1;

  while (ends [i] <= x1) i ++;
  while (ends [i] <= x2)
    {
      memset (buffer + old, datas [i], ends [i] - old);
      old = ends [i];
      i ++;
    }
  memset (buffer + old, datas [i], x2 - old);
}
//  else memset (buffer, 0, x2 - x1);

static void
render_line_4 (struct vcl_rle_data * rle, u32 * buffer, int y, int x1, int x2)
{
  int *ends = rle -> lines [y] . ends;
  u32 *datas = (u32 *) (rle -> lines [y] . datas);
  int i = 0;
  int old = x1;

  while (ends [i] <= x1) i ++;
  while (ends [i] <= x2)
    {
      vcl_memset_32 (buffer + old, datas [i], ends [i] - old);
      old = ends [i];
      i ++;
    }
  vcl_memset_32 (buffer + old, datas [i], x2 - old);
}
//   else vcl_memset_32 (buffer, 0, x2 - x1);

static void
render_line_x (struct vcl_rle_data * rle, u8 * buffer, int y, int x1, int x2)
{
  int ps = rle -> channels * rle -> depth;
  int *ends = rle -> lines [y] . ends;
  u8 *datas = rle -> lines [y] . datas;
  int i = 0;
  int old = x1;

  while (ends [i] <= x1) i ++;
  while (ends [i] <= x2)
    {
      vcl_memset_univ (buffer + old * ps, datas + i * ps, ends [i] - old, ps);
      old = ends [i];
      i ++;
    }
  vcl_memset_univ (buffer + old * os, datas + i * ps, x2 - old, ps);
}

//  else memset (buffer, 0, (x2 - x1) * ps);

void
vcl_rle_data_crop_bbox (struct vcl_rle_data *rd, struct vcl_rectangle *src, 
			struct vcl_rectangle *dst)
{
  
}

void
vcl_rle_data_render (struct vcl_rle_data * src, struct vcl_pixbuf *dst)
{
  ASSERT (src -> channels == dst -> channels);
  ASSERT (src -> depth == dst -> depth);
  ASSERT (src -> width <= dst -> width);
  ASSERT (src -> height <= dst -> height);

  int ps = src -> channels * src -> depth;
  uns x1 = src -> offset_x;
  uns x2 = src -> offset_x + src -> width;
  uns y1 = src -> offset_y;
  uns y2 = src -> offset_y + src -> height;

  u8 *data = dst -> data - x1;
  
  if (ps == 1)
    for (l = y1; l < y2 ; l++, data += dst -> rowstride)
      render_line_1 (src, data, l, x1, x2);
  else if (ps == 4)
    for (l = y1; l < y2 ; l++, data += dst -> rowstride)
      render_line_4 (src, (u32 *) data, l, x1, x2);
  else 
    for (l = y1; l < y2 ; l++, data += dst -> rowstride)
      render_line_x (src, data, l, x1, x2);
}
