
struct vcl_rle_data *
compute_fill_outline (void *shape, struct geom_transform *t)
{
}

// compose source masked by rle (which is situated in coords by rle_bbox)
// over b (which is situated in coords by pb_bbox)
static void
rasterize_area (struct vcl_pixbuf *b, struct vcl_rectangle *pb_bbox,
		struct vcl_rle_data *rle, struct vcl_rectangle rle_bbox,
		void * source)
{
  struct vcl_pixbuf bg, fg, ol;
  struct vcl_rle_data rdc;
  struct vcl_rectangle bbox;

  vcl_rectangle_intersection (pb_bbox, rle_bbox, &bbox);
  if (vcl_rectangle_is_empty (&bbox)) return;

  uns d1x = bbox . lx - pb_bbox -> lx;
  uns d1y = bbox . ly - pb_bbox -> ly;
  uns d2x = bbox . lx - rle_bbox -> lx;
  uns d2y = bbox . ly - rle_bbox -> ly;
  uns w = bbox . hx - bbox . lx;
  uns h = bbox . hy - bbox . ly;
  u8 fg_data [w * h * 4];
  u8 ol_data [w * h];
	  
  vcl_pixbuf_init_subblock (&bg, b, d1x, d1y, w, h);
  vcl_pixbuf_init_from_data (&fg, 4, 3, 1, w, h, w * 4, fg_data);
  vcl_pixbuf_init_from_data (&ol, 1, 0, 1, w, h, w, ol_data);
  vcl_rle_data_init_subblock (&rdc, rd, d2x, d2y, w, h);
	  
  vcl_rle_data_rasterize (&rdc, &ol);
  vcl_source_rasterize (source, &fg, &bbox, t, 1);
  vcl_dissolve_ll (&fg, &ol, &fg);
  vcl_compose_over_ll (&fg, &bg, &bg);
}

static void
rasterize_container (struct vcl_pixbuf *b, const struct vcl_rectangle *pb_bbox,
		 struct vcl_context *ct, void *cnt)
{
  u32 zero = 0;
  struct vcl_pixbuf bg, fg;
  struct vcl_rectangle bbox;
  strut geom_rectangle r;

  vcl_item_get_bbox (container, &bbox);
  vcl_rectangle_intersection (pb_bbox, &bbox, &bbox);
  if (vcl_rectangle_is_empty (&bbox)) return;

  uns dx = bbox . lx - pb_bbox -> lx;
  uns dy = bbox . ly - pb_bbox -> ly;
  uns w = bbox . hx - bbox . lx;
  uns h = bbox . hy - bbox . ly;
  u8 fg_data [w * h * 4];

  vcl_pixbuf_init_subblock (&bg, b, dx, dy, w, h);
  vcl_pixbuf_init_from_data (&fg, b -> channels, b -> alpha, b -> depth,
  			     w, h, w * b -> channels * b -> depth, fg_data);
  vcl_pixbuf_fill (&fg, &zero);

  vcl_rectangle_to_geom_rectangle (&bbox, &r);
  geom_rectangle_transform (&r, &(ct -> t . inverted), &r);

  void *node;
  VCL_CONTAINER_WALK_CHILDREN_BACKWARDS_IN_BBOX (container, node, &r)
    rasterize_node (&fg, &bbox, ct, node);
	  
  //  vcl_dissolve_ll (&fg, &ol, &fg);
  vcl_compose_over_ll (&fg, &bg, &bg);
}


}

static void
rasterize_shape (struct vcl_pixbuf *b, const struct vcl_rectangle *pb_bbox,
		 struct vcl_context *ct, void *shape)
{
  u32 ft = vcl_context_get_u32_property (ct, VCL_PROP_FILL_STYLE);
  void *fs = vcl_context_get_ptr_property (ct, VCL_PROP_FILL_SOURCE);

  u32 sw = vcl_context_get_u32_property (ct, VCL_PROP_STROKE_WIDTH);
  void *ss = vcl_context_get_ptr_property (ct, VCL_PROP_STROKE_SOURCE);

  if (fs != NULL)
    {
      struct vcl_rectangle rle_bbox; 
      struct vcl_rle_data *rd;

      rd = vcl_shape_render_area (shape, &(ct -> t . primary), pb_bbox, &rle_bbox);
      rasterize_area (b, pb_bbox, rd, rle_bbox);
    }
}

static void
rasterize_node (struct vcl_pixbuf *b, const struct vcl_rectangle *pb_bbox,
		const vcl_context *ct, void *node)
{
  if (vcl_property_p (node))
    {
      int i;
      int j = vcl_property_get_count (node);
      for (i = 0; i < j; i ++)
	{
	  vcl_context_set_property (ct, vcl_property_get_type (node, i),
				    vcl_property_get_value (node, i));
	}
    }

  if (vcl_transform_p (node))
    vcl_context_apply_transform (ct, vcl_transform_get_primary);
  
  if (vcl_composite_p (node))
    rasterize_composite (b, pb_bbox, ct, node);
  else if (vcl_enclosure_p (node))
    rasterize_node (b, pb_bbox, ct, vcl_enclosure_get_child (node));
  else if (vcl_shape_p (node))
    rasterize_shape (b, pb_bbox, ct, node);

  if (vcl_transform_p (node))
    vcl_context_unapply_transform (ct);

  if (vcl_property_p (node))
    {
      int i;
      int j = vcl_property_get_count (node);
      for (i = j - 1; i >= 0; i --)
	{
	  vcl_context_unset_property (ct, vcl_property_get_type (node, i));
				      
	}
    }
  
  
}
