#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vcl_vtable *vcl_vt_affinity;


static void
changed_new_bbox (struct vcl_affinity * obj, const struct geom_rectangle *bbox_of_child)
{
  struct geom_rectangle old_bbox = obj -> bbox;
  geom_rectangle_transform (bbox_of_child, &(obj -> coords . primary), &(obj -> bbox));

  if (obj -> parent)
    vcl_container_child_changed (obj -> parent, obj, &old_bbox, &(obj -> bbox));
}

static void
changed (struct vcl_affinity * obj)
{
  struct geom_rectangle bbox;

  if (obj -> child)
    vcl_node_get_bbox (obj -> child, &bbox);
  else
    geom_rectangle_set_empty (&bbox);

  changed_new_bbox (obj, &bbox);
}


// OBJECT interface

static void
vcl_affinity_destroy (struct vcl_affinity * obj)
{
  ASSERT (obj -> parent == NULL);

  if (obj -> child != NULL)
  {
    vcl_node_set_parent (obj -> child, NULL);
    vcl_object_destroy (obj -> child);
  }

  xfree (obj);
}


// NODE interface

static void
vcl_affinity_set_data (struct vcl_affinity * obj, int data)
{ obj -> data = data; }

static int
vcl_affinity_get_data (struct vcl_affinity * obj)
{ return (obj -> data); }

static void
vcl_affinity_set_parent (struct vcl_affinity * obj, void * parent)
{ obj -> parent = parent; }

static void *
vcl_affinity_get_parent (struct vcl_affinity * obj)
{ return (obj -> parent); }

static void
vcl_affinity_get_bbox (struct vcl_affinity * obj, struct geom_rectangle * bbox)
{ *bbox = obj -> bbox; }


// CONTAINER interface

static void
vcl_affinity_child_altered (struct vcl_affinity * obj, UNUSED void * child,
			 const struct geom_rectangle *where)
{
  if (obj -> parent)
    {
      struct geom_rectangle new_where;
      geom_rectangle_transform (where, &(obj -> coords . primary), &new_where);
      vcl_container_child_altered (obj -> parent, obj, &new_where);
    }
}

static void
vcl_affinity_child_changed (struct vcl_affinity * obj, UNUSED void * child,
			    UNUSED const struct geom_rectangle *old_bbox,
			    const struct geom_rectangle *new_bbox)
{ changed_new_bbox (obj, new_bbox); }


// ENCLOSURE interface

static void *
vcl_affinity_get_child (struct vcl_affinity * obj)
{ return (obj -> child); }

static void
vcl_affinity_set_child (struct vcl_affinity * obj, void * child)
{
  if (obj -> child != NULL)
    vcl_node_set_parent (obj -> child, NULL);
  obj -> child = child;
  vcl_node_set_parent (child, obj);
  changed (obj);
}


// TRANSFORMATION interface

static int
vcl_affinity_primary_apply_matrix (struct vcl_affinity * obj, const struct geom_transform *src,
		    struct geom_transform *dst)
{ return (geom_transform_merge (src, &(obj -> coords . primary), dst)); }

static int
vcl_affinity_inverted_apply_matrix (struct vcl_affinity * obj, const struct geom_transform *src,
		    struct geom_transform *dst)
{ return (geom_transform_merge (&(obj -> coords . inverted), src, dst)); }

static int
vcl_affinity_primary_preply_matrix (struct vcl_affinity * obj, const struct geom_transform *src,
		    struct geom_transform *dst)
{ return (geom_transform_merge (&(obj -> coords . primary), src, dst)); }

static int
vcl_affinity_inverted_preply_matrix (struct vcl_affinity * obj, const struct geom_transform *src,
		    struct geom_transform *dst)
{ return (geom_transform_merge (src, &(obj -> coords . inverted), dst)); }

static int
vcl_affinity_primary_apply_point (struct vcl_affinity * obj, const struct geom_point *src,
		    struct geom_point *dst)
{ return (geom_point_transform (src, &(obj -> coords . primary), dst)); }

static int
vcl_affinity_inverted_apply_point (struct vcl_affinity * obj, const struct geom_point *src,
		    struct geom_point *dst)
{ return (geom_point_transform (src, &(obj -> coords . inverted), dst)); }


// specific functions

struct vcl_affinity *
vcl_affinity_new (void)
{
  struct vcl_affinity *obj = xmalloc (sizeof (struct vcl_affinity));
  vcl_object_init (obj, vcl_vt_affinity);

  geom_transform2_set_identity (&(obj -> coords));
  geom_rectangle_set_empty (&(obj -> bbox));

  obj -> child = NULL;
  obj -> parent = NULL;
  obj -> data = 0;

  return (obj);
}

void
vcl_affinity_set (struct vcl_affinity *obj, const struct geom_transform2 *t)
{
  obj -> coords = *t;
  changed (obj);
}

void
vcl_affinity_inner_merge (struct vcl_affinity *obj, const struct geom_transform2 *t)
{
  geom_transform2_merge (t, &(obj -> coords), &(obj -> coords));
  changed (obj);
}

void
vcl_affinity_outer_merge (struct vcl_affinity *obj, const struct geom_transform2 *t)
{
  geom_transform2_merge (&(obj -> coords), t, &(obj -> coords));
  changed (obj);
}

void
vcl_affinity_main (void)
{
  vcl_vt_affinity = vcl_class_new ("affinity");

  vcl_class_add_interface (vcl_vt_affinity, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_affinity_destroy);

  vcl_class_add_interface (vcl_vt_affinity, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_affinity_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_affinity_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_affinity_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_affinity_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_affinity_get_bbox);

  vcl_class_add_interface (vcl_vt_affinity, vcl_if_container);
  vcl_class_add_method_li (vcl_mid_container_child_altered, vcl_affinity_child_altered);
  vcl_class_add_method_li (vcl_mid_container_child_changed, vcl_affinity_child_changed);

  vcl_class_add_interface (vcl_vt_affinity, vcl_if_enclosure);
  vcl_class_add_method_li (vcl_mid_enclosure_set_child, vcl_affinity_set_child);
  vcl_class_add_method_li (vcl_mid_enclosure_get_child, vcl_affinity_get_child);

  vcl_class_add_interface (vcl_vt_affinity, vcl_if_transformation);
  vcl_class_add_method_li (vcl_mid_transformation_primary_apply_matrix, vcl_affinity_primary_apply_matrix);
  vcl_class_add_method_li (vcl_mid_transformation_inverted_apply_matrix, vcl_affinity_inverted_apply_matrix);
  vcl_class_add_method_li (vcl_mid_transformation_primary_preply_matrix, vcl_affinity_primary_preply_matrix);
  vcl_class_add_method_li (vcl_mid_transformation_inverted_preply_matrix, vcl_affinity_inverted_preply_matrix);

  vcl_class_add_method_li (vcl_mid_transformation_primary_apply_point, vcl_affinity_primary_apply_point);
  vcl_class_add_method_li (vcl_mid_transformation_inverted_apply_point, vcl_affinity_inverted_apply_point);
}
