#ifndef _VCL_GAS_H_
#define _VCL_GAS_H_

#include <stdlib.h>

struct vcl_growing_array
{
  void *array;
  int used;
  int size;
  int item_size;
};


#define VCL_GA_REF(GA, INDEX, TYPE) (((TYPE *) ((GA) -> array)) [INDEX])
#define VCL_GA_REF_LAST(GA, TYPE) (((TYPE *) ((GA) -> array)) [(GA) -> used - 1])


extern void
vcl_ga_init (struct vcl_growing_array *ga, int init_items, int item_size);


extern void
vcl_ga_ensure (struct vcl_growing_array *ga, int pos);


static inline struct vcl_growing_array *
vcl_ga_new (int init_items, int item_size)
{
  struct vcl_growing_array *ga = xmalloc (sizeof (struct vcl_growing_array));
  vcl_ga_init (ga, init_items, item_size);
  return (ga);
}

static inline void
vcl_ga_flush (struct vcl_growing_array *ga)
{
  ga -> used = 0;
}

static inline void
vcl_ga_finish (struct vcl_growing_array *ga)
{
  xfree (ga -> array);
}

static inline void
vcl_ga_destroy (struct vcl_growing_array *ga)
{
  vcl_ga_finish (ga);
  xfree (ga);
}

static inline int
vcl_ga_new_item (struct vcl_growing_array *ga)
{
  int i = ga -> used ++;
  vcl_ga_ensure (ga, i);
  return (i);
}

static inline void *
vcl_ga_allocate (struct vcl_growing_array *ga, int items)
{
  int i = ga -> used;
  ga -> used += items;
  vcl_ga_ensure (ga, ga -> used - 1);
  return ((void *) (((char *) ga -> array) + (i * ga -> item_size)));
}

static inline void *
vcl_ga_unallocate (struct vcl_growing_array *ga, int items)
{
  ga -> used -= items;
  ASSERT (ga -> used >= 0);
  return ((void *) (((char *) ga -> array) + (ga -> used * ga -> item_size)));
}


#endif
