#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vcl_vtable *vcl_vt_le_area;

typedef struct childtab_item
{
  struct childtab_item* next;
  void* child_id;
  void* child;
  int locked;
} childtab_item;

#define HT_PREFIX(x) cht_ ## x
#define HT_KEY void*
#define HT_KEY_POINTER child_id
#define HT_TYPE void*
#define HT_STRUCT childtab_item
#define HT_ALLOC(_struct, _key, _type) \
	(_struct) = (childtab_item *) xmalloc (sizeof (struct childtab_item)); \
	(_struct)->child_id = _key; \
	(_struct)->child = _type;
#include "lib/hashtable.h"

static inline int
want_redraw (struct vcl_le_area * obj)
{ return  (! (obj -> parent == NULL)); }

static inline void
cb_get_bbox_of_child (void * obj, struct vcl_le_area * questioner,
		      struct geom_rectangle *bbox)
{ questioner -> gboc_cb (obj, questioner, bbox, questioner -> gboc_data); }

static inline void *
cb_expand_child (void * obj, struct vcl_le_area * questioner)
{ return questioner -> ec_cb (obj, questioner, questioner -> ec_data); }

static inline  void
cb_get_children (struct vcl_le_area * questioner, uns backwards,
			 const struct geom_rectangle * bbox)
{ questioner -> gc_cb (questioner, backwards, bbox, questioner -> gc_data); }


static void
altered (struct vcl_le_area * obj, const struct geom_rectangle *where)
{
  if (want_redraw (obj) && (! geom_rectangle_is_empty (where)))
    vcl_container_child_altered (obj -> parent, obj, where);
}

static void
changed (struct vcl_le_area * obj,
	 const struct geom_rectangle *old_bbox,
	 const struct geom_rectangle *new_bbox)
{
  struct geom_rectangle temp;

  geom_rectangle_merge (old_bbox, new_bbox, &temp);

  if (geom_rectangle_is_in (&temp, &(obj -> bbox)))
    altered (obj, &temp);
  else /* new bbox is not empty, because old_bbox is always in lc_bbox */
    {
      struct geom_rectangle old_bbox = obj -> bbox;

      geom_rectangle_merge (&(obj -> bbox), new_bbox, &(obj -> bbox));
      if (want_redraw (obj))
	vcl_container_child_changed (obj -> parent, obj, &old_bbox, &(obj -> bbox));
    }
}

static void
cache_add (struct vcl_le_area * obj, void * child_id, void * child)
{
  /* suppose not in table */
  struct childtab_item *cht_item = cht_insert((cht_ht *) &(obj -> cht), child_id, child);
  cht_item -> locked = 0;
}

static void
cache_invalidate (struct vcl_le_area * obj, void * child_id)
{
  struct childtab_item *cht_item = cht_get ((void*)&(obj -> cht), child_id);
  if (cht_item != NULL)
    {
      void * child = cht_item -> child;
      ASSERT (! cht_item -> locked);
      cht_delete ((void*)&(obj -> cht), child_id);

      vcl_node_set_parent (child, NULL);
      vcl_object_destroy (child);
    }
}

static void *
cache_lookup (struct vcl_le_area * obj, void * child_id)
{
  struct childtab_item *cht_item = cht_get ((void*)&(obj -> cht), child_id);
  return ((cht_item == NULL) ? NULL : cht_item -> child);
}

static void
cache_lock (struct vcl_le_area * obj, void * child_id)
{
  struct childtab_item *cht_item = cht_get ((void*)&(obj -> cht), child_id);
  ASSERT (cht_item != NULL);
  cht_item -> locked = 1;
}

static void
cache_unlock (struct vcl_le_area * obj, void * child_id)
{
  struct childtab_item *cht_item = cht_get ((void*)&(obj -> cht), child_id);
  ASSERT (cht_item != NULL);
  cht_item -> locked = 0;
}


// OBJECT interface

static void
vcl_le_area_destroy (struct vcl_le_area * obj)
{
  ASSERT (obj -> parent == NULL);

  HT_DO_ALL(((cht_ht *) &(obj -> cht)), walk)
    {
      void * child = walk -> child;
      ASSERT (! walk -> locked);
      vcl_node_set_parent (child, NULL);
      vcl_object_destroy (child);
    }

  cht_finish((cht_ht *) &(obj -> cht));
  xfree (obj);
}

static void
vcl_le_area_set_data (struct vcl_le_area * obj, int data)
{ obj -> data = data; }

static int
vcl_le_area_get_data (struct vcl_le_area * obj)
{ return (obj -> data); }

static void
vcl_le_area_set_parent (struct vcl_le_area * obj, void * parent)
{ obj -> parent = parent; }

static void *
vcl_le_area_get_parent (struct vcl_le_area * obj)
{ return (obj -> parent); }

static void
vcl_le_area_get_bbox (struct vcl_le_area * obj, struct geom_rectangle * bbox)
{ *bbox = obj -> bbox; }


// CONTAINER interface

static void
vcl_le_area_child_altered (struct vcl_le_area * obj, void * child UNUSED,
			 const struct geom_rectangle *where)
{ altered (obj, where); }

static void
vcl_le_area_child_changed (struct vcl_le_area * obj, void * child UNUSED,
			 const struct geom_rectangle *old_bbox,
			 const struct geom_rectangle *new_bbox)
{ changed (obj, old_bbox, new_bbox); }


// COMPOSITE interface

struct push_args_to_answer
{
  void (* cb)(void *, void *);
  void * cb_data;
};

void
vcl_le_area_answer_child_in_bbox (struct vcl_le_area * obj, void * child_id)
{
  struct push_args_to_answer *pata = (struct push_args_to_answer *) obj -> temporary;
  void * child = cache_lookup (obj, child_id);

  ASSERT (pata != NULL);

  if (child == NULL)
    {
      child = cb_expand_child (child_id, obj);

      vcl_node_set_parent (child, obj);
      cache_add (obj, child_id, child);
    }

  (pata -> cb) (child, (pata -> cb_data));
}

static void
universal_get_children  (struct vcl_le_area * obj, const struct geom_rectangle * wanted_bbox, uns backwards, void (* cb)(void *, void *), void * cb_data)
{
  if ((wanted_bbox == NULL) || (geom_rectangle_have_intersection (&(obj -> bbox), wanted_bbox)))
    {
      struct push_args_to_answer pata;

      pata . cb = cb;
      pata . cb_data = cb_data;

      obj -> temporary = &pata;
      cb_get_children (obj, backwards, wanted_bbox);
      obj -> temporary = NULL;
    }
}

static void
vcl_le_area_get_children (struct vcl_le_area * obj, void (* cb)(void *, void *), void * cb_data)
{ universal_get_children (obj, NULL, 0, cb, cb_data); }

static void
vcl_le_area_get_children_backwards (struct vcl_le_area * obj, void (* cb)(void *, void *), void * cb_data)
{ universal_get_children (obj, NULL, 1, cb, cb_data); }

static void
vcl_le_area_get_children_in_bbox (struct vcl_le_area * obj, const struct geom_rectangle * wanted_bbox, void (* cb)(void *, void *), void * cb_data)
{ if (wanted_bbox != NULL) universal_get_children (obj, wanted_bbox, 0, cb, cb_data); }

static void
vcl_le_area_get_children_in_bbox_backwards (struct vcl_le_area * obj, const struct geom_rectangle * wanted_bbox, void (* cb)(void *, void *), void * cb_data)
{ if (wanted_bbox != NULL) universal_get_children (obj, wanted_bbox, 1, cb, cb_data); }


// specific functions

struct vcl_le_area *
vcl_le_area_new (get_bbox_of_child_fn_t cb1, void * cb1_data,
		 expand_child_fn_t cb2, void * cb2_data,
		 get_children_fn_t cb3, void * cb3_data)
{
  struct vcl_le_area *obj = xmalloc (sizeof (struct vcl_le_area));
  vcl_object_init (obj, vcl_vt_le_area);
  geom_rectangle_set_empty (&(obj -> bbox));

  cht_init ((void*)&(obj -> cht), 201);

  obj -> gboc_cb = cb1;
  obj -> gboc_data = cb1_data;
  obj -> ec_cb = cb2;
  obj -> ec_data = cb2_data;
  obj -> gc_cb = cb3;
  obj -> gc_data = cb3_data;

  return (obj);
}

void
vcl_le_area_flush (struct vcl_le_area * obj)
{
  HT_DO_ALL(((cht_ht *) &(obj -> cht)), walk)
    {
      void * child = walk -> child;
      ASSERT (! walk -> locked);
      vcl_node_set_parent (child, NULL);
      vcl_object_destroy (child);
    }

  cht_finish((cht_ht *) &(obj -> cht));
  cht_init ((void*)&(obj -> cht), 201);

  altered (obj, &(obj -> bbox));
}


void
vcl_le_area_child_altered_notification (struct vcl_le_area * obj, void * child_id,
					const struct geom_rectangle *where)
{
  cache_invalidate (obj, child_id);
  altered (obj, where);
}

void
vcl_le_area_child_changed_notification (struct vcl_le_area * obj, void * child_id,
					const struct geom_rectangle *old_bbox,
					const struct geom_rectangle *new_bbox)
{
  cache_invalidate (obj, child_id);
  changed (obj, old_bbox, new_bbox);
}


void
vcl_le_area_child_transformed_notification (struct vcl_le_area * obj, void * child_id,
					    const struct geom_rectangle *old_bbox,
					    const struct geom_rectangle *new_bbox,
					    UNUSED const struct geom_transform2 * trans)
{
  cache_invalidate (obj, child_id);
  changed (obj, old_bbox, new_bbox);
}

void
vcl_le_area_child_appeared_notification (struct vcl_le_area * obj, void * child_id)
{
  struct geom_rectangle child_bbox;
  cb_get_bbox_of_child (child_id, obj, &child_bbox);
  changed (obj, NULL, &child_bbox);
}

void
vcl_le_area_child_disappeared_notification (struct vcl_le_area * obj, void * child_id)
{
  void * child = cache_lookup (obj, child_id);

  if (child != NULL)
    {
      struct geom_rectangle child_bbox;
      vcl_node_get_bbox (child, &child_bbox);
      cache_invalidate (obj, child_id);
      altered (obj, &child_bbox);
    }
  /* If child isn't in cache, then it shouldn't be on viewable position,
     so no need to redraw */
}

void *
vcl_le_area_child_lookup (struct vcl_le_area * obj, void * child_id)
{
  void * child = cache_lookup (obj, child_id);
  if (child != NULL) cache_lock (obj, child_id);
  return (child);
}

void
vcl_le_area_child_unlock (struct vcl_le_area * obj, void * child_id)
{
  cache_unlock (obj, child_id);
}

void
vcl_le_area_main (void)
{
  vcl_vt_le_area = vcl_class_new ("lazy-expanding-area");

  vcl_class_add_interface (vcl_vt_le_area, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_le_area_destroy);

  vcl_class_add_interface (vcl_vt_le_area, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_le_area_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_le_area_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_le_area_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_le_area_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_le_area_get_bbox);

  vcl_class_add_interface (vcl_vt_le_area, vcl_if_container);
  vcl_class_add_method_li (vcl_mid_container_child_altered, vcl_le_area_child_altered);
  vcl_class_add_method_li (vcl_mid_container_child_changed, vcl_le_area_child_changed);

  vcl_class_add_interface (vcl_vt_le_area, vcl_if_composite);
  vcl_class_add_method_li (vcl_mid_composite_get_children, vcl_le_area_get_children);
  vcl_class_add_method_li (vcl_mid_composite_get_children_backwards, vcl_le_area_get_children_backwards);
  vcl_class_add_method_li (vcl_mid_composite_get_children_in_bbox, vcl_le_area_get_children_in_bbox);
  vcl_class_add_method_li (vcl_mid_composite_get_children_in_bbox_backwards, vcl_le_area_get_children_in_bbox_backwards);
}

