#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vcl_vtable *vcl_vt_offset;


// NODE interface

static void
vcl_offset_set_data (UNUSED struct vcl_offset * obj, UNUSED int data)
{ BUG (); }

static int
vcl_offset_get_data (UNUSED struct vcl_offset * obj)
{ BUG (); }

static void
vcl_offset_set_parent (UNUSED struct vcl_offset * obj, UNUSED void * parent)
{ BUG (); }

static void *
vcl_offset_get_parent (UNUSED struct vcl_offset * obj)
{ BUG (); }

static void
vcl_offset_get_bbox (struct vcl_offset * obj, struct geom_rectangle * bbox)
{
  vcl_node_get_bbox (obj -> child, bbox);
  bbox -> ll . x += obj -> dx;
  bbox -> ll . y += obj -> dy;
  bbox -> ur . x += obj -> dx;
  bbox -> ur . y += obj -> dy;
}


// CONTAINER interface

static void
vcl_offset_child_altered (UNUSED struct vcl_offset * obj, UNUSED void * child,
			  UNUSED const struct geom_rectangle *where)
{ BUG (); }

static void
vcl_offset_child_changed (UNUSED struct vcl_offset * obj, UNUSED void * child,
			  UNUSED const struct geom_rectangle *old_bbox,
			  UNUSED const struct geom_rectangle *new_bbox)
{ BUG (); }


// ENCLOSURE interface

static void *
vcl_offset_get_child (struct vcl_offset * obj)
{ return (obj -> child); }

static void
vcl_offset_set_child (UNUSED struct vcl_offset * obj, UNUSED void * child)
{ BUG (); }


// TRANSFORMATION interface

static int
vcl_offset_primary_apply_matrix (struct vcl_offset * obj, const struct geom_transform *src,
				 struct geom_transform *dst)
{
  if (src != dst) *dst = *src;
  dst -> flags = 0;
  dst -> coef [0] [2] += obj -> dx;
  dst -> coef [1] [2] += obj -> dy;
  return ((isfinite (dst -> coef [0] [2]) && isfinite (dst -> coef [1] [2]))
	  ? 0 : GEOM_ERR_NUMERIC);
}

static int
vcl_offset_inverted_apply_matrix (struct vcl_offset * obj, const struct geom_transform *src,
				  struct geom_transform *dst)
{
  if (src != dst) *dst = *src;
  dst -> flags = 0;
  dst -> coef [0] [2] -= obj -> dx * dst -> coef [0] [0] + obj -> dy * dst -> coef [0] [1];
  dst -> coef [1] [2] -= obj -> dx * dst -> coef [1] [0] + obj -> dy * dst -> coef [1] [1];
  return ((isfinite (dst -> coef [0] [2]) && isfinite (dst -> coef [1] [2]))
	  ? 0 : GEOM_ERR_NUMERIC);
}

static int
vcl_offset_primary_preply_matrix (struct vcl_offset * obj, const struct geom_transform *src,
				  struct geom_transform *dst)
{
  if (src != dst) *dst = *src;
  dst -> flags = 0;
  dst -> coef [0] [2] += obj -> dx * dst -> coef [0] [0] + obj -> dy * dst -> coef [0] [1];
  dst -> coef [1] [2] += obj -> dx * dst -> coef [1] [0] + obj -> dy * dst -> coef [1] [1];
  return ((isfinite (dst -> coef [0] [2]) && isfinite (dst -> coef [1] [2]))
	  ? 0 : GEOM_ERR_NUMERIC);
}

static int
vcl_offset_inverted_preply_matrix (struct vcl_offset * obj, const struct geom_transform *src,
				  struct geom_transform *dst)
{
  if (src != dst) *dst = *src;
  dst -> flags = 0;
  dst -> coef [0] [2] -= obj -> dx;
  dst -> coef [1] [2] -= obj -> dy;
  return ((isfinite (dst -> coef [0] [2]) && isfinite (dst -> coef [1] [2]))
	  ? 0 : GEOM_ERR_NUMERIC);
}

static int
vcl_offset_primary_apply_point (struct vcl_offset * obj, const struct geom_point *src,
			  struct geom_point *dst)
{
  if (src != dst) *dst = *src;
  dst -> x += obj -> dx;
  dst -> y += obj -> dy;
  return ((isfinite (dst -> x) && isfinite (dst -> y))
	  ? 0 : GEOM_ERR_NUMERIC);
}

static int
vcl_offset_inverted_apply_point (struct vcl_offset * obj, const struct geom_point *src,
				struct geom_point *dst)
{
  if (src != dst) *dst = *src;
  dst -> x -= obj -> dx;
  dst -> y -= obj -> dy;
  return ((isfinite (dst -> x) && isfinite (dst -> y))
	  ? 0 : GEOM_ERR_NUMERIC);
}


// specific functions


struct vcl_offset *
vcl_offset_init (struct vcl_offset * obj, real dx, real dy, void * child)
{
  vcl_object_init (obj, vcl_vt_offset);

  obj -> dx = dx;
  obj -> dy = dy;
  obj -> child = child;

  return (obj);
}

void
vcl_offset_main (void)
{
  vcl_vt_offset = vcl_class_new ("offset");

  vcl_class_add_interface (vcl_vt_offset, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_offset_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_offset_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_offset_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_offset_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_offset_get_bbox);

  vcl_class_add_interface (vcl_vt_offset, vcl_if_container);
  vcl_class_add_method_li (vcl_mid_container_child_altered, vcl_offset_child_altered);
  vcl_class_add_method_li (vcl_mid_container_child_changed, vcl_offset_child_changed);

  vcl_class_add_interface (vcl_vt_offset, vcl_if_enclosure);
  vcl_class_add_method_li (vcl_mid_enclosure_set_child, vcl_offset_set_child);
  vcl_class_add_method_li (vcl_mid_enclosure_get_child, vcl_offset_get_child);

  vcl_class_add_interface (vcl_vt_offset, vcl_if_transformation);
  vcl_class_add_method_li (vcl_mid_transformation_primary_apply_matrix, vcl_offset_primary_apply_matrix);
  vcl_class_add_method_li (vcl_mid_transformation_inverted_apply_matrix, vcl_offset_inverted_apply_matrix);
  vcl_class_add_method_li (vcl_mid_transformation_primary_preply_matrix, vcl_offset_primary_preply_matrix);
  vcl_class_add_method_li (vcl_mid_transformation_inverted_preply_matrix, vcl_offset_inverted_preply_matrix);
  vcl_class_add_method_li (vcl_mid_transformation_primary_apply_point, vcl_offset_primary_apply_point);
  vcl_class_add_method_li (vcl_mid_transformation_inverted_apply_point, vcl_offset_inverted_apply_point);
}
