#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vcl_vtable *vcl_vt_path;


// OBJECT interface

static void
vcl_path_destroy (struct vcl_path * obj)
{
  ASSERT (obj -> parent == NULL);
  geom_path_destroy (&(obj -> embedded));
  xfree (obj);
}


// NODE interface

static void
vcl_path_set_data (struct vcl_path * obj, int data)
{ obj -> data = data; }

static int
vcl_path_get_data (struct vcl_path * obj)
{ return (obj -> data); }

static void
vcl_path_set_parent (struct vcl_path * obj, void * parent)
{ obj -> parent = parent; }

static void *
vcl_path_get_parent (struct vcl_path * obj)
{ return (obj -> parent); }

static void
vcl_path_get_bbox (struct vcl_path * obj, struct geom_rectangle * bbox)
{ geom_path_bbox (obj -> used, bbox); }


// SHAPE interface

static inline void
add_point (int append, uns *used, u8 **buff, uns *buff_size, real x, real y)
{
  if ((! append) && *used)
    (*used) --;

  if ((*used * sizeof (struct geom_point)) == *buff_size)
    {
      if (*buff_size < 64)
	*buff_size = 64;
      else
	*buff_size *= 2;

      *buff = xrealloc (*buff, *buff_size);
    }

  struct geom_point *xbuff = (struct geom_point *) *buff;
  xbuff [*used] . x = x;
  xbuff [*used] . y = y;

  (*used) ++;
}

static void *
vcl_path_polygonize (void * a, void * it, double delta,
		     const struct geom_transform *t,
		     const struct geom_rectangle *bbox,
		     u8 ** buff, uns * buff_size,
		     uns *closed, uns *points, UNUSED uns hints)
{
  struct vcl_path *obj = a;
  static double *int_buff = NULL;
  static int int_buff_size = 0;

  struct geom_bezier dst;
  struct geom_path *p = obj -> used;
  struct geom_bezier *bez;

  ASSERT (it == NULL);

  *points = 0;

  GEOM_GROUP_WALK(p, bez)
    {
      struct geom_rectangle bezier_bbox;
      geom_bezier_control_box (bez, &bezier_bbox);
      GEOM_ITEM_INIT(&dst, bezier);
      geom_bezier_transform (bez, (void *)t, &dst);

      if (geom_rectangle_have_intersection (&bezier_bbox, bbox))
	{
	  uns pts;

	  geom_bezier_approximate (&dst, delta, &int_buff, &int_buff_size, &pts);
	  for (uns j = 0; j < pts; j ++)
	    add_point (j, points, buff, buff_size, int_buff [j * 2], int_buff [j * 2 + 1]);
	}
      else
	{
	  add_point (0, points, buff, buff_size, dst . pt [0] . x, dst . pt [0] . y);
	  add_point (1, points, buff, buff_size, dst . pt [GEOM_BEZIER_DEGREE(&dst)] . x,
		     dst . pt [GEOM_BEZIER_DEGREE(&dst)] . y);
	}
      geom_bezier_destroy(&dst);
    }

  struct geom_point *xbuff = (struct geom_point *) *buff;

  *closed = ((*points > 1)
      && (IS_EQUAL (xbuff [0] . x, xbuff [*points - 1] . x))
      && (IS_EQUAL (xbuff [0] . y, xbuff [*points - 1] . y)));

  return (NULL);
}


// specific functions

struct vcl_path *
vcl_path_new (struct geom_path * p)
{
  struct vcl_path *obj = xmalloc (sizeof (struct vcl_path));
  vcl_object_init (obj, vcl_vt_path);
  obj -> parent = NULL;
  obj -> data = 0;

  if (p == NULL)
    {
      GEOM_ITEM_CREATE (&(obj -> embedded), path);
      obj -> used = &(obj -> embedded);
    }
  else obj -> used = p;

  return (obj);
}

void
vcl_path_source_changed (struct vcl_path * obj)
{
  struct geom_rectangle new_bbox;
  vcl_path_get_bbox (obj, &new_bbox);

  if (obj -> parent != NULL)
    vcl_container_child_altered (obj -> parent, obj, &new_bbox);
}

void
vcl_path_main (void)
{
  vcl_vt_path = vcl_class_new ("path");

  vcl_class_add_interface (vcl_vt_path, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_path_destroy);

  vcl_class_add_interface (vcl_vt_path, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_path_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_path_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_path_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_path_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_path_get_bbox);

  vcl_class_add_interface (vcl_vt_path, vcl_if_shape);
  vcl_class_add_method_li (vcl_mid_shape_polygonize, vcl_path_polygonize);
}
