#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vcl_vtable *vcl_vt_painter_plainx;

int vcl_to_gdk_cap_styles [VCL_CAP_COUNT] =
  {[VCL_CAP_BUTT] = GDK_CAP_BUTT,
   [VCL_CAP_ROUND] = GDK_CAP_ROUND,
   [VCL_CAP_PROJECTING] = GDK_CAP_PROJECTING};

int vcl_to_gdk_join_styles [VCL_JOIN_COUNT] =
  {[VCL_JOIN_MITER] = GDK_JOIN_MITER,
   [VCL_JOIN_ROUND] = GDK_JOIN_ROUND,
   [VCL_JOIN_BEVEL] = GDK_JOIN_BEVEL};


// OBJECT interface

static void
vcl_painter_plainx_destroy (struct vcl_painter_plainx * obj)
{
  vcl_context_destroy (obj -> ct);
  xfree (obj);
}


// PAINTER_PLAINX interface

/*
void
convert_point (struct vcl_painter_plainx * obj, double x, double y, GdkPoint *p)
{
  struct geom_point p1, p2;
  p1 . x = x;
  p1 . y = y;
  geom_point_transform (&p1, &(obj -> ct -> t . primary), &p2);
  p -> x = p2 . x;
  p -> y = p2 . y;
}

void
draw_bbox (struct vcl_painter_plainx * obj, const struct geom_rectangle *r)
{
  GdkPoint pts [5];

  convert_point (obj, r -> ll . x, r -> ll . y, pts + 0);
  convert_point (obj, r -> ll . x, r -> ur . y, pts + 1);
  convert_point (obj, r -> ur . x, r -> ur . y, pts + 2);
  convert_point (obj, r -> ur . x, r -> ll . y, pts + 3);
  convert_point (obj, r -> ll . x, r -> ll . y, pts + 4);

  gdk_draw_lines (obj -> win, obj -> gc, pts, 5);
}

*/


static void
vcl_painter_plainx_draw_node (struct vcl_painter_plainx * obj, void * node, const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local)
{ vcl_context_walk_node (obj -> ct, obj, node, bbox, bbox_local); }

static void
fill_x_color (u32 color, GdkColor *dst)
{
  dst -> pixel = 0;
  dst -> red   = 257 * get_r (color);
  dst -> green = 257 * get_g (color);
  dst -> blue  = 257 * get_b (color);
}

static inline uns
get_cs_code (struct geom_point p, const struct vcl_rectangle *bbox)
{
  return (((p . y > bbox -> hy) ? 8 : 0) +
	  ((p . y < bbox -> ly) ? 4 : 0) +
	  ((p . x > bbox -> hx) ? 2 : 0) +
	  ((p . x < bbox -> lx) ? 1 : 0));
}

static inline void
intersect_v (const struct geom_point *p, double dx, double dy,
             double at, const struct vcl_rectangle *bbox, GdkPoint *dst)
{
  double par = (at - p -> x) / dx;
  double val = p -> y + par * dy;

  DBG ("V AT: %f PAR: %f VAL: %f\n", at, par, val);

  if ((bbox -> ly <= val) && (bbox -> hy >= val))
    {
      dst -> x = at;
      dst -> y = val;
    }
}

static inline void
intersect_h (const struct geom_point *p, double dx, double dy,
             double at, const struct vcl_rectangle *bbox, GdkPoint *dst)
{
  double par = (at - p -> y) / dy;
  double val = p -> x + par * dx;

  DBG ("H AT: %f PAR: %f VAL: %f\n", at, par, val);

  if ((bbox -> lx <= val) && (bbox -> hx >= val))
    {
      dst -> x = val;
      dst -> y = at;
    }
}

static inline void
clip_line (uns oc, uns nc, GdkPoint *d1, GdkPoint *d2,
	   const struct geom_point *p1, const struct geom_point *p2,
	   const struct vcl_rectangle *bbox, uns clear)
{
  d1 -> x = p1 -> x;
  d1 -> y = p1 -> y;
  d2 -> x = p2 -> x;
  d2 -> y = p2 -> y;

  if ((oc | nc) != 0)
    {
      real dx = p2 -> x - p1 -> x;
      real dy = p2 -> y - p1 -> y;

      if (clear)
	{
	  if (oc & 1) intersect_v (p1, dx, dy, bbox -> lx, bbox, d1);
	  else if (oc & 2) intersect_v (p1, dx, dy, bbox -> hx, bbox, d1);

	  if (oc & 4) intersect_h (p1, dx, dy, bbox -> ly, bbox, d1);
	  else if (oc & 8) intersect_h (p1, dx, dy, bbox -> hy, bbox, d1);
	}

      if (nc & 1) intersect_v (p1, dx, dy, bbox -> lx, bbox, d2);
      else if (nc & 2) intersect_v (p1, dx, dy, bbox -> hx, bbox, d2);

      if (nc & 4) intersect_h (p1, dx, dy, bbox -> ly, bbox, d2);
      else if (nc & 8) intersect_h (p1, dx, dy, bbox -> hy, bbox, d2);
    }
}

static inline void
draw_lines (struct vcl_painter_plainx * obj)
{
  DBG ("DRAW %d PLINES\n", obj -> point_buffer . used);

  GdkColor clr;

  if (get_a (obj -> ct -> fill_color) > 127)
    {
      fill_x_color (obj -> ct -> fill_color, &clr);
      gdk_gc_set_rgb_fg_color (obj -> gc, &clr);

      gdk_draw_polygon (obj -> win, obj -> gc, TRUE, obj -> point_buffer . array,
			obj -> point_buffer . used);
    }

  if (get_a (obj -> ct -> stroke_color) > 127)
    {
      fill_x_color (obj -> ct -> stroke_color, &clr);
      gdk_gc_set_rgb_fg_color (obj -> gc, &clr);

      gdk_draw_lines (obj -> win, obj -> gc, obj -> point_buffer . array,
		      obj -> point_buffer . used);
    }

  vcl_ga_flush (&(obj -> point_buffer));
}

static inline void
add_point (struct vcl_painter_plainx * obj, GdkPoint p)
{
  int i = vcl_ga_new_item (&(obj -> point_buffer));
  VCL_GA_REF (&(obj -> point_buffer), i, GdkPoint) = p;
}

static void
draw_shape (struct vcl_painter_plainx * obj, void * node, const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local)
{
  struct vcl_rectangle clip = *bbox;
  struct geom_point * buff;
  void * it = NULL;
  uns closed = 0;
  uns points = 0;

  // FIXME this is ugly
  if (clip . lx > - 10000) clip . lx = - 10000;
  if (clip . ly > - 10000) clip . ly = - 10000;
  if (clip . hx < 10000) clip . hx = 10000;
  if (clip . hy < 10000) clip . hy = 10000;

  gdk_gc_set_line_attributes (obj -> gc, round (obj -> ct -> stroke_width), GDK_LINE_SOLID,
			      vcl_to_gdk_cap_styles [obj -> ct -> stroke_cap_style],
			      vcl_to_gdk_join_styles [obj -> ct -> stroke_join_style]);

  do
    {
      it = vcl_shape_polygonize (node, it, 0.2, &(obj -> ct -> t . primary), bbox_local, &(obj -> buff), &(obj -> buff_len), &closed, &points, 0);

      if (points == 0) continue;
      buff = (struct geom_point *) obj -> buff;
      uns nc, oc; // new code, old code

      // Cohen - Sutherland algorithm for line clipping
      // needed because otherwise we get garbage in conversion
      // from float to unsigned

      /* // No clipping
      uns clear = 1;
      for (uns i = 0; i < points; i ++)
	{
	  GdkPoint d1;
	  d1 . x = buff [i] . x;

	  add_point (obj, d1);
	  clear = 0;
	}
      */

      uns clear = 1;
      oc = get_cs_code (buff [0], &clip);
      for (uns i = 0; i < (points - 1); i ++)
	{
	  nc = get_cs_code (buff [i+1], &clip);

	  if ((oc & nc) == 0)
	    {
	      GdkPoint d1, d2;

	      clip_line (oc, nc, &d1, &d2, buff + i, buff + i + 1, &clip, clear);
	      DBG ("%d %d %d %d\n", d1 . x, d1 . y, d2 . x, d2 . y);

	      if (clear)
		{
		  add_point (obj, d1);
		  clear = 0;
		}

	      add_point (obj, d2);

	      if (nc != 0)
		{
		  draw_lines (obj);
		  clear = 1;
		}
	    }
	  oc = nc;
	}

      if (! clear) draw_lines (obj);
    } while (it != NULL);
}

static void
draw_mask_data (GdkWindow * win, int x, int y, int w, int h, u8 *data)
{
  //a heuristic: do not try to allocate more than 1MB
  if ( w * h > (1 << 18) ) { err("Too large bitmap!\n"); return; }

  if ( w == 0 || h == 0 ) return; //nothing to do

  u32 * buff = xmalloc_nonfail(w * h * 4);
  if (!buff) { err("Cannot allocate memory, too large bitmap!\n"); return; }

  int i;
  for (i = 0; i < (w*h); i++)
    {
	buff [i] = data[i] << 24;
    }

  GdkPixbuf *pb = gdk_pixbuf_new_from_data ((char *) buff, GDK_COLORSPACE_RGB,
					    1, 8, w, h, w*4, NULL, NULL);
  gdk_pixbuf_render_to_drawable_alpha (pb, win, 0, 0, x, y, w, h,
				       GDK_PIXBUF_ALPHA_FULL, 128,
				       GDK_RGB_DITHER_NORMAL, 0, 0);
  gdk_pixbuf_unref (pb);
  xfree(buff);
}

static void
draw_mask (struct vcl_painter_plainx * obj, void * node, const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local)
{
  struct vcl_rectangle rb;
  vcl_mask_render (node, &(obj -> ct -> t . primary), bbox, bbox_local, &(obj -> buff), &(obj -> buff_len), &rb);
  draw_mask_data (obj -> win, rb . lx, rb . ly, rb . hx - rb . lx, rb . hy - rb . ly, obj -> buff);
}

static void
vcl_painter_plainx_draw_leaf (struct vcl_painter_plainx * obj, void * node, const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local)
{
  if (vcl_if_shape_p (node))
    draw_shape (obj, node, bbox, bbox_local);
  else if (vcl_if_mask_p (node))
    draw_mask (obj, node, bbox, bbox_local);
}

static void
vcl_painter_plainx_draw_composite (struct vcl_painter_plainx * obj, void * node, const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local)
{ vcl_context_walk_children (obj -> ct, obj, node, bbox, bbox_local); }

// specific functions

struct vcl_painter_plainx *
vcl_painter_plainx_new (GtkWidget *wg)
{
  struct vcl_painter_plainx *obj = xmalloc (sizeof (struct vcl_painter_plainx));
  vcl_object_init (obj, vcl_vt_painter_plainx);

  vcl_ga_init (&(obj -> point_buffer), 32, sizeof (GdkPoint));

  obj -> win = wg -> window;
  obj -> gc = gdk_gc_new (obj -> win);
  gdk_gc_copy (obj -> gc, gtk_widget_get_style (wg) -> fg_gc [0]);

  obj -> buff = NULL;
  obj -> buff_len = 0;

  obj -> ct = vcl_context_new ();

  return (obj);
}

void
vcl_painter_plainx_main (void)
{
  vcl_vt_painter_plainx = vcl_class_new ("painter-plainx");

  vcl_class_add_interface (vcl_vt_painter_plainx, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_painter_plainx_destroy);

  vcl_class_add_interface (vcl_vt_painter_plainx, vcl_if_painter);
  vcl_class_add_method_li (vcl_mid_painter_draw_node, vcl_painter_plainx_draw_node);
  vcl_class_add_method_li (vcl_mid_painter_draw_leaf, vcl_painter_plainx_draw_leaf);
  vcl_class_add_method_li (vcl_mid_painter_draw_composite, vcl_painter_plainx_draw_composite);
}
