#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vcl_vtable *vcl_vt_segment;


// OBJECT interface

static void
vcl_segment_destroy (struct vcl_segment * obj)
{
  ASSERT (obj -> parent == NULL);
  xfree (obj);
}


// NODE interface

static void
vcl_segment_set_data (struct vcl_segment * obj, int data)
{ obj -> data = data; }

static int
vcl_segment_get_data (struct vcl_segment * obj)
{ return (obj -> data); }

static void
vcl_segment_set_parent (struct vcl_segment * obj, void * parent)
{ obj -> parent = parent; }

static void *
vcl_segment_get_parent (struct vcl_segment * obj)
{ return (obj -> parent); }

static void
vcl_segment_get_bbox (struct vcl_segment * obj, struct geom_rectangle * bbox)
{
  bbox -> ll . x = MIN (obj -> p1 . x, obj -> p2 . x);
  bbox -> ll . y = MIN (obj -> p1 . y, obj -> p2 . y);
  bbox -> ur . x = MAX (obj -> p1 . x, obj -> p2 . x);
  bbox -> ur . y = MAX (obj -> p1 . y, obj -> p2 . y);
}


// SHAPE interface

static void *
vcl_segment_polygonize (void * a, void * it, UNUSED double delta,
			const struct geom_transform *t,
			UNUSED const struct geom_rectangle *bbox,
			u8 ** buff, uns * buff_size,
			uns *closed, uns *points, UNUSED uns hints)
{
  struct vcl_segment *obj = a;

  ASSERT (it == NULL);

  if (*buff_size < (2 * sizeof (struct geom_point)))
    {
      if (*buff_size < 64)
	*buff_size = 64;
      else
	*buff_size *= 2;

      *buff = xrealloc (*buff, *buff_size);
    }

  struct geom_point *xbuff = (struct geom_point *) *buff;
  geom_point_transform (&(obj -> p1), t, xbuff + 0);
  geom_point_transform (&(obj -> p2), t, xbuff + 1);
  *points = 2;
  *closed = 0;

  return (NULL);
}


// specific functions

struct vcl_segment *
vcl_segment_new (struct geom_point p1, struct geom_point p2)
{
  struct vcl_segment *obj = xmalloc (sizeof (struct vcl_segment));
  vcl_object_init (obj, vcl_vt_segment);
  obj -> p1 = p1;
  obj -> p2 = p2;
  obj -> parent = NULL;
  obj -> data = 0;

  return (obj);
}

void
vcl_segment_main (void)
{
  vcl_vt_segment = vcl_class_new ("segment");

  vcl_class_add_interface (vcl_vt_segment, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_segment_destroy);

  vcl_class_add_interface (vcl_vt_segment, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_segment_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_segment_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_segment_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_segment_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_segment_get_bbox);

  vcl_class_add_interface (vcl_vt_segment, vcl_if_shape);
  vcl_class_add_method_li (vcl_mid_shape_polygonize, vcl_segment_polygonize);
}
