#undef LOCAL_DEBUG

#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "kernel/kernel.h"
#include "font/font.h"

vcl_vtable *vcl_vt_string;


// NODE interface

static void
vcl_string_set_data (UNUSED struct vcl_string * obj, UNUSED int data)
{ BUG (); }

static int
vcl_string_get_data (UNUSED struct vcl_string * obj)
{ BUG (); }

static void
vcl_string_set_parent (UNUSED struct vcl_string * obj, UNUSED void * parent)
{ BUG (); }

static void *
vcl_string_get_parent (UNUSED struct vcl_string * obj)
{ BUG (); }

static void
vcl_string_get_bbox (struct vcl_string * obj, struct geom_rectangle * bbox)
{ *bbox = obj -> bbox; }


// MASK interface

static void *
vcl_string_render (struct vcl_string * obj, const struct geom_transform *t,
		   UNUSED const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local,
		   u8** buff, uns * buff_len, struct vcl_rectangle *retbox)
{
  if (geom_rectangle_have_intersection (&(obj -> bbox), bbox_local))
    {
      struct font_ctl ctl;
      font_ctl_init(&ctl);

      ctl . charmap = obj -> charmap;
      ctl . font_id = obj -> font_id;
      ctl . font_size = obj -> font_size;
      ctl . transform = *t;

      ctl . bitmap = *buff;
      ctl . bsize = *buff_len;

      if (font_render_string (&ctl, obj -> string) >= 0)
	{
	  retbox -> lx = ctl . xoff;
	  retbox -> ly = ctl . yoff;
	  retbox -> hx = ctl . xoff + ctl . width;
	  retbox -> hy = ctl . yoff + ctl . height;
	}
      else vcl_rectangle_set_empty (retbox);

      *buff = ctl . bitmap;
      *buff_len = ctl . bsize;

      ctl . bitmap = NULL;
      ctl . bsize = 0;
      font_ctl_cleanup(&ctl);
    }
  else vcl_rectangle_set_empty (retbox);

  return (NULL);
}


// specific functions

extern void
vcl_string_init (struct vcl_string *obj,  const char *string, uns charmap,  int font_id, real font_size, const struct geom_rectangle *bbox)
{
  vcl_object_init (obj, vcl_vt_string);
  obj -> string = string;
  obj -> charmap = charmap;
  obj -> font_id = font_id;
  obj -> font_size = font_size;
  obj -> bbox = *bbox;
}

void
vcl_string_main (void)
{
  vcl_vt_string = vcl_class_new ("string");

  vcl_class_add_interface (vcl_vt_string, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_string_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_string_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_string_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_string_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_string_get_bbox);

  vcl_class_add_interface (vcl_vt_string, vcl_if_mask);
  vcl_class_add_method_li (vcl_mid_mask_render, vcl_string_render);
}
