#undef LOCAL_DEBUG

#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "import/import.h"
#include "kernel/kernel.h"
#include "font/font.h"

vcl_vtable *vcl_vt_tex_layout;


// OBJECT interface

static void
vcl_tex_layout_destroy (struct vcl_tex_layout * obj)
{
  ASSERT (obj -> parent == NULL);
  if (obj -> copy) xfree (obj -> glyphs);
  xfree (obj);
}


// NODE interface

static void
vcl_tex_layout_set_data (struct vcl_tex_layout * obj, int data)
{ obj -> data = data; }

static int
vcl_tex_layout_get_data (struct vcl_tex_layout * obj)
{ return (obj -> data); }

static void
vcl_tex_layout_set_parent (struct vcl_tex_layout * obj, void * parent)
{ obj -> parent = parent; }

static void *
vcl_tex_layout_get_parent (struct vcl_tex_layout * obj)
{ return (obj -> parent); }

static void
vcl_tex_layout_get_bbox (struct vcl_tex_layout * obj, struct geom_rectangle * bbox)
{ *bbox = obj -> bbox; }


// COMPOSITE interface

static void
vcl_tex_layout_get_children (struct vcl_tex_layout * obj, void (* cb)(void *, void * ), void * cb_data)
{
  struct vcl_char ch;
  struct vcl_offset of;
  for (int i = obj -> glyphs_cnt - 1; i >= 0 ; i --)
    {
      struct tex_glyph * gl = obj -> glyphs + i;
      vcl_char_init (&ch, gl -> code, FONT_CHARMAP_TEX, gl -> font_id, gl -> font_size, gl -> bbox);
      vcl_offset_init (&of, gl -> xpos, gl -> ypos, &ch);
      cb (&of, cb_data);
    }
}

static void
vcl_tex_layout_get_children_backwards (struct vcl_tex_layout * obj, void (* cb)(void *, void * ), void * cb_data)
{
  struct vcl_char ch;
  struct vcl_offset of;

  for (uns i = 0; i < obj -> glyphs_cnt; i ++)
    {
      struct tex_glyph * gl = obj -> glyphs + i;
      vcl_char_init (&ch, gl -> code, FONT_CHARMAP_TEX, gl -> font_id, gl -> font_size, gl -> bbox);
      vcl_offset_init (&of, gl -> xpos, gl -> ypos, &ch);
      cb (&of, cb_data);
    }
}

static void
vcl_tex_layout_get_children_in_bbox (struct vcl_tex_layout * obj, const struct geom_rectangle * wanted_bbox, void (* cb)(void *, void * ), void * cb_data)
{
  if (geom_rectangle_have_intersection (&(obj -> bbox), wanted_bbox))
    vcl_tex_layout_get_children (obj, cb, cb_data);
}

static void
vcl_tex_layout_get_children_in_bbox_backwards (struct vcl_tex_layout * obj, const struct geom_rectangle * wanted_bbox, void (* cb)(void *, void * ), void * cb_data)
{
  if (geom_rectangle_have_intersection (&(obj -> bbox), wanted_bbox))
    vcl_tex_layout_get_children_backwards (obj, cb, cb_data);
}


// specific functions

struct vcl_tex_layout *
vcl_tex_layout_new (uns copy, struct tex_glyph *glyphs, uns glyphs_cnt, const struct geom_rectangle *bbox)
{
  struct vcl_tex_layout *obj = xmalloc (sizeof (struct vcl_tex_layout));
  vcl_object_init (obj, vcl_vt_tex_layout);

  obj -> bbox = *bbox;
  obj -> copy = copy;
  obj -> glyphs = glyphs;
  obj -> glyphs_cnt = glyphs_cnt;

  return (obj);
}

void
vcl_tex_layout_main (void)
{
  vcl_vt_tex_layout = vcl_class_new ("tex-layout");

  vcl_class_add_interface (vcl_vt_tex_layout, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_tex_layout_destroy);

  vcl_class_add_interface (vcl_vt_tex_layout, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_tex_layout_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_tex_layout_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_tex_layout_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_tex_layout_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_tex_layout_get_bbox);

  vcl_class_add_interface (vcl_vt_tex_layout, vcl_if_composite);
  vcl_class_add_method_li (vcl_mid_composite_get_children, vcl_tex_layout_get_children);
  vcl_class_add_method_li (vcl_mid_composite_get_children_backwards, vcl_tex_layout_get_children_backwards);
  vcl_class_add_method_li (vcl_mid_composite_get_children_in_bbox, vcl_tex_layout_get_children_in_bbox);
  vcl_class_add_method_li (vcl_mid_composite_get_children_in_bbox_backwards, vcl_tex_layout_get_children_in_bbox_backwards);
}
