#include "vdl_internal.h"

#include <stdio.h>
#include <stdlib.h>

static int context_item_table_size = 0;
static int context_item_table_used = 0;
static const char **context_item_names = NULL;
static int *context_item_sizes = NULL;

short int vdl_cid_add_point = 0;
short int vdl_cid_add_line = 0;
short int vdl_cid_add_bezier_2n = 0;
short int vdl_cid_add_bezier_2r = 0;
short int vdl_cid_add_bezier_3n = 0;
short int vdl_cid_add_bezier_3r = 0;
short int vdl_cid_start_path = 0;
short int vdl_cid_close_path = 0;

void
vdl_context_init (void)
{
  vdl_cid_add_point = vdl_register_context_item
    ("add_point", sizeof (struct ci_point));
  vdl_cid_add_line = vdl_register_context_item ("add_line", 0);
  vdl_cid_add_bezier_2n = vdl_register_context_item
    ("add_quadratic_bezier", sizeof (struct ci_bezier_2n));
  vdl_cid_add_bezier_2r = vdl_register_context_item
    ("add_quadratic_rational_bezier", sizeof (struct ci_bezier_2r));
  vdl_cid_add_bezier_3n = vdl_register_context_item
    ("add_cubic_bezier", sizeof (struct ci_bezier_3n));
  vdl_cid_add_bezier_3r = vdl_register_context_item
    ("add_cubic_rational_bezier", sizeof (struct ci_bezier_3r));
  vdl_cid_start_path = vdl_register_context_item ("start_path", 0);
  vdl_cid_close_path = vdl_register_context_item ("close_path", 0);
}


void
vdl_fill_bezier (struct vdl_context *ct, struct geom_bezier *b, int offset,
		  double x1, double y1, double x2, double y2)
{
  int type = CI_TYPE (ct, offset);

  if (type == vdl_cid_add_line)
    {
      b -> degree = GEOM_BEZIER_LINEAR;
      b -> flags = GEOM_BEZIER_NONRATIONAL;
  }
  else if (type == vdl_cid_add_bezier_2n)
    {
      struct ci_bezier_2n *op = (struct ci_bezier_2n *) CI_ADDR (ct, offset);
      b -> degree = GEOM_BEZIER_QUADRATIC;
      b -> flags = GEOM_BEZIER_NONRATIONAL;
      b -> pt [1] . x = op -> x1;
      b -> pt [1] . y = op -> y1;
    }
  else if (type == vdl_cid_add_bezier_2r)
    {
      struct ci_bezier_2r *op = (struct ci_bezier_2r *) CI_ADDR (ct, offset);
      b -> degree = GEOM_BEZIER_QUADRATIC;
      b -> flags = 0;
      b -> pt [1] . x = op -> x1;
      b -> pt [1] . y = op -> y1;
      b -> pt [0] . w = op -> wi;
      b -> pt [1] . w = op -> w1;
      b -> pt [2] . w = op -> wo;
    }
  else if (type == vdl_cid_add_bezier_3n)
    {
      struct ci_bezier_3n *op = (struct ci_bezier_3n *) CI_ADDR (ct, offset);
      b -> degree = GEOM_BEZIER_CUBIC;
      b -> flags = GEOM_BEZIER_NONRATIONAL;
      b -> pt [1] . x = op -> x1;
      b -> pt [1] . y = op -> y1;
      b -> pt [2] . x = op -> x2;
      b -> pt [2] . y = op -> y2;
    }
  else if (type == vdl_cid_add_bezier_3r)
    {
      struct ci_bezier_3r *op = (struct ci_bezier_3r *) CI_ADDR (ct, offset);
      b -> degree = GEOM_BEZIER_CUBIC;
      b -> flags = 0;
      b -> pt [1] . x = op -> x1;
      b -> pt [1] . y = op -> y1;
      b -> pt [2] . x = op -> x2;
      b -> pt [2] . y = op -> y2;
      b -> pt [0] . w = op -> wi;
      b -> pt [1] . w = op -> w1;
      b -> pt [2] . w = op -> w2;
      b -> pt [3] . w = op -> wo;
    }
  else BUG();

  b -> pt [0] . x = x1;
  b -> pt [0] . y = y1;
  b -> pt [b -> degree] . x = x2;
  b -> pt [b -> degree] . y = y2;

  if (b -> flags == GEOM_BEZIER_NONRATIONAL)
    {
      int i;
      for (i = 0; i <= b -> degree; i ++)
	b -> pt [i] . w = 1;
    }
}


int
vdl_register_context_item (const char *name, int size)
{
  int id;

  if (context_item_table_size == context_item_table_used)
    {
      if (context_item_table_size == 0)
	context_item_table_size = 16;
      else context_item_table_size *= 2;
      context_item_names = (const char **)
	xrealloc (context_item_names,
		 context_item_table_size * sizeof (char *));
      context_item_sizes = (int *)
	xrealloc (context_item_sizes,
		 context_item_table_size * sizeof (int));
    }

  id = context_item_table_used ++;
  context_item_names [id] = name;
  context_item_sizes [id] = size;
  return (id);
}

int
vdl_new_context_item (struct vdl_context *ct, short int cid)
{
  int id = ct -> command_buffer_used;
  int new_size = context_item_sizes [cid];

  if (ct -> command_buffer_size == ct -> command_buffer_used)
    {
      if (ct -> command_buffer_size == 0)
	ct -> command_buffer_size = 8;
      else ct -> command_buffer_size *= 2;

      ct -> command_buffer_cids = (short int *)
	xrealloc (ct -> command_buffer_cids,
		 ct -> command_buffer_size * sizeof (short int));
      ct -> command_buffer_offsets = (int *)
	xrealloc (ct -> command_buffer_offsets,
		 ct -> command_buffer_size * sizeof (int));
    }

  while ((ct -> command_buffer_args_used + new_size) > ct -> command_buffer_args_size)
    {
      if (ct -> command_buffer_args_size == 0)
	ct -> command_buffer_args_size = 64;
      else ct -> command_buffer_args_size *= 2;
      ct -> command_buffer_args = xrealloc (ct -> command_buffer_args,
					   ct -> command_buffer_args_size);
    }

  ct -> command_buffer_cids [id] = cid;
  ct -> command_buffer_offsets [id] = ct -> command_buffer_args_used;
  ct -> command_buffer_args_used += new_size;
  ct -> command_buffer_used ++;
  return (id);
}

struct vdl_context *
vdl_new_context (struct vdl_surface *sf)
{
  MALLOCATED (struct vdl_context, ct);
  ct -> command_buffer_cids = NULL;
  ct -> command_buffer_offsets = NULL;
  ct -> command_buffer_size = 0;
  ct -> command_buffer_used = 0;
  ct -> command_buffer_args = NULL;
  ct -> command_buffer_args_size = 0;
  ct -> command_buffer_args_used = 0;
  ct -> surface = sf;
  return (ct);
}

void
vdl_flush_context (struct vdl_context *ct)
{
  ct -> command_buffer_used = 0;
  ct -> command_buffer_args_used = 0;
}

void
vdl_add_point (struct vdl_context *ct, xxxnum_t x, xxxnum_t y)
{
  int id = vdl_new_context_item (ct, vdl_cid_add_point);
  struct ci_point *op = (struct ci_point *) CI_ADDR (ct, id);
  op -> x = x;
  op -> y = y;
}

void
vdl_add_line (struct vdl_context *ct)
{
  vdl_new_context_item (ct, vdl_cid_add_line);
}

void
vdl_add_cubic_bezier (struct vdl_context *ct, xxxnum_t x1, xxxnum_t y1,
		       xxxnum_t x2, xxxnum_t y2)
{
  int id = vdl_new_context_item (ct, vdl_cid_add_bezier_2n);
  struct ci_bezier_3n *op = (struct ci_bezier_3n *) CI_ADDR (ct, id);
  op -> x1 = x1;
  op -> y1 = y1;
  op -> x2 = x2;
  op -> y2 = y2;
}

void
vdl_add_cubic_rational_bezier (struct vdl_context *ct, xxxnum_t x1, xxxnum_t y1,
				xxxnum_t x2, xxxnum_t y2, xxxnum_t wi, xxxnum_t w1,
				xxxnum_t w2, xxxnum_t wo)
{
  int id = vdl_new_context_item (ct, vdl_cid_add_bezier_3r);
  struct ci_bezier_3r *op = (struct ci_bezier_3r *) CI_ADDR (ct, id);
  op -> x1 = x1;
  op -> y1 = y1;
  op -> x2 = x2;
  op -> y2 = y2;
  op -> wi = wi;
  op -> w1 = w1;
  op -> w2 = w2;
  op -> wo = wo;
}

void
vdl_add_quadratic_bezier (struct vdl_context *ct, xxxnum_t x1, xxxnum_t y1)
{
  int id = vdl_new_context_item (ct, vdl_cid_add_bezier_2n);
  struct ci_bezier_2n *op = (struct ci_bezier_2n *) CI_ADDR (ct, id);
  op -> x1 = x1;
  op -> y1 = y1;
}

void
vdl_add_quadratic_rational_bezier (struct vdl_context *ct, xxxnum_t x1, xxxnum_t y1,
				    xxxnum_t wi, xxxnum_t w1, xxxnum_t wo)
{
  int id = vdl_new_context_item (ct, vdl_cid_add_bezier_2r);
  struct ci_bezier_2r *op = (struct ci_bezier_2r *) CI_ADDR (ct, id);
  op -> x1 = x1;
  op -> y1 = y1;
  op -> wi = wi;
  op -> w1 = w1;
  op -> wo = wo;
}


void
vdl_start_path (struct vdl_context *ct)
{
  vdl_new_context_item (ct, vdl_cid_start_path);
}

void
vdl_close_path (struct vdl_context *ct)
{
  vdl_new_context_item (ct, vdl_cid_close_path);
}



