#ifndef _VDL_CONTEXT_H_
#define _VDL_CONTEXT_H_

struct vdl_context
{
  struct vdl_surface *surface;
  short int *command_buffer_cids;
  int *command_buffer_offsets;
  int command_buffer_used;
  int command_buffer_size;
  void *command_buffer_args;
  int command_buffer_args_used;
  int command_buffer_args_size;
  xxxnum_t *curve_approx_buffer;
  int curve_approx_buffer_size;
};

extern struct vdl_context * vdl_new_context (struct vdl_surface *sf);
extern void vdl_flush_context (struct vdl_context *ct);

extern void vdl_add_point (struct vdl_context *ct, xxxnum_t x, xxxnum_t y);
extern void vdl_add_line (struct vdl_context *ct);

extern void vdl_add_cubic_bezier (struct vdl_context *ct, xxxnum_t x1, xxxnum_t y1, xxxnum_t x2, xxxnum_t y2);
extern void vdl_add_cubic_rational_bezier (struct vdl_context *ct, xxxnum_t x1, xxxnum_t y1, xxxnum_t x2, xxxnum_t y2, xxxnum_t wi, xxxnum_t w1, xxxnum_t w2, xxxnum_t wo);
extern void vdl_add_quadratic_bezier (struct vdl_context *ct, xxxnum_t x1, xxxnum_t y1);
extern void vdl_add_quadratic_rational_bezier (struct vdl_context *ct, xxxnum_t x1, xxxnum_t y1, xxxnum_t wi, xxxnum_t w1, xxxnum_t wo);

extern void vdl_start_path (struct vdl_context *ct);
extern void vdl_close_path (struct vdl_context *ct);

#ifdef _VDL_INTERNAL_

#define CI_ADDR(CT, ID) ((CT) -> command_buffer_args + (CT) -> command_buffer_offsets [ID])
#define CI_TYPE(CT, ID) ((CT) -> command_buffer_cids [ID])

extern short int vdl_cid_add_point;
extern short int vdl_cid_add_line;
extern short int vdl_cid_add_bezier_2n;
extern short int vdl_cid_add_bezier_2r;
extern short int vdl_cid_add_bezier_3n;
extern short int vdl_cid_add_bezier_3r;
extern short int vdl_cid_start_path;
extern short int vdl_cid_close_path;

struct ci_point
{
  xxxnum_t x, y;
};

struct ci_bezier_2n
{
  xxxnum_t x1, y1;
};

struct ci_bezier_2r
{
  xxxnum_t x1, y1, wi, w1, wo;
};

struct ci_bezier_3n
{
  xxxnum_t x1, y1, x2, y2;
};

struct ci_bezier_3r
{
  xxxnum_t x1, y1, x2, y2, wi, w1, w2, wo;
};

extern void vdl_context_init (void);
extern void vdl_fill_bezier (struct vdl_context *ct, struct geom_bezier *b, int offset, double x1, double y1, double x2, double y2);
extern int vdl_register_context_item (const char *name, int size);
extern int vdl_new_context_item (struct vdl_context *ct, short int cid);

#endif
#endif
