/* Jednoduch program v X s pouitm pouze Xlib, bez toolkitu. Ilustruje
 * zkladn inicializaci X klienta, vytvoen okna, kreslen a event loop. */

/* Zkladn deklarace pro Xlib */
#include <X11/Xlib.h>
/* Dal deklarace pro Xlib */
#include <X11/Xutil.h>
/* Peddefinovan atomy */
#include <X11/Xatom.h>
/* Msto include pro asto pouvan systmov hlavikov soubory */
#include <X11/Xos.h>

/* Dal hlaviky */
#include <stdlib.h>
#include <stdio.h>
#include "vdl_internal.h"

struct vdl_surface *sf;
struct vdl_context *ct;

char *progname; /* Jmno programu z argv[0] */

char *display_name = NULL; /* Jmeno displeje (X serveru), NULL => vezme se
                              z environmentove promenne DISPLAY */
Display *display; /* Struktura popisujc pouvan displej */
int screen_num; /* slo obrazovky, kterou budeme pouvat */
Screen *screen_ptr; /* Struktura popisujc obrazovku */
unsigned int display_width, display_height; /* Rozmry obrazovky */

Window win; /* Top-level okno */
Pixmap icon_pixmap; /* Pixmapa pro ikonu okna */
unsigned int border_width = 4; /* ka okraje okna */
int x, y; /* Pozice okna na obrazovce */
unsigned int width, height; /* Rozmry okna */

void
draw_graphics (int x, int y, int w, int h)
{
  int lns = 31;
  int skip = 15;
  double lnm = (3.14159 * 2.0 / (lns));
  double eps = 0.2;

  int i;



    vdl_start_path (ct);
  vdl_add_point (ct, 10, 10);
  vdl_add_line (ct);
  vdl_add_point (ct, 70, 20);
    vdl_add_line (ct);
    vdl_add_point (ct, 100, 50);
    vdl_add_quadratic_bezier (ct, 160, 40);
  vdl_add_point (ct, 105, 100);
  vdl_add_line (ct);
 vdl_add_point (ct, 80, 100);
  vdl_add_line (ct);
  vdl_add_point (ct, 70, 150);
  vdl_add_line (ct);
  vdl_add_point (ct, 20, 120);
    vdl_close_path (ct);

    /*
  vdl_start_path (ct);
  for (i = 0; i < lns; i ++)
    {
      vdl_add_point (ct, 200 + sin (i * (3.14159 * 2.0 / lns)) * 80, 200 + cos (i * (3.14159 * 2.0 / lns)) * 80);

      vdl_add_line (ct);
      vdl_add_point (ct, 200 + sin (i * (3.14159 * 2.0 / lns)) * 180, 200 + cos (i * (3.14159 * 2.0 / lns)) * 180);
      vdl_add_line (ct);
}
  vdl_close_path (ct);

*/
  vdl_start_path (ct);
  for (i = 0; i < lns; i ++)
    {
      int j = (i * skip) % lns;
      vdl_add_point (ct, 200 + sin (j * lnm + eps) * 180, 200 + cos (j * lnm + eps) * 180);

      vdl_add_line (ct);
    }
  vdl_close_path (ct);


    vdl_start_path (ct);
  vdl_add_point (ct, 200, 200);
    vdl_add_quadratic_bezier (ct, 275, 100);
  vdl_add_point (ct, 350, 200);
    vdl_add_quadratic_bezier (ct, 275, 150);
      vdl_close_path (ct);


  vdl_xrender_fill_hack (ct, 0xFFFF, 0, 0, 0x7FFF);
  //  vdl_xrender_stroke_hack (ct, 5, 0, 0x0000, 0xFFFF, 0x7FFF);
  vdl_flush_context (ct);

}

int main(int argc, char *argv[])
{
    XEvent event; /* Pro zpracovn pijatch udlost */
    static Bool window_too_small = False;

    /* Inicializace */
    progname = argv[0];

    /* Pipojen k X serveru */
    if(!(display = XOpenDisplay(display_name))) {
        fprintf(stderr, "%s: cannot connect to X server %s\n",
                progname, XDisplayName(display_name));
        exit(-1);
    }
    printf("Connected to X server %s\n", XDisplayName(display_name));
    screen_num = DefaultScreen(display);
    printf("Default screen number is %d\n", screen_num);
    screen_ptr = DefaultScreenOfDisplay(display);

    display_width = DisplayWidth(display, screen_num);
    display_height = DisplayHeight(display, screen_num);
    printf("Display is %u x %u pixels\n", display_width, display_height);

    /* Vytvoen okna */
    x = y = 0; /* Okno umst window manager */
    width = display_width/2;
    height = display_height/2;
    win = XCreateSimpleWindow(display, RootWindow(display, screen_num),
                              x, y, width, height, border_width,
                              BlackPixel(display, screen_num),
                              WhitePixel(display, screen_num));
    printf("Created window with ID %#lx\n", win);

    /* Vbr typ udlost ke zpracovn */
    XSelectInput(display, win, ExposureMask | KeyPressMask |
                 ButtonPressMask | StructureNotifyMask);

    /* Namapovat okno na obrazovku (zobrazit) */
    XMapWindow(display, win);

    vdl_init ();
    sf = vdl_new_surface_xrender (display, win);
    ct = vdl_new_context (sf);

    while(1) {
        XNextEvent(display, &event);
        switch(event.type) {
            case Expose:
                printf("Drawing (serial=%lu send=%d count=%d x=%d y=%d "
                       "w=%d h=%d)\n",
                       event.xexpose.serial,
                       event.xexpose.send_event,
                       event.xexpose.count,
                       event.xexpose.x,
                       event.xexpose.y,
                       event.xexpose.width,
                       event.xexpose.height);
                if(event.xexpose.count == 0) {
                    if(window_too_small) { /* Pli zmenen okno */
                        printf("Drawing small\n");
                    } else { /* Nakreslit obsah okna */
                        printf("Drawing normal (w=%d h=%d)\n", width, height);
                        draw_graphics(event.xexpose.x, event.xexpose.y, event.xexpose.width, event.xexpose.height);
                    }
                }
                break;
            case ConfigureNotify:
		if(width != event.xconfigure.width || height != event.xconfigure.height)
		    XClearWindow(display, win);
                width = event.xconfigure.width;
                height = event.xconfigure.height;
                printf("ConfigureNotify (serial=%lu send=%d w=%d h=%d)\n",
                       event.xconfigure.serial,
                       event.xconfigure.send_event,
                       event.xconfigure.width,
                       event.xconfigure.height);
                break;
            case ButtonPress:
            case KeyPress:
                XCloseDisplay(display);
                return 0;
                break;
            default:
                break;
        }
    }

    return 0;
}
