#include "vdl_internal.h"

#include <stdio.h>
#include <stdlib.h>

struct vdl_surface_xrender
{
  struct vdl_surface parent;
  XRenderPictFormat *pf;
  Display *dpy;
  Drawable drw;
  Picture pict;
};

static XTriangle * triangle_buffer = NULL;
static int triangle_buffer_used = 0;
static int triangle_buffer_size = 0;


static double *approximation_buffer = NULL;
static int approximation_buffer_size = 0;

static  Picture px ;

static void
add_triangle (xxxnum_t x1, xxxnum_t y1, xxxnum_t x2, xxxnum_t y2, xxxnum_t x3, xxxnum_t y3)
{
  if (triangle_buffer_used == triangle_buffer_size)
    {
      if (triangle_buffer_size == 0)
	triangle_buffer_size = 16;
      else triangle_buffer_size *= 2;
      triangle_buffer = (XTriangle *)
	xrealloc (triangle_buffer, triangle_buffer_size * sizeof (XTriangle));
    }
  //  printf ("QQQ %f, %f, %f, %f, %f, %f\n", x1, y1, x2, y2, x3, y3);
  triangle_buffer [triangle_buffer_used] . p1 . x = x1 * 0xFFFF;
  triangle_buffer [triangle_buffer_used] . p2 . x = x2 * 0xFFFF;
  triangle_buffer [triangle_buffer_used] . p3 . x = x3 * 0xFFFF;
  triangle_buffer [triangle_buffer_used] . p1 . y = y1 * 0xFFFF;
  triangle_buffer [triangle_buffer_used] . p2 . y = y2 * 0xFFFF;
  triangle_buffer [triangle_buffer_used] . p3 . y = y3 * 0xFFFF;
  triangle_buffer_used ++;
}

void
vdl_xrender_init_surface (struct vdl_surface_xrender *sf, Display *dpy, Drawable drw)
{
  XWindowAttributes attr;
  XGetWindowAttributes (dpy, drw, &attr);

  sf -> dpy = dpy;
  sf -> drw = drw;
  sf -> pf = XRenderFindVisualFormat (dpy, attr . visual);
  sf -> pict = XRenderCreatePicture (dpy, drw, sf -> pf, 0, NULL);

  {
  Pixmap p = XCreatePixmap (dpy, drw, 1, 1, 32);
  XRenderPictFormat *pf2 ;
  XRenderPictureAttributes at;

  at . repeat = 1;
  pf2 = XRenderFindStandardFormat (dpy, PictStandardARGB32);
  px = XRenderCreatePicture (dpy, p, pf2, CPRepeat, &at);

  }


}

struct vdl_surface *
vdl_new_surface_xrender (Display *dpy, Drawable drw)
{
  MALLOCATED (struct vdl_surface_xrender, sf);
  vdl_init_surface ((struct vdl_surface *) sf);
  vdl_xrender_init_surface (sf, dpy, drw);
  return ((struct vdl_surface *) sf);
}

static void set_color (struct vdl_surface_xrender *sf, int r, int g, int b, int a)
{
  XRenderColor c;
  c . red = r;
  c . green = g;
  c . blue = b;
  c . alpha = a;

  XRenderComposite (sf -> dpy, PictOpClear,   px, 0, px, 0, 0, 0, 0, 0, 0, 1, 1);
  XRenderFillRectangle (sf -> dpy, PictOpSrc, px, &c, 0, 0, 1, 1);

}

static struct geom_point segment_mps [7];
static struct geom_point segment_l1ps [4];
static struct geom_point segment_r1ps [4];
static struct geom_point segment_l2ps [4];
static struct geom_point segment_r2ps [4];
static struct geom_vector segment_lvs [4];
static struct geom_vector segment_rvs [4];
static xxxnum_t segment_wls [7];
static xxxnum_t segment_wrs [7];
static int segment_state = 0;
static int segment_offset = 0;

#define SEGMENT(WHAT, NUM) ((segment_ ## WHAT ## s) [((NUM) + segment_offset) % 4])

static inline void
start_segment (void)
{
  segment_state = 0;
}

static void
add_segment (xxxnum_t x, xxxnum_t y, xxxnum_t wl, xxxnum_t wr)
{
  SEGMENT (mp, 0) . x = x;
  SEGMENT (mp, 0) . y = y;
  SEGMENT (wl, 0) = wl;
  SEGMENT (wr, 0) = wr;

  printf ("%f, %f\n", x, y);
  segment_offset = (segment_offset + 1) % 4;

  if ((segment_state > 0) && (isfinite (SEGMENT (mp, 2) . x)) && (isfinite (SEGMENT (mp, 3) . x)))
    {
      struct geom_vector tmp, gv;
      struct geom_point *p1 = &(SEGMENT (mp, 2));
      struct geom_point *p2 = &(SEGMENT (mp, 3));

      geom_vector_init_two_points (p1, p2, &tmp);
      geom_vector_left_normal (&tmp, &gv);
      geom_vector_point_in_direction (p1, &gv, SEGMENT (wl, 2), &(SEGMENT (l1p, 3)));
      geom_vector_point_in_direction (p2, &gv, SEGMENT (wl, 3), &(SEGMENT (l2p, 3)));
      geom_vector_point_in_direction (p1, &gv, - SEGMENT (wr, 2), &(SEGMENT (r1p, 3)));
      geom_vector_point_in_direction (p2, &gv, - SEGMENT (wr, 3), &(SEGMENT (r2p, 3)));
      geom_vector_init_two_points (&(SEGMENT (l1p, 3)), &(SEGMENT (l2p, 3)), &(SEGMENT (lv, 3)));
      geom_vector_init_two_points (&(SEGMENT (r1p, 3)), &(SEGMENT (r2p, 3)), &(SEGMENT (rv, 3)));
    }

  if (segment_state < 3)
    {
      segment_mps [4 + segment_state] = SEGMENT (mp, 3);
      segment_wls [4 + segment_state] = wl;
      segment_wrs [4 + segment_state] = wr;
      segment_state ++;
    }
  else if (isfinite (SEGMENT (mp, 1) . x) && isfinite (SEGMENT (mp, 2) . x))
    {
      struct geom_point f1, f2, f3, f4;

      if (isfinite (SEGMENT (mp, 0) . x))
	{
	  int e = 0;
	  e = e || geom_vector_intersection (&(SEGMENT (l1p, 1)), &(SEGMENT (lv, 1)),
					     &(SEGMENT (l1p, 2)), &(SEGMENT (lv, 2)), &f1);
	  e = e || geom_vector_intersection (&(SEGMENT (r1p, 1)), &(SEGMENT (rv, 1)),
					     &(SEGMENT (r1p, 2)), &(SEGMENT (rv, 2)), &f2);
	  if (e)
	    {

	      f1 = SEGMENT (l1p, 2);
	      f2 = SEGMENT (r1p, 2);
	    }
	}
      else
	{
	  printf ("No intersecion \n");
	  f1 = SEGMENT (l1p, 2);
	  f2 = SEGMENT (r1p, 2);
	}

      if (isfinite (SEGMENT (mp, 3) . x))
	{
	  int e = 0;
	  e = e || geom_vector_intersection (&(SEGMENT (l1p, 3)), &(SEGMENT (lv, 3)),
					     &(SEGMENT (l1p, 2)), &(SEGMENT (lv, 2)), &f3);
	  e = e || geom_vector_intersection (&(SEGMENT (r1p, 3)), &(SEGMENT (rv, 3)),
					     &(SEGMENT (r1p, 2)), &(SEGMENT (rv, 2)), &f4);
	  if (e)
	    {
	      printf ("No intersecion \n");
	      f3 = SEGMENT (l2p, 2);
	      f4 = SEGMENT (r2p, 2);
	    }
	}
      else
	{
	  f3 = SEGMENT (l2p, 2);
	  f4 = SEGMENT (r2p, 2);
	}

      add_triangle (f1 . x, f1 . y, f2 . x, f2 . y, f3 . x, f3 . y);
      add_triangle (f2 . x, f2 . y, f3 . x, f3 . y, f4 . x, f4 . y);
    }
}

static inline void
break_segment (void)
{
  printf ("BREAK\n");
  add_segment (NAN, 0, 0, 0);
}

static void
close_segment (void)
{
  add_segment (segment_mps [4] . x, segment_mps [4] . y, segment_wls [4], segment_wrs [4]);
  add_segment (segment_mps [5] . x, segment_mps [5] . y, segment_wls [5], segment_wrs [5]);
  add_segment (segment_mps [6] . x, segment_mps [6] . y, segment_wls [6], segment_wrs [6]);
  segment_state = 0;
}

static inline void
end_segment (void)
{
  break_segment ();
  close_segment ();
}

static void
xrender_stroke_hack_add_mid_points (struct vdl_context *ct, xxxnum_t ox, xxxnum_t oy,
				    xxxnum_t ax, xxxnum_t ay, int stored_action,
				    xxxnum_t owl, xxxnum_t owr, xxxnum_t awl, xxxnum_t awr)
{
  double delta = 0.25;

  if ((stored_action != -1) &&
      (CI_TYPE (ct, stored_action) != vdl_cid_add_line))
    {
      struct geom_bezier b;
      int j, dp;
      xxxnum_t dwl = awl - owl;
      xxxnum_t dwr = awr - owr;

      vdl_fill_bezier (ct, &b, stored_action, ox, oy, ax, ay);

      geom_bezier_approximate (&b, delta, &approximation_buffer,
			       &approximation_buffer_size, &dp);

      dp = (dp - 1) * 2;
      if (dp > 0)
	    {
	      dwl /= dp;
	      dwr /= dp;
	    }

      for (j = 2; j < dp; j += 2)
	add_segment (approximation_buffer [j],
		     approximation_buffer [j + 1],
		     owl + dwl * j,
		     owr + dwr * j);
    }
}

void
vdl_xrender_stroke_hack (struct vdl_context *ct, xxxnum_t width, int r, int g, int b, int a)
{
  struct vdl_surface_xrender *su = (struct vdl_surface_xrender *) (ct -> surface);

  int i;
  xxxnum_t ox, oy, sx, sy;
  xxxnum_t owl, owr, swl, swr;
  xxxnum_t awl = 4;
  xxxnum_t awr = 4;


  int state = -1;
  int stored_action = -1;

  set_color (su, r, g, b, a);

    for (i = 0; i < ct -> command_buffer_used; i ++)
    {
      int type = CI_TYPE (ct, i);
      if (type == vdl_cid_add_point)
	{
	  struct ci_point *p = CI_ADDR (ct, i);

	  if (state == -1) BUG();
	  if (state == 1) /* continue path */
	    xrender_stroke_hack_add_mid_points (ct, ox, oy, p -> x, p -> y, stored_action,
						owl, owr, awl, awr);

	  add_segment (p -> x, p-> y, awl, awr);

	  if (state == 0)
	    {
	      sx = p -> x;
	      sy = p -> y;
	      swl = awl;
	      swr = awr;
	    }

	  ox = p -> x;
	  oy = p -> y;
	  owl = awl;
	  owr = awr;
	  stored_action = -1;
	  state = 1;
	}
      else if ((type == vdl_cid_add_line) ||
	       (type == vdl_cid_add_bezier_2n) ||
	       (type == vdl_cid_add_bezier_2r) ||
	       (type == vdl_cid_add_bezier_3n) ||
	       (type == vdl_cid_add_bezier_3r))
	{
	  stored_action = i;
	}
      else if (type == vdl_cid_start_path)
	{
	  if (state == 1) end_segment ();

	  start_segment ();
	  state = 0;
	}
      else if (type == vdl_cid_close_path)
	{
	  if (state == 1) /* continue path */
	    xrender_stroke_hack_add_mid_points (ct, ox, oy, sx, sy, stored_action,
						owl, owr, swl, swr);
	  close_segment ();
	  state = -1;
	}
    }

  if (state != -1) end_segment ();

  //  XRenderComposite (su -> dpy, PictOpClear, su -> pict, 0, su -> pict,
  //	    0, 0, 0, 0, 0, 0, 400, 400);
  XRenderCompositeTriangles (su -> dpy, PictOpOver, px, su -> pict,
			     XRenderFindStandardFormat (su ->dpy, PictStandardA8),
			     0, 0, triangle_buffer, triangle_buffer_used);

  triangle_buffer_used = 0;
}

void
vdl_xrender_fill_hack (struct vdl_context *ct, int r, int g, int b, int a)
{
  struct vdl_surface_xrender *su = (struct vdl_surface_xrender *) (ct -> surface);

  double delta = 0.25;
  xxxnum_t sx, sy, ox, oy, ax, ay;
  int state, last_pgn, stored_action;
  int i;

  struct vdl_polygon *pg = vdl_polygon_new (0);
  struct vdl_traps *tr = vdl_traps_new (0);

  set_color (su, r, g, b, a);

  state = -1;
  for (i = 0; i < ct -> command_buffer_used; i ++)
    {
      int type = CI_TYPE (ct, i);
      if ((type == vdl_cid_add_point) ||
	  (type == vdl_cid_close_path))
	{
	  if (type == vdl_cid_add_point)
	    {
	      struct ci_point *p = CI_ADDR (ct, i);
	      ax = p -> x;
	      ay = p -> y;
	    }
	  else
	    {
	      ax = sx;
	      ay = sy;
	      state = 2;
	    }

	  switch (state)
	    {
	    case 0:
	      sx = ox = ax;
	      sy = oy = ay;
	      last_pgn = pg -> edges_used;
	      state = 1;
	      break;
	    case 1:
	    case 2:
	      if ((stored_action == -1) ||
		  (CI_TYPE (ct, stored_action) == vdl_cid_add_line))
		{
		  vdl_polygon_add_edge (pg, ox, oy, ax, ay);
		}
	      else
		{
		  struct geom_bezier b;
		  int j, dp;

		  vdl_fill_bezier (ct, &b, stored_action, ox, oy, ax, ay);

		  geom_bezier_approximate (&b, delta, &approximation_buffer,
					   &approximation_buffer_size, &dp);

		  dp = (dp - 1) * 2;

		  for (j = 0; j < dp; j += 2)
		    vdl_polygon_add_edge (pg, approximation_buffer [j],
					   approximation_buffer [j + 1],
					   approximation_buffer [j + 2],
					   approximation_buffer [j + 3]);
		}

	      ox = ax;
	      oy = ay;

	      break;
	    default:
	      BUG ();
	    }
	}
      else if ((type == vdl_cid_add_line) ||
	       (type == vdl_cid_add_bezier_2n) ||
	       (type == vdl_cid_add_bezier_2r) ||
	       (type == vdl_cid_add_bezier_3n) ||
	       (type == vdl_cid_add_bezier_3r))
	{
	  stored_action = i;
	}
      else if (type == vdl_cid_start_path)
	{
	  if (state == 1) pg -> edges_used = last_pgn;
	  state = 0;
	}
    }
  if (state == 1) pg -> edges_used = last_pgn;

  vdl_polygon_tesselate_to_trapezoids (pg, tr, VDL_FILL_RULE_EVEN_ODD);


  XRenderComposite (su -> dpy, PictOpClear, su -> pict, 0, su -> pict,
		    0, 0, 0, 0, 0, 0, 400, 400);
  XRenderCompositeTrapezoids (su -> dpy, PictOpOver, px, su -> pict,
			      XRenderFindStandardFormat (su ->dpy, PictStandardA8),
			      0, 0, tr-> traps, tr -> traps_used);

}
